/* Generated by Cython 0.29.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-fopenmp"
        ],
        "extra_link_args": [
            "-fopenmp"
        ],
        "language": "c++",
        "name": "cyext_acv",
        "sources": [
            "acv_explainers/cyext_acv/cyext_acv.pyx"
        ]
    },
    "module_name": "cyext_acv"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cyext_acv
#define __PYX_HAVE_API__cyext_acv
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include <utility>

    #if __cplusplus > 199711L
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <set>
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include <omp.h>
#include <algorithm>
#include "limits.h"
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "acv_explainers/cyext_acv/cyext_acv.pyx",
  "__init__.pxd",
  "stringsource",
  "type.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "cyext_acv.pyx":7
 * import numpy as np
 * cimport numpy as np
 * ctypedef np.float64_t double             # <<<<<<<<<<<<<<
 * cimport cython
 * from scipy.special import comb
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_9cyext_acv_double;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset;
struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */
struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset {
  PyObject_HEAD
  PyObject *__pyx_v_s;
};


/* "cyext_acv.pyx":457
 * def powerset(iterable):
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *__pyx_outer_scope;
  Py_ssize_t __pyx_v_r;
  Py_ssize_t __pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* None.proto */
static void __Pyx_RaiseUnboundMemoryviewSliceNogil(const char *varname);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* None.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_long(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_long(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *, int writable_flag);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_int(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(PyObject *, int writable_flag);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* None.proto */
static CYTHON_INLINE long __Pyx_pow_long(long, long);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_long(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_long(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_long__const__(const char *itemp);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_long(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_long(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsdsds_double(PyObject *, int writable_flag);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libcpp.utility' */

/* Module declarations from 'libcpp.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'openmp' */

/* Module declarations from 'cyext_acv' */
static PyTypeObject *__pyx_ptype_9cyext_acv___pyx_scope_struct__powerset = 0;
static PyTypeObject *__pyx_ptype_9cyext_acv___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_9cyext_acv_compute_exp(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_normalized(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_cat(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_clf_cat(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_clf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_reg(__Pyx_memviewslice, __Pyx_memviewslice, double, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_reg_cat(__Pyx_memviewslice, __Pyx_memviewslice, double, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg_cat(__Pyx_memviewslice, __Pyx_memviewslice, double, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg(__Pyx_memviewslice, __Pyx_memviewslice, double, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_leaves_pa(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, long const , std::vector<std::vector<std::vector<long> > >  const , std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_leaves_normalized(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, long const , std::vector<std::vector<std::vector<long> > >  const , std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_acv_leaves(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, std::vector<std::vector<std::vector<long> > > , __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_swing(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, unsigned long, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double const , int); /*proto*/
static PyObject *__pyx_f_9cyext_acv_swing_sv_clf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_swing_diff(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, long, __Pyx_memviewslice, unsigned long, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, double const , int); /*proto*/
static PyObject *__pyx_f_9cyext_acv_swing_sv_clf_direct(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg_cpp_pa_coal(__Pyx_memviewslice, __Pyx_memviewslice, double, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf_ptrees(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_acv_leaves_adap(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int &, std::vector<std::vector<std::vector<long> > > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_single_shap_values_acv_leaves(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, std::vector<std::vector<std::vector<long> > > , __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_single_shap_values(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, std::vector<std::vector<std::vector<long> > > , std::vector<std::vector<long> >  const , int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf_approx(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, PyObject *, double, int, PyObject *, int, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_binomialC(unsigned long, unsigned long, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv__comb_int_long(unsigned long, unsigned long, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rf_v0(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf_v0(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<int> , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_rf_v0(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, PyObject *, double, int, int, PyObject *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_exp_rf(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_quantile_rf(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_quantile_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_quantile_diff_rf(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_quantile_diff_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_cdf_rf(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_cdf_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_sufficient_expl_rf_v0(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, PyObject *, double, int, int, PyObject *, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule_v0(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_v0(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__Pyx_memviewslice &, double &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compile_rules(PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule_biased(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_biased(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rule_v0_weights(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_opti(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_biased_v2(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rule(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule(__Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<std::vector<int> >  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rule_weights(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_verbose(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, double &, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_9cyext_acv_single_compute_sdp_rf(__Pyx_memviewslice &, double &, __Pyx_memviewslice &, __Pyx_memviewslice &, std::vector<int>  &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, __Pyx_memviewslice &, int &, int &, int &, double &); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<std::vector<int> > , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf_same_set(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, std::vector<int> , __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, PyObject *, double, int, int, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_9cyext_acv_sufficient_expl_rf(__Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, int, int, int &, double &, PyObject *, double, int, int, PyObject *, int __pyx_skip_dispatch); /*proto*/
static std::vector<long>  __pyx_convert_vector_from_py_long(PyObject *); /*proto*/
static std::vector<std::vector<long> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(PyObject *); /*proto*/
static std::vector<std::vector<std::vector<long> > >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(PyObject *); /*proto*/
static std::vector<int>  __pyx_convert_vector_from_py_int(PyObject *); /*proto*/
static std::vector<std::vector<int> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_int(const std::vector<int>  &); /*proto*/
static std::vector<double>  __pyx_convert_vector_from_py_double(PyObject *); /*proto*/
static std::vector<std::vector<double> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_long(const std::vector<long>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___(const std::vector<std::vector<long> >  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(const std::vector<std::vector<std::vector<long> > >  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_double(const std::vector<double>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(const std::vector<std::vector<double> >  &); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_long = { "long", NULL, sizeof(long), { 0 }, 0, IS_UNSIGNED(long) ? 'U' : 'I', IS_UNSIGNED(long), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double__const__ = { "const double", NULL, sizeof(double const ), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_long__const__ = { "const long", NULL, sizeof(long const ), { 0 }, 0, IS_UNSIGNED(long const ) ? 'U' : 'I', IS_UNSIGNED(long const ), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_int = { "int", NULL, sizeof(int), { 0 }, 0, IS_UNSIGNED(int) ? 'U' : 'I', IS_UNSIGNED(int), 0 };
#define __Pyx_MODULE_NAME "cyext_acv"
extern int __pyx_module_is_main_cyext_acv;
int __pyx_module_is_main_cyext_acv = 0;

/* Implementation of 'cyext_acv' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_C[] = "C";
static const char __pyx_k_N[] = "N";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_X[] = "X";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_fX[] = "fX";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_pi[] = "pi";
static const char __pyx_k_tX[] = "tX";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_y_X[] = "y_X";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_axis[] = "axis";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_comb[] = "comb";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_long[] = "long";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_ones[] = "ones";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_prod[] = "prod";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tqdm[] = "tqdm";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_chain[] = "chain";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_N_star[] = "N_star";
static const char __pyx_k_S_star[] = "S_star";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_y_data[] = "y_data";
static const char __pyx_k_y_pred[] = "y_pred";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_minimal[] = "minimal";
static const char __pyx_k_scaling[] = "scaling";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_iterable[] = "iterable";
static const char __pyx_k_pi_level[] = "pi_level";
static const char __pyx_k_powerset[] = "powerset";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_quantile[] = "quantile";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_cyext_acv[] = "cyext_acv";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_leaves_nb[] = "leaves_nb";
static const char __pyx_k_max_depth[] = "max_depth";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_rule_data[] = "rule_data";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_classifier[] = "classifier";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_thresholds[] = "thresholds";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_expand_dims[] = "expand_dims";
static const char __pyx_k_num_threads[] = "num_threads";
static const char __pyx_k_combinations[] = "combinations";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_search_space[] = "search_space";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_children_left[] = "children_left";
static const char __pyx_k_from_iterable[] = "from_iterable";
static const char __pyx_k_min_node_size[] = "min_node_size";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_scipy_special[] = "scipy.special";
static const char __pyx_k_children_right[] = "children_right";
static const char __pyx_k_leaf_idx_trees[] = "leaf_idx_trees";
static const char __pyx_k_node_idx_trees[] = "node_idx_trees";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_weighted_percentile[] = "weighted_percentile";
static const char __pyx_k_acv_explainers_utils[] = "acv_explainers.utils";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_partition_leaves_trees[] = "partition_leaves_trees";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_powerset_locals_genexpr[] = "powerset.<locals>.genexpr";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Check_condition_in_the_computati[] = "Check condition in the computation of SDP";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_acv_explainers_cyext_acv_cyext_a[] = "acv_explainers/cyext_acv/cyext_acv.pyx";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_kp_s_Check_condition_in_the_computati;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_s_N;
static PyObject *__pyx_n_s_N_star;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_n_s_S_star;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_n_s_X;
static PyObject *__pyx_kp_s_acv_explainers_cyext_acv_cyext_a;
static PyObject *__pyx_n_s_acv_explainers_utils;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_axis;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_chain;
static PyObject *__pyx_n_s_children_left;
static PyObject *__pyx_n_s_children_right;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_classifier;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_comb;
static PyObject *__pyx_n_s_combinations;
static PyObject *__pyx_n_s_concatenate;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_cyext_acv;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_double;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_expand_dims;
static PyObject *__pyx_n_s_fX;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_from_iterable;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_iterable;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_leaf_idx_trees;
static PyObject *__pyx_n_s_leaves_nb;
static PyObject *__pyx_n_s_long;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_depth;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_min_node_size;
static PyObject *__pyx_n_s_minimal;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_node_idx_trees;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_num_threads;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_partition_leaves_trees;
static PyObject *__pyx_n_s_pi;
static PyObject *__pyx_n_s_pi_level;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_powerset;
static PyObject *__pyx_n_s_powerset_locals_genexpr;
static PyObject *__pyx_n_s_prod;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_quantile;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_rule_data;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_scaling;
static PyObject *__pyx_n_s_scipy_special;
static PyObject *__pyx_n_s_search_space;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_tX;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_thresholds;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tqdm;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_weighted_percentile;
static PyObject *__pyx_n_s_y_X;
static PyObject *__pyx_n_s_y_data;
static PyObject *__pyx_n_s_y_pred;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_9cyext_acv_compute_exp(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_2compute_exp_normalized(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_4compute_exp_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_6compute_sdp_clf_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_8compute_sdp_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_8powerset_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9cyext_acv_10powerset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_iterable); /* proto */
static PyObject *__pyx_pf_9cyext_acv_12compute_sdp_reg(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_14compute_sdp_reg_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_16global_sdp_reg_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_18global_sdp_reg(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_20shap_values_leaves_pa(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_22shap_values_leaves_normalized(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_24shap_values_acv_leaves(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_26swing_sv_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_thresholds, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_28swing_sv_clf_direct(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_thresholds, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_30global_sdp_reg_cpp_pa_coal(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_num_threads, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_32global_sdp_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_34global_sdp_clf_ptrees(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_36shap_values_acv_leaves_adap(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, __Pyx_memviewslice __pyx_v_size, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_38single_shap_values_acv_leaves(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_40single_shap_values(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads); /* proto */
static PyObject *__pyx_pf_9cyext_acv_42global_sdp_clf_approx(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, PyObject *__pyx_v_search_space, int __pyx_v_stop); /* proto */
static PyObject *__pyx_pf_9cyext_acv_44binomialC(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_N, unsigned long __pyx_v_k); /* proto */
static PyObject *__pyx_pf_9cyext_acv_46_comb_int_long(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_N, unsigned long __pyx_v_k); /* proto */
static PyObject *__pyx_pf_9cyext_acv_48compute_sdp_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_50global_sdp_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space); /* proto */
static PyObject *__pyx_pf_9cyext_acv_52compute_exp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_54compute_quantile_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_quantile); /* proto */
static PyObject *__pyx_pf_9cyext_acv_56compute_quantile_diff_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_quantile); /* proto */
static PyObject *__pyx_pf_9cyext_acv_58compute_cdf_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_60sufficient_expl_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space); /* proto */
static PyObject *__pyx_pf_9cyext_acv_62compute_sdp_rule_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_64compute_sdp_maxrule_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_66compile_rules(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rule_data, PyObject *__pyx_v_data, PyObject *__pyx_v_S, PyObject *__pyx_v_min_node_size); /* proto */
static PyObject *__pyx_pf_9cyext_acv_68compute_sdp_rule_biased(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_70compute_sdp_maxrule_biased(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_72compute_sdp_maxrule_opti(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_74compute_sdp_maxrule_biased_v2(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_76compute_sdp_rule(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_78compute_sdp_maxrule(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_80compute_sdp_maxrule_verbose(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi); /* proto */
static PyObject *__pyx_pf_9cyext_acv_82compute_sdp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_84compute_sdp_rf_same_set(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t); /* proto */
static PyObject *__pyx_pf_9cyext_acv_86global_sdp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space); /* proto */
static PyObject *__pyx_pf_9cyext_acv_88sufficient_expl_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_9cyext_acv___pyx_scope_struct__powerset(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9cyext_acv___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_copy = {0, &__pyx_n_s_copy, 0, 0, 0};
static PyObject *__pyx_float_1e_10;
static PyObject *__pyx_float_neg_1e_10;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__34;
/* Late includes */

/* "cyext_acv.pyx":32
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

static PyObject *__pyx_pw_9cyext_acv_1compute_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_it;
  unsigned int __pyx_v_it_s;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_lm;
  unsigned int __pyx_v_p_ss;
  unsigned int __pyx_v_o_all;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  unsigned int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  int __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  double __pyx_t_33;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp", 0);

  /* "cyext_acv.pyx":36
 *         int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":37
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":38
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":39
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":40
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":41
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":47
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":50
 * 
 *     cdef double[:, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((max_leaves, N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, nb_leaf, lm, p_ss, o_all
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":55
 *     cdef int b, leaf_numb, i, s, j
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  __pyx_t_9 = __pyx_v_n_trees;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_b = __pyx_t_11;

    /* "cyext_acv.pyx":56
 * 
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0
 */
    __pyx_t_12 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_12 * __pyx_v_leaves_nb.strides[0]) )));

    /* "cyext_acv.pyx":57
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             lm = 0
 *             p_ss = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_13 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_15 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_i) lastprivate(__pyx_v_it) lastprivate(__pyx_v_it_s) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_ss) lastprivate(__pyx_v_s) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                          {
                              __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_14);
                              /* Initialize private variables to invalid values */
                              __pyx_v_i = ((int)0xbad0bad0);
                              __pyx_v_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_it_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_j = ((int)0xbad0bad0);
                              __pyx_v_lm = ((unsigned int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_ss = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":58
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0             # <<<<<<<<<<<<<<
 *             p_ss = 0
 *             for i in range(data.shape[0]):
 */
                              __pyx_v_lm = 0;

                              /* "cyext_acv.pyx":59
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0
 *             p_ss = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 it = 0
 */
                              __pyx_v_p_ss = 0;

                              /* "cyext_acv.pyx":60
 *             lm = 0
 *             p_ss = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 it = 0
 *                 it_s = 0
 */
                              __pyx_t_16 = (__pyx_v_data.shape[0]);
                              __pyx_t_17 = __pyx_t_16;
                              for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
                                __pyx_v_i = __pyx_t_18;

                                /* "cyext_acv.pyx":61
 *             p_ss = 0
 *             for i in range(data.shape[0]):
 *                 it = 0             # <<<<<<<<<<<<<<
 *                 it_s = 0
 *                 for s in range(m):
 */
                                __pyx_v_it = 0;

                                /* "cyext_acv.pyx":62
 *             for i in range(data.shape[0]):
 *                 it = 0
 *                 it_s = 0             # <<<<<<<<<<<<<<
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                __pyx_v_it_s = 0;

                                /* "cyext_acv.pyx":63
 *                 it = 0
 *                 it_s = 0
 *                 for s in range(m):             # <<<<<<<<<<<<<<
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1
 */
                                __pyx_t_19 = __pyx_v_m;
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                                  __pyx_v_s = __pyx_t_21;

                                  /* "cyext_acv.pyx":64
 *                 it_s = 0
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                         it = it + 1
 *                 for s in range(S_size):
 */
                                  __pyx_t_12 = __pyx_v_i;
                                  __pyx_t_23 = __pyx_v_s;
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_26 = __pyx_v_s;
                                  __pyx_t_27 = 1;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_12 * __pyx_v_data.strides[0]) ) + __pyx_t_23 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_28) {
                                  } else {
                                    __pyx_t_22 = __pyx_t_28;
                                    goto __pyx_L19_bool_binop_done;
                                  }
                                  __pyx_t_27 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_s;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_24 = __pyx_v_leaf_numb;
                                  __pyx_t_23 = __pyx_v_s;
                                  __pyx_t_12 = 0;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_22 = __pyx_t_28;
                                  __pyx_L19_bool_binop_done:;
                                  if (__pyx_t_22) {

                                    /* "cyext_acv.pyx":65
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                    __pyx_v_it = (__pyx_v_it + 1);

                                    /* "cyext_acv.pyx":64
 *                 it_s = 0
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                         it = it + 1
 *                 for s in range(S_size):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":66
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 */
                                __pyx_t_19 = __pyx_v_S_size;
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                                  __pyx_v_s = __pyx_t_21;

                                  /* "cyext_acv.pyx":67
 *                         it = it + 1
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         it_s = it_s + 1
 *                 if it == m:
 */
                                  __pyx_t_12 = __pyx_v_s;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_30 = 1;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_23 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_30 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_28) {
                                  } else {
                                    __pyx_t_22 = __pyx_t_28;
                                    goto __pyx_L24_bool_binop_done;
                                  }
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_30 = __pyx_v_i;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_23 = 0;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_30 * __pyx_v_data.strides[0]) ) + __pyx_t_29 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_22 = __pyx_t_28;
                                  __pyx_L24_bool_binop_done:;
                                  if (__pyx_t_22) {

                                    /* "cyext_acv.pyx":68
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1             # <<<<<<<<<<<<<<
 *                 if it == m:
 *                     lm = lm + 1
 */
                                    __pyx_v_it_s = (__pyx_v_it_s + 1);

                                    /* "cyext_acv.pyx":67
 *                         it = it + 1
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         it_s = it_s + 1
 *                 if it == m:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":69
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 *                 if it == m:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 *                 if it_s == S_size:
 */
                                __pyx_t_22 = ((__pyx_v_it == __pyx_v_m) != 0);
                                if (__pyx_t_22) {

                                  /* "cyext_acv.pyx":70
 *                         it_s = it_s + 1
 *                 if it == m:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 *                 if it_s == S_size:
 *                     p_ss = p_ss + 1
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1);

                                  /* "cyext_acv.pyx":69
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 *                 if it == m:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 *                 if it_s == S_size:
 */
                                }

                                /* "cyext_acv.pyx":71
 *                 if it == m:
 *                     lm = lm + 1
 *                 if it_s == S_size:             # <<<<<<<<<<<<<<
 *                     p_ss = p_ss + 1
 * 
 */
                                __pyx_t_22 = ((__pyx_v_it_s == __pyx_v_S_size) != 0);
                                if (__pyx_t_22) {

                                  /* "cyext_acv.pyx":72
 *                     lm = lm + 1
 *                 if it_s == S_size:
 *                     p_ss = p_ss + 1             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                  __pyx_v_p_ss = (__pyx_v_p_ss + 1);

                                  /* "cyext_acv.pyx":71
 *                 if it == m:
 *                     lm = lm + 1
 *                 if it_s == S_size:             # <<<<<<<<<<<<<<
 *                     p_ss = p_ss + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":74
 *                     p_ss = p_ss + 1
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                              __pyx_t_19 = __pyx_v_N;
                              __pyx_t_20 = __pyx_t_19;
                              for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_20; __pyx_t_18+=1) {
                                __pyx_v_i = __pyx_t_18;

                                /* "cyext_acv.pyx":75
 * 
 *             for i in range(N):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":76
 *             for i in range(N):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_31 = __pyx_v_S_size;
                                __pyx_t_32 = __pyx_t_31;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_32; __pyx_t_21+=1) {
                                  __pyx_v_s = __pyx_t_21;

                                  /* "cyext_acv.pyx":77
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_12 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_30 = 1;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_23 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_30 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_28) {
                                  } else {
                                    __pyx_t_22 = __pyx_t_28;
                                    goto __pyx_L33_bool_binop_done;
                                  }
                                  __pyx_t_12 = __pyx_v_s;
                                  __pyx_t_30 = __pyx_v_i;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_23 = 0;
                                  __pyx_t_28 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_30 * __pyx_v_X.strides[0]) ) + __pyx_t_29 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_22 = __pyx_t_28;
                                  __pyx_L33_bool_binop_done:;
                                  if (__pyx_t_22) {

                                    /* "cyext_acv.pyx":78
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":77
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":79
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for j in range(d):
 */
                                __pyx_t_22 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_22) {

                                  /* "cyext_acv.pyx":80
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 for j in range(d):
 *                     mean_forest_b[leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0
 */
                                  goto __pyx_L28_continue;

                                  /* "cyext_acv.pyx":79
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 for j in range(d):
 */
                                }

                                /* "cyext_acv.pyx":81
 *                 if o_all > 0:
 *                     continue
 *                 for j in range(d):             # <<<<<<<<<<<<<<
 *                     mean_forest_b[leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0
 * 
 */
                                __pyx_t_31 = __pyx_v_d;
                                __pyx_t_32 = __pyx_t_31;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_32; __pyx_t_21+=1) {
                                  __pyx_v_j = __pyx_t_21;

                                  /* "cyext_acv.pyx":82
 *                     continue
 *                 for j in range(d):
 *                     mean_forest_b[leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(N):
 */
                                  if (((__pyx_v_p_ss != 0) != 0)) {
                                    __pyx_t_25 = __pyx_v_b;
                                    __pyx_t_23 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = __pyx_v_b;
                                    __pyx_t_27 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_23 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                    __pyx_t_26 = __pyx_v_j;
                                    __pyx_t_33 = ((__pyx_v_lm * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_24 * __pyx_v_values.strides[0]) ) + __pyx_t_27 * __pyx_v_values.strides[1]) ) + __pyx_t_26 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_ss);
                                  } else {
                                    __pyx_t_33 = 0.0;
                                  }
                                  __pyx_t_23 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_j;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_23 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[2]) )) += __pyx_t_33;
                                }
                                __pyx_L28_continue:;
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":57
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             lm = 0
 *             p_ss = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":84
 *                     mean_forest_b[leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):
 */
  __pyx_t_15 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_15;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_14; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "cyext_acv.pyx":85
 * 
 *     for i in range(N):
 *         for j in range(d):             # <<<<<<<<<<<<<<
 *             for leaf_numb in range(max_leaves):
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 */
    __pyx_t_13 = __pyx_v_d;
    __pyx_t_19 = __pyx_t_13;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_19; __pyx_t_10+=1) {
      __pyx_v_j = __pyx_t_10;

      /* "cyext_acv.pyx":86
 *     for i in range(N):
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 * 
 */
      __pyx_t_20 = __pyx_v_max_leaves;
      __pyx_t_31 = __pyx_t_20;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_31; __pyx_t_11+=1) {
        __pyx_v_leaf_numb = __pyx_t_11;

        /* "cyext_acv.pyx":87
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(mean_forest)
 */
        __pyx_t_26 = __pyx_v_leaf_numb;
        __pyx_t_25 = __pyx_v_i;
        __pyx_t_23 = __pyx_v_j;
        __pyx_t_27 = __pyx_v_i;
        __pyx_t_24 = __pyx_v_j;
        *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_27 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[1]) )) += (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_26 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_23 * __pyx_v_mean_forest_b.strides[2]) )));
      }
    }
  }

  /* "cyext_acv.pyx":89
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 * 
 *     return np.asarray(mean_forest)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_mean_forest, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":32
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("cyext_acv.compute_exp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_1compute_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_1compute_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_exp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 1); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 2); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 3); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 4); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 5); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 6); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 7); __PYX_ERR(0, 32, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, 8); __PYX_ERR(0, 32, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_exp") < 0)) __PYX_ERR(0, 32, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 32, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 32, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 32, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 32, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 33, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 33, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 33, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 33, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_exp", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 32, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_exp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_compute_exp(__pyx_self, __pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_compute_exp(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 32, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 32, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_exp(__pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_exp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":96
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_normalized(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

static PyObject *__pyx_pw_9cyext_acv_3compute_exp_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_normalized(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_it;
  unsigned int __pyx_v_it_s;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_lm;
  unsigned int __pyx_v_p_ss;
  unsigned int __pyx_v_o_all;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_z;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_t_20;
  unsigned int __pyx_t_21;
  unsigned int __pyx_t_22;
  int __pyx_t_23;
  int __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  int __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  double __pyx_t_35;
  long __pyx_t_36;
  long __pyx_t_37;
  int __pyx_t_38;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_normalized", 0);

  /* "cyext_acv.pyx":100
 *         int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":101
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":102
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":103
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":104
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":105
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":111
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":114
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, max_leaves, N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, nb_leaf, lm, p_ss, o_all
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":119
 *     cdef int b, leaf_numb, i, s, j
 *     cdef double[:, :, :] p_n
 *     p_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double z
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":122
 *     cdef double z
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  __pyx_t_11 = __pyx_v_n_trees;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_b = __pyx_t_13;

    /* "cyext_acv.pyx":123
 * 
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0
 */
    __pyx_t_14 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_14 * __pyx_v_leaves_nb.strides[0]) )));

    /* "cyext_acv.pyx":124
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             lm = 0
 *             p_ss = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_15 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_17 = (__pyx_t_15 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_17 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_14, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_i) lastprivate(__pyx_v_it) lastprivate(__pyx_v_it_s) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_ss) lastprivate(__pyx_v_s) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_17; __pyx_t_16++){
                          {
                              __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_16);
                              /* Initialize private variables to invalid values */
                              __pyx_v_i = ((int)0xbad0bad0);
                              __pyx_v_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_it_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_j = ((int)0xbad0bad0);
                              __pyx_v_lm = ((unsigned int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_ss = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":125
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0             # <<<<<<<<<<<<<<
 *             p_ss = 0
 *             for i in range(data.shape[0]):
 */
                              __pyx_v_lm = 0;

                              /* "cyext_acv.pyx":126
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             lm = 0
 *             p_ss = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 it = 0
 */
                              __pyx_v_p_ss = 0;

                              /* "cyext_acv.pyx":127
 *             lm = 0
 *             p_ss = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 it = 0
 *                 it_s = 0
 */
                              __pyx_t_18 = (__pyx_v_data.shape[0]);
                              __pyx_t_19 = __pyx_t_18;
                              for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
                                __pyx_v_i = __pyx_t_20;

                                /* "cyext_acv.pyx":128
 *             p_ss = 0
 *             for i in range(data.shape[0]):
 *                 it = 0             # <<<<<<<<<<<<<<
 *                 it_s = 0
 *                 for s in range(m):
 */
                                __pyx_v_it = 0;

                                /* "cyext_acv.pyx":129
 *             for i in range(data.shape[0]):
 *                 it = 0
 *                 it_s = 0             # <<<<<<<<<<<<<<
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                __pyx_v_it_s = 0;

                                /* "cyext_acv.pyx":130
 *                 it = 0
 *                 it_s = 0
 *                 for s in range(m):             # <<<<<<<<<<<<<<
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1
 */
                                __pyx_t_21 = __pyx_v_m;
                                __pyx_t_22 = __pyx_t_21;
                                for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
                                  __pyx_v_s = __pyx_t_23;

                                  /* "cyext_acv.pyx":131
 *                 it_s = 0
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                         it = it + 1
 *                 for s in range(S_size):
 */
                                  __pyx_t_14 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_28 = __pyx_v_s;
                                  __pyx_t_29 = 1;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_14 * __pyx_v_data.strides[0]) ) + __pyx_t_25 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_30) {
                                  } else {
                                    __pyx_t_24 = __pyx_t_30;
                                    goto __pyx_L19_bool_binop_done;
                                  }
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_28 = __pyx_v_s;
                                  __pyx_t_27 = __pyx_v_b;
                                  __pyx_t_26 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_14 = 0;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_29 * __pyx_v_data.strides[0]) ) + __pyx_t_28 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_14 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_24 = __pyx_t_30;
                                  __pyx_L19_bool_binop_done:;
                                  if (__pyx_t_24) {

                                    /* "cyext_acv.pyx":132
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                    __pyx_v_it = (__pyx_v_it + 1);

                                    /* "cyext_acv.pyx":131
 *                 it_s = 0
 *                 for s in range(m):
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                         it = it + 1
 *                 for s in range(S_size):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":133
 *                     if((data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                         it = it + 1
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 */
                                __pyx_t_21 = __pyx_v_S_size;
                                __pyx_t_22 = __pyx_t_21;
                                for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
                                  __pyx_v_s = __pyx_t_23;

                                  /* "cyext_acv.pyx":134
 *                         it = it + 1
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         it_s = it_s + 1
 *                 if it == m:
 */
                                  __pyx_t_14 = __pyx_v_s;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_b;
                                  __pyx_t_29 = __pyx_v_leaf_numb;
                                  __pyx_t_31 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_32 = 1;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_31 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_32 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_30) {
                                  } else {
                                    __pyx_t_24 = __pyx_t_30;
                                    goto __pyx_L24_bool_binop_done;
                                  }
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_32 = __pyx_v_i;
                                  __pyx_t_31 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_29 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_b;
                                  __pyx_t_14 = __pyx_v_leaf_numb;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_29 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = 0;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_32 * __pyx_v_data.strides[0]) ) + __pyx_t_31 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_14 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_24 = __pyx_t_30;
                                  __pyx_L24_bool_binop_done:;
                                  if (__pyx_t_24) {

                                    /* "cyext_acv.pyx":135
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1             # <<<<<<<<<<<<<<
 *                 if it == m:
 *                     lm = lm + 1
 */
                                    __pyx_v_it_s = (__pyx_v_it_s + 1);

                                    /* "cyext_acv.pyx":134
 *                         it = it + 1
 *                 for s in range(S_size):
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         it_s = it_s + 1
 *                 if it == m:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":136
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 *                 if it == m:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 *                 if it_s == S_size:
 */
                                __pyx_t_24 = ((__pyx_v_it == __pyx_v_m) != 0);
                                if (__pyx_t_24) {

                                  /* "cyext_acv.pyx":137
 *                         it_s = it_s + 1
 *                 if it == m:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 *                 if it_s == S_size:
 *                     p_ss = p_ss + 1
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1);

                                  /* "cyext_acv.pyx":136
 *                     if((data[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         it_s = it_s + 1
 *                 if it == m:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 *                 if it_s == S_size:
 */
                                }

                                /* "cyext_acv.pyx":138
 *                 if it == m:
 *                     lm = lm + 1
 *                 if it_s == S_size:             # <<<<<<<<<<<<<<
 *                     p_ss = p_ss + 1
 * 
 */
                                __pyx_t_24 = ((__pyx_v_it_s == __pyx_v_S_size) != 0);
                                if (__pyx_t_24) {

                                  /* "cyext_acv.pyx":139
 *                     lm = lm + 1
 *                 if it_s == S_size:
 *                     p_ss = p_ss + 1             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                  __pyx_v_p_ss = (__pyx_v_p_ss + 1);

                                  /* "cyext_acv.pyx":138
 *                 if it == m:
 *                     lm = lm + 1
 *                 if it_s == S_size:             # <<<<<<<<<<<<<<
 *                     p_ss = p_ss + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":141
 *                     p_ss = p_ss + 1
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                              __pyx_t_21 = __pyx_v_N;
                              __pyx_t_22 = __pyx_t_21;
                              for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_22; __pyx_t_20+=1) {
                                __pyx_v_i = __pyx_t_20;

                                /* "cyext_acv.pyx":142
 * 
 *             for i in range(N):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":143
 *             for i in range(N):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_33 = __pyx_v_S_size;
                                __pyx_t_34 = __pyx_t_33;
                                for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_34; __pyx_t_23+=1) {
                                  __pyx_v_s = __pyx_t_23;

                                  /* "cyext_acv.pyx":144
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_29 = __pyx_v_s;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_29 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_14 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_31 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_32 = 1;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_25 * __pyx_v_X.strides[0]) ) + __pyx_t_26 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_31 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_32 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_30) {
                                  } else {
                                    __pyx_t_24 = __pyx_t_30;
                                    goto __pyx_L33_bool_binop_done;
                                  }
                                  __pyx_t_14 = __pyx_v_s;
                                  __pyx_t_32 = __pyx_v_i;
                                  __pyx_t_31 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_b;
                                  __pyx_t_29 = __pyx_v_leaf_numb;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = 0;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_32 * __pyx_v_X.strides[0]) ) + __pyx_t_31 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_24 = __pyx_t_30;
                                  __pyx_L33_bool_binop_done:;
                                  if (__pyx_t_24) {

                                    /* "cyext_acv.pyx":145
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":144
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":146
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_24 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_24) {

                                  /* "cyext_acv.pyx":147
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 p_n[b, leaf_numb, i] += (1.*lm)/p_ss
 */
                                  goto __pyx_L28_continue;

                                  /* "cyext_acv.pyx":146
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":149
 *                     continue
 * 
 *                 p_n[b, leaf_numb, i] += (1.*lm)/p_ss             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(d):
 */
                                __pyx_t_27 = __pyx_v_b;
                                __pyx_t_25 = __pyx_v_leaf_numb;
                                __pyx_t_26 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_27 * __pyx_v_p_n.strides[0]) ) + __pyx_t_25 * __pyx_v_p_n.strides[1]) ) + __pyx_t_26 * __pyx_v_p_n.strides[2]) )) += ((1. * __pyx_v_lm) / __pyx_v_p_ss);

                                /* "cyext_acv.pyx":151
 *                 p_n[b, leaf_numb, i] += (1.*lm)/p_ss
 * 
 *                 for j in range(d):             # <<<<<<<<<<<<<<
 *                     mean_forest_b[b, leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0
 * 
 */
                                __pyx_t_33 = __pyx_v_d;
                                __pyx_t_34 = __pyx_t_33;
                                for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_34; __pyx_t_23+=1) {
                                  __pyx_v_j = __pyx_t_23;

                                  /* "cyext_acv.pyx":152
 * 
 *                 for j in range(d):
 *                     mean_forest_b[b, leaf_numb, i, j] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j]) / p_ss if p_ss != 0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  if (((__pyx_v_p_ss != 0) != 0)) {
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_27 = __pyx_v_b;
                                    __pyx_t_29 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_26 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                    __pyx_t_28 = __pyx_v_j;
                                    __pyx_t_35 = ((__pyx_v_lm * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_27 * __pyx_v_values.strides[0]) ) + __pyx_t_29 * __pyx_v_values.strides[1]) ) + __pyx_t_28 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_ss);
                                  } else {
                                    __pyx_t_35 = 0.0;
                                  }
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_26 = __pyx_v_leaf_numb;
                                  __pyx_t_28 = __pyx_v_i;
                                  __pyx_t_29 = __pyx_v_j;
                                  *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_28 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_29 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_35;
                                }
                                __pyx_L28_continue:;
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":124
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             lm = 0
 *             p_ss = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":155
 * 
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         for b in range(n_trees):
 *             z = 0
 */
  __pyx_t_17 = __pyx_v_N;
  __pyx_t_16 = __pyx_t_17;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_16; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "cyext_acv.pyx":156
 * 
 *     for i in range(N):
 *         for b in range(n_trees):             # <<<<<<<<<<<<<<
 *             z = 0
 *             for leaf_numb in range(leaves_nb[b]):
 */
    __pyx_t_12 = __pyx_v_n_trees;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_13; __pyx_t_20+=1) {
      __pyx_v_b = __pyx_t_20;

      /* "cyext_acv.pyx":157
 *     for i in range(N):
 *         for b in range(n_trees):
 *             z = 0             # <<<<<<<<<<<<<<
 *             for leaf_numb in range(leaves_nb[b]):
 *                 z += p_n[b, leaf_numb, i]
 */
      __pyx_v_z = 0.0;

      /* "cyext_acv.pyx":158
 *         for b in range(n_trees):
 *             z = 0
 *             for leaf_numb in range(leaves_nb[b]):             # <<<<<<<<<<<<<<
 *                 z += p_n[b, leaf_numb, i]
 *             for leaf_numb in range(leaves_nb[b]):
 */
      __pyx_t_29 = __pyx_v_b;
      __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_29 * __pyx_v_leaves_nb.strides[0]) )));
      __pyx_t_37 = __pyx_t_36;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_37; __pyx_t_23+=1) {
        __pyx_v_leaf_numb = __pyx_t_23;

        /* "cyext_acv.pyx":159
 *             z = 0
 *             for leaf_numb in range(leaves_nb[b]):
 *                 z += p_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *             for leaf_numb in range(leaves_nb[b]):
 *                 for j in range(d):
 */
        __pyx_t_29 = __pyx_v_b;
        __pyx_t_28 = __pyx_v_leaf_numb;
        __pyx_t_26 = __pyx_v_i;
        __pyx_v_z = (__pyx_v_z + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_29 * __pyx_v_p_n.strides[0]) ) + __pyx_t_28 * __pyx_v_p_n.strides[1]) ) + __pyx_t_26 * __pyx_v_p_n.strides[2]) ))));
      }

      /* "cyext_acv.pyx":160
 *             for leaf_numb in range(leaves_nb[b]):
 *                 z += p_n[b, leaf_numb, i]
 *             for leaf_numb in range(leaves_nb[b]):             # <<<<<<<<<<<<<<
 *                 for j in range(d):
 *                     mean_forest[i, j] += mean_forest_b[b, leaf_numb, i, j]/z if z !=0 else 0
 */
      __pyx_t_26 = __pyx_v_b;
      __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_26 * __pyx_v_leaves_nb.strides[0]) )));
      __pyx_t_37 = __pyx_t_36;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_37; __pyx_t_23+=1) {
        __pyx_v_leaf_numb = __pyx_t_23;

        /* "cyext_acv.pyx":161
 *                 z += p_n[b, leaf_numb, i]
 *             for leaf_numb in range(leaves_nb[b]):
 *                 for j in range(d):             # <<<<<<<<<<<<<<
 *                     mean_forest[i, j] += mean_forest_b[b, leaf_numb, i, j]/z if z !=0 else 0
 * 
 */
        __pyx_t_15 = __pyx_v_d;
        __pyx_t_21 = __pyx_t_15;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_21; __pyx_t_38+=1) {
          __pyx_v_j = __pyx_t_38;

          /* "cyext_acv.pyx":162
 *             for leaf_numb in range(leaves_nb[b]):
 *                 for j in range(d):
 *                     mean_forest[i, j] += mean_forest_b[b, leaf_numb, i, j]/z if z !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
          if (((__pyx_v_z != 0.0) != 0)) {
            __pyx_t_26 = __pyx_v_b;
            __pyx_t_28 = __pyx_v_leaf_numb;
            __pyx_t_29 = __pyx_v_i;
            __pyx_t_25 = __pyx_v_j;
            __pyx_t_35 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_26 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_28 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_29 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_z);
          } else {
            __pyx_t_35 = 0.0;
          }
          __pyx_t_25 = __pyx_v_i;
          __pyx_t_29 = __pyx_v_j;
          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_25 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_29 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_35;
        }
      }
    }
  }

  /* "cyext_acv.pyx":165
 * 
 * 
 *     return np.asarray(mean_forest)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_mean_forest, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":96
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_normalized(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_3compute_exp_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_3compute_exp_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_exp_normalized (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 1); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 2); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 3); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 4); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 5); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 6); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 7); __PYX_ERR(0, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, 8); __PYX_ERR(0, 96, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_exp_normalized") < 0)) __PYX_ERR(0, 96, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 96, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 96, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 96, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 96, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 97, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 97, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 97, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 97, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_exp_normalized", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 96, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_exp_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_2compute_exp_normalized(__pyx_self, __pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_2compute_exp_normalized(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_normalized", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 96, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 96, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_exp_normalized(__pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":172
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_cat(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees,  long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

static PyObject *__pyx_pw_9cyext_acv_5compute_exp_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_cat(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_it;
  unsigned int __pyx_v_it_s;
  unsigned int __pyx_v_p_ss;
  unsigned int __pyx_v_o_all;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_lm;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j1;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  int __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  int __pyx_t_33;
  double __pyx_t_34;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_cat", 0);

  /* "cyext_acv.pyx":176
 *         int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":177
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":178
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":179
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":180
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":181
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":185
 *     cdef double[:, :] leaf_part
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":188
 * 
 *     cdef double[:, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((max_leaves, N, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, p_ss, o_all, nb_leaf, lm
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":193
 *     cdef int b, leaf_numb, i, s, j1, j
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  __pyx_t_9 = __pyx_v_n_trees;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_b = __pyx_t_11;

    /* "cyext_acv.pyx":194
 * 
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 */
    __pyx_t_12 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_12 * __pyx_v_leaves_nb.strides[0]) )));

    /* "cyext_acv.pyx":195
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_13 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_15 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_i) lastprivate(__pyx_v_it) lastprivate(__pyx_v_it_s) lastprivate(__pyx_v_j) lastprivate(__pyx_v_j1) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_ss) lastprivate(__pyx_v_s) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                          {
                              __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_14);
                              /* Initialize private variables to invalid values */
                              __pyx_v_i = ((int)0xbad0bad0);
                              __pyx_v_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_it_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_j = ((int)0xbad0bad0);
                              __pyx_v_j1 = ((int)0xbad0bad0);
                              __pyx_v_lm = ((unsigned int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_ss = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":196
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                              __pyx_t_16 = __pyx_v_N;
                              __pyx_t_17 = __pyx_t_16;
                              for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
                                __pyx_v_i = __pyx_t_18;

                                /* "cyext_acv.pyx":197
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":198
 *             for i in range(N):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_19 = __pyx_v_S_size;
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                                  __pyx_v_s = __pyx_t_21;

                                  /* "cyext_acv.pyx":199
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_12 = __pyx_v_s;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_29 = 1;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_23 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_30) {
                                  } else {
                                    __pyx_t_22 = __pyx_t_30;
                                    goto __pyx_L19_bool_binop_done;
                                  }
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_23 = 0;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_29 * __pyx_v_X.strides[0]) ) + __pyx_t_28 * __pyx_v_X.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_22 = __pyx_t_30;
                                  __pyx_L19_bool_binop_done:;
                                  if (__pyx_t_22) {

                                    /* "cyext_acv.pyx":200
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":199
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":201
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_22 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_22) {

                                  /* "cyext_acv.pyx":202
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 lm = 0
 */
                                  goto __pyx_L14_continue;

                                  /* "cyext_acv.pyx":201
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":204
 *                     continue
 * 
 *                 lm = 0             # <<<<<<<<<<<<<<
 *                 p_ss = 0
 *                 for j in range(data.shape[0]):
 */
                                __pyx_v_lm = 0;

                                /* "cyext_acv.pyx":205
 * 
 *                 lm = 0
 *                 p_ss = 0             # <<<<<<<<<<<<<<
 *                 for j in range(data.shape[0]):
 * 
 */
                                __pyx_v_p_ss = 0;

                                /* "cyext_acv.pyx":206
 *                 lm = 0
 *                 p_ss = 0
 *                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 * 
 *                     it = 0
 */
                                __pyx_t_31 = (__pyx_v_data.shape[0]);
                                __pyx_t_32 = __pyx_t_31;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_32; __pyx_t_21+=1) {
                                  __pyx_v_j = __pyx_t_21;

                                  /* "cyext_acv.pyx":208
 *                 for j in range(data.shape[0]):
 * 
 *                     it = 0             # <<<<<<<<<<<<<<
 *                     it_s = 0
 *                     for s in range(m):
 */
                                  __pyx_v_it = 0;

                                  /* "cyext_acv.pyx":209
 * 
 *                     it = 0
 *                     it_s = 0             # <<<<<<<<<<<<<<
 *                     for s in range(m):
 *                         if((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] > partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                  __pyx_v_it_s = 0;

                                  /* "cyext_acv.pyx":210
 *                     it = 0
 *                     it_s = 0
 *                     for s in range(m):             # <<<<<<<<<<<<<<
 *                         if((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] > partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             it = it +  1
 */
                                  __pyx_t_19 = __pyx_v_m;
                                  __pyx_t_20 = __pyx_t_19;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_20; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":211
 *                     it_s = 0
 *                     for s in range(m):
 *                         if((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] > partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             it = it +  1
 * 
 */
                                    __pyx_t_27 = __pyx_v_j;
                                    __pyx_t_23 = __pyx_v_s;
                                    __pyx_t_24 = __pyx_v_b;
                                    __pyx_t_12 = __pyx_v_leaf_numb;
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_25 = 1;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_23 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_30) {
                                    } else {
                                      __pyx_t_22 = __pyx_t_30;
                                      goto __pyx_L27_bool_binop_done;
                                    }
                                    __pyx_t_25 = __pyx_v_j;
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_12 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_23 = __pyx_v_s;
                                    __pyx_t_27 = 0;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_22 = __pyx_t_30;
                                    __pyx_L27_bool_binop_done:;
                                    if (__pyx_t_22) {

                                      /* "cyext_acv.pyx":212
 *                     for s in range(m):
 *                         if((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] > partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             it = it +  1             # <<<<<<<<<<<<<<
 * 
 *                     for s in range(S_size):
 */
                                      __pyx_v_it = (__pyx_v_it + 1);

                                      /* "cyext_acv.pyx":211
 *                     it_s = 0
 *                     for s in range(m):
 *                         if((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] > partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             it = it +  1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":214
 *                             it = it +  1
 * 
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if(data[j, S[s]] == X[i, S[s]]):
 *                             it_s = it_s +  1
 */
                                  __pyx_t_19 = __pyx_v_S_size;
                                  __pyx_t_20 = __pyx_t_19;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_20; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":215
 * 
 *                     for s in range(S_size):
 *                         if(data[j, S[s]] == X[i, S[s]]):             # <<<<<<<<<<<<<<
 *                             it_s = it_s +  1
 * 
 */
                                    __pyx_t_27 = __pyx_v_s;
                                    __pyx_t_23 = __pyx_v_j;
                                    __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_12 = __pyx_v_s;
                                    __pyx_t_26 = __pyx_v_i;
                                    __pyx_t_25 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_22 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_23 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) == (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_26 * __pyx_v_X.strides[0]) ) + __pyx_t_25 * __pyx_v_X.strides[1]) )))) != 0);
                                    if (__pyx_t_22) {

                                      /* "cyext_acv.pyx":216
 *                     for s in range(S_size):
 *                         if(data[j, S[s]] == X[i, S[s]]):
 *                             it_s = it_s +  1             # <<<<<<<<<<<<<<
 * 
 *                     if it_s == S_size:
 */
                                      __pyx_v_it_s = (__pyx_v_it_s + 1);

                                      /* "cyext_acv.pyx":215
 * 
 *                     for s in range(S_size):
 *                         if(data[j, S[s]] == X[i, S[s]]):             # <<<<<<<<<<<<<<
 *                             it_s = it_s +  1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":218
 *                             it_s = it_s +  1
 * 
 *                     if it_s == S_size:             # <<<<<<<<<<<<<<
 *                         p_ss =  p_ss + 1
 *                         if it == m:
 */
                                  __pyx_t_22 = ((__pyx_v_it_s == __pyx_v_S_size) != 0);
                                  if (__pyx_t_22) {

                                    /* "cyext_acv.pyx":219
 * 
 *                     if it_s == S_size:
 *                         p_ss =  p_ss + 1             # <<<<<<<<<<<<<<
 *                         if it == m:
 *                             lm = lm + 1
 */
                                    __pyx_v_p_ss = (__pyx_v_p_ss + 1);

                                    /* "cyext_acv.pyx":220
 *                     if it_s == S_size:
 *                         p_ss =  p_ss + 1
 *                         if it == m:             # <<<<<<<<<<<<<<
 *                             lm = lm + 1
 * 
 */
                                    __pyx_t_22 = ((__pyx_v_it == __pyx_v_m) != 0);
                                    if (__pyx_t_22) {

                                      /* "cyext_acv.pyx":221
 *                         p_ss =  p_ss + 1
 *                         if it == m:
 *                             lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 *                 for j1 in range(d):
 */
                                      __pyx_v_lm = (__pyx_v_lm + 1);

                                      /* "cyext_acv.pyx":220
 *                     if it_s == S_size:
 *                         p_ss =  p_ss + 1
 *                         if it == m:             # <<<<<<<<<<<<<<
 *                             lm = lm + 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":218
 *                             it_s = it_s +  1
 * 
 *                     if it_s == S_size:             # <<<<<<<<<<<<<<
 *                         p_ss =  p_ss + 1
 *                         if it == m:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":223
 *                             lm = lm + 1
 * 
 *                 for j1 in range(d):             # <<<<<<<<<<<<<<
 *                     mean_forest_b[leaf_numb, i, j1] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j1]) / (p_ss) if p_ss != 0 else 0
 * 
 */
                                __pyx_t_19 = __pyx_v_d;
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                                  __pyx_v_j1 = __pyx_t_21;

                                  /* "cyext_acv.pyx":224
 * 
 *                 for j1 in range(d):
 *                     mean_forest_b[leaf_numb, i, j1] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j1]) / (p_ss) if p_ss != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(N):
 */
                                  if (((__pyx_v_p_ss != 0) != 0)) {
                                    __pyx_t_12 = __pyx_v_b;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_27 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_12 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                    __pyx_t_24 = __pyx_v_j1;
                                    __pyx_t_34 = ((__pyx_v_lm * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_27 * __pyx_v_values.strides[1]) ) + __pyx_t_24 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_ss);
                                  } else {
                                    __pyx_t_34 = 0.0;
                                  }
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_12 = __pyx_v_i;
                                  __pyx_t_24 = __pyx_v_j1;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) )) += __pyx_t_34;
                                }
                                __pyx_L14_continue:;
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":195
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":226
 *                     mean_forest_b[leaf_numb, i, j1] += (lm * values[b, leaf_idx_trees[b, leaf_numb], j1]) / (p_ss) if p_ss != 0 else 0
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):
 */
  __pyx_t_15 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_15;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_14; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "cyext_acv.pyx":227
 * 
 *     for i in range(N):
 *         for j in range(d):             # <<<<<<<<<<<<<<
 *             for leaf_numb in range(max_leaves):
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 */
    __pyx_t_13 = __pyx_v_d;
    __pyx_t_16 = __pyx_t_13;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_16; __pyx_t_10+=1) {
      __pyx_v_j = __pyx_t_10;

      /* "cyext_acv.pyx":228
 *     for i in range(N):
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 * 
 */
      __pyx_t_17 = __pyx_v_max_leaves;
      __pyx_t_19 = __pyx_t_17;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_19; __pyx_t_11+=1) {
        __pyx_v_leaf_numb = __pyx_t_11;

        /* "cyext_acv.pyx":229
 *         for j in range(d):
 *             for leaf_numb in range(max_leaves):
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(mean_forest)
 */
        __pyx_t_24 = __pyx_v_leaf_numb;
        __pyx_t_12 = __pyx_v_i;
        __pyx_t_25 = __pyx_v_j;
        __pyx_t_27 = __pyx_v_i;
        __pyx_t_26 = __pyx_v_j;
        *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_27 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_26 * __pyx_v_mean_forest.strides[1]) )) += (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[2]) )));
      }
    }
  }

  /* "cyext_acv.pyx":231
 *                 mean_forest[i, j] += mean_forest_b[leaf_numb, i, j]
 * 
 *     return np.asarray(mean_forest)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_mean_forest, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":172
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_cat(double[:, :] X, long[:] S, double[:, :] data, double[:, :, :] values,             # <<<<<<<<<<<<<<
 *         double[:, :, :, :] partition_leaves_trees, long[:, :] leaf_idx_trees,  long[:] leaves_nb, double[:] scaling,
 *         int num_threads):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_5compute_exp_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_5compute_exp_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_exp_cat (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 1); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 2); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 3); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 4); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 5); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 6); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 7); __PYX_ERR(0, 172, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, 8); __PYX_ERR(0, 172, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_exp_cat") < 0)) __PYX_ERR(0, 172, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 172, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 172, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 172, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 172, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 173, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 173, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 173, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 173, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_exp_cat", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_exp_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_4compute_exp_cat(__pyx_self, __pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_4compute_exp_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_cat", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 172, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 172, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_exp_cat(__pyx_v_X, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":237
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_clf_cat(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_7compute_sdp_clf_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_clf_cat(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_down;
  unsigned int __pyx_v_up;
  double __pyx_v_ss;
  double __pyx_v_ss_a;
  double __pyx_v_ss_u;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  int __pyx_v_lm;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  int __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  int __pyx_t_32;
  double __pyx_t_33;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_clf_cat", 0);

  /* "cyext_acv.pyx":243
 *             int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":244
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":245
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]
 *     if S_size == m:
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":246
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     if S_size == m:
 *         return np.ones(shape=(N))
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":247
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 */
  __pyx_t_1 = ((__pyx_v_S_size == __pyx_v_m) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":248
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]
 *     if S_size == m:
 *         return np.ones(shape=(N))             # <<<<<<<<<<<<<<
 *     elif S_size == 0:
 *         return np.zeros(shape=(N))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_4) < 0) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":247
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves =partition_leaves_trees.shape[1]
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 */
  }

  /* "cyext_acv.pyx":249
 *     if S_size == m:
 *         return np.ones(shape=(N))
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(shape=(N))
 * 
 */
  __pyx_t_1 = ((__pyx_v_S_size == 0) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":250
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 *         return np.zeros(shape=(N))             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_trees = values.shape[0]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":249
 *     if S_size == m:
 *         return np.ones(shape=(N))
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(shape=(N))
 * 
 */
  }

  /* "cyext_acv.pyx":252
 *         return np.zeros(shape=(N))
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":258
 * 
 *     cdef double[:, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_int_3);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_forest_b = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":261
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sdp = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":268
 *     cdef int b, leaf_numb, i, s, j, lm
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  __pyx_t_9 = __pyx_v_n_trees;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_b = __pyx_t_11;

    /* "cyext_acv.pyx":269
 * 
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 */
    __pyx_t_12 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_12 * __pyx_v_leaves_nb.strides[0]) )));

    /* "cyext_acv.pyx":270
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_13 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_15 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_1, __pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_down) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p) lastprivate(__pyx_v_p_d) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_sd) lastprivate(__pyx_v_p_su) lastprivate(__pyx_v_p_u) lastprivate(__pyx_v_s) lastprivate(__pyx_v_up) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                          {
                              __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_14);
                              /* Initialize private variables to invalid values */
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_down = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((int)0xbad0bad0);
                              __pyx_v_j = ((int)0xbad0bad0);
                              __pyx_v_lm = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_d = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_sd = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_su = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_u = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((int)0xbad0bad0);
                              __pyx_v_up = ((unsigned int)0xbad0bad0);

                              /* "cyext_acv.pyx":271
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                              __pyx_t_16 = __pyx_v_N;
                              __pyx_t_17 = __pyx_t_16;
                              for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
                                __pyx_v_i = __pyx_t_18;

                                /* "cyext_acv.pyx":272
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":273
 *             for i in range(N):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_19 = __pyx_v_S_size;
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_20; __pyx_t_21+=1) {
                                  __pyx_v_s = __pyx_t_21;

                                  /* "cyext_acv.pyx":274
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_12 = __pyx_v_s;
                                  __pyx_t_22 = __pyx_v_i;
                                  __pyx_t_23 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_24 = __pyx_v_s;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_26 = __pyx_v_leaf_numb;
                                  __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_28 = 1;
                                  __pyx_t_29 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_22 * __pyx_v_X.strides[0]) ) + __pyx_t_23 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_29) {
                                  } else {
                                    __pyx_t_1 = __pyx_t_29;
                                    goto __pyx_L20_bool_binop_done;
                                  }
                                  __pyx_t_24 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_i;
                                  __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_26 = __pyx_v_s;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_23 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_22 = 0;
                                  __pyx_t_29 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_28 * __pyx_v_X.strides[0]) ) + __pyx_t_27 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_1 = __pyx_t_29;
                                  __pyx_L20_bool_binop_done:;
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":275
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":274
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":276
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_1 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_1) {

                                  /* "cyext_acv.pyx":277
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 p = 0
 */
                                  goto __pyx_L15_continue;

                                  /* "cyext_acv.pyx":276
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":279
 *                     continue
 * 
 *                 p = 0             # <<<<<<<<<<<<<<
 *                 p_u = 0
 *                 p_d = 0
 */
                                __pyx_v_p = 0;

                                /* "cyext_acv.pyx":280
 * 
 *                 p = 0
 *                 p_u = 0             # <<<<<<<<<<<<<<
 *                 p_d = 0
 *                 p_s = 0
 */
                                __pyx_v_p_u = 0;

                                /* "cyext_acv.pyx":281
 *                 p = 0
 *                 p_u = 0
 *                 p_d = 0             # <<<<<<<<<<<<<<
 *                 p_s = 0
 *                 p_su = 0
 */
                                __pyx_v_p_d = 0;

                                /* "cyext_acv.pyx":282
 *                 p_u = 0
 *                 p_d = 0
 *                 p_s = 0             # <<<<<<<<<<<<<<
 *                 p_su = 0
 *                 p_sd = 0
 */
                                __pyx_v_p_s = 0;

                                /* "cyext_acv.pyx":283
 *                 p_d = 0
 *                 p_s = 0
 *                 p_su = 0             # <<<<<<<<<<<<<<
 *                 p_sd = 0
 * 
 */
                                __pyx_v_p_su = 0;

                                /* "cyext_acv.pyx":284
 *                 p_s = 0
 *                 p_su = 0
 *                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(data.shape[0]):
 */
                                __pyx_v_p_sd = 0;

                                /* "cyext_acv.pyx":286
 *                 p_sd = 0
 * 
 *                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                     up = 0
 *                     down = 0
 */
                                __pyx_t_30 = (__pyx_v_data.shape[0]);
                                __pyx_t_31 = __pyx_t_30;
                                for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_31; __pyx_t_21+=1) {
                                  __pyx_v_j = __pyx_t_21;

                                  /* "cyext_acv.pyx":287
 * 
 *                 for j in range(data.shape[0]):
 *                     up = 0             # <<<<<<<<<<<<<<
 *                     down = 0
 *                     b_it = 0
 */
                                  __pyx_v_up = 0;

                                  /* "cyext_acv.pyx":288
 *                 for j in range(data.shape[0]):
 *                     up = 0
 *                     down = 0             # <<<<<<<<<<<<<<
 *                     b_it = 0
 *                     lm = 0
 */
                                  __pyx_v_down = 0;

                                  /* "cyext_acv.pyx":289
 *                     up = 0
 *                     down = 0
 *                     b_it = 0             # <<<<<<<<<<<<<<
 *                     lm = 0
 *                     a_it = 0
 */
                                  __pyx_v_b_it = 0;

                                  /* "cyext_acv.pyx":290
 *                     down = 0
 *                     b_it = 0
 *                     lm = 0             # <<<<<<<<<<<<<<
 *                     a_it = 0
 *                     for s in range(m):
 */
                                  __pyx_v_lm = 0;

                                  /* "cyext_acv.pyx":291
 *                     b_it = 0
 *                     lm = 0
 *                     a_it = 0             # <<<<<<<<<<<<<<
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                  __pyx_v_a_it = 0;

                                  /* "cyext_acv.pyx":292
 *                     lm = 0
 *                     a_it = 0
 *                     for s in range(m):             # <<<<<<<<<<<<<<
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 */
                                  __pyx_t_19 = __pyx_v_m;
                                  __pyx_t_20 = __pyx_t_19;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_20; __pyx_t_32+=1) {
                                    __pyx_v_s = __pyx_t_32;

                                    /* "cyext_acv.pyx":293
 *                     a_it = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     if a_it == m:
 */
                                    __pyx_t_26 = __pyx_v_j;
                                    __pyx_t_22 = __pyx_v_s;
                                    __pyx_t_23 = __pyx_v_b;
                                    __pyx_t_12 = __pyx_v_leaf_numb;
                                    __pyx_t_25 = __pyx_v_s;
                                    __pyx_t_24 = 1;
                                    __pyx_t_29 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_26 * __pyx_v_data.strides[0]) ) + __pyx_t_22 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_29) {
                                    } else {
                                      __pyx_t_1 = __pyx_t_29;
                                      goto __pyx_L28_bool_binop_done;
                                    }
                                    __pyx_t_24 = __pyx_v_j;
                                    __pyx_t_25 = __pyx_v_s;
                                    __pyx_t_12 = __pyx_v_b;
                                    __pyx_t_23 = __pyx_v_leaf_numb;
                                    __pyx_t_22 = __pyx_v_s;
                                    __pyx_t_26 = 0;
                                    __pyx_t_29 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_24 * __pyx_v_data.strides[0]) ) + __pyx_t_25 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_12 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_1 = __pyx_t_29;
                                    __pyx_L28_bool_binop_done:;
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":294
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                     if a_it == m:
 *                         lm = 1
 */
                                      __pyx_v_a_it = (__pyx_v_a_it + 1);

                                      /* "cyext_acv.pyx":293
 *                     a_it = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     if a_it == m:
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":295
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  __pyx_t_1 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":296
 *                             a_it = a_it + 1
 *                     if a_it == m:
 *                         lm = 1             # <<<<<<<<<<<<<<
 * 
 *                     for s in range(S_size):
 */
                                    __pyx_v_lm = 1;

                                    /* "cyext_acv.pyx":295
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":298
 *                         lm = 1
 * 
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if data[j, S[s]] == X[i, S[s]]:
 *                             b_it = b_it + 1
 */
                                  __pyx_t_19 = __pyx_v_S_size;
                                  __pyx_t_20 = __pyx_t_19;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_20; __pyx_t_32+=1) {
                                    __pyx_v_s = __pyx_t_32;

                                    /* "cyext_acv.pyx":299
 * 
 *                     for s in range(S_size):
 *                         if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_22 = __pyx_v_j;
                                    __pyx_t_23 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_12 = __pyx_v_s;
                                    __pyx_t_25 = __pyx_v_i;
                                    __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_12 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_1 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_22 * __pyx_v_data.strides[0]) ) + __pyx_t_23 * __pyx_v_data.strides[1]) ))) == (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_25 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) )))) != 0);
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":300
 *                     for s in range(S_size):
 *                         if data[j, S[s]] == X[i, S[s]]:
 *                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                     if b_it == S_size:
 */
                                      __pyx_v_b_it = (__pyx_v_b_it + 1);

                                      /* "cyext_acv.pyx":299
 * 
 *                     for s in range(S_size):
 *                         if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":302
 *                             b_it = b_it + 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         if fX[i] == y_pred[j]:
 *                             up = 1
 */
                                  __pyx_t_1 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":303
 * 
 *                     if b_it == S_size:
 *                         if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                             up = 1
 *                         else:
 */
                                    __pyx_t_12 = __pyx_v_i;
                                    __pyx_t_24 = __pyx_v_j;
                                    __pyx_t_1 = (((*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_12 * __pyx_v_fX.strides[0]) ))) == (*((long *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_24 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":304
 *                     if b_it == S_size:
 *                         if fX[i] == y_pred[j]:
 *                             up = 1             # <<<<<<<<<<<<<<
 *                         else:
 *                             down = 1
 */
                                      __pyx_v_up = 1;

                                      /* "cyext_acv.pyx":303
 * 
 *                     if b_it == S_size:
 *                         if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                             up = 1
 *                         else:
 */
                                      goto __pyx_L35;
                                    }

                                    /* "cyext_acv.pyx":306
 *                             up = 1
 *                         else:
 *                             down = 1             # <<<<<<<<<<<<<<
 * 
 *                         p = p + lm
 */
                                    /*else*/ {
                                      __pyx_v_down = 1;
                                    }
                                    __pyx_L35:;

                                    /* "cyext_acv.pyx":308
 *                             down = 1
 * 
 *                         p = p + lm             # <<<<<<<<<<<<<<
 *                         p_u = p_u + lm * up
 *                         p_d = p_d + lm * down
 */
                                    __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                    /* "cyext_acv.pyx":309
 * 
 *                         p = p + lm
 *                         p_u = p_u + lm * up             # <<<<<<<<<<<<<<
 *                         p_d = p_d + lm * down
 *                         p_s = p_s + 1
 */
                                    __pyx_v_p_u = (__pyx_v_p_u + (__pyx_v_lm * __pyx_v_up));

                                    /* "cyext_acv.pyx":310
 *                         p = p + lm
 *                         p_u = p_u + lm * up
 *                         p_d = p_d + lm * down             # <<<<<<<<<<<<<<
 *                         p_s = p_s + 1
 *                         p_su = p_su + up
 */
                                    __pyx_v_p_d = (__pyx_v_p_d + (__pyx_v_lm * __pyx_v_down));

                                    /* "cyext_acv.pyx":311
 *                         p_u = p_u + lm * up
 *                         p_d = p_d + lm * down
 *                         p_s = p_s + 1             # <<<<<<<<<<<<<<
 *                         p_su = p_su + up
 *                         p_sd = p_sd + down
 */
                                    __pyx_v_p_s = (__pyx_v_p_s + 1);

                                    /* "cyext_acv.pyx":312
 *                         p_d = p_d + lm * down
 *                         p_s = p_s + 1
 *                         p_su = p_su + up             # <<<<<<<<<<<<<<
 *                         p_sd = p_sd + down
 * 
 */
                                    __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_up);

                                    /* "cyext_acv.pyx":313
 *                         p_s = p_s + 1
 *                         p_su = p_su + up
 *                         p_sd = p_sd + down             # <<<<<<<<<<<<<<
 * 
 *                 mean_forest_b[leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 */
                                    __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_down);

                                    /* "cyext_acv.pyx":302
 *                             b_it = b_it + 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         if fX[i] == y_pred[j]:
 *                             up = 1
 */
                                  }
                                }

                                /* "cyext_acv.pyx":315
 *                         p_sd = p_sd + down
 * 
 *                 mean_forest_b[leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_b[leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest_b[leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_23 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_33 = ((__pyx_v_p * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_23 * __pyx_v_values.strides[1]) ) + __pyx_t_22 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_33 = 0.0;
                                }
                                __pyx_t_25 = __pyx_v_leaf_numb;
                                __pyx_t_12 = __pyx_v_i;
                                __pyx_t_24 = 0;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) )) += __pyx_t_33;

                                /* "cyext_acv.pyx":316
 * 
 *                 mean_forest_b[leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest_b[leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_b[leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_22 = __pyx_v_b;
                                  __pyx_t_23 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_33 = ((__pyx_v_p_u * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_22 * __pyx_v_values.strides[0]) ) + __pyx_t_23 * __pyx_v_values.strides[1]) ) + __pyx_t_26 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_33 = 0.0;
                                }
                                __pyx_t_25 = __pyx_v_leaf_numb;
                                __pyx_t_12 = __pyx_v_i;
                                __pyx_t_24 = 1;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) )) += __pyx_t_33;

                                /* "cyext_acv.pyx":317
 *                 mean_forest_b[leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest_b[leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest_b[leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(N):
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_12 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_23 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_12 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_33 = ((__pyx_v_p_d * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_23 * __pyx_v_values.strides[1]) ) + __pyx_t_22 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_33 = 0.0;
                                }
                                __pyx_t_25 = __pyx_v_leaf_numb;
                                __pyx_t_12 = __pyx_v_i;
                                __pyx_t_24 = 2;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) )) += __pyx_t_33;
                                __pyx_L15_continue:;
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":270
 *     for b in range(n_trees):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L10;
          }
          __pyx_L10:;
        }
    }
  }

  /* "cyext_acv.pyx":319
 *                 mean_forest_b[leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         ss_u = 0
 *         ss_d = 0
 */
  __pyx_t_15 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_15;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_14; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "cyext_acv.pyx":320
 * 
 *     for i in range(N):
 *         ss_u = 0             # <<<<<<<<<<<<<<
 *         ss_d = 0
 *         ss_a = 0
 */
    __pyx_v_ss_u = 0.0;

    /* "cyext_acv.pyx":321
 *     for i in range(N):
 *         ss_u = 0
 *         ss_d = 0             # <<<<<<<<<<<<<<
 *         ss_a = 0
 *         for leaf_numb in range(max_leaves):
 */
    __pyx_v_ss_d = 0.0;

    /* "cyext_acv.pyx":322
 *         ss_u = 0
 *         ss_d = 0
 *         ss_a = 0             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(max_leaves):
 *             ss_a += mean_forest_b[leaf_numb, i, 0]
 */
    __pyx_v_ss_a = 0.0;

    /* "cyext_acv.pyx":323
 *         ss_d = 0
 *         ss_a = 0
 *         for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *             ss_a += mean_forest_b[leaf_numb, i, 0]
 *             ss_u += mean_forest_b[leaf_numb, i, 1]
 */
    __pyx_t_13 = __pyx_v_max_leaves;
    __pyx_t_16 = __pyx_t_13;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_16; __pyx_t_10+=1) {
      __pyx_v_leaf_numb = __pyx_t_10;

      /* "cyext_acv.pyx":324
 *         ss_a = 0
 *         for leaf_numb in range(max_leaves):
 *             ss_a += mean_forest_b[leaf_numb, i, 0]             # <<<<<<<<<<<<<<
 *             ss_u += mean_forest_b[leaf_numb, i, 1]
 *             ss_d += mean_forest_b[leaf_numb, i, 2]
 */
      __pyx_t_24 = __pyx_v_leaf_numb;
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_25 = 0;
      __pyx_v_ss_a = (__pyx_v_ss_a + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[2]) ))));

      /* "cyext_acv.pyx":325
 *         for leaf_numb in range(max_leaves):
 *             ss_a += mean_forest_b[leaf_numb, i, 0]
 *             ss_u += mean_forest_b[leaf_numb, i, 1]             # <<<<<<<<<<<<<<
 *             ss_d += mean_forest_b[leaf_numb, i, 2]
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
      __pyx_t_25 = __pyx_v_leaf_numb;
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_24 = 1;
      __pyx_v_ss_u = (__pyx_v_ss_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) ))));

      /* "cyext_acv.pyx":326
 *             ss_a += mean_forest_b[leaf_numb, i, 0]
 *             ss_u += mean_forest_b[leaf_numb, i, 1]
 *             ss_d += mean_forest_b[leaf_numb, i, 2]             # <<<<<<<<<<<<<<
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *         if((ss <= 1) and (ss>=0)):
 */
      __pyx_t_24 = __pyx_v_leaf_numb;
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_25 = 2;
      __pyx_v_ss_d = (__pyx_v_ss_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_12 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[2]) ))));
    }

    /* "cyext_acv.pyx":327
 *             ss_u += mean_forest_b[leaf_numb, i, 1]
 *             ss_d += mean_forest_b[leaf_numb, i, 2]
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 *         if((ss <= 1) and (ss>=0)):
 *             sdp[i] = ss
 */
    if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
      __pyx_t_33 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
    } else {
      __pyx_t_33 = 0.0;
    }
    __pyx_v_ss = __pyx_t_33;

    /* "cyext_acv.pyx":328
 *             ss_d += mean_forest_b[leaf_numb, i, 2]
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *         if((ss <= 1) and (ss>=0)):             # <<<<<<<<<<<<<<
 *             sdp[i] = ss
 *         elif(ss > 1):
 */
    __pyx_t_29 = ((__pyx_v_ss <= 1.0) != 0);
    if (__pyx_t_29) {
    } else {
      __pyx_t_1 = __pyx_t_29;
      goto __pyx_L43_bool_binop_done;
    }
    __pyx_t_29 = ((__pyx_v_ss >= 0.0) != 0);
    __pyx_t_1 = __pyx_t_29;
    __pyx_L43_bool_binop_done:;
    if (__pyx_t_1) {

      /* "cyext_acv.pyx":329
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *         if((ss <= 1) and (ss>=0)):
 *             sdp[i] = ss             # <<<<<<<<<<<<<<
 *         elif(ss > 1):
 *             sdp[i] = 1
 */
      __pyx_t_25 = __pyx_v_i;
      *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_25 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

      /* "cyext_acv.pyx":328
 *             ss_d += mean_forest_b[leaf_numb, i, 2]
 *         ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *         if((ss <= 1) and (ss>=0)):             # <<<<<<<<<<<<<<
 *             sdp[i] = ss
 *         elif(ss > 1):
 */
      goto __pyx_L42;
    }

    /* "cyext_acv.pyx":330
 *         if((ss <= 1) and (ss>=0)):
 *             sdp[i] = ss
 *         elif(ss > 1):             # <<<<<<<<<<<<<<
 *             sdp[i] = 1
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_ss > 1.0) != 0);
    if (__pyx_t_1) {

      /* "cyext_acv.pyx":331
 *             sdp[i] = ss
 *         elif(ss > 1):
 *             sdp[i] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             sdp[i] = 0
 */
      __pyx_t_25 = __pyx_v_i;
      *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_25 * __pyx_v_sdp.strides[0]) )) = 1.0;

      /* "cyext_acv.pyx":330
 *         if((ss <= 1) and (ss>=0)):
 *             sdp[i] = ss
 *         elif(ss > 1):             # <<<<<<<<<<<<<<
 *             sdp[i] = 1
 *         else:
 */
      goto __pyx_L42;
    }

    /* "cyext_acv.pyx":333
 *             sdp[i] = 1
 *         else:
 *             sdp[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp)
 */
    /*else*/ {
      __pyx_t_25 = __pyx_v_i;
      *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_25 * __pyx_v_sdp.strides[0]) )) = 0.0;
    }
    __pyx_L42:;
  }

  /* "cyext_acv.pyx":335
 *             sdp[i] = 0
 * 
 *     return np.asarray(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":237
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_clf_cat(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_7compute_sdp_clf_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_7compute_sdp_clf_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_clf_cat (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 1); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 2); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 3); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 4); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 5); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 6); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 7); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 8); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 9); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, 10); __PYX_ERR(0, 237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_clf_cat") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 237, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 237, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 238, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 238, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 238, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 239, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 239, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 240, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 240, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 240, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 241, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_clf_cat", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_6compute_sdp_clf_cat(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_6compute_sdp_clf_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_clf_cat", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 237, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 237, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_clf_cat(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":341
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_clf(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_9compute_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_clf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_o_all;
  double __pyx_v_ss_u;
  double __pyx_v_ss_a;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  int __pyx_v_lm;
  int __pyx_v_lm_s;
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  unsigned int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  unsigned int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  int __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  int __pyx_t_33;
  double __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_clf", 0);

  /* "cyext_acv.pyx":347
 *             int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":348
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":349
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":350
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     if S_size == m:
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":352
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 */
  __pyx_t_1 = ((__pyx_v_S_size == __pyx_v_m) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":353
 * 
 *     if S_size == m:
 *         return np.ones(shape=(N))             # <<<<<<<<<<<<<<
 *     elif S_size == 0:
 *         return np.zeros(shape=(N))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_4) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":352
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 */
  }

  /* "cyext_acv.pyx":354
 *     if S_size == m:
 *         return np.ones(shape=(N))
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(shape=(N))
 * 
 */
  __pyx_t_1 = ((__pyx_v_S_size == 0) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":355
 *         return np.ones(shape=(N))
 *     elif S_size == 0:
 *         return np.zeros(shape=(N))             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_trees = values.shape[0]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":354
 *     if S_size == m:
 *         return np.ones(shape=(N))
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         return np.zeros(shape=(N))
 * 
 */
  }

  /* "cyext_acv.pyx":357
 *         return np.zeros(shape=(N))
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":363
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_3);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":366
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":374
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_5);
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_p_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":375
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_p_u_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":376
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_p_d_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":379
 *     cdef double n, n_u, n_d
 * 
 *     for b in tqdm(range(n_trees)):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_7 = __pyx_t_3; __Pyx_INCREF(__pyx_t_7); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 379, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 379, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 379, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_12(__pyx_t_7);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 379, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_b = __pyx_t_13;

    /* "cyext_acv.pyx":380
 * 
 *     for b in tqdm(range(n_trees)):
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 */
    __pyx_t_14 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_14 * __pyx_v_leaves_nb.strides[0]) )));

    /* "cyext_acv.pyx":381
 *     for b in tqdm(range(n_trees)):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_15 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_17 = (__pyx_t_15 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_17 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_1, __pyx_t_13, __pyx_t_14, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p) lastprivate(__pyx_v_p_d) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_sd) lastprivate(__pyx_v_p_su) lastprivate(__pyx_v_p_u) lastprivate(__pyx_v_s) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_17; __pyx_t_16++){
                          {
                              __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_16);
                              /* Initialize private variables to invalid values */
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((int)0xbad0bad0);
                              __pyx_v_j = ((int)0xbad0bad0);
                              __pyx_v_lm = ((int)0xbad0bad0);
                              __pyx_v_lm_s = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_d = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_sd = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_su = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_u = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":382
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                              __pyx_t_18 = __pyx_v_N;
                              __pyx_t_19 = __pyx_t_18;
                              for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_19; __pyx_t_13+=1) {
                                __pyx_v_i = __pyx_t_13;

                                /* "cyext_acv.pyx":383
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             for i in range(N):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":384
 *             for i in range(N):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_20 = __pyx_v_S_size;
                                __pyx_t_21 = __pyx_t_20;
                                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                                  __pyx_v_s = __pyx_t_22;

                                  /* "cyext_acv.pyx":385
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_14 = __pyx_v_s;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_29 = 1;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_23 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_30) {
                                  } else {
                                    __pyx_t_1 = __pyx_t_30;
                                    goto __pyx_L20_bool_binop_done;
                                  }
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_14 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                  __pyx_t_23 = 0;
                                  __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_29 * __pyx_v_X.strides[0]) ) + __pyx_t_28 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_14 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_1 = __pyx_t_30;
                                  __pyx_L20_bool_binop_done:;
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":386
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":385
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":387
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_1 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_1) {

                                  /* "cyext_acv.pyx":388
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 p = 0
 */
                                  goto __pyx_L15_continue;

                                  /* "cyext_acv.pyx":387
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":390
 *                     continue
 * 
 *                 p = 0             # <<<<<<<<<<<<<<
 *                 p_u = 0
 *                 p_d = 0
 */
                                __pyx_v_p = 0;

                                /* "cyext_acv.pyx":391
 * 
 *                 p = 0
 *                 p_u = 0             # <<<<<<<<<<<<<<
 *                 p_d = 0
 *                 p_s = 0
 */
                                __pyx_v_p_u = 0;

                                /* "cyext_acv.pyx":392
 *                 p = 0
 *                 p_u = 0
 *                 p_d = 0             # <<<<<<<<<<<<<<
 *                 p_s = 0
 *                 p_su = 0
 */
                                __pyx_v_p_d = 0;

                                /* "cyext_acv.pyx":393
 *                 p_u = 0
 *                 p_d = 0
 *                 p_s = 0             # <<<<<<<<<<<<<<
 *                 p_su = 0
 *                 p_sd = 0
 */
                                __pyx_v_p_s = 0;

                                /* "cyext_acv.pyx":394
 *                 p_d = 0
 *                 p_s = 0
 *                 p_su = 0             # <<<<<<<<<<<<<<
 *                 p_sd = 0
 * 
 */
                                __pyx_v_p_su = 0;

                                /* "cyext_acv.pyx":395
 *                 p_s = 0
 *                 p_su = 0
 *                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(data.shape[0]):
 */
                                __pyx_v_p_sd = 0;

                                /* "cyext_acv.pyx":397
 *                 p_sd = 0
 * 
 *                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                     a_it = 0
 *                     b_it = 0
 */
                                __pyx_t_31 = (__pyx_v_data.shape[0]);
                                __pyx_t_32 = __pyx_t_31;
                                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_32; __pyx_t_22+=1) {
                                  __pyx_v_j = __pyx_t_22;

                                  /* "cyext_acv.pyx":398
 * 
 *                 for j in range(data.shape[0]):
 *                     a_it = 0             # <<<<<<<<<<<<<<
 *                     b_it = 0
 *                     lm = 0
 */
                                  __pyx_v_a_it = 0;

                                  /* "cyext_acv.pyx":399
 *                 for j in range(data.shape[0]):
 *                     a_it = 0
 *                     b_it = 0             # <<<<<<<<<<<<<<
 *                     lm = 0
 *                     lm_s = 0
 */
                                  __pyx_v_b_it = 0;

                                  /* "cyext_acv.pyx":400
 *                     a_it = 0
 *                     b_it = 0
 *                     lm = 0             # <<<<<<<<<<<<<<
 *                     lm_s = 0
 *                     for s in range(m):
 */
                                  __pyx_v_lm = 0;

                                  /* "cyext_acv.pyx":401
 *                     b_it = 0
 *                     lm = 0
 *                     lm_s = 0             # <<<<<<<<<<<<<<
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                  __pyx_v_lm_s = 0;

                                  /* "cyext_acv.pyx":402
 *                     lm = 0
 *                     lm_s = 0
 *                     for s in range(m):             # <<<<<<<<<<<<<<
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 */
                                  __pyx_t_20 = __pyx_v_m;
                                  __pyx_t_21 = __pyx_t_20;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_21; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":403
 *                     lm_s = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    __pyx_t_27 = __pyx_v_j;
                                    __pyx_t_23 = __pyx_v_s;
                                    __pyx_t_24 = __pyx_v_b;
                                    __pyx_t_14 = __pyx_v_leaf_numb;
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_25 = 1;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_23 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_14 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_30) {
                                    } else {
                                      __pyx_t_1 = __pyx_t_30;
                                      goto __pyx_L28_bool_binop_done;
                                    }
                                    __pyx_t_25 = __pyx_v_j;
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_14 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_23 = __pyx_v_s;
                                    __pyx_t_27 = 0;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_14 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_1 = __pyx_t_30;
                                    __pyx_L28_bool_binop_done:;
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":404
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                      __pyx_v_a_it = (__pyx_v_a_it + 1);

                                      /* "cyext_acv.pyx":403
 *                     lm_s = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":405
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1
 */
                                  __pyx_t_20 = __pyx_v_S_size;
                                  __pyx_t_21 = __pyx_t_20;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_21; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":406
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    __pyx_t_27 = __pyx_v_s;
                                    __pyx_t_23 = __pyx_v_j;
                                    __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_14 = __pyx_v_s;
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_29 = 1;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_23 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_30) {
                                    } else {
                                      __pyx_t_1 = __pyx_t_30;
                                      goto __pyx_L33_bool_binop_done;
                                    }
                                    __pyx_t_14 = __pyx_v_s;
                                    __pyx_t_29 = __pyx_v_j;
                                    __pyx_t_28 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_14 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_25 = __pyx_v_s;
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_27 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                    __pyx_t_23 = 0;
                                    __pyx_t_30 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_29 * __pyx_v_data.strides[0]) ) + __pyx_t_28 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_1 = __pyx_t_30;
                                    __pyx_L33_bool_binop_done:;
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":407
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                     if a_it == m:
 */
                                      __pyx_v_b_it = (__pyx_v_b_it + 1);

                                      /* "cyext_acv.pyx":406
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":409
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  __pyx_t_1 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":410
 * 
 *                     if a_it == m:
 *                         lm = 1             # <<<<<<<<<<<<<<
 * 
 *                     if b_it == S_size:
 */
                                    __pyx_v_lm = 1;

                                    /* "cyext_acv.pyx":409
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":412
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  __pyx_t_1 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":413
 * 
 *                     if b_it == S_size:
 *                         lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                     p = p + lm
 */
                                    __pyx_v_lm_s = 1;

                                    /* "cyext_acv.pyx":412
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":415
 *                         lm_s = 1
 * 
 *                     p = p + lm             # <<<<<<<<<<<<<<
 *                     p_s = p_s + lm_s
 *                     if fX[i] == y_pred[j]:
 */
                                  __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                  /* "cyext_acv.pyx":416
 * 
 *                     p = p + lm
 *                     p_s = p_s + lm_s             # <<<<<<<<<<<<<<
 *                     if fX[i] == y_pred[j]:
 *                         p_u = p_u + lm
 */
                                  __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                  /* "cyext_acv.pyx":417
 *                     p = p + lm
 *                     p_s = p_s + lm_s
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u = p_u + lm
 *                         p_su = p_su + lm_s
 */
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_23 = __pyx_v_j;
                                  __pyx_t_1 = (((*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) ))) == (*((long *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_23 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":418
 *                     p_s = p_s + lm_s
 *                     if fX[i] == y_pred[j]:
 *                         p_u = p_u + lm             # <<<<<<<<<<<<<<
 *                         p_su = p_su + lm_s
 *                     else:
 */
                                    __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                    /* "cyext_acv.pyx":419
 *                     if fX[i] == y_pred[j]:
 *                         p_u = p_u + lm
 *                         p_su = p_su + lm_s             # <<<<<<<<<<<<<<
 *                     else:
 *                         p_d = p_d + lm
 */
                                    __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                    /* "cyext_acv.pyx":417
 *                     p = p + lm
 *                     p_s = p_s + lm_s
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u = p_u + lm
 *                         p_su = p_su + lm_s
 */
                                    goto __pyx_L37;
                                  }

                                  /* "cyext_acv.pyx":421
 *                         p_su = p_su + lm_s
 *                     else:
 *                         p_d = p_d + lm             # <<<<<<<<<<<<<<
 *                         p_sd = p_sd + lm_s
 * 
 */
                                  /*else*/ {
                                    __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                    /* "cyext_acv.pyx":422
 *                     else:
 *                         p_d = p_d + lm
 *                         p_sd = p_sd + lm_s             # <<<<<<<<<<<<<<
 * 
 *                 mean_forest_b[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 */
                                    __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);
                                  }
                                  __pyx_L37:;
                                }

                                /* "cyext_acv.pyx":424
 *                         p_sd = p_sd + lm_s
 * 
 *                 mean_forest_b[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_b[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest_b[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_23 = __pyx_v_b;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = __pyx_v_i;
                                  __pyx_t_27 = __pyx_v_b;
                                  __pyx_t_26 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_23 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_14 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_27 * __pyx_v_values.strides[0]) ) + __pyx_t_26 * __pyx_v_values.strides[1]) ) + __pyx_t_14 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_25 = __pyx_v_leaf_numb;
                                __pyx_t_23 = __pyx_v_i;
                                __pyx_t_14 = 0;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_23 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":425
 * 
 *                 mean_forest_b[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest_b[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_b[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_14 = __pyx_v_b;
                                  __pyx_t_23 = __pyx_v_leaf_numb;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_26 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_14 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_23 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p_u * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_24 * __pyx_v_values.strides[0]) ) + __pyx_t_26 * __pyx_v_values.strides[1]) ) + __pyx_t_27 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_25 = __pyx_v_b;
                                __pyx_t_23 = __pyx_v_leaf_numb;
                                __pyx_t_14 = __pyx_v_i;
                                __pyx_t_27 = 1;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_23 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":426
 *                 mean_forest_b[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest_b[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest_b[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_27 = __pyx_v_b;
                                  __pyx_t_14 = __pyx_v_leaf_numb;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_26 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_14 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_23 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p_d * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_25 * __pyx_v_values.strides[0]) ) + __pyx_t_26 * __pyx_v_values.strides[1]) ) + __pyx_t_24 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_23 = __pyx_v_b;
                                __pyx_t_14 = __pyx_v_leaf_numb;
                                __pyx_t_27 = __pyx_v_i;
                                __pyx_t_24 = 2;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_23 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":428
 *                 mean_forest_b[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_34 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_14 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_24 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_14 * __pyx_v_p_n.strides[2]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":429
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_34 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_14 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_14 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[2]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":430
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(N):
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_34 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_14 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_24 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_14 * __pyx_v_p_d_n.strides[2]) )) += __pyx_t_34;
                                __pyx_L15_continue:;
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":381
 *     for b in tqdm(range(n_trees)):
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 o_all = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L10;
          }
          __pyx_L10:;
        }
    }

    /* "cyext_acv.pyx":379
 *     cdef double n, n_u, n_d
 * 
 *     for b in tqdm(range(n_trees)):             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cyext_acv.pyx":432
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         ss_u = 0
 *         ss_d = 0
 */
  __pyx_t_17 = __pyx_v_N;
  __pyx_t_16 = __pyx_t_17;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_16; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":433
 * 
 *     for i in range(N):
 *         ss_u = 0             # <<<<<<<<<<<<<<
 *         ss_d = 0
 *         ss_a = 0
 */
    __pyx_v_ss_u = 0.0;

    /* "cyext_acv.pyx":434
 *     for i in range(N):
 *         ss_u = 0
 *         ss_d = 0             # <<<<<<<<<<<<<<
 *         ss_a = 0
 *         for b in range(n_trees):
 */
    __pyx_v_ss_d = 0.0;

    /* "cyext_acv.pyx":435
 *         ss_u = 0
 *         ss_d = 0
 *         ss_a = 0             # <<<<<<<<<<<<<<
 *         for b in range(n_trees):
 *             nb_leaf = leaves_nb[b]
 */
    __pyx_v_ss_a = 0.0;

    /* "cyext_acv.pyx":436
 *         ss_d = 0
 *         ss_a = 0
 *         for b in range(n_trees):             # <<<<<<<<<<<<<<
 *             nb_leaf = leaves_nb[b]
 *             n = 0
 */
    __pyx_t_22 = __pyx_v_n_trees;
    __pyx_t_33 = __pyx_t_22;
    for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_33; __pyx_t_35+=1) {
      __pyx_v_b = __pyx_t_35;

      /* "cyext_acv.pyx":437
 *         ss_a = 0
 *         for b in range(n_trees):
 *             nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *             n = 0
 *             n_u = 0
 */
      __pyx_t_14 = __pyx_v_b;
      __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_14 * __pyx_v_leaves_nb.strides[0]) )));

      /* "cyext_acv.pyx":438
 *         for b in range(n_trees):
 *             nb_leaf = leaves_nb[b]
 *             n = 0             # <<<<<<<<<<<<<<
 *             n_u = 0
 *             n_d = 0
 */
      __pyx_v_n = 0.0;

      /* "cyext_acv.pyx":439
 *             nb_leaf = leaves_nb[b]
 *             n = 0
 *             n_u = 0             # <<<<<<<<<<<<<<
 *             n_d = 0
 *             for leaf_numb in range(nb_leaf):
 */
      __pyx_v_n_u = 0.0;

      /* "cyext_acv.pyx":440
 *             n = 0
 *             n_u = 0
 *             n_d = 0             # <<<<<<<<<<<<<<
 *             for leaf_numb in range(nb_leaf):
 *                 n += p_n[b, leaf_numb, i]
 */
      __pyx_v_n_d = 0.0;

      /* "cyext_acv.pyx":441
 *             n_u = 0
 *             n_d = 0
 *             for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                 n += p_n[b, leaf_numb, i]
 *                 n_u += p_u_n[b, leaf_numb, i]
 */
      __pyx_t_15 = __pyx_v_nb_leaf;
      __pyx_t_18 = __pyx_t_15;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_18; __pyx_t_36+=1) {
        __pyx_v_leaf_numb = __pyx_t_36;

        /* "cyext_acv.pyx":442
 *             n_d = 0
 *             for leaf_numb in range(nb_leaf):
 *                 n += p_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                 n_u += p_u_n[b, leaf_numb, i]
 *                 n_d += p_d_n[b, leaf_numb, i]
 */
        __pyx_t_14 = __pyx_v_b;
        __pyx_t_27 = __pyx_v_leaf_numb;
        __pyx_t_24 = __pyx_v_i;
        __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_14 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_n.strides[2]) ))));

        /* "cyext_acv.pyx":443
 *             for leaf_numb in range(nb_leaf):
 *                 n += p_n[b, leaf_numb, i]
 *                 n_u += p_u_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                 n_d += p_d_n[b, leaf_numb, i]
 * 
 */
        __pyx_t_24 = __pyx_v_b;
        __pyx_t_27 = __pyx_v_leaf_numb;
        __pyx_t_14 = __pyx_v_i;
        __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_24 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_14 * __pyx_v_p_u_n.strides[2]) ))));

        /* "cyext_acv.pyx":444
 *                 n += p_n[b, leaf_numb, i]
 *                 n_u += p_u_n[b, leaf_numb, i]
 *                 n_d += p_d_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 * 
 *             for leaf_numb in range(max_leaves):
 */
        __pyx_t_14 = __pyx_v_b;
        __pyx_t_27 = __pyx_v_leaf_numb;
        __pyx_t_24 = __pyx_v_i;
        __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_14 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_d_n.strides[2]) ))));
      }

      /* "cyext_acv.pyx":446
 *                 n_d += p_d_n[b, leaf_numb, i]
 * 
 *             for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                 ss_a += mean_forest_b[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                 ss_u += mean_forest_b[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 */
      __pyx_t_15 = __pyx_v_max_leaves;
      __pyx_t_18 = __pyx_t_15;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_18; __pyx_t_36+=1) {
        __pyx_v_leaf_numb = __pyx_t_36;

        /* "cyext_acv.pyx":447
 * 
 *             for leaf_numb in range(max_leaves):
 *                 ss_a += mean_forest_b[b, leaf_numb, i, 0]/n if n !=0 else 0             # <<<<<<<<<<<<<<
 *                 ss_u += mean_forest_b[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                 ss_d += mean_forest_b[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 */
        if (((__pyx_v_n != 0.0) != 0)) {
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_23 = 0;
          __pyx_t_34 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_23 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n);
        } else {
          __pyx_t_34 = 0.0;
        }
        __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_34);

        /* "cyext_acv.pyx":448
 *             for leaf_numb in range(max_leaves):
 *                 ss_a += mean_forest_b[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                 ss_u += mean_forest_b[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                 ss_d += mean_forest_b[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 * 
 */
        if (((__pyx_v_n_u != 0.0) != 0)) {
          __pyx_t_23 = __pyx_v_b;
          __pyx_t_14 = __pyx_v_leaf_numb;
          __pyx_t_27 = __pyx_v_i;
          __pyx_t_24 = 1;
          __pyx_t_34 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_23 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_u);
        } else {
          __pyx_t_34 = 0.0;
        }
        __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_34);

        /* "cyext_acv.pyx":449
 *                 ss_a += mean_forest_b[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                 ss_u += mean_forest_b[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                 ss_d += mean_forest_b[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *         sdp[i] = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
        if (((__pyx_v_n_d != 0.0) != 0)) {
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_14 = __pyx_v_i;
          __pyx_t_23 = 2;
          __pyx_t_34 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_24 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_14 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_23 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_d);
        } else {
          __pyx_t_34 = 0.0;
        }
        __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_34);
      }
    }

    /* "cyext_acv.pyx":451
 *                 ss_d += mean_forest_b[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 * 
 *         sdp[i] = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *     return np.array(sdp)
 */
    if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
      __pyx_t_34 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
    } else {
      __pyx_t_34 = 0.0;
    }
    __pyx_t_23 = __pyx_v_i;
    *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_23 * __pyx_v_sdp.strides[0]) )) = __pyx_t_34;
  }

  /* "cyext_acv.pyx":453
 *         sdp[i] = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 *     return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * def powerset(iterable):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":341
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_clf(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_9compute_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_9compute_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_clf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 1); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 2); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 3); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 4); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 5); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 6); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 7); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 8); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 9); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, 10); __PYX_ERR(0, 341, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_clf") < 0)) __PYX_ERR(0, 341, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 341, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 341, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 342, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 342, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 342, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_clf", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_8compute_sdp_clf(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_8compute_sdp_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_clf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 341, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 341, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_clf(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_11powerset(PyObject *__pyx_self, PyObject *__pyx_v_iterable); /*proto*/
static PyMethodDef __pyx_mdef_9cyext_acv_11powerset = {"powerset", (PyCFunction)__pyx_pw_9cyext_acv_11powerset, METH_O, 0};
static PyObject *__pyx_pw_9cyext_acv_11powerset(PyObject *__pyx_self, PyObject *__pyx_v_iterable) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("powerset (wrapper)", 0);
  __pyx_r = __pyx_pf_9cyext_acv_10powerset(__pyx_self, ((PyObject *)__pyx_v_iterable));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9cyext_acv_8powerset_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cyext_acv.pyx":457
 * def powerset(iterable):
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))             # <<<<<<<<<<<<<<
 * 
 * 
 */

static PyObject *__pyx_pf_9cyext_acv_8powerset_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)__pyx_tp_new_9cyext_acv___pyx_scope_struct_1_genexpr(__pyx_ptype_9cyext_acv___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 457, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9cyext_acv_8powerset_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_powerset_locals_genexpr, __pyx_n_s_cyext_acv); if (unlikely(!gen)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cyext_acv.powerset.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9cyext_acv_8powerset_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 457, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_s)) { __Pyx_RaiseClosureNameError("s"); __PYX_ERR(0, 457, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_s;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 457, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 + 1);
  __pyx_t_2 = __pyx_t_3;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_2; __pyx_t_4+=1) {
    __pyx_cur_scope->__pyx_v_r = __pyx_t_4;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_itertools); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_combinations); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_s)) { __Pyx_RaiseClosureNameError("s"); __PYX_ERR(0, 457, __pyx_L1_error) }
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_r); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_s, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_s, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_s);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_s);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 457, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */

static PyObject *__pyx_pf_9cyext_acv_10powerset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_iterable) {
  struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *__pyx_cur_scope;
  PyObject *__pyx_gb_9cyext_acv_8powerset_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("powerset", 0);
  __pyx_cur_scope = (struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)__pyx_tp_new_9cyext_acv___pyx_scope_struct__powerset(__pyx_ptype_9cyext_acv___pyx_scope_struct__powerset, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 455, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "cyext_acv.pyx":456
 * 
 * def powerset(iterable):
 *     s = list(iterable)             # <<<<<<<<<<<<<<
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 * 
 */
  __pyx_t_1 = PySequence_List(__pyx_v_iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":457
 * def powerset(iterable):
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_chain); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_from_iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_pf_9cyext_acv_8powerset_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cyext_acv.powerset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9cyext_acv_8powerset_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":464
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_reg(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_13compute_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_reg(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_leaves_tree = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_tree_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_value;
  double __pyx_v_value_l;
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm_u = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm_d = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_nb_leaf_l;
  unsigned int __pyx_v_o_all;
  int __pyx_v_b;
  int __pyx_v_l;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  int __pyx_v_leaf_numb_l;
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  PyObject *__pyx_v_ss_u = NULL;
  PyObject *__pyx_v_ss_d = NULL;
  PyObject *__pyx_v_ss_a = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_19;
  unsigned int __pyx_t_20;
  unsigned int __pyx_t_21;
  int __pyx_t_22;
  __Pyx_memviewslice __pyx_t_23 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  __Pyx_memviewslice __pyx_t_28 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  int __pyx_t_34;
  int __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  unsigned int __pyx_t_37;
  unsigned int __pyx_t_38;
  unsigned int __pyx_t_39;
  double __pyx_t_40;
  int __pyx_t_41;
  unsigned int __pyx_t_42;
  unsigned int __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  double __pyx_t_49;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_reg", 0);

  /* "cyext_acv.pyx":469
 *             long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling, int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":470
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":471
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":472
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":473
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :, :] leaves_tree_l
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":482
 * 
 *     cdef double[:, :, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, max_leaves, max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":485
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] norm, norm_u, norm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":488
 * 
 *     cdef double[:] norm, norm_u, norm_d
 *     norm = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 *     norm_u = np.zeros((n_trees))
 *     norm_d = np.zeros((n_trees))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_norm = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":489
 *     cdef double[:] norm, norm_u, norm_d
 *     norm = np.zeros((n_trees))
 *     norm_u = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 *     norm_d = np.zeros((n_trees))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_norm_u = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":490
 *     norm = np.zeros((n_trees))
 *     norm_u = np.zeros((n_trees))
 *     norm_d = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm, lm_u, lm_d, lm_s
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_norm_d = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":498
 * 
 *     cdef double[:, :, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":499
 *     cdef double[:, :, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_6);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_u_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":500
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_d_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":503
 *     cdef double n, n_u, n_d
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for l in range(n_trees):
 *             if b == l:
 */
  __pyx_t_11 = __pyx_v_n_trees;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_b = __pyx_t_13;

    /* "cyext_acv.pyx":504
 * 
 *     for b in range(n_trees):
 *         for l in range(n_trees):             # <<<<<<<<<<<<<<
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]
 */
    __pyx_t_14 = __pyx_v_n_trees;
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_l = __pyx_t_16;

      /* "cyext_acv.pyx":505
 *     for b in range(n_trees):
 *         for l in range(n_trees):
 *             if b == l:             # <<<<<<<<<<<<<<
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 */
      __pyx_t_17 = ((__pyx_v_b == __pyx_v_l) != 0);
      if (__pyx_t_17) {

        /* "cyext_acv.pyx":506
 *         for l in range(n_trees):
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 * 
 */
        __pyx_t_18.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_18.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_18.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_18.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_18.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_18.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
        __pyx_v_leaves_tree = __pyx_t_18;
        __pyx_t_18.memview = NULL;
        __pyx_t_18.data = NULL;

        /* "cyext_acv.pyx":507
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_t_19 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_19 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":509
 *                 nb_leaf = leaves_nb[b]
 * 
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     leaf_part = leaves_tree[leaf_numb]
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
        __pyx_t_20 = __pyx_v_nb_leaf;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_leaf_numb = __pyx_t_22;

          /* "cyext_acv.pyx":510
 * 
 *                 for leaf_numb in range(nb_leaf):
 *                     leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 */
          __pyx_t_23.data = __pyx_v_leaves_tree.data;
          __pyx_t_23.memview = __pyx_v_leaves_tree.memview;
          __PYX_INC_MEMVIEW(&__pyx_t_23, 0);
          {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_23.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_23.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_23.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_23.suboffsets[0] = -1;

__pyx_t_23.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_23.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_23.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
          __pyx_v_leaf_part = __pyx_t_23;
          __pyx_t_23.memview = NULL;
          __pyx_t_23.data = NULL;

          /* "cyext_acv.pyx":511
 *                 for leaf_numb in range(nb_leaf):
 *                     leaf_part = leaves_tree[leaf_numb]
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 * 
 *                     lm = np.zeros(data.shape[0], dtype=np.int)
 */
          __pyx_t_19 = __pyx_v_b;
          __pyx_t_24 = __pyx_v_leaf_numb;
          __pyx_t_25 = __pyx_v_b;
          __pyx_t_26 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[1]) )));
          __pyx_t_27 = 0;
          __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_25 * __pyx_v_values.strides[0]) ) + __pyx_t_26 * __pyx_v_values.strides[1]) ) + __pyx_t_27 * __pyx_v_values.strides[2]) )));

          /* "cyext_acv.pyx":513
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 *                     lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                     lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 513, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
          __pyx_v_lm = __pyx_t_28;
          __pyx_t_28.memview = NULL;
          __pyx_t_28.data = NULL;

          /* "cyext_acv.pyx":514
 * 
 *                     lm = np.zeros(data.shape[0], dtype=np.int)
 *                     lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 514, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
          __pyx_v_lm_s = __pyx_t_28;
          __pyx_t_28.memview = NULL;
          __pyx_t_28.data = NULL;

          /* "cyext_acv.pyx":516
 *                     lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         a_it = 0
 *                         b_it = 0
 */
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              __Pyx_FastGIL_Remember();
              #endif
              /*try:*/ {
                if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 516, __pyx_L13_error) }
                __pyx_t_29 = (__pyx_v_data.shape[0]);
                if ((1 == 0)) abort();
                {
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   (x)
                        #define unlikely(x) (x)
                    #endif
                    __pyx_t_31 = (__pyx_t_29 - 0 + 1 - 1/abs(1)) / 1;
                    if (__pyx_t_31 > 0)
                    {
                        #ifdef _OPENMP
                        #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_17, __pyx_t_19, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36)
                        #endif /* _OPENMP */
                        {
                            #ifdef _OPENMP
                            #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                            #endif /* _OPENMP */
                            for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_31; __pyx_t_30++){
                                {
                                    __pyx_v_i = (int)(0 + 1 * __pyx_t_30);
                                    /* Initialize private variables to invalid values */
                                    __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                    __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                    __pyx_v_s = ((int)0xbad0bad0);

                                    /* "cyext_acv.pyx":517
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                         a_it = 0             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(m):
 */
                                    __pyx_v_a_it = 0;

                                    /* "cyext_acv.pyx":518
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                         a_it = 0
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":519
 *                         a_it = 0
 *                         b_it = 0
 *                         for s in range(m):             # <<<<<<<<<<<<<<
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1
 */
                                    __pyx_t_32 = __pyx_v_m;
                                    __pyx_t_33 = __pyx_t_32;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                      __pyx_v_s = __pyx_t_34;

                                      /* "cyext_acv.pyx":520
 *                         b_it = 0
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 */
                                      __pyx_t_24 = __pyx_v_i;
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_27 = __pyx_v_s;
                                      __pyx_t_26 = 1;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_24 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_27 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_26 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (__pyx_t_35) {
                                      } else {
                                        __pyx_t_17 = __pyx_t_35;
                                        goto __pyx_L22_bool_binop_done;
                                      }
                                      __pyx_t_26 = __pyx_v_i;
                                      __pyx_t_27 = __pyx_v_s;
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_24 = 0;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_26 * __pyx_v_data.strides[0]) ) + __pyx_t_27 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_19 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_24 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_17 = __pyx_t_35;
                                      __pyx_L22_bool_binop_done:;
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":521
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 */
                                        __pyx_v_a_it = (__pyx_v_a_it + 1);

                                        /* "cyext_acv.pyx":520
 *                         b_it = 0
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":522
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_32 = __pyx_v_S_size;
                                    __pyx_t_33 = __pyx_t_32;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                      __pyx_v_s = __pyx_t_34;

                                      /* "cyext_acv.pyx":523
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_19 = __pyx_v_i;
                                      __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_26 = __pyx_v_s;
                                      __pyx_t_25 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_36 = 1;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_27 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_25 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (__pyx_t_35) {
                                      } else {
                                        __pyx_t_17 = __pyx_t_35;
                                        goto __pyx_L27_bool_binop_done;
                                      }
                                      __pyx_t_26 = __pyx_v_s;
                                      __pyx_t_36 = __pyx_v_i;
                                      __pyx_t_25 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_19 = 0;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_25 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_27 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_17 = __pyx_t_35;
                                      __pyx_L27_bool_binop_done:;
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":524
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if a_it == m:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":523
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":526
 *                                 b_it = b_it + 1
 * 
 *                         if a_it == m:             # <<<<<<<<<<<<<<
 *                             lm[i] = 1
 * 
 */
                                    __pyx_t_17 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":527
 * 
 *                         if a_it == m:
 *                             lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == S_size:
 */
                                      __pyx_t_24 = __pyx_v_i;
                                      *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_24 * __pyx_v_lm.strides[0]) )) = 1;

                                      /* "cyext_acv.pyx":526
 *                                 b_it = b_it + 1
 * 
 *                         if a_it == m:             # <<<<<<<<<<<<<<
 *                             lm[i] = 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":529
 *                             lm[i] = 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[i] = 1
 * 
 */
                                    __pyx_t_17 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":530
 * 
 *                         if b_it == S_size:
 *                             lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                     for i in prange(N, nogil=True, num_threads=num_threads):
 */
                                      __pyx_t_24 = __pyx_v_i;
                                      *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_24 * __pyx_v_lm_s.strides[0]) )) = 1;

                                      /* "cyext_acv.pyx":529
 *                             lm[i] = 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[i] = 1
 * 
 */
                                    }
                                }
                            }
                        }
                    }
                }
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   __builtin_expect(!!(x), 1)
                    #define unlikely(x) __builtin_expect(!!(x), 0)
                #endif
              }

              /* "cyext_acv.pyx":516
 *                     lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         a_it = 0
 *                         b_it = 0
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L14;
                }
                __pyx_L13_error: {
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L1_error;
                }
                __pyx_L14:;
              }
          }

          /* "cyext_acv.pyx":532
 *                             lm_s[i] = 1
 * 
 *                     for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              __Pyx_FastGIL_Remember();
              #endif
              /*try:*/ {
                __pyx_t_32 = __pyx_v_N;
                if ((1 == 0)) abort();
                {
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   (x)
                        #define unlikely(x) (x)
                    #endif
                    __pyx_t_37 = (__pyx_t_32 - 0 + 1 - 1/abs(1)) / 1;
                    if (__pyx_t_37 > 0)
                    {
                        #ifdef _OPENMP
                        #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) num_threads(__pyx_v_num_threads) private(__pyx_t_17, __pyx_t_19, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_30, __pyx_t_31, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_38, __pyx_t_39, __pyx_t_40)
                        #endif /* _OPENMP */
                        {
                            #ifdef _OPENMP
                            #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                            #endif /* _OPENMP */
                            for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_37; __pyx_t_33++){
                                {
                                    __pyx_v_i = (int)(0 + 1 * __pyx_t_33);
                                    /* Initialize private variables to invalid values */
                                    __pyx_v_j = ((int)0xbad0bad0);
                                    __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                    __pyx_v_s = ((int)0xbad0bad0);

                                    /* "cyext_acv.pyx":533
 * 
 *                     for i in prange(N, nogil=True, num_threads=num_threads):
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":534
 *                     for i in prange(N, nogil=True, num_threads=num_threads):
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_38 = __pyx_v_S_size;
                                    __pyx_t_39 = __pyx_t_38;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_39; __pyx_t_34+=1) {
                                      __pyx_v_s = __pyx_t_34;

                                      /* "cyext_acv.pyx":535
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_19 = __pyx_v_i;
                                      __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_26 = __pyx_v_s;
                                      __pyx_t_25 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_36 = 1;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_19 * __pyx_v_X.strides[0]) ) + __pyx_t_27 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_25 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (!__pyx_t_35) {
                                      } else {
                                        __pyx_t_17 = __pyx_t_35;
                                        goto __pyx_L45_bool_binop_done;
                                      }
                                      __pyx_t_26 = __pyx_v_s;
                                      __pyx_t_36 = __pyx_v_i;
                                      __pyx_t_25 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_26 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_27 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_24 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_19 = 0;
                                      __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_36 * __pyx_v_X.strides[0]) ) + __pyx_t_25 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_27 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_17 = __pyx_t_35;
                                      __pyx_L45_bool_binop_done:;
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":536
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                         if o_all > 0:
 *                             continue
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":535
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":537
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    __pyx_t_17 = ((__pyx_v_o_all > 0) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":538
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         p = 0
 */
                                      goto __pyx_L38_continue;

                                      /* "cyext_acv.pyx":537
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":540
 *                             continue
 * 
 *                         p = 0             # <<<<<<<<<<<<<<
 *                         p_u = 0
 *                         p_d = 0
 */
                                    __pyx_v_p = 0;

                                    /* "cyext_acv.pyx":541
 * 
 *                         p = 0
 *                         p_u = 0             # <<<<<<<<<<<<<<
 *                         p_d = 0
 *                         p_s = 0
 */
                                    __pyx_v_p_u = 0;

                                    /* "cyext_acv.pyx":542
 *                         p = 0
 *                         p_u = 0
 *                         p_d = 0             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_su = 0
 */
                                    __pyx_v_p_d = 0;

                                    /* "cyext_acv.pyx":543
 *                         p_u = 0
 *                         p_d = 0
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_su = 0
 *                         p_sd = 0
 */
                                    __pyx_v_p_s = 0;

                                    /* "cyext_acv.pyx":544
 *                         p_d = 0
 *                         p_s = 0
 *                         p_su = 0             # <<<<<<<<<<<<<<
 *                         p_sd = 0
 * 
 */
                                    __pyx_v_p_su = 0;

                                    /* "cyext_acv.pyx":545
 *                         p_s = 0
 *                         p_su = 0
 *                         p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                         for j in range(data.shape[0]):
 */
                                    __pyx_v_p_sd = 0;

                                    /* "cyext_acv.pyx":547
 *                         p_sd = 0
 * 
 *                         for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                             p += lm[j]
 *                             p_s += lm_s[j]
 */
                                    __pyx_t_31 = (__pyx_v_data.shape[0]);
                                    __pyx_t_30 = __pyx_t_31;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_30; __pyx_t_34+=1) {
                                      __pyx_v_j = __pyx_t_34;

                                      /* "cyext_acv.pyx":548
 * 
 *                         for j in range(data.shape[0]):
 *                             p += lm[j]             # <<<<<<<<<<<<<<
 *                             p_s += lm_s[j]
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 */
                                      __pyx_t_24 = __pyx_v_j;
                                      __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_24 * __pyx_v_lm.strides[0]) ))));

                                      /* "cyext_acv.pyx":549
 *                         for j in range(data.shape[0]):
 *                             p += lm[j]
 *                             p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                 p_u += lm[j]
 */
                                      __pyx_t_24 = __pyx_v_j;
                                      __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_24 * __pyx_v_lm_s.strides[0]) ))));

                                      /* "cyext_acv.pyx":550
 *                             p += lm[j]
 *                             p_s += lm_s[j]
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                 p_u += lm[j]
 *                                 p_su += lm_s[j]
 */
                                      __pyx_t_24 = __pyx_v_i;
                                      __pyx_t_19 = __pyx_v_j;
                                      __pyx_t_27 = __pyx_v_i;
                                      __pyx_t_26 = __pyx_v_j;
                                      __pyx_t_17 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_19 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_27 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_26 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":551
 *                             p_s += lm_s[j]
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                 p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                 p_su += lm_s[j]
 *                             else:
 */
                                        __pyx_t_26 = __pyx_v_j;
                                        __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_26 * __pyx_v_lm.strides[0]) ))));

                                        /* "cyext_acv.pyx":552
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                 p_u += lm[j]
 *                                 p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                             else:
 *                                 p_d += lm[j]
 */
                                        __pyx_t_26 = __pyx_v_j;
                                        __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_26 * __pyx_v_lm_s.strides[0]) ))));

                                        /* "cyext_acv.pyx":550
 *                             p += lm[j]
 *                             p_s += lm_s[j]
 *                             if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                 p_u += lm[j]
 *                                 p_su += lm_s[j]
 */
                                        goto __pyx_L50;
                                      }

                                      /* "cyext_acv.pyx":554
 *                                 p_su += lm_s[j]
 *                             else:
 *                                 p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                 p_sd += lm_s[j]
 * 
 */
                                      /*else*/ {
                                        __pyx_t_26 = __pyx_v_j;
                                        __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_26 * __pyx_v_lm.strides[0]) ))));

                                        /* "cyext_acv.pyx":555
 *                             else:
 *                                 p_d += lm[j]
 *                                 p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 */
                                        __pyx_t_26 = __pyx_v_j;
                                        __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_26 * __pyx_v_lm_s.strides[0]) ))));
                                      }
                                      __pyx_L50:;
                                    }

                                    /* "cyext_acv.pyx":557
 *                                 p_sd += lm_s[j]
 * 
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_26 = __pyx_v_i;
                                      __pyx_t_40 = ((((__pyx_v_p * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_s) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_26 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p) * __pyx_v_value) / __pyx_v_p_s));
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_27 = __pyx_v_leaf_numb;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = __pyx_v_i;
                                    __pyx_t_25 = 0;
                                    *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_26 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                    /* "cyext_acv.pyx":558
 * 
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_25 = __pyx_v_i;
                                      __pyx_t_40 = ((((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_su) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_25 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_u) * __pyx_v_value) / __pyx_v_p_su));
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_25 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_27 = __pyx_v_i;
                                    __pyx_t_26 = 1;
                                    *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                    /* "cyext_acv.pyx":559
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         p_n[b, leaf_numb, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_26 = __pyx_v_i;
                                      __pyx_t_40 = ((((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_sd) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_26 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_d) * __pyx_v_value) / __pyx_v_p_sd));
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_27 = __pyx_v_leaf_numb;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = __pyx_v_i;
                                    __pyx_t_25 = 2;
                                    *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_26 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                    /* "cyext_acv.pyx":561
 *                         mean_forest_b[b, leaf_numb, leaf_numb, i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 * 
 *                         p_n[b, leaf_numb, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_u_n[b, leaf_numb, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_40 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_25 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_27 = __pyx_v_i;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_25 * __pyx_v_p_n.strides[0]) ) + __pyx_t_24 * __pyx_v_p_n.strides[1]) ) + __pyx_t_19 * __pyx_v_p_n.strides[2]) ) + __pyx_t_27 * __pyx_v_p_n.strides[3]) )) += __pyx_t_40;

                                    /* "cyext_acv.pyx":562
 * 
 *                         p_n[b, leaf_numb, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_d_n[b, leaf_numb, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *             else:
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_40 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_27 = __pyx_v_b;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_25 = __pyx_v_i;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_27 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_19 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_25 * __pyx_v_p_u_n.strides[3]) )) += __pyx_t_40;

                                    /* "cyext_acv.pyx":563
 *                         p_n[b, leaf_numb, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_40 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_40 = 0.0;
                                    }
                                    __pyx_t_25 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_19 = __pyx_v_leaf_numb;
                                    __pyx_t_27 = __pyx_v_i;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_25 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_24 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_19 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[3]) )) += __pyx_t_40;
                                    goto __pyx_L52;
                                    __pyx_L38_continue:;
                                    goto __pyx_L52;
                                    __pyx_L52:;
                                }
                            }
                        }
                    }
                }
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   __builtin_expect(!!(x), 1)
                    #define unlikely(x) __builtin_expect(!!(x), 0)
                #endif
              }

              /* "cyext_acv.pyx":532
 *                             lm_s[i] = 1
 * 
 *                     for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L37;
                }
                __pyx_L37:;
              }
          }
        }

        /* "cyext_acv.pyx":505
 *     for b in range(n_trees):
 *         for l in range(n_trees):
 *             if b == l:             # <<<<<<<<<<<<<<
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 */
        goto __pyx_L7;
      }

      /* "cyext_acv.pyx":565
 *                         p_d_n[b, leaf_numb, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]
 */
      /*else*/ {
        __pyx_t_18.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_18.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_18.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_18.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_18.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_18.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
        __pyx_v_leaves_tree = __pyx_t_18;
        __pyx_t_18.memview = NULL;
        __pyx_t_18.data = NULL;

        /* "cyext_acv.pyx":566
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 leaves_tree_l = partition_leaves_trees[l]
 *                 nb_leaf_l = leaves_nb[l]
 */
        __pyx_t_27 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_27 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":567
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]             # <<<<<<<<<<<<<<
 *                 nb_leaf_l = leaves_nb[l]
 * 
 */
        __pyx_t_18.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_18.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_18.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_18.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_18.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_18.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
        __pyx_v_leaves_tree_l = __pyx_t_18;
        __pyx_t_18.memview = NULL;
        __pyx_t_18.data = NULL;

        /* "cyext_acv.pyx":568
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]
 *                 nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_t_27 = __pyx_v_l;
        __pyx_v_nb_leaf_l = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_27 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":570
 *                 nb_leaf_l = leaves_nb[l]
 * 
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for leaf_numb_l in range(nb_leaf_l):
 * 
 */
        __pyx_t_20 = __pyx_v_nb_leaf;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_leaf_numb = __pyx_t_22;

          /* "cyext_acv.pyx":571
 * 
 *                 for leaf_numb in range(nb_leaf):
 *                     for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 * 
 *                         leaf_part = leaves_tree[leaf_numb]
 */
          __pyx_t_37 = __pyx_v_nb_leaf_l;
          __pyx_t_33 = __pyx_t_37;
          for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
            __pyx_v_leaf_numb_l = __pyx_t_34;

            /* "cyext_acv.pyx":573
 *                     for leaf_numb_l in range(nb_leaf_l):
 * 
 *                         leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
            __pyx_t_23.data = __pyx_v_leaves_tree.data;
            __pyx_t_23.memview = __pyx_v_leaves_tree.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_23, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_23.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_23.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_23.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_23.suboffsets[0] = -1;

__pyx_t_23.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_23.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_23.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
            __pyx_v_leaf_part = __pyx_t_23;
            __pyx_t_23.memview = NULL;
            __pyx_t_23.data = NULL;

            /* "cyext_acv.pyx":574
 * 
 *                         leaf_part = leaves_tree[leaf_numb]
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]             # <<<<<<<<<<<<<<
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 */
            __pyx_t_23.data = __pyx_v_leaves_tree_l.data;
            __pyx_t_23.memview = __pyx_v_leaves_tree_l.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_23, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree_l.strides[0];
        __pyx_t_23.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_23.shape[0] = __pyx_v_leaves_tree_l.shape[1];
__pyx_t_23.strides[0] = __pyx_v_leaves_tree_l.strides[1];
    __pyx_t_23.suboffsets[0] = -1;

__pyx_t_23.shape[1] = __pyx_v_leaves_tree_l.shape[2];
__pyx_t_23.strides[1] = __pyx_v_leaves_tree_l.strides[2];
    __pyx_t_23.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
            __pyx_v_leaf_part_l = __pyx_t_23;
            __pyx_t_23.memview = NULL;
            __pyx_t_23.data = NULL;

            /* "cyext_acv.pyx":575
 *                         leaf_part = leaves_tree[leaf_numb]
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 */
            __pyx_t_27 = __pyx_v_b;
            __pyx_t_19 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_25 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[1]) )));
            __pyx_t_26 = 0;
            __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_24 * __pyx_v_values.strides[0]) ) + __pyx_t_25 * __pyx_v_values.strides[1]) ) + __pyx_t_26 * __pyx_v_values.strides[2]) )));

            /* "cyext_acv.pyx":576
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]             # <<<<<<<<<<<<<<
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)
 */
            __pyx_t_19 = __pyx_v_l;
            __pyx_t_27 = __pyx_v_leaf_numb_l;
            __pyx_t_26 = __pyx_v_l;
            __pyx_t_25 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
            __pyx_t_24 = 0;
            __pyx_v_value_l = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_25 * __pyx_v_values.strides[1]) ) + __pyx_t_24 * __pyx_v_values.strides[2]) )));

            /* "cyext_acv.pyx":578
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 578, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
            __pyx_v_lm = __pyx_t_28;
            __pyx_t_28.memview = NULL;
            __pyx_t_28.data = NULL;

            /* "cyext_acv.pyx":579
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 579, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
            __pyx_v_lm_s = __pyx_t_28;
            __pyx_t_28.memview = NULL;
            __pyx_t_28.data = NULL;

            /* "cyext_acv.pyx":581
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
            {
                #ifdef WITH_THREAD
                PyThreadState *_save;
                Py_UNBLOCK_THREADS
                __Pyx_FastGIL_Remember();
                #endif
                /*try:*/ {
                  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 581, __pyx_L60_error) }
                  __pyx_t_31 = (__pyx_v_data.shape[0]);
                  if ((1 == 0)) abort();
                  {
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   (x)
                          #define unlikely(x) (x)
                      #endif
                      __pyx_t_29 = (__pyx_t_31 - 0 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_29 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_17, __pyx_t_19, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_32, __pyx_t_35, __pyx_t_36, __pyx_t_38, __pyx_t_41)
                          #endif /* _OPENMP */
                          {
                              #ifdef _OPENMP
                              #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                              #endif /* _OPENMP */
                              for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30++){
                                  {
                                      __pyx_v_i = (int)(0 + 1 * __pyx_t_30);
                                      /* Initialize private variables to invalid values */
                                      __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                      __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                      __pyx_v_s = ((int)0xbad0bad0);

                                      /* "cyext_acv.pyx":582
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             b_it = 0
 *                             for s in range(m):
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":583
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                             a_it = 0
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":584
 *                             a_it = 0
 *                             b_it = 0
 *                             for s in range(m):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_32 = __pyx_v_m;
                                      __pyx_t_38 = __pyx_t_32;
                                      for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_38; __pyx_t_41+=1) {
                                        __pyx_v_s = __pyx_t_41;

                                        /* "cyext_acv.pyx":585
 *                             b_it = 0
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        __pyx_t_27 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_24 = __pyx_v_s;
                                        __pyx_t_25 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_24 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L69_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_i;
                                        __pyx_t_24 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_27 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_19 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L69_bool_binop_done;
                                        }
                                        __pyx_t_27 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_24 = __pyx_v_s;
                                        __pyx_t_25 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_24 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L69_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_i;
                                        __pyx_t_24 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_27 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_19 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        __pyx_t_17 = __pyx_t_35;
                                        __pyx_L69_bool_binop_done:;
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":586
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":585
 *                             b_it = 0
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":587
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                     b_it = b_it + 1
 */
                                      __pyx_t_32 = __pyx_v_S_size;
                                      __pyx_t_38 = __pyx_t_32;
                                      for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_38; __pyx_t_41+=1) {
                                        __pyx_v_s = __pyx_t_41;

                                        /* "cyext_acv.pyx":588
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_36 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_26 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L76_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_24 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L76_bool_binop_done;
                                        }
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_36 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_26 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        if (__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L76_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_24 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        __pyx_t_17 = __pyx_t_35;
                                        __pyx_L76_bool_binop_done:;
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":589
 *                             for s in range(S_size):
 *                                 if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                     b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if a_it == m:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":588
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":591
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == m:             # <<<<<<<<<<<<<<
 *                                 lm[i] = 1
 * 
 */
                                      __pyx_t_17 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":592
 * 
 *                             if a_it == m:
 *                                 lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                             if b_it == S_size:
 */
                                        __pyx_t_27 = __pyx_v_i;
                                        *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_27 * __pyx_v_lm.strides[0]) )) = 1;

                                        /* "cyext_acv.pyx":591
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == m:             # <<<<<<<<<<<<<<
 *                                 lm[i] = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":594
 *                                 lm[i] = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s[i] = 1
 * 
 */
                                      __pyx_t_17 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":595
 * 
 *                             if b_it == S_size:
 *                                 lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                         for i in prange(N, nogil=True, num_threads=num_threads):
 */
                                        __pyx_t_27 = __pyx_v_i;
                                        *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_27 * __pyx_v_lm_s.strides[0]) )) = 1;

                                        /* "cyext_acv.pyx":594
 *                                 lm[i] = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s[i] = 1
 * 
 */
                                      }
                                  }
                              }
                          }
                      }
                  }
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   __builtin_expect(!!(x), 1)
                      #define unlikely(x) __builtin_expect(!!(x), 0)
                  #endif
                }

                /* "cyext_acv.pyx":581
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L61;
                  }
                  __pyx_L60_error: {
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L1_error;
                  }
                  __pyx_L61:;
                }
            }

            /* "cyext_acv.pyx":597
 *                                 lm_s[i] = 1
 * 
 *                         for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             o_all = 0
 *                             for s in range(S_size):
 */
            {
                #ifdef WITH_THREAD
                PyThreadState *_save;
                Py_UNBLOCK_THREADS
                __Pyx_FastGIL_Remember();
                #endif
                /*try:*/ {
                  __pyx_t_32 = __pyx_v_N;
                  if ((1 == 0)) abort();
                  {
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   (x)
                          #define unlikely(x) (x)
                      #endif
                      __pyx_t_39 = (__pyx_t_32 - 0 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_39 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) num_threads(__pyx_v_num_threads) private(__pyx_t_17, __pyx_t_19, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_29, __pyx_t_30, __pyx_t_35, __pyx_t_36, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43)
                          #endif /* _OPENMP */
                          {
                              #ifdef _OPENMP
                              #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                              #endif /* _OPENMP */
                              for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_39; __pyx_t_38++){
                                  {
                                      __pyx_v_i = (int)(0 + 1 * __pyx_t_38);
                                      /* Initialize private variables to invalid values */
                                      __pyx_v_j = ((int)0xbad0bad0);
                                      __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                      __pyx_v_s = ((int)0xbad0bad0);

                                      /* "cyext_acv.pyx":598
 * 
 *                         for i in prange(N, nogil=True, num_threads=num_threads):
 *                             o_all = 0             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 */
                                      __pyx_v_o_all = 0;

                                      /* "cyext_acv.pyx":599
 *                         for i in prange(N, nogil=True, num_threads=num_threads):
 *                             o_all = 0
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 */
                                      __pyx_t_42 = __pyx_v_S_size;
                                      __pyx_t_43 = __pyx_t_42;
                                      for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_43; __pyx_t_41+=1) {
                                        __pyx_v_s = __pyx_t_41;

                                        /* "cyext_acv.pyx":600
 *                             o_all = 0
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 */
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_36 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_19 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_26 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (!__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L96_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_36 * __pyx_v_X.strides[0]) ) + __pyx_t_26 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_24 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (!__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L96_bool_binop_done;
                                        }
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_36 = 1;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_19 * __pyx_v_X.strides[0]) ) + __pyx_t_24 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_26 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        if (!__pyx_t_35) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_35;
                                          goto __pyx_L96_bool_binop_done;
                                        }
                                        __pyx_t_25 = __pyx_v_s;
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_25 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_27 = __pyx_v_s;
                                        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_27 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_35 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_36 * __pyx_v_X.strides[0]) ) + __pyx_t_26 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_24 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        __pyx_t_17 = __pyx_t_35;
                                        __pyx_L96_bool_binop_done:;
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":601
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                             if o_all > 0:
 *                                 continue
 */
                                          __pyx_v_o_all = (__pyx_v_o_all + 1);

                                          /* "cyext_acv.pyx":600
 *                             o_all = 0
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":602
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 *                             if o_all > 0:             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                      __pyx_t_17 = ((__pyx_v_o_all > 0) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":603
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 *                                 continue             # <<<<<<<<<<<<<<
 * 
 *                             p = 0
 */
                                        goto __pyx_L89_continue;

                                        /* "cyext_acv.pyx":602
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 *                             if o_all > 0:             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":605
 *                                 continue
 * 
 *                             p = 0             # <<<<<<<<<<<<<<
 *                             p_u = 0
 *                             p_d = 0
 */
                                      __pyx_v_p = 0;

                                      /* "cyext_acv.pyx":606
 * 
 *                             p = 0
 *                             p_u = 0             # <<<<<<<<<<<<<<
 *                             p_d = 0
 *                             p_s = 0
 */
                                      __pyx_v_p_u = 0;

                                      /* "cyext_acv.pyx":607
 *                             p = 0
 *                             p_u = 0
 *                             p_d = 0             # <<<<<<<<<<<<<<
 *                             p_s = 0
 *                             p_su = 0
 */
                                      __pyx_v_p_d = 0;

                                      /* "cyext_acv.pyx":608
 *                             p_u = 0
 *                             p_d = 0
 *                             p_s = 0             # <<<<<<<<<<<<<<
 *                             p_su = 0
 *                             p_sd = 0
 */
                                      __pyx_v_p_s = 0;

                                      /* "cyext_acv.pyx":609
 *                             p_d = 0
 *                             p_s = 0
 *                             p_su = 0             # <<<<<<<<<<<<<<
 *                             p_sd = 0
 * 
 */
                                      __pyx_v_p_su = 0;

                                      /* "cyext_acv.pyx":610
 *                             p_s = 0
 *                             p_su = 0
 *                             p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                             for j in range(data.shape[0]):
 */
                                      __pyx_v_p_sd = 0;

                                      /* "cyext_acv.pyx":612
 *                             p_sd = 0
 * 
 *                             for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                 p += lm[j]
 *                                 p_s += lm_s[j]
 */
                                      __pyx_t_29 = (__pyx_v_data.shape[0]);
                                      __pyx_t_30 = __pyx_t_29;
                                      for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_30; __pyx_t_41+=1) {
                                        __pyx_v_j = __pyx_t_41;

                                        /* "cyext_acv.pyx":613
 * 
 *                             for j in range(data.shape[0]):
 *                                 p += lm[j]             # <<<<<<<<<<<<<<
 *                                 p_s += lm_s[j]
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 */
                                        __pyx_t_27 = __pyx_v_j;
                                        __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_27 * __pyx_v_lm.strides[0]) ))));

                                        /* "cyext_acv.pyx":614
 *                             for j in range(data.shape[0]):
 *                                 p += lm[j]
 *                                 p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                     p_u += lm[j]
 */
                                        __pyx_t_27 = __pyx_v_j;
                                        __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_27 * __pyx_v_lm_s.strides[0]) ))));

                                        /* "cyext_acv.pyx":615
 *                                 p += lm[j]
 *                                 p_s += lm_s[j]
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                     p_u += lm[j]
 *                                     p_su += lm_s[j]
 */
                                        __pyx_t_27 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_j;
                                        __pyx_t_24 = __pyx_v_i;
                                        __pyx_t_25 = __pyx_v_j;
                                        __pyx_t_17 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_27 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_19 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_25 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":616
 *                                 p_s += lm_s[j]
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                     p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_su += lm_s[j]
 *                                 else:
 */
                                          __pyx_t_25 = __pyx_v_j;
                                          __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_25 * __pyx_v_lm.strides[0]) ))));

                                          /* "cyext_acv.pyx":617
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:
 *                                     p_u += lm[j]
 *                                     p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     p_d += lm[j]
 */
                                          __pyx_t_25 = __pyx_v_j;
                                          __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_25 * __pyx_v_lm_s.strides[0]) ))));

                                          /* "cyext_acv.pyx":615
 *                                 p += lm[j]
 *                                 p_s += lm_s[j]
 *                                 if (fX[i] - y_pred[j])*(fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                     p_u += lm[j]
 *                                     p_su += lm_s[j]
 */
                                          goto __pyx_L103;
                                        }

                                        /* "cyext_acv.pyx":619
 *                                     p_su += lm_s[j]
 *                                 else:
 *                                     p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_sd += lm_s[j]
 * 
 */
                                        /*else*/ {
                                          __pyx_t_25 = __pyx_v_j;
                                          __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_25 * __pyx_v_lm.strides[0]) ))));

                                          /* "cyext_acv.pyx":620
 *                                 else:
 *                                     p_d += lm[j]
 *                                     p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 */
                                          __pyx_t_25 = __pyx_v_j;
                                          __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_25 * __pyx_v_lm_s.strides[0]) ))));
                                        }
                                        __pyx_L103:;
                                      }

                                      /* "cyext_acv.pyx":622
 *                                     p_sd += lm_s[j]
 * 
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 */
                                      if (((__pyx_v_p_s != 0) != 0)) {
                                        __pyx_t_40 = (((__pyx_v_p * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_s);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_25 = __pyx_v_b;
                                      __pyx_t_24 = __pyx_v_leaf_numb;
                                      __pyx_t_19 = __pyx_v_leaf_numb_l;
                                      __pyx_t_27 = __pyx_v_i;
                                      __pyx_t_26 = 0;
                                      *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                      /* "cyext_acv.pyx":623
 * 
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                      if (((__pyx_v_p_su != 0) != 0)) {
                                        __pyx_t_40 = (((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_su);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_26 = __pyx_v_b;
                                      __pyx_t_27 = __pyx_v_leaf_numb;
                                      __pyx_t_19 = __pyx_v_leaf_numb_l;
                                      __pyx_t_24 = __pyx_v_i;
                                      __pyx_t_25 = 1;
                                      *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_26 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                      /* "cyext_acv.pyx":624
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                             p_n[b, leaf_numb, leaf_numb_l, i] += (1.*p)/p_s if p_s != 0 else 0
 */
                                      if (((__pyx_v_p_sd != 0) != 0)) {
                                        __pyx_t_40 = (((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_sd);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_25 = __pyx_v_b;
                                      __pyx_t_24 = __pyx_v_leaf_numb;
                                      __pyx_t_19 = __pyx_v_leaf_numb_l;
                                      __pyx_t_27 = __pyx_v_i;
                                      __pyx_t_26 = 2;
                                      *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_25 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_27 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_40;

                                      /* "cyext_acv.pyx":626
 *                             mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 *                             p_n[b, leaf_numb, leaf_numb_l, i] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                             p_u_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                      if (((__pyx_v_p_s != 0) != 0)) {
                                        __pyx_t_40 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_26 = __pyx_v_b;
                                      __pyx_t_27 = __pyx_v_leaf_numb;
                                      __pyx_t_19 = __pyx_v_leaf_numb_l;
                                      __pyx_t_24 = __pyx_v_i;
                                      *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_26 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_19 * __pyx_v_p_n.strides[2]) ) + __pyx_t_24 * __pyx_v_p_n.strides[3]) )) += __pyx_t_40;

                                      /* "cyext_acv.pyx":627
 * 
 *                             p_n[b, leaf_numb, leaf_numb_l, i] += (1.*p)/p_s if p_s != 0 else 0
 *                             p_u_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                             p_d_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *     for i in range(N):
 */
                                      if (((__pyx_v_p_su != 0) != 0)) {
                                        __pyx_t_40 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_24 = __pyx_v_b;
                                      __pyx_t_19 = __pyx_v_leaf_numb;
                                      __pyx_t_27 = __pyx_v_leaf_numb_l;
                                      __pyx_t_26 = __pyx_v_i;
                                      *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_24 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_19 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_26 * __pyx_v_p_u_n.strides[3]) )) += __pyx_t_40;

                                      /* "cyext_acv.pyx":628
 *                             p_n[b, leaf_numb, leaf_numb_l, i] += (1.*p)/p_s if p_s != 0 else 0
 *                             p_u_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         ss_u = 0
 */
                                      if (((__pyx_v_p_sd != 0) != 0)) {
                                        __pyx_t_40 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                      } else {
                                        __pyx_t_40 = 0.0;
                                      }
                                      __pyx_t_26 = __pyx_v_b;
                                      __pyx_t_27 = __pyx_v_leaf_numb;
                                      __pyx_t_19 = __pyx_v_leaf_numb_l;
                                      __pyx_t_24 = __pyx_v_i;
                                      *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_26 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_19 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_24 * __pyx_v_p_d_n.strides[3]) )) += __pyx_t_40;
                                      goto __pyx_L105;
                                      __pyx_L89_continue:;
                                      goto __pyx_L105;
                                      __pyx_L105:;
                                  }
                              }
                          }
                      }
                  }
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   __builtin_expect(!!(x), 1)
                      #define unlikely(x) __builtin_expect(!!(x), 0)
                  #endif
                }

                /* "cyext_acv.pyx":597
 *                                 lm_s[i] = 1
 * 
 *                         for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             o_all = 0
 *                             for s in range(S_size):
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L88;
                  }
                  __pyx_L88:;
                }
            }
          }
        }
      }
      __pyx_L7:;
    }
  }

  /* "cyext_acv.pyx":629
 *                             p_u_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         ss_u = 0
 *         ss_d = 0
 */
  __pyx_t_20 = __pyx_v_N;
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_21; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "cyext_acv.pyx":630
 *                             p_d_n[b, leaf_numb, leaf_numb_l, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *     for i in range(N):
 *         ss_u = 0             # <<<<<<<<<<<<<<
 *         ss_d = 0
 *         ss_a = 0
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_XDECREF_SET(__pyx_v_ss_u, __pyx_int_0);

    /* "cyext_acv.pyx":631
 *     for i in range(N):
 *         ss_u = 0
 *         ss_d = 0             # <<<<<<<<<<<<<<
 *         ss_a = 0
 *         for b in range(n_trees):
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_XDECREF_SET(__pyx_v_ss_d, __pyx_int_0);

    /* "cyext_acv.pyx":632
 *         ss_u = 0
 *         ss_d = 0
 *         ss_a = 0             # <<<<<<<<<<<<<<
 *         for b in range(n_trees):
 *             for l in range(n_trees):
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_XDECREF_SET(__pyx_v_ss_a, __pyx_int_0);

    /* "cyext_acv.pyx":633
 *         ss_d = 0
 *         ss_a = 0
 *         for b in range(n_trees):             # <<<<<<<<<<<<<<
 *             for l in range(n_trees):
 *                 if b == l:
 */
    __pyx_t_12 = __pyx_v_n_trees;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_b = __pyx_t_14;

      /* "cyext_acv.pyx":634
 *         ss_a = 0
 *         for b in range(n_trees):
 *             for l in range(n_trees):             # <<<<<<<<<<<<<<
 *                 if b == l:
 *                     n = 0
 */
      __pyx_t_15 = __pyx_v_n_trees;
      __pyx_t_16 = __pyx_t_15;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_16; __pyx_t_22+=1) {
        __pyx_v_l = __pyx_t_22;

        /* "cyext_acv.pyx":635
 *         for b in range(n_trees):
 *             for l in range(n_trees):
 *                 if b == l:             # <<<<<<<<<<<<<<
 *                     n = 0
 *                     n_u = 0
 */
        __pyx_t_17 = ((__pyx_v_b == __pyx_v_l) != 0);
        if (__pyx_t_17) {

          /* "cyext_acv.pyx":636
 *             for l in range(n_trees):
 *                 if b == l:
 *                     n = 0             # <<<<<<<<<<<<<<
 *                     n_u = 0
 *                     n_d = 0
 */
          __pyx_v_n = 0.0;

          /* "cyext_acv.pyx":637
 *                 if b == l:
 *                     n = 0
 *                     n_u = 0             # <<<<<<<<<<<<<<
 *                     n_d = 0
 *                     nb_leaf = leaves_nb[b]
 */
          __pyx_v_n_u = 0.0;

          /* "cyext_acv.pyx":638
 *                     n = 0
 *                     n_u = 0
 *                     n_d = 0             # <<<<<<<<<<<<<<
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):
 */
          __pyx_v_n_d = 0.0;

          /* "cyext_acv.pyx":639
 *                     n_u = 0
 *                     n_d = 0
 *                     nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, leaf_numb, i]
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_24 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":640
 *                     n_d = 0
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         n += p_n[b, leaf_numb, leaf_numb, i]
 *                         n_u += p_u_n[b, leaf_numb, leaf_numb, i]
 */
          __pyx_t_37 = __pyx_v_nb_leaf;
          __pyx_t_33 = __pyx_t_37;
          for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
            __pyx_v_leaf_numb = __pyx_t_34;

            /* "cyext_acv.pyx":641
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                         n_u += p_u_n[b, leaf_numb, leaf_numb, i]
 *                         n_d += p_d_n[b, leaf_numb, leaf_numb, i]
 */
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_19 = __pyx_v_leaf_numb;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_26 = __pyx_v_i;
            __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_24 * __pyx_v_p_n.strides[0]) ) + __pyx_t_19 * __pyx_v_p_n.strides[1]) ) + __pyx_t_27 * __pyx_v_p_n.strides[2]) ) + __pyx_t_26 * __pyx_v_p_n.strides[3]) ))));

            /* "cyext_acv.pyx":642
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, leaf_numb, i]
 *                         n_u += p_u_n[b, leaf_numb, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                         n_d += p_d_n[b, leaf_numb, leaf_numb, i]
 *                     norm[b] = n
 */
            __pyx_t_26 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_19 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_26 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_19 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[3]) ))));

            /* "cyext_acv.pyx":643
 *                         n += p_n[b, leaf_numb, leaf_numb, i]
 *                         n_u += p_u_n[b, leaf_numb, leaf_numb, i]
 *                         n_d += p_d_n[b, leaf_numb, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     norm[b] = n
 *                     norm_u[b] = n_u
 */
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_19 = __pyx_v_leaf_numb;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_26 = __pyx_v_i;
            __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_24 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_19 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_26 * __pyx_v_p_d_n.strides[3]) ))));
          }

          /* "cyext_acv.pyx":644
 *                         n_u += p_u_n[b, leaf_numb, leaf_numb, i]
 *                         n_d += p_d_n[b, leaf_numb, leaf_numb, i]
 *                     norm[b] = n             # <<<<<<<<<<<<<<
 *                     norm_u[b] = n_u
 *                     norm_d[b] = n_d
 */
          __pyx_t_26 = __pyx_v_b;
          *((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_26 * __pyx_v_norm.strides[0]) )) = __pyx_v_n;

          /* "cyext_acv.pyx":645
 *                         n_d += p_d_n[b, leaf_numb, leaf_numb, i]
 *                     norm[b] = n
 *                     norm_u[b] = n_u             # <<<<<<<<<<<<<<
 *                     norm_d[b] = n_d
 *                     for leaf_numb in range(nb_leaf):
 */
          __pyx_t_26 = __pyx_v_b;
          *((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_26 * __pyx_v_norm_u.strides[0]) )) = __pyx_v_n_u;

          /* "cyext_acv.pyx":646
 *                     norm[b] = n
 *                     norm_u[b] = n_u
 *                     norm_d[b] = n_d             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  i, 0]/norm[b] if norm[b] !=0 else 0
 */
          __pyx_t_26 = __pyx_v_b;
          *((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_26 * __pyx_v_norm_d.strides[0]) )) = __pyx_v_n_d;

          /* "cyext_acv.pyx":647
 *                     norm_u[b] = n_u
 *                     norm_d[b] = n_d
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  i, 0]/norm[b] if norm[b] !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, leaf_numb, i, 1]/norm_u[b] if norm_u[b] !=0 else 0
 */
          __pyx_t_37 = __pyx_v_nb_leaf;
          __pyx_t_33 = __pyx_t_37;
          for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
            __pyx_v_leaf_numb = __pyx_t_34;

            /* "cyext_acv.pyx":648
 *                     norm_d[b] = n_d
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  i, 0]/norm[b] if norm[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_u += mean_forest_b[b, leaf_numb, leaf_numb, i, 1]/norm_u[b] if norm_u[b] !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, leaf_numb, i, 2]/norm_d[b] if norm_d[b] !=0 else 0
 */
            __pyx_t_26 = __pyx_v_b;
            if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_26 * __pyx_v_norm.strides[0]) ))) != 0.0) != 0)) {
              __pyx_t_27 = __pyx_v_b;
              __pyx_t_19 = __pyx_v_leaf_numb;
              __pyx_t_24 = __pyx_v_leaf_numb;
              __pyx_t_25 = __pyx_v_i;
              __pyx_t_36 = 0;
              __pyx_t_44 = __pyx_v_b;
              __pyx_t_3 = PyFloat_FromDouble(((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_27 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_36 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_44 * __pyx_v_norm.strides[0]) ))))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_1 = __pyx_t_3;
              __pyx_t_3 = 0;
            } else {
              __Pyx_INCREF(__pyx_int_0);
              __pyx_t_1 = __pyx_int_0;
            }
            __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_ss_a, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v_ss_a, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "cyext_acv.pyx":649
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  i, 0]/norm[b] if norm[b] !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, leaf_numb, i, 1]/norm_u[b] if norm_u[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_d += mean_forest_b[b, leaf_numb, leaf_numb, i, 2]/norm_d[b] if norm_d[b] !=0 else 0
 *                 else:
 */
            __pyx_t_26 = __pyx_v_b;
            if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_26 * __pyx_v_norm_u.strides[0]) ))) != 0.0) != 0)) {
              __pyx_t_44 = __pyx_v_b;
              __pyx_t_36 = __pyx_v_leaf_numb;
              __pyx_t_25 = __pyx_v_leaf_numb;
              __pyx_t_24 = __pyx_v_i;
              __pyx_t_19 = 1;
              __pyx_t_27 = __pyx_v_b;
              __pyx_t_1 = PyFloat_FromDouble(((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_44 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_36 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_27 * __pyx_v_norm_u.strides[0]) ))))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_3 = __pyx_t_1;
              __pyx_t_1 = 0;
            } else {
              __Pyx_INCREF(__pyx_int_0);
              __pyx_t_3 = __pyx_int_0;
            }
            __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_ss_u, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v_ss_u, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "cyext_acv.pyx":650
 *                         ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  i, 0]/norm[b] if norm[b] !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, leaf_numb, i, 1]/norm_u[b] if norm_u[b] !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, leaf_numb, i, 2]/norm_d[b] if norm_d[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     nb_leaf = leaves_nb[b]
 */
            __pyx_t_26 = __pyx_v_b;
            if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_26 * __pyx_v_norm_d.strides[0]) ))) != 0.0) != 0)) {
              __pyx_t_27 = __pyx_v_b;
              __pyx_t_19 = __pyx_v_leaf_numb;
              __pyx_t_24 = __pyx_v_leaf_numb;
              __pyx_t_25 = __pyx_v_i;
              __pyx_t_36 = 2;
              __pyx_t_44 = __pyx_v_b;
              __pyx_t_3 = PyFloat_FromDouble(((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_27 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_19 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_25 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_36 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_44 * __pyx_v_norm_d.strides[0]) ))))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_1 = __pyx_t_3;
              __pyx_t_3 = 0;
            } else {
              __Pyx_INCREF(__pyx_int_0);
              __pyx_t_1 = __pyx_int_0;
            }
            __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_ss_d, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 650, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v_ss_d, __pyx_t_3);
            __pyx_t_3 = 0;
          }

          /* "cyext_acv.pyx":635
 *         for b in range(n_trees):
 *             for l in range(n_trees):
 *                 if b == l:             # <<<<<<<<<<<<<<
 *                     n = 0
 *                     n_u = 0
 */
          goto __pyx_L112;
        }

        /* "cyext_acv.pyx":652
 *                         ss_d += mean_forest_b[b, leaf_numb, leaf_numb, i, 2]/norm_d[b] if norm_d[b] !=0 else 0
 *                 else:
 *                     nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                     nb_leaf_l = leaves_nb[l]
 * 
 */
        /*else*/ {
          __pyx_t_26 = __pyx_v_b;
          __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_26 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":653
 *                 else:
 *                     nb_leaf = leaves_nb[b]
 *                     nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                     for leaf_numb in range(nb_leaf):
 */
          __pyx_t_26 = __pyx_v_l;
          __pyx_v_nb_leaf_l = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_26 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":655
 *                     nb_leaf_l = leaves_nb[l]
 * 
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         n = 0
 *                         n_u = 0
 */
          __pyx_t_37 = __pyx_v_nb_leaf;
          __pyx_t_33 = __pyx_t_37;
          for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
            __pyx_v_leaf_numb = __pyx_t_34;

            /* "cyext_acv.pyx":656
 * 
 *                     for leaf_numb in range(nb_leaf):
 *                         n = 0             # <<<<<<<<<<<<<<
 *                         n_u = 0
 *                         n_d = 0
 */
            __pyx_v_n = 0.0;

            /* "cyext_acv.pyx":657
 *                     for leaf_numb in range(nb_leaf):
 *                         n = 0
 *                         n_u = 0             # <<<<<<<<<<<<<<
 *                         n_d = 0
 *                         for leaf_numb_l in range(nb_leaf_l):
 */
            __pyx_v_n_u = 0.0;

            /* "cyext_acv.pyx":658
 *                         n = 0
 *                         n_u = 0
 *                         n_d = 0             # <<<<<<<<<<<<<<
 *                         for leaf_numb_l in range(nb_leaf_l):
 *                             n += p_n[b, leaf_numb, leaf_numb_l, i]
 */
            __pyx_v_n_d = 0.0;

            /* "cyext_acv.pyx":659
 *                         n_u = 0
 *                         n_d = 0
 *                         for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 *                             n += p_n[b, leaf_numb, leaf_numb_l, i]
 *                             n_u += p_u_n[b, leaf_numb, leaf_numb_l, i]
 */
            __pyx_t_39 = __pyx_v_nb_leaf_l;
            __pyx_t_38 = __pyx_t_39;
            for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_38; __pyx_t_41+=1) {
              __pyx_v_leaf_numb_l = __pyx_t_41;

              /* "cyext_acv.pyx":660
 *                         n_d = 0
 *                         for leaf_numb_l in range(nb_leaf_l):
 *                             n += p_n[b, leaf_numb, leaf_numb_l, i]             # <<<<<<<<<<<<<<
 *                             n_u += p_u_n[b, leaf_numb, leaf_numb_l, i]
 *                             n_d += p_d_n[b, leaf_numb, leaf_numb_l, i]
 */
              __pyx_t_26 = __pyx_v_b;
              __pyx_t_44 = __pyx_v_leaf_numb;
              __pyx_t_36 = __pyx_v_leaf_numb_l;
              __pyx_t_25 = __pyx_v_i;
              __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_26 * __pyx_v_p_n.strides[0]) ) + __pyx_t_44 * __pyx_v_p_n.strides[1]) ) + __pyx_t_36 * __pyx_v_p_n.strides[2]) ) + __pyx_t_25 * __pyx_v_p_n.strides[3]) ))));

              /* "cyext_acv.pyx":661
 *                         for leaf_numb_l in range(nb_leaf_l):
 *                             n += p_n[b, leaf_numb, leaf_numb_l, i]
 *                             n_u += p_u_n[b, leaf_numb, leaf_numb_l, i]             # <<<<<<<<<<<<<<
 *                             n_d += p_d_n[b, leaf_numb, leaf_numb_l, i]
 * 
 */
              __pyx_t_25 = __pyx_v_b;
              __pyx_t_36 = __pyx_v_leaf_numb;
              __pyx_t_44 = __pyx_v_leaf_numb_l;
              __pyx_t_26 = __pyx_v_i;
              __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_25 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_36 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_44 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_26 * __pyx_v_p_u_n.strides[3]) ))));

              /* "cyext_acv.pyx":662
 *                             n += p_n[b, leaf_numb, leaf_numb_l, i]
 *                             n_u += p_u_n[b, leaf_numb, leaf_numb_l, i]
 *                             n_d += p_d_n[b, leaf_numb, leaf_numb_l, i]             # <<<<<<<<<<<<<<
 * 
 *                         for leaf_numb_l in range(nb_leaf_l):
 */
              __pyx_t_26 = __pyx_v_b;
              __pyx_t_44 = __pyx_v_leaf_numb;
              __pyx_t_36 = __pyx_v_leaf_numb_l;
              __pyx_t_25 = __pyx_v_i;
              __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_26 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_44 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_36 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_25 * __pyx_v_p_d_n.strides[3]) ))));
            }

            /* "cyext_acv.pyx":664
 *                             n_d += p_d_n[b, leaf_numb, leaf_numb_l, i]
 * 
 *                         for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 *                             ss_a += (p_n[b, leaf_numb, leaf_numb, i]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  i, 0]/n) if n*norm[b] !=0  else 0
 *                             ss_u += (p_u_n[b, leaf_numb, leaf_numb, i]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1]/n_u) if n_u*norm_u[b] !=0 else 0
 */
            __pyx_t_39 = __pyx_v_nb_leaf_l;
            __pyx_t_38 = __pyx_t_39;
            for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_38; __pyx_t_41+=1) {
              __pyx_v_leaf_numb_l = __pyx_t_41;

              /* "cyext_acv.pyx":665
 * 
 *                         for leaf_numb_l in range(nb_leaf_l):
 *                             ss_a += (p_n[b, leaf_numb, leaf_numb, i]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  i, 0]/n) if n*norm[b] !=0  else 0             # <<<<<<<<<<<<<<
 *                             ss_u += (p_u_n[b, leaf_numb, leaf_numb, i]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1]/n_u) if n_u*norm_u[b] !=0 else 0
 *                             ss_d += (p_d_n[b, leaf_numb, leaf_numb, i]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2]/n_d) if n_d*norm_d[b] !=0 else 0
 */
              __pyx_t_25 = __pyx_v_b;
              if ((((__pyx_v_n * (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_25 * __pyx_v_norm.strides[0]) )))) != 0.0) != 0)) {
                __pyx_t_36 = __pyx_v_b;
                __pyx_t_44 = __pyx_v_leaf_numb;
                __pyx_t_26 = __pyx_v_leaf_numb;
                __pyx_t_24 = __pyx_v_i;
                __pyx_t_19 = __pyx_v_b;
                __pyx_t_27 = __pyx_v_b;
                __pyx_t_45 = __pyx_v_leaf_numb;
                __pyx_t_46 = __pyx_v_leaf_numb_l;
                __pyx_t_47 = __pyx_v_i;
                __pyx_t_48 = 0;
                __pyx_t_1 = PyFloat_FromDouble((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_36 * __pyx_v_p_n.strides[0]) ) + __pyx_t_44 * __pyx_v_p_n.strides[1]) ) + __pyx_t_26 * __pyx_v_p_n.strides[2]) ) + __pyx_t_24 * __pyx_v_p_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_19 * __pyx_v_norm.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_27 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_3 = __pyx_t_1;
                __pyx_t_1 = 0;
              } else {
                __Pyx_INCREF(__pyx_int_0);
                __pyx_t_3 = __pyx_int_0;
              }
              __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_ss_a, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF_SET(__pyx_v_ss_a, __pyx_t_1);
              __pyx_t_1 = 0;

              /* "cyext_acv.pyx":666
 *                         for leaf_numb_l in range(nb_leaf_l):
 *                             ss_a += (p_n[b, leaf_numb, leaf_numb, i]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  i, 0]/n) if n*norm[b] !=0  else 0
 *                             ss_u += (p_u_n[b, leaf_numb, leaf_numb, i]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1]/n_u) if n_u*norm_u[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                             ss_d += (p_d_n[b, leaf_numb, leaf_numb, i]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2]/n_d) if n_d*norm_d[b] !=0 else 0
 * 
 */
              __pyx_t_25 = __pyx_v_b;
              if ((((__pyx_v_n_u * (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_25 * __pyx_v_norm_u.strides[0]) )))) != 0.0) != 0)) {
                __pyx_t_48 = __pyx_v_b;
                __pyx_t_47 = __pyx_v_leaf_numb;
                __pyx_t_46 = __pyx_v_leaf_numb;
                __pyx_t_45 = __pyx_v_i;
                __pyx_t_27 = __pyx_v_b;
                __pyx_t_19 = __pyx_v_b;
                __pyx_t_24 = __pyx_v_leaf_numb;
                __pyx_t_26 = __pyx_v_leaf_numb_l;
                __pyx_t_44 = __pyx_v_i;
                __pyx_t_36 = 1;
                __pyx_t_3 = PyFloat_FromDouble((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_48 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_47 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_46 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_45 * __pyx_v_p_u_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_27 * __pyx_v_norm_u.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_19 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_24 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_26 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_36 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n_u))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_1 = __pyx_t_3;
                __pyx_t_3 = 0;
              } else {
                __Pyx_INCREF(__pyx_int_0);
                __pyx_t_1 = __pyx_int_0;
              }
              __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_ss_u, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF_SET(__pyx_v_ss_u, __pyx_t_3);
              __pyx_t_3 = 0;

              /* "cyext_acv.pyx":667
 *                             ss_a += (p_n[b, leaf_numb, leaf_numb, i]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  i, 0]/n) if n*norm[b] !=0  else 0
 *                             ss_u += (p_u_n[b, leaf_numb, leaf_numb, i]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 1]/n_u) if n_u*norm_u[b] !=0 else 0
 *                             ss_d += (p_d_n[b, leaf_numb, leaf_numb, i]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, i, 2]/n_d) if n_d*norm_d[b] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
              __pyx_t_25 = __pyx_v_b;
              if ((((__pyx_v_n_d * (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_25 * __pyx_v_norm_d.strides[0]) )))) != 0.0) != 0)) {
                __pyx_t_36 = __pyx_v_b;
                __pyx_t_44 = __pyx_v_leaf_numb;
                __pyx_t_26 = __pyx_v_leaf_numb;
                __pyx_t_24 = __pyx_v_i;
                __pyx_t_19 = __pyx_v_b;
                __pyx_t_27 = __pyx_v_b;
                __pyx_t_45 = __pyx_v_leaf_numb;
                __pyx_t_46 = __pyx_v_leaf_numb_l;
                __pyx_t_47 = __pyx_v_i;
                __pyx_t_48 = 2;
                __pyx_t_1 = PyFloat_FromDouble((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_36 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_44 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_26 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_24 * __pyx_v_p_d_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_19 * __pyx_v_norm_d.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_27 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n_d))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_3 = __pyx_t_1;
                __pyx_t_1 = 0;
              } else {
                __Pyx_INCREF(__pyx_int_0);
                __pyx_t_3 = __pyx_int_0;
              }
              __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_ss_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF_SET(__pyx_v_ss_d, __pyx_t_1);
              __pyx_t_1 = 0;
            }
          }
        }
        __pyx_L112:;
      }
    }

    /* "cyext_acv.pyx":670
 * 
 * 
 *         sdp[i] = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *     return np.array(sdp)
 */
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_ss_u, __pyx_v_ss_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_17) {
      __pyx_t_3 = PyNumber_Subtract(__pyx_v_ss_u, __pyx_v_ss_a); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_ss_u, __pyx_v_ss_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_49 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_49 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 670, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_40 = __pyx_t_49;
    } else {
      __pyx_t_40 = 0.0;
    }
    __pyx_t_25 = __pyx_v_i;
    *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_25 * __pyx_v_sdp.strides[0]) )) = __pyx_t_40;
  }

  /* "cyext_acv.pyx":672
 *         sdp[i] = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 *     return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":464
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_reg(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_23, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_28, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm_u, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm_d, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __Pyx_XDECREF(__pyx_v_ss_u);
  __Pyx_XDECREF(__pyx_v_ss_d);
  __Pyx_XDECREF(__pyx_v_ss_a);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_13compute_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_13compute_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_tX;
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_reg (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_tX,&__pyx_n_s_y_pred,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 1); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 2); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 3); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 4); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 5); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 6); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 7); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 8); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 9); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 10); __PYX_ERR(0, 464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, 11); __PYX_ERR(0, 464, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_reg") < 0)) __PYX_ERR(0, 464, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_tX = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tX == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 466, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 466, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 467, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 467, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[10], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 467, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_reg", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_12compute_sdp_reg(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_12compute_sdp_reg(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_reg", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 464, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 464, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_reg(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":679
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_reg_cat(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_15compute_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_reg_cat(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_S_size;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_leaves_tree = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_tree_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_value;
  double __pyx_v_value_l;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_nb_leaf_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_down;
  unsigned int __pyx_v_up;
  int __pyx_v_b;
  int __pyx_v_l;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  int __pyx_v_leaf_numb_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  __Pyx_memviewslice __pyx_t_24 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  unsigned int __pyx_t_35;
  int __pyx_t_36;
  double __pyx_t_37;
  unsigned int __pyx_t_38;
  unsigned int __pyx_t_39;
  int __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_reg_cat", 0);

  /* "cyext_acv.pyx":684
 *             long[:, :] leaf_idx_trees, long[:] leaves_nb, double[:] scaling, int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":685
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":686
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_S_size = (__pyx_v_S.shape[0]);

  /* "cyext_acv.pyx":687
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int S_size = S.shape[0]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :, :] leaves_tree_l
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":696
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":699
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm, lm_u, lm_d, lm_s
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":706
 *     cdef int b, l, leaf_numb, i, s, j, leaf_numb_l
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for l in range(n_trees):
 *             if b == l:
 */
  __pyx_t_7 = __pyx_v_n_trees;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_b = __pyx_t_9;

    /* "cyext_acv.pyx":707
 * 
 *     for b in range(n_trees):
 *         for l in range(n_trees):             # <<<<<<<<<<<<<<
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]
 */
    __pyx_t_10 = __pyx_v_n_trees;
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_l = __pyx_t_12;

      /* "cyext_acv.pyx":708
 *     for b in range(n_trees):
 *         for l in range(n_trees):
 *             if b == l:             # <<<<<<<<<<<<<<
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 */
      __pyx_t_13 = ((__pyx_v_b == __pyx_v_l) != 0);
      if (__pyx_t_13) {

        /* "cyext_acv.pyx":709
 *         for l in range(n_trees):
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 * 
 */
        __pyx_t_14.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_14.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_14.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_14.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_14.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_14.suboffsets[1] = -1;

__pyx_t_14.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_14.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_14.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
        __pyx_v_leaves_tree = __pyx_t_14;
        __pyx_t_14.memview = NULL;
        __pyx_t_14.data = NULL;

        /* "cyext_acv.pyx":710
 *             if b == l:
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_t_15 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_15 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":712
 *                 nb_leaf = leaves_nb[b]
 * 
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     leaf_part = leaves_tree[leaf_numb]
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
        __pyx_t_16 = __pyx_v_nb_leaf;
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
          __pyx_v_leaf_numb = __pyx_t_18;

          /* "cyext_acv.pyx":713
 * 
 *                 for leaf_numb in range(nb_leaf):
 *                     leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 */
          __pyx_t_5.data = __pyx_v_leaves_tree.data;
          __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
          __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
          {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
          __pyx_v_leaf_part = __pyx_t_5;
          __pyx_t_5.memview = NULL;
          __pyx_t_5.data = NULL;

          /* "cyext_acv.pyx":714
 *                 for leaf_numb in range(nb_leaf):
 *                     leaf_part = leaves_tree[leaf_numb]
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 * 
 *                     lm = np.zeros(data.shape[0], dtype=np.int)
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_19 = __pyx_v_leaf_numb;
          __pyx_t_20 = __pyx_v_b;
          __pyx_t_21 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_15 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[1]) )));
          __pyx_t_22 = 0;
          __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) ) + __pyx_t_21 * __pyx_v_values.strides[1]) ) + __pyx_t_22 * __pyx_v_values.strides[2]) )));

          /* "cyext_acv.pyx":716
 *                     value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 *                     lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_23) < 0) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_24 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_23, PyBUF_WRITABLE); if (unlikely(!__pyx_t_24.memview)) __PYX_ERR(0, 716, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
          __pyx_v_lm = __pyx_t_24;
          __pyx_t_24.memview = NULL;
          __pyx_t_24.data = NULL;

          /* "cyext_acv.pyx":718
 *                     lm = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         a_it = 0
 *                         b_it = 0
 */
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              __Pyx_FastGIL_Remember();
              #endif
              /*try:*/ {
                if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 718, __pyx_L13_error) }
                __pyx_t_25 = (__pyx_v_data.shape[0]);
                if ((1 == 0)) abort();
                {
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   (x)
                        #define unlikely(x) (x)
                    #endif
                    __pyx_t_27 = (__pyx_t_25 - 0 + 1 - 1/abs(1)) / 1;
                    if (__pyx_t_27 > 0)
                    {
                        #ifdef _OPENMP
                        #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_13, __pyx_t_15, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32)
                        #endif /* _OPENMP */
                        {
                            #ifdef _OPENMP
                            #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                            #endif /* _OPENMP */
                            for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_27; __pyx_t_26++){
                                {
                                    __pyx_v_i = (int)(0 + 1 * __pyx_t_26);
                                    /* Initialize private variables to invalid values */
                                    __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                    __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                    __pyx_v_s = ((int)0xbad0bad0);

                                    /* "cyext_acv.pyx":719
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                         a_it = 0             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(m):
 */
                                    __pyx_v_a_it = 0;

                                    /* "cyext_acv.pyx":720
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                         a_it = 0
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":721
 *                         a_it = 0
 *                         b_it = 0
 *                         for s in range(m):             # <<<<<<<<<<<<<<
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1
 */
                                    __pyx_t_28 = __pyx_v_m;
                                    __pyx_t_29 = __pyx_t_28;
                                    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                                      __pyx_v_s = __pyx_t_30;

                                      /* "cyext_acv.pyx":722
 *                         b_it = 0
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 */
                                      __pyx_t_19 = __pyx_v_i;
                                      __pyx_t_15 = __pyx_v_s;
                                      __pyx_t_22 = __pyx_v_s;
                                      __pyx_t_21 = 1;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_15 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_22 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_21 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (__pyx_t_31) {
                                      } else {
                                        __pyx_t_13 = __pyx_t_31;
                                        goto __pyx_L22_bool_binop_done;
                                      }
                                      __pyx_t_21 = __pyx_v_i;
                                      __pyx_t_22 = __pyx_v_s;
                                      __pyx_t_15 = __pyx_v_s;
                                      __pyx_t_19 = 0;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_21 * __pyx_v_data.strides[0]) ) + __pyx_t_22 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_15 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_13 = __pyx_t_31;
                                      __pyx_L22_bool_binop_done:;
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":723
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 */
                                        __pyx_v_a_it = (__pyx_v_a_it + 1);

                                        /* "cyext_acv.pyx":722
 *                         b_it = 0
 *                         for s in range(m):
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":724
 *                             if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_28 = __pyx_v_S_size;
                                    __pyx_t_29 = __pyx_t_28;
                                    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                                      __pyx_v_s = __pyx_t_30;

                                      /* "cyext_acv.pyx":725
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_15 = __pyx_v_i;
                                      __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_19 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_21 = __pyx_v_s;
                                      __pyx_t_20 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_32 = 1;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_15 * __pyx_v_data.strides[0]) ) + __pyx_t_22 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_20 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_32 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (__pyx_t_31) {
                                      } else {
                                        __pyx_t_13 = __pyx_t_31;
                                        goto __pyx_L27_bool_binop_done;
                                      }
                                      __pyx_t_21 = __pyx_v_s;
                                      __pyx_t_32 = __pyx_v_i;
                                      __pyx_t_20 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_19 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_15 = 0;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_32 * __pyx_v_data.strides[0]) ) + __pyx_t_20 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_22 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_15 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_13 = __pyx_t_31;
                                      __pyx_L27_bool_binop_done:;
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":726
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if a_it == m:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":725
 *                                 a_it = a_it + 1
 *                         for s in range(S_size):
 *                             if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":728
 *                                 b_it = b_it + 1
 * 
 *                         if a_it == m:             # <<<<<<<<<<<<<<
 *                             lm[i] = 1
 * 
 */
                                    __pyx_t_13 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                    if (__pyx_t_13) {

                                      /* "cyext_acv.pyx":729
 * 
 *                         if a_it == m:
 *                             lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                     for i in prange(N, nogil=True):
 */
                                      __pyx_t_19 = __pyx_v_i;
                                      *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_19 * __pyx_v_lm.strides[0]) )) = 1;

                                      /* "cyext_acv.pyx":728
 *                                 b_it = b_it + 1
 * 
 *                         if a_it == m:             # <<<<<<<<<<<<<<
 *                             lm[i] = 1
 * 
 */
                                    }
                                }
                            }
                        }
                    }
                }
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   __builtin_expect(!!(x), 1)
                    #define unlikely(x) __builtin_expect(!!(x), 0)
                #endif
              }

              /* "cyext_acv.pyx":718
 *                     lm = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                     for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                         a_it = 0
 *                         b_it = 0
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L14;
                }
                __pyx_L13_error: {
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L1_error;
                }
                __pyx_L14:;
              }
          }

          /* "cyext_acv.pyx":731
 *                             lm[i] = 1
 * 
 *                     for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
          {
              #ifdef WITH_THREAD
              PyThreadState *_save;
              Py_UNBLOCK_THREADS
              __Pyx_FastGIL_Remember();
              #endif
              /*try:*/ {
                __pyx_t_28 = __pyx_v_N;
                if ((1 == 0)) abort();
                {
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   (x)
                        #define unlikely(x) (x)
                    #endif
                    __pyx_t_33 = (__pyx_t_28 - 0 + 1 - 1/abs(1)) / 1;
                    if (__pyx_t_33 > 0)
                    {
                        #ifdef _OPENMP
                        #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_13, __pyx_t_15, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_26, __pyx_t_27, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37)
                        #endif /* _OPENMP */
                        {
                            #ifdef _OPENMP
                            #pragma omp for lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_down) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) lastprivate(__pyx_v_up)
                            #endif /* _OPENMP */
                            for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_33; __pyx_t_29++){
                                {
                                    __pyx_v_i = (int)(0 + 1 * __pyx_t_29);
                                    /* Initialize private variables to invalid values */
                                    __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                    __pyx_v_down = ((unsigned int)0xbad0bad0);
                                    __pyx_v_j = ((int)0xbad0bad0);
                                    __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                    __pyx_v_s = ((int)0xbad0bad0);
                                    __pyx_v_up = ((unsigned int)0xbad0bad0);

                                    /* "cyext_acv.pyx":732
 * 
 *                     for i in prange(N, nogil=True):
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":733
 *                     for i in prange(N, nogil=True):
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_34 = __pyx_v_S_size;
                                    __pyx_t_35 = __pyx_t_34;
                                    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_35; __pyx_t_30+=1) {
                                      __pyx_v_s = __pyx_t_30;

                                      /* "cyext_acv.pyx":734
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_15 = __pyx_v_i;
                                      __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_19 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_21 = __pyx_v_s;
                                      __pyx_t_20 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_32 = 1;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_15 * __pyx_v_X.strides[0]) ) + __pyx_t_22 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_20 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_32 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      if (!__pyx_t_31) {
                                      } else {
                                        __pyx_t_13 = __pyx_t_31;
                                        goto __pyx_L44_bool_binop_done;
                                      }
                                      __pyx_t_21 = __pyx_v_s;
                                      __pyx_t_32 = __pyx_v_i;
                                      __pyx_t_20 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_19 = __pyx_v_s;
                                      __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_19 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_15 = 0;
                                      __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_32 * __pyx_v_X.strides[0]) ) + __pyx_t_20 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_22 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_15 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                      __pyx_t_13 = __pyx_t_31;
                                      __pyx_L44_bool_binop_done:;
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":735
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                         if o_all > 0:
 *                             continue
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":734
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":736
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    __pyx_t_13 = ((__pyx_v_o_all > 0) != 0);
                                    if (__pyx_t_13) {

                                      /* "cyext_acv.pyx":737
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         p = 0
 */
                                      goto __pyx_L37_continue;

                                      /* "cyext_acv.pyx":736
 *                             if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":739
 *                             continue
 * 
 *                         p = 0             # <<<<<<<<<<<<<<
 *                         p_u = 0
 *                         p_d = 0
 */
                                    __pyx_v_p = 0;

                                    /* "cyext_acv.pyx":740
 * 
 *                         p = 0
 *                         p_u = 0             # <<<<<<<<<<<<<<
 *                         p_d = 0
 *                         p_s = 0
 */
                                    __pyx_v_p_u = 0;

                                    /* "cyext_acv.pyx":741
 *                         p = 0
 *                         p_u = 0
 *                         p_d = 0             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_su = 0
 */
                                    __pyx_v_p_d = 0;

                                    /* "cyext_acv.pyx":742
 *                         p_u = 0
 *                         p_d = 0
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_su = 0
 *                         p_sd = 0
 */
                                    __pyx_v_p_s = 0;

                                    /* "cyext_acv.pyx":743
 *                         p_d = 0
 *                         p_s = 0
 *                         p_su = 0             # <<<<<<<<<<<<<<
 *                         p_sd = 0
 * 
 */
                                    __pyx_v_p_su = 0;

                                    /* "cyext_acv.pyx":744
 *                         p_s = 0
 *                         p_su = 0
 *                         p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                         for j in range(data.shape[0]):
 */
                                    __pyx_v_p_sd = 0;

                                    /* "cyext_acv.pyx":746
 *                         p_sd = 0
 * 
 *                         for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                 up = 0
 *                                 down = 0
 */
                                    __pyx_t_27 = (__pyx_v_data.shape[0]);
                                    __pyx_t_26 = __pyx_t_27;
                                    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_26; __pyx_t_30+=1) {
                                      __pyx_v_j = __pyx_t_30;

                                      /* "cyext_acv.pyx":747
 * 
 *                         for j in range(data.shape[0]):
 *                                 up = 0             # <<<<<<<<<<<<<<
 *                                 down = 0
 *                                 b_it = 0
 */
                                      __pyx_v_up = 0;

                                      /* "cyext_acv.pyx":748
 *                         for j in range(data.shape[0]):
 *                                 up = 0
 *                                 down = 0             # <<<<<<<<<<<<<<
 *                                 b_it = 0
 *                                 for s in range(S_size):
 */
                                      __pyx_v_down = 0;

                                      /* "cyext_acv.pyx":749
 *                                 up = 0
 *                                 down = 0
 *                                 b_it = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":750
 *                                 down = 0
 *                                 b_it = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if data[j, S[s]] == X[i, S[s]]:
 *                                         b_it = b_it + 1
 */
                                      __pyx_t_34 = __pyx_v_S_size;
                                      __pyx_t_35 = __pyx_t_34;
                                      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
                                        __pyx_v_s = __pyx_t_36;

                                        /* "cyext_acv.pyx":751
 *                                 b_it = 0
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_15 = __pyx_v_j;
                                        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_19 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_21 = __pyx_v_s;
                                        __pyx_t_20 = __pyx_v_i;
                                        __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_15 * __pyx_v_data.strides[0]) ) + __pyx_t_22 * __pyx_v_data.strides[1]) ))) == (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_20 * __pyx_v_X.strides[0]) ) + __pyx_t_32 * __pyx_v_X.strides[1]) )))) != 0);
                                        if (__pyx_t_13) {

                                          /* "cyext_acv.pyx":752
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:
 *                                         b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if b_it == S_size:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":751
 *                                 b_it = 0
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":754
 *                                         b_it = b_it + 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1
 */
                                      __pyx_t_13 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":755
 * 
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         up = 1
 *                                     else:
 */
                                        __pyx_t_21 = __pyx_v_i;
                                        __pyx_t_32 = __pyx_v_j;
                                        __pyx_t_20 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_j;
                                        __pyx_t_13 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_21 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_32 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_20 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_19 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                        if (__pyx_t_13) {

                                          /* "cyext_acv.pyx":756
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         down = 1
 */
                                          __pyx_v_up = 1;

                                          /* "cyext_acv.pyx":755
 * 
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         up = 1
 *                                     else:
 */
                                          goto __pyx_L53;
                                        }

                                        /* "cyext_acv.pyx":758
 *                                         up = 1
 *                                     else:
 *                                         down = 1             # <<<<<<<<<<<<<<
 * 
 *                                     p += lm[j]
 */
                                        /*else*/ {
                                          __pyx_v_down = 1;
                                        }
                                        __pyx_L53:;

                                        /* "cyext_acv.pyx":760
 *                                         down = 1
 * 
 *                                     p += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down
 */
                                        __pyx_t_19 = __pyx_v_j;
                                        __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_19 * __pyx_v_lm.strides[0]) ))));

                                        /* "cyext_acv.pyx":761
 * 
 *                                     p += lm[j]
 *                                     p_u += lm[j] * up             # <<<<<<<<<<<<<<
 *                                     p_d += lm[j] * down
 *                                     p_s += 1
 */
                                        __pyx_t_19 = __pyx_v_j;
                                        __pyx_v_p_u = (__pyx_v_p_u + ((*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_19 * __pyx_v_lm.strides[0]) ))) * __pyx_v_up));

                                        /* "cyext_acv.pyx":762
 *                                     p += lm[j]
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down             # <<<<<<<<<<<<<<
 *                                     p_s += 1
 *                                     p_su += up
 */
                                        __pyx_t_19 = __pyx_v_j;
                                        __pyx_v_p_d = (__pyx_v_p_d + ((*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_19 * __pyx_v_lm.strides[0]) ))) * __pyx_v_down));

                                        /* "cyext_acv.pyx":763
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down
 *                                     p_s += 1             # <<<<<<<<<<<<<<
 *                                     p_su += up
 *                                     p_sd += down
 */
                                        __pyx_v_p_s = (__pyx_v_p_s + 1);

                                        /* "cyext_acv.pyx":764
 *                                     p_d += lm[j] * down
 *                                     p_s += 1
 *                                     p_su += up             # <<<<<<<<<<<<<<
 *                                     p_sd += down
 * 
 */
                                        __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_up);

                                        /* "cyext_acv.pyx":765
 *                                     p_s += 1
 *                                     p_su += up
 *                                     p_sd += down             # <<<<<<<<<<<<<<
 * 
 *                         mean_forest[i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 */
                                        __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_down);

                                        /* "cyext_acv.pyx":754
 *                                         b_it = b_it + 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":767
 *                                     p_sd += down
 * 
 *                         mean_forest[i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest[i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0
 *                         mean_forest[i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_19 = __pyx_v_i;
                                      __pyx_t_37 = ((((__pyx_v_p * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_s) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_19 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p) * __pyx_v_value) / __pyx_v_p_s));
                                    } else {
                                      __pyx_t_37 = 0.0;
                                    }
                                    __pyx_t_19 = __pyx_v_i;
                                    __pyx_t_20 = 0;
                                    *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_19 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_20 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;

                                    /* "cyext_acv.pyx":768
 * 
 *                         mean_forest[i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 *                         mean_forest[i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest[i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 *             else:
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_20 = __pyx_v_i;
                                      __pyx_t_37 = ((((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_su) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_20 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_u) * __pyx_v_value) / __pyx_v_p_su));
                                    } else {
                                      __pyx_t_37 = 0.0;
                                    }
                                    __pyx_t_20 = __pyx_v_i;
                                    __pyx_t_19 = 1;
                                    *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_20 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_19 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;

                                    /* "cyext_acv.pyx":769
 *                         mean_forest[i, 0] += (p * value*value) / (p_s) - (2 * fX[i] * p * value)/(p_s) if p_s != 0 else 0
 *                         mean_forest[i, 1] += (p_u * value*value) / (p_su) - (2 * fX[i] * p_u * value)/(p_su) if p_su != 0 else 0
 *                         mean_forest[i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_19 = __pyx_v_i;
                                      __pyx_t_37 = ((((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_sd) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_19 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_d) * __pyx_v_value) / __pyx_v_p_sd));
                                    } else {
                                      __pyx_t_37 = 0.0;
                                    }
                                    __pyx_t_19 = __pyx_v_i;
                                    __pyx_t_20 = 2;
                                    *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_19 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_20 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;
                                    goto __pyx_L55;
                                    __pyx_L37_continue:;
                                    goto __pyx_L55;
                                    __pyx_L55:;
                                }
                            }
                        }
                    }
                }
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   __builtin_expect(!!(x), 1)
                    #define unlikely(x) __builtin_expect(!!(x), 0)
                #endif
              }

              /* "cyext_acv.pyx":731
 *                             lm[i] = 1
 * 
 *                     for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
              /*finally:*/ {
                /*normal exit:*/{
                  #ifdef WITH_THREAD
                  __Pyx_FastGIL_Forget();
                  Py_BLOCK_THREADS
                  #endif
                  goto __pyx_L36;
                }
                __pyx_L36:;
              }
          }
        }

        /* "cyext_acv.pyx":708
 *     for b in range(n_trees):
 *         for l in range(n_trees):
 *             if b == l:             # <<<<<<<<<<<<<<
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 */
        goto __pyx_L7;
      }

      /* "cyext_acv.pyx":771
 *                         mean_forest[i, 2] += (p_d * value*value) / (p_sd) - (2 * fX[i] * p_d * value)/(p_sd) if p_sd != 0 else 0
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]
 */
      /*else*/ {
        __pyx_t_14.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_14.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_14.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_14.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_14.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_14.suboffsets[1] = -1;

__pyx_t_14.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_14.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_14.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
        __pyx_v_leaves_tree = __pyx_t_14;
        __pyx_t_14.memview = NULL;
        __pyx_t_14.data = NULL;

        /* "cyext_acv.pyx":772
 *             else:
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 leaves_tree_l = partition_leaves_trees[l]
 *                 nb_leaf_l = leaves_nb[l]
 */
        __pyx_t_20 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_20 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":773
 *                 leaves_tree = partition_leaves_trees[b]
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]             # <<<<<<<<<<<<<<
 *                 nb_leaf_l = leaves_nb[l]
 * 
 */
        __pyx_t_14.data = __pyx_v_partition_leaves_trees.data;
        __pyx_t_14.memview = __pyx_v_partition_leaves_trees.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_14.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_14.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_14.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_14.suboffsets[1] = -1;

__pyx_t_14.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_14.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_14.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
        __pyx_v_leaves_tree_l = __pyx_t_14;
        __pyx_t_14.memview = NULL;
        __pyx_t_14.data = NULL;

        /* "cyext_acv.pyx":774
 *                 nb_leaf = leaves_nb[b]
 *                 leaves_tree_l = partition_leaves_trees[l]
 *                 nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_t_20 = __pyx_v_l;
        __pyx_v_nb_leaf_l = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_20 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":776
 *                 nb_leaf_l = leaves_nb[l]
 * 
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for leaf_numb_l in range(nb_leaf_l):
 * 
 */
        __pyx_t_16 = __pyx_v_nb_leaf;
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
          __pyx_v_leaf_numb = __pyx_t_18;

          /* "cyext_acv.pyx":777
 * 
 *                 for leaf_numb in range(nb_leaf):
 *                     for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 * 
 *                         leaf_part = leaves_tree[leaf_numb]
 */
          __pyx_t_33 = __pyx_v_nb_leaf_l;
          __pyx_t_29 = __pyx_t_33;
          for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
            __pyx_v_leaf_numb_l = __pyx_t_30;

            /* "cyext_acv.pyx":779
 *                     for leaf_numb_l in range(nb_leaf_l):
 * 
 *                         leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
            __pyx_t_5.data = __pyx_v_leaves_tree.data;
            __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
            __pyx_v_leaf_part = __pyx_t_5;
            __pyx_t_5.memview = NULL;
            __pyx_t_5.data = NULL;

            /* "cyext_acv.pyx":780
 * 
 *                         leaf_part = leaves_tree[leaf_numb]
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]             # <<<<<<<<<<<<<<
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 */
            __pyx_t_5.data = __pyx_v_leaves_tree_l.data;
            __pyx_t_5.memview = __pyx_v_leaves_tree_l.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree_l.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree_l.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree_l.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree_l.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree_l.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
            __pyx_v_leaf_part_l = __pyx_t_5;
            __pyx_t_5.memview = NULL;
            __pyx_t_5.data = NULL;

            /* "cyext_acv.pyx":781
 *                         leaf_part = leaves_tree[leaf_numb]
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 */
            __pyx_t_20 = __pyx_v_b;
            __pyx_t_19 = __pyx_v_leaf_numb;
            __pyx_t_32 = __pyx_v_b;
            __pyx_t_21 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_20 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[1]) )));
            __pyx_t_22 = 0;
            __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_32 * __pyx_v_values.strides[0]) ) + __pyx_t_21 * __pyx_v_values.strides[1]) ) + __pyx_t_22 * __pyx_v_values.strides[2]) )));

            /* "cyext_acv.pyx":782
 *                         leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                         value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]             # <<<<<<<<<<<<<<
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)
 */
            __pyx_t_19 = __pyx_v_l;
            __pyx_t_20 = __pyx_v_leaf_numb_l;
            __pyx_t_22 = __pyx_v_l;
            __pyx_t_21 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_19 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_20 * __pyx_v_leaf_idx_trees.strides[1]) )));
            __pyx_t_32 = 0;
            __pyx_v_value_l = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_22 * __pyx_v_values.strides[0]) ) + __pyx_t_21 * __pyx_v_values.strides[1]) ) + __pyx_t_32 * __pyx_v_values.strides[2]) )));

            /* "cyext_acv.pyx":784
 *                         value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
            __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_np); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_23);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
            __pyx_t_23 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_23);
            __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_23);
            PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_23);
            __pyx_t_23 = 0;
            __pyx_t_23 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_23);
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PyDict_SetItem(__pyx_t_23, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_23); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
            __pyx_t_24 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_24.memview)) __PYX_ERR(0, 784, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
            __pyx_v_lm = __pyx_t_24;
            __pyx_t_24.memview = NULL;
            __pyx_t_24.data = NULL;

            /* "cyext_acv.pyx":785
 * 
 *                         lm = np.zeros(data.shape[0], dtype=np.int)
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
            __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_23);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
            __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_24 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_24.memview)) __PYX_ERR(0, 785, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
            __pyx_v_lm_s = __pyx_t_24;
            __pyx_t_24.memview = NULL;
            __pyx_t_24.data = NULL;

            /* "cyext_acv.pyx":787
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
            {
                #ifdef WITH_THREAD
                PyThreadState *_save;
                Py_UNBLOCK_THREADS
                __Pyx_FastGIL_Remember();
                #endif
                /*try:*/ {
                  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 787, __pyx_L63_error) }
                  __pyx_t_27 = (__pyx_v_data.shape[0]);
                  if ((1 == 0)) abort();
                  {
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   (x)
                          #define unlikely(x) (x)
                      #endif
                      __pyx_t_25 = (__pyx_t_27 - 0 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_25 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_13, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_28, __pyx_t_31, __pyx_t_32, __pyx_t_34, __pyx_t_36)
                          #endif /* _OPENMP */
                          {
                              #ifdef _OPENMP
                              #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                              #endif /* _OPENMP */
                              for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26++){
                                  {
                                      __pyx_v_i = (int)(0 + 1 * __pyx_t_26);
                                      /* Initialize private variables to invalid values */
                                      __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                      __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                      __pyx_v_s = ((int)0xbad0bad0);

                                      /* "cyext_acv.pyx":788
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             b_it = 0
 *                             for s in range(m):
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":789
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                             a_it = 0
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":790
 *                             a_it = 0
 *                             b_it = 0
 *                             for s in range(m):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_28 = __pyx_v_m;
                                      __pyx_t_34 = __pyx_t_28;
                                      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_34; __pyx_t_36+=1) {
                                        __pyx_v_s = __pyx_t_36;

                                        /* "cyext_acv.pyx":791
 *                             b_it = 0
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 * 
 */
                                        __pyx_t_20 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_32 = __pyx_v_s;
                                        __pyx_t_21 = 1;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_20 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_32 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_21 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L72_bool_binop_done;
                                        }
                                        __pyx_t_21 = __pyx_v_i;
                                        __pyx_t_32 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_20 = 0;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_21 * __pyx_v_data.strides[0]) ) + __pyx_t_32 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_19 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_20 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L72_bool_binop_done;
                                        }
                                        __pyx_t_20 = __pyx_v_i;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_32 = __pyx_v_s;
                                        __pyx_t_21 = 1;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_20 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_32 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_21 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        if (__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L72_bool_binop_done;
                                        }
                                        __pyx_t_21 = __pyx_v_i;
                                        __pyx_t_32 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_s;
                                        __pyx_t_20 = 0;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_21 * __pyx_v_data.strides[0]) ) + __pyx_t_32 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_19 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_20 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        __pyx_t_13 = __pyx_t_31;
                                        __pyx_L72_bool_binop_done:;
                                        if (__pyx_t_13) {

                                          /* "cyext_acv.pyx":792
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if a_it == m:
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":791
 *                             b_it = 0
 *                             for s in range(m):
 *                                 if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":794
 *                                     a_it = a_it + 1
 * 
 *                             if a_it == m:             # <<<<<<<<<<<<<<
 *                                 lm[i] = 1
 * 
 */
                                      __pyx_t_13 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":795
 * 
 *                             if a_it == m:
 *                                 lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                         for i in prange(N, nogil=True):
 */
                                        __pyx_t_20 = __pyx_v_i;
                                        *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_20 * __pyx_v_lm.strides[0]) )) = 1;

                                        /* "cyext_acv.pyx":794
 *                                     a_it = a_it + 1
 * 
 *                             if a_it == m:             # <<<<<<<<<<<<<<
 *                                 lm[i] = 1
 * 
 */
                                      }
                                  }
                              }
                          }
                      }
                  }
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   __builtin_expect(!!(x), 1)
                      #define unlikely(x) __builtin_expect(!!(x), 0)
                  #endif
                }

                /* "cyext_acv.pyx":787
 *                         lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                         for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L64;
                  }
                  __pyx_L63_error: {
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L1_error;
                  }
                  __pyx_L64:;
                }
            }

            /* "cyext_acv.pyx":797
 *                                 lm[i] = 1
 * 
 *                         for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                             o_all = 0
 *                             for s in range(S_size):
 */
            {
                #ifdef WITH_THREAD
                PyThreadState *_save;
                Py_UNBLOCK_THREADS
                __Pyx_FastGIL_Remember();
                #endif
                /*try:*/ {
                  __pyx_t_28 = __pyx_v_N;
                  if ((1 == 0)) abort();
                  {
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   (x)
                          #define unlikely(x) (x)
                      #endif
                      __pyx_t_35 = (__pyx_t_28 - 0 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_35 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_13, __pyx_t_15, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_25, __pyx_t_26, __pyx_t_31, __pyx_t_32, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40)
                          #endif /* _OPENMP */
                          {
                              #ifdef _OPENMP
                              #pragma omp for lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_down) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) lastprivate(__pyx_v_up)
                              #endif /* _OPENMP */
                              for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_35; __pyx_t_34++){
                                  {
                                      __pyx_v_i = (int)(0 + 1 * __pyx_t_34);
                                      /* Initialize private variables to invalid values */
                                      __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                      __pyx_v_down = ((unsigned int)0xbad0bad0);
                                      __pyx_v_j = ((int)0xbad0bad0);
                                      __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                      __pyx_v_s = ((int)0xbad0bad0);
                                      __pyx_v_up = ((unsigned int)0xbad0bad0);

                                      /* "cyext_acv.pyx":798
 * 
 *                         for i in prange(N, nogil=True):
 *                             o_all = 0             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 */
                                      __pyx_v_o_all = 0;

                                      /* "cyext_acv.pyx":799
 *                         for i in prange(N, nogil=True):
 *                             o_all = 0
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 */
                                      __pyx_t_38 = __pyx_v_S_size;
                                      __pyx_t_39 = __pyx_t_38;
                                      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_39; __pyx_t_36+=1) {
                                        __pyx_v_s = __pyx_t_36;

                                        /* "cyext_acv.pyx":800
 *                             o_all = 0
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 */
                                        __pyx_t_20 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_20 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_21 = __pyx_v_s;
                                        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_15 = 1;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_19 * __pyx_v_X.strides[0]) ) + __pyx_t_32 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_22 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_15 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (!__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L91_bool_binop_done;
                                        }
                                        __pyx_t_21 = __pyx_v_s;
                                        __pyx_t_15 = __pyx_v_i;
                                        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_20 = __pyx_v_s;
                                        __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_20 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_15 * __pyx_v_X.strides[0]) ) + __pyx_t_22 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_32 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                        if (!__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L91_bool_binop_done;
                                        }
                                        __pyx_t_20 = __pyx_v_s;
                                        __pyx_t_19 = __pyx_v_i;
                                        __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_20 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_21 = __pyx_v_s;
                                        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_15 = 1;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_19 * __pyx_v_X.strides[0]) ) + __pyx_t_32 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_22 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_15 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        if (!__pyx_t_31) {
                                        } else {
                                          __pyx_t_13 = __pyx_t_31;
                                          goto __pyx_L91_bool_binop_done;
                                        }
                                        __pyx_t_21 = __pyx_v_s;
                                        __pyx_t_15 = __pyx_v_i;
                                        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_20 = __pyx_v_s;
                                        __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_20 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_19 = 0;
                                        __pyx_t_31 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_15 * __pyx_v_X.strides[0]) ) + __pyx_t_22 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_32 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_19 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                        __pyx_t_13 = __pyx_t_31;
                                        __pyx_L91_bool_binop_done:;
                                        if (__pyx_t_13) {

                                          /* "cyext_acv.pyx":801
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                             if o_all > 0:
 *                                 continue
 */
                                          __pyx_v_o_all = (__pyx_v_o_all + 1);

                                          /* "cyext_acv.pyx":800
 *                             o_all = 0
 *                             for s in range(S_size):
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":802
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 *                             if o_all > 0:             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                      __pyx_t_13 = ((__pyx_v_o_all > 0) != 0);
                                      if (__pyx_t_13) {

                                        /* "cyext_acv.pyx":803
 *                                     o_all = o_all + 1
 *                             if o_all > 0:
 *                                 continue             # <<<<<<<<<<<<<<
 * 
 *                             p = 0
 */
                                        goto __pyx_L84_continue;

                                        /* "cyext_acv.pyx":802
 *                                 if ((X[i, S[s]] > leaf_part[S[s], 1]) or (X[i, S[s]] < leaf_part[S[s], 0]) or (X[i, S[s]] > leaf_part_l[S[s], 1]) or (X[i, S[s]] < leaf_part_l[S[s], 0])):
 *                                     o_all = o_all + 1
 *                             if o_all > 0:             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":805
 *                                 continue
 * 
 *                             p = 0             # <<<<<<<<<<<<<<
 *                             p_u = 0
 *                             p_d = 0
 */
                                      __pyx_v_p = 0;

                                      /* "cyext_acv.pyx":806
 * 
 *                             p = 0
 *                             p_u = 0             # <<<<<<<<<<<<<<
 *                             p_d = 0
 *                             p_s = 0
 */
                                      __pyx_v_p_u = 0;

                                      /* "cyext_acv.pyx":807
 *                             p = 0
 *                             p_u = 0
 *                             p_d = 0             # <<<<<<<<<<<<<<
 *                             p_s = 0
 *                             p_su = 0
 */
                                      __pyx_v_p_d = 0;

                                      /* "cyext_acv.pyx":808
 *                             p_u = 0
 *                             p_d = 0
 *                             p_s = 0             # <<<<<<<<<<<<<<
 *                             p_su = 0
 *                             p_sd = 0
 */
                                      __pyx_v_p_s = 0;

                                      /* "cyext_acv.pyx":809
 *                             p_d = 0
 *                             p_s = 0
 *                             p_su = 0             # <<<<<<<<<<<<<<
 *                             p_sd = 0
 * 
 */
                                      __pyx_v_p_su = 0;

                                      /* "cyext_acv.pyx":810
 *                             p_s = 0
 *                             p_su = 0
 *                             p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                             for j in range(data.shape[0]):
 */
                                      __pyx_v_p_sd = 0;

                                      /* "cyext_acv.pyx":812
 *                             p_sd = 0
 * 
 *                             for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                 up = 0
 *                                 down = 0
 */
                                      __pyx_t_25 = (__pyx_v_data.shape[0]);
                                      __pyx_t_26 = __pyx_t_25;
                                      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_26; __pyx_t_36+=1) {
                                        __pyx_v_j = __pyx_t_36;

                                        /* "cyext_acv.pyx":813
 * 
 *                             for j in range(data.shape[0]):
 *                                 up = 0             # <<<<<<<<<<<<<<
 *                                 down = 0
 *                                 b_it = 0
 */
                                        __pyx_v_up = 0;

                                        /* "cyext_acv.pyx":814
 *                             for j in range(data.shape[0]):
 *                                 up = 0
 *                                 down = 0             # <<<<<<<<<<<<<<
 *                                 b_it = 0
 *                                 for s in range(S_size):
 */
                                        __pyx_v_down = 0;

                                        /* "cyext_acv.pyx":815
 *                                 up = 0
 *                                 down = 0
 *                                 b_it = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:
 */
                                        __pyx_v_b_it = 0;

                                        /* "cyext_acv.pyx":816
 *                                 down = 0
 *                                 b_it = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if data[j, S[s]] == X[i, S[s]]:
 *                                         b_it = b_it + 1
 */
                                        __pyx_t_38 = __pyx_v_S_size;
                                        __pyx_t_39 = __pyx_t_38;
                                        for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
                                          __pyx_v_s = __pyx_t_40;

                                          /* "cyext_acv.pyx":817
 *                                 b_it = 0
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          __pyx_t_20 = __pyx_v_s;
                                          __pyx_t_19 = __pyx_v_j;
                                          __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_20 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_21 = __pyx_v_s;
                                          __pyx_t_22 = __pyx_v_i;
                                          __pyx_t_15 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_21 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_32 * __pyx_v_data.strides[1]) ))) == (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_22 * __pyx_v_X.strides[0]) ) + __pyx_t_15 * __pyx_v_X.strides[1]) )))) != 0);
                                          if (__pyx_t_13) {

                                            /* "cyext_acv.pyx":818
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:
 *                                         b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if b_it == S_size:
 */
                                            __pyx_v_b_it = (__pyx_v_b_it + 1);

                                            /* "cyext_acv.pyx":817
 *                                 b_it = 0
 *                                 for s in range(S_size):
 *                                     if data[j, S[s]] == X[i, S[s]]:             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":820
 *                                         b_it = b_it + 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1
 */
                                        __pyx_t_13 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                        if (__pyx_t_13) {

                                          /* "cyext_acv.pyx":821
 * 
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         up = 1
 *                                     else:
 */
                                          __pyx_t_21 = __pyx_v_i;
                                          __pyx_t_15 = __pyx_v_j;
                                          __pyx_t_22 = __pyx_v_i;
                                          __pyx_t_20 = __pyx_v_j;
                                          __pyx_t_13 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_21 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_15 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_22 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_20 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                          if (__pyx_t_13) {

                                            /* "cyext_acv.pyx":822
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         down = 1
 */
                                            __pyx_v_up = 1;

                                            /* "cyext_acv.pyx":821
 * 
 *                                 if b_it == S_size:
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         up = 1
 *                                     else:
 */
                                            goto __pyx_L102;
                                          }

                                          /* "cyext_acv.pyx":824
 *                                         up = 1
 *                                     else:
 *                                         down = 1             # <<<<<<<<<<<<<<
 * 
 *                                     p += lm[j]
 */
                                          /*else*/ {
                                            __pyx_v_down = 1;
                                          }
                                          __pyx_L102:;

                                          /* "cyext_acv.pyx":826
 *                                         down = 1
 * 
 *                                     p += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down
 */
                                          __pyx_t_20 = __pyx_v_j;
                                          __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_20 * __pyx_v_lm.strides[0]) ))));

                                          /* "cyext_acv.pyx":827
 * 
 *                                     p += lm[j]
 *                                     p_u += lm[j] * up             # <<<<<<<<<<<<<<
 *                                     p_d += lm[j] * down
 *                                     p_s += 1
 */
                                          __pyx_t_20 = __pyx_v_j;
                                          __pyx_v_p_u = (__pyx_v_p_u + ((*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_20 * __pyx_v_lm.strides[0]) ))) * __pyx_v_up));

                                          /* "cyext_acv.pyx":828
 *                                     p += lm[j]
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down             # <<<<<<<<<<<<<<
 *                                     p_s += 1
 *                                     p_su += up
 */
                                          __pyx_t_20 = __pyx_v_j;
                                          __pyx_v_p_d = (__pyx_v_p_d + ((*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_20 * __pyx_v_lm.strides[0]) ))) * __pyx_v_down));

                                          /* "cyext_acv.pyx":829
 *                                     p_u += lm[j] * up
 *                                     p_d += lm[j] * down
 *                                     p_s += 1             # <<<<<<<<<<<<<<
 *                                     p_su += up
 *                                     p_sd += down
 */
                                          __pyx_v_p_s = (__pyx_v_p_s + 1);

                                          /* "cyext_acv.pyx":830
 *                                     p_d += lm[j] * down
 *                                     p_s += 1
 *                                     p_su += up             # <<<<<<<<<<<<<<
 *                                     p_sd += down
 * 
 */
                                          __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_up);

                                          /* "cyext_acv.pyx":831
 *                                     p_s += 1
 *                                     p_su += up
 *                                     p_sd += down             # <<<<<<<<<<<<<<
 * 
 *                             mean_forest[i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 */
                                          __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_down);

                                          /* "cyext_acv.pyx":820
 *                                         b_it = b_it + 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     if (fX[i] - y_pred[j]) * (fX[i] - y_pred[j]) > tX:
 *                                         up = 1
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":833
 *                                     p_sd += down
 * 
 *                             mean_forest[i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                             mean_forest[i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                             mean_forest[i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 */
                                      if (((__pyx_v_p_s != 0) != 0)) {
                                        __pyx_t_37 = (((__pyx_v_p * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_s);
                                      } else {
                                        __pyx_t_37 = 0.0;
                                      }
                                      __pyx_t_20 = __pyx_v_i;
                                      __pyx_t_22 = 0;
                                      *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_20 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_22 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;

                                      /* "cyext_acv.pyx":834
 * 
 *                             mean_forest[i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                             mean_forest[i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                             mean_forest[i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                      if (((__pyx_v_p_su != 0) != 0)) {
                                        __pyx_t_37 = (((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_su);
                                      } else {
                                        __pyx_t_37 = 0.0;
                                      }
                                      __pyx_t_22 = __pyx_v_i;
                                      __pyx_t_20 = 1;
                                      *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_22 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_20 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;

                                      /* "cyext_acv.pyx":835
 *                             mean_forest[i, 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                             mean_forest[i, 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                             mean_forest[i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(N):
 */
                                      if (((__pyx_v_p_sd != 0) != 0)) {
                                        __pyx_t_37 = (((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_sd);
                                      } else {
                                        __pyx_t_37 = 0.0;
                                      }
                                      __pyx_t_20 = __pyx_v_i;
                                      __pyx_t_22 = 2;
                                      *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_20 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_22 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_37;
                                      goto __pyx_L104;
                                      __pyx_L84_continue:;
                                      goto __pyx_L104;
                                      __pyx_L104:;
                                  }
                              }
                          }
                      }
                  }
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   __builtin_expect(!!(x), 1)
                      #define unlikely(x) __builtin_expect(!!(x), 0)
                  #endif
                }

                /* "cyext_acv.pyx":797
 *                                 lm[i] = 1
 * 
 *                         for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                             o_all = 0
 *                             for s in range(S_size):
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    #ifdef WITH_THREAD
                    __Pyx_FastGIL_Forget();
                    Py_BLOCK_THREADS
                    #endif
                    goto __pyx_L83;
                  }
                  __pyx_L83:;
                }
            }
          }
        }
      }
      __pyx_L7:;
    }
  }

  /* "cyext_acv.pyx":837
 *                             mean_forest[i, 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         sdp[i]  = (mean_forest[i, 1] - mean_forest[i, 0])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 * 
 */
  __pyx_t_16 = __pyx_v_N;
  __pyx_t_17 = __pyx_t_16;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_17; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "cyext_acv.pyx":838
 * 
 *     for i in range(N):
 *         sdp[i]  = (mean_forest[i, 1] - mean_forest[i, 0])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp)
 */
    __pyx_t_22 = __pyx_v_i;
    __pyx_t_20 = 1;
    __pyx_t_15 = __pyx_v_i;
    __pyx_t_21 = 2;
    if (((((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_22 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_20 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_15 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_21 * __pyx_v_mean_forest.strides[1]) )))) != 0.0) != 0)) {
      __pyx_t_32 = __pyx_v_i;
      __pyx_t_19 = 1;
      __pyx_t_41 = __pyx_v_i;
      __pyx_t_42 = 0;
      __pyx_t_43 = __pyx_v_i;
      __pyx_t_44 = 1;
      __pyx_t_45 = __pyx_v_i;
      __pyx_t_46 = 2;
      __pyx_t_37 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_32 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_19 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_41 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest.strides[1]) )))) / ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_43 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_44 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_45 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_46 * __pyx_v_mean_forest.strides[1]) )))));
    } else {
      __pyx_t_37 = 0.0;
    }
    __pyx_t_21 = __pyx_v_i;
    *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_21 * __pyx_v_sdp.strides[0]) )) = __pyx_t_37;
  }

  /* "cyext_acv.pyx":840
 *         sdp[i]  = (mean_forest[i, 1] - mean_forest[i, 0])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 * 
 *     return np.asarray(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_23)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_23);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_23, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":679
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_reg_cat(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, long[:] S, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __Pyx_XDECREF(__pyx_t_23);
  __PYX_XDEC_MEMVIEW(&__pyx_t_24, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_15compute_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_15compute_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_tX;
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_reg_cat (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_tX,&__pyx_n_s_y_pred,&__pyx_n_s_S,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 1); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 2); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 3); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 4); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 5); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 6); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 7); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 8); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 9); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 10); __PYX_ERR(0, 679, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, 11); __PYX_ERR(0, 679, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_reg_cat") < 0)) __PYX_ERR(0, 679, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_tX = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tX == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_S.memview)) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 681, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 681, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[10], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_reg_cat", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 679, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_14compute_sdp_reg_cat(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_14compute_sdp_reg_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_reg_cat", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S.memview)) { __Pyx_RaiseUnboundLocalError("S"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 679, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 679, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_reg_cat(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":846
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg_cat(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_17global_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg_cat(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  CYTHON_UNUSED unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_max_size;
  int __pyx_v_size;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  long __pyx_t_22;
  long __pyx_t_23;
  int __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_t_26;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_27;
  __Pyx_memviewslice __pyx_t_28 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<std::vector<long> > ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  std::vector<long> ::size_type __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  PyObject *__pyx_t_41 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg_cat", 0);

  /* "cyext_acv.pyx":852
 *             int minimal, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":853
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":854
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef vector[int].iterator t
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":855
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef vector[int].iterator t
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":865
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":868
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, leaf_idx_trees.shape[1], N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp, sdp_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_leaf_idx_trees.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":872
 *     cdef double[:] sdp, sdp_b
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":873
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm, lm_u, lm_d, lm_s
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":881
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_len_s_star = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":886
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 886, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":887
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":888
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":889
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 889, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":892
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_R_buf = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":894
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 894, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_14) {

    /* "cyext_acv.pyx":895
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 895, __pyx_L1_error)
    }
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 895, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_ci = __pyx_t_17;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 895, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_18 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_cj = __pyx_t_20;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 895, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 895, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":896
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in range(m)]
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_21 = (__pyx_t_14 != 0);
      if (__pyx_t_21) {
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6);
        PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
        __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_v_C); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":894
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":898
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":900
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 900, __pyx_L1_error)
  __pyx_v_m = __pyx_t_15;

  /* "cyext_acv.pyx":901
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_power = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "cyext_acv.pyx":902
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":903
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_22 = (__pyx_v_m + 1);
  __pyx_t_23 = __pyx_t_22;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_23; __pyx_t_13+=1) {
    __pyx_v_size = __pyx_t_13;

    /* "cyext_acv.pyx":904
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":905
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_itertools); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_combinations); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_24 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_24 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_24, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_24, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
      __pyx_t_25 = NULL;
    } else {
      __pyx_t_15 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_25 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 905, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_25)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 905, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 905, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 905, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_25(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 905, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "cyext_acv.pyx":906
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_7); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "cyext_acv.pyx":907
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":905
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyext_acv.pyx":908
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 908, __pyx_L1_error)

    /* "cyext_acv.pyx":909
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_21 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":910
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L16_break;

      /* "cyext_acv.pyx":909
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L16_break:;

  /* "cyext_acv.pyx":912
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_27;

  /* "cyext_acv.pyx":914
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_28;
  __pyx_t_28.memview = NULL;
  __pyx_t_28.data = NULL;

  /* "cyext_acv.pyx":917
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":918
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_S = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":920
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __pyx_t_6 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_15 = 0;
    __pyx_t_25 = NULL;
  } else {
    __pyx_t_15 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 920, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_25)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 920, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 920, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_25(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 920, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_s_0 = __pyx_t_13;

    /* "cyext_acv.pyx":921
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_29 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_30 = __pyx_t_29;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_30; __pyx_t_13+=1) {
      __pyx_v_s_1 = __pyx_t_13;

      /* "cyext_acv.pyx":922
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_31 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_32 = __pyx_t_31;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_32; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":923
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":925
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":926
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":927
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":928
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 * 
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":929
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 * 
 *             sdp_b = compute_sdp_reg_cat(X, fX, tX, y_pred, S[:S_size], data, values, partition_leaves_trees, leaf_idx_trees,
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":931
 *                 R_buf[i] = R[i]
 * 
 *             sdp_b = compute_sdp_reg_cat(X, fX, tX, y_pred, S[:S_size], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                         leaves_nb, scaling, 0)
 *             for i in range(N):
 */
      __pyx_t_10.data = __pyx_v_S.data;
      __pyx_t_10.memview = __pyx_v_S.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
      __pyx_t_24 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_10,
    __pyx_v_S.shape[0], __pyx_v_S.strides[0], __pyx_v_S.suboffsets[0],
    0,
    0,
    &__pyx_t_24,
    0,
    __pyx_v_S_size,
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 931, __pyx_L1_error)
}

__pyx_t_2 = __pyx_f_9cyext_acv_compute_sdp_reg_cat(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_t_10, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 931, __pyx_L1_error)

      /* "cyext_acv.pyx":932
 * 
 *             sdp_b = compute_sdp_reg_cat(X, fX, tX, y_pred, S[:S_size], data, values, partition_leaves_trees, leaf_idx_trees,
 *                         leaves_nb, scaling, 0)             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 */
      __Pyx_GOTREF(__pyx_t_2);
      __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
      __pyx_t_10.memview = NULL;
      __pyx_t_10.data = NULL;

      /* "cyext_acv.pyx":931
 *                 R_buf[i] = R[i]
 * 
 *             sdp_b = compute_sdp_reg_cat(X, fX, tX, y_pred, S[:S_size], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                         leaves_nb, scaling, 0)
 *             for i in range(N):
 */
      __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 931, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_b, 1);
      __pyx_v_sdp_b = __pyx_t_9;
      __pyx_t_9.memview = NULL;
      __pyx_t_9.data = NULL;

      /* "cyext_acv.pyx":933
 *             sdp_b = compute_sdp_reg_cat(X, fX, tX, y_pred, S[:S_size], data, values, partition_leaves_trees, leaf_idx_trees,
 *                         leaves_nb, scaling, 0)
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":934
 *                         leaves_nb, scaling, 0)
 *             for i in range(N):
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_33 = __pyx_v_i;
        __pyx_t_34 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_35 = __pyx_v_i;
        __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_21 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_34 * __pyx_v_sdp_b.strides[0]) ))) >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_36 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":935
 *             for i in range(N):
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_33 = __pyx_v_i;
          __pyx_t_34 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_34 * __pyx_v_sdp.strides[0]) )) = (*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_36 * __pyx_v_sdp_b.strides[0]) )));

          /* "cyext_acv.pyx":936
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_36 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":937
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_37 = __pyx_v_S_size;
          __pyx_t_38 = __pyx_t_37;
          for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
            __pyx_v_s = __pyx_t_39;

            /* "cyext_acv.pyx":938
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_35 = __pyx_v_s;
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_34 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_33 * __pyx_v_s_star.strides[0]) ) + __pyx_t_34 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":934
 *                         leaves_nb, scaling, 0)
 *             for i in range(N):
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":940
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_21 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":941
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_36 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":942
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_36 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":943
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_37 = __pyx_v_S_size;
          __pyx_t_38 = __pyx_t_37;
          for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
            __pyx_v_s = __pyx_t_39;

            /* "cyext_acv.pyx":944
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_35 = __pyx_v_s;
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_34 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_33 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_34 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":945
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_16 = (__pyx_v_X.shape[1]);
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_17 = __pyx_t_16;
          for (__pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_36 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_37 < __pyx_t_17; __pyx_t_37+=1) {
            __pyx_v_s = __pyx_t_37;

            /* "cyext_acv.pyx":946
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_33 = __pyx_v_i;
            __pyx_t_34 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_40 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_34 * __pyx_v_s_star.strides[0]) ) + __pyx_t_40 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":940
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":948
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_11 = __pyx_v_N;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":949
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_35 = __pyx_v_i;
      __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_21 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_36 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_21) {

        /* "cyext_acv.pyx":950
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 950, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":951
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_35 = __pyx_v_i;
        __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_36 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_23 = __pyx_t_22;
        for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
          __pyx_v_s = __pyx_t_24;

          /* "cyext_acv.pyx":952
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_33 = __pyx_v_s;
          __pyx_t_40 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_36 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_40 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":949
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":954
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_31 = __pyx_v_r.size();
    __pyx_t_32 = __pyx_t_31;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_32; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":955
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":956
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":958
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_14 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_14) {
    } else {
      goto __pyx_L47_next_and;
    }
    __pyx_t_14 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_14) {
    } else {
      __pyx_t_21 = __pyx_t_14;
      goto __pyx_L46_bool_binop_done;
    }
    __pyx_L47_next_and:;
    __pyx_t_14 = (__pyx_v_stop != 0);
    __pyx_t_21 = __pyx_t_14;
    __pyx_L46_bool_binop_done:;
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":959
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L21_break;

      /* "cyext_acv.pyx":958
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":920
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L21_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":961
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_long); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_41 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_long); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_41) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_41 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":846
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg_cat(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_28, 1);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_17global_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_17global_sdp_reg_cat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_tX;
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_reg_cat (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_tX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 1); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 2); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 3); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 4); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 5); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 6); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 7); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 8); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 9); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 10); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 11); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 12); __PYX_ERR(0, 846, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, 13); __PYX_ERR(0, 846, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_reg_cat") < 0)) __PYX_ERR(0, 846, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_tX = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tX == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 848, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 848, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 849, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 849, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 849, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[10]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[13]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cat", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 846, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 849, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_16global_sdp_reg_cat(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_16global_sdp_reg_cat(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg_cat", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 846, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 846, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_reg_cat(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":968
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_19global_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_leaves_tree = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_tree_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_value;
  double __pyx_v_value_l;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  double __pyx_v_ss;
  double __pyx_v_ss_a;
  double __pyx_v_ss_u;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_max_size;
  int __pyx_v_size;
  int __pyx_v_leaf_numb_l;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm_u = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_norm_d = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  int __pyx_v_l;
  PyObject *__pyx_v_nb_leaf_l = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  int __pyx_t_23;
  long __pyx_t_24;
  long __pyx_t_25;
  int __pyx_t_26;
  PyObject *(*__pyx_t_27)(PyObject *);
  int __pyx_t_28;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_29;
  __Pyx_memviewslice __pyx_t_30 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_31;
  std::vector<std::vector<long> > ::size_type __pyx_t_32;
  std::vector<long> ::size_type __pyx_t_33;
  std::vector<long> ::size_type __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  unsigned int __pyx_t_39;
  unsigned int __pyx_t_40;
  int __pyx_t_41;
  unsigned int __pyx_t_42;
  unsigned int __pyx_t_43;
  int __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  int __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  __Pyx_memviewslice __pyx_t_51 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_52;
  int __pyx_t_53;
  int __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  double __pyx_t_56;
  int __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  Py_ssize_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg", 0);

  /* "cyext_acv.pyx":974
 *             int minimal, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":975
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":976
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef vector[int].iterator t
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":977
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef vector[int].iterator t
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":987
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":990
 * 
 *     cdef double[:, :, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, max_leaves, max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_7);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":994
 *     cdef double[:] sdp
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":995
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm, lm_u, lm_d, lm_s
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1003
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_len_s_star = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1005
 *     len_s_star = np.zeros((N), dtype=np.int)
 *     cdef double[:, :, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_p_n = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":1006
 *     cdef double[:, :, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_p_u_n = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":1007
 *     p_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_p_d_n = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":1012
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1012, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1013
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_13 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "cyext_acv.pyx":1014
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":1015
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1015, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1018
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] norm, norm_u, norm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_R_buf = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1021
 * 
 *     cdef double[:] norm, norm_u, norm_d
 *     norm = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 *     norm_u = np.zeros((n_trees))
 *     norm_d = np.zeros((n_trees))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_norm = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1022
 *     cdef double[:] norm, norm_u, norm_d
 *     norm = np.zeros((n_trees))
 *     norm_u = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 *     norm_d = np.zeros((n_trees))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_norm_u = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1023
 *     norm = np.zeros((n_trees))
 *     norm_u = np.zeros((n_trees))
 *     norm_d = np.zeros((n_trees))             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_norm_d = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1025
 *     norm_d = np.zeros((n_trees))
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 1025, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_16) {

    /* "cyext_acv.pyx":1026
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 */
    __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1026, __pyx_L1_error)
    }
    __pyx_t_17 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1026, __pyx_L1_error)
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
      __pyx_v_ci = __pyx_t_19;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1026, __pyx_L1_error)
      }
      __pyx_t_7 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_20 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1026, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_v_cj = __pyx_t_22;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 1026, __pyx_L1_error)
        }
        __pyx_t_7 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1026, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 1026, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "cyext_acv.pyx":1027
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in range(m)]
 */
    __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_13 = __pyx_v_m;
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_16 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_23 = (__pyx_t_16 != 0);
      if (__pyx_t_23) {
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1027, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1027, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
        __pyx_t_7 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1027, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_t_6 = PyNumber_Add(__pyx_t_8, __pyx_v_C); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "cyext_acv.pyx":1025
 *     norm_d = np.zeros((n_trees))
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":1029
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __pyx_v_m;
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_8);
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
      __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":1031
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_17 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1031, __pyx_L1_error)
  __pyx_v_m = __pyx_t_17;

  /* "cyext_acv.pyx":1032
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_power = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "cyext_acv.pyx":1033
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":1034
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_24 = (__pyx_v_m + 1);
  __pyx_t_25 = __pyx_t_24;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_25; __pyx_t_15+=1) {
    __pyx_v_size = __pyx_t_15;

    /* "cyext_acv.pyx":1035
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "cyext_acv.pyx":1036
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_itertools); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_combinations); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = NULL;
    __pyx_t_26 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_26 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_7};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_26, 2+__pyx_t_26); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_7};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_26, 2+__pyx_t_26); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_26); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_26, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_26, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_8 = __pyx_t_6; __Pyx_INCREF(__pyx_t_8); __pyx_t_17 = 0;
      __pyx_t_27 = NULL;
    } else {
      __pyx_t_17 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_27 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1036, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_27)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_17); __Pyx_INCREF(__pyx_t_6); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1036, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_17); __Pyx_INCREF(__pyx_t_6); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1036, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1036, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_27(__pyx_t_8);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1036, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "cyext_acv.pyx":1037
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_28 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_6); if (unlikely(__pyx_t_28 == ((int)-1))) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "cyext_acv.pyx":1038
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":1036
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "cyext_acv.pyx":1039
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_28 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_28 == ((int)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)

    /* "cyext_acv.pyx":1040
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_23 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_23) {

      /* "cyext_acv.pyx":1041
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L16_break;

      /* "cyext_acv.pyx":1040
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L16_break:;

  /* "cyext_acv.pyx":1043
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_29 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1043, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_29;

  /* "cyext_acv.pyx":1045
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_7);
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_30 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_30.memview)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_30;
  __pyx_t_30.memview = NULL;
  __pyx_t_30.data = NULL;

  /* "cyext_acv.pyx":1048
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":1049
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_S = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1051
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
    __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1); __pyx_t_17 = 0;
    __pyx_t_27 = NULL;
  } else {
    __pyx_t_17 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_27 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1051, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  for (;;) {
    if (likely(!__pyx_t_27)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1051, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1051, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_8); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1051, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1051, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_27(__pyx_t_1);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1051, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_s_0 = __pyx_t_15;

    /* "cyext_acv.pyx":1052
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_31 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_32 = __pyx_t_31;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_32; __pyx_t_15+=1) {
      __pyx_v_s_1 = __pyx_t_15;

      /* "cyext_acv.pyx":1053
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_33 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_34 = __pyx_t_33;
      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_34; __pyx_t_26+=1) {
        __pyx_v_i = __pyx_t_26;

        /* "cyext_acv.pyx":1054
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_35 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":1056
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":1057
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":1058
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":1059
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 * 
 */
      __pyx_t_13 = __pyx_v_N;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_14; __pyx_t_26+=1) {
        __pyx_v_i = __pyx_t_26;

        /* "cyext_acv.pyx":1060
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
        __pyx_t_35 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":1062
 *                 R_buf[i] = R[i]
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(max_leaves):
 *                     for leaf_numb_l in range(max_leaves):
 */
      __pyx_t_26 = __pyx_v_n_trees;
      __pyx_t_36 = __pyx_t_26;
      for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
        __pyx_v_b = __pyx_t_37;

        /* "cyext_acv.pyx":1063
 * 
 *             for b in range(n_trees):
 *                 for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                     for leaf_numb_l in range(max_leaves):
 *                         for i in range(N):
 */
        __pyx_t_13 = __pyx_v_max_leaves;
        __pyx_t_14 = __pyx_t_13;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_14; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":1064
 *             for b in range(n_trees):
 *                 for leaf_numb in range(max_leaves):
 *                     for leaf_numb_l in range(max_leaves):             # <<<<<<<<<<<<<<
 *                         for i in range(N):
 *                             p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 */
          __pyx_t_39 = __pyx_v_max_leaves;
          __pyx_t_40 = __pyx_t_39;
          for (__pyx_t_41 = 0; __pyx_t_41 < __pyx_t_40; __pyx_t_41+=1) {
            __pyx_v_leaf_numb_l = __pyx_t_41;

            /* "cyext_acv.pyx":1065
 *                 for leaf_numb in range(max_leaves):
 *                     for leaf_numb_l in range(max_leaves):
 *                         for i in range(N):             # <<<<<<<<<<<<<<
 *                             p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 */
            __pyx_t_42 = __pyx_v_N;
            __pyx_t_43 = __pyx_t_42;
            for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_43; __pyx_t_44+=1) {
              __pyx_v_i = __pyx_t_44;

              /* "cyext_acv.pyx":1066
 *                     for leaf_numb_l in range(max_leaves):
 *                         for i in range(N):
 *                             p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                             p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 */
              __pyx_t_35 = __pyx_v_i;
              __pyx_t_45 = __pyx_v_b;
              __pyx_t_46 = __pyx_v_leaf_numb;
              __pyx_t_47 = __pyx_v_leaf_numb_l;
              __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_45 * __pyx_v_p_n.strides[0]) ) + __pyx_t_46 * __pyx_v_p_n.strides[1]) ) + __pyx_t_47 * __pyx_v_p_n.strides[2]) ) + __pyx_t_48 * __pyx_v_p_n.strides[3]) )) = 0.0;

              /* "cyext_acv.pyx":1067
 *                         for i in range(N):
 *                             p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                             p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             for j in range(3):
 */
              __pyx_t_35 = __pyx_v_i;
              __pyx_t_48 = __pyx_v_b;
              __pyx_t_47 = __pyx_v_leaf_numb;
              __pyx_t_46 = __pyx_v_leaf_numb_l;
              __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_48 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_47 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_46 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_45 * __pyx_v_p_u_n.strides[3]) )) = 0.0;

              /* "cyext_acv.pyx":1068
 *                             p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                             for j in range(3):
 *                                 mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], j] = 0
 */
              __pyx_t_35 = __pyx_v_i;
              __pyx_t_45 = __pyx_v_b;
              __pyx_t_46 = __pyx_v_leaf_numb;
              __pyx_t_47 = __pyx_v_leaf_numb_l;
              __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_45 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_46 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_47 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_48 * __pyx_v_p_d_n.strides[3]) )) = 0.0;

              /* "cyext_acv.pyx":1069
 *                             p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             for j in range(3):             # <<<<<<<<<<<<<<
 *                                 mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], j] = 0
 *             for b in range(n_trees):
 */
              for (__pyx_t_49 = 0; __pyx_t_49 < 3; __pyx_t_49+=1) {
                __pyx_v_j = __pyx_t_49;

                /* "cyext_acv.pyx":1070
 *                             p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] = 0
 *                             for j in range(3):
 *                                 mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], j] = 0             # <<<<<<<<<<<<<<
 *             for b in range(n_trees):
 *                 for l in range(n_trees):
 */
                __pyx_t_35 = __pyx_v_i;
                __pyx_t_48 = __pyx_v_b;
                __pyx_t_47 = __pyx_v_leaf_numb;
                __pyx_t_46 = __pyx_v_leaf_numb_l;
                __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                __pyx_t_50 = __pyx_v_j;
                *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[4]) )) = 0.0;
              }
            }
          }
        }
      }

      /* "cyext_acv.pyx":1071
 *                             for j in range(3):
 *                                 mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], j] = 0
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 for l in range(n_trees):
 *                     if b == l:
 */
      __pyx_t_26 = __pyx_v_n_trees;
      __pyx_t_36 = __pyx_t_26;
      for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
        __pyx_v_b = __pyx_t_37;

        /* "cyext_acv.pyx":1072
 *                                 mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], j] = 0
 *             for b in range(n_trees):
 *                 for l in range(n_trees):             # <<<<<<<<<<<<<<
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]
 */
        __pyx_t_38 = __pyx_v_n_trees;
        __pyx_t_41 = __pyx_t_38;
        for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_41; __pyx_t_44+=1) {
          __pyx_v_l = __pyx_t_44;

          /* "cyext_acv.pyx":1073
 *             for b in range(n_trees):
 *                 for l in range(n_trees):
 *                     if b == l:             # <<<<<<<<<<<<<<
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 */
          __pyx_t_23 = ((__pyx_v_b == __pyx_v_l) != 0);
          if (__pyx_t_23) {

            /* "cyext_acv.pyx":1074
 *                 for l in range(n_trees):
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                         nb_leaf = leaves_nb[b]
 * 
 */
            __pyx_t_51.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_51.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_51, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_51.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_51.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_51.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_51.suboffsets[0] = -1;

__pyx_t_51.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_51.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_51.suboffsets[1] = -1;

__pyx_t_51.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_51.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_51.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
            __pyx_v_leaves_tree = __pyx_t_51;
            __pyx_t_51.memview = NULL;
            __pyx_t_51.data = NULL;

            /* "cyext_acv.pyx":1075
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 * 
 *                         for leaf_numb in range(nb_leaf):
 */
            __pyx_t_35 = __pyx_v_b;
            __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_35 * __pyx_v_leaves_nb.strides[0]) )));

            /* "cyext_acv.pyx":1077
 *                         nb_leaf = leaves_nb[b]
 * 
 *                         for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                             leaf_part = leaves_tree[leaf_numb]
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
            __pyx_t_13 = __pyx_v_nb_leaf;
            __pyx_t_14 = __pyx_t_13;
            for (__pyx_t_49 = 0; __pyx_t_49 < __pyx_t_14; __pyx_t_49+=1) {
              __pyx_v_leaf_numb = __pyx_t_49;

              /* "cyext_acv.pyx":1078
 * 
 *                         for leaf_numb in range(nb_leaf):
 *                             leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 */
              __pyx_t_5.data = __pyx_v_leaves_tree.data;
              __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
              __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
              __pyx_v_leaf_part = __pyx_t_5;
              __pyx_t_5.memview = NULL;
              __pyx_t_5.data = NULL;

              /* "cyext_acv.pyx":1079
 *                         for leaf_numb in range(nb_leaf):
 *                             leaf_part = leaves_tree[leaf_numb]
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)
 */
              __pyx_t_35 = __pyx_v_b;
              __pyx_t_50 = __pyx_v_leaf_numb;
              __pyx_t_45 = __pyx_v_b;
              __pyx_t_46 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_35 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_idx_trees.strides[1]) )));
              __pyx_t_47 = 0;
              __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_45 * __pyx_v_values.strides[0]) ) + __pyx_t_46 * __pyx_v_values.strides[1]) ) + __pyx_t_47 * __pyx_v_values.strides[2]) )));

              /* "cyext_acv.pyx":1081
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_8 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_8);
              PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
              __pyx_t_8 = 0;
              __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1081, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
              __pyx_v_lm = __pyx_t_11;
              __pyx_t_11.memview = NULL;
              __pyx_t_11.data = NULL;

              /* "cyext_acv.pyx":1082
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                             for i in prange(data.shape[0], nogil=True):
 */
              __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
              __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1082, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
              __pyx_v_lm_s = __pyx_t_11;
              __pyx_t_11.memview = NULL;
              __pyx_t_11.data = NULL;

              /* "cyext_acv.pyx":1084
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                             for i in prange(data.shape[0], nogil=True):             # <<<<<<<<<<<<<<
 *                                 a_it = 0
 *                                 b_it = 0
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {
                    if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 1084, __pyx_L48_error) }
                    __pyx_t_18 = (__pyx_v_data.shape[0]);
                    if ((1 == 0)) abort();
                    {
                        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                            #undef likely
                            #undef unlikely
                            #define likely(x)   (x)
                            #define unlikely(x) (x)
                        #endif
                        __pyx_t_20 = (__pyx_t_18 - 0 + 1 - 1/abs(1)) / 1;
                        if (__pyx_t_20 > 0)
                        {
                            #ifdef _OPENMP
                            #pragma omp parallel private(__pyx_t_16, __pyx_t_23, __pyx_t_35, __pyx_t_39, __pyx_t_40, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_50, __pyx_t_52, __pyx_t_53, __pyx_t_54)
                            #endif /* _OPENMP */
                            {
                                #ifdef _OPENMP
                                #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                                #endif /* _OPENMP */
                                for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_20; __pyx_t_19++){
                                    {
                                        __pyx_v_i = (int)(0 + 1 * __pyx_t_19);
                                        /* Initialize private variables to invalid values */
                                        __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                        __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                        __pyx_v_s = ((int)0xbad0bad0);

                                        /* "cyext_acv.pyx":1085
 * 
 *                             for i in prange(data.shape[0], nogil=True):
 *                                 a_it = 0             # <<<<<<<<<<<<<<
 *                                 b_it = 0
 *                                 for s in range(m):
 */
                                        __pyx_v_a_it = 0;

                                        /* "cyext_acv.pyx":1086
 *                             for i in prange(data.shape[0], nogil=True):
 *                                 a_it = 0
 *                                 b_it = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(m):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 */
                                        __pyx_v_b_it = 0;

                                        /* "cyext_acv.pyx":1087
 *                                 a_it = 0
 *                                 b_it = 0
 *                                 for s in range(m):             # <<<<<<<<<<<<<<
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1
 */
                                        __pyx_t_39 = __pyx_v_m;
                                        __pyx_t_40 = __pyx_t_39;
                                        for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_40; __pyx_t_52+=1) {
                                          __pyx_v_s = __pyx_t_52;

                                          /* "cyext_acv.pyx":1088
 *                                 b_it = 0
 *                                 for s in range(m):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 */
                                          __pyx_t_50 = __pyx_v_i;
                                          __pyx_t_35 = __pyx_v_s;
                                          __pyx_t_47 = __pyx_v_s;
                                          __pyx_t_46 = 1;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_50 * __pyx_v_data.strides[0]) ) + __pyx_t_35 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_47 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_46 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (__pyx_t_16) {
                                          } else {
                                            __pyx_t_23 = __pyx_t_16;
                                            goto __pyx_L57_bool_binop_done;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_47 = __pyx_v_s;
                                          __pyx_t_35 = __pyx_v_s;
                                          __pyx_t_50 = 0;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_46 * __pyx_v_data.strides[0]) ) + __pyx_t_47 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_35 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_23 = __pyx_t_16;
                                          __pyx_L57_bool_binop_done:;
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1089
 *                                 for s in range(m):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 */
                                            __pyx_v_a_it = (__pyx_v_a_it + 1);

                                            /* "cyext_acv.pyx":1088
 *                                 b_it = 0
 *                                 for s in range(m):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":1090
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                         b_it = b_it + 1
 */
                                        __pyx_t_52 = __pyx_v_S_size;
                                        __pyx_t_53 = __pyx_t_52;
                                        for (__pyx_t_54 = 0; __pyx_t_54 < __pyx_t_53; __pyx_t_54+=1) {
                                          __pyx_v_s = __pyx_t_54;

                                          /* "cyext_acv.pyx":1091
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          __pyx_t_50 = __pyx_v_s;
                                          __pyx_t_35 = __pyx_v_i;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_46 = __pyx_v_s;
                                          __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_48 = 1;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_47 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_45 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_48 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (__pyx_t_16) {
                                          } else {
                                            __pyx_t_23 = __pyx_t_16;
                                            goto __pyx_L62_bool_binop_done;
                                          }
                                          __pyx_t_46 = __pyx_v_s;
                                          __pyx_t_48 = __pyx_v_i;
                                          __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_50 = __pyx_v_s;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_35 = 0;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_48 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_47 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_23 = __pyx_t_16;
                                          __pyx_L62_bool_binop_done:;
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1092
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                         b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if a_it == m:
 */
                                            __pyx_v_b_it = (__pyx_v_b_it + 1);

                                            /* "cyext_acv.pyx":1091
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":1094
 *                                         b_it = b_it + 1
 * 
 *                                 if a_it == m:             # <<<<<<<<<<<<<<
 *                                     lm[i] = 1
 * 
 */
                                        __pyx_t_23 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                        if (__pyx_t_23) {

                                          /* "cyext_acv.pyx":1095
 * 
 *                                 if a_it == m:
 *                                     lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 if b_it == S_size:
 */
                                          __pyx_t_50 = __pyx_v_i;
                                          *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_50 * __pyx_v_lm.strides[0]) )) = 1;

                                          /* "cyext_acv.pyx":1094
 *                                         b_it = b_it + 1
 * 
 *                                 if a_it == m:             # <<<<<<<<<<<<<<
 *                                     lm[i] = 1
 * 
 */
                                        }

                                        /* "cyext_acv.pyx":1097
 *                                     lm[i] = 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     lm_s[i] = 1
 * 
 */
                                        __pyx_t_23 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                        if (__pyx_t_23) {

                                          /* "cyext_acv.pyx":1098
 * 
 *                                 if b_it == S_size:
 *                                     lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                             for i in prange(N, nogil=True):
 */
                                          __pyx_t_50 = __pyx_v_i;
                                          *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_50 * __pyx_v_lm_s.strides[0]) )) = 1;

                                          /* "cyext_acv.pyx":1097
 *                                     lm[i] = 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     lm_s[i] = 1
 * 
 */
                                        }
                                    }
                                }
                            }
                        }
                    }
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   __builtin_expect(!!(x), 1)
                        #define unlikely(x) __builtin_expect(!!(x), 0)
                    #endif
                  }

                  /* "cyext_acv.pyx":1084
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                             for i in prange(data.shape[0], nogil=True):             # <<<<<<<<<<<<<<
 *                                 a_it = 0
 *                                 b_it = 0
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L49;
                    }
                    __pyx_L48_error: {
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L1_error;
                    }
                    __pyx_L49:;
                  }
              }

              /* "cyext_acv.pyx":1100
 *                                     lm_s[i] = 1
 * 
 *                             for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                                 o_all = 0
 *                                 for s in range(S_size):
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {
                    __pyx_t_39 = __pyx_v_N;
                    if ((1 == 0)) abort();
                    {
                        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                            #undef likely
                            #undef unlikely
                            #define likely(x)   (x)
                            #define unlikely(x) (x)
                        #endif
                        __pyx_t_42 = (__pyx_t_39 - 0 + 1 - 1/abs(1)) / 1;
                        if (__pyx_t_42 > 0)
                        {
                            #ifdef _OPENMP
                            #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_16, __pyx_t_19, __pyx_t_20, __pyx_t_23, __pyx_t_35, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_50, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56)
                            #endif /* _OPENMP */
                            {
                                #ifdef _OPENMP
                                #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                                #endif /* _OPENMP */
                                for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_42; __pyx_t_40++){
                                    {
                                        __pyx_v_i = (int)(0 + 1 * __pyx_t_40);
                                        /* Initialize private variables to invalid values */
                                        __pyx_v_j = ((int)0xbad0bad0);
                                        __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                        __pyx_v_s = ((int)0xbad0bad0);

                                        /* "cyext_acv.pyx":1101
 * 
 *                             for i in prange(N, nogil=True):
 *                                 o_all = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 */
                                        __pyx_v_o_all = 0;

                                        /* "cyext_acv.pyx":1102
 *                             for i in prange(N, nogil=True):
 *                                 o_all = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 */
                                        __pyx_t_52 = __pyx_v_S_size;
                                        __pyx_t_53 = __pyx_t_52;
                                        for (__pyx_t_54 = 0; __pyx_t_54 < __pyx_t_53; __pyx_t_54+=1) {
                                          __pyx_v_s = __pyx_t_54;

                                          /* "cyext_acv.pyx":1103
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 */
                                          __pyx_t_50 = __pyx_v_i;
                                          __pyx_t_35 = __pyx_v_s;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_50 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_45 = __pyx_v_s;
                                          __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_45 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_55 = 1;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_47 * __pyx_v_X.strides[0]) ) + __pyx_t_46 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_48 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_55 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (!__pyx_t_16) {
                                          } else {
                                            __pyx_t_23 = __pyx_t_16;
                                            goto __pyx_L80_bool_binop_done;
                                          }
                                          __pyx_t_45 = __pyx_v_i;
                                          __pyx_t_55 = __pyx_v_s;
                                          __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_45 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_55 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_50 = __pyx_v_s;
                                          __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_47 = 0;
                                          __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_48 * __pyx_v_X.strides[0]) ) + __pyx_t_35 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_46 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_47 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_23 = __pyx_t_16;
                                          __pyx_L80_bool_binop_done:;
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1104
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                                 if o_all > 0:
 *                                     continue
 */
                                            __pyx_v_o_all = (__pyx_v_o_all + 1);

                                            /* "cyext_acv.pyx":1103
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":1105
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                        __pyx_t_23 = ((__pyx_v_o_all > 0) != 0);
                                        if (__pyx_t_23) {

                                          /* "cyext_acv.pyx":1106
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                                 p = 0
 */
                                          goto __pyx_L73_continue;

                                          /* "cyext_acv.pyx":1105
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                        }

                                        /* "cyext_acv.pyx":1108
 *                                     continue
 * 
 *                                 p = 0             # <<<<<<<<<<<<<<
 *                                 p_u = 0
 *                                 p_d = 0
 */
                                        __pyx_v_p = 0;

                                        /* "cyext_acv.pyx":1109
 * 
 *                                 p = 0
 *                                 p_u = 0             # <<<<<<<<<<<<<<
 *                                 p_d = 0
 *                                 p_s = 0
 */
                                        __pyx_v_p_u = 0;

                                        /* "cyext_acv.pyx":1110
 *                                 p = 0
 *                                 p_u = 0
 *                                 p_d = 0             # <<<<<<<<<<<<<<
 *                                 p_s = 0
 *                                 p_su = 0
 */
                                        __pyx_v_p_d = 0;

                                        /* "cyext_acv.pyx":1111
 *                                 p_u = 0
 *                                 p_d = 0
 *                                 p_s = 0             # <<<<<<<<<<<<<<
 *                                 p_su = 0
 *                                 p_sd = 0
 */
                                        __pyx_v_p_s = 0;

                                        /* "cyext_acv.pyx":1112
 *                                 p_d = 0
 *                                 p_s = 0
 *                                 p_su = 0             # <<<<<<<<<<<<<<
 *                                 p_sd = 0
 * 
 */
                                        __pyx_v_p_su = 0;

                                        /* "cyext_acv.pyx":1113
 *                                 p_s = 0
 *                                 p_su = 0
 *                                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                                 for j in range(data.shape[0]):
 */
                                        __pyx_v_p_sd = 0;

                                        /* "cyext_acv.pyx":1115
 *                                 p_sd = 0
 * 
 *                                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 */
                                        __pyx_t_20 = (__pyx_v_data.shape[0]);
                                        __pyx_t_19 = __pyx_t_20;
                                        for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_19; __pyx_t_52+=1) {
                                          __pyx_v_j = __pyx_t_52;

                                          /* "cyext_acv.pyx":1116
 * 
 *                                 for j in range(data.shape[0]):
 *                                     p += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 */
                                          __pyx_t_50 = __pyx_v_j;
                                          __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_50 * __pyx_v_lm.strides[0]) ))));

                                          /* "cyext_acv.pyx":1117
 *                                 for j in range(data.shape[0]):
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]
 */
                                          __pyx_t_50 = __pyx_v_j;
                                          __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_50 * __pyx_v_lm_s.strides[0]) ))));

                                          /* "cyext_acv.pyx":1118
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]
 */
                                          __pyx_t_50 = __pyx_v_i;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_50 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_46 = __pyx_v_j;
                                          __pyx_t_55 = __pyx_v_i;
                                          __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_55 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_35 = __pyx_v_j;
                                          __pyx_t_23 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_47 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_46 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_45 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_35 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1119
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_su += lm_s[j]
 *                                     else:
 */
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_35 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":1120
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         p_d += lm[j]
 */
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_35 * __pyx_v_lm_s.strides[0]) ))));

                                            /* "cyext_acv.pyx":1118
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]
 */
                                            goto __pyx_L85;
                                          }

                                          /* "cyext_acv.pyx":1122
 *                                         p_su += lm_s[j]
 *                                     else:
 *                                         p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_sd += lm_s[j]
 * 
 */
                                          /*else*/ {
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_35 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":1123
 *                                     else:
 *                                         p_d += lm[j]
 *                                         p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 */
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_35 * __pyx_v_lm_s.strides[0]) ))));
                                          }
                                          __pyx_L85:;
                                        }

                                        /* "cyext_acv.pyx":1125
 *                                         p_sd += lm_s[j]
 * 
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 */
                                        if (((__pyx_v_p_s != 0) != 0)) {
                                          __pyx_t_35 = __pyx_v_i;
                                          __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_56 = ((((__pyx_v_p * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_s) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_55 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p) * __pyx_v_value) / __pyx_v_p_s));
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_55 = __pyx_v_b;
                                        __pyx_t_45 = __pyx_v_leaf_numb;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_47 = 0;
                                        *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_55 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                        /* "cyext_acv.pyx":1126
 * 
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 * 
 */
                                        if (((__pyx_v_p_su != 0) != 0)) {
                                          __pyx_t_35 = __pyx_v_i;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_56 = ((((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_su) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_47 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_u) * __pyx_v_value) / __pyx_v_p_su));
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_47 = __pyx_v_b;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_55 = 1;
                                        *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_47 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_55 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                        /* "cyext_acv.pyx":1127
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                                 p_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 */
                                        if (((__pyx_v_p_sd != 0) != 0)) {
                                          __pyx_t_35 = __pyx_v_i;
                                          __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_56 = ((((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_sd) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_55 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_d) * __pyx_v_value) / __pyx_v_p_sd));
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_55 = __pyx_v_b;
                                        __pyx_t_45 = __pyx_v_leaf_numb;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_47 = 2;
                                        *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_55 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_46 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                        /* "cyext_acv.pyx":1129
 *                                 mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 * 
 *                                 p_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                 p_u_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                                 p_d_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                        if (((__pyx_v_p_s != 0) != 0)) {
                                          __pyx_t_56 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_47 = __pyx_v_b;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_47 * __pyx_v_p_n.strides[0]) ) + __pyx_t_50 * __pyx_v_p_n.strides[1]) ) + __pyx_t_46 * __pyx_v_p_n.strides[2]) ) + __pyx_t_45 * __pyx_v_p_n.strides[3]) )) += __pyx_t_56;

                                        /* "cyext_acv.pyx":1130
 * 
 *                                 p_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                                 p_u_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                 p_d_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *                     else:
 */
                                        if (((__pyx_v_p_su != 0) != 0)) {
                                          __pyx_t_56 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_45 = __pyx_v_b;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_45 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_46 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_50 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_u_n.strides[3]) )) += __pyx_t_56;

                                        /* "cyext_acv.pyx":1131
 *                                 p_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                                 p_u_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                                 p_d_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]
 */
                                        if (((__pyx_v_p_sd != 0) != 0)) {
                                          __pyx_t_56 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                        } else {
                                          __pyx_t_56 = 0.0;
                                        }
                                        __pyx_t_35 = __pyx_v_i;
                                        __pyx_t_47 = __pyx_v_b;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_46 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_47 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_50 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_46 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_45 * __pyx_v_p_d_n.strides[3]) )) += __pyx_t_56;
                                        goto __pyx_L87;
                                        __pyx_L73_continue:;
                                        goto __pyx_L87;
                                        __pyx_L87:;
                                    }
                                }
                            }
                        }
                    }
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   __builtin_expect(!!(x), 1)
                        #define unlikely(x) __builtin_expect(!!(x), 0)
                    #endif
                  }

                  /* "cyext_acv.pyx":1100
 *                                     lm_s[i] = 1
 * 
 *                             for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                                 o_all = 0
 *                                 for s in range(S_size):
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L72;
                    }
                    __pyx_L72:;
                  }
              }
            }

            /* "cyext_acv.pyx":1073
 *             for b in range(n_trees):
 *                 for l in range(n_trees):
 *                     if b == l:             # <<<<<<<<<<<<<<
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 */
            goto __pyx_L42;
          }

          /* "cyext_acv.pyx":1133
 *                                 p_d_n[b, leaf_numb, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]
 */
          /*else*/ {
            __pyx_t_51.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_51.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_51, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_51.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_51.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_51.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_51.suboffsets[0] = -1;

__pyx_t_51.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_51.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_51.suboffsets[1] = -1;

__pyx_t_51.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_51.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_51.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
            __pyx_v_leaves_tree = __pyx_t_51;
            __pyx_t_51.memview = NULL;
            __pyx_t_51.data = NULL;

            /* "cyext_acv.pyx":1134
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                         leaves_tree_l = partition_leaves_trees[l]
 *                         nb_leaf_l = leaves_nb[l]
 */
            __pyx_t_35 = __pyx_v_b;
            __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_35 * __pyx_v_leaves_nb.strides[0]) )));

            /* "cyext_acv.pyx":1135
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]             # <<<<<<<<<<<<<<
 *                         nb_leaf_l = leaves_nb[l]
 * 
 */
            __pyx_t_51.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_51.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_51, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_51.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_51.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_51.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_51.suboffsets[0] = -1;

__pyx_t_51.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_51.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_51.suboffsets[1] = -1;

__pyx_t_51.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_51.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_51.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
            __pyx_v_leaves_tree_l = __pyx_t_51;
            __pyx_t_51.memview = NULL;
            __pyx_t_51.data = NULL;

            /* "cyext_acv.pyx":1136
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]
 *                         nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                         for leaf_numb in range(nb_leaf):
 */
            __pyx_t_35 = __pyx_v_l;
            __pyx_t_7 = __Pyx_PyInt_From_long((*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_35 * __pyx_v_leaves_nb.strides[0]) )))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1136, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_XDECREF_SET(__pyx_v_nb_leaf_l, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "cyext_acv.pyx":1138
 *                         nb_leaf_l = leaves_nb[l]
 * 
 *                         for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                             for leaf_numb_l in range(nb_leaf_l):
 * 
 */
            __pyx_t_13 = __pyx_v_nb_leaf;
            __pyx_t_14 = __pyx_t_13;
            for (__pyx_t_49 = 0; __pyx_t_49 < __pyx_t_14; __pyx_t_49+=1) {
              __pyx_v_leaf_numb = __pyx_t_49;

              /* "cyext_acv.pyx":1139
 * 
 *                         for leaf_numb in range(nb_leaf):
 *                             for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]
 */
              __pyx_t_24 = __Pyx_PyInt_As_long(__pyx_v_nb_leaf_l); if (unlikely((__pyx_t_24 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L1_error)
              __pyx_t_25 = __pyx_t_24;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_25; __pyx_t_52+=1) {
                __pyx_v_leaf_numb_l = __pyx_t_52;

                /* "cyext_acv.pyx":1141
 *                             for leaf_numb_l in range(nb_leaf_l):
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
                __pyx_t_5.data = __pyx_v_leaves_tree.data;
                __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
                __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
                __pyx_v_leaf_part = __pyx_t_5;
                __pyx_t_5.memview = NULL;
                __pyx_t_5.data = NULL;

                /* "cyext_acv.pyx":1142
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]             # <<<<<<<<<<<<<<
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 */
                __pyx_t_5.data = __pyx_v_leaves_tree_l.data;
                __pyx_t_5.memview = __pyx_v_leaves_tree_l.memview;
                __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree_l.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree_l.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree_l.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree_l.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree_l.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
                __pyx_v_leaf_part_l = __pyx_t_5;
                __pyx_t_5.memview = NULL;
                __pyx_t_5.data = NULL;

                /* "cyext_acv.pyx":1143
 *                                 leaf_part = leaves_tree[leaf_numb]
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 */
                __pyx_t_35 = __pyx_v_b;
                __pyx_t_45 = __pyx_v_leaf_numb;
                __pyx_t_46 = __pyx_v_b;
                __pyx_t_50 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_35 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_45 * __pyx_v_leaf_idx_trees.strides[1]) )));
                __pyx_t_47 = 0;
                __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_46 * __pyx_v_values.strides[0]) ) + __pyx_t_50 * __pyx_v_values.strides[1]) ) + __pyx_t_47 * __pyx_v_values.strides[2]) )));

                /* "cyext_acv.pyx":1144
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]             # <<<<<<<<<<<<<<
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)
 */
                __pyx_t_45 = __pyx_v_l;
                __pyx_t_35 = __pyx_v_leaf_numb_l;
                __pyx_t_47 = __pyx_v_l;
                __pyx_t_50 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_45 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_idx_trees.strides[1]) )));
                __pyx_t_46 = 0;
                __pyx_v_value_l = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_47 * __pyx_v_values.strides[0]) ) + __pyx_t_50 * __pyx_v_values.strides[1]) ) + __pyx_t_46 * __pyx_v_values.strides[2]) )));

                /* "cyext_acv.pyx":1146
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
                __pyx_t_7 = 0;
                __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1146, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
                __pyx_v_lm = __pyx_t_11;
                __pyx_t_11.memview = NULL;
                __pyx_t_11.data = NULL;

                /* "cyext_acv.pyx":1147
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                                 for i in prange(data.shape[0], nogil=True):
 */
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
                __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1147, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
                __pyx_v_lm_s = __pyx_t_11;
                __pyx_t_11.memview = NULL;
                __pyx_t_11.data = NULL;

                /* "cyext_acv.pyx":1149
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                                 for i in prange(data.shape[0], nogil=True):             # <<<<<<<<<<<<<<
 *                                     a_it = 0
 *                                     b_it = 0
 */
                {
                    #ifdef WITH_THREAD
                    PyThreadState *_save;
                    Py_UNBLOCK_THREADS
                    __Pyx_FastGIL_Remember();
                    #endif
                    /*try:*/ {
                      if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 1149, __pyx_L95_error) }
                      __pyx_t_20 = (__pyx_v_data.shape[0]);
                      if ((1 == 0)) abort();
                      {
                          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                              #undef likely
                              #undef unlikely
                              #define likely(x)   (x)
                              #define unlikely(x) (x)
                          #endif
                          __pyx_t_18 = (__pyx_t_20 - 0 + 1 - 1/abs(1)) / 1;
                          if (__pyx_t_18 > 0)
                          {
                              #ifdef _OPENMP
                              #pragma omp parallel private(__pyx_t_16, __pyx_t_23, __pyx_t_35, __pyx_t_40, __pyx_t_42, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_50, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_57)
                              #endif /* _OPENMP */
                              {
                                  #ifdef _OPENMP
                                  #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                                  #endif /* _OPENMP */
                                  for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19++){
                                      {
                                          __pyx_v_i = (int)(0 + 1 * __pyx_t_19);
                                          /* Initialize private variables to invalid values */
                                          __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                          __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                          __pyx_v_s = ((int)0xbad0bad0);

                                          /* "cyext_acv.pyx":1150
 * 
 *                                 for i in prange(data.shape[0], nogil=True):
 *                                     a_it = 0             # <<<<<<<<<<<<<<
 *                                     b_it = 0
 *                                     for s in range(m):
 */
                                          __pyx_v_a_it = 0;

                                          /* "cyext_acv.pyx":1151
 *                                 for i in prange(data.shape[0], nogil=True):
 *                                     a_it = 0
 *                                     b_it = 0             # <<<<<<<<<<<<<<
 *                                     for s in range(m):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 */
                                          __pyx_v_b_it = 0;

                                          /* "cyext_acv.pyx":1152
 *                                     a_it = 0
 *                                     b_it = 0
 *                                     for s in range(m):             # <<<<<<<<<<<<<<
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1
 */
                                          __pyx_t_42 = __pyx_v_m;
                                          __pyx_t_40 = __pyx_t_42;
                                          for (__pyx_t_53 = 0; __pyx_t_53 < __pyx_t_40; __pyx_t_53+=1) {
                                            __pyx_v_s = __pyx_t_53;

                                            /* "cyext_acv.pyx":1153
 *                                     b_it = 0
 *                                     for s in range(m):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 */
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_45 = __pyx_v_s;
                                            __pyx_t_46 = __pyx_v_s;
                                            __pyx_t_50 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_46 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L104_bool_binop_done;
                                            }
                                            __pyx_t_50 = __pyx_v_i;
                                            __pyx_t_46 = __pyx_v_s;
                                            __pyx_t_45 = __pyx_v_s;
                                            __pyx_t_35 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_50 * __pyx_v_data.strides[0]) ) + __pyx_t_46 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_45 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L104_bool_binop_done;
                                            }
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_45 = __pyx_v_s;
                                            __pyx_t_46 = __pyx_v_s;
                                            __pyx_t_50 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_46 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L104_bool_binop_done;
                                            }
                                            __pyx_t_50 = __pyx_v_i;
                                            __pyx_t_46 = __pyx_v_s;
                                            __pyx_t_45 = __pyx_v_s;
                                            __pyx_t_35 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_50 * __pyx_v_data.strides[0]) ) + __pyx_t_46 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_45 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_23 = __pyx_t_16;
                                            __pyx_L104_bool_binop_done:;
                                            if (__pyx_t_23) {

                                              /* "cyext_acv.pyx":1154
 *                                     for s in range(m):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 */
                                              __pyx_v_a_it = (__pyx_v_a_it + 1);

                                              /* "cyext_acv.pyx":1153
 *                                     b_it = 0
 *                                     for s in range(m):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":1155
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                             b_it = b_it + 1
 */
                                          __pyx_t_53 = __pyx_v_S_size;
                                          __pyx_t_54 = __pyx_t_53;
                                          for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_54; __pyx_t_57+=1) {
                                            __pyx_v_s = __pyx_t_57;

                                            /* "cyext_acv.pyx":1156
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             b_it = b_it + 1
 * 
 */
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_45 = __pyx_v_i;
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_50 = __pyx_v_s;
                                            __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_55 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_45 * __pyx_v_data.strides[0]) ) + __pyx_t_46 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_47 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_55 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L111_bool_binop_done;
                                            }
                                            __pyx_t_50 = __pyx_v_s;
                                            __pyx_t_55 = __pyx_v_i;
                                            __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_45 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_55 * __pyx_v_data.strides[0]) ) + __pyx_t_47 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_46 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_45 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L111_bool_binop_done;
                                            }
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_45 = __pyx_v_i;
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_50 = __pyx_v_s;
                                            __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_55 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_45 * __pyx_v_data.strides[0]) ) + __pyx_t_46 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_47 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_55 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L111_bool_binop_done;
                                            }
                                            __pyx_t_50 = __pyx_v_s;
                                            __pyx_t_55 = __pyx_v_i;
                                            __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_45 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_55 * __pyx_v_data.strides[0]) ) + __pyx_t_47 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_46 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_45 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_23 = __pyx_t_16;
                                            __pyx_L111_bool_binop_done:;
                                            if (__pyx_t_23) {

                                              /* "cyext_acv.pyx":1157
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                     if a_it == m:
 */
                                              __pyx_v_b_it = (__pyx_v_b_it + 1);

                                              /* "cyext_acv.pyx":1156
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             b_it = b_it + 1
 * 
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":1159
 *                                             b_it = b_it + 1
 * 
 *                                     if a_it == m:             # <<<<<<<<<<<<<<
 *                                         lm[i] = 1
 * 
 */
                                          __pyx_t_23 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1160
 * 
 *                                     if a_it == m:
 *                                         lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                     if b_it == S_size:
 */
                                            __pyx_t_35 = __pyx_v_i;
                                            *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_35 * __pyx_v_lm.strides[0]) )) = 1;

                                            /* "cyext_acv.pyx":1159
 *                                             b_it = b_it + 1
 * 
 *                                     if a_it == m:             # <<<<<<<<<<<<<<
 *                                         lm[i] = 1
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":1162
 *                                         lm[i] = 1
 * 
 *                                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                         lm_s[i] = 1
 * 
 */
                                          __pyx_t_23 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1163
 * 
 *                                     if b_it == S_size:
 *                                         lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 for i in prange(N, nogil=True):
 */
                                            __pyx_t_35 = __pyx_v_i;
                                            *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_35 * __pyx_v_lm_s.strides[0]) )) = 1;

                                            /* "cyext_acv.pyx":1162
 *                                         lm[i] = 1
 * 
 *                                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                         lm_s[i] = 1
 * 
 */
                                          }
                                      }
                                  }
                              }
                          }
                      }
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   __builtin_expect(!!(x), 1)
                          #define unlikely(x) __builtin_expect(!!(x), 0)
                      #endif
                    }

                    /* "cyext_acv.pyx":1149
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                                 for i in prange(data.shape[0], nogil=True):             # <<<<<<<<<<<<<<
 *                                     a_it = 0
 *                                     b_it = 0
 */
                    /*finally:*/ {
                      /*normal exit:*/{
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L96;
                      }
                      __pyx_L95_error: {
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L1_error;
                      }
                      __pyx_L96:;
                    }
                }

                /* "cyext_acv.pyx":1165
 *                                         lm_s[i] = 1
 * 
 *                                 for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                                     o_all = 0
 *                                     for s in range(S_size):
 */
                {
                    #ifdef WITH_THREAD
                    PyThreadState *_save;
                    Py_UNBLOCK_THREADS
                    __Pyx_FastGIL_Remember();
                    #endif
                    /*try:*/ {
                      __pyx_t_42 = __pyx_v_N;
                      if ((1 == 0)) abort();
                      {
                          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                              #undef likely
                              #undef unlikely
                              #define likely(x)   (x)
                              #define unlikely(x) (x)
                          #endif
                          __pyx_t_39 = (__pyx_t_42 - 0 + 1 - 1/abs(1)) / 1;
                          if (__pyx_t_39 > 0)
                          {
                              #ifdef _OPENMP
                              #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_16, __pyx_t_18, __pyx_t_19, __pyx_t_23, __pyx_t_35, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_50, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57)
                              #endif /* _OPENMP */
                              {
                                  #ifdef _OPENMP
                                  #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                                  #endif /* _OPENMP */
                                  for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40++){
                                      {
                                          __pyx_v_i = (int)(0 + 1 * __pyx_t_40);
                                          /* Initialize private variables to invalid values */
                                          __pyx_v_j = ((int)0xbad0bad0);
                                          __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                          __pyx_v_s = ((int)0xbad0bad0);

                                          /* "cyext_acv.pyx":1166
 * 
 *                                 for i in prange(N, nogil=True):
 *                                     o_all = 0             # <<<<<<<<<<<<<<
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 */
                                          __pyx_v_o_all = 0;

                                          /* "cyext_acv.pyx":1167
 *                                 for i in prange(N, nogil=True):
 *                                     o_all = 0
 *                                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 */
                                          __pyx_t_53 = __pyx_v_S_size;
                                          __pyx_t_54 = __pyx_t_53;
                                          for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_54; __pyx_t_57+=1) {
                                            __pyx_v_s = __pyx_t_57;

                                            /* "cyext_acv.pyx":1168
 *                                     o_all = 0
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 */
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_45 = __pyx_v_s;
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_45 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_47 = __pyx_v_s;
                                            __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_47 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_48 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_46 * __pyx_v_X.strides[0]) ) + __pyx_t_50 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_55 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_48 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (!__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L131_bool_binop_done;
                                            }
                                            __pyx_t_47 = __pyx_v_i;
                                            __pyx_t_48 = __pyx_v_s;
                                            __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_47 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_46 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_55 * __pyx_v_X.strides[0]) ) + __pyx_t_45 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_50 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_46 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (!__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L131_bool_binop_done;
                                            }
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_46 = __pyx_v_s;
                                            __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_47 = __pyx_v_s;
                                            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_47 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_55 = 1;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_50 * __pyx_v_X.strides[0]) ) + __pyx_t_48 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_45 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_55 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (!__pyx_t_16) {
                                            } else {
                                              __pyx_t_23 = __pyx_t_16;
                                              goto __pyx_L131_bool_binop_done;
                                            }
                                            __pyx_t_47 = __pyx_v_i;
                                            __pyx_t_55 = __pyx_v_s;
                                            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_47 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_55 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_50 = 0;
                                            __pyx_t_16 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_45 * __pyx_v_X.strides[0]) ) + __pyx_t_46 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_48 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_23 = __pyx_t_16;
                                            __pyx_L131_bool_binop_done:;
                                            if (__pyx_t_23) {

                                              /* "cyext_acv.pyx":1169
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                                     if o_all > 0:
 *                                         continue
 */
                                              __pyx_v_o_all = (__pyx_v_o_all + 1);

                                              /* "cyext_acv.pyx":1168
 *                                     o_all = 0
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":1170
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:             # <<<<<<<<<<<<<<
 *                                         continue
 * 
 */
                                          __pyx_t_23 = ((__pyx_v_o_all > 0) != 0);
                                          if (__pyx_t_23) {

                                            /* "cyext_acv.pyx":1171
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 *                                         continue             # <<<<<<<<<<<<<<
 * 
 *                                     p = 0
 */
                                            goto __pyx_L124_continue;

                                            /* "cyext_acv.pyx":1170
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:             # <<<<<<<<<<<<<<
 *                                         continue
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":1173
 *                                         continue
 * 
 *                                     p = 0             # <<<<<<<<<<<<<<
 *                                     p_u = 0
 *                                     p_d = 0
 */
                                          __pyx_v_p = 0;

                                          /* "cyext_acv.pyx":1174
 * 
 *                                     p = 0
 *                                     p_u = 0             # <<<<<<<<<<<<<<
 *                                     p_d = 0
 *                                     p_s = 0
 */
                                          __pyx_v_p_u = 0;

                                          /* "cyext_acv.pyx":1175
 *                                     p = 0
 *                                     p_u = 0
 *                                     p_d = 0             # <<<<<<<<<<<<<<
 *                                     p_s = 0
 *                                     p_su = 0
 */
                                          __pyx_v_p_d = 0;

                                          /* "cyext_acv.pyx":1176
 *                                     p_u = 0
 *                                     p_d = 0
 *                                     p_s = 0             # <<<<<<<<<<<<<<
 *                                     p_su = 0
 *                                     p_sd = 0
 */
                                          __pyx_v_p_s = 0;

                                          /* "cyext_acv.pyx":1177
 *                                     p_d = 0
 *                                     p_s = 0
 *                                     p_su = 0             # <<<<<<<<<<<<<<
 *                                     p_sd = 0
 * 
 */
                                          __pyx_v_p_su = 0;

                                          /* "cyext_acv.pyx":1178
 *                                     p_s = 0
 *                                     p_su = 0
 *                                     p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                                     for j in range(data.shape[0]):
 */
                                          __pyx_v_p_sd = 0;

                                          /* "cyext_acv.pyx":1180
 *                                     p_sd = 0
 * 
 *                                     for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 */
                                          __pyx_t_18 = (__pyx_v_data.shape[0]);
                                          __pyx_t_19 = __pyx_t_18;
                                          for (__pyx_t_53 = 0; __pyx_t_53 < __pyx_t_19; __pyx_t_53+=1) {
                                            __pyx_v_j = __pyx_t_53;

                                            /* "cyext_acv.pyx":1181
 * 
 *                                     for j in range(data.shape[0]):
 *                                         p += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 */
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_35 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":1182
 *                                     for j in range(data.shape[0]):
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]
 */
                                            __pyx_t_35 = __pyx_v_j;
                                            __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_35 * __pyx_v_lm_s.strides[0]) ))));

                                            /* "cyext_acv.pyx":1183
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]
 */
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_48 = __pyx_v_j;
                                            __pyx_t_55 = __pyx_v_i;
                                            __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_55 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_46 = __pyx_v_j;
                                            __pyx_t_23 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_50 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_48 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_47 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_46 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                            if (__pyx_t_23) {

                                              /* "cyext_acv.pyx":1184
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                             p_su += lm_s[j]
 *                                         else:
 */
                                              __pyx_t_46 = __pyx_v_j;
                                              __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_46 * __pyx_v_lm.strides[0]) ))));

                                              /* "cyext_acv.pyx":1185
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                                         else:
 *                                             p_d += lm[j]
 */
                                              __pyx_t_46 = __pyx_v_j;
                                              __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_46 * __pyx_v_lm_s.strides[0]) ))));

                                              /* "cyext_acv.pyx":1183
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]
 */
                                              goto __pyx_L138;
                                            }

                                            /* "cyext_acv.pyx":1187
 *                                             p_su += lm_s[j]
 *                                         else:
 *                                             p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                             p_sd += lm_s[j]
 * 
 */
                                            /*else*/ {
                                              __pyx_t_46 = __pyx_v_j;
                                              __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_46 * __pyx_v_lm.strides[0]) ))));

                                              /* "cyext_acv.pyx":1188
 *                                         else:
 *                                             p_d += lm[j]
 *                                             p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 */
                                              __pyx_t_46 = __pyx_v_j;
                                              __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_46 * __pyx_v_lm_s.strides[0]) ))));
                                            }
                                            __pyx_L138:;
                                          }

                                          /* "cyext_acv.pyx":1190
 *                                             p_sd += lm_s[j]
 * 
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 */
                                          if (((__pyx_v_p_s != 0) != 0)) {
                                            __pyx_t_56 = (((__pyx_v_p * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_s);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_55 = __pyx_v_b;
                                          __pyx_t_47 = __pyx_v_leaf_numb;
                                          __pyx_t_48 = __pyx_v_leaf_numb_l;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_50 = 0;
                                          *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_55 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_35 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                          /* "cyext_acv.pyx":1191
 * 
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                          if (((__pyx_v_p_su != 0) != 0)) {
                                            __pyx_t_56 = (((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_su);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_50 = __pyx_v_b;
                                          __pyx_t_35 = __pyx_v_leaf_numb;
                                          __pyx_t_48 = __pyx_v_leaf_numb_l;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_55 = 1;
                                          *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_50 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_35 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_55 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                          /* "cyext_acv.pyx":1192
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                                     p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 */
                                          if (((__pyx_v_p_sd != 0) != 0)) {
                                            __pyx_t_56 = (((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_sd);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_55 = __pyx_v_b;
                                          __pyx_t_47 = __pyx_v_leaf_numb;
                                          __pyx_t_48 = __pyx_v_leaf_numb_l;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_50 = 2;
                                          *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_55 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_35 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[4]) )) += __pyx_t_56;

                                          /* "cyext_acv.pyx":1194
 *                                     mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 *                                     p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                     p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                                     p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                          if (((__pyx_v_p_s != 0) != 0)) {
                                            __pyx_t_56 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_50 = __pyx_v_b;
                                          __pyx_t_35 = __pyx_v_leaf_numb;
                                          __pyx_t_48 = __pyx_v_leaf_numb_l;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_50 * __pyx_v_p_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_n.strides[3]) )) += __pyx_t_56;

                                          /* "cyext_acv.pyx":1195
 * 
 *                                     p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                                     p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                     p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *             for i in range(N):
 */
                                          if (((__pyx_v_p_su != 0) != 0)) {
                                            __pyx_t_56 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_47 = __pyx_v_b;
                                          __pyx_t_48 = __pyx_v_leaf_numb;
                                          __pyx_t_35 = __pyx_v_leaf_numb_l;
                                          __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_47 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_48 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_35 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_50 * __pyx_v_p_u_n.strides[3]) )) += __pyx_t_56;

                                          /* "cyext_acv.pyx":1196
 *                                     p_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                                     p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                                     p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 ss_u = 0
 */
                                          if (((__pyx_v_p_sd != 0) != 0)) {
                                            __pyx_t_56 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                          } else {
                                            __pyx_t_56 = 0.0;
                                          }
                                          __pyx_t_46 = __pyx_v_i;
                                          __pyx_t_50 = __pyx_v_b;
                                          __pyx_t_35 = __pyx_v_leaf_numb;
                                          __pyx_t_48 = __pyx_v_leaf_numb_l;
                                          __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_50 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_d_n.strides[3]) )) += __pyx_t_56;
                                          goto __pyx_L140;
                                          __pyx_L124_continue:;
                                          goto __pyx_L140;
                                          __pyx_L140:;
                                      }
                                  }
                              }
                          }
                      }
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   __builtin_expect(!!(x), 1)
                          #define unlikely(x) __builtin_expect(!!(x), 0)
                      #endif
                    }

                    /* "cyext_acv.pyx":1165
 *                                         lm_s[i] = 1
 * 
 *                                 for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *                                     o_all = 0
 *                                     for s in range(S_size):
 */
                    /*finally:*/ {
                      /*normal exit:*/{
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L123;
                      }
                      __pyx_L123:;
                    }
                }
              }
            }
          }
          __pyx_L42:;
        }
      }

      /* "cyext_acv.pyx":1197
 *                                     p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                                     p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 ss_u = 0
 *                 ss_d = 0
 */
      __pyx_t_13 = __pyx_v_N;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_14; __pyx_t_26+=1) {
        __pyx_v_i = __pyx_t_26;

        /* "cyext_acv.pyx":1198
 *                                     p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 *             for i in range(N):
 *                 ss_u = 0             # <<<<<<<<<<<<<<
 *                 ss_d = 0
 *                 ss_a = 0
 */
        __pyx_v_ss_u = 0.0;

        /* "cyext_acv.pyx":1199
 *             for i in range(N):
 *                 ss_u = 0
 *                 ss_d = 0             # <<<<<<<<<<<<<<
 *                 ss_a = 0
 *                 for b in range(n_trees):
 */
        __pyx_v_ss_d = 0.0;

        /* "cyext_acv.pyx":1200
 *                 ss_u = 0
 *                 ss_d = 0
 *                 ss_a = 0             # <<<<<<<<<<<<<<
 *                 for b in range(n_trees):
 *                     for l in range(n_trees):
 */
        __pyx_v_ss_a = 0.0;

        /* "cyext_acv.pyx":1201
 *                 ss_d = 0
 *                 ss_a = 0
 *                 for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                     for l in range(n_trees):
 *                         if b == l:
 */
        __pyx_t_36 = __pyx_v_n_trees;
        __pyx_t_37 = __pyx_t_36;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
          __pyx_v_b = __pyx_t_38;

          /* "cyext_acv.pyx":1202
 *                 ss_a = 0
 *                 for b in range(n_trees):
 *                     for l in range(n_trees):             # <<<<<<<<<<<<<<
 *                         if b == l:
 *                             n = 0
 */
          __pyx_t_41 = __pyx_v_n_trees;
          __pyx_t_44 = __pyx_t_41;
          for (__pyx_t_49 = 0; __pyx_t_49 < __pyx_t_44; __pyx_t_49+=1) {
            __pyx_v_l = __pyx_t_49;

            /* "cyext_acv.pyx":1203
 *                 for b in range(n_trees):
 *                     for l in range(n_trees):
 *                         if b == l:             # <<<<<<<<<<<<<<
 *                             n = 0
 *                             n_u = 0
 */
            __pyx_t_23 = ((__pyx_v_b == __pyx_v_l) != 0);
            if (__pyx_t_23) {

              /* "cyext_acv.pyx":1204
 *                     for l in range(n_trees):
 *                         if b == l:
 *                             n = 0             # <<<<<<<<<<<<<<
 *                             n_u = 0
 *                             n_d = 0
 */
              __pyx_v_n = 0.0;

              /* "cyext_acv.pyx":1205
 *                         if b == l:
 *                             n = 0
 *                             n_u = 0             # <<<<<<<<<<<<<<
 *                             n_d = 0
 *                             nb_leaf = leaves_nb[b]
 */
              __pyx_v_n_u = 0.0;

              /* "cyext_acv.pyx":1206
 *                             n = 0
 *                             n_u = 0
 *                             n_d = 0             # <<<<<<<<<<<<<<
 *                             nb_leaf = leaves_nb[b]
 *                             for leaf_numb in range(nb_leaf):
 */
              __pyx_v_n_d = 0.0;

              /* "cyext_acv.pyx":1207
 *                             n_u = 0
 *                             n_d = 0
 *                             nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                             for leaf_numb in range(nb_leaf):
 *                                 n += p_n[b, leaf_numb, leaf_numb, R_buf[i]]
 */
              __pyx_t_46 = __pyx_v_b;
              __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_46 * __pyx_v_leaves_nb.strides[0]) )));

              /* "cyext_acv.pyx":1208
 *                             n_d = 0
 *                             nb_leaf = leaves_nb[b]
 *                             for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                                 n += p_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_u += p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]
 */
              __pyx_t_39 = __pyx_v_nb_leaf;
              __pyx_t_40 = __pyx_t_39;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_40; __pyx_t_52+=1) {
                __pyx_v_leaf_numb = __pyx_t_52;

                /* "cyext_acv.pyx":1209
 *                             nb_leaf = leaves_nb[b]
 *                             for leaf_numb in range(nb_leaf):
 *                                 n += p_n[b, leaf_numb, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                                 n_u += p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_d += p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]
 */
                __pyx_t_46 = __pyx_v_i;
                __pyx_t_47 = __pyx_v_b;
                __pyx_t_48 = __pyx_v_leaf_numb;
                __pyx_t_35 = __pyx_v_leaf_numb;
                __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_47 * __pyx_v_p_n.strides[0]) ) + __pyx_t_48 * __pyx_v_p_n.strides[1]) ) + __pyx_t_35 * __pyx_v_p_n.strides[2]) ) + __pyx_t_50 * __pyx_v_p_n.strides[3]) ))));

                /* "cyext_acv.pyx":1210
 *                             for leaf_numb in range(nb_leaf):
 *                                 n += p_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_u += p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                                 n_d += p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                             norm[b] = n
 */
                __pyx_t_46 = __pyx_v_i;
                __pyx_t_50 = __pyx_v_b;
                __pyx_t_35 = __pyx_v_leaf_numb;
                __pyx_t_48 = __pyx_v_leaf_numb;
                __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_50 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_u_n.strides[3]) ))));

                /* "cyext_acv.pyx":1211
 *                                 n += p_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_u += p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_d += p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                             norm[b] = n
 *                             norm_u[b] = n_u
 */
                __pyx_t_46 = __pyx_v_i;
                __pyx_t_47 = __pyx_v_b;
                __pyx_t_48 = __pyx_v_leaf_numb;
                __pyx_t_35 = __pyx_v_leaf_numb;
                __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_47 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_48 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_35 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_50 * __pyx_v_p_d_n.strides[3]) ))));
              }

              /* "cyext_acv.pyx":1212
 *                                 n_u += p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                                 n_d += p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                             norm[b] = n             # <<<<<<<<<<<<<<
 *                             norm_u[b] = n_u
 *                             norm_d[b] = n_d
 */
              __pyx_t_46 = __pyx_v_b;
              *((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_46 * __pyx_v_norm.strides[0]) )) = __pyx_v_n;

              /* "cyext_acv.pyx":1213
 *                                 n_d += p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]
 *                             norm[b] = n
 *                             norm_u[b] = n_u             # <<<<<<<<<<<<<<
 *                             norm_d[b] = n_d
 *                             for leaf_numb in range(nb_leaf):
 */
              __pyx_t_46 = __pyx_v_b;
              *((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_46 * __pyx_v_norm_u.strides[0]) )) = __pyx_v_n_u;

              /* "cyext_acv.pyx":1214
 *                             norm[b] = n
 *                             norm_u[b] = n_u
 *                             norm_d[b] = n_d             # <<<<<<<<<<<<<<
 *                             for leaf_numb in range(nb_leaf):
 *                                 ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  R_buf[i], 0]/norm[b] if norm[b] !=0 else 0
 */
              __pyx_t_46 = __pyx_v_b;
              *((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_46 * __pyx_v_norm_d.strides[0]) )) = __pyx_v_n_d;

              /* "cyext_acv.pyx":1215
 *                             norm_u[b] = n_u
 *                             norm_d[b] = n_d
 *                             for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                                 ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  R_buf[i], 0]/norm[b] if norm[b] !=0 else 0
 *                                 ss_u += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1]/norm_u[b] if norm_u[b] !=0 else 0
 */
              __pyx_t_39 = __pyx_v_nb_leaf;
              __pyx_t_40 = __pyx_t_39;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_40; __pyx_t_52+=1) {
                __pyx_v_leaf_numb = __pyx_t_52;

                /* "cyext_acv.pyx":1216
 *                             norm_d[b] = n_d
 *                             for leaf_numb in range(nb_leaf):
 *                                 ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  R_buf[i], 0]/norm[b] if norm[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                                 ss_u += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1]/norm_u[b] if norm_u[b] !=0 else 0
 *                                 ss_d += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2]/norm_d[b] if norm_d[b] !=0 else 0
 */
                __pyx_t_46 = __pyx_v_b;
                if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_46 * __pyx_v_norm.strides[0]) ))) != 0.0) != 0)) {
                  __pyx_t_50 = __pyx_v_i;
                  __pyx_t_35 = __pyx_v_b;
                  __pyx_t_48 = __pyx_v_leaf_numb;
                  __pyx_t_47 = __pyx_v_leaf_numb;
                  __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_50 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_t_45 = 0;
                  __pyx_t_58 = __pyx_v_b;
                  __pyx_t_56 = ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_35 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_55 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_58 * __pyx_v_norm.strides[0]) ))));
                } else {
                  __pyx_t_56 = 0.0;
                }
                __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_56);

                /* "cyext_acv.pyx":1217
 *                             for leaf_numb in range(nb_leaf):
 *                                 ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  R_buf[i], 0]/norm[b] if norm[b] !=0 else 0
 *                                 ss_u += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1]/norm_u[b] if norm_u[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                                 ss_d += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2]/norm_d[b] if norm_d[b] !=0 else 0
 *                         else:
 */
                __pyx_t_46 = __pyx_v_b;
                if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_46 * __pyx_v_norm_u.strides[0]) ))) != 0.0) != 0)) {
                  __pyx_t_58 = __pyx_v_i;
                  __pyx_t_50 = __pyx_v_b;
                  __pyx_t_45 = __pyx_v_leaf_numb;
                  __pyx_t_55 = __pyx_v_leaf_numb;
                  __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_58 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_t_48 = 1;
                  __pyx_t_35 = __pyx_v_b;
                  __pyx_t_56 = ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_50 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_55 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_35 * __pyx_v_norm_u.strides[0]) ))));
                } else {
                  __pyx_t_56 = 0.0;
                }
                __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_56);

                /* "cyext_acv.pyx":1218
 *                                 ss_a += mean_forest_b[b, leaf_numb, leaf_numb,  R_buf[i], 0]/norm[b] if norm[b] !=0 else 0
 *                                 ss_u += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 1]/norm_u[b] if norm_u[b] !=0 else 0
 *                                 ss_d += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2]/norm_d[b] if norm_d[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                         else:
 *                             nb_leaf = leaves_nb[b]
 */
                __pyx_t_46 = __pyx_v_b;
                if ((((*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_46 * __pyx_v_norm_d.strides[0]) ))) != 0.0) != 0)) {
                  __pyx_t_35 = __pyx_v_i;
                  __pyx_t_58 = __pyx_v_b;
                  __pyx_t_48 = __pyx_v_leaf_numb;
                  __pyx_t_47 = __pyx_v_leaf_numb;
                  __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_t_45 = 2;
                  __pyx_t_50 = __pyx_v_b;
                  __pyx_t_56 = ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_58 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_55 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[4]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_50 * __pyx_v_norm_d.strides[0]) ))));
                } else {
                  __pyx_t_56 = 0.0;
                }
                __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_56);
              }

              /* "cyext_acv.pyx":1203
 *                 for b in range(n_trees):
 *                     for l in range(n_trees):
 *                         if b == l:             # <<<<<<<<<<<<<<
 *                             n = 0
 *                             n_u = 0
 */
              goto __pyx_L147;
            }

            /* "cyext_acv.pyx":1220
 *                                 ss_d += mean_forest_b[b, leaf_numb, leaf_numb, R_buf[i], 2]/norm_d[b] if norm_d[b] !=0 else 0
 *                         else:
 *                             nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                             nb_leaf_l = leaves_nb[l]
 * 
 */
            /*else*/ {
              __pyx_t_46 = __pyx_v_b;
              __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_46 * __pyx_v_leaves_nb.strides[0]) )));

              /* "cyext_acv.pyx":1221
 *                         else:
 *                             nb_leaf = leaves_nb[b]
 *                             nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                             for leaf_numb in range(nb_leaf):
 */
              __pyx_t_46 = __pyx_v_l;
              __pyx_t_8 = __Pyx_PyInt_From_long((*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_46 * __pyx_v_leaves_nb.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1221, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_XDECREF_SET(__pyx_v_nb_leaf_l, __pyx_t_8);
              __pyx_t_8 = 0;

              /* "cyext_acv.pyx":1223
 *                             nb_leaf_l = leaves_nb[l]
 * 
 *                             for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                                 n = 0
 *                                 n_u = 0
 */
              __pyx_t_39 = __pyx_v_nb_leaf;
              __pyx_t_40 = __pyx_t_39;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_40; __pyx_t_52+=1) {
                __pyx_v_leaf_numb = __pyx_t_52;

                /* "cyext_acv.pyx":1224
 * 
 *                             for leaf_numb in range(nb_leaf):
 *                                 n = 0             # <<<<<<<<<<<<<<
 *                                 n_u = 0
 *                                 n_d = 0
 */
                __pyx_v_n = 0.0;

                /* "cyext_acv.pyx":1225
 *                             for leaf_numb in range(nb_leaf):
 *                                 n = 0
 *                                 n_u = 0             # <<<<<<<<<<<<<<
 *                                 n_d = 0
 *                                 for leaf_numb_l in range(nb_leaf_l):
 */
                __pyx_v_n_u = 0.0;

                /* "cyext_acv.pyx":1226
 *                                 n = 0
 *                                 n_u = 0
 *                                 n_d = 0             # <<<<<<<<<<<<<<
 *                                 for leaf_numb_l in range(nb_leaf_l):
 *                                     n += p_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 */
                __pyx_v_n_d = 0.0;

                /* "cyext_acv.pyx":1227
 *                                 n_u = 0
 *                                 n_d = 0
 *                                 for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 *                                     n += p_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 *                                     n_u += p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 */
                __pyx_t_24 = __Pyx_PyInt_As_long(__pyx_v_nb_leaf_l); if (unlikely((__pyx_t_24 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L1_error)
                __pyx_t_25 = __pyx_t_24;
                for (__pyx_t_53 = 0; __pyx_t_53 < __pyx_t_25; __pyx_t_53+=1) {
                  __pyx_v_leaf_numb_l = __pyx_t_53;

                  /* "cyext_acv.pyx":1228
 *                                 n_d = 0
 *                                 for leaf_numb_l in range(nb_leaf_l):
 *                                     n += p_n[b, leaf_numb, leaf_numb_l, R_buf[i]]             # <<<<<<<<<<<<<<
 *                                     n_u += p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 *                                     n_d += p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 */
                  __pyx_t_46 = __pyx_v_i;
                  __pyx_t_50 = __pyx_v_b;
                  __pyx_t_35 = __pyx_v_leaf_numb;
                  __pyx_t_45 = __pyx_v_leaf_numb_l;
                  __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_50 * __pyx_v_p_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_n.strides[2]) ) + __pyx_t_55 * __pyx_v_p_n.strides[3]) ))));

                  /* "cyext_acv.pyx":1229
 *                                 for leaf_numb_l in range(nb_leaf_l):
 *                                     n += p_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 *                                     n_u += p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]]             # <<<<<<<<<<<<<<
 *                                     n_d += p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 * 
 */
                  __pyx_t_46 = __pyx_v_i;
                  __pyx_t_55 = __pyx_v_b;
                  __pyx_t_45 = __pyx_v_leaf_numb;
                  __pyx_t_35 = __pyx_v_leaf_numb_l;
                  __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_55 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_45 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_35 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_50 * __pyx_v_p_u_n.strides[3]) ))));

                  /* "cyext_acv.pyx":1230
 *                                     n += p_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 *                                     n_u += p_u_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 *                                     n_d += p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]]             # <<<<<<<<<<<<<<
 * 
 *                                 for leaf_numb_l in range(nb_leaf_l):
 */
                  __pyx_t_46 = __pyx_v_i;
                  __pyx_t_50 = __pyx_v_b;
                  __pyx_t_35 = __pyx_v_leaf_numb;
                  __pyx_t_45 = __pyx_v_leaf_numb_l;
                  __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
                  __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_50 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_55 * __pyx_v_p_d_n.strides[3]) ))));
                }

                /* "cyext_acv.pyx":1232
 *                                     n_d += p_d_n[b, leaf_numb, leaf_numb_l, R_buf[i]]
 * 
 *                                 for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 *                                     ss_a += (p_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  R_buf[i], 0]/n) if n*norm[b] !=0  else 0
 *                                     ss_u += (p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1]/n_u) if n_u*norm_u[b] !=0 else 0
 */
                __pyx_t_24 = __Pyx_PyInt_As_long(__pyx_v_nb_leaf_l); if (unlikely((__pyx_t_24 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1232, __pyx_L1_error)
                __pyx_t_25 = __pyx_t_24;
                for (__pyx_t_53 = 0; __pyx_t_53 < __pyx_t_25; __pyx_t_53+=1) {
                  __pyx_v_leaf_numb_l = __pyx_t_53;

                  /* "cyext_acv.pyx":1233
 * 
 *                                 for leaf_numb_l in range(nb_leaf_l):
 *                                     ss_a += (p_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  R_buf[i], 0]/n) if n*norm[b] !=0  else 0             # <<<<<<<<<<<<<<
 *                                     ss_u += (p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1]/n_u) if n_u*norm_u[b] !=0 else 0
 *                                     ss_d += (p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2]/n_d) if n_d*norm_d[b] !=0 else 0
 */
                  __pyx_t_46 = __pyx_v_b;
                  if ((((__pyx_v_n * (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_46 * __pyx_v_norm.strides[0]) )))) != 0.0) != 0)) {
                    __pyx_t_55 = __pyx_v_i;
                    __pyx_t_45 = __pyx_v_b;
                    __pyx_t_35 = __pyx_v_leaf_numb;
                    __pyx_t_50 = __pyx_v_leaf_numb;
                    __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_55 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_48 = __pyx_v_b;
                    __pyx_t_58 = __pyx_v_i;
                    __pyx_t_59 = __pyx_v_b;
                    __pyx_t_60 = __pyx_v_leaf_numb;
                    __pyx_t_61 = __pyx_v_leaf_numb_l;
                    __pyx_t_62 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_58 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_63 = 0;
                    __pyx_t_56 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_45 * __pyx_v_p_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_n.strides[1]) ) + __pyx_t_50 * __pyx_v_p_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm.data + __pyx_t_48 * __pyx_v_norm.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_59 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_60 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_61 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_62 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_63 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n));
                  } else {
                    __pyx_t_56 = 0.0;
                  }
                  __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_56);

                  /* "cyext_acv.pyx":1234
 *                                 for leaf_numb_l in range(nb_leaf_l):
 *                                     ss_a += (p_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  R_buf[i], 0]/n) if n*norm[b] !=0  else 0
 *                                     ss_u += (p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1]/n_u) if n_u*norm_u[b] !=0 else 0             # <<<<<<<<<<<<<<
 *                                     ss_d += (p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2]/n_d) if n_d*norm_d[b] !=0 else 0
 * 
 */
                  __pyx_t_46 = __pyx_v_b;
                  if ((((__pyx_v_n_u * (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_46 * __pyx_v_norm_u.strides[0]) )))) != 0.0) != 0)) {
                    __pyx_t_58 = __pyx_v_i;
                    __pyx_t_63 = __pyx_v_b;
                    __pyx_t_62 = __pyx_v_leaf_numb;
                    __pyx_t_61 = __pyx_v_leaf_numb;
                    __pyx_t_60 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_58 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_59 = __pyx_v_b;
                    __pyx_t_48 = __pyx_v_i;
                    __pyx_t_55 = __pyx_v_b;
                    __pyx_t_47 = __pyx_v_leaf_numb;
                    __pyx_t_50 = __pyx_v_leaf_numb_l;
                    __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_45 = 1;
                    __pyx_t_56 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_63 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_62 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_61 * __pyx_v_p_u_n.strides[2]) ) + __pyx_t_60 * __pyx_v_p_u_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_u.data + __pyx_t_59 * __pyx_v_norm_u.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_55 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_35 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n_u));
                  } else {
                    __pyx_t_56 = 0.0;
                  }
                  __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_56);

                  /* "cyext_acv.pyx":1235
 *                                     ss_a += (p_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l,  R_buf[i], 0]/n) if n*norm[b] !=0  else 0
 *                                     ss_u += (p_u_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_u[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 1]/n_u) if n_u*norm_u[b] !=0 else 0
 *                                     ss_d += (p_d_n[b, leaf_numb, leaf_numb, R_buf[i]]/norm_d[b])*(mean_forest_b[b, leaf_numb, leaf_numb_l, R_buf[i], 2]/n_d) if n_d*norm_d[b] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                  __pyx_t_46 = __pyx_v_b;
                  if ((((__pyx_v_n_d * (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_46 * __pyx_v_norm_d.strides[0]) )))) != 0.0) != 0)) {
                    __pyx_t_48 = __pyx_v_i;
                    __pyx_t_45 = __pyx_v_b;
                    __pyx_t_35 = __pyx_v_leaf_numb;
                    __pyx_t_50 = __pyx_v_leaf_numb;
                    __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_55 = __pyx_v_b;
                    __pyx_t_59 = __pyx_v_i;
                    __pyx_t_58 = __pyx_v_b;
                    __pyx_t_60 = __pyx_v_leaf_numb;
                    __pyx_t_61 = __pyx_v_leaf_numb_l;
                    __pyx_t_62 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_59 * __pyx_v_R_buf.strides[0]) )));
                    __pyx_t_63 = 2;
                    __pyx_t_56 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_45 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_35 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_50 * __pyx_v_p_d_n.strides[2]) ) + __pyx_t_47 * __pyx_v_p_d_n.strides[3]) ))) / (*((double *) ( /* dim=0 */ (__pyx_v_norm_d.data + __pyx_t_55 * __pyx_v_norm_d.strides[0]) )))) * ((*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_58 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_60 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_61 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_62 * __pyx_v_mean_forest_b.strides[3]) ) + __pyx_t_63 * __pyx_v_mean_forest_b.strides[4]) ))) / __pyx_v_n_d));
                  } else {
                    __pyx_t_56 = 0.0;
                  }
                  __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_56);
                }
              }
            }
            __pyx_L147:;
          }
        }

        /* "cyext_acv.pyx":1238
 * 
 * 
 *                 ss = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 */
        if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
          __pyx_t_56 = ((__pyx_v_ss_u - __pyx_v_ss_a) / (__pyx_v_ss_u - __pyx_v_ss_d));
        } else {
          __pyx_t_56 = 0.0;
        }
        __pyx_v_ss = __pyx_t_56;

        /* "cyext_acv.pyx":1239
 * 
 *                 ss = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_46 = __pyx_v_i;
        __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_23 = ((__pyx_v_ss >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_59 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_23) {

          /* "cyext_acv.pyx":1240
 *                 ss = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_59 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

          /* "cyext_acv.pyx":1241
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_59 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":1242
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_36 = __pyx_v_S_size;
          __pyx_t_37 = __pyx_t_36;
          for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
            __pyx_v_s = __pyx_t_38;

            /* "cyext_acv.pyx":1243
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_46 = __pyx_v_s;
            __pyx_t_59 = __pyx_v_i;
            __pyx_t_63 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_59 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_62 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_63 * __pyx_v_s_star.strides[0]) ) + __pyx_t_62 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":1239
 * 
 *                 ss = (ss_u - ss_a)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":1245
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_23 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_23) {

          /* "cyext_acv.pyx":1246
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_59 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":1247
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_59 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":1248
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_36 = __pyx_v_S_size;
          __pyx_t_37 = __pyx_t_36;
          for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
            __pyx_v_s = __pyx_t_38;

            /* "cyext_acv.pyx":1249
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_46 = __pyx_v_s;
            __pyx_t_59 = __pyx_v_i;
            __pyx_t_62 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_59 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_63 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_62 * __pyx_v_s_star.strides[0]) ) + __pyx_t_63 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":1250
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_18 = (__pyx_v_X.shape[1]);
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_19 = __pyx_t_18;
          for (__pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_59 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_36 < __pyx_t_19; __pyx_t_36+=1) {
            __pyx_v_s = __pyx_t_36;

            /* "cyext_acv.pyx":1251
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_63 = __pyx_v_i;
            __pyx_t_62 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_63 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_61 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_62 * __pyx_v_s_star.strides[0]) ) + __pyx_t_61 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":1245
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":1253
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_13 = __pyx_v_N;
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;

      /* "cyext_acv.pyx":1254
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_46 = __pyx_v_i;
      __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_23 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_59 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":1255
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1255, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1256
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_46 = __pyx_v_i;
        __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_24 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_59 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_25 = __pyx_t_24;
        for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
          __pyx_v_s = __pyx_t_26;

          /* "cyext_acv.pyx":1257
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_46 = __pyx_v_i;
          __pyx_t_59 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_46 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_63 = __pyx_v_s;
          __pyx_t_61 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_59 * __pyx_v_s_star.strides[0]) ) + __pyx_t_63 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_61 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":1254
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":1259
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_33 = __pyx_v_r.size();
    __pyx_t_34 = __pyx_t_33;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_34; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;

      /* "cyext_acv.pyx":1260
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":1261
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":1263
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_16 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_16) {
    } else {
      goto __pyx_L175_next_and;
    }
    __pyx_t_16 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_16) {
    } else {
      __pyx_t_23 = __pyx_t_16;
      goto __pyx_L174_bool_binop_done;
    }
    __pyx_L175_next_and:;
    __pyx_t_16 = (__pyx_v_stop != 0);
    __pyx_t_23 = __pyx_t_16;
    __pyx_L174_bool_binop_done:;
    if (__pyx_t_23) {

      /* "cyext_acv.pyx":1264
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L21_break;

      /* "cyext_acv.pyx":1263
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":1051
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L21_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":1266
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_long); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_long); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":968
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_30, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_51, 1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm_u, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_norm_d, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XDECREF(__pyx_v_nb_leaf_l);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_19global_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_19global_sdp_reg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_tX;
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_reg (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_tX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 1); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 2); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 3); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 4); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 5); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 6); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 7); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 8); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 9); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 10); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 11); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 12); __PYX_ERR(0, 968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, 13); __PYX_ERR(0, 968, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_reg") < 0)) __PYX_ERR(0, 968, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_tX = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tX == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 971, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 971, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 971, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[10]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[13]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_reg", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 968, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 971, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_18global_sdp_reg(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_18global_sdp_reg(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 968, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 968, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_reg(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":1272
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_leaves_pa(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_21shap_values_leaves_pa(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_leaves_pa(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long const __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  const __pyx_v_node_idx_trees, std::vector<std::vector<long> >  const __pyx_v_C, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_ns;
  unsigned int __pyx_v_add;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_i2;
  unsigned int __pyx_v_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_csi;
  unsigned int __pyx_v_cs;
  double __pyx_v_lm;
  std::vector<std::vector<long> >  __pyx_v_C_buff;
  std::vector<std::vector<long> >  __pyx_v_va_id;
  std::vector<std::vector<long> >  __pyx_v_buff;
  std::vector<long>  __pyx_v_coal_va;
  std::vector<long>  __pyx_v_remove_va;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id_v2;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_C_b;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_Sm;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id;
  std::vector<std::vector<std::vector<std::vector<long> > > >  __pyx_v_node_id_va;
  int __pyx_v_S_size;
  int __pyx_v_va_size;
  int __pyx_v_b_it;
  int __pyx_v_nv_bool;
  long __pyx_v_set_size;
  double __pyx_v_p_s;
  double __pyx_v_p_si;
  double __pyx_v_coef;
  double __pyx_v_coef_0;
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_si = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  std::vector<std::vector<long> > ::size_type __pyx_t_16;
  std::vector<std::vector<long> > ::size_type __pyx_t_17;
  std::vector<long> ::size_type __pyx_t_18;
  std::vector<long> ::size_type __pyx_t_19;
  unsigned int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  unsigned int __pyx_t_25;
  unsigned int __pyx_t_26;
  unsigned int __pyx_t_27;
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  size_t __pyx_t_35;
  size_t __pyx_t_36;
  size_t __pyx_t_37;
  size_t __pyx_t_38;
  size_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  int __pyx_t_41;
  std::vector<long> ::size_type __pyx_t_42;
  std::vector<long> ::size_type __pyx_t_43;
  long __pyx_t_44;
  long __pyx_t_45;
  int __pyx_t_46;
  int __pyx_t_47;
  size_t __pyx_t_48;
  size_t __pyx_t_49;
  size_t __pyx_t_50;
  size_t __pyx_t_51;
  size_t __pyx_t_52;
  size_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  size_t __pyx_t_56;
  size_t __pyx_t_57;
  size_t __pyx_t_58;
  size_t __pyx_t_59;
  size_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  size_t __pyx_t_62;
  size_t __pyx_t_63;
  size_t __pyx_t_64;
  size_t __pyx_t_65;
  size_t __pyx_t_66;
  size_t __pyx_t_67;
  Py_ssize_t __pyx_t_68;
  Py_ssize_t __pyx_t_69;
  double __pyx_t_70;
  unsigned int __pyx_t_71;
  unsigned int __pyx_t_72;
  unsigned int __pyx_t_73;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_leaves_pa", 0);

  /* "cyext_acv.pyx":1283
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":1284
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":1285
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":1286
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":1287
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] phi
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":1290
 * 
 *     cdef double[:, :, :] phi
 *     phi = np.zeros((N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1293
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 *     phi_b = np.zeros((max_leaves, 2**scaling, N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :, :, ::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_8);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_b = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1296
 * 
 *     cdef long[:, :, :, ::1] S
 *     S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_S = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1304
 *     cdef vector[vector[long]] C_buff, va_id, buff
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     for i in range(max_leaves):
 *         buff[i].resize(1)
 */
  try {
    __pyx_v_buff.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1304, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1305
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):             # <<<<<<<<<<<<<<
 *         buff[i].resize(1)
 * 
 */
  __pyx_t_12 = __pyx_v_max_leaves;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":1306
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):
 *         buff[i].resize(1)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 */
    try {
      (__pyx_v_buff[__pyx_v_i]).resize(1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1306, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1309
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 */
  try {
    __pyx_v_node_id_v2.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1309, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1310
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)
 */
  try {
    __pyx_v_C_b.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1310, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1311
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     node_id.resize(max_leaves)
 * 
 */
  try {
    __pyx_v_Sm.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1311, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1312
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 */
  try {
    __pyx_v_node_id.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1312, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1315
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)             # <<<<<<<<<<<<<<
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)
 */
  try {
    __pyx_v_node_id_va.resize(__pyx_v_m);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1315, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1316
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         node_id_va[i].resize(max_leaves)
 * 
 */
  __pyx_t_12 = __pyx_v_m;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":1317
 *     node_id_va.resize(m)
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef int S_size, va_size,  b_it, nv_bool
 */
    try {
      (__pyx_v_node_id_va[__pyx_v_i]).resize(__pyx_v_max_leaves);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1317, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1323
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
  __pyx_t_15 = (((__pyx_v_C[0]).size() != 0) != 0);
  if (__pyx_t_15) {

    /* "cyext_acv.pyx":1324
 * 
 *     if C[0].size() != 0:
 *         C_buff = C             # <<<<<<<<<<<<<<
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 */
    __pyx_v_C_buff = __pyx_v_C;

    /* "cyext_acv.pyx":1325
 *     if C[0].size() != 0:
 *         C_buff = C
 *         for ci in range(C.size()):             # <<<<<<<<<<<<<<
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])
 */
    __pyx_t_16 = __pyx_v_C.size();
    __pyx_t_17 = __pyx_t_16;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_17; __pyx_t_12+=1) {
      __pyx_v_ci = __pyx_t_12;

      /* "cyext_acv.pyx":1326
 *         C_buff = C
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):             # <<<<<<<<<<<<<<
 *                 coal_va.push_back(C[ci][cj])
 * 
 */
      __pyx_t_18 = (__pyx_v_C[__pyx_v_ci]).size();
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_19; __pyx_t_13+=1) {
        __pyx_v_cj = __pyx_t_13;

        /* "cyext_acv.pyx":1327
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(m):
 */
        try {
          __pyx_v_coal_va.push_back(((__pyx_v_C[__pyx_v_ci])[__pyx_v_cj]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1327, __pyx_L1_error)
        }
      }
    }

    /* "cyext_acv.pyx":1329
 *                 coal_va.push_back(C[ci][cj])
 * 
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 */
    __pyx_t_12 = __pyx_v_m;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "cyext_acv.pyx":1330
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
      __pyx_t_15 = ((!((std::find<std::vector<long> ::iterator,long>(__pyx_v_coal_va.begin(), __pyx_v_coal_va.end(), __pyx_v_i) != __pyx_v_coal_va.end()) != 0)) != 0);
      if (__pyx_t_15) {

        /* "cyext_acv.pyx":1331
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)             # <<<<<<<<<<<<<<
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])
 */
        try {
          __pyx_v_remove_va.push_back(__pyx_v_i);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1331, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1332
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i             # <<<<<<<<<<<<<<
 *                 va_id.push_back(buff[0])
 *             else:
 */
        ((__pyx_v_buff[0])[0]) = __pyx_v_i;

        /* "cyext_acv.pyx":1333
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 for ci in range(C_buff.size()):
 */
        try {
          __pyx_v_va_id.push_back((__pyx_v_buff[0]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1333, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1330
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
        goto __pyx_L14;
      }

      /* "cyext_acv.pyx":1335
 *                 va_id.push_back(buff[0])
 *             else:
 *                 for ci in range(C_buff.size()):             # <<<<<<<<<<<<<<
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 */
      /*else*/ {
        __pyx_t_16 = __pyx_v_C_buff.size();
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_17; __pyx_t_20+=1) {
          __pyx_v_ci = __pyx_t_20;

          /* "cyext_acv.pyx":1336
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          __pyx_t_15 = ((std::find<std::vector<long> ::iterator,long>((__pyx_v_C_buff[__pyx_v_ci]).begin(), (__pyx_v_C_buff[__pyx_v_ci]).end(), __pyx_v_i) != (__pyx_v_C_buff[__pyx_v_ci]).end()) != 0);
          if (__pyx_t_15) {

            /* "cyext_acv.pyx":1337
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])             # <<<<<<<<<<<<<<
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 */
            try {
              __pyx_v_va_id.push_back((__pyx_v_C_buff[__pyx_v_ci]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 1337, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":1338
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])             # <<<<<<<<<<<<<<
 *                         C_buff.pop_back()
 *                         break
 */
            (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(__pyx_v_C_buff.begin(), __pyx_v_C_buff.end(), (__pyx_v_C_buff[__pyx_v_ci])));

            /* "cyext_acv.pyx":1339
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_v_C_buff.pop_back();

            /* "cyext_acv.pyx":1340
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 *                         break             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L16_break;

            /* "cyext_acv.pyx":1336
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          }
        }
        __pyx_L16_break:;
      }
      __pyx_L14:;
    }

    /* "cyext_acv.pyx":1323
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
    goto __pyx_L7;
  }

  /* "cyext_acv.pyx":1343
 * 
 *     else:
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])
 */
  /*else*/ {
    __pyx_t_12 = __pyx_v_m;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "cyext_acv.pyx":1344
 *     else:
 *         for i in range(m):
 *             buff[0][0] = i             # <<<<<<<<<<<<<<
 *             va_id.push_back(buff[0])
 * 
 */
      ((__pyx_v_buff[0])[0]) = __pyx_v_i;

      /* "cyext_acv.pyx":1345
 *         for i in range(m):
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef double p_s, p_si, coef, coef_0
 */
      try {
        __pyx_v_va_id.push_back((__pyx_v_buff[0]));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1345, __pyx_L1_error)
      }
    }
  }
  __pyx_L7:;

  /* "cyext_acv.pyx":1349
 *     cdef double p_s, p_si, coef, coef_0
 *     cdef double[:, :, :] lm_s, lm_si
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_8);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_9 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lm_s = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1350
 *     cdef double[:, :, :] lm_s, lm_si
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lm_si = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1352
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 */
  __pyx_t_12 = __pyx_v_n_trees;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_b = __pyx_t_14;

    /* "cyext_acv.pyx":1353
 * 
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):
 */
    __pyx_t_21 = (__pyx_v_phi_b.shape[0]);
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_22; __pyx_t_20+=1) {
      __pyx_v_leaf_numb = __pyx_t_20;

      /* "cyext_acv.pyx":1354
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                 for i in range(N):
 *                     for j in range(m):
 */
      __pyx_t_23 = (__pyx_v_phi_b.shape[1]);
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_24; __pyx_t_25+=1) {
        __pyx_v_counter = __pyx_t_25;

        /* "cyext_acv.pyx":1355
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):             # <<<<<<<<<<<<<<
 *                     for j in range(m):
 *                         for i2 in range(d):
 */
        __pyx_t_26 = __pyx_v_N;
        __pyx_t_27 = __pyx_t_26;
        for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
          __pyx_v_i = __pyx_t_28;

          /* "cyext_acv.pyx":1356
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):
 *                     for j in range(m):             # <<<<<<<<<<<<<<
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 */
          __pyx_t_29 = __pyx_v_m;
          __pyx_t_30 = __pyx_t_29;
          for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
            __pyx_v_j = __pyx_t_31;

            /* "cyext_acv.pyx":1357
 *                 for i in range(N):
 *                     for j in range(m):
 *                         for i2 in range(d):             # <<<<<<<<<<<<<<
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 */
            __pyx_t_32 = __pyx_v_d;
            __pyx_t_33 = __pyx_t_32;
            for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
              __pyx_v_i2 = __pyx_t_34;

              /* "cyext_acv.pyx":1358
 *                     for j in range(m):
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
              __pyx_t_35 = __pyx_v_leaf_numb;
              __pyx_t_36 = __pyx_v_counter;
              __pyx_t_37 = __pyx_v_i;
              __pyx_t_38 = __pyx_v_j;
              __pyx_t_39 = __pyx_v_i2;
              *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_35 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_36 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_37 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[4]) )) = 0.0;
            }
          }
        }
      }
    }

    /* "cyext_acv.pyx":1359
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 */
    __pyx_t_39 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ ((char *) (((long const  *) __pyx_v_leaves_nb.data) + __pyx_t_39)) )));

    /* "cyext_acv.pyx":1360
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_20 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              int __pyx_parallel_temp0 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp1 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp2 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp3 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp4 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp5 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp6 = ((double)__PYX_NAN());
              double __pyx_parallel_temp7 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp8 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp9 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp10 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp11 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp12 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp13 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp14 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp15 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp16 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp17 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp18 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp19 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp20 = ((double)__PYX_NAN());
              double __pyx_parallel_temp21 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp22 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp23 = ((unsigned int)0xbad0bad0);
              long __pyx_parallel_temp24 = ((long)0xbad0bad0);
              unsigned int __pyx_parallel_temp25 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp26 = ((int)0xbad0bad0);
              const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
              PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
              int __pyx_parallel_why;
              __pyx_parallel_why = 0;
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_26 = (__pyx_t_20 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_26 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57, __pyx_t_58, __pyx_t_59, __pyx_t_60, __pyx_t_61, __pyx_t_62, __pyx_t_63, __pyx_t_64, __pyx_t_65, __pyx_t_66, __pyx_t_67, __pyx_t_68, __pyx_t_69, __pyx_t_70, __pyx_t_71, __pyx_t_72, __pyx_t_73) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      Py_BEGIN_ALLOW_THREADS
                      #endif /* _OPENMP */
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_S_size) lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_add) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_ci) lastprivate(__pyx_v_cj) lastprivate(__pyx_v_coef) lastprivate(__pyx_v_coef_0) lastprivate(__pyx_v_counter) lastprivate(__pyx_v_cs) lastprivate(__pyx_v_csi) lastprivate(__pyx_v_i) lastprivate(__pyx_v_i2) lastprivate(__pyx_v_l) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_ns) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_si) lastprivate(__pyx_v_pow_set_size) lastprivate(__pyx_v_s) lastprivate(__pyx_v_set_size) lastprivate(__pyx_v_va) lastprivate(__pyx_v_va_size) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_26; __pyx_t_25++){
                          if (__pyx_parallel_why < 2)
                          {
                              __pyx_v_leaf_numb = (unsigned int)(0 + 1 * __pyx_t_25);
                              /* Initialize private variables to invalid values */
                              __pyx_v_S_size = ((int)0xbad0bad0);
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_add = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((int)0xbad0bad0);
                              __pyx_v_ci = ((unsigned int)0xbad0bad0);
                              __pyx_v_cj = ((unsigned int)0xbad0bad0);
                              __pyx_v_coef = ((double)__PYX_NAN());
                              __pyx_v_coef_0 = ((double)__PYX_NAN());
                              __pyx_v_counter = ((unsigned int)0xbad0bad0);
                              __pyx_v_cs = ((unsigned int)0xbad0bad0);
                              __pyx_v_csi = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((unsigned int)0xbad0bad0);
                              __pyx_v_i2 = ((unsigned int)0xbad0bad0);
                              __pyx_v_l = ((unsigned int)0xbad0bad0);
                              __pyx_v_lm = ((double)__PYX_NAN());
                              __pyx_v_ns = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv_bool = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_s = ((double)__PYX_NAN());
                              __pyx_v_p_si = ((double)__PYX_NAN());
                              __pyx_v_pow_set_size = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_set_size = ((long)0xbad0bad0);
                              __pyx_v_va = ((unsigned int)0xbad0bad0);
                              __pyx_v_va_size = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":1361
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *             lm = 0
 *             for i in range(data.shape[0]):
 */
                              (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).clear();

                              /* "cyext_acv.pyx":1362
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 */
                              __pyx_v_lm = 0.0;

                              /* "cyext_acv.pyx":1363
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 */
                              __pyx_t_21 = (__pyx_v_data.shape[0]);
                              __pyx_t_22 = __pyx_t_21;
                              for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_22; __pyx_t_27+=1) {
                                __pyx_v_i = __pyx_t_27;

                                /* "cyext_acv.pyx":1364
 *             lm = 0
 *             for i in range(data.shape[0]):
 *                 a_it = 0             # <<<<<<<<<<<<<<
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 */
                                __pyx_v_a_it = 0;

                                /* "cyext_acv.pyx":1365
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 *                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 */
                                __pyx_t_23 = (__pyx_v_data.shape[1]);
                                __pyx_t_24 = __pyx_t_23;
                                for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_24; __pyx_t_28+=1) {
                                  __pyx_v_s = __pyx_t_28;

                                  /* "cyext_acv.pyx":1366
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  __pyx_t_39 = __pyx_v_i;
                                  __pyx_t_38 = __pyx_v_s;
                                  __pyx_t_37 = __pyx_v_b;
                                  __pyx_t_36 = __pyx_v_leaf_numb;
                                  __pyx_t_35 = __pyx_v_s;
                                  __pyx_t_40 = 1;
                                  __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_39 * __pyx_v_data.strides[0]) ) + __pyx_t_38 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_36 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_41) {
                                  } else {
                                    __pyx_t_15 = __pyx_t_41;
                                    goto __pyx_L46_bool_binop_done;
                                  }
                                  __pyx_t_35 = __pyx_v_i;
                                  __pyx_t_36 = __pyx_v_s;
                                  __pyx_t_37 = __pyx_v_b;
                                  __pyx_t_38 = __pyx_v_leaf_numb;
                                  __pyx_t_39 = __pyx_v_s;
                                  __pyx_t_40 = 0;
                                  __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_36 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_15 = __pyx_t_41;
                                  __pyx_L46_bool_binop_done:;
                                  if (__pyx_t_15) {

                                    /* "cyext_acv.pyx":1367
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1
 */
                                    __pyx_v_a_it = (__pyx_v_a_it + 1);

                                    /* "cyext_acv.pyx":1366
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1368
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                __pyx_t_15 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                if (__pyx_t_15) {

                                  /* "cyext_acv.pyx":1369
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 *             if C[0].size() != 0:
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1.0);

                                  /* "cyext_acv.pyx":1368
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":1371
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                              __pyx_t_15 = (((__pyx_v_C[0]).size() != 0) != 0);
                              if (__pyx_t_15) {

                                /* "cyext_acv.pyx":1372
 * 
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C             # <<<<<<<<<<<<<<
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 */
                                (__pyx_v_C_b[__pyx_v_leaf_numb]) = __pyx_v_C;

                                /* "cyext_acv.pyx":1373
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 */
                                __pyx_t_18 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_19 = __pyx_t_18;
                                for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_19; __pyx_t_27+=1) {
                                  __pyx_v_nv = __pyx_t_27;

                                  /* "cyext_acv.pyx":1374
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0             # <<<<<<<<<<<<<<
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 */
                                  __pyx_v_add = 0;

                                  /* "cyext_acv.pyx":1375
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 *                     for ns in range(remove_va.size()):             # <<<<<<<<<<<<<<
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 */
                                  __pyx_t_42 = __pyx_v_remove_va.size();
                                  __pyx_t_43 = __pyx_t_42;
                                  for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_43; __pyx_t_28+=1) {
                                    __pyx_v_ns = __pyx_t_28;

                                    /* "cyext_acv.pyx":1376
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    __pyx_t_15 = (((((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]) == (__pyx_v_remove_va[__pyx_v_ns])) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1377
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1             # <<<<<<<<<<<<<<
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 */
                                      __pyx_v_add = 1;

                                      /* "cyext_acv.pyx":1378
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]             # <<<<<<<<<<<<<<
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 */
                                      ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]);

                                      /* "cyext_acv.pyx":1379
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 *                             break
 *                     if add == 0:
 */
                                      try {
                                        (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                      } catch(...) {
                                        #ifdef WITH_THREAD
                                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                        #endif
                                        __Pyx_CppExn2PyErr();
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        __PYX_ERR(0, 1379, __pyx_L39_error)
                                      }

                                      /* "cyext_acv.pyx":1380
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break             # <<<<<<<<<<<<<<
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 */
                                      goto __pyx_L53_break;

                                      /* "cyext_acv.pyx":1376
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    }
                                  }
                                  __pyx_L53_break:;

                                  /* "cyext_acv.pyx":1381
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  __pyx_t_15 = ((__pyx_v_add == 0) != 0);
                                  if (__pyx_t_15) {

                                    /* "cyext_acv.pyx":1382
 *                             break
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 */
                                    __pyx_t_16 = (__pyx_v_C_b[__pyx_v_leaf_numb]).size();
                                    __pyx_t_17 = __pyx_t_16;
                                    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_17; __pyx_t_28+=1) {
                                      __pyx_v_ci = __pyx_t_28;

                                      /* "cyext_acv.pyx":1383
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 */
                                      __pyx_t_42 = ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_43 = __pyx_t_42;
                                      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_43; __pyx_t_29+=1) {
                                        __pyx_v_cj = __pyx_t_29;

                                        /* "cyext_acv.pyx":1384
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        __pyx_t_15 = (((((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]) == (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv])) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1385
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1             # <<<<<<<<<<<<<<
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 */
                                          __pyx_v_add = 1;

                                          /* "cyext_acv.pyx":1386
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                     break
 *                             if add == 1:
 */
                                          try {
                                            (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back(((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]));
                                          } catch(...) {
                                            #ifdef WITH_THREAD
                                            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                            #endif
                                            __Pyx_CppExn2PyErr();
                                            #ifdef WITH_THREAD
                                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                            #endif
                                            __PYX_ERR(0, 1386, __pyx_L39_error)
                                          }

                                          /* "cyext_acv.pyx":1387
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break             # <<<<<<<<<<<<<<
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 */
                                          goto __pyx_L59_break;

                                          /* "cyext_acv.pyx":1384
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        }
                                      }
                                      __pyx_L59_break:;

                                      /* "cyext_acv.pyx":1388
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      __pyx_t_15 = ((__pyx_v_add == 1) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1389
 *                                     break
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 */
                                        (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >((__pyx_v_C_b[__pyx_v_leaf_numb]).begin(), (__pyx_v_C_b[__pyx_v_leaf_numb]).end(), ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])));

                                        /* "cyext_acv.pyx":1390
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 */
                                        (__pyx_v_C_b[__pyx_v_leaf_numb]).pop_back();

                                        /* "cyext_acv.pyx":1391
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 *                                 break             # <<<<<<<<<<<<<<
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 */
                                        goto __pyx_L57_break;

                                        /* "cyext_acv.pyx":1388
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      }
                                    }
                                    __pyx_L57_break:;

                                    /* "cyext_acv.pyx":1381
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1392
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]             # <<<<<<<<<<<<<<
 *             else:
 *                 node_id[leaf_numb].clear()
 */
                                (__pyx_v_node_id[__pyx_v_leaf_numb]) = (__pyx_v_node_id_v2[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1371
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                                goto __pyx_L49;
                              }

                              /* "cyext_acv.pyx":1394
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 *                 node_id[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 */
                              /*else*/ {
                                (__pyx_v_node_id[__pyx_v_leaf_numb]).clear();

                                /* "cyext_acv.pyx":1395
 *             else:
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 */
                                __pyx_t_18 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_19 = __pyx_t_18;
                                for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_19; __pyx_t_27+=1) {
                                  __pyx_v_i = __pyx_t_27;

                                  /* "cyext_acv.pyx":1396
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]             # <<<<<<<<<<<<<<
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 */
                                  ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_i]);

                                  /* "cyext_acv.pyx":1397
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  try {
                                    (__pyx_v_node_id[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                  } catch(...) {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    __Pyx_CppExn2PyErr();
                                    #ifdef WITH_THREAD
                                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                    #endif
                                    __PYX_ERR(0, 1397, __pyx_L39_error)
                                  }
                                }
                              }
                              __pyx_L49:;

                              /* "cyext_acv.pyx":1400
 * 
 * 
 *             for va in range(va_id.size()):             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 */
                              __pyx_t_16 = __pyx_v_va_id.size();
                              __pyx_t_17 = __pyx_t_16;
                              for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_17; __pyx_t_27+=1) {
                                __pyx_v_va = __pyx_t_27;

                                /* "cyext_acv.pyx":1401
 * 
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]             # <<<<<<<<<<<<<<
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]) = (__pyx_v_node_id[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1402
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_15 = ((!((std::find<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])) != ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end()) != 0)) != 0);
                                if (__pyx_t_15) {

                                  /* "cyext_acv.pyx":1403
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 */
                                  goto __pyx_L64_continue;

                                  /* "cyext_acv.pyx":1402
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":1405
 *                     continue
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 */
                                (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])));

                                /* "cyext_acv.pyx":1406
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 *                 node_id_va[va][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).pop_back();

                                /* "cyext_acv.pyx":1408
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()             # <<<<<<<<<<<<<<
 *                 pow_set_size = 2**set_size
 * 
 */
                                __pyx_v_set_size = ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).size();

                                /* "cyext_acv.pyx":1409
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 *                 pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *                 for counter in range(0, pow_set_size):
 */
                                __pyx_v_pow_set_size = __Pyx_pow_long(2, __pyx_v_set_size);

                                /* "cyext_acv.pyx":1411
 *                 pow_set_size = 2**set_size
 * 
 *                 for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *                     va_size = 0
 *                     S_size = 0
 */
                                __pyx_t_28 = __pyx_v_pow_set_size;
                                __pyx_t_29 = __pyx_t_28;
                                for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                                  __pyx_v_counter = __pyx_t_30;

                                  /* "cyext_acv.pyx":1412
 * 
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0             # <<<<<<<<<<<<<<
 *                     S_size = 0
 *                     for ci in range(set_size):
 */
                                  __pyx_v_va_size = 0;

                                  /* "cyext_acv.pyx":1413
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0
 *                     S_size = 0             # <<<<<<<<<<<<<<
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 */
                                  __pyx_v_S_size = 0;

                                  /* "cyext_acv.pyx":1414
 *                     va_size = 0
 *                     S_size = 0
 *                     for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 */
                                  __pyx_t_44 = __pyx_v_set_size;
                                  __pyx_t_45 = __pyx_t_44;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_45; __pyx_t_31+=1) {
                                    __pyx_v_ci = __pyx_t_31;

                                    /* "cyext_acv.pyx":1415
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    __pyx_t_15 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1416
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 */
                                      __pyx_t_18 = (((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_19; __pyx_t_32+=1) {
                                        __pyx_v_cj = __pyx_t_32;

                                        /* "cyext_acv.pyx":1417
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]             # <<<<<<<<<<<<<<
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1
 */
                                        __pyx_t_39 = __pyx_v_va;
                                        __pyx_t_38 = __pyx_v_counter;
                                        __pyx_t_37 = __pyx_v_leaf_numb;
                                        __pyx_t_40 = __pyx_v_S_size;
                                        *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_39 * __pyx_v_S.strides[0]) ) + __pyx_t_38 * __pyx_v_S.strides[1]) ) + __pyx_t_37 * __pyx_v_S.strides[2]) )) + __pyx_t_40)) )) = ((((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]);

                                        /* "cyext_acv.pyx":1418
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                             va_size = va_size + 1
 * 
 */
                                        __pyx_v_S_size = (__pyx_v_S_size + 1);
                                      }

                                      /* "cyext_acv.pyx":1419
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 */
                                      __pyx_v_va_size = (__pyx_v_va_size + 1);

                                      /* "cyext_acv.pyx":1415
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1421
 *                             va_size = va_size + 1
 * 
 *                     lm_s[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):
 */
                                  __pyx_t_37 = __pyx_v_va;
                                  __pyx_t_38 = __pyx_v_leaf_numb;
                                  __pyx_t_39 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_37 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_39 * __pyx_v_lm_s.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":1422
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 */
                                  __pyx_t_39 = __pyx_v_va;
                                  __pyx_t_38 = __pyx_v_leaf_numb;
                                  __pyx_t_37 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_39 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_37 * __pyx_v_lm_si.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":1423
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(S_size):
 */
                                  __pyx_t_21 = (__pyx_v_data.shape[0]);
                                  __pyx_t_22 = __pyx_t_21;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_22; __pyx_t_31+=1) {
                                    __pyx_v_i = __pyx_t_31;

                                    /* "cyext_acv.pyx":1424
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":1425
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_46 = __pyx_v_S_size;
                                    __pyx_t_47 = __pyx_t_46;
                                    for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_47; __pyx_t_32+=1) {
                                      __pyx_v_s = __pyx_t_32;

                                      /* "cyext_acv.pyx":1426
 *                         b_it = 0
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_37 = __pyx_v_va;
                                      __pyx_t_38 = __pyx_v_counter;
                                      __pyx_t_39 = __pyx_v_leaf_numb;
                                      __pyx_t_36 = __pyx_v_s;
                                      __pyx_t_35 = __pyx_v_i;
                                      __pyx_t_40 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_37 * __pyx_v_S.strides[0]) ) + __pyx_t_38 * __pyx_v_S.strides[1]) ) + __pyx_t_39 * __pyx_v_S.strides[2]) )) + __pyx_t_36)) )));
                                      __pyx_t_48 = __pyx_v_va;
                                      __pyx_t_49 = __pyx_v_counter;
                                      __pyx_t_50 = __pyx_v_leaf_numb;
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_52 = __pyx_v_b;
                                      __pyx_t_53 = __pyx_v_leaf_numb;
                                      __pyx_t_54 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) ) + __pyx_t_49 * __pyx_v_S.strides[1]) ) + __pyx_t_50 * __pyx_v_S.strides[2]) )) + __pyx_t_51)) )));
                                      __pyx_t_55 = 1;
                                      __pyx_t_56 = __pyx_v_va;
                                      __pyx_t_57 = __pyx_v_counter;
                                      __pyx_t_58 = __pyx_v_leaf_numb;
                                      __pyx_t_59 = __pyx_v_s;
                                      __pyx_t_60 = __pyx_v_i;
                                      __pyx_t_61 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_56 * __pyx_v_S.strides[0]) ) + __pyx_t_57 * __pyx_v_S.strides[1]) ) + __pyx_t_58 * __pyx_v_S.strides[2]) )) + __pyx_t_59)) )));
                                      __pyx_t_62 = __pyx_v_va;
                                      __pyx_t_63 = __pyx_v_counter;
                                      __pyx_t_64 = __pyx_v_leaf_numb;
                                      __pyx_t_65 = __pyx_v_s;
                                      __pyx_t_66 = __pyx_v_b;
                                      __pyx_t_67 = __pyx_v_leaf_numb;
                                      __pyx_t_68 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_62 * __pyx_v_S.strides[0]) ) + __pyx_t_63 * __pyx_v_S.strides[1]) ) + __pyx_t_64 * __pyx_v_S.strides[2]) )) + __pyx_t_65)) )));
                                      __pyx_t_69 = 0;
                                      __pyx_t_15 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_40 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_60 * __pyx_v_data.strides[0]) ) + __pyx_t_61 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_67 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1427
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == S_size:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":1426
 *                         b_it = 0
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1429
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 */
                                    __pyx_t_15 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1430
 * 
 *                         if b_it == S_size:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                             nv_bool = 0
 */
                                      __pyx_t_65 = __pyx_v_va;
                                      __pyx_t_64 = __pyx_v_leaf_numb;
                                      __pyx_t_63 = __pyx_v_counter;
                                      __pyx_t_62 = __pyx_v_va;
                                      __pyx_t_67 = __pyx_v_leaf_numb;
                                      __pyx_t_66 = __pyx_v_counter;
                                      *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_62 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_67 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_66 * __pyx_v_lm_s.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_65 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_64 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_63 * __pyx_v_lm_s.strides[2]) ))) + 1.0);

                                      /* "cyext_acv.pyx":1432
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":1433
 * 
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_19; __pyx_t_32+=1) {
                                        __pyx_v_nv = __pyx_t_32;

                                        /* "cyext_acv.pyx":1434
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_63 = __pyx_v_i;
                                        __pyx_t_69 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_64 = __pyx_v_b;
                                        __pyx_t_65 = __pyx_v_leaf_numb;
                                        __pyx_t_68 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_61 = 1;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_63 * __pyx_v_data.strides[0]) ) + __pyx_t_69 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_65 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_61 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_41) {
                                        } else {
                                          __pyx_t_15 = __pyx_t_41;
                                          goto __pyx_L83_bool_binop_done;
                                        }
                                        __pyx_t_65 = __pyx_v_i;
                                        __pyx_t_61 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_64 = __pyx_v_b;
                                        __pyx_t_63 = __pyx_v_leaf_numb;
                                        __pyx_t_68 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_69 = 0;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_65 * __pyx_v_data.strides[0]) ) + __pyx_t_61 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_15 = __pyx_t_41;
                                        __pyx_L83_bool_binop_done:;
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1435
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":1436
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L80_continue;

                                          /* "cyext_acv.pyx":1434
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L80_continue:;
                                      }

                                      /* "cyext_acv.pyx":1438
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      __pyx_t_15 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1439
 * 
 *                             if nv_bool == 0:
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(N):
 */
                                        __pyx_t_63 = __pyx_v_va;
                                        __pyx_t_64 = __pyx_v_leaf_numb;
                                        __pyx_t_65 = __pyx_v_counter;
                                        __pyx_t_66 = __pyx_v_va;
                                        __pyx_t_67 = __pyx_v_leaf_numb;
                                        __pyx_t_62 = __pyx_v_counter;
                                        *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_66 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_67 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_62 * __pyx_v_lm_si.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_63 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_64 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_65 * __pyx_v_lm_si.strides[2]) ))) + 1.0);

                                        /* "cyext_acv.pyx":1438
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":1429
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1441
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 * 
 *                         csi = 0
 */
                                  __pyx_t_31 = __pyx_v_N;
                                  __pyx_t_32 = __pyx_t_31;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
                                    __pyx_v_i = __pyx_t_33;

                                    /* "cyext_acv.pyx":1443
 *                     for i in range(N):
 * 
 *                         csi = 0             # <<<<<<<<<<<<<<
 *                         cs = 0
 * 
 */
                                    __pyx_v_csi = 0;

                                    /* "cyext_acv.pyx":1444
 * 
 *                         csi = 0
 *                         cs = 0             # <<<<<<<<<<<<<<
 * 
 *                         o_all = 0
 */
                                    __pyx_v_cs = 0;

                                    /* "cyext_acv.pyx":1446
 *                         cs = 0
 * 
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":1447
 * 
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_46 = __pyx_v_S_size;
                                    __pyx_t_47 = __pyx_t_46;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_47; __pyx_t_34+=1) {
                                      __pyx_v_s = __pyx_t_34;

                                      /* "cyext_acv.pyx":1448
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      __pyx_t_65 = __pyx_v_va;
                                      __pyx_t_64 = __pyx_v_counter;
                                      __pyx_t_63 = __pyx_v_leaf_numb;
                                      __pyx_t_62 = __pyx_v_s;
                                      __pyx_t_67 = __pyx_v_i;
                                      __pyx_t_69 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_65 * __pyx_v_S.strides[0]) ) + __pyx_t_64 * __pyx_v_S.strides[1]) ) + __pyx_t_63 * __pyx_v_S.strides[2]) )) + __pyx_t_62)) )));
                                      __pyx_t_66 = __pyx_v_va;
                                      __pyx_t_59 = __pyx_v_counter;
                                      __pyx_t_58 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = __pyx_v_s;
                                      __pyx_t_56 = __pyx_v_b;
                                      __pyx_t_60 = __pyx_v_leaf_numb;
                                      __pyx_t_68 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_66 * __pyx_v_S.strides[0]) ) + __pyx_t_59 * __pyx_v_S.strides[1]) ) + __pyx_t_58 * __pyx_v_S.strides[2]) )) + __pyx_t_57)) )));
                                      __pyx_t_61 = 1;
                                      __pyx_t_51 = __pyx_v_va;
                                      __pyx_t_50 = __pyx_v_counter;
                                      __pyx_t_49 = __pyx_v_leaf_numb;
                                      __pyx_t_48 = __pyx_v_s;
                                      __pyx_t_53 = __pyx_v_i;
                                      __pyx_t_55 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) ) + __pyx_t_50 * __pyx_v_S.strides[1]) ) + __pyx_t_49 * __pyx_v_S.strides[2]) )) + __pyx_t_48)) )));
                                      __pyx_t_52 = __pyx_v_va;
                                      __pyx_t_36 = __pyx_v_counter;
                                      __pyx_t_39 = __pyx_v_leaf_numb;
                                      __pyx_t_38 = __pyx_v_s;
                                      __pyx_t_37 = __pyx_v_b;
                                      __pyx_t_35 = __pyx_v_leaf_numb;
                                      __pyx_t_54 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_52 * __pyx_v_S.strides[0]) ) + __pyx_t_36 * __pyx_v_S.strides[1]) ) + __pyx_t_39 * __pyx_v_S.strides[2]) )) + __pyx_t_38)) )));
                                      __pyx_t_40 = 0;
                                      __pyx_t_15 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_67 * __pyx_v_X.strides[0]) ) + __pyx_t_69 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_60 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_61 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_53 * __pyx_v_X.strides[0]) ) + __pyx_t_55 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1449
 *                         for s in range(S_size):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                         if o_all == S_size:
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":1448
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1451
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == S_size:             # <<<<<<<<<<<<<<
 *                             cs = 1
 *                             nv_bool = 0
 */
                                    __pyx_t_15 = ((__pyx_v_o_all == __pyx_v_S_size) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1452
 * 
 *                         if o_all == S_size:
 *                             cs = 1             # <<<<<<<<<<<<<<
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 */
                                      __pyx_v_cs = 1;

                                      /* "cyext_acv.pyx":1453
 *                         if o_all == S_size:
 *                             cs = 1
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":1454
 *                             cs = 1
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_19; __pyx_t_34+=1) {
                                        __pyx_v_nv = __pyx_t_34;

                                        /* "cyext_acv.pyx":1455
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_38 = __pyx_v_i;
                                        __pyx_t_40 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_39 = __pyx_v_b;
                                        __pyx_t_36 = __pyx_v_leaf_numb;
                                        __pyx_t_54 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_55 = 1;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_38 * __pyx_v_X.strides[0]) ) + __pyx_t_40 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_36 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_41) {
                                        } else {
                                          __pyx_t_15 = __pyx_t_41;
                                          goto __pyx_L95_bool_binop_done;
                                        }
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_55 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_39 = __pyx_v_b;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_54 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_40 = 0;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_36 * __pyx_v_X.strides[0]) ) + __pyx_t_55 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_15 = __pyx_t_41;
                                        __pyx_L95_bool_binop_done:;
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1456
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":1457
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L92_continue;

                                          /* "cyext_acv.pyx":1455
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L92_continue:;
                                      }

                                      /* "cyext_acv.pyx":1459
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 csi = 1
 * 
 */
                                      __pyx_t_15 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1460
 * 
 *                             if nv_bool == 0:
 *                                 csi = 1             # <<<<<<<<<<<<<<
 * 
 *                         coef = 0
 */
                                        __pyx_v_csi = 1;

                                        /* "cyext_acv.pyx":1459
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 csi = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":1451
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == S_size:             # <<<<<<<<<<<<<<
 *                             cs = 1
 *                             nv_bool = 0
 */
                                    }

                                    /* "cyext_acv.pyx":1462
 *                                 csi = 1
 * 
 *                         coef = 0             # <<<<<<<<<<<<<<
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 */
                                    __pyx_v_coef = 0.0;

                                    /* "cyext_acv.pyx":1463
 * 
 *                         coef = 0
 *                         for l in range(1, m - set_size):             # <<<<<<<<<<<<<<
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 */
                                    __pyx_t_44 = (__pyx_v_m - __pyx_v_set_size);
                                    __pyx_t_45 = __pyx_t_44;
                                    for (__pyx_t_34 = 1; __pyx_t_34 < __pyx_t_45; __pyx_t_34+=1) {
                                      __pyx_v_l = __pyx_t_34;

                                      /* "cyext_acv.pyx":1464
 *                         coef = 0
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 */
                                      if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0) != 0.0) != 0)) {
                                        __pyx_t_70 = (__pyx_v_coef + ((1. * __pyx_f_9cyext_acv_binomialC(((__pyx_v_m - __pyx_v_set_size) - 1), __pyx_v_l, 0)) / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0)));
                                      } else {
                                        __pyx_t_70 = 0.0;
                                      }
                                      __pyx_v_coef = __pyx_t_70;
                                    }

                                    /* "cyext_acv.pyx":1466
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         if S_size == 0:
 */
                                    if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0) != 0.0) != 0)) {
                                      __pyx_t_70 = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));
                                    } else {
                                      __pyx_t_70 = 0.0;
                                    }
                                    __pyx_v_coef_0 = __pyx_t_70;

                                    /* "cyext_acv.pyx":1468
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 * 
 *                         if S_size == 0:             # <<<<<<<<<<<<<<
 *                             p_s = lm/data.shape[0]
 *                         else:
 */
                                    __pyx_t_15 = ((__pyx_v_S_size == 0) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1469
 * 
 *                         if S_size == 0:
 *                             p_s = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                         else:
 *                             p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0
 */
                                      __pyx_v_p_s = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                      /* "cyext_acv.pyx":1468
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 * 
 *                         if S_size == 0:             # <<<<<<<<<<<<<<
 *                             p_s = lm/data.shape[0]
 *                         else:
 */
                                      goto __pyx_L100;
                                    }

                                    /* "cyext_acv.pyx":1471
 *                             p_s = lm/data.shape[0]
 *                         else:
 *                             p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0
 * 
 */
                                    /*else*/ {
                                      __pyx_t_38 = __pyx_v_va;
                                      __pyx_t_39 = __pyx_v_leaf_numb;
                                      __pyx_t_36 = __pyx_v_counter;
                                      if ((((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_38 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_39 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_36 * __pyx_v_lm_s.strides[2]) ))) != 0.0) != 0)) {
                                        __pyx_t_52 = __pyx_v_va;
                                        __pyx_t_35 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_counter;
                                        __pyx_t_70 = ((__pyx_v_cs * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_52 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_35 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_37 * __pyx_v_lm_s.strides[2]) ))));
                                      } else {
                                        __pyx_t_70 = 0.0;
                                      }
                                      __pyx_v_p_s = __pyx_t_70;
                                    }
                                    __pyx_L100:;

                                    /* "cyext_acv.pyx":1472
 *                         else:
 *                             p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         for nv in range(va_id[va].size()):
 */
                                    __pyx_t_36 = __pyx_v_va;
                                    __pyx_t_39 = __pyx_v_leaf_numb;
                                    __pyx_t_38 = __pyx_v_counter;
                                    if ((((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_36 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_39 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[2]) ))) != 0.0) != 0)) {
                                      __pyx_t_37 = __pyx_v_va;
                                      __pyx_t_35 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = __pyx_v_counter;
                                      __pyx_t_70 = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_37 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_35 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_52 * __pyx_v_lm_si.strides[2]) ))));
                                    } else {
                                      __pyx_t_70 = 0.0;
                                    }
                                    __pyx_v_p_si = __pyx_t_70;

                                    /* "cyext_acv.pyx":1474
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0
 * 
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                    __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_19 = __pyx_t_18;
                                    for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_19; __pyx_t_34+=1) {
                                      __pyx_v_nv = __pyx_t_34;

                                      /* "cyext_acv.pyx":1475
 * 
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                      __pyx_t_71 = __pyx_v_d;
                                      __pyx_t_72 = __pyx_t_71;
                                      for (__pyx_t_73 = 0; __pyx_t_73 < __pyx_t_72; __pyx_t_73+=1) {
                                        __pyx_v_i2 = __pyx_t_73;

                                        /* "cyext_acv.pyx":1476
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
                                        __pyx_t_38 = __pyx_v_b;
                                        __pyx_t_39 = __pyx_v_leaf_numb;
                                        __pyx_t_36 = __pyx_v_b;
                                        __pyx_t_40 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_38 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_52 = __pyx_v_i2;
                                        __pyx_t_35 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_counter;
                                        __pyx_t_48 = __pyx_v_i;
                                        __pyx_t_54 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_49 = __pyx_v_i2;
                                        *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_35 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_37 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_48 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_54 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_49 * __pyx_v_phi_b.strides[4]) )) += (((__pyx_v_coef_0 + __pyx_v_coef) * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_36 * __pyx_v_values.strides[0]) ) + __pyx_t_40 * __pyx_v_values.strides[1]) ) + __pyx_t_52 * __pyx_v_values.strides[2]) ))));
                                      }
                                    }
                                  }
                                }
                                __pyx_L64_continue:;
                              }
                              goto __pyx_L106;
                              __pyx_L39_error:;
                              {
                                  #ifdef WITH_THREAD
                                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                  #endif
                                  #ifdef _OPENMP
                                  #pragma omp flush(__pyx_parallel_exc_type)
                                  #endif /* _OPENMP */
                                  if (!__pyx_parallel_exc_type) {
                                    __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                    __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                    __Pyx_GOTREF(__pyx_parallel_exc_type);
                                  }
                                  #ifdef WITH_THREAD
                                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                  #endif
                              }
                              __pyx_parallel_why = 4;
                              goto __pyx_L105;
                              __pyx_L105:;
                              #ifdef _OPENMP
                              #pragma omp critical(__pyx_parallel_lastprivates0)
                              #endif /* _OPENMP */
                              {
                                  __pyx_parallel_temp0 = __pyx_v_S_size;
                                  __pyx_parallel_temp1 = __pyx_v_a_it;
                                  __pyx_parallel_temp2 = __pyx_v_add;
                                  __pyx_parallel_temp3 = __pyx_v_b_it;
                                  __pyx_parallel_temp4 = __pyx_v_ci;
                                  __pyx_parallel_temp5 = __pyx_v_cj;
                                  __pyx_parallel_temp6 = __pyx_v_coef;
                                  __pyx_parallel_temp7 = __pyx_v_coef_0;
                                  __pyx_parallel_temp8 = __pyx_v_counter;
                                  __pyx_parallel_temp9 = __pyx_v_cs;
                                  __pyx_parallel_temp10 = __pyx_v_csi;
                                  __pyx_parallel_temp11 = __pyx_v_i;
                                  __pyx_parallel_temp12 = __pyx_v_i2;
                                  __pyx_parallel_temp13 = __pyx_v_l;
                                  __pyx_parallel_temp14 = __pyx_v_leaf_numb;
                                  __pyx_parallel_temp15 = __pyx_v_lm;
                                  __pyx_parallel_temp16 = __pyx_v_ns;
                                  __pyx_parallel_temp17 = __pyx_v_nv;
                                  __pyx_parallel_temp18 = __pyx_v_nv_bool;
                                  __pyx_parallel_temp19 = __pyx_v_o_all;
                                  __pyx_parallel_temp20 = __pyx_v_p_s;
                                  __pyx_parallel_temp21 = __pyx_v_p_si;
                                  __pyx_parallel_temp22 = __pyx_v_pow_set_size;
                                  __pyx_parallel_temp23 = __pyx_v_s;
                                  __pyx_parallel_temp24 = __pyx_v_set_size;
                                  __pyx_parallel_temp25 = __pyx_v_va;
                                  __pyx_parallel_temp26 = __pyx_v_va_size;
                              }
                              __pyx_L106:;
                              #ifdef _OPENMP
                              #pragma omp flush(__pyx_parallel_why)
                              #endif /* _OPENMP */
                          }
                      }
                      #ifdef _OPENMP
                      Py_END_ALLOW_THREADS
                      #else
{
#ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      #endif /* _OPENMP */
                      /* Clean up any temporaries */
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      #ifndef _OPENMP
}
#endif /* _OPENMP */
                  }
              }
              if (__pyx_parallel_exc_type) {
                /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
                __pyx_parallel_why = 4;
              }
              if (__pyx_parallel_why) {
                __pyx_v_S_size = __pyx_parallel_temp0;
                __pyx_v_a_it = __pyx_parallel_temp1;
                __pyx_v_add = __pyx_parallel_temp2;
                __pyx_v_b_it = __pyx_parallel_temp3;
                __pyx_v_ci = __pyx_parallel_temp4;
                __pyx_v_cj = __pyx_parallel_temp5;
                __pyx_v_coef = __pyx_parallel_temp6;
                __pyx_v_coef_0 = __pyx_parallel_temp7;
                __pyx_v_counter = __pyx_parallel_temp8;
                __pyx_v_cs = __pyx_parallel_temp9;
                __pyx_v_csi = __pyx_parallel_temp10;
                __pyx_v_i = __pyx_parallel_temp11;
                __pyx_v_i2 = __pyx_parallel_temp12;
                __pyx_v_l = __pyx_parallel_temp13;
                __pyx_v_leaf_numb = __pyx_parallel_temp14;
                __pyx_v_lm = __pyx_parallel_temp15;
                __pyx_v_ns = __pyx_parallel_temp16;
                __pyx_v_nv = __pyx_parallel_temp17;
                __pyx_v_nv_bool = __pyx_parallel_temp18;
                __pyx_v_o_all = __pyx_parallel_temp19;
                __pyx_v_p_s = __pyx_parallel_temp20;
                __pyx_v_p_si = __pyx_parallel_temp21;
                __pyx_v_pow_set_size = __pyx_parallel_temp22;
                __pyx_v_s = __pyx_parallel_temp23;
                __pyx_v_set_size = __pyx_parallel_temp24;
                __pyx_v_va = __pyx_parallel_temp25;
                __pyx_v_va_size = __pyx_parallel_temp26;
                switch (__pyx_parallel_why) {
                      case 4:
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      __Pyx_GIVEREF(__pyx_parallel_exc_type);
                      __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                      __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                  }
                  goto __pyx_L35_error;
                }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":1360
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L36;
          }
          __pyx_L35_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L36:;
        }
    }

    /* "cyext_acv.pyx":1478
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             for j in range(m):
 *                 for i2 in range(d):
 */
    __pyx_t_26 = __pyx_v_N;
    __pyx_t_25 = __pyx_t_26;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_25; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;

      /* "cyext_acv.pyx":1479
 * 
 *         for i in range(N):
 *             for j in range(m):             # <<<<<<<<<<<<<<
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 */
      __pyx_t_27 = __pyx_v_m;
      __pyx_t_28 = __pyx_t_27;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_28; __pyx_t_29+=1) {
        __pyx_v_j = __pyx_t_29;

        /* "cyext_acv.pyx":1480
 *         for i in range(N):
 *             for j in range(m):
 *                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):
 */
        __pyx_t_30 = __pyx_v_d;
        __pyx_t_31 = __pyx_t_30;
        for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
          __pyx_v_i2 = __pyx_t_32;

          /* "cyext_acv.pyx":1481
 *             for j in range(m):
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 */
          __pyx_t_21 = (__pyx_v_phi_b.shape[0]);
          __pyx_t_22 = __pyx_t_21;
          for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_22; __pyx_t_33+=1) {
            __pyx_v_leaf_numb = __pyx_t_33;

            /* "cyext_acv.pyx":1482
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 * 
 */
            __pyx_t_23 = (__pyx_v_phi_b.shape[1]);
            __pyx_t_24 = __pyx_t_23;
            for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_24; __pyx_t_34+=1) {
              __pyx_v_counter = __pyx_t_34;

              /* "cyext_acv.pyx":1483
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(phi)/m
 */
              __pyx_t_39 = __pyx_v_leaf_numb;
              __pyx_t_38 = __pyx_v_counter;
              __pyx_t_52 = __pyx_v_i;
              __pyx_t_36 = __pyx_v_j;
              __pyx_t_49 = __pyx_v_i2;
              __pyx_t_48 = __pyx_v_i;
              __pyx_t_37 = __pyx_v_j;
              __pyx_t_35 = __pyx_v_i2;
              *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_48 * __pyx_v_phi.strides[0]) ) + __pyx_t_37 * __pyx_v_phi.strides[1]) ) + __pyx_t_35 * __pyx_v_phi.strides[2]) )) += (*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_39 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_52 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_36 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_49 * __pyx_v_phi_b.strides[4]) )));
            }
          }
        }
      }
    }
  }

  /* "cyext_acv.pyx":1485
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 * 
 *     return np.asarray(phi)/m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_phi, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":1272
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_leaves_pa(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_pa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_si, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_21shap_values_leaves_pa(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_21shap_values_leaves_pa(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  long __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shap_values_leaves_pa (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 1); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 2); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 3); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 4); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 5); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 6); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 7); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 8); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, 9); __PYX_ERR(0, 1272, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "shap_values_leaves_pa") < 0)) __PYX_ERR(0, 1272, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 1272, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 1273, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 1275, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 1276, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 1277, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_long(values[6]); if (unlikely((__pyx_v_scaling == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1279, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[8]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1280, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shap_values_leaves_pa", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1272, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_pa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_20shap_values_leaves_pa(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_20shap_values_leaves_pa(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_leaves_pa", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 1272, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_shap_values_leaves_pa(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_pa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":1492
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_leaves_normalized(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_23shap_values_leaves_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_leaves_normalized(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long const __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  const __pyx_v_node_idx_trees, std::vector<std::vector<long> >  const __pyx_v_C, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_ns;
  unsigned int __pyx_v_add;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_i2;
  unsigned int __pyx_v_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_leaf_n;
  double __pyx_v_lm;
  std::vector<std::vector<long> >  __pyx_v_C_buff;
  std::vector<std::vector<long> >  __pyx_v_va_id;
  std::vector<std::vector<long> >  __pyx_v_buff;
  std::vector<long>  __pyx_v_coal_va;
  std::vector<long>  __pyx_v_remove_va;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id_v2;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_C_b;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_Sm;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id;
  std::vector<std::vector<std::vector<std::vector<long> > > >  __pyx_v_node_id_va;
  int __pyx_v_S_size;
  int __pyx_v_va_size;
  int __pyx_v_b_it;
  int __pyx_v_nv_bool;
  long __pyx_v_set_size;
  double __pyx_v_p_s;
  double __pyx_v_p_si;
  double __pyx_v_coef;
  double __pyx_v_coef_0;
  double __pyx_v_n_s;
  double __pyx_v_n_si;
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_si = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_csi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_cs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  std::vector<std::vector<long> > ::size_type __pyx_t_16;
  std::vector<std::vector<long> > ::size_type __pyx_t_17;
  std::vector<long> ::size_type __pyx_t_18;
  std::vector<long> ::size_type __pyx_t_19;
  unsigned int __pyx_t_20;
  __Pyx_memviewslice __pyx_t_21 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  unsigned int __pyx_t_26;
  unsigned int __pyx_t_27;
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  unsigned int __pyx_t_35;
  size_t __pyx_t_36;
  size_t __pyx_t_37;
  size_t __pyx_t_38;
  size_t __pyx_t_39;
  size_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  int __pyx_t_42;
  std::vector<long> ::size_type __pyx_t_43;
  std::vector<long> ::size_type __pyx_t_44;
  long __pyx_t_45;
  long __pyx_t_46;
  unsigned int __pyx_t_47;
  size_t __pyx_t_48;
  size_t __pyx_t_49;
  size_t __pyx_t_50;
  int __pyx_t_51;
  int __pyx_t_52;
  size_t __pyx_t_53;
  size_t __pyx_t_54;
  size_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  size_t __pyx_t_58;
  size_t __pyx_t_59;
  size_t __pyx_t_60;
  size_t __pyx_t_61;
  size_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  size_t __pyx_t_64;
  size_t __pyx_t_65;
  size_t __pyx_t_66;
  size_t __pyx_t_67;
  size_t __pyx_t_68;
  size_t __pyx_t_69;
  Py_ssize_t __pyx_t_70;
  Py_ssize_t __pyx_t_71;
  double __pyx_t_72;
  unsigned int __pyx_t_73;
  unsigned int __pyx_t_74;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_leaves_normalized", 0);

  /* "cyext_acv.pyx":1503
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":1504
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":1505
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":1506
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":1507
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] phi
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":1510
 * 
 *     cdef double[:, :, :] phi
 *     phi = np.zeros((N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1513
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 *     phi_b = np.zeros((max_leaves, 2**scaling, N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :, :, ::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_8);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_b = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1516
 * 
 *     cdef long[:, :, :, ::1] S
 *     S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_S = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1524
 *     cdef vector[vector[long]] C_buff, va_id, buff
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     for i in range(max_leaves):
 *         buff[i].resize(1)
 */
  try {
    __pyx_v_buff.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1524, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1525
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):             # <<<<<<<<<<<<<<
 *         buff[i].resize(1)
 * 
 */
  __pyx_t_12 = __pyx_v_max_leaves;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":1526
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):
 *         buff[i].resize(1)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 */
    try {
      (__pyx_v_buff[__pyx_v_i]).resize(1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1526, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1529
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 */
  try {
    __pyx_v_node_id_v2.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1529, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1530
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)
 */
  try {
    __pyx_v_C_b.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1530, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1531
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     node_id.resize(max_leaves)
 * 
 */
  try {
    __pyx_v_Sm.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1531, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1532
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 */
  try {
    __pyx_v_node_id.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1532, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1535
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)             # <<<<<<<<<<<<<<
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)
 */
  try {
    __pyx_v_node_id_va.resize(__pyx_v_m);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1535, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1536
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         node_id_va[i].resize(max_leaves)
 * 
 */
  __pyx_t_12 = __pyx_v_m;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":1537
 *     node_id_va.resize(m)
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef int S_size, va_size,  b_it, nv_bool
 */
    try {
      (__pyx_v_node_id_va[__pyx_v_i]).resize(__pyx_v_max_leaves);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1537, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1543
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
  __pyx_t_15 = (((__pyx_v_C[0]).size() != 0) != 0);
  if (__pyx_t_15) {

    /* "cyext_acv.pyx":1544
 * 
 *     if C[0].size() != 0:
 *         C_buff = C             # <<<<<<<<<<<<<<
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 */
    __pyx_v_C_buff = __pyx_v_C;

    /* "cyext_acv.pyx":1545
 *     if C[0].size() != 0:
 *         C_buff = C
 *         for ci in range(C.size()):             # <<<<<<<<<<<<<<
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])
 */
    __pyx_t_16 = __pyx_v_C.size();
    __pyx_t_17 = __pyx_t_16;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_17; __pyx_t_12+=1) {
      __pyx_v_ci = __pyx_t_12;

      /* "cyext_acv.pyx":1546
 *         C_buff = C
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):             # <<<<<<<<<<<<<<
 *                 coal_va.push_back(C[ci][cj])
 * 
 */
      __pyx_t_18 = (__pyx_v_C[__pyx_v_ci]).size();
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_19; __pyx_t_13+=1) {
        __pyx_v_cj = __pyx_t_13;

        /* "cyext_acv.pyx":1547
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(m):
 */
        try {
          __pyx_v_coal_va.push_back(((__pyx_v_C[__pyx_v_ci])[__pyx_v_cj]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1547, __pyx_L1_error)
        }
      }
    }

    /* "cyext_acv.pyx":1549
 *                 coal_va.push_back(C[ci][cj])
 * 
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 */
    __pyx_t_12 = __pyx_v_m;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "cyext_acv.pyx":1550
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
      __pyx_t_15 = ((!((std::find<std::vector<long> ::iterator,long>(__pyx_v_coal_va.begin(), __pyx_v_coal_va.end(), __pyx_v_i) != __pyx_v_coal_va.end()) != 0)) != 0);
      if (__pyx_t_15) {

        /* "cyext_acv.pyx":1551
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)             # <<<<<<<<<<<<<<
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])
 */
        try {
          __pyx_v_remove_va.push_back(__pyx_v_i);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1551, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1552
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i             # <<<<<<<<<<<<<<
 *                 va_id.push_back(buff[0])
 *             else:
 */
        ((__pyx_v_buff[0])[0]) = __pyx_v_i;

        /* "cyext_acv.pyx":1553
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 for ci in range(C_buff.size()):
 */
        try {
          __pyx_v_va_id.push_back((__pyx_v_buff[0]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1553, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1550
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
        goto __pyx_L14;
      }

      /* "cyext_acv.pyx":1555
 *                 va_id.push_back(buff[0])
 *             else:
 *                 for ci in range(C_buff.size()):             # <<<<<<<<<<<<<<
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 */
      /*else*/ {
        __pyx_t_16 = __pyx_v_C_buff.size();
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_17; __pyx_t_20+=1) {
          __pyx_v_ci = __pyx_t_20;

          /* "cyext_acv.pyx":1556
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          __pyx_t_15 = ((std::find<std::vector<long> ::iterator,long>((__pyx_v_C_buff[__pyx_v_ci]).begin(), (__pyx_v_C_buff[__pyx_v_ci]).end(), __pyx_v_i) != (__pyx_v_C_buff[__pyx_v_ci]).end()) != 0);
          if (__pyx_t_15) {

            /* "cyext_acv.pyx":1557
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])             # <<<<<<<<<<<<<<
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 */
            try {
              __pyx_v_va_id.push_back((__pyx_v_C_buff[__pyx_v_ci]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 1557, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":1558
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])             # <<<<<<<<<<<<<<
 *                         C_buff.pop_back()
 *                         break
 */
            (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(__pyx_v_C_buff.begin(), __pyx_v_C_buff.end(), (__pyx_v_C_buff[__pyx_v_ci])));

            /* "cyext_acv.pyx":1559
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_v_C_buff.pop_back();

            /* "cyext_acv.pyx":1560
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 *                         break             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L16_break;

            /* "cyext_acv.pyx":1556
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          }
        }
        __pyx_L16_break:;
      }
      __pyx_L14:;
    }

    /* "cyext_acv.pyx":1543
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
    goto __pyx_L7;
  }

  /* "cyext_acv.pyx":1563
 * 
 *     else:
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])
 */
  /*else*/ {
    __pyx_t_12 = __pyx_v_m;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "cyext_acv.pyx":1564
 *     else:
 *         for i in range(m):
 *             buff[0][0] = i             # <<<<<<<<<<<<<<
 *             va_id.push_back(buff[0])
 * 
 */
      ((__pyx_v_buff[0])[0]) = __pyx_v_i;

      /* "cyext_acv.pyx":1565
 *         for i in range(m):
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef double p_s, p_si, coef, coef_0, n_s, n_si
 */
      try {
        __pyx_v_va_id.push_back((__pyx_v_buff[0]));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 1565, __pyx_L1_error)
      }
    }
  }
  __pyx_L7:;

  /* "cyext_acv.pyx":1569
 *     cdef double p_s, p_si, coef, coef_0, n_s, n_si
 *     cdef double[:, :, :, :] lm_s, lm_si, lm_n, csi, cs
 *     lm_n = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_s = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_1);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_21 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_21.memview)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lm_n = __pyx_t_21;
  __pyx_t_21.memview = NULL;
  __pyx_t_21.data = NULL;

  /* "cyext_acv.pyx":1570
 *     cdef double[:, :, :, :] lm_s, lm_si, lm_n, csi, cs
 *     lm_n = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     lm_s = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_si = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     csi = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_21 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_21.memview)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lm_s = __pyx_t_21;
  __pyx_t_21.memview = NULL;
  __pyx_t_21.data = NULL;

  /* "cyext_acv.pyx":1571
 *     lm_n = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     lm_s = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     csi = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     cs = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_1);
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_21 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_21.memview)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_lm_si = __pyx_t_21;
  __pyx_t_21.memview = NULL;
  __pyx_t_21.data = NULL;

  /* "cyext_acv.pyx":1572
 *     lm_s = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     csi = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     cs = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_9 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_21 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_21.memview)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_csi = __pyx_t_21;
  __pyx_t_21.memview = NULL;
  __pyx_t_21.data = NULL;

  /* "cyext_acv.pyx":1573
 *     lm_si = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     csi = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 *     cs = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_21 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_21.memview)) __PYX_ERR(0, 1573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_cs = __pyx_t_21;
  __pyx_t_21.memview = NULL;
  __pyx_t_21.data = NULL;

  /* "cyext_acv.pyx":1575
 *     cs = np.zeros((va_id.size(), max_leaves, max_leaves, 2**scaling))
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 */
  __pyx_t_12 = __pyx_v_n_trees;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_b = __pyx_t_14;

    /* "cyext_acv.pyx":1576
 * 
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):
 */
    __pyx_t_22 = (__pyx_v_phi_b.shape[0]);
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_23; __pyx_t_20+=1) {
      __pyx_v_leaf_numb = __pyx_t_20;

      /* "cyext_acv.pyx":1577
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                 for i in range(N):
 *                     for j in range(m):
 */
      __pyx_t_24 = (__pyx_v_phi_b.shape[1]);
      __pyx_t_25 = __pyx_t_24;
      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
        __pyx_v_counter = __pyx_t_26;

        /* "cyext_acv.pyx":1578
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):             # <<<<<<<<<<<<<<
 *                     for j in range(m):
 *                         for i2 in range(d):
 */
        __pyx_t_27 = __pyx_v_N;
        __pyx_t_28 = __pyx_t_27;
        for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_28; __pyx_t_29+=1) {
          __pyx_v_i = __pyx_t_29;

          /* "cyext_acv.pyx":1579
 *             for counter in range(phi_b.shape[1]):
 *                 for i in range(N):
 *                     for j in range(m):             # <<<<<<<<<<<<<<
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 */
          __pyx_t_30 = __pyx_v_m;
          __pyx_t_31 = __pyx_t_30;
          for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
            __pyx_v_j = __pyx_t_32;

            /* "cyext_acv.pyx":1580
 *                 for i in range(N):
 *                     for j in range(m):
 *                         for i2 in range(d):             # <<<<<<<<<<<<<<
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 */
            __pyx_t_33 = __pyx_v_d;
            __pyx_t_34 = __pyx_t_33;
            for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_34; __pyx_t_35+=1) {
              __pyx_v_i2 = __pyx_t_35;

              /* "cyext_acv.pyx":1581
 *                     for j in range(m):
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
              __pyx_t_36 = __pyx_v_leaf_numb;
              __pyx_t_37 = __pyx_v_counter;
              __pyx_t_38 = __pyx_v_i;
              __pyx_t_39 = __pyx_v_j;
              __pyx_t_40 = __pyx_v_i2;
              *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_36 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_37 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_40 * __pyx_v_phi_b.strides[4]) )) = 0.0;
            }
          }
        }
      }
    }

    /* "cyext_acv.pyx":1582
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 */
    __pyx_t_40 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ ((char *) (((long const  *) __pyx_v_leaves_nb.data) + __pyx_t_40)) )));

    /* "cyext_acv.pyx":1583
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_20 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              int __pyx_parallel_temp0 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp1 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp2 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp3 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp4 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp5 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp6 = ((double)__PYX_NAN());
              double __pyx_parallel_temp7 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp8 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp9 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp10 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp11 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp12 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp13 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp14 = ((double)__PYX_NAN());
              double __pyx_parallel_temp15 = ((double)__PYX_NAN());
              double __pyx_parallel_temp16 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp17 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp18 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp19 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp20 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp21 = ((double)__PYX_NAN());
              double __pyx_parallel_temp22 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp23 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp24 = ((unsigned int)0xbad0bad0);
              long __pyx_parallel_temp25 = ((long)0xbad0bad0);
              unsigned int __pyx_parallel_temp26 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp27 = ((int)0xbad0bad0);
              const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
              PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
              int __pyx_parallel_why;
              __pyx_parallel_why = 0;
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_27 = (__pyx_t_20 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_27 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57, __pyx_t_58, __pyx_t_59, __pyx_t_60, __pyx_t_61, __pyx_t_62, __pyx_t_63, __pyx_t_64, __pyx_t_65, __pyx_t_66, __pyx_t_67, __pyx_t_68, __pyx_t_69, __pyx_t_70, __pyx_t_71, __pyx_t_72, __pyx_t_73, __pyx_t_74) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      Py_BEGIN_ALLOW_THREADS
                      #endif /* _OPENMP */
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_S_size) lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_add) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_ci) lastprivate(__pyx_v_cj) lastprivate(__pyx_v_coef) lastprivate(__pyx_v_coef_0) lastprivate(__pyx_v_counter) lastprivate(__pyx_v_i) lastprivate(__pyx_v_i2) lastprivate(__pyx_v_l) lastprivate(__pyx_v_leaf_n) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_n_s) lastprivate(__pyx_v_n_si) lastprivate(__pyx_v_ns) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_si) lastprivate(__pyx_v_pow_set_size) lastprivate(__pyx_v_s) lastprivate(__pyx_v_set_size) lastprivate(__pyx_v_va) lastprivate(__pyx_v_va_size) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_27; __pyx_t_26++){
                          if (__pyx_parallel_why < 2)
                          {
                              __pyx_v_leaf_numb = (unsigned int)(0 + 1 * __pyx_t_26);
                              /* Initialize private variables to invalid values */
                              __pyx_v_S_size = ((int)0xbad0bad0);
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_add = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((int)0xbad0bad0);
                              __pyx_v_ci = ((unsigned int)0xbad0bad0);
                              __pyx_v_cj = ((unsigned int)0xbad0bad0);
                              __pyx_v_coef = ((double)__PYX_NAN());
                              __pyx_v_coef_0 = ((double)__PYX_NAN());
                              __pyx_v_counter = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((unsigned int)0xbad0bad0);
                              __pyx_v_i2 = ((unsigned int)0xbad0bad0);
                              __pyx_v_l = ((unsigned int)0xbad0bad0);
                              __pyx_v_leaf_n = ((unsigned int)0xbad0bad0);
                              __pyx_v_lm = ((double)__PYX_NAN());
                              __pyx_v_n_s = ((double)__PYX_NAN());
                              __pyx_v_n_si = ((double)__PYX_NAN());
                              __pyx_v_ns = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv_bool = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_s = ((double)__PYX_NAN());
                              __pyx_v_p_si = ((double)__PYX_NAN());
                              __pyx_v_pow_set_size = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_set_size = ((long)0xbad0bad0);
                              __pyx_v_va = ((unsigned int)0xbad0bad0);
                              __pyx_v_va_size = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":1584
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *             lm = 0
 *             for i in range(data.shape[0]):
 */
                              (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).clear();

                              /* "cyext_acv.pyx":1585
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 */
                              __pyx_v_lm = 0.0;

                              /* "cyext_acv.pyx":1586
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 */
                              __pyx_t_22 = (__pyx_v_data.shape[0]);
                              __pyx_t_23 = __pyx_t_22;
                              for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_23; __pyx_t_28+=1) {
                                __pyx_v_i = __pyx_t_28;

                                /* "cyext_acv.pyx":1587
 *             lm = 0
 *             for i in range(data.shape[0]):
 *                 a_it = 0             # <<<<<<<<<<<<<<
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 */
                                __pyx_v_a_it = 0;

                                /* "cyext_acv.pyx":1588
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 *                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 */
                                __pyx_t_24 = (__pyx_v_data.shape[1]);
                                __pyx_t_25 = __pyx_t_24;
                                for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_25; __pyx_t_29+=1) {
                                  __pyx_v_s = __pyx_t_29;

                                  /* "cyext_acv.pyx":1589
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  __pyx_t_40 = __pyx_v_i;
                                  __pyx_t_39 = __pyx_v_s;
                                  __pyx_t_38 = __pyx_v_b;
                                  __pyx_t_37 = __pyx_v_leaf_numb;
                                  __pyx_t_36 = __pyx_v_s;
                                  __pyx_t_41 = 1;
                                  __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_40 * __pyx_v_data.strides[0]) ) + __pyx_t_39 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_36 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_42) {
                                  } else {
                                    __pyx_t_15 = __pyx_t_42;
                                    goto __pyx_L46_bool_binop_done;
                                  }
                                  __pyx_t_36 = __pyx_v_i;
                                  __pyx_t_37 = __pyx_v_s;
                                  __pyx_t_38 = __pyx_v_b;
                                  __pyx_t_39 = __pyx_v_leaf_numb;
                                  __pyx_t_40 = __pyx_v_s;
                                  __pyx_t_41 = 0;
                                  __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_37 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_15 = __pyx_t_42;
                                  __pyx_L46_bool_binop_done:;
                                  if (__pyx_t_15) {

                                    /* "cyext_acv.pyx":1590
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1
 */
                                    __pyx_v_a_it = (__pyx_v_a_it + 1);

                                    /* "cyext_acv.pyx":1589
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1591
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                __pyx_t_15 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                if (__pyx_t_15) {

                                  /* "cyext_acv.pyx":1592
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 *             if C[0].size() != 0:
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1.0);

                                  /* "cyext_acv.pyx":1591
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":1594
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                              __pyx_t_15 = (((__pyx_v_C[0]).size() != 0) != 0);
                              if (__pyx_t_15) {

                                /* "cyext_acv.pyx":1595
 * 
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C             # <<<<<<<<<<<<<<
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 */
                                (__pyx_v_C_b[__pyx_v_leaf_numb]) = __pyx_v_C;

                                /* "cyext_acv.pyx":1596
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 */
                                __pyx_t_18 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_19 = __pyx_t_18;
                                for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_19; __pyx_t_28+=1) {
                                  __pyx_v_nv = __pyx_t_28;

                                  /* "cyext_acv.pyx":1597
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0             # <<<<<<<<<<<<<<
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 */
                                  __pyx_v_add = 0;

                                  /* "cyext_acv.pyx":1598
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 *                     for ns in range(remove_va.size()):             # <<<<<<<<<<<<<<
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 */
                                  __pyx_t_43 = __pyx_v_remove_va.size();
                                  __pyx_t_44 = __pyx_t_43;
                                  for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_44; __pyx_t_29+=1) {
                                    __pyx_v_ns = __pyx_t_29;

                                    /* "cyext_acv.pyx":1599
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    __pyx_t_15 = (((((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]) == (__pyx_v_remove_va[__pyx_v_ns])) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1600
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1             # <<<<<<<<<<<<<<
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 */
                                      __pyx_v_add = 1;

                                      /* "cyext_acv.pyx":1601
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]             # <<<<<<<<<<<<<<
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 */
                                      ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]);

                                      /* "cyext_acv.pyx":1602
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 *                             break
 *                     if add == 0:
 */
                                      try {
                                        (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                      } catch(...) {
                                        #ifdef WITH_THREAD
                                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                        #endif
                                        __Pyx_CppExn2PyErr();
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        __PYX_ERR(0, 1602, __pyx_L39_error)
                                      }

                                      /* "cyext_acv.pyx":1603
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break             # <<<<<<<<<<<<<<
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 */
                                      goto __pyx_L53_break;

                                      /* "cyext_acv.pyx":1599
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    }
                                  }
                                  __pyx_L53_break:;

                                  /* "cyext_acv.pyx":1604
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  __pyx_t_15 = ((__pyx_v_add == 0) != 0);
                                  if (__pyx_t_15) {

                                    /* "cyext_acv.pyx":1605
 *                             break
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 */
                                    __pyx_t_16 = (__pyx_v_C_b[__pyx_v_leaf_numb]).size();
                                    __pyx_t_17 = __pyx_t_16;
                                    for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_17; __pyx_t_29+=1) {
                                      __pyx_v_ci = __pyx_t_29;

                                      /* "cyext_acv.pyx":1606
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 */
                                      __pyx_t_43 = ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_44 = __pyx_t_43;
                                      for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_44; __pyx_t_30+=1) {
                                        __pyx_v_cj = __pyx_t_30;

                                        /* "cyext_acv.pyx":1607
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        __pyx_t_15 = (((((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]) == (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv])) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1608
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1             # <<<<<<<<<<<<<<
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 */
                                          __pyx_v_add = 1;

                                          /* "cyext_acv.pyx":1609
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                     break
 *                             if add == 1:
 */
                                          try {
                                            (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back(((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]));
                                          } catch(...) {
                                            #ifdef WITH_THREAD
                                            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                            #endif
                                            __Pyx_CppExn2PyErr();
                                            #ifdef WITH_THREAD
                                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                            #endif
                                            __PYX_ERR(0, 1609, __pyx_L39_error)
                                          }

                                          /* "cyext_acv.pyx":1610
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break             # <<<<<<<<<<<<<<
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 */
                                          goto __pyx_L59_break;

                                          /* "cyext_acv.pyx":1607
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        }
                                      }
                                      __pyx_L59_break:;

                                      /* "cyext_acv.pyx":1611
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      __pyx_t_15 = ((__pyx_v_add == 1) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1612
 *                                     break
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 */
                                        (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >((__pyx_v_C_b[__pyx_v_leaf_numb]).begin(), (__pyx_v_C_b[__pyx_v_leaf_numb]).end(), ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])));

                                        /* "cyext_acv.pyx":1613
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 */
                                        (__pyx_v_C_b[__pyx_v_leaf_numb]).pop_back();

                                        /* "cyext_acv.pyx":1614
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 *                                 break             # <<<<<<<<<<<<<<
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 */
                                        goto __pyx_L57_break;

                                        /* "cyext_acv.pyx":1611
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      }
                                    }
                                    __pyx_L57_break:;

                                    /* "cyext_acv.pyx":1604
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1615
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]             # <<<<<<<<<<<<<<
 *             else:
 *                 node_id[leaf_numb].clear()
 */
                                (__pyx_v_node_id[__pyx_v_leaf_numb]) = (__pyx_v_node_id_v2[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1594
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                                goto __pyx_L49;
                              }

                              /* "cyext_acv.pyx":1617
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 *                 node_id[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 */
                              /*else*/ {
                                (__pyx_v_node_id[__pyx_v_leaf_numb]).clear();

                                /* "cyext_acv.pyx":1618
 *             else:
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 */
                                __pyx_t_18 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_19 = __pyx_t_18;
                                for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_19; __pyx_t_28+=1) {
                                  __pyx_v_i = __pyx_t_28;

                                  /* "cyext_acv.pyx":1619
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]             # <<<<<<<<<<<<<<
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 */
                                  ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_i]);

                                  /* "cyext_acv.pyx":1620
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  try {
                                    (__pyx_v_node_id[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                  } catch(...) {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    __Pyx_CppExn2PyErr();
                                    #ifdef WITH_THREAD
                                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                    #endif
                                    __PYX_ERR(0, 1620, __pyx_L39_error)
                                  }
                                }
                              }
                              __pyx_L49:;

                              /* "cyext_acv.pyx":1623
 * 
 * 
 *             for va in range(va_id.size()):             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 */
                              __pyx_t_16 = __pyx_v_va_id.size();
                              __pyx_t_17 = __pyx_t_16;
                              for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_17; __pyx_t_28+=1) {
                                __pyx_v_va = __pyx_t_28;

                                /* "cyext_acv.pyx":1624
 * 
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]             # <<<<<<<<<<<<<<
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]) = (__pyx_v_node_id[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1625
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_15 = ((!((std::find<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])) != ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end()) != 0)) != 0);
                                if (__pyx_t_15) {

                                  /* "cyext_acv.pyx":1626
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 */
                                  goto __pyx_L64_continue;

                                  /* "cyext_acv.pyx":1625
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":1628
 *                     continue
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 */
                                (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])));

                                /* "cyext_acv.pyx":1629
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 *                 node_id_va[va][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).pop_back();

                                /* "cyext_acv.pyx":1631
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()             # <<<<<<<<<<<<<<
 *                 pow_set_size = 2**set_size
 * 
 */
                                __pyx_v_set_size = ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).size();

                                /* "cyext_acv.pyx":1632
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 *                 pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *                 for counter in range(0, pow_set_size):
 */
                                __pyx_v_pow_set_size = __Pyx_pow_long(2, __pyx_v_set_size);

                                /* "cyext_acv.pyx":1634
 *                 pow_set_size = 2**set_size
 * 
 *                 for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *                     va_size = 0
 *                     S_size = 0
 */
                                __pyx_t_29 = __pyx_v_pow_set_size;
                                __pyx_t_30 = __pyx_t_29;
                                for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
                                  __pyx_v_counter = __pyx_t_31;

                                  /* "cyext_acv.pyx":1635
 * 
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0             # <<<<<<<<<<<<<<
 *                     S_size = 0
 *                     for ci in range(set_size):
 */
                                  __pyx_v_va_size = 0;

                                  /* "cyext_acv.pyx":1636
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0
 *                     S_size = 0             # <<<<<<<<<<<<<<
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 */
                                  __pyx_v_S_size = 0;

                                  /* "cyext_acv.pyx":1637
 *                     va_size = 0
 *                     S_size = 0
 *                     for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 */
                                  __pyx_t_45 = __pyx_v_set_size;
                                  __pyx_t_46 = __pyx_t_45;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_46; __pyx_t_32+=1) {
                                    __pyx_v_ci = __pyx_t_32;

                                    /* "cyext_acv.pyx":1638
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    __pyx_t_15 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1639
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 */
                                      __pyx_t_18 = (((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_19; __pyx_t_33+=1) {
                                        __pyx_v_cj = __pyx_t_33;

                                        /* "cyext_acv.pyx":1640
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]             # <<<<<<<<<<<<<<
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1
 */
                                        __pyx_t_40 = __pyx_v_va;
                                        __pyx_t_39 = __pyx_v_counter;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_41 = __pyx_v_S_size;
                                        *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_40 * __pyx_v_S.strides[0]) ) + __pyx_t_39 * __pyx_v_S.strides[1]) ) + __pyx_t_38 * __pyx_v_S.strides[2]) )) + __pyx_t_41)) )) = ((((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]);

                                        /* "cyext_acv.pyx":1641
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                             va_size = va_size + 1
 * 
 */
                                        __pyx_v_S_size = (__pyx_v_S_size + 1);
                                      }

                                      /* "cyext_acv.pyx":1642
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *                     for leaf_n in range(nb_leaf):
 */
                                      __pyx_v_va_size = (__pyx_v_va_size + 1);

                                      /* "cyext_acv.pyx":1638
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1644
 *                             va_size = va_size + 1
 * 
 *                     for leaf_n in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         lm_s[va, leaf_numb, leaf_n, counter] = 0
 *                         lm_si[va, leaf_numb, leaf_n,  counter] = 0
 */
                                  __pyx_t_32 = __pyx_v_nb_leaf;
                                  __pyx_t_33 = __pyx_t_32;
                                  for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                    __pyx_v_leaf_n = __pyx_t_34;

                                    /* "cyext_acv.pyx":1645
 * 
 *                     for leaf_n in range(nb_leaf):
 *                         lm_s[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 *                         lm_si[va, leaf_numb, leaf_n,  counter] = 0
 *                         lm_n[va, leaf_numb, leaf_n, counter] = 0
 */
                                    __pyx_t_38 = __pyx_v_va;
                                    __pyx_t_39 = __pyx_v_leaf_numb;
                                    __pyx_t_40 = __pyx_v_leaf_n;
                                    __pyx_t_37 = __pyx_v_counter;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_38 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_39 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_40 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_37 * __pyx_v_lm_s.strides[3]) )) = 0.0;

                                    /* "cyext_acv.pyx":1646
 *                     for leaf_n in range(nb_leaf):
 *                         lm_s[va, leaf_numb, leaf_n, counter] = 0
 *                         lm_si[va, leaf_numb, leaf_n,  counter] = 0             # <<<<<<<<<<<<<<
 *                         lm_n[va, leaf_numb, leaf_n, counter] = 0
 *                         for i in range(data.shape[0]):
 */
                                    __pyx_t_37 = __pyx_v_va;
                                    __pyx_t_40 = __pyx_v_leaf_numb;
                                    __pyx_t_39 = __pyx_v_leaf_n;
                                    __pyx_t_38 = __pyx_v_counter;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_37 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_40 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_39 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[3]) )) = 0.0;

                                    /* "cyext_acv.pyx":1647
 *                         lm_s[va, leaf_numb, leaf_n, counter] = 0
 *                         lm_si[va, leaf_numb, leaf_n,  counter] = 0
 *                         lm_n[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 *                         for i in range(data.shape[0]):
 *                             a_it = 0
 */
                                    __pyx_t_38 = __pyx_v_va;
                                    __pyx_t_39 = __pyx_v_leaf_numb;
                                    __pyx_t_40 = __pyx_v_leaf_n;
                                    __pyx_t_37 = __pyx_v_counter;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_38 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_39 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_40 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_37 * __pyx_v_lm_n.strides[3]) )) = 0.0;

                                    /* "cyext_acv.pyx":1648
 *                         lm_si[va, leaf_numb, leaf_n,  counter] = 0
 *                         lm_n[va, leaf_numb, leaf_n, counter] = 0
 *                         for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             for s in range(data.shape[1]):
 */
                                    __pyx_t_22 = (__pyx_v_data.shape[0]);
                                    __pyx_t_23 = __pyx_t_22;
                                    for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_23; __pyx_t_35+=1) {
                                      __pyx_v_i = __pyx_t_35;

                                      /* "cyext_acv.pyx":1649
 *                         lm_n[va, leaf_numb, leaf_n, counter] = 0
 *                         for i in range(data.shape[0]):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             for s in range(data.shape[1]):
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":1650
 *                         for i in range(data.shape[0]):
 *                             a_it = 0
 *                             for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_24 = (__pyx_v_data.shape[1]);
                                      __pyx_t_25 = __pyx_t_24;
                                      for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_25; __pyx_t_47+=1) {
                                        __pyx_v_s = __pyx_t_47;

                                        /* "cyext_acv.pyx":1651
 *                             a_it = 0
 *                             for s in range(data.shape[1]):
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             if a_it == data.shape[1]:
 */
                                        __pyx_t_37 = __pyx_v_i;
                                        __pyx_t_40 = __pyx_v_s;
                                        __pyx_t_39 = __pyx_v_b;
                                        __pyx_t_38 = __pyx_v_leaf_n;
                                        __pyx_t_36 = __pyx_v_s;
                                        __pyx_t_41 = 1;
                                        __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_37 * __pyx_v_data.strides[0]) ) + __pyx_t_40 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_36 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_42) {
                                        } else {
                                          __pyx_t_15 = __pyx_t_42;
                                          goto __pyx_L81_bool_binop_done;
                                        }
                                        __pyx_t_36 = __pyx_v_i;
                                        __pyx_t_38 = __pyx_v_s;
                                        __pyx_t_39 = __pyx_v_b;
                                        __pyx_t_40 = __pyx_v_leaf_n;
                                        __pyx_t_37 = __pyx_v_s;
                                        __pyx_t_41 = 0;
                                        __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_38 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_15 = __pyx_t_42;
                                        __pyx_L81_bool_binop_done:;
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1652
 *                             for s in range(data.shape[1]):
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                             if a_it == data.shape[1]:
 *                                 lm_n[va, leaf_numb, leaf_n, counter] = lm_n[va, leaf_numb, leaf_n, counter] + 1
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":1651
 *                             a_it = 0
 *                             for s in range(data.shape[1]):
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             if a_it == data.shape[1]:
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":1653
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
 *                                     a_it = a_it + 1
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm_n[va, leaf_numb, leaf_n, counter] = lm_n[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                      __pyx_t_15 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1654
 *                                     a_it = a_it + 1
 *                             if a_it == data.shape[1]:
 *                                 lm_n[va, leaf_numb, leaf_n, counter] = lm_n[va, leaf_numb, leaf_n, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                             b_it = 0
 */
                                        __pyx_t_37 = __pyx_v_va;
                                        __pyx_t_40 = __pyx_v_leaf_numb;
                                        __pyx_t_39 = __pyx_v_leaf_n;
                                        __pyx_t_38 = __pyx_v_counter;
                                        __pyx_t_36 = __pyx_v_va;
                                        __pyx_t_48 = __pyx_v_leaf_numb;
                                        __pyx_t_49 = __pyx_v_leaf_n;
                                        __pyx_t_50 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_36 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_48 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_49 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_50 * __pyx_v_lm_n.strides[3]) )) = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_37 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_40 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_39 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_38 * __pyx_v_lm_n.strides[3]) ))) + 1.0);

                                        /* "cyext_acv.pyx":1653
 *                                 if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
 *                                     a_it = a_it + 1
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm_n[va, leaf_numb, leaf_n, counter] = lm_n[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":1656
 *                                 lm_n[va, leaf_numb, leaf_n, counter] = lm_n[va, leaf_numb, leaf_n, counter] + 1
 * 
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":1657
 * 
 *                             b_it = 0
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                     b_it = b_it + 1
 */
                                      __pyx_t_51 = __pyx_v_S_size;
                                      __pyx_t_52 = __pyx_t_51;
                                      for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_52; __pyx_t_47+=1) {
                                        __pyx_v_s = __pyx_t_47;

                                        /* "cyext_acv.pyx":1658
 *                             b_it = 0
 *                             for s in range(S_size):
 *                                 if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        __pyx_t_38 = __pyx_v_va;
                                        __pyx_t_39 = __pyx_v_counter;
                                        __pyx_t_40 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_s;
                                        __pyx_t_50 = __pyx_v_i;
                                        __pyx_t_41 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_38 * __pyx_v_S.strides[0]) ) + __pyx_t_39 * __pyx_v_S.strides[1]) ) + __pyx_t_40 * __pyx_v_S.strides[2]) )) + __pyx_t_37)) )));
                                        __pyx_t_49 = __pyx_v_va;
                                        __pyx_t_48 = __pyx_v_counter;
                                        __pyx_t_36 = __pyx_v_leaf_numb;
                                        __pyx_t_53 = __pyx_v_s;
                                        __pyx_t_54 = __pyx_v_b;
                                        __pyx_t_55 = __pyx_v_leaf_n;
                                        __pyx_t_56 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_49 * __pyx_v_S.strides[0]) ) + __pyx_t_48 * __pyx_v_S.strides[1]) ) + __pyx_t_36 * __pyx_v_S.strides[2]) )) + __pyx_t_53)) )));
                                        __pyx_t_57 = 1;
                                        __pyx_t_58 = __pyx_v_va;
                                        __pyx_t_59 = __pyx_v_counter;
                                        __pyx_t_60 = __pyx_v_leaf_numb;
                                        __pyx_t_61 = __pyx_v_s;
                                        __pyx_t_62 = __pyx_v_i;
                                        __pyx_t_63 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_58 * __pyx_v_S.strides[0]) ) + __pyx_t_59 * __pyx_v_S.strides[1]) ) + __pyx_t_60 * __pyx_v_S.strides[2]) )) + __pyx_t_61)) )));
                                        __pyx_t_64 = __pyx_v_va;
                                        __pyx_t_65 = __pyx_v_counter;
                                        __pyx_t_66 = __pyx_v_leaf_numb;
                                        __pyx_t_67 = __pyx_v_s;
                                        __pyx_t_68 = __pyx_v_b;
                                        __pyx_t_69 = __pyx_v_leaf_n;
                                        __pyx_t_70 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_64 * __pyx_v_S.strides[0]) ) + __pyx_t_65 * __pyx_v_S.strides[1]) ) + __pyx_t_66 * __pyx_v_S.strides[2]) )) + __pyx_t_67)) )));
                                        __pyx_t_71 = 0;
                                        __pyx_t_15 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_50 * __pyx_v_data.strides[0]) ) + __pyx_t_41 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_62 * __pyx_v_data.strides[0]) ) + __pyx_t_63 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1659
 *                             for s in range(S_size):
 *                                 if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                     b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if b_it == S_size:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":1658
 *                             b_it = 0
 *                             for s in range(S_size):
 *                                 if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":1661
 *                                     b_it = b_it + 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s[va, leaf_numb, leaf_n, counter] = lm_s[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                      __pyx_t_15 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_15) {

                                        /* "cyext_acv.pyx":1662
 * 
 *                             if b_it == S_size:
 *                                 lm_s[va, leaf_numb, leaf_n, counter] = lm_s[va, leaf_numb, leaf_n, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                                 nv_bool = 0
 */
                                        __pyx_t_67 = __pyx_v_va;
                                        __pyx_t_66 = __pyx_v_leaf_numb;
                                        __pyx_t_65 = __pyx_v_leaf_n;
                                        __pyx_t_64 = __pyx_v_counter;
                                        __pyx_t_69 = __pyx_v_va;
                                        __pyx_t_68 = __pyx_v_leaf_numb;
                                        __pyx_t_61 = __pyx_v_leaf_n;
                                        __pyx_t_60 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_69 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_68 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_61 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_60 * __pyx_v_lm_s.strides[3]) )) = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_67 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_66 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_65 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_64 * __pyx_v_lm_s.strides[3]) ))) + 1.0);

                                        /* "cyext_acv.pyx":1664
 *                                 lm_s[va, leaf_numb, leaf_n, counter] = lm_s[va, leaf_numb, leaf_n, counter] + 1
 * 
 *                                 nv_bool = 0             # <<<<<<<<<<<<<<
 *                                 for nv in range(va_id[va].size()):
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 */
                                        __pyx_v_nv_bool = 0;

                                        /* "cyext_acv.pyx":1665
 * 
 *                                 nv_bool = 0
 *                                 for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                         nv_bool = nv_bool + 1
 */
                                        __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                        __pyx_t_19 = __pyx_t_18;
                                        for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_19; __pyx_t_47+=1) {
                                          __pyx_v_nv = __pyx_t_47;

                                          /* "cyext_acv.pyx":1666
 *                                 nv_bool = 0
 *                                 for nv in range(va_id[va].size()):
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                         nv_bool = nv_bool + 1
 *                                         continue
 */
                                          __pyx_t_64 = __pyx_v_i;
                                          __pyx_t_71 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_65 = __pyx_v_b;
                                          __pyx_t_66 = __pyx_v_leaf_n;
                                          __pyx_t_70 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_63 = 1;
                                          __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_64 * __pyx_v_data.strides[0]) ) + __pyx_t_71 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_65 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                          if (!__pyx_t_42) {
                                          } else {
                                            __pyx_t_15 = __pyx_t_42;
                                            goto __pyx_L91_bool_binop_done;
                                          }
                                          __pyx_t_66 = __pyx_v_i;
                                          __pyx_t_63 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_65 = __pyx_v_b;
                                          __pyx_t_64 = __pyx_v_leaf_n;
                                          __pyx_t_70 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_71 = 0;
                                          __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_66 * __pyx_v_data.strides[0]) ) + __pyx_t_63 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_65 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                          __pyx_t_15 = __pyx_t_42;
                                          __pyx_L91_bool_binop_done:;
                                          if (__pyx_t_15) {

                                            /* "cyext_acv.pyx":1667
 *                                 for nv in range(va_id[va].size()):
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                         nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                         continue
 * 
 */
                                            __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                            /* "cyext_acv.pyx":1668
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                         nv_bool = nv_bool + 1
 *                                         continue             # <<<<<<<<<<<<<<
 * 
 *                                 if nv_bool == 0:
 */
                                            goto __pyx_L88_continue;

                                            /* "cyext_acv.pyx":1666
 *                                 nv_bool = 0
 *                                 for nv in range(va_id[va].size()):
 *                                     if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                         nv_bool = nv_bool + 1
 *                                         continue
 */
                                          }
                                          __pyx_L88_continue:;
                                        }

                                        /* "cyext_acv.pyx":1670
 *                                         continue
 * 
 *                                 if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                     lm_si[va, leaf_numb, leaf_n, counter] = lm_si[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                        __pyx_t_15 = ((__pyx_v_nv_bool == 0) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1671
 * 
 *                                 if nv_bool == 0:
 *                                     lm_si[va, leaf_numb, leaf_n, counter] = lm_si[va, leaf_numb, leaf_n, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(N):
 */
                                          __pyx_t_64 = __pyx_v_va;
                                          __pyx_t_65 = __pyx_v_leaf_numb;
                                          __pyx_t_66 = __pyx_v_leaf_n;
                                          __pyx_t_67 = __pyx_v_counter;
                                          __pyx_t_60 = __pyx_v_va;
                                          __pyx_t_61 = __pyx_v_leaf_numb;
                                          __pyx_t_68 = __pyx_v_leaf_n;
                                          __pyx_t_69 = __pyx_v_counter;
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_60 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_61 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_68 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_69 * __pyx_v_lm_si.strides[3]) )) = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_64 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_65 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_66 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_67 * __pyx_v_lm_si.strides[3]) ))) + 1.0);

                                          /* "cyext_acv.pyx":1670
 *                                         continue
 * 
 *                                 if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                     lm_si[va, leaf_numb, leaf_n, counter] = lm_si[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                        }

                                        /* "cyext_acv.pyx":1661
 *                                     b_it = b_it + 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s[va, leaf_numb, leaf_n, counter] = lm_s[va, leaf_numb, leaf_n, counter] + 1
 * 
 */
                                      }
                                    }
                                  }

                                  /* "cyext_acv.pyx":1673
 *                                     lm_si[va, leaf_numb, leaf_n, counter] = lm_si[va, leaf_numb, leaf_n, counter] + 1
 * 
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 * 
 *                         coef = 0
 */
                                  __pyx_t_32 = __pyx_v_N;
                                  __pyx_t_33 = __pyx_t_32;
                                  for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                    __pyx_v_i = __pyx_t_34;

                                    /* "cyext_acv.pyx":1675
 *                     for i in range(N):
 * 
 *                         coef = 0             # <<<<<<<<<<<<<<
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 */
                                    __pyx_v_coef = 0.0;

                                    /* "cyext_acv.pyx":1676
 * 
 *                         coef = 0
 *                         for l in range(1, m - set_size):             # <<<<<<<<<<<<<<
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 */
                                    __pyx_t_45 = (__pyx_v_m - __pyx_v_set_size);
                                    __pyx_t_46 = __pyx_t_45;
                                    for (__pyx_t_35 = 1; __pyx_t_35 < __pyx_t_46; __pyx_t_35+=1) {
                                      __pyx_v_l = __pyx_t_35;

                                      /* "cyext_acv.pyx":1677
 *                         coef = 0
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 */
                                      if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0) != 0.0) != 0)) {
                                        __pyx_t_72 = (__pyx_v_coef + ((1. * __pyx_f_9cyext_acv_binomialC(((__pyx_v_m - __pyx_v_set_size) - 1), __pyx_v_l, 0)) / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0)));
                                      } else {
                                        __pyx_t_72 = 0.0;
                                      }
                                      __pyx_v_coef = __pyx_t_72;
                                    }

                                    /* "cyext_acv.pyx":1679
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         n_s = 0
 */
                                    if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0) != 0.0) != 0)) {
                                      __pyx_t_72 = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));
                                    } else {
                                      __pyx_t_72 = 0.0;
                                    }
                                    __pyx_v_coef_0 = __pyx_t_72;

                                    /* "cyext_acv.pyx":1681
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 * 
 *                         n_s = 0             # <<<<<<<<<<<<<<
 *                         n_si = 0
 *                         if S_size == 0:
 */
                                    __pyx_v_n_s = 0.0;

                                    /* "cyext_acv.pyx":1682
 * 
 *                         n_s = 0
 *                         n_si = 0             # <<<<<<<<<<<<<<
 *                         if S_size == 0:
 *                             p_s = lm/data.shape[0]
 */
                                    __pyx_v_n_si = 0.0;

                                    /* "cyext_acv.pyx":1683
 *                         n_s = 0
 *                         n_si = 0
 *                         if S_size == 0:             # <<<<<<<<<<<<<<
 *                             p_s = lm/data.shape[0]
 *                             n_s = 0
 */
                                    __pyx_t_15 = ((__pyx_v_S_size == 0) != 0);
                                    if (__pyx_t_15) {

                                      /* "cyext_acv.pyx":1684
 *                         n_si = 0
 *                         if S_size == 0:
 *                             p_s = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                             n_s = 0
 *                             n_si = 0
 */
                                      __pyx_v_p_s = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                      /* "cyext_acv.pyx":1685
 *                         if S_size == 0:
 *                             p_s = lm/data.shape[0]
 *                             n_s = 0             # <<<<<<<<<<<<<<
 *                             n_si = 0
 *                             for leaf_n in range(nb_leaf):
 */
                                      __pyx_v_n_s = 0.0;

                                      /* "cyext_acv.pyx":1686
 *                             p_s = lm/data.shape[0]
 *                             n_s = 0
 *                             n_si = 0             # <<<<<<<<<<<<<<
 *                             for leaf_n in range(nb_leaf):
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0
 */
                                      __pyx_v_n_si = 0.0;

                                      /* "cyext_acv.pyx":1687
 *                             n_s = 0
 *                             n_si = 0
 *                             for leaf_n in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 */
                                      __pyx_t_35 = __pyx_v_nb_leaf;
                                      __pyx_t_47 = __pyx_t_35;
                                      for (__pyx_t_73 = 0; __pyx_t_73 < __pyx_t_47; __pyx_t_73+=1) {
                                        __pyx_v_leaf_n = __pyx_t_73;

                                        /* "cyext_acv.pyx":1688
 *                             n_si = 0
 *                             for leaf_n in range(nb_leaf):
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 * 
 */
                                        __pyx_t_67 = __pyx_v_va;
                                        __pyx_t_66 = __pyx_v_leaf_numb;
                                        __pyx_t_65 = __pyx_v_leaf_n;
                                        __pyx_t_64 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_67 * __pyx_v_csi.strides[0]) ) + __pyx_t_66 * __pyx_v_csi.strides[1]) ) + __pyx_t_65 * __pyx_v_csi.strides[2]) ) + __pyx_t_64 * __pyx_v_csi.strides[3]) )) = 0.0;

                                        /* "cyext_acv.pyx":1689
 *                             for leaf_n in range(nb_leaf):
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 * 
 *                                 o_all = 0
 */
                                        __pyx_t_64 = __pyx_v_va;
                                        __pyx_t_65 = __pyx_v_leaf_numb;
                                        __pyx_t_66 = __pyx_v_leaf_n;
                                        __pyx_t_67 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_64 * __pyx_v_cs.strides[0]) ) + __pyx_t_65 * __pyx_v_cs.strides[1]) ) + __pyx_t_66 * __pyx_v_cs.strides[2]) ) + __pyx_t_67 * __pyx_v_cs.strides[3]) )) = 0.0;

                                        /* "cyext_acv.pyx":1691
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 * 
 *                                 o_all = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 */
                                        __pyx_v_o_all = 0;

                                        /* "cyext_acv.pyx":1692
 * 
 *                                 o_all = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                         o_all = o_all + 1
 */
                                        __pyx_t_51 = __pyx_v_S_size;
                                        __pyx_t_52 = __pyx_t_51;
                                        for (__pyx_t_74 = 0; __pyx_t_74 < __pyx_t_52; __pyx_t_74+=1) {
                                          __pyx_v_s = __pyx_t_74;

                                          /* "cyext_acv.pyx":1693
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 * 
 */
                                          __pyx_t_67 = __pyx_v_va;
                                          __pyx_t_66 = __pyx_v_counter;
                                          __pyx_t_65 = __pyx_v_leaf_numb;
                                          __pyx_t_64 = __pyx_v_s;
                                          __pyx_t_69 = __pyx_v_i;
                                          __pyx_t_71 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_67 * __pyx_v_S.strides[0]) ) + __pyx_t_66 * __pyx_v_S.strides[1]) ) + __pyx_t_65 * __pyx_v_S.strides[2]) )) + __pyx_t_64)) )));
                                          __pyx_t_68 = __pyx_v_va;
                                          __pyx_t_61 = __pyx_v_counter;
                                          __pyx_t_60 = __pyx_v_leaf_numb;
                                          __pyx_t_59 = __pyx_v_s;
                                          __pyx_t_58 = __pyx_v_b;
                                          __pyx_t_62 = __pyx_v_leaf_n;
                                          __pyx_t_70 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_68 * __pyx_v_S.strides[0]) ) + __pyx_t_61 * __pyx_v_S.strides[1]) ) + __pyx_t_60 * __pyx_v_S.strides[2]) )) + __pyx_t_59)) )));
                                          __pyx_t_63 = 1;
                                          __pyx_t_53 = __pyx_v_va;
                                          __pyx_t_36 = __pyx_v_counter;
                                          __pyx_t_48 = __pyx_v_leaf_numb;
                                          __pyx_t_49 = __pyx_v_s;
                                          __pyx_t_55 = __pyx_v_i;
                                          __pyx_t_57 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_53 * __pyx_v_S.strides[0]) ) + __pyx_t_36 * __pyx_v_S.strides[1]) ) + __pyx_t_48 * __pyx_v_S.strides[2]) )) + __pyx_t_49)) )));
                                          __pyx_t_54 = __pyx_v_va;
                                          __pyx_t_37 = __pyx_v_counter;
                                          __pyx_t_40 = __pyx_v_leaf_numb;
                                          __pyx_t_39 = __pyx_v_s;
                                          __pyx_t_38 = __pyx_v_b;
                                          __pyx_t_50 = __pyx_v_leaf_n;
                                          __pyx_t_56 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_54 * __pyx_v_S.strides[0]) ) + __pyx_t_37 * __pyx_v_S.strides[1]) ) + __pyx_t_40 * __pyx_v_S.strides[2]) )) + __pyx_t_39)) )));
                                          __pyx_t_41 = 0;
                                          __pyx_t_15 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_69 * __pyx_v_X.strides[0]) ) + __pyx_t_71 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_58 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_62 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_55 * __pyx_v_X.strides[0]) ) + __pyx_t_57 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                          if (__pyx_t_15) {

                                            /* "cyext_acv.pyx":1694
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if o_all == S_size:
 */
                                            __pyx_v_o_all = (__pyx_v_o_all + 1);

                                            /* "cyext_acv.pyx":1693
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 * 
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":1696
 *                                         o_all = o_all + 1
 * 
 *                                 if o_all == S_size:             # <<<<<<<<<<<<<<
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 */
                                        __pyx_t_15 = ((__pyx_v_o_all == __pyx_v_S_size) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1697
 * 
 *                                 if o_all == S_size:
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1             # <<<<<<<<<<<<<<
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 */
                                          __pyx_t_39 = __pyx_v_va;
                                          __pyx_t_40 = __pyx_v_leaf_numb;
                                          __pyx_t_37 = __pyx_v_leaf_n;
                                          __pyx_t_54 = __pyx_v_counter;
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_39 * __pyx_v_cs.strides[0]) ) + __pyx_t_40 * __pyx_v_cs.strides[1]) ) + __pyx_t_37 * __pyx_v_cs.strides[2]) ) + __pyx_t_54 * __pyx_v_cs.strides[3]) )) = 1.0;

                                          /* "cyext_acv.pyx":1698
 *                                 if o_all == S_size:
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0             # <<<<<<<<<<<<<<
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 */
                                          __pyx_v_nv_bool = 0;

                                          /* "cyext_acv.pyx":1699
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1
 */
                                          __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                          __pyx_t_19 = __pyx_t_18;
                                          for (__pyx_t_74 = 0; __pyx_t_74 < __pyx_t_19; __pyx_t_74+=1) {
                                            __pyx_v_nv = __pyx_t_74;

                                            /* "cyext_acv.pyx":1700
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                             nv_bool = nv_bool + 1
 *                                             continue
 */
                                            __pyx_t_54 = __pyx_v_i;
                                            __pyx_t_41 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_37 = __pyx_v_b;
                                            __pyx_t_40 = __pyx_v_leaf_n;
                                            __pyx_t_56 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_57 = 1;
                                            __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_54 * __pyx_v_X.strides[0]) ) + __pyx_t_41 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                            if (!__pyx_t_42) {
                                            } else {
                                              __pyx_t_15 = __pyx_t_42;
                                              goto __pyx_L108_bool_binop_done;
                                            }
                                            __pyx_t_40 = __pyx_v_i;
                                            __pyx_t_57 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_37 = __pyx_v_b;
                                            __pyx_t_54 = __pyx_v_leaf_n;
                                            __pyx_t_56 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_41 = 0;
                                            __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_40 * __pyx_v_X.strides[0]) ) + __pyx_t_57 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                            __pyx_t_15 = __pyx_t_42;
                                            __pyx_L108_bool_binop_done:;
                                            if (__pyx_t_15) {

                                              /* "cyext_acv.pyx":1701
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                             continue
 * 
 */
                                              __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                              /* "cyext_acv.pyx":1702
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1
 *                                             continue             # <<<<<<<<<<<<<<
 * 
 *                                     if nv_bool == 0:
 */
                                              goto __pyx_L105_continue;

                                              /* "cyext_acv.pyx":1700
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                             nv_bool = nv_bool + 1
 *                                             continue
 */
                                            }
                                            __pyx_L105_continue:;
                                          }

                                          /* "cyext_acv.pyx":1704
 *                                             continue
 * 
 *                                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 */
                                          __pyx_t_15 = ((__pyx_v_nv_bool == 0) != 0);
                                          if (__pyx_t_15) {

                                            /* "cyext_acv.pyx":1705
 * 
 *                                     if nv_bool == 0:
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0
 */
                                            __pyx_t_54 = __pyx_v_va;
                                            __pyx_t_37 = __pyx_v_leaf_numb;
                                            __pyx_t_40 = __pyx_v_leaf_n;
                                            __pyx_t_39 = __pyx_v_counter;
                                            *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_54 * __pyx_v_csi.strides[0]) ) + __pyx_t_37 * __pyx_v_csi.strides[1]) ) + __pyx_t_40 * __pyx_v_csi.strides[2]) ) + __pyx_t_39 * __pyx_v_csi.strides[3]) )) = 1.0;

                                            /* "cyext_acv.pyx":1704
 *                                             continue
 * 
 *                                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":1696
 *                                         o_all = o_all + 1
 * 
 *                                 if o_all == S_size:             # <<<<<<<<<<<<<<
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 */
                                        }

                                        /* "cyext_acv.pyx":1707
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                             p_si = (csi[va, leaf_numb, leaf_numb, counter] * lm)/(lm_si[va, leaf_numb, leaf_numb, counter] * n_si) if n_si*lm_si[va, leaf_numb, leaf_numb, counter] !=0 else 0
 */
                                        __pyx_t_39 = __pyx_v_va;
                                        __pyx_t_40 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_leaf_n;
                                        __pyx_t_54 = __pyx_v_counter;
                                        if ((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_39 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_40 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_37 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_54 * __pyx_v_lm_si.strides[3]) ))) != 0.0) != 0)) {
                                          __pyx_t_50 = __pyx_v_va;
                                          __pyx_t_38 = __pyx_v_leaf_numb;
                                          __pyx_t_49 = __pyx_v_leaf_n;
                                          __pyx_t_48 = __pyx_v_counter;
                                          __pyx_t_36 = __pyx_v_va;
                                          __pyx_t_53 = __pyx_v_leaf_numb;
                                          __pyx_t_55 = __pyx_v_leaf_n;
                                          __pyx_t_59 = __pyx_v_counter;
                                          __pyx_t_60 = __pyx_v_va;
                                          __pyx_t_61 = __pyx_v_leaf_numb;
                                          __pyx_t_68 = __pyx_v_leaf_n;
                                          __pyx_t_62 = __pyx_v_counter;
                                          __pyx_t_72 = (__pyx_v_n_si + (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_50 * __pyx_v_csi.strides[0]) ) + __pyx_t_38 * __pyx_v_csi.strides[1]) ) + __pyx_t_49 * __pyx_v_csi.strides[2]) ) + __pyx_t_48 * __pyx_v_csi.strides[3]) ))) * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_36 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_53 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_55 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_59 * __pyx_v_lm_n.strides[3]) )))) / (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_60 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_61 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_68 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_62 * __pyx_v_lm_si.strides[3]) )))));
                                        } else {
                                          __pyx_t_72 = 0.0;
                                        }
                                        __pyx_v_n_si = __pyx_t_72;
                                      }

                                      /* "cyext_acv.pyx":1709
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0
 * 
 *                             p_si = (csi[va, leaf_numb, leaf_numb, counter] * lm)/(lm_si[va, leaf_numb, leaf_numb, counter] * n_si) if n_si*lm_si[va, leaf_numb, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 *                         else:
 * 
 */
                                      __pyx_t_54 = __pyx_v_va;
                                      __pyx_t_37 = __pyx_v_leaf_numb;
                                      __pyx_t_40 = __pyx_v_leaf_numb;
                                      __pyx_t_39 = __pyx_v_counter;
                                      if ((((__pyx_v_n_si * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_54 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_37 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_40 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_39 * __pyx_v_lm_si.strides[3]) )))) != 0.0) != 0)) {
                                        __pyx_t_62 = __pyx_v_va;
                                        __pyx_t_68 = __pyx_v_leaf_numb;
                                        __pyx_t_61 = __pyx_v_leaf_numb;
                                        __pyx_t_60 = __pyx_v_counter;
                                        __pyx_t_59 = __pyx_v_va;
                                        __pyx_t_55 = __pyx_v_leaf_numb;
                                        __pyx_t_53 = __pyx_v_leaf_numb;
                                        __pyx_t_36 = __pyx_v_counter;
                                        __pyx_t_72 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_62 * __pyx_v_csi.strides[0]) ) + __pyx_t_68 * __pyx_v_csi.strides[1]) ) + __pyx_t_61 * __pyx_v_csi.strides[2]) ) + __pyx_t_60 * __pyx_v_csi.strides[3]) ))) * __pyx_v_lm) / ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_59 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_55 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_53 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_36 * __pyx_v_lm_si.strides[3]) ))) * __pyx_v_n_si));
                                      } else {
                                        __pyx_t_72 = 0.0;
                                      }
                                      __pyx_v_p_si = __pyx_t_72;

                                      /* "cyext_acv.pyx":1683
 *                         n_s = 0
 *                         n_si = 0
 *                         if S_size == 0:             # <<<<<<<<<<<<<<
 *                             p_s = lm/data.shape[0]
 *                             n_s = 0
 */
                                      goto __pyx_L98;
                                    }

                                    /* "cyext_acv.pyx":1712
 *                         else:
 * 
 *                             for leaf_n in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 */
                                    /*else*/ {
                                      __pyx_t_35 = __pyx_v_nb_leaf;
                                      __pyx_t_47 = __pyx_t_35;
                                      for (__pyx_t_73 = 0; __pyx_t_73 < __pyx_t_47; __pyx_t_73+=1) {
                                        __pyx_v_leaf_n = __pyx_t_73;

                                        /* "cyext_acv.pyx":1713
 * 
 *                             for leaf_n in range(nb_leaf):
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 * 
 */
                                        __pyx_t_39 = __pyx_v_va;
                                        __pyx_t_40 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_leaf_n;
                                        __pyx_t_54 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_39 * __pyx_v_csi.strides[0]) ) + __pyx_t_40 * __pyx_v_csi.strides[1]) ) + __pyx_t_37 * __pyx_v_csi.strides[2]) ) + __pyx_t_54 * __pyx_v_csi.strides[3]) )) = 0.0;

                                        /* "cyext_acv.pyx":1714
 *                             for leaf_n in range(nb_leaf):
 *                                 csi[va, leaf_numb, leaf_n, counter] = 0
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0             # <<<<<<<<<<<<<<
 * 
 *                                 o_all = 0
 */
                                        __pyx_t_54 = __pyx_v_va;
                                        __pyx_t_37 = __pyx_v_leaf_numb;
                                        __pyx_t_40 = __pyx_v_leaf_n;
                                        __pyx_t_39 = __pyx_v_counter;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_54 * __pyx_v_cs.strides[0]) ) + __pyx_t_37 * __pyx_v_cs.strides[1]) ) + __pyx_t_40 * __pyx_v_cs.strides[2]) ) + __pyx_t_39 * __pyx_v_cs.strides[3]) )) = 0.0;

                                        /* "cyext_acv.pyx":1716
 *                                 cs[va, leaf_numb, leaf_n, counter] = 0
 * 
 *                                 o_all = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 */
                                        __pyx_v_o_all = 0;

                                        /* "cyext_acv.pyx":1717
 * 
 *                                 o_all = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                         o_all = o_all + 1
 */
                                        __pyx_t_51 = __pyx_v_S_size;
                                        __pyx_t_52 = __pyx_t_51;
                                        for (__pyx_t_74 = 0; __pyx_t_74 < __pyx_t_52; __pyx_t_74+=1) {
                                          __pyx_v_s = __pyx_t_74;

                                          /* "cyext_acv.pyx":1718
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 * 
 */
                                          __pyx_t_39 = __pyx_v_va;
                                          __pyx_t_40 = __pyx_v_counter;
                                          __pyx_t_37 = __pyx_v_leaf_numb;
                                          __pyx_t_54 = __pyx_v_s;
                                          __pyx_t_36 = __pyx_v_i;
                                          __pyx_t_41 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_39 * __pyx_v_S.strides[0]) ) + __pyx_t_40 * __pyx_v_S.strides[1]) ) + __pyx_t_37 * __pyx_v_S.strides[2]) )) + __pyx_t_54)) )));
                                          __pyx_t_53 = __pyx_v_va;
                                          __pyx_t_55 = __pyx_v_counter;
                                          __pyx_t_59 = __pyx_v_leaf_numb;
                                          __pyx_t_60 = __pyx_v_s;
                                          __pyx_t_61 = __pyx_v_b;
                                          __pyx_t_68 = __pyx_v_leaf_n;
                                          __pyx_t_56 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_53 * __pyx_v_S.strides[0]) ) + __pyx_t_55 * __pyx_v_S.strides[1]) ) + __pyx_t_59 * __pyx_v_S.strides[2]) )) + __pyx_t_60)) )));
                                          __pyx_t_57 = 1;
                                          __pyx_t_62 = __pyx_v_va;
                                          __pyx_t_48 = __pyx_v_counter;
                                          __pyx_t_49 = __pyx_v_leaf_numb;
                                          __pyx_t_38 = __pyx_v_s;
                                          __pyx_t_50 = __pyx_v_i;
                                          __pyx_t_63 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_62 * __pyx_v_S.strides[0]) ) + __pyx_t_48 * __pyx_v_S.strides[1]) ) + __pyx_t_49 * __pyx_v_S.strides[2]) )) + __pyx_t_38)) )));
                                          __pyx_t_58 = __pyx_v_va;
                                          __pyx_t_64 = __pyx_v_counter;
                                          __pyx_t_65 = __pyx_v_leaf_numb;
                                          __pyx_t_66 = __pyx_v_s;
                                          __pyx_t_67 = __pyx_v_b;
                                          __pyx_t_69 = __pyx_v_leaf_n;
                                          __pyx_t_70 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_58 * __pyx_v_S.strides[0]) ) + __pyx_t_64 * __pyx_v_S.strides[1]) ) + __pyx_t_65 * __pyx_v_S.strides[2]) )) + __pyx_t_66)) )));
                                          __pyx_t_71 = 0;
                                          __pyx_t_15 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_36 * __pyx_v_X.strides[0]) ) + __pyx_t_41 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_61 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_50 * __pyx_v_X.strides[0]) ) + __pyx_t_63 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_67 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                          if (__pyx_t_15) {

                                            /* "cyext_acv.pyx":1719
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
 *                                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if o_all == S_size:
 */
                                            __pyx_v_o_all = (__pyx_v_o_all + 1);

                                            /* "cyext_acv.pyx":1718
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 * 
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":1721
 *                                         o_all = o_all + 1
 * 
 *                                 if o_all == S_size:             # <<<<<<<<<<<<<<
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 */
                                        __pyx_t_15 = ((__pyx_v_o_all == __pyx_v_S_size) != 0);
                                        if (__pyx_t_15) {

                                          /* "cyext_acv.pyx":1722
 * 
 *                                 if o_all == S_size:
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1             # <<<<<<<<<<<<<<
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 */
                                          __pyx_t_66 = __pyx_v_va;
                                          __pyx_t_65 = __pyx_v_leaf_numb;
                                          __pyx_t_64 = __pyx_v_leaf_n;
                                          __pyx_t_58 = __pyx_v_counter;
                                          *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_66 * __pyx_v_cs.strides[0]) ) + __pyx_t_65 * __pyx_v_cs.strides[1]) ) + __pyx_t_64 * __pyx_v_cs.strides[2]) ) + __pyx_t_58 * __pyx_v_cs.strides[3]) )) = 1.0;

                                          /* "cyext_acv.pyx":1723
 *                                 if o_all == S_size:
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0             # <<<<<<<<<<<<<<
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 */
                                          __pyx_v_nv_bool = 0;

                                          /* "cyext_acv.pyx":1724
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1
 */
                                          __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                          __pyx_t_19 = __pyx_t_18;
                                          for (__pyx_t_74 = 0; __pyx_t_74 < __pyx_t_19; __pyx_t_74+=1) {
                                            __pyx_v_nv = __pyx_t_74;

                                            /* "cyext_acv.pyx":1725
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                             nv_bool = nv_bool + 1
 *                                             continue
 */
                                            __pyx_t_58 = __pyx_v_i;
                                            __pyx_t_71 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_64 = __pyx_v_b;
                                            __pyx_t_65 = __pyx_v_leaf_n;
                                            __pyx_t_70 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_63 = 1;
                                            __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_58 * __pyx_v_X.strides[0]) ) + __pyx_t_71 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_65 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                            if (!__pyx_t_42) {
                                            } else {
                                              __pyx_t_15 = __pyx_t_42;
                                              goto __pyx_L120_bool_binop_done;
                                            }
                                            __pyx_t_65 = __pyx_v_i;
                                            __pyx_t_63 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_64 = __pyx_v_b;
                                            __pyx_t_58 = __pyx_v_leaf_n;
                                            __pyx_t_70 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                            __pyx_t_71 = 0;
                                            __pyx_t_42 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_65 * __pyx_v_X.strides[0]) ) + __pyx_t_63 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_58 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                            __pyx_t_15 = __pyx_t_42;
                                            __pyx_L120_bool_binop_done:;
                                            if (__pyx_t_15) {

                                              /* "cyext_acv.pyx":1726
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                             continue
 * 
 */
                                              __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                              /* "cyext_acv.pyx":1727
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
 *                                             nv_bool = nv_bool + 1
 *                                             continue             # <<<<<<<<<<<<<<
 * 
 *                                     if nv_bool == 0:
 */
                                              goto __pyx_L117_continue;

                                              /* "cyext_acv.pyx":1725
 *                                     nv_bool = 0
 *                                     for nv in range(va_id[va].size()):
 *                                         if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                             nv_bool = nv_bool + 1
 *                                             continue
 */
                                            }
                                            __pyx_L117_continue:;
                                          }

                                          /* "cyext_acv.pyx":1729
 *                                             continue
 * 
 *                                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 */
                                          __pyx_t_15 = ((__pyx_v_nv_bool == 0) != 0);
                                          if (__pyx_t_15) {

                                            /* "cyext_acv.pyx":1730
 * 
 *                                     if nv_bool == 0:
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 n_s = n_s + (cs[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_s[va, leaf_numb, leaf_n, counter] if lm_s[va, leaf_numb, leaf_n, counter] !=0 else 0
 */
                                            __pyx_t_58 = __pyx_v_va;
                                            __pyx_t_64 = __pyx_v_leaf_numb;
                                            __pyx_t_65 = __pyx_v_leaf_n;
                                            __pyx_t_66 = __pyx_v_counter;
                                            *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_58 * __pyx_v_csi.strides[0]) ) + __pyx_t_64 * __pyx_v_csi.strides[1]) ) + __pyx_t_65 * __pyx_v_csi.strides[2]) ) + __pyx_t_66 * __pyx_v_csi.strides[3]) )) = 1.0;

                                            /* "cyext_acv.pyx":1729
 *                                             continue
 * 
 *                                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":1721
 *                                         o_all = o_all + 1
 * 
 *                                 if o_all == S_size:             # <<<<<<<<<<<<<<
 *                                     cs[va, leaf_numb, leaf_n, counter] = 1
 *                                     nv_bool = 0
 */
                                        }

                                        /* "cyext_acv.pyx":1732
 *                                         csi[va, leaf_numb, leaf_n, counter] = 1
 * 
 *                                 n_s = n_s + (cs[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_s[va, leaf_numb, leaf_n, counter] if lm_s[va, leaf_numb, leaf_n, counter] !=0 else 0             # <<<<<<<<<<<<<<
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0
 * 
 */
                                        __pyx_t_66 = __pyx_v_va;
                                        __pyx_t_65 = __pyx_v_leaf_numb;
                                        __pyx_t_64 = __pyx_v_leaf_n;
                                        __pyx_t_58 = __pyx_v_counter;
                                        if ((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_66 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_65 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_64 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_58 * __pyx_v_lm_s.strides[3]) ))) != 0.0) != 0)) {
                                          __pyx_t_69 = __pyx_v_va;
                                          __pyx_t_67 = __pyx_v_leaf_numb;
                                          __pyx_t_38 = __pyx_v_leaf_n;
                                          __pyx_t_49 = __pyx_v_counter;
                                          __pyx_t_48 = __pyx_v_va;
                                          __pyx_t_62 = __pyx_v_leaf_numb;
                                          __pyx_t_50 = __pyx_v_leaf_n;
                                          __pyx_t_60 = __pyx_v_counter;
                                          __pyx_t_59 = __pyx_v_va;
                                          __pyx_t_55 = __pyx_v_leaf_numb;
                                          __pyx_t_53 = __pyx_v_leaf_n;
                                          __pyx_t_68 = __pyx_v_counter;
                                          __pyx_t_72 = (__pyx_v_n_s + (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_69 * __pyx_v_cs.strides[0]) ) + __pyx_t_67 * __pyx_v_cs.strides[1]) ) + __pyx_t_38 * __pyx_v_cs.strides[2]) ) + __pyx_t_49 * __pyx_v_cs.strides[3]) ))) * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_48 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_62 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_50 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_60 * __pyx_v_lm_n.strides[3]) )))) / (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_59 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_55 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_53 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_68 * __pyx_v_lm_s.strides[3]) )))));
                                        } else {
                                          __pyx_t_72 = 0.0;
                                        }
                                        __pyx_v_n_s = __pyx_t_72;

                                        /* "cyext_acv.pyx":1733
 * 
 *                                 n_s = n_s + (cs[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_s[va, leaf_numb, leaf_n, counter] if lm_s[va, leaf_numb, leaf_n, counter] !=0 else 0
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                             p_s = (cs[va, leaf_numb, leaf_numb, counter] * lm)/(lm_s[va, leaf_numb, leaf_numb, counter] * n_s) if n_s*lm_s[va, leaf_numb, leaf_numb, counter] !=0 else 0
 */
                                        __pyx_t_58 = __pyx_v_va;
                                        __pyx_t_64 = __pyx_v_leaf_numb;
                                        __pyx_t_65 = __pyx_v_leaf_n;
                                        __pyx_t_66 = __pyx_v_counter;
                                        if ((((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_58 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_64 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_65 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_66 * __pyx_v_lm_si.strides[3]) ))) != 0.0) != 0)) {
                                          __pyx_t_68 = __pyx_v_va;
                                          __pyx_t_53 = __pyx_v_leaf_numb;
                                          __pyx_t_55 = __pyx_v_leaf_n;
                                          __pyx_t_59 = __pyx_v_counter;
                                          __pyx_t_60 = __pyx_v_va;
                                          __pyx_t_50 = __pyx_v_leaf_numb;
                                          __pyx_t_62 = __pyx_v_leaf_n;
                                          __pyx_t_48 = __pyx_v_counter;
                                          __pyx_t_49 = __pyx_v_va;
                                          __pyx_t_38 = __pyx_v_leaf_numb;
                                          __pyx_t_67 = __pyx_v_leaf_n;
                                          __pyx_t_69 = __pyx_v_counter;
                                          __pyx_t_72 = (__pyx_v_n_si + (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_68 * __pyx_v_csi.strides[0]) ) + __pyx_t_53 * __pyx_v_csi.strides[1]) ) + __pyx_t_55 * __pyx_v_csi.strides[2]) ) + __pyx_t_59 * __pyx_v_csi.strides[3]) ))) * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_n.data + __pyx_t_60 * __pyx_v_lm_n.strides[0]) ) + __pyx_t_50 * __pyx_v_lm_n.strides[1]) ) + __pyx_t_62 * __pyx_v_lm_n.strides[2]) ) + __pyx_t_48 * __pyx_v_lm_n.strides[3]) )))) / (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_49 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_67 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_69 * __pyx_v_lm_si.strides[3]) )))));
                                        } else {
                                          __pyx_t_72 = 0.0;
                                        }
                                        __pyx_v_n_si = __pyx_t_72;
                                      }

                                      /* "cyext_acv.pyx":1735
 *                                 n_si = n_si + (csi[va, leaf_numb, leaf_n, counter]*lm_n[va, leaf_numb, leaf_n, counter])/lm_si[va, leaf_numb, leaf_n, counter] if lm_si[va, leaf_numb, leaf_n, counter] !=0 else 0
 * 
 *                             p_s = (cs[va, leaf_numb, leaf_numb, counter] * lm)/(lm_s[va, leaf_numb, leaf_numb, counter] * n_s) if n_s*lm_s[va, leaf_numb, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 *                             p_si = (csi[va, leaf_numb, leaf_numb, counter] * lm)/(lm_si[va, leaf_numb,  leaf_numb, counter] * n_si) if n_si*lm_si[va, leaf_numb, leaf_numb, counter] !=0 else 0
 * 
 */
                                      __pyx_t_66 = __pyx_v_va;
                                      __pyx_t_65 = __pyx_v_leaf_numb;
                                      __pyx_t_64 = __pyx_v_leaf_numb;
                                      __pyx_t_58 = __pyx_v_counter;
                                      if ((((__pyx_v_n_s * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_66 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_65 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_64 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_58 * __pyx_v_lm_s.strides[3]) )))) != 0.0) != 0)) {
                                        __pyx_t_69 = __pyx_v_va;
                                        __pyx_t_67 = __pyx_v_leaf_numb;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_49 = __pyx_v_counter;
                                        __pyx_t_48 = __pyx_v_va;
                                        __pyx_t_62 = __pyx_v_leaf_numb;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_60 = __pyx_v_counter;
                                        __pyx_t_72 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_cs.data + __pyx_t_69 * __pyx_v_cs.strides[0]) ) + __pyx_t_67 * __pyx_v_cs.strides[1]) ) + __pyx_t_38 * __pyx_v_cs.strides[2]) ) + __pyx_t_49 * __pyx_v_cs.strides[3]) ))) * __pyx_v_lm) / ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_48 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_62 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_50 * __pyx_v_lm_s.strides[2]) ) + __pyx_t_60 * __pyx_v_lm_s.strides[3]) ))) * __pyx_v_n_s));
                                      } else {
                                        __pyx_t_72 = 0.0;
                                      }
                                      __pyx_v_p_s = __pyx_t_72;

                                      /* "cyext_acv.pyx":1736
 * 
 *                             p_s = (cs[va, leaf_numb, leaf_numb, counter] * lm)/(lm_s[va, leaf_numb, leaf_numb, counter] * n_s) if n_s*lm_s[va, leaf_numb, leaf_numb, counter] !=0 else 0
 *                             p_si = (csi[va, leaf_numb, leaf_numb, counter] * lm)/(lm_si[va, leaf_numb,  leaf_numb, counter] * n_si) if n_si*lm_si[va, leaf_numb, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         for nv in range(va_id[va].size()):
 */
                                      __pyx_t_58 = __pyx_v_va;
                                      __pyx_t_64 = __pyx_v_leaf_numb;
                                      __pyx_t_65 = __pyx_v_leaf_numb;
                                      __pyx_t_66 = __pyx_v_counter;
                                      if ((((__pyx_v_n_si * (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_58 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_64 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_65 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_66 * __pyx_v_lm_si.strides[3]) )))) != 0.0) != 0)) {
                                        __pyx_t_60 = __pyx_v_va;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_62 = __pyx_v_leaf_numb;
                                        __pyx_t_48 = __pyx_v_counter;
                                        __pyx_t_49 = __pyx_v_va;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_67 = __pyx_v_leaf_numb;
                                        __pyx_t_69 = __pyx_v_counter;
                                        __pyx_t_72 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_csi.data + __pyx_t_60 * __pyx_v_csi.strides[0]) ) + __pyx_t_50 * __pyx_v_csi.strides[1]) ) + __pyx_t_62 * __pyx_v_csi.strides[2]) ) + __pyx_t_48 * __pyx_v_csi.strides[3]) ))) * __pyx_v_lm) / ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_49 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_67 * __pyx_v_lm_si.strides[2]) ) + __pyx_t_69 * __pyx_v_lm_si.strides[3]) ))) * __pyx_v_n_si));
                                      } else {
                                        __pyx_t_72 = 0.0;
                                      }
                                      __pyx_v_p_si = __pyx_t_72;
                                    }
                                    __pyx_L98:;

                                    /* "cyext_acv.pyx":1738
 *                             p_si = (csi[va, leaf_numb, leaf_numb, counter] * lm)/(lm_si[va, leaf_numb,  leaf_numb, counter] * n_si) if n_si*lm_si[va, leaf_numb, leaf_numb, counter] !=0 else 0
 * 
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                    __pyx_t_18 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_19 = __pyx_t_18;
                                    for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_19; __pyx_t_35+=1) {
                                      __pyx_v_nv = __pyx_t_35;

                                      /* "cyext_acv.pyx":1739
 * 
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                      __pyx_t_47 = __pyx_v_d;
                                      __pyx_t_73 = __pyx_t_47;
                                      for (__pyx_t_74 = 0; __pyx_t_74 < __pyx_t_73; __pyx_t_74+=1) {
                                        __pyx_v_i2 = __pyx_t_74;

                                        /* "cyext_acv.pyx":1740
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
                                        __pyx_t_66 = __pyx_v_b;
                                        __pyx_t_65 = __pyx_v_leaf_numb;
                                        __pyx_t_64 = __pyx_v_b;
                                        __pyx_t_71 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_66 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_65 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_58 = __pyx_v_i2;
                                        __pyx_t_69 = __pyx_v_leaf_numb;
                                        __pyx_t_67 = __pyx_v_counter;
                                        __pyx_t_38 = __pyx_v_i;
                                        __pyx_t_70 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_49 = __pyx_v_i2;
                                        *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_69 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_67 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_70 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_49 * __pyx_v_phi_b.strides[4]) )) += (((__pyx_v_coef_0 + __pyx_v_coef) * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_64 * __pyx_v_values.strides[0]) ) + __pyx_t_71 * __pyx_v_values.strides[1]) ) + __pyx_t_58 * __pyx_v_values.strides[2]) ))));
                                      }
                                    }
                                  }
                                }
                                __pyx_L64_continue:;
                              }
                              goto __pyx_L128;
                              __pyx_L39_error:;
                              {
                                  #ifdef WITH_THREAD
                                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                  #endif
                                  #ifdef _OPENMP
                                  #pragma omp flush(__pyx_parallel_exc_type)
                                  #endif /* _OPENMP */
                                  if (!__pyx_parallel_exc_type) {
                                    __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                    __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                    __Pyx_GOTREF(__pyx_parallel_exc_type);
                                  }
                                  #ifdef WITH_THREAD
                                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                  #endif
                              }
                              __pyx_parallel_why = 4;
                              goto __pyx_L127;
                              __pyx_L127:;
                              #ifdef _OPENMP
                              #pragma omp critical(__pyx_parallel_lastprivates1)
                              #endif /* _OPENMP */
                              {
                                  __pyx_parallel_temp0 = __pyx_v_S_size;
                                  __pyx_parallel_temp1 = __pyx_v_a_it;
                                  __pyx_parallel_temp2 = __pyx_v_add;
                                  __pyx_parallel_temp3 = __pyx_v_b_it;
                                  __pyx_parallel_temp4 = __pyx_v_ci;
                                  __pyx_parallel_temp5 = __pyx_v_cj;
                                  __pyx_parallel_temp6 = __pyx_v_coef;
                                  __pyx_parallel_temp7 = __pyx_v_coef_0;
                                  __pyx_parallel_temp8 = __pyx_v_counter;
                                  __pyx_parallel_temp9 = __pyx_v_i;
                                  __pyx_parallel_temp10 = __pyx_v_i2;
                                  __pyx_parallel_temp11 = __pyx_v_l;
                                  __pyx_parallel_temp12 = __pyx_v_leaf_n;
                                  __pyx_parallel_temp13 = __pyx_v_leaf_numb;
                                  __pyx_parallel_temp14 = __pyx_v_lm;
                                  __pyx_parallel_temp15 = __pyx_v_n_s;
                                  __pyx_parallel_temp16 = __pyx_v_n_si;
                                  __pyx_parallel_temp17 = __pyx_v_ns;
                                  __pyx_parallel_temp18 = __pyx_v_nv;
                                  __pyx_parallel_temp19 = __pyx_v_nv_bool;
                                  __pyx_parallel_temp20 = __pyx_v_o_all;
                                  __pyx_parallel_temp21 = __pyx_v_p_s;
                                  __pyx_parallel_temp22 = __pyx_v_p_si;
                                  __pyx_parallel_temp23 = __pyx_v_pow_set_size;
                                  __pyx_parallel_temp24 = __pyx_v_s;
                                  __pyx_parallel_temp25 = __pyx_v_set_size;
                                  __pyx_parallel_temp26 = __pyx_v_va;
                                  __pyx_parallel_temp27 = __pyx_v_va_size;
                              }
                              __pyx_L128:;
                              #ifdef _OPENMP
                              #pragma omp flush(__pyx_parallel_why)
                              #endif /* _OPENMP */
                          }
                      }
                      #ifdef _OPENMP
                      Py_END_ALLOW_THREADS
                      #else
{
#ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      #endif /* _OPENMP */
                      /* Clean up any temporaries */
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      #ifndef _OPENMP
}
#endif /* _OPENMP */
                  }
              }
              if (__pyx_parallel_exc_type) {
                /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
                __pyx_parallel_why = 4;
              }
              if (__pyx_parallel_why) {
                __pyx_v_S_size = __pyx_parallel_temp0;
                __pyx_v_a_it = __pyx_parallel_temp1;
                __pyx_v_add = __pyx_parallel_temp2;
                __pyx_v_b_it = __pyx_parallel_temp3;
                __pyx_v_ci = __pyx_parallel_temp4;
                __pyx_v_cj = __pyx_parallel_temp5;
                __pyx_v_coef = __pyx_parallel_temp6;
                __pyx_v_coef_0 = __pyx_parallel_temp7;
                __pyx_v_counter = __pyx_parallel_temp8;
                __pyx_v_i = __pyx_parallel_temp9;
                __pyx_v_i2 = __pyx_parallel_temp10;
                __pyx_v_l = __pyx_parallel_temp11;
                __pyx_v_leaf_n = __pyx_parallel_temp12;
                __pyx_v_leaf_numb = __pyx_parallel_temp13;
                __pyx_v_lm = __pyx_parallel_temp14;
                __pyx_v_n_s = __pyx_parallel_temp15;
                __pyx_v_n_si = __pyx_parallel_temp16;
                __pyx_v_ns = __pyx_parallel_temp17;
                __pyx_v_nv = __pyx_parallel_temp18;
                __pyx_v_nv_bool = __pyx_parallel_temp19;
                __pyx_v_o_all = __pyx_parallel_temp20;
                __pyx_v_p_s = __pyx_parallel_temp21;
                __pyx_v_p_si = __pyx_parallel_temp22;
                __pyx_v_pow_set_size = __pyx_parallel_temp23;
                __pyx_v_s = __pyx_parallel_temp24;
                __pyx_v_set_size = __pyx_parallel_temp25;
                __pyx_v_va = __pyx_parallel_temp26;
                __pyx_v_va_size = __pyx_parallel_temp27;
                switch (__pyx_parallel_why) {
                      case 4:
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      __Pyx_GIVEREF(__pyx_parallel_exc_type);
                      __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                      __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                  }
                  goto __pyx_L35_error;
                }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":1583
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L36;
          }
          __pyx_L35_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L36:;
        }
    }

    /* "cyext_acv.pyx":1742
 *                                 phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             for j in range(m):
 *                 for i2 in range(d):
 */
    __pyx_t_27 = __pyx_v_N;
    __pyx_t_26 = __pyx_t_27;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_26; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;

      /* "cyext_acv.pyx":1743
 * 
 *         for i in range(N):
 *             for j in range(m):             # <<<<<<<<<<<<<<
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 */
      __pyx_t_28 = __pyx_v_m;
      __pyx_t_29 = __pyx_t_28;
      for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
        __pyx_v_j = __pyx_t_30;

        /* "cyext_acv.pyx":1744
 *         for i in range(N):
 *             for j in range(m):
 *                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):
 */
        __pyx_t_31 = __pyx_v_d;
        __pyx_t_32 = __pyx_t_31;
        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
          __pyx_v_i2 = __pyx_t_33;

          /* "cyext_acv.pyx":1745
 *             for j in range(m):
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 */
          __pyx_t_22 = (__pyx_v_phi_b.shape[0]);
          __pyx_t_23 = __pyx_t_22;
          for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_23; __pyx_t_34+=1) {
            __pyx_v_leaf_numb = __pyx_t_34;

            /* "cyext_acv.pyx":1746
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 * 
 */
            __pyx_t_24 = (__pyx_v_phi_b.shape[1]);
            __pyx_t_25 = __pyx_t_24;
            for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_25; __pyx_t_35+=1) {
              __pyx_v_counter = __pyx_t_35;

              /* "cyext_acv.pyx":1747
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(phi)/m
 */
              __pyx_t_65 = __pyx_v_leaf_numb;
              __pyx_t_66 = __pyx_v_counter;
              __pyx_t_58 = __pyx_v_i;
              __pyx_t_64 = __pyx_v_j;
              __pyx_t_49 = __pyx_v_i2;
              __pyx_t_38 = __pyx_v_i;
              __pyx_t_67 = __pyx_v_j;
              __pyx_t_69 = __pyx_v_i2;
              *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_38 * __pyx_v_phi.strides[0]) ) + __pyx_t_67 * __pyx_v_phi.strides[1]) ) + __pyx_t_69 * __pyx_v_phi.strides[2]) )) += (*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_65 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_66 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_58 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_64 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_49 * __pyx_v_phi_b.strides[4]) )));
            }
          }
        }
      }
    }
  }

  /* "cyext_acv.pyx":1749
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 * 
 *     return np.asarray(phi)/m             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_phi, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_9 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":1492
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_leaves_normalized(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_21, 1);
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_si, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_csi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_cs, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_23shap_values_leaves_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_23shap_values_leaves_normalized(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  long __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shap_values_leaves_normalized (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 1); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 2); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 3); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 4); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 5); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 6); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 7); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 8); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, 9); __PYX_ERR(0, 1492, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "shap_values_leaves_normalized") < 0)) __PYX_ERR(0, 1492, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 1492, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 1493, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 1494, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 1495, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 1496, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 1497, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_long(values[6]); if (unlikely((__pyx_v_scaling == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1498, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1499, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[8]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1500, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shap_values_leaves_normalized", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1492, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_22shap_values_leaves_normalized(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_22shap_values_leaves_normalized(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, long __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_leaves_normalized", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 1492, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_shap_values_leaves_normalized(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.shap_values_leaves_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":1755
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_acv_leaves(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_25shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_acv_leaves(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  const __pyx_v_C, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_ns;
  unsigned int __pyx_v_add;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_i2;
  unsigned int __pyx_v_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_csi;
  unsigned int __pyx_v_cs;
  unsigned int __pyx_v_len_n_star;
  double __pyx_v_lm;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_va_c = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_N_star_a = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_S_size;
  int __pyx_v_va_size;
  int __pyx_v_b_it;
  int __pyx_v_nv_bool;
  long __pyx_v_set_size;
  std::vector<std::vector<long> >  __pyx_v_C_buff;
  std::vector<std::vector<long> >  __pyx_v_va_id;
  std::vector<std::vector<long> >  __pyx_v_buff;
  std::vector<long>  __pyx_v_coal_va;
  std::vector<long>  __pyx_v_remove_va;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id_v2;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_C_b;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id;
  std::vector<std::vector<std::vector<std::vector<long> > > >  __pyx_v_node_id_va;
  double __pyx_v_p_s;
  double __pyx_v_p_si;
  double __pyx_v_coef;
  double __pyx_v_coef_0;
  double __pyx_v_p_off;
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_si = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  int __pyx_t_17;
  std::vector<std::vector<long> > ::size_type __pyx_t_18;
  std::vector<std::vector<long> > ::size_type __pyx_t_19;
  std::vector<long> ::size_type __pyx_t_20;
  std::vector<long> ::size_type __pyx_t_21;
  std::vector<long>  __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  size_t __pyx_t_25;
  std::vector<std::vector<long> >  __pyx_t_26;
  __Pyx_memviewslice __pyx_t_27 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  unsigned int __pyx_t_37;
  size_t __pyx_t_38;
  size_t __pyx_t_39;
  size_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  unsigned int __pyx_t_43;
  size_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  int __pyx_t_46;
  std::vector<long> ::size_type __pyx_t_47;
  std::vector<long> ::size_type __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  size_t __pyx_t_51;
  size_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  size_t __pyx_t_54;
  size_t __pyx_t_55;
  size_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  unsigned int __pyx_t_59;
  unsigned int __pyx_t_60;
  long __pyx_t_61;
  long __pyx_t_62;
  size_t __pyx_t_63;
  size_t __pyx_t_64;
  size_t __pyx_t_65;
  size_t __pyx_t_66;
  size_t __pyx_t_67;
  size_t __pyx_t_68;
  size_t __pyx_t_69;
  size_t __pyx_t_70;
  size_t __pyx_t_71;
  size_t __pyx_t_72;
  size_t __pyx_t_73;
  size_t __pyx_t_74;
  unsigned int __pyx_t_75;
  unsigned int __pyx_t_76;
  double __pyx_t_77;
  unsigned int __pyx_t_78;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_acv_leaves", 0);

  /* "cyext_acv.pyx":1766
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":1767
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":1768
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":1769
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":1770
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] phi
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":1773
 * 
 *     cdef double[:, :, :] phi
 *     phi = np.zeros((N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1776
 * 
 *     cdef double[ :, :, :,  :, :] phi_b
 *     phi_b = np.zeros((max_leaves, 2**scaling, N, m, d))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_8);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_9);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1776, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_b = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":1780
 * 
 *     cdef double[:, :, :,  :] phi_n
 *     phi_n = np.zeros((max_leaves, N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :, :, ::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_6 = 0;
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_n = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":1783
 * 
 *     cdef long[:, :, :, ::1] S
 *     S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_S = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":1792
 *     cdef int len_va_c, S_size, va_size,  b_it, nv_bool
 *     cdef long set_size
 *     va_c = np.zeros((m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[long]] C_buff, va_id, buff, Sm
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_13 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_13.memview)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_va_c = __pyx_t_13;
  __pyx_t_13.memview = NULL;
  __pyx_t_13.data = NULL;

  /* "cyext_acv.pyx":1796
 *     cdef vector[vector[long]] C_buff, va_id, buff, Sm
 *     cdef vector[long] coal_va, remove_va, node_id_b
 *     buff.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     for i in range(max_leaves):
 *         buff[i].resize(1)
 */
  try {
    __pyx_v_buff.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1796, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1797
 *     cdef vector[long] coal_va, remove_va, node_id_b
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):             # <<<<<<<<<<<<<<
 *         buff[i].resize(1)
 * 
 */
  __pyx_t_14 = __pyx_v_max_leaves;
  __pyx_t_15 = __pyx_t_14;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "cyext_acv.pyx":1798
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):
 *         buff[i].resize(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    try {
      (__pyx_v_buff[__pyx_v_i]).resize(1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1798, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1802
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, node_id
 *     node_id_v2.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     C_b.resize(max_leaves)
 *     # Sm.resize(max_leaves)
 */
  try {
    __pyx_v_node_id_v2.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1802, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1803
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, node_id
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     # Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)
 */
  try {
    __pyx_v_C_b.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1803, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1805
 *     C_b.resize(max_leaves)
 *     # Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 */
  try {
    __pyx_v_node_id.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1805, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1808
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)             # <<<<<<<<<<<<<<
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)
 */
  try {
    __pyx_v_node_id_va.resize(__pyx_v_m);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1808, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":1809
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         node_id_va[i].resize(max_leaves)
 * 
 */
  __pyx_t_14 = __pyx_v_m;
  __pyx_t_15 = __pyx_t_14;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "cyext_acv.pyx":1810
 *     node_id_va.resize(m)
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     N_star_a = np.array(N_star)
 */
    try {
      (__pyx_v_node_id_va[__pyx_v_i]).resize(__pyx_v_max_leaves);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1810, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":1812
 *         node_id_va[i].resize(max_leaves)
 * 
 *     N_star_a = np.array(N_star)             # <<<<<<<<<<<<<<
 *     len_n_star = N_star_a.shape[0]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_N_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_13 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_13.memview)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_N_star_a = __pyx_t_13;
  __pyx_t_13.memview = NULL;
  __pyx_t_13.data = NULL;

  /* "cyext_acv.pyx":1813
 * 
 *     N_star_a = np.array(N_star)
 *     len_n_star = N_star_a.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if C[0].size() != 0:
 */
  __pyx_v_len_n_star = (__pyx_v_N_star_a.shape[0]);

  /* "cyext_acv.pyx":1815
 *     len_n_star = N_star_a.shape[0]
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 */
  __pyx_t_17 = (((__pyx_v_C[0]).size() != 0) != 0);
  if (__pyx_t_17) {

    /* "cyext_acv.pyx":1816
 * 
 *     if C[0].size() != 0:
 *         C_buff = C             # <<<<<<<<<<<<<<
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 * 
 */
    __pyx_v_C_buff = __pyx_v_C;

    /* "cyext_acv.pyx":1817
 *     if C[0].size() != 0:
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(S_star.shape[0]):
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_18 = __pyx_v_C.size();
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_19; __pyx_t_14+=1) {
      __pyx_v_ci = __pyx_t_14;
      __pyx_t_20 = (__pyx_v_C[__pyx_v_ci]).size();
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_21; __pyx_t_15+=1) {
        __pyx_v_cj = __pyx_t_15;
        __pyx_t_2 = __Pyx_PyInt_From_long(((__pyx_v_C[__pyx_v_ci])[__pyx_v_cj])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1817, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 1817, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __pyx_t_22 = __pyx_convert_vector_from_py_long(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1817, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_coal_va = __pyx_t_22;

    /* "cyext_acv.pyx":1819
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 * 
 *         for i in range(S_star.shape[0]):             # <<<<<<<<<<<<<<
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])
 */
    __pyx_t_23 = (__pyx_v_S_star.shape[0]);
    __pyx_t_24 = __pyx_t_23;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_24; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;

      /* "cyext_acv.pyx":1820
 * 
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 */
      __pyx_t_25 = __pyx_v_i;
      __pyx_t_17 = ((!((std::find<std::vector<long> ::iterator,long>(__pyx_v_coal_va.begin(), __pyx_v_coal_va.end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_25 * __pyx_v_S_star.strides[0]) )))) != __pyx_v_coal_va.end()) != 0)) != 0);
      if (__pyx_t_17) {

        /* "cyext_acv.pyx":1821
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])             # <<<<<<<<<<<<<<
 *                 buff[0][0] = S_star[i]
 *                 va_id.push_back(buff[0])
 */
        __pyx_t_25 = __pyx_v_i;
        try {
          __pyx_v_remove_va.push_back((*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_25 * __pyx_v_S_star.strides[0]) ))));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1821, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1822
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]             # <<<<<<<<<<<<<<
 *                 va_id.push_back(buff[0])
 *             else:
 */
        __pyx_t_25 = __pyx_v_i;
        ((__pyx_v_buff[0])[0]) = (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_25 * __pyx_v_S_star.strides[0]) )));

        /* "cyext_acv.pyx":1823
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 *                 va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 for ci in range(C_buff.size()):
 */
        try {
          __pyx_v_va_id.push_back((__pyx_v_buff[0]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 1823, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":1820
 * 
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 */
        goto __pyx_L14;
      }

      /* "cyext_acv.pyx":1825
 *                 va_id.push_back(buff[0])
 *             else:
 *                 for ci in range(C_buff.size()):             # <<<<<<<<<<<<<<
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 */
      /*else*/ {
        __pyx_t_18 = __pyx_v_C_buff.size();
        __pyx_t_19 = __pyx_t_18;
        for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_19; __pyx_t_15+=1) {
          __pyx_v_ci = __pyx_t_15;

          /* "cyext_acv.pyx":1826
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          __pyx_t_25 = __pyx_v_i;
          __pyx_t_17 = ((std::find<std::vector<long> ::iterator,long>((__pyx_v_C_buff[__pyx_v_ci]).begin(), (__pyx_v_C_buff[__pyx_v_ci]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_25 * __pyx_v_S_star.strides[0]) )))) != (__pyx_v_C_buff[__pyx_v_ci]).end()) != 0);
          if (__pyx_t_17) {

            /* "cyext_acv.pyx":1827
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])             # <<<<<<<<<<<<<<
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 */
            try {
              __pyx_v_va_id.push_back((__pyx_v_C_buff[__pyx_v_ci]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 1827, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":1828
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])             # <<<<<<<<<<<<<<
 *                         C_buff.pop_back()
 *                         break
 */
            (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(__pyx_v_C_buff.begin(), __pyx_v_C_buff.end(), (__pyx_v_C_buff[__pyx_v_ci])));

            /* "cyext_acv.pyx":1829
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_v_C_buff.pop_back();

            /* "cyext_acv.pyx":1830
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 *                         break             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L16_break;

            /* "cyext_acv.pyx":1826
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          }
        }
        __pyx_L16_break:;
      }
      __pyx_L14:;
    }

    /* "cyext_acv.pyx":1815
 *     len_n_star = N_star_a.shape[0]
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 */
    goto __pyx_L7;
  }

  /* "cyext_acv.pyx":1833
 * 
 *     else:
 *         va_id = [[S_star[i]] for i in range(S_star.shape[0])]             # <<<<<<<<<<<<<<
 * 
 *     m = va_id.size()
 */
  /*else*/ {
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_23 = (__pyx_v_S_star.shape[0]);
    __pyx_t_24 = __pyx_t_23;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_24; __pyx_t_14+=1) {
      __pyx_v_i = __pyx_t_14;
      __pyx_t_25 = __pyx_v_i;
      __pyx_t_2 = __Pyx_PyInt_From_long((*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_25 * __pyx_v_S_star.strides[0]) )))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_26 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1833, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_va_id = __pyx_t_26;
  }
  __pyx_L7:;

  /* "cyext_acv.pyx":1835
 *         va_id = [[S_star[i]] for i in range(S_star.shape[0])]
 * 
 *     m = va_id.size()             # <<<<<<<<<<<<<<
 *     cdef double p_s, p_si, coef, coef_0, p_off
 *     cdef double[:, :, :] lm_s, lm_si
 */
  __pyx_v_m = __pyx_v_va_id.size();

  /* "cyext_acv.pyx":1839
 *     cdef double[:, :, :] lm_s, lm_si
 *     cdef double[:, :] lm_star
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_9);
  __pyx_t_5 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_s = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1840
 *     cdef double[:, :] lm_star
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
  __pyx_t_2 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_si = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":1841
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_star = np.zeros((va_id.size(), max_leaves))             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_27 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_27.memview)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_star = __pyx_t_27;
  __pyx_t_27.memview = NULL;
  __pyx_t_27.data = NULL;

  /* "cyext_acv.pyx":1843
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             for j in range(data.shape[1]):
 */
  __pyx_t_14 = __pyx_v_n_trees;
  __pyx_t_15 = __pyx_t_14;
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
    __pyx_v_b = __pyx_t_16;

    /* "cyext_acv.pyx":1844
 * 
 *     for b in range(n_trees):
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):
 */
    __pyx_t_28 = __pyx_v_N;
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
      __pyx_v_i = __pyx_t_30;

      /* "cyext_acv.pyx":1845
 *     for b in range(n_trees):
 *         for i in range(N):
 *             for j in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 */
      __pyx_t_23 = (__pyx_v_data.shape[1]);
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_24; __pyx_t_31+=1) {
        __pyx_v_j = __pyx_t_31;

        /* "cyext_acv.pyx":1846
 *         for i in range(N):
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi_n[leaf_numb, i, j, i2] = 0
 */
        __pyx_t_32 = __pyx_v_d;
        __pyx_t_33 = __pyx_t_32;
        for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
          __pyx_v_i2 = __pyx_t_34;

          /* "cyext_acv.pyx":1847
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                         phi_n[leaf_numb, i, j, i2] = 0
 *                         for counter in range(phi_b.shape[1]):
 */
          __pyx_t_35 = (__pyx_v_phi_b.shape[0]);
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_leaf_numb = __pyx_t_37;

            /* "cyext_acv.pyx":1848
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi_n[leaf_numb, i, j, i2] = 0             # <<<<<<<<<<<<<<
 *                         for counter in range(phi_b.shape[1]):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 */
            __pyx_t_25 = __pyx_v_leaf_numb;
            __pyx_t_38 = __pyx_v_i;
            __pyx_t_39 = __pyx_v_j;
            __pyx_t_40 = __pyx_v_i2;
            *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_25 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_38 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_39 * __pyx_v_phi_n.strides[2]) ) + __pyx_t_40 * __pyx_v_phi_n.strides[3]) )) = 0.0;

            /* "cyext_acv.pyx":1849
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi_n[leaf_numb, i, j, i2] = 0
 *                         for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 */
            __pyx_t_41 = (__pyx_v_phi_b.shape[1]);
            __pyx_t_42 = __pyx_t_41;
            for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_42; __pyx_t_43+=1) {
              __pyx_v_counter = __pyx_t_43;

              /* "cyext_acv.pyx":1850
 *                         phi_n[leaf_numb, i, j, i2] = 0
 *                         for counter in range(phi_b.shape[1]):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
              __pyx_t_40 = __pyx_v_leaf_numb;
              __pyx_t_39 = __pyx_v_counter;
              __pyx_t_38 = __pyx_v_i;
              __pyx_t_25 = __pyx_v_j;
              __pyx_t_44 = __pyx_v_i2;
              *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_40 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_25 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_44 * __pyx_v_phi_b.strides[4]) )) = 0.0;
            }
          }
        }
      }
    }

    /* "cyext_acv.pyx":1851
 *                         for counter in range(phi_b.shape[1]):
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 */
    __pyx_t_44 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ ((char *) (((long const  *) __pyx_v_leaves_nb.data) + __pyx_t_44)) )));

    /* "cyext_acv.pyx":1852
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_28 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              int __pyx_parallel_temp0 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp1 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp2 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp3 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp4 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp5 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp6 = ((double)__PYX_NAN());
              double __pyx_parallel_temp7 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp8 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp9 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp10 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp11 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp12 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp13 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp14 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp15 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp16 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp17 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp18 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp19 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp20 = ((double)__PYX_NAN());
              double __pyx_parallel_temp21 = ((double)__PYX_NAN());
              double __pyx_parallel_temp22 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp23 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp24 = ((unsigned int)0xbad0bad0);
              long __pyx_parallel_temp25 = ((long)0xbad0bad0);
              unsigned int __pyx_parallel_temp26 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp27 = ((int)0xbad0bad0);
              const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
              PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
              int __pyx_parallel_why;
              __pyx_parallel_why = 0;
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_30 = (__pyx_t_28 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_30 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57, __pyx_t_58, __pyx_t_59, __pyx_t_60, __pyx_t_61, __pyx_t_62, __pyx_t_63, __pyx_t_64, __pyx_t_65, __pyx_t_66, __pyx_t_67, __pyx_t_68, __pyx_t_69, __pyx_t_70, __pyx_t_71, __pyx_t_72, __pyx_t_73, __pyx_t_74, __pyx_t_75, __pyx_t_76, __pyx_t_77, __pyx_t_78) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      Py_BEGIN_ALLOW_THREADS
                      #endif /* _OPENMP */
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_S_size) lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_add) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_ci) lastprivate(__pyx_v_cj) lastprivate(__pyx_v_coef) lastprivate(__pyx_v_coef_0) lastprivate(__pyx_v_counter) lastprivate(__pyx_v_cs) lastprivate(__pyx_v_csi) lastprivate(__pyx_v_i) lastprivate(__pyx_v_i2) lastprivate(__pyx_v_l) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_ns) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_off) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_si) lastprivate(__pyx_v_pow_set_size) lastprivate(__pyx_v_s) lastprivate(__pyx_v_set_size) lastprivate(__pyx_v_va) lastprivate(__pyx_v_va_size) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_30; __pyx_t_29++){
                          if (__pyx_parallel_why < 2)
                          {
                              __pyx_v_leaf_numb = (unsigned int)(0 + 1 * __pyx_t_29);
                              /* Initialize private variables to invalid values */
                              __pyx_v_S_size = ((int)0xbad0bad0);
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_add = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((int)0xbad0bad0);
                              __pyx_v_ci = ((unsigned int)0xbad0bad0);
                              __pyx_v_cj = ((unsigned int)0xbad0bad0);
                              __pyx_v_coef = ((double)__PYX_NAN());
                              __pyx_v_coef_0 = ((double)__PYX_NAN());
                              __pyx_v_counter = ((unsigned int)0xbad0bad0);
                              __pyx_v_cs = ((unsigned int)0xbad0bad0);
                              __pyx_v_csi = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((unsigned int)0xbad0bad0);
                              __pyx_v_i2 = ((unsigned int)0xbad0bad0);
                              __pyx_v_l = ((unsigned int)0xbad0bad0);
                              __pyx_v_lm = ((double)__PYX_NAN());
                              __pyx_v_ns = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv_bool = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_off = ((double)__PYX_NAN());
                              __pyx_v_p_s = ((double)__PYX_NAN());
                              __pyx_v_p_si = ((double)__PYX_NAN());
                              __pyx_v_pow_set_size = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_set_size = ((long)0xbad0bad0);
                              __pyx_v_va = ((unsigned int)0xbad0bad0);
                              __pyx_v_va_size = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":1853
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 */
                              (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).clear();

                              /* "cyext_acv.pyx":1854
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):             # <<<<<<<<<<<<<<
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 */
                              __pyx_t_23 = (__pyx_v_N_star.shape[0]);
                              __pyx_t_24 = __pyx_t_23;
                              for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_24; __pyx_t_31+=1) {
                                __pyx_v_i = __pyx_t_31;

                                /* "cyext_acv.pyx":1855
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):             # <<<<<<<<<<<<<<
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 */
                                __pyx_t_44 = __pyx_v_i;
                                __pyx_t_17 = ((std::find<std::vector<long> ::iterator,long>(((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_N_star.data + __pyx_t_44 * __pyx_v_N_star.strides[0]) )))) != ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end()) != 0);
                                if (__pyx_t_17) {

                                  /* "cyext_acv.pyx":1856
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])             # <<<<<<<<<<<<<<
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0
 */
                                  __pyx_t_44 = __pyx_v_i;
                                  (void)(std::remove<std::vector<long> ::iterator,long>(((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_N_star.data + __pyx_t_44 * __pyx_v_N_star.strides[0]) )))));

                                  /* "cyext_acv.pyx":1857
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *             lm = 0
 *             for i in range(data.shape[0]):
 */
                                  ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).pop_back();

                                  /* "cyext_acv.pyx":1855
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):             # <<<<<<<<<<<<<<
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 */
                                }
                              }

                              /* "cyext_acv.pyx":1858
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 */
                              __pyx_v_lm = 0.0;

                              /* "cyext_acv.pyx":1859
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 */
                              __pyx_t_23 = (__pyx_v_data.shape[0]);
                              __pyx_t_24 = __pyx_t_23;
                              for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_24; __pyx_t_31+=1) {
                                __pyx_v_i = __pyx_t_31;

                                /* "cyext_acv.pyx":1860
 *             lm = 0
 *             for i in range(data.shape[0]):
 *                 a_it = 0             # <<<<<<<<<<<<<<
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 */
                                __pyx_v_a_it = 0;

                                /* "cyext_acv.pyx":1861
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 *                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 */
                                __pyx_t_35 = (__pyx_v_data.shape[1]);
                                __pyx_t_36 = __pyx_t_35;
                                for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_36; __pyx_t_32+=1) {
                                  __pyx_v_s = __pyx_t_32;

                                  /* "cyext_acv.pyx":1862
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  __pyx_t_44 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_38 = __pyx_v_b;
                                  __pyx_t_39 = __pyx_v_leaf_numb;
                                  __pyx_t_40 = __pyx_v_s;
                                  __pyx_t_45 = 1;
                                  __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_44 * __pyx_v_data.strides[0]) ) + __pyx_t_25 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_46) {
                                  } else {
                                    __pyx_t_17 = __pyx_t_46;
                                    goto __pyx_L49_bool_binop_done;
                                  }
                                  __pyx_t_40 = __pyx_v_i;
                                  __pyx_t_39 = __pyx_v_s;
                                  __pyx_t_38 = __pyx_v_b;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_44 = __pyx_v_s;
                                  __pyx_t_45 = 0;
                                  __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_40 * __pyx_v_data.strides[0]) ) + __pyx_t_39 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_17 = __pyx_t_46;
                                  __pyx_L49_bool_binop_done:;
                                  if (__pyx_t_17) {

                                    /* "cyext_acv.pyx":1863
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1
 */
                                    __pyx_v_a_it = (__pyx_v_a_it + 1);

                                    /* "cyext_acv.pyx":1862
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1864
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                __pyx_t_17 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                if (__pyx_t_17) {

                                  /* "cyext_acv.pyx":1865
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1.0);

                                  /* "cyext_acv.pyx":1864
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":1868
 * 
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                              __pyx_t_17 = (((__pyx_v_C[0]).size() != 0) != 0);
                              if (__pyx_t_17) {

                                /* "cyext_acv.pyx":1869
 * 
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C             # <<<<<<<<<<<<<<
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 */
                                (__pyx_v_C_b[__pyx_v_leaf_numb]) = __pyx_v_C;

                                /* "cyext_acv.pyx":1870
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 */
                                __pyx_t_20 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_21 = __pyx_t_20;
                                for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_21; __pyx_t_31+=1) {
                                  __pyx_v_nv = __pyx_t_31;

                                  /* "cyext_acv.pyx":1871
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0             # <<<<<<<<<<<<<<
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 */
                                  __pyx_v_add = 0;

                                  /* "cyext_acv.pyx":1872
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 *                     for ns in range(remove_va.size()):             # <<<<<<<<<<<<<<
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 */
                                  __pyx_t_47 = __pyx_v_remove_va.size();
                                  __pyx_t_48 = __pyx_t_47;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_48; __pyx_t_32+=1) {
                                    __pyx_v_ns = __pyx_t_32;

                                    /* "cyext_acv.pyx":1873
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    __pyx_t_17 = (((((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]) == (__pyx_v_remove_va[__pyx_v_ns])) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1874
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1             # <<<<<<<<<<<<<<
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 */
                                      __pyx_v_add = 1;

                                      /* "cyext_acv.pyx":1875
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]             # <<<<<<<<<<<<<<
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 */
                                      ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]);

                                      /* "cyext_acv.pyx":1876
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 *                             break
 *                     if add == 0:
 */
                                      try {
                                        (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                      } catch(...) {
                                        #ifdef WITH_THREAD
                                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                        #endif
                                        __Pyx_CppExn2PyErr();
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        __PYX_ERR(0, 1876, __pyx_L39_error)
                                      }

                                      /* "cyext_acv.pyx":1877
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break             # <<<<<<<<<<<<<<
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 */
                                      goto __pyx_L56_break;

                                      /* "cyext_acv.pyx":1873
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    }
                                  }
                                  __pyx_L56_break:;

                                  /* "cyext_acv.pyx":1878
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  __pyx_t_17 = ((__pyx_v_add == 0) != 0);
                                  if (__pyx_t_17) {

                                    /* "cyext_acv.pyx":1879
 *                             break
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 */
                                    __pyx_t_18 = (__pyx_v_C_b[__pyx_v_leaf_numb]).size();
                                    __pyx_t_19 = __pyx_t_18;
                                    for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_19; __pyx_t_32+=1) {
                                      __pyx_v_ci = __pyx_t_32;

                                      /* "cyext_acv.pyx":1880
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 */
                                      __pyx_t_47 = ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_48 = __pyx_t_47;
                                      for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_48; __pyx_t_33+=1) {
                                        __pyx_v_cj = __pyx_t_33;

                                        /* "cyext_acv.pyx":1881
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        __pyx_t_17 = (((((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]) == (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv])) != 0);
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":1882
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1             # <<<<<<<<<<<<<<
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 */
                                          __pyx_v_add = 1;

                                          /* "cyext_acv.pyx":1883
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                     break
 *                             if add == 1:
 */
                                          try {
                                            (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back(((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]));
                                          } catch(...) {
                                            #ifdef WITH_THREAD
                                            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                            #endif
                                            __Pyx_CppExn2PyErr();
                                            #ifdef WITH_THREAD
                                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                            #endif
                                            __PYX_ERR(0, 1883, __pyx_L39_error)
                                          }

                                          /* "cyext_acv.pyx":1884
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break             # <<<<<<<<<<<<<<
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 */
                                          goto __pyx_L62_break;

                                          /* "cyext_acv.pyx":1881
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        }
                                      }
                                      __pyx_L62_break:;

                                      /* "cyext_acv.pyx":1885
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      __pyx_t_17 = ((__pyx_v_add == 1) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1886
 *                                     break
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 */
                                        (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >((__pyx_v_C_b[__pyx_v_leaf_numb]).begin(), (__pyx_v_C_b[__pyx_v_leaf_numb]).end(), ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])));

                                        /* "cyext_acv.pyx":1887
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 */
                                        (__pyx_v_C_b[__pyx_v_leaf_numb]).pop_back();

                                        /* "cyext_acv.pyx":1888
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 *                                 break             # <<<<<<<<<<<<<<
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 */
                                        goto __pyx_L60_break;

                                        /* "cyext_acv.pyx":1885
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      }
                                    }
                                    __pyx_L60_break:;

                                    /* "cyext_acv.pyx":1878
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":1889
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]             # <<<<<<<<<<<<<<
 *             else:
 *                 node_id[leaf_numb].clear()
 */
                                (__pyx_v_node_id[__pyx_v_leaf_numb]) = (__pyx_v_node_id_v2[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1868
 * 
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                                goto __pyx_L52;
                              }

                              /* "cyext_acv.pyx":1891
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 *                 node_id[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 */
                              /*else*/ {
                                (__pyx_v_node_id[__pyx_v_leaf_numb]).clear();

                                /* "cyext_acv.pyx":1892
 *             else:
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 */
                                __pyx_t_20 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_21 = __pyx_t_20;
                                for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_21; __pyx_t_31+=1) {
                                  __pyx_v_i = __pyx_t_31;

                                  /* "cyext_acv.pyx":1893
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]             # <<<<<<<<<<<<<<
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 */
                                  ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_i]);

                                  /* "cyext_acv.pyx":1894
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 * 
 *             for va in range(va_id.size()):
 */
                                  try {
                                    (__pyx_v_node_id[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                  } catch(...) {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    __Pyx_CppExn2PyErr();
                                    #ifdef WITH_THREAD
                                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                    #endif
                                    __PYX_ERR(0, 1894, __pyx_L39_error)
                                  }
                                }
                              }
                              __pyx_L52:;

                              /* "cyext_acv.pyx":1896
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 *             for va in range(va_id.size()):             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 */
                              __pyx_t_18 = __pyx_v_va_id.size();
                              __pyx_t_19 = __pyx_t_18;
                              for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_19; __pyx_t_31+=1) {
                                __pyx_v_va = __pyx_t_31;

                                /* "cyext_acv.pyx":1897
 * 
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]             # <<<<<<<<<<<<<<
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     lm_star[va, leaf_numb] = 0
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]) = (__pyx_v_node_id[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":1898
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):
 */
                                __pyx_t_17 = ((!((std::find<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])) != ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end()) != 0)) != 0);
                                if (__pyx_t_17) {

                                  /* "cyext_acv.pyx":1899
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     lm_star[va, leaf_numb] = 0             # <<<<<<<<<<<<<<
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 */
                                  __pyx_t_44 = __pyx_v_va;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_44 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_25 * __pyx_v_lm_star.strides[1]) )) = 0.0;

                                  /* "cyext_acv.pyx":1900
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(len_n_star):
 */
                                  __pyx_t_23 = (__pyx_v_data.shape[0]);
                                  __pyx_t_24 = __pyx_t_23;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_24; __pyx_t_32+=1) {
                                    __pyx_v_i = __pyx_t_32;

                                    /* "cyext_acv.pyx":1901
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":1902
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_33 = __pyx_v_len_n_star;
                                    __pyx_t_34 = __pyx_t_33;
                                    for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_34; __pyx_t_37+=1) {
                                      __pyx_v_s = __pyx_t_37;

                                      /* "cyext_acv.pyx":1903
 *                         b_it = 0
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_25 = __pyx_v_s;
                                      __pyx_t_44 = __pyx_v_i;
                                      __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_25 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_38 = __pyx_v_s;
                                      __pyx_t_39 = __pyx_v_b;
                                      __pyx_t_40 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_38 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_50 = 1;
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_52 = __pyx_v_i;
                                      __pyx_t_53 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_51 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_54 = __pyx_v_s;
                                      __pyx_t_55 = __pyx_v_b;
                                      __pyx_t_56 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_54 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_58 = 0;
                                      __pyx_t_17 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_44 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_52 * __pyx_v_data.strides[0]) ) + __pyx_t_53 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_56 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_58 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1904
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == len_n_star:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":1903
 *                         b_it = 0
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1906
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1
 * 
 */
                                    __pyx_t_17 = ((__pyx_v_b_it == __pyx_v_len_n_star) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1907
 * 
 *                         if b_it == len_n_star:
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(N):
 */
                                      __pyx_t_54 = __pyx_v_va;
                                      __pyx_t_56 = __pyx_v_leaf_numb;
                                      __pyx_t_55 = __pyx_v_va;
                                      __pyx_t_51 = __pyx_v_leaf_numb;
                                      *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_55 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_51 * __pyx_v_lm_star.strides[1]) )) = ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_54 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_56 * __pyx_v_lm_star.strides[1]) ))) + 1.0);

                                      /* "cyext_acv.pyx":1906
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1909
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1
 * 
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_si = 0
 */
                                  __pyx_t_32 = __pyx_v_N;
                                  __pyx_t_33 = __pyx_t_32;
                                  for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                    __pyx_v_i = __pyx_t_34;

                                    /* "cyext_acv.pyx":1910
 * 
 *                     for i in range(N):
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_si = 0
 * 
 */
                                    __pyx_v_p_s = 0.0;

                                    /* "cyext_acv.pyx":1911
 *                     for i in range(N):
 *                         p_s = 0
 *                         p_si = 0             # <<<<<<<<<<<<<<
 * 
 *                         o_all = 0
 */
                                    __pyx_v_p_si = 0.0;

                                    /* "cyext_acv.pyx":1913
 *                         p_si = 0
 * 
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         csi = 0
 *                         for s in range(len_n_star):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":1914
 * 
 *                         o_all = 0
 *                         csi = 0             # <<<<<<<<<<<<<<
 *                         for s in range(len_n_star):
 *                             if ((X[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 */
                                    __pyx_v_csi = 0;

                                    /* "cyext_acv.pyx":1915
 *                         o_all = 0
 *                         csi = 0
 *                         for s in range(len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((X[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_37 = __pyx_v_len_n_star;
                                    __pyx_t_43 = __pyx_t_37;
                                    for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_43; __pyx_t_59+=1) {
                                      __pyx_v_s = __pyx_t_59;

                                      /* "cyext_acv.pyx":1916
 *                         csi = 0
 *                         for s in range(len_n_star):
 *                             if ((X[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      __pyx_t_56 = __pyx_v_s;
                                      __pyx_t_54 = __pyx_v_i;
                                      __pyx_t_58 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_56 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_55 = __pyx_v_b;
                                      __pyx_t_52 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_51 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_53 = 1;
                                      __pyx_t_38 = __pyx_v_s;
                                      __pyx_t_40 = __pyx_v_i;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_38 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_39 = __pyx_v_s;
                                      __pyx_t_25 = __pyx_v_b;
                                      __pyx_t_44 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_39 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_45 = 0;
                                      __pyx_t_17 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_54 * __pyx_v_X.strides[0]) ) + __pyx_t_58 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_40 * __pyx_v_X.strides[0]) ) + __pyx_t_50 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1917
 *                         for s in range(len_n_star):
 *                             if ((X[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                         if o_all == len_n_star:
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":1916
 *                         csi = 0
 *                         for s in range(len_n_star):
 *                             if ((X[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1919
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == len_n_star:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    __pyx_t_17 = ((__pyx_v_o_all == __pyx_v_len_n_star) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1920
 * 
 *                         if o_all == len_n_star:
 *                             csi = 1             # <<<<<<<<<<<<<<
 * 
 *                         p_s = lm/data.shape[0]
 */
                                      __pyx_v_csi = 1;

                                      /* "cyext_acv.pyx":1919
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == len_n_star:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":1922
 *                             csi = 1
 * 
 *                         p_s = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                         p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                         for nv in range(va_id[va].size()):
 */
                                    __pyx_v_p_s = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                    /* "cyext_acv.pyx":1923
 * 
 *                         p_s = lm/data.shape[0]
 *                         p_si = (csi * lm)/lm_star[va, leaf_numb]             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 */
                                    __pyx_t_39 = __pyx_v_va;
                                    __pyx_t_44 = __pyx_v_leaf_numb;
                                    __pyx_v_p_si = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_39 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_44 * __pyx_v_lm_star.strides[1]) ))));

                                    /* "cyext_acv.pyx":1924
 *                         p_s = lm/data.shape[0]
 *                         p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             for i2 in range(d):
 *                                 phi_n[leaf_numb, i, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                    __pyx_t_20 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_21 = __pyx_t_20;
                                    for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_21; __pyx_t_37+=1) {
                                      __pyx_v_nv = __pyx_t_37;

                                      /* "cyext_acv.pyx":1925
 *                         p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                 phi_n[leaf_numb, i, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 *                     continue
 */
                                      __pyx_t_43 = __pyx_v_d;
                                      __pyx_t_59 = __pyx_t_43;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_59; __pyx_t_60+=1) {
                                        __pyx_v_i2 = __pyx_t_60;

                                        /* "cyext_acv.pyx":1926
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 *                                 phi_n[leaf_numb, i, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                        __pyx_t_44 = __pyx_v_b;
                                        __pyx_t_39 = __pyx_v_leaf_numb;
                                        __pyx_t_25 = __pyx_v_b;
                                        __pyx_t_45 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_44 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_38 = __pyx_v_i2;
                                        __pyx_t_40 = __pyx_v_leaf_numb;
                                        __pyx_t_51 = __pyx_v_i;
                                        __pyx_t_49 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_52 = __pyx_v_i2;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_40 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_51 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_49 * __pyx_v_phi_n.strides[2]) ) + __pyx_t_52 * __pyx_v_phi_n.strides[3]) )) += ((__pyx_v_p_si - __pyx_v_p_s) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_25 * __pyx_v_values.strides[0]) ) + __pyx_t_45 * __pyx_v_values.strides[1]) ) + __pyx_t_38 * __pyx_v_values.strides[2]) ))));
                                      }
                                    }
                                  }

                                  /* "cyext_acv.pyx":1927
 *                             for i2 in range(d):
 *                                 phi_n[leaf_numb, i, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 */
                                  goto __pyx_L67_continue;

                                  /* "cyext_acv.pyx":1898
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):
 */
                                }

                                /* "cyext_acv.pyx":1929
 *                     continue
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 */
                                (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])));

                                /* "cyext_acv.pyx":1930
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 *                 node_id_va[va][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).pop_back();

                                /* "cyext_acv.pyx":1932
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()             # <<<<<<<<<<<<<<
 *                 pow_set_size = 2**set_size
 * 
 */
                                __pyx_v_set_size = ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).size();

                                /* "cyext_acv.pyx":1933
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 *                 pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *                 for counter in range(0, pow_set_size):
 */
                                __pyx_v_pow_set_size = __Pyx_pow_long(2, __pyx_v_set_size);

                                /* "cyext_acv.pyx":1935
 *                 pow_set_size = 2**set_size
 * 
 *                 for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *                     va_size = 0
 *                     S_size = 0
 */
                                __pyx_t_32 = __pyx_v_pow_set_size;
                                __pyx_t_33 = __pyx_t_32;
                                for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                                  __pyx_v_counter = __pyx_t_34;

                                  /* "cyext_acv.pyx":1936
 * 
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0             # <<<<<<<<<<<<<<
 *                     S_size = 0
 *                     for ci in range(set_size):
 */
                                  __pyx_v_va_size = 0;

                                  /* "cyext_acv.pyx":1937
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0
 *                     S_size = 0             # <<<<<<<<<<<<<<
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 */
                                  __pyx_v_S_size = 0;

                                  /* "cyext_acv.pyx":1938
 *                     va_size = 0
 *                     S_size = 0
 *                     for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 */
                                  __pyx_t_61 = __pyx_v_set_size;
                                  __pyx_t_62 = __pyx_t_61;
                                  for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_62; __pyx_t_37+=1) {
                                    __pyx_v_ci = __pyx_t_37;

                                    /* "cyext_acv.pyx":1939
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    __pyx_t_17 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1940
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 */
                                      __pyx_t_20 = (((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_21 = __pyx_t_20;
                                      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_21; __pyx_t_43+=1) {
                                        __pyx_v_cj = __pyx_t_43;

                                        /* "cyext_acv.pyx":1941
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]             # <<<<<<<<<<<<<<
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1
 */
                                        __pyx_t_39 = __pyx_v_va;
                                        __pyx_t_44 = __pyx_v_counter;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = __pyx_v_S_size;
                                        *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_39 * __pyx_v_S.strides[0]) ) + __pyx_t_44 * __pyx_v_S.strides[1]) ) + __pyx_t_38 * __pyx_v_S.strides[2]) )) + __pyx_t_45)) )) = ((((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]);

                                        /* "cyext_acv.pyx":1942
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                             va_size = va_size + 1
 * 
 */
                                        __pyx_v_S_size = (__pyx_v_S_size + 1);
                                      }

                                      /* "cyext_acv.pyx":1943
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 */
                                      __pyx_v_va_size = (__pyx_v_va_size + 1);

                                      /* "cyext_acv.pyx":1939
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1945
 *                             va_size = va_size + 1
 * 
 *                     lm_s[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     lm_si[va, leaf_numb, counter] = 0
 * 
 */
                                  __pyx_t_38 = __pyx_v_va;
                                  __pyx_t_44 = __pyx_v_leaf_numb;
                                  __pyx_t_39 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_38 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_44 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_39 * __pyx_v_lm_s.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":1946
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 * 
 *                     for nv in range(len_n_star):
 */
                                  __pyx_t_39 = __pyx_v_va;
                                  __pyx_t_44 = __pyx_v_leaf_numb;
                                  __pyx_t_38 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_39 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_44 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":1948
 *                     lm_si[va, leaf_numb, counter] = 0
 * 
 *                     for nv in range(len_n_star):             # <<<<<<<<<<<<<<
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]
 * 
 */
                                  __pyx_t_37 = __pyx_v_len_n_star;
                                  __pyx_t_43 = __pyx_t_37;
                                  for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_43; __pyx_t_59+=1) {
                                    __pyx_v_nv = __pyx_t_59;

                                    /* "cyext_acv.pyx":1949
 * 
 *                     for nv in range(len_n_star):
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(data.shape[0]):
 */
                                    __pyx_t_38 = __pyx_v_nv;
                                    __pyx_t_44 = __pyx_v_va;
                                    __pyx_t_39 = __pyx_v_counter;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_52 = (__pyx_v_S_size + __pyx_v_nv);
                                    *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_44 * __pyx_v_S.strides[0]) ) + __pyx_t_39 * __pyx_v_S.strides[1]) ) + __pyx_t_25 * __pyx_v_S.strides[2]) )) + __pyx_t_52)) )) = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_38 * __pyx_v_N_star_a.strides[0]) )));
                                  }

                                  /* "cyext_acv.pyx":1951
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]
 * 
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 */
                                  __pyx_t_23 = (__pyx_v_data.shape[0]);
                                  __pyx_t_24 = __pyx_t_23;
                                  for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_24; __pyx_t_37+=1) {
                                    __pyx_v_i = __pyx_t_37;

                                    /* "cyext_acv.pyx":1952
 * 
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":1953
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_43 = (__pyx_v_S_size + __pyx_v_len_n_star);
                                    __pyx_t_59 = __pyx_t_43;
                                    for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_59; __pyx_t_60+=1) {
                                      __pyx_v_s = __pyx_t_60;

                                      /* "cyext_acv.pyx":1954
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_38 = __pyx_v_va;
                                      __pyx_t_52 = __pyx_v_counter;
                                      __pyx_t_25 = __pyx_v_leaf_numb;
                                      __pyx_t_39 = __pyx_v_s;
                                      __pyx_t_44 = __pyx_v_i;
                                      __pyx_t_45 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_38 * __pyx_v_S.strides[0]) ) + __pyx_t_52 * __pyx_v_S.strides[1]) ) + __pyx_t_25 * __pyx_v_S.strides[2]) )) + __pyx_t_39)) )));
                                      __pyx_t_51 = __pyx_v_va;
                                      __pyx_t_40 = __pyx_v_counter;
                                      __pyx_t_55 = __pyx_v_leaf_numb;
                                      __pyx_t_56 = __pyx_v_s;
                                      __pyx_t_54 = __pyx_v_b;
                                      __pyx_t_63 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) ) + __pyx_t_40 * __pyx_v_S.strides[1]) ) + __pyx_t_55 * __pyx_v_S.strides[2]) )) + __pyx_t_56)) )));
                                      __pyx_t_50 = 1;
                                      __pyx_t_64 = __pyx_v_va;
                                      __pyx_t_65 = __pyx_v_counter;
                                      __pyx_t_66 = __pyx_v_leaf_numb;
                                      __pyx_t_67 = __pyx_v_s;
                                      __pyx_t_68 = __pyx_v_i;
                                      __pyx_t_53 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_64 * __pyx_v_S.strides[0]) ) + __pyx_t_65 * __pyx_v_S.strides[1]) ) + __pyx_t_66 * __pyx_v_S.strides[2]) )) + __pyx_t_67)) )));
                                      __pyx_t_69 = __pyx_v_va;
                                      __pyx_t_70 = __pyx_v_counter;
                                      __pyx_t_71 = __pyx_v_leaf_numb;
                                      __pyx_t_72 = __pyx_v_s;
                                      __pyx_t_73 = __pyx_v_b;
                                      __pyx_t_74 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_69 * __pyx_v_S.strides[0]) ) + __pyx_t_70 * __pyx_v_S.strides[1]) ) + __pyx_t_71 * __pyx_v_S.strides[2]) )) + __pyx_t_72)) )));
                                      __pyx_t_58 = 0;
                                      __pyx_t_17 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_44 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_68 * __pyx_v_data.strides[0]) ) + __pyx_t_53 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_73 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_74 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_58 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1955
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == S_size + len_n_star:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":1954
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1957
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 */
                                    __pyx_t_17 = ((__pyx_v_b_it == (__pyx_v_S_size + __pyx_v_len_n_star)) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1958
 * 
 *                         if b_it == S_size + len_n_star:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 */
                                      __pyx_t_72 = __pyx_v_va;
                                      __pyx_t_71 = __pyx_v_leaf_numb;
                                      __pyx_t_70 = __pyx_v_counter;
                                      __pyx_t_69 = __pyx_v_va;
                                      __pyx_t_74 = __pyx_v_leaf_numb;
                                      __pyx_t_73 = __pyx_v_counter;
                                      *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_69 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_74 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_73 * __pyx_v_lm_s.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_72 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_71 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_70 * __pyx_v_lm_s.strides[2]) ))) + 1.0);

                                      /* "cyext_acv.pyx":1959
 *                         if b_it == S_size + len_n_star:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":1960
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_20 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_21 = __pyx_t_20;
                                      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_21; __pyx_t_43+=1) {
                                        __pyx_v_nv = __pyx_t_43;

                                        /* "cyext_acv.pyx":1961
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_70 = __pyx_v_i;
                                        __pyx_t_58 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_71 = __pyx_v_b;
                                        __pyx_t_72 = __pyx_v_leaf_numb;
                                        __pyx_t_57 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_53 = 1;
                                        __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_70 * __pyx_v_data.strides[0]) ) + __pyx_t_58 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_72 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_46) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_46;
                                          goto __pyx_L104_bool_binop_done;
                                        }
                                        __pyx_t_72 = __pyx_v_i;
                                        __pyx_t_53 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_71 = __pyx_v_b;
                                        __pyx_t_70 = __pyx_v_leaf_numb;
                                        __pyx_t_57 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_58 = 0;
                                        __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_72 * __pyx_v_data.strides[0]) ) + __pyx_t_53 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_58 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_17 = __pyx_t_46;
                                        __pyx_L104_bool_binop_done:;
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":1962
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":1963
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L101_continue;

                                          /* "cyext_acv.pyx":1961
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L101_continue:;
                                      }

                                      /* "cyext_acv.pyx":1965
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      __pyx_t_17 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1966
 * 
 *                             if nv_bool == 0:
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(N):
 */
                                        __pyx_t_70 = __pyx_v_va;
                                        __pyx_t_71 = __pyx_v_leaf_numb;
                                        __pyx_t_72 = __pyx_v_counter;
                                        __pyx_t_73 = __pyx_v_va;
                                        __pyx_t_74 = __pyx_v_leaf_numb;
                                        __pyx_t_69 = __pyx_v_counter;
                                        *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_73 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_74 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_69 * __pyx_v_lm_si.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_70 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_71 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_72 * __pyx_v_lm_si.strides[2]) ))) + 1.0);

                                        /* "cyext_acv.pyx":1965
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":1957
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":1968
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 * 
 *                         p_s = 0
 */
                                  __pyx_t_37 = __pyx_v_N;
                                  __pyx_t_43 = __pyx_t_37;
                                  for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_43; __pyx_t_59+=1) {
                                    __pyx_v_i = __pyx_t_59;

                                    /* "cyext_acv.pyx":1970
 *                     for i in range(N):
 * 
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_si = 0
 * 
 */
                                    __pyx_v_p_s = 0.0;

                                    /* "cyext_acv.pyx":1971
 * 
 *                         p_s = 0
 *                         p_si = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                    __pyx_v_p_si = 0.0;

                                    /* "cyext_acv.pyx":1974
 * 
 * 
 *                         csi = 0             # <<<<<<<<<<<<<<
 *                         cs = 0
 *                         o_all = 0
 */
                                    __pyx_v_csi = 0;

                                    /* "cyext_acv.pyx":1975
 * 
 *                         csi = 0
 *                         cs = 0             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size + len_n_star):
 */
                                    __pyx_v_cs = 0;

                                    /* "cyext_acv.pyx":1976
 *                         csi = 0
 *                         cs = 0
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size + len_n_star):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":1977
 *                         cs = 0
 *                         o_all = 0
 *                         for s in range(S_size + len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_60 = (__pyx_v_S_size + __pyx_v_len_n_star);
                                    __pyx_t_75 = __pyx_t_60;
                                    for (__pyx_t_76 = 0; __pyx_t_76 < __pyx_t_75; __pyx_t_76+=1) {
                                      __pyx_v_s = __pyx_t_76;

                                      /* "cyext_acv.pyx":1978
 *                         o_all = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      __pyx_t_72 = __pyx_v_va;
                                      __pyx_t_71 = __pyx_v_counter;
                                      __pyx_t_70 = __pyx_v_leaf_numb;
                                      __pyx_t_69 = __pyx_v_s;
                                      __pyx_t_74 = __pyx_v_i;
                                      __pyx_t_58 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_72 * __pyx_v_S.strides[0]) ) + __pyx_t_71 * __pyx_v_S.strides[1]) ) + __pyx_t_70 * __pyx_v_S.strides[2]) )) + __pyx_t_69)) )));
                                      __pyx_t_73 = __pyx_v_va;
                                      __pyx_t_67 = __pyx_v_counter;
                                      __pyx_t_66 = __pyx_v_leaf_numb;
                                      __pyx_t_65 = __pyx_v_s;
                                      __pyx_t_64 = __pyx_v_b;
                                      __pyx_t_68 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_73 * __pyx_v_S.strides[0]) ) + __pyx_t_67 * __pyx_v_S.strides[1]) ) + __pyx_t_66 * __pyx_v_S.strides[2]) )) + __pyx_t_65)) )));
                                      __pyx_t_53 = 1;
                                      __pyx_t_56 = __pyx_v_va;
                                      __pyx_t_55 = __pyx_v_counter;
                                      __pyx_t_40 = __pyx_v_leaf_numb;
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_63 = __pyx_v_i;
                                      __pyx_t_50 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_56 * __pyx_v_S.strides[0]) ) + __pyx_t_55 * __pyx_v_S.strides[1]) ) + __pyx_t_40 * __pyx_v_S.strides[2]) )) + __pyx_t_51)) )));
                                      __pyx_t_54 = __pyx_v_va;
                                      __pyx_t_39 = __pyx_v_counter;
                                      __pyx_t_25 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = __pyx_v_s;
                                      __pyx_t_38 = __pyx_v_b;
                                      __pyx_t_44 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_54 * __pyx_v_S.strides[0]) ) + __pyx_t_39 * __pyx_v_S.strides[1]) ) + __pyx_t_25 * __pyx_v_S.strides[2]) )) + __pyx_t_52)) )));
                                      __pyx_t_45 = 0;
                                      __pyx_t_17 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_74 * __pyx_v_X.strides[0]) ) + __pyx_t_58 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_57 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_63 * __pyx_v_X.strides[0]) ) + __pyx_t_50 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1979
 *                         for s in range(S_size + len_n_star):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                         if o_all == S_size + len_n_star:
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":1978
 *                         o_all = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":1981
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             cs = 1
 *                             nv_bool = 0
 */
                                    __pyx_t_17 = ((__pyx_v_o_all == (__pyx_v_S_size + __pyx_v_len_n_star)) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":1982
 * 
 *                         if o_all == S_size + len_n_star:
 *                             cs = 1             # <<<<<<<<<<<<<<
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 */
                                      __pyx_v_cs = 1;

                                      /* "cyext_acv.pyx":1983
 *                         if o_all == S_size + len_n_star:
 *                             cs = 1
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":1984
 *                             cs = 1
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_20 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_21 = __pyx_t_20;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_21; __pyx_t_60+=1) {
                                        __pyx_v_nv = __pyx_t_60;

                                        /* "cyext_acv.pyx":1985
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_52 = __pyx_v_i;
                                        __pyx_t_45 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_25 = __pyx_v_b;
                                        __pyx_t_39 = __pyx_v_leaf_numb;
                                        __pyx_t_49 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_50 = 1;
                                        __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_52 * __pyx_v_X.strides[0]) ) + __pyx_t_45 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_46) {
                                        } else {
                                          __pyx_t_17 = __pyx_t_46;
                                          goto __pyx_L116_bool_binop_done;
                                        }
                                        __pyx_t_39 = __pyx_v_i;
                                        __pyx_t_50 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_25 = __pyx_v_b;
                                        __pyx_t_52 = __pyx_v_leaf_numb;
                                        __pyx_t_49 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_45 = 0;
                                        __pyx_t_46 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_39 * __pyx_v_X.strides[0]) ) + __pyx_t_50 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_17 = __pyx_t_46;
                                        __pyx_L116_bool_binop_done:;
                                        if (__pyx_t_17) {

                                          /* "cyext_acv.pyx":1986
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":1987
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L113_continue;

                                          /* "cyext_acv.pyx":1985
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L113_continue:;
                                      }

                                      /* "cyext_acv.pyx":1989
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 csi = 1
 * 
 */
                                      __pyx_t_17 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_17) {

                                        /* "cyext_acv.pyx":1990
 * 
 *                             if nv_bool == 0:
 *                                 csi = 1             # <<<<<<<<<<<<<<
 * 
 *                         coef = 0
 */
                                        __pyx_v_csi = 1;

                                        /* "cyext_acv.pyx":1989
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 csi = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":1981
 *                                 o_all = o_all + 1
 * 
 *                         if o_all == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             cs = 1
 *                             nv_bool = 0
 */
                                    }

                                    /* "cyext_acv.pyx":1992
 *                                 csi = 1
 * 
 *                         coef = 0             # <<<<<<<<<<<<<<
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 */
                                    __pyx_v_coef = 0.0;

                                    /* "cyext_acv.pyx":1993
 * 
 *                         coef = 0
 *                         for l in range(1, m - set_size):             # <<<<<<<<<<<<<<
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 */
                                    __pyx_t_61 = (__pyx_v_m - __pyx_v_set_size);
                                    __pyx_t_62 = __pyx_t_61;
                                    for (__pyx_t_60 = 1; __pyx_t_60 < __pyx_t_62; __pyx_t_60+=1) {
                                      __pyx_v_l = __pyx_t_60;

                                      /* "cyext_acv.pyx":1994
 *                         coef = 0
 *                         for l in range(1, m - set_size):
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 */
                                      if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0) != 0.0) != 0)) {
                                        __pyx_t_77 = (__pyx_v_coef + ((1. * __pyx_f_9cyext_acv_binomialC(((__pyx_v_m - __pyx_v_set_size) - 1), __pyx_v_l, 0)) / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0)));
                                      } else {
                                        __pyx_t_77 = 0.0;
                                      }
                                      __pyx_v_coef = __pyx_t_77;
                                    }

                                    /* "cyext_acv.pyx":1996
 *                             coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 *                         coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                    if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0) != 0.0) != 0)) {
                                      __pyx_t_77 = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));
                                    } else {
                                      __pyx_t_77 = 0.0;
                                    }
                                    __pyx_v_coef_0 = __pyx_t_77;

                                    /* "cyext_acv.pyx":1999
 * 
 * 
 *                         p_s = (cs * lm)/lm_s[va, leaf_numb, counter]             # <<<<<<<<<<<<<<
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 */
                                    __pyx_t_52 = __pyx_v_va;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_39 = __pyx_v_counter;
                                    __pyx_v_p_s = ((__pyx_v_cs * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_52 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_25 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_39 * __pyx_v_lm_s.strides[2]) ))));

                                    /* "cyext_acv.pyx":2000
 * 
 *                         p_s = (cs * lm)/lm_s[va, leaf_numb, counter]
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter]             # <<<<<<<<<<<<<<
 * 
 *                         if S_size != 0:
 */
                                    __pyx_t_39 = __pyx_v_va;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_52 = __pyx_v_counter;
                                    __pyx_v_p_si = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_39 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_25 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_52 * __pyx_v_lm_si.strides[2]) ))));

                                    /* "cyext_acv.pyx":2002
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 *                         if S_size != 0:             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):
 */
                                    __pyx_t_17 = ((__pyx_v_S_size != 0) != 0);
                                    if (__pyx_t_17) {

                                      /* "cyext_acv.pyx":2003
 * 
 *                         if S_size != 0:
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 for i2 in range(d):
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                      __pyx_t_20 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_21 = __pyx_t_20;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_21; __pyx_t_60+=1) {
                                        __pyx_v_nv = __pyx_t_60;

                                        /* "cyext_acv.pyx":2004
 *                         if S_size != 0:
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                        __pyx_t_75 = __pyx_v_d;
                                        __pyx_t_76 = __pyx_t_75;
                                        for (__pyx_t_78 = 0; __pyx_t_78 < __pyx_t_76; __pyx_t_78+=1) {
                                          __pyx_v_i2 = __pyx_t_78;

                                          /* "cyext_acv.pyx":2005
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *                         else:
 */
                                          __pyx_t_52 = __pyx_v_b;
                                          __pyx_t_25 = __pyx_v_leaf_numb;
                                          __pyx_t_39 = __pyx_v_b;
                                          __pyx_t_45 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_52 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                          __pyx_t_54 = __pyx_v_i2;
                                          __pyx_t_44 = __pyx_v_leaf_numb;
                                          __pyx_t_38 = __pyx_v_counter;
                                          __pyx_t_51 = __pyx_v_i;
                                          __pyx_t_49 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_40 = __pyx_v_i2;
                                          *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_44 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_51 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_49 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_40 * __pyx_v_phi_b.strides[4]) )) += (((__pyx_v_coef_0 + __pyx_v_coef) * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_39 * __pyx_v_values.strides[0]) ) + __pyx_t_45 * __pyx_v_values.strides[1]) ) + __pyx_t_54 * __pyx_v_values.strides[2]) ))));
                                        }
                                      }

                                      /* "cyext_acv.pyx":2002
 *                         p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 *                         if S_size != 0:             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):
 */
                                      goto __pyx_L121;
                                    }

                                    /* "cyext_acv.pyx":2008
 * 
 *                         else:
 *                             p_off = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):
 */
                                    /*else*/ {
                                      __pyx_v_p_off = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                      /* "cyext_acv.pyx":2009
 *                         else:
 *                             p_off = lm/data.shape[0]
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 for i2 in range(d):
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                      __pyx_t_20 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_21 = __pyx_t_20;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_21; __pyx_t_60+=1) {
                                        __pyx_v_nv = __pyx_t_60;

                                        /* "cyext_acv.pyx":2010
 *                             p_off = lm/data.shape[0]
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                        __pyx_t_75 = __pyx_v_d;
                                        __pyx_t_76 = __pyx_t_75;
                                        for (__pyx_t_78 = 0; __pyx_t_78 < __pyx_t_76; __pyx_t_78+=1) {
                                          __pyx_v_i2 = __pyx_t_78;

                                          /* "cyext_acv.pyx":2011
 *                             for nv in range(va_id[va].size()):
 *                                 for i2 in range(d):
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
                                          __pyx_t_25 = __pyx_v_b;
                                          __pyx_t_52 = __pyx_v_leaf_numb;
                                          __pyx_t_54 = __pyx_v_b;
                                          __pyx_t_45 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_25 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_52 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                          __pyx_t_39 = __pyx_v_i2;
                                          __pyx_t_40 = __pyx_v_b;
                                          __pyx_t_51 = __pyx_v_leaf_numb;
                                          __pyx_t_38 = __pyx_v_b;
                                          __pyx_t_49 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_40 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                          __pyx_t_44 = __pyx_v_i2;
                                          __pyx_t_55 = __pyx_v_leaf_numb;
                                          __pyx_t_56 = __pyx_v_counter;
                                          __pyx_t_63 = __pyx_v_i;
                                          __pyx_t_50 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                          __pyx_t_65 = __pyx_v_i2;
                                          *((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_55 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_56 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_63 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_50 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_65 * __pyx_v_phi_b.strides[4]) )) += (((__pyx_v_p_si - __pyx_v_p_off) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_54 * __pyx_v_values.strides[0]) ) + __pyx_t_45 * __pyx_v_values.strides[1]) ) + __pyx_t_39 * __pyx_v_values.strides[2]) )))) + ((__pyx_v_coef * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_38 * __pyx_v_values.strides[0]) ) + __pyx_t_49 * __pyx_v_values.strides[1]) ) + __pyx_t_44 * __pyx_v_values.strides[2]) )))));
                                        }
                                      }
                                    }
                                    __pyx_L121:;
                                  }
                                }
                                __pyx_L67_continue:;
                              }
                              goto __pyx_L131;
                              __pyx_L39_error:;
                              {
                                  #ifdef WITH_THREAD
                                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                  #endif
                                  #ifdef _OPENMP
                                  #pragma omp flush(__pyx_parallel_exc_type)
                                  #endif /* _OPENMP */
                                  if (!__pyx_parallel_exc_type) {
                                    __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                    __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                    __Pyx_GOTREF(__pyx_parallel_exc_type);
                                  }
                                  #ifdef WITH_THREAD
                                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                  #endif
                              }
                              __pyx_parallel_why = 4;
                              goto __pyx_L130;
                              __pyx_L130:;
                              #ifdef _OPENMP
                              #pragma omp critical(__pyx_parallel_lastprivates2)
                              #endif /* _OPENMP */
                              {
                                  __pyx_parallel_temp0 = __pyx_v_S_size;
                                  __pyx_parallel_temp1 = __pyx_v_a_it;
                                  __pyx_parallel_temp2 = __pyx_v_add;
                                  __pyx_parallel_temp3 = __pyx_v_b_it;
                                  __pyx_parallel_temp4 = __pyx_v_ci;
                                  __pyx_parallel_temp5 = __pyx_v_cj;
                                  __pyx_parallel_temp6 = __pyx_v_coef;
                                  __pyx_parallel_temp7 = __pyx_v_coef_0;
                                  __pyx_parallel_temp8 = __pyx_v_counter;
                                  __pyx_parallel_temp9 = __pyx_v_cs;
                                  __pyx_parallel_temp10 = __pyx_v_csi;
                                  __pyx_parallel_temp11 = __pyx_v_i;
                                  __pyx_parallel_temp12 = __pyx_v_i2;
                                  __pyx_parallel_temp13 = __pyx_v_l;
                                  __pyx_parallel_temp14 = __pyx_v_leaf_numb;
                                  __pyx_parallel_temp15 = __pyx_v_lm;
                                  __pyx_parallel_temp16 = __pyx_v_ns;
                                  __pyx_parallel_temp17 = __pyx_v_nv;
                                  __pyx_parallel_temp18 = __pyx_v_nv_bool;
                                  __pyx_parallel_temp19 = __pyx_v_o_all;
                                  __pyx_parallel_temp20 = __pyx_v_p_off;
                                  __pyx_parallel_temp21 = __pyx_v_p_s;
                                  __pyx_parallel_temp22 = __pyx_v_p_si;
                                  __pyx_parallel_temp23 = __pyx_v_pow_set_size;
                                  __pyx_parallel_temp24 = __pyx_v_s;
                                  __pyx_parallel_temp25 = __pyx_v_set_size;
                                  __pyx_parallel_temp26 = __pyx_v_va;
                                  __pyx_parallel_temp27 = __pyx_v_va_size;
                              }
                              __pyx_L131:;
                              #ifdef _OPENMP
                              #pragma omp flush(__pyx_parallel_why)
                              #endif /* _OPENMP */
                          }
                      }
                      #ifdef _OPENMP
                      Py_END_ALLOW_THREADS
                      #else
{
#ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      #endif /* _OPENMP */
                      /* Clean up any temporaries */
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      #ifndef _OPENMP
}
#endif /* _OPENMP */
                  }
              }
              if (__pyx_parallel_exc_type) {
                /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
                __pyx_parallel_why = 4;
              }
              if (__pyx_parallel_why) {
                __pyx_v_S_size = __pyx_parallel_temp0;
                __pyx_v_a_it = __pyx_parallel_temp1;
                __pyx_v_add = __pyx_parallel_temp2;
                __pyx_v_b_it = __pyx_parallel_temp3;
                __pyx_v_ci = __pyx_parallel_temp4;
                __pyx_v_cj = __pyx_parallel_temp5;
                __pyx_v_coef = __pyx_parallel_temp6;
                __pyx_v_coef_0 = __pyx_parallel_temp7;
                __pyx_v_counter = __pyx_parallel_temp8;
                __pyx_v_cs = __pyx_parallel_temp9;
                __pyx_v_csi = __pyx_parallel_temp10;
                __pyx_v_i = __pyx_parallel_temp11;
                __pyx_v_i2 = __pyx_parallel_temp12;
                __pyx_v_l = __pyx_parallel_temp13;
                __pyx_v_leaf_numb = __pyx_parallel_temp14;
                __pyx_v_lm = __pyx_parallel_temp15;
                __pyx_v_ns = __pyx_parallel_temp16;
                __pyx_v_nv = __pyx_parallel_temp17;
                __pyx_v_nv_bool = __pyx_parallel_temp18;
                __pyx_v_o_all = __pyx_parallel_temp19;
                __pyx_v_p_off = __pyx_parallel_temp20;
                __pyx_v_p_s = __pyx_parallel_temp21;
                __pyx_v_p_si = __pyx_parallel_temp22;
                __pyx_v_pow_set_size = __pyx_parallel_temp23;
                __pyx_v_s = __pyx_parallel_temp24;
                __pyx_v_set_size = __pyx_parallel_temp25;
                __pyx_v_va = __pyx_parallel_temp26;
                __pyx_v_va_size = __pyx_parallel_temp27;
                switch (__pyx_parallel_why) {
                      case 4:
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      __Pyx_GIVEREF(__pyx_parallel_exc_type);
                      __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                      __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                  }
                  goto __pyx_L35_error;
                }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":1852
 *                             phi_b[leaf_numb, counter, i, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L36;
          }
          __pyx_L35_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L36:;
        }
    }

    /* "cyext_acv.pyx":2013
 *                                     phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):
 */
    __pyx_t_30 = __pyx_v_N;
    __pyx_t_29 = __pyx_t_30;
    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_29; __pyx_t_28+=1) {
      __pyx_v_i = __pyx_t_28;

      /* "cyext_acv.pyx":2014
 * 
 *         for i in range(N):
 *             for j in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 */
      __pyx_t_23 = (__pyx_v_data.shape[1]);
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_24; __pyx_t_31+=1) {
        __pyx_v_j = __pyx_t_31;

        /* "cyext_acv.pyx":2015
 *         for i in range(N):
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi[i, j, i2] += phi_n[leaf_numb, i, j, i2]
 */
        __pyx_t_32 = __pyx_v_d;
        __pyx_t_33 = __pyx_t_32;
        for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
          __pyx_v_i2 = __pyx_t_34;

          /* "cyext_acv.pyx":2016
 *             for j in range(data.shape[1]):
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                         phi[i, j, i2] += phi_n[leaf_numb, i, j, i2]
 *                         for counter in range(phi_b.shape[1]):
 */
          __pyx_t_35 = (__pyx_v_phi_b.shape[0]);
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_leaf_numb = __pyx_t_37;

            /* "cyext_acv.pyx":2017
 *                 for i2 in range(d):
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi[i, j, i2] += phi_n[leaf_numb, i, j, i2]             # <<<<<<<<<<<<<<
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 */
            __pyx_t_51 = __pyx_v_leaf_numb;
            __pyx_t_40 = __pyx_v_i;
            __pyx_t_44 = __pyx_v_j;
            __pyx_t_38 = __pyx_v_i2;
            __pyx_t_52 = __pyx_v_i;
            __pyx_t_25 = __pyx_v_j;
            __pyx_t_39 = __pyx_v_i2;
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_52 * __pyx_v_phi.strides[0]) ) + __pyx_t_25 * __pyx_v_phi.strides[1]) ) + __pyx_t_39 * __pyx_v_phi.strides[2]) )) += (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_51 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_40 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_44 * __pyx_v_phi_n.strides[2]) ) + __pyx_t_38 * __pyx_v_phi_n.strides[3]) )));

            /* "cyext_acv.pyx":2018
 *                     for leaf_numb in range(phi_b.shape[0]):
 *                         phi[i, j, i2] += phi_n[leaf_numb, i, j, i2]
 *                         for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 *     return np.asarray(phi)/m
 */
            __pyx_t_41 = (__pyx_v_phi_b.shape[1]);
            __pyx_t_42 = __pyx_t_41;
            for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_42; __pyx_t_43+=1) {
              __pyx_v_counter = __pyx_t_43;

              /* "cyext_acv.pyx":2019
 *                         phi[i, j, i2] += phi_n[leaf_numb, i, j, i2]
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]             # <<<<<<<<<<<<<<
 *     return np.asarray(phi)/m
 * 
 */
              __pyx_t_38 = __pyx_v_leaf_numb;
              __pyx_t_44 = __pyx_v_counter;
              __pyx_t_40 = __pyx_v_i;
              __pyx_t_51 = __pyx_v_j;
              __pyx_t_39 = __pyx_v_i2;
              __pyx_t_25 = __pyx_v_i;
              __pyx_t_52 = __pyx_v_j;
              __pyx_t_54 = __pyx_v_i2;
              *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_25 * __pyx_v_phi.strides[0]) ) + __pyx_t_52 * __pyx_v_phi.strides[1]) ) + __pyx_t_54 * __pyx_v_phi.strides[2]) )) += (*((double *) ( /* dim=4 */ (( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_38 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_44 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_40 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_51 * __pyx_v_phi_b.strides[3]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[4]) )));
            }
          }
        }
      }
    }
  }

  /* "cyext_acv.pyx":2020
 *                         for counter in range(phi_b.shape[1]):
 *                             phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]
 *     return np.asarray(phi)/m             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_phi, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":1755
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_acv_leaves(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_27, 1);
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_va_c, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_N_star_a, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_si, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_25shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_25shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  __Pyx_memviewslice __pyx_v_S_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_N_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shap_values_acv_leaves (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_S_star,&__pyx_n_s_N_star,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 1); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 2); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 3); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 4); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 5); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 6); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 7); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 8); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_N_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 9); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 10); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, 11); __PYX_ERR(0, 1755, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "shap_values_acv_leaves") < 0)) __PYX_ERR(0, 1755, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 1756, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 1757, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 1758, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 1759, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 1760, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_scaling == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1761, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1762, __pyx_L3_error)
    __pyx_v_S_star = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[8], 0); if (unlikely(!__pyx_v_S_star.memview)) __PYX_ERR(0, 1762, __pyx_L3_error)
    __pyx_v_N_star = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[9], 0); if (unlikely(!__pyx_v_N_star.memview)) __PYX_ERR(0, 1762, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[10]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1763, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1763, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1755, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_24shap_values_acv_leaves(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_24shap_values_acv_leaves(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_acv_leaves", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S_star.memview)) { __Pyx_RaiseUnboundLocalError("S_star"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  if (unlikely(!__pyx_v_N_star.memview)) { __Pyx_RaiseUnboundLocalError("N_star"); __PYX_ERR(0, 1755, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_shap_values_acv_leaves(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_N_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2026
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef compute_sdp_swing(const double[:, :] X, const long[:] fX,             # <<<<<<<<<<<<<<
 *             const long[:] y_pred, long[::1] S, unsigned long S_size, const double[:, :] data,
 *             const double[:, :, :] values,const  double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_f_9cyext_acv_compute_sdp_swing(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, unsigned long __pyx_v_S_size, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, double const __pyx_v_thresholds, CYTHON_UNUSED int __pyx_v_num_threads) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_out = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  unsigned int __pyx_v_o_all;
  double __pyx_v_ss;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_j;
  int __pyx_v_lm;
  int __pyx_v_lm_s;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  int __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  unsigned long __pyx_t_18;
  unsigned long __pyx_t_19;
  int __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  int __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  int __pyx_t_33;
  double __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_swing", 0);

  /* "cyext_acv.pyx":2032
 *             const double thresholds, int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef double[:] out
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2033
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef double[:] out
 * 
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2036
 *     cdef double[:] out
 * 
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         out = np.ones(shape=(N))
 *         return out
 */
  __pyx_t_1 = ((__pyx_v_S_size == __pyx_v_m) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":2037
 * 
 *     if S_size == m:
 *         out = np.ones(shape=(N))             # <<<<<<<<<<<<<<
 *         return out
 *     elif S_size == 0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_4) < 0) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_out = __pyx_t_5;
    __pyx_t_5.memview = NULL;
    __pyx_t_5.data = NULL;

    /* "cyext_acv.pyx":2038
 *     if S_size == m:
 *         out = np.ones(shape=(N))
 *         return out             # <<<<<<<<<<<<<<
 *     elif S_size == 0:
 *         out = np.zeros(shape=(N))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_out, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":2036
 *     cdef double[:] out
 * 
 *     if S_size == m:             # <<<<<<<<<<<<<<
 *         out = np.ones(shape=(N))
 *         return out
 */
  }

  /* "cyext_acv.pyx":2039
 *         out = np.ones(shape=(N))
 *         return out
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         out = np.zeros(shape=(N))
 *         return out
 */
  __pyx_t_1 = ((__pyx_v_S_size == 0) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":2040
 *         return out
 *     elif S_size == 0:
 *         out = np.zeros(shape=(N))             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2040, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_out = __pyx_t_5;
    __pyx_t_5.memview = NULL;
    __pyx_t_5.data = NULL;

    /* "cyext_acv.pyx":2041
 *     elif S_size == 0:
 *         out = np.zeros(shape=(N))
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_trees = values.shape[0]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_out, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":2039
 *         out = np.ones(shape=(N))
 *         return out
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         out = np.zeros(shape=(N))
 *         return out
 */
  }

  /* "cyext_acv.pyx":2043
 *         return out
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":2049
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_3);
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_mean_forest = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2052
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":2059
 *     cdef int b, leaf_numb, i, s, j, lm, lm_s
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 */
  __pyx_t_8 = __pyx_v_n_trees;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_b = __pyx_t_10;

    /* "cyext_acv.pyx":2060
 * 
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):             # <<<<<<<<<<<<<<
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 o_all = 0
 */
    __pyx_t_11 = __pyx_v_b;
    __pyx_t_12 = (*((long const  *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_11 * __pyx_v_leaves_nb.strides[0]) )));
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
      __pyx_v_leaf_numb = __pyx_t_14;

      /* "cyext_acv.pyx":2061
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {
            __pyx_t_15 = __pyx_v_N;
            if ((1 == 0)) abort();
            {
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   (x)
                    #define unlikely(x) (x)
                #endif
                __pyx_t_17 = (__pyx_t_15 - 0 + 1 - 1/abs(1)) / 1;
                if (__pyx_t_17 > 0)
                {
                    #ifdef _OPENMP
                    #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) num_threads(__pyx_v_num_threads) private(__pyx_t_1, __pyx_t_11, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34)
                    #endif /* _OPENMP */
                    {
                        #ifdef _OPENMP
                        #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                        #endif /* _OPENMP */
                        for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_17; __pyx_t_16++){
                            {
                                __pyx_v_i = (int)(0 + 1 * __pyx_t_16);
                                /* Initialize private variables to invalid values */
                                __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_j = ((int)0xbad0bad0);
                                __pyx_v_lm = ((int)0xbad0bad0);
                                __pyx_v_lm_s = ((int)0xbad0bad0);
                                __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                __pyx_v_s = ((int)0xbad0bad0);

                                /* "cyext_acv.pyx":2062
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":2063
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_18 = __pyx_v_S_size;
                                __pyx_t_19 = __pyx_t_18;
                                for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
                                  __pyx_v_s = __pyx_t_20;

                                  /* "cyext_acv.pyx":2064
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_11 = __pyx_v_s;
                                  __pyx_t_21 = __pyx_v_i;
                                  __pyx_t_22 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_11)) )));
                                  __pyx_t_23 = __pyx_v_s;
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_25 = __pyx_v_leaf_numb;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_23)) )));
                                  __pyx_t_27 = 1;
                                  __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_21 * __pyx_v_X.strides[0]) ) + __pyx_t_22 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_28) {
                                  } else {
                                    __pyx_t_1 = __pyx_t_28;
                                    goto __pyx_L20_bool_binop_done;
                                  }
                                  __pyx_t_23 = __pyx_v_s;
                                  __pyx_t_27 = __pyx_v_i;
                                  __pyx_t_26 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_23)) )));
                                  __pyx_t_25 = __pyx_v_s;
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_11 = __pyx_v_leaf_numb;
                                  __pyx_t_22 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_25)) )));
                                  __pyx_t_21 = 0;
                                  __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_27 * __pyx_v_X.strides[0]) ) + __pyx_t_26 * __pyx_v_X.strides[1]) ))) < (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_11 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_21 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_1 = __pyx_t_28;
                                  __pyx_L20_bool_binop_done:;
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":2065
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":2064
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":2066
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_1 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_1) {

                                  /* "cyext_acv.pyx":2067
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 p = 0
 */
                                  goto __pyx_L13_continue;

                                  /* "cyext_acv.pyx":2066
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":2069
 *                     continue
 * 
 *                 p = 0             # <<<<<<<<<<<<<<
 *                 p_u = 0
 *                 p_d = 0
 */
                                __pyx_v_p = 0;

                                /* "cyext_acv.pyx":2070
 * 
 *                 p = 0
 *                 p_u = 0             # <<<<<<<<<<<<<<
 *                 p_d = 0
 *                 p_s = 0
 */
                                __pyx_v_p_u = 0;

                                /* "cyext_acv.pyx":2071
 *                 p = 0
 *                 p_u = 0
 *                 p_d = 0             # <<<<<<<<<<<<<<
 *                 p_s = 0
 *                 p_su = 0
 */
                                __pyx_v_p_d = 0;

                                /* "cyext_acv.pyx":2072
 *                 p_u = 0
 *                 p_d = 0
 *                 p_s = 0             # <<<<<<<<<<<<<<
 *                 p_su = 0
 *                 p_sd = 0
 */
                                __pyx_v_p_s = 0;

                                /* "cyext_acv.pyx":2073
 *                 p_d = 0
 *                 p_s = 0
 *                 p_su = 0             # <<<<<<<<<<<<<<
 *                 p_sd = 0
 * 
 */
                                __pyx_v_p_su = 0;

                                /* "cyext_acv.pyx":2074
 *                 p_s = 0
 *                 p_su = 0
 *                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(data.shape[0]):
 */
                                __pyx_v_p_sd = 0;

                                /* "cyext_acv.pyx":2076
 *                 p_sd = 0
 * 
 *                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                     a_it = 0
 *                     b_it = 0
 */
                                __pyx_t_29 = (__pyx_v_data.shape[0]);
                                __pyx_t_30 = __pyx_t_29;
                                for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_30; __pyx_t_20+=1) {
                                  __pyx_v_j = __pyx_t_20;

                                  /* "cyext_acv.pyx":2077
 * 
 *                 for j in range(data.shape[0]):
 *                     a_it = 0             # <<<<<<<<<<<<<<
 *                     b_it = 0
 *                     lm = 0
 */
                                  __pyx_v_a_it = 0;

                                  /* "cyext_acv.pyx":2078
 *                 for j in range(data.shape[0]):
 *                     a_it = 0
 *                     b_it = 0             # <<<<<<<<<<<<<<
 *                     lm = 0
 *                     lm_s = 0
 */
                                  __pyx_v_b_it = 0;

                                  /* "cyext_acv.pyx":2079
 *                     a_it = 0
 *                     b_it = 0
 *                     lm = 0             # <<<<<<<<<<<<<<
 *                     lm_s = 0
 *                     for s in range(m):
 */
                                  __pyx_v_lm = 0;

                                  /* "cyext_acv.pyx":2080
 *                     b_it = 0
 *                     lm = 0
 *                     lm_s = 0             # <<<<<<<<<<<<<<
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                  __pyx_v_lm_s = 0;

                                  /* "cyext_acv.pyx":2081
 *                     lm = 0
 *                     lm_s = 0
 *                     for s in range(m):             # <<<<<<<<<<<<<<
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 */
                                  __pyx_t_31 = __pyx_v_m;
                                  __pyx_t_32 = __pyx_t_31;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":2082
 *                     lm_s = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    __pyx_t_25 = __pyx_v_j;
                                    __pyx_t_21 = __pyx_v_s;
                                    __pyx_t_22 = __pyx_v_b;
                                    __pyx_t_11 = __pyx_v_leaf_numb;
                                    __pyx_t_24 = __pyx_v_s;
                                    __pyx_t_23 = 1;
                                    __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_25 * __pyx_v_data.strides[0]) ) + __pyx_t_21 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_11 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_28) {
                                    } else {
                                      __pyx_t_1 = __pyx_t_28;
                                      goto __pyx_L28_bool_binop_done;
                                    }
                                    __pyx_t_23 = __pyx_v_j;
                                    __pyx_t_24 = __pyx_v_s;
                                    __pyx_t_11 = __pyx_v_b;
                                    __pyx_t_22 = __pyx_v_leaf_numb;
                                    __pyx_t_21 = __pyx_v_s;
                                    __pyx_t_25 = 0;
                                    __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_23 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) >= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_11 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_21 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_1 = __pyx_t_28;
                                    __pyx_L28_bool_binop_done:;
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":2083
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                      __pyx_v_a_it = (__pyx_v_a_it + 1);

                                      /* "cyext_acv.pyx":2082
 *                     lm_s = 0
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":2084
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1
 */
                                  __pyx_t_18 = __pyx_v_S_size;
                                  __pyx_t_19 = __pyx_t_18;
                                  for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_19; __pyx_t_33+=1) {
                                    __pyx_v_s = __pyx_t_33;

                                    /* "cyext_acv.pyx":2085
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    __pyx_t_25 = __pyx_v_s;
                                    __pyx_t_21 = __pyx_v_j;
                                    __pyx_t_22 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_25)) )));
                                    __pyx_t_11 = __pyx_v_s;
                                    __pyx_t_24 = __pyx_v_b;
                                    __pyx_t_23 = __pyx_v_leaf_numb;
                                    __pyx_t_26 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_11)) )));
                                    __pyx_t_27 = 1;
                                    __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_21 * __pyx_v_data.strides[0]) ) + __pyx_t_22 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_23 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_28) {
                                    } else {
                                      __pyx_t_1 = __pyx_t_28;
                                      goto __pyx_L33_bool_binop_done;
                                    }
                                    __pyx_t_11 = __pyx_v_s;
                                    __pyx_t_27 = __pyx_v_j;
                                    __pyx_t_26 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_11)) )));
                                    __pyx_t_23 = __pyx_v_s;
                                    __pyx_t_24 = __pyx_v_b;
                                    __pyx_t_25 = __pyx_v_leaf_numb;
                                    __pyx_t_22 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_23)) )));
                                    __pyx_t_21 = 0;
                                    __pyx_t_28 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_27 * __pyx_v_data.strides[0]) ) + __pyx_t_26 * __pyx_v_data.strides[1]) ))) >= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_22 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_21 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_1 = __pyx_t_28;
                                    __pyx_L33_bool_binop_done:;
                                    if (__pyx_t_1) {

                                      /* "cyext_acv.pyx":2086
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                     if a_it == m:
 */
                                      __pyx_v_b_it = (__pyx_v_b_it + 1);

                                      /* "cyext_acv.pyx":2085
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":2088
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  __pyx_t_1 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":2089
 * 
 *                     if a_it == m:
 *                         lm = 1             # <<<<<<<<<<<<<<
 * 
 *                     if b_it == S_size:
 */
                                    __pyx_v_lm = 1;

                                    /* "cyext_acv.pyx":2088
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":2091
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  __pyx_t_1 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":2092
 * 
 *                     if b_it == S_size:
 *                         lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                     p += lm
 */
                                    __pyx_v_lm_s = 1;

                                    /* "cyext_acv.pyx":2091
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":2094
 *                         lm_s = 1
 * 
 *                     p += lm             # <<<<<<<<<<<<<<
 *                     p_s += lm_s
 *                     if fX[i] == y_pred[j]:
 */
                                  __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                  /* "cyext_acv.pyx":2095
 * 
 *                     p += lm
 *                     p_s += lm_s             # <<<<<<<<<<<<<<
 *                     if fX[i] == y_pred[j]:
 *                         p_u += lm
 */
                                  __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                  /* "cyext_acv.pyx":2096
 *                     p += lm
 *                     p_s += lm_s
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u += lm
 *                         p_su += lm_s
 */
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_21 = __pyx_v_j;
                                  __pyx_t_1 = (((*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_23 * __pyx_v_fX.strides[0]) ))) == (*((long const  *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_21 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                  if (__pyx_t_1) {

                                    /* "cyext_acv.pyx":2097
 *                     p_s += lm_s
 *                     if fX[i] == y_pred[j]:
 *                         p_u += lm             # <<<<<<<<<<<<<<
 *                         p_su += lm_s
 *                     else:
 */
                                    __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                    /* "cyext_acv.pyx":2098
 *                     if fX[i] == y_pred[j]:
 *                         p_u += lm
 *                         p_su += lm_s             # <<<<<<<<<<<<<<
 *                     else:
 *                         p_d += lm
 */
                                    __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                    /* "cyext_acv.pyx":2096
 *                     p += lm
 *                     p_s += lm_s
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u += lm
 *                         p_su += lm_s
 */
                                    goto __pyx_L37;
                                  }

                                  /* "cyext_acv.pyx":2100
 *                         p_su += lm_s
 *                     else:
 *                         p_d += lm             # <<<<<<<<<<<<<<
 *                         p_sd += lm_s
 * 
 */
                                  /*else*/ {
                                    __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                    /* "cyext_acv.pyx":2101
 *                     else:
 *                         p_d += lm
 *                         p_sd += lm_s             # <<<<<<<<<<<<<<
 * 
 *                 mean_forest[i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 */
                                    __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);
                                  }
                                  __pyx_L37:;
                                }

                                /* "cyext_acv.pyx":2103
 *                         p_sd += lm_s
 * 
 *                 mean_forest[i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest[i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest[i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_21 = __pyx_v_b;
                                  __pyx_t_23 = __pyx_v_leaf_numb;
                                  __pyx_t_22 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_24 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_21 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_23 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_11 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_22 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_25 * __pyx_v_values.strides[0]) ) + __pyx_t_24 * __pyx_v_values.strides[1]) ) + __pyx_t_11 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_22 = __pyx_v_i;
                                __pyx_t_23 = 0;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_22 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_23 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":2104
 * 
 *                 mean_forest[i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest[i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest[i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_23 = __pyx_v_b;
                                  __pyx_t_22 = __pyx_v_leaf_numb;
                                  __pyx_t_21 = __pyx_v_i;
                                  __pyx_t_11 = __pyx_v_b;
                                  __pyx_t_24 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_23 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_22 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_25 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_21 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p_u * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_11 * __pyx_v_values.strides[0]) ) + __pyx_t_24 * __pyx_v_values.strides[1]) ) + __pyx_t_25 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_21 = __pyx_v_i;
                                __pyx_t_22 = 1;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_21 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_22 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_34;

                                /* "cyext_acv.pyx":2105
 *                 mean_forest[i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest[i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest[i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *     for i in prange(N, nogil=True, num_threads=num_threads):
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_22 = __pyx_v_b;
                                  __pyx_t_21 = __pyx_v_leaf_numb;
                                  __pyx_t_23 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_b;
                                  __pyx_t_24 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_22 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_21 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_11 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_23 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_34 = ((__pyx_v_p_d * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_25 * __pyx_v_values.strides[0]) ) + __pyx_t_24 * __pyx_v_values.strides[1]) ) + __pyx_t_11 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_34 = 0.0;
                                }
                                __pyx_t_23 = __pyx_v_i;
                                __pyx_t_21 = 2;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_23 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_21 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_34;
                                goto __pyx_L39;
                                __pyx_L13_continue:;
                                goto __pyx_L39;
                                __pyx_L39:;
                            }
                        }
                    }
                }
            }
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   __builtin_expect(!!(x), 1)
                #define unlikely(x) __builtin_expect(!!(x), 0)
            #endif
          }

          /* "cyext_acv.pyx":2061
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L12;
            }
            __pyx_L12:;
          }
      }
    }
  }

  /* "cyext_acv.pyx":2107
 *                 mean_forest[i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 *     for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 *         if ss >= thresholds:
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_17 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_15 = (__pyx_t_17 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_15 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_1, __pyx_t_11, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_ss)
                    #endif /* _OPENMP */
                    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_16);
                            /* Initialize private variables to invalid values */
                            __pyx_v_ss = ((double)__PYX_NAN());

                            /* "cyext_acv.pyx":2108
 * 
 *     for i in prange(N, nogil=True, num_threads=num_threads):
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0             # <<<<<<<<<<<<<<
 *         if ss >= thresholds:
 *             sdp[i] = 1
 */
                            __pyx_t_21 = __pyx_v_i;
                            __pyx_t_23 = 1;
                            __pyx_t_22 = __pyx_v_i;
                            __pyx_t_11 = 2;
                            if (((((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_21 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_23 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_22 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_11 * __pyx_v_mean_forest.strides[1]) )))) != 0.0) != 0)) {
                              __pyx_t_24 = __pyx_v_i;
                              __pyx_t_25 = 0;
                              __pyx_t_26 = __pyx_v_i;
                              __pyx_t_27 = 2;
                              __pyx_t_35 = __pyx_v_i;
                              __pyx_t_36 = 1;
                              __pyx_t_37 = __pyx_v_i;
                              __pyx_t_38 = 2;
                              __pyx_t_34 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_24 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_25 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_26 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) )))) / ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_35 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_36 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_37 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_38 * __pyx_v_mean_forest.strides[1]) )))));
                            } else {
                              __pyx_t_34 = 0.0;
                            }
                            __pyx_v_ss = __pyx_t_34;

                            /* "cyext_acv.pyx":2109
 *     for i in prange(N, nogil=True, num_threads=num_threads):
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 *         if ss >= thresholds:             # <<<<<<<<<<<<<<
 *             sdp[i] = 1
 *         else:
 */
                            __pyx_t_1 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
                            if (__pyx_t_1) {

                              /* "cyext_acv.pyx":2110
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 *         if ss >= thresholds:
 *             sdp[i] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             sdp[i] = 0
 */
                              __pyx_t_11 = __pyx_v_i;
                              *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_11 * __pyx_v_sdp.strides[0]) )) = 1.0;

                              /* "cyext_acv.pyx":2109
 *     for i in prange(N, nogil=True, num_threads=num_threads):
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 *         if ss >= thresholds:             # <<<<<<<<<<<<<<
 *             sdp[i] = 1
 *         else:
 */
                              goto __pyx_L47;
                            }

                            /* "cyext_acv.pyx":2112
 *             sdp[i] = 1
 *         else:
 *             sdp[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
                            /*else*/ {
                              __pyx_t_11 = __pyx_v_i;
                              *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_11 * __pyx_v_sdp.strides[0]) )) = 0.0;
                            }
                            __pyx_L47:;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":2107
 *                 mean_forest[i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 *     for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *         ss = (mean_forest[i, 0] - mean_forest[i, 2])/(mean_forest[i, 1] - mean_forest[i, 2]) if mean_forest[i, 1] - mean_forest[i, 2] !=0 else 0
 *         if ss >= thresholds:
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L42;
        }
        __pyx_L42:;
      }
  }

  /* "cyext_acv.pyx":2114
 *             sdp[i] = 0
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2026
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef compute_sdp_swing(const double[:, :] X, const long[:] fX,             # <<<<<<<<<<<<<<
 *             const long[:] y_pred, long[::1] S, unsigned long S_size, const double[:, :] data,
 *             const double[:, :, :] values,const  double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_swing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_out, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2121
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef swing_sv_clf(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const long[:] fX,
 *     const long[:] y_pred,
 */

static PyObject *__pyx_pw_9cyext_acv_27swing_sv_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_swing_sv_clf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double const __pyx_v_thresholds, int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  CYTHON_UNUSED unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_v_plus = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_v_minus = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_dif_pos;
  unsigned int __pyx_v_dif_neg;
  unsigned int __pyx_v_dif_null;
  unsigned int __pyx_v_va_size;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_set_size;
  std::vector<std::vector<int> >  __pyx_v_Sm;
  std::vector<std::vector<int> >  __pyx_v_va_id_cpp;
  double __pyx_v_weight;
  __Pyx_memviewslice __pyx_v_swings_prop = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_swings = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sm_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sm_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C_buff = NULL;
  PyObject *__pyx_v_coal_va = NULL;
  PyObject *__pyx_v_va_id = NULL;
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_c = NULL;
  __Pyx_memviewslice __pyx_v_va_id_a = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_va_id = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  unsigned int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  std::vector<std::vector<int> >  __pyx_t_24;
  size_t __pyx_t_25;
  long __pyx_t_26;
  long __pyx_t_27;
  unsigned int __pyx_t_28;
  size_t __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  unsigned int __pyx_t_35;
  size_t __pyx_t_36;
  size_t __pyx_t_37;
  size_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  PyObject *__pyx_t_41 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("swing_sv_clf", 0);

  /* "cyext_acv.pyx":2132
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":2133
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2134
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :] phi
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2137
 * 
 *     cdef double[:, :] phi
 *     phi = np.zeros((N, m))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] v_plus, v_minus
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 2137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":2140
 * 
 *     cdef double[:] v_plus, v_minus
 *     v_plus = np.zeros((N))             # <<<<<<<<<<<<<<
 *     v_minus = np.zeros((N))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v_plus = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2141
 *     cdef double[:] v_plus, v_minus
 *     v_plus = np.zeros((N))
 *     v_minus = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v_minus = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2144
 * 
 *     cdef long[::1] S
 *     S = np.zeros((m), dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size, dif_pos, dif_neg, dif_null, va_size, pow_set_size
 *     cdef unsigned int counter, i, va, j, ci, cj, set_size
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dc_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_S = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2152
 * 
 *     cdef long[:, :, :] swings_prop
 *     swings_prop = np.zeros((N, m, 3), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] swings,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_swings_prop = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2155
 * 
 *     cdef double[:, :, :] swings,
 *     swings = np.zeros((N, m, 2))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :] sm_buf
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_swings = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":2158
 * 
 *     cdef long[:, :] sm_buf
 *     sm_buf = np.zeros((data.shape[1], data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] sm_size
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 2158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sm_buf = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":2161
 * 
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 *     if C[0] != []:
 *         C_buff = C.copy()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 2161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sm_size = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":2162
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2162, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_13) {

    /* "cyext_acv.pyx":2163
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:
 *         C_buff = C.copy()             # <<<<<<<<<<<<<<
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []
 */
    __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_copy, __pyx_v_C); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_C_buff = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cyext_acv.pyx":2164
 *     if C[0] != []:
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = []
 *         remove_va = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 2164, __pyx_L1_error)
    }
    __pyx_t_14 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2164, __pyx_L1_error)
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_ci = __pyx_t_16;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 2164, __pyx_L1_error)
      }
      __pyx_t_4 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_17 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_cj = __pyx_t_19;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 2164, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 2164, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __pyx_v_coal_va = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "cyext_acv.pyx":2165
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []             # <<<<<<<<<<<<<<
 *         remove_va = []
 * 
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_va_id = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "cyext_acv.pyx":2166
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []
 *         remove_va = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(m):
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_remove_va = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "cyext_acv.pyx":2168
 *         remove_va = []
 * 
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if i not in coal_va:
 *                 remove_va.append([i])
 */
    __pyx_t_16 = __pyx_v_m;
    __pyx_t_19 = __pyx_t_16;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;

      /* "cyext_acv.pyx":2169
 * 
 *         for i in range(m):
 *             if i not in coal_va:             # <<<<<<<<<<<<<<
 *                 remove_va.append([i])
 *                 va_id.append([i])
 */
      __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_coal_va, Py_NE)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_21 = (__pyx_t_13 != 0);
      if (__pyx_t_21) {

        /* "cyext_acv.pyx":2170
 *         for i in range(m):
 *             if i not in coal_va:
 *                 remove_va.append([i])             # <<<<<<<<<<<<<<
 *                 va_id.append([i])
 *             else:
 */
        __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2170, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_remove_va, __pyx_t_4); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2170, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "cyext_acv.pyx":2171
 *             if i not in coal_va:
 *                 remove_va.append([i])
 *                 va_id.append([i])             # <<<<<<<<<<<<<<
 *             else:
 *                 for c in C_buff:
 */
        __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_va_id, __pyx_t_3); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2171, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "cyext_acv.pyx":2169
 * 
 *         for i in range(m):
 *             if i not in coal_va:             # <<<<<<<<<<<<<<
 *                 remove_va.append([i])
 *                 va_id.append([i])
 */
        goto __pyx_L10;
      }

      /* "cyext_acv.pyx":2173
 *                 va_id.append([i])
 *             else:
 *                 for c in C_buff:             # <<<<<<<<<<<<<<
 *                     if i in c:
 *                         va_id.append(c)
 */
      /*else*/ {
        if (likely(PyList_CheckExact(__pyx_v_C_buff)) || PyTuple_CheckExact(__pyx_v_C_buff)) {
          __pyx_t_3 = __pyx_v_C_buff; __Pyx_INCREF(__pyx_t_3); __pyx_t_14 = 0;
          __pyx_t_23 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_C_buff); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2173, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_23 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 2173, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_23)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2173, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2173, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_23(__pyx_t_3);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 2173, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "cyext_acv.pyx":2174
 *             else:
 *                 for c in C_buff:
 *                     if i in c:             # <<<<<<<<<<<<<<
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 */
          __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2174, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_21 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_v_c, Py_EQ)); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 2174, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_13 = (__pyx_t_21 != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":2175
 *                 for c in C_buff:
 *                     if i in c:
 *                         va_id.append(c)             # <<<<<<<<<<<<<<
 *                         C_buff.remove(c)
 *                         continue
 */
            __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_va_id, __pyx_v_c); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2175, __pyx_L1_error)

            /* "cyext_acv.pyx":2176
 *                     if i in c:
 *                         va_id.append(c)
 *                         C_buff.remove(c)             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_C_buff, __pyx_n_s_remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2176, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_c);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2176, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "cyext_acv.pyx":2177
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L11_continue;

            /* "cyext_acv.pyx":2174
 *             else:
 *                 for c in C_buff:
 *                     if i in c:             # <<<<<<<<<<<<<<
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 */
          }

          /* "cyext_acv.pyx":2173
 *                 va_id.append([i])
 *             else:
 *                 for c in C_buff:             # <<<<<<<<<<<<<<
 *                     if i in c:
 *                         va_id.append(c)
 */
          __pyx_L11_continue:;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_L10:;
    }

    /* "cyext_acv.pyx":2162
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 */
    goto __pyx_L3;
  }

  /* "cyext_acv.pyx":2180
 * 
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_16 = __pyx_v_m;
    __pyx_t_19 = __pyx_t_16;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;
      __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2180, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "cyext_acv.pyx":2182
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     va_id_cpp = va_id
 * 
 */
  __pyx_t_14 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2182, __pyx_L1_error)
  __pyx_v_m = __pyx_t_14;

  /* "cyext_acv.pyx":2183
 * 
 *     m = len(va_id)
 *     va_id_cpp = va_id             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :] va_id_a
 */
  __pyx_t_24 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(__pyx_v_va_id); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2183, __pyx_L1_error)
  __pyx_v_va_id_cpp = __pyx_t_24;

  /* "cyext_acv.pyx":2188
 *     cdef long[:] len_va_id
 * 
 *     va_id_a = np.empty((data.shape[1], data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 2188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_va_id_a = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":2189
 * 
 *     va_id_a = np.empty((data.shape[1], data.shape[1]), dtype=np.int)
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(m):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 2189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_len_va_id = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":2191
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)
 * 
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):
 */
  __pyx_t_16 = __pyx_v_m;
  __pyx_t_19 = __pyx_t_16;
  for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
    __pyx_v_i = __pyx_t_20;

    /* "cyext_acv.pyx":2192
 * 
 *     for i in range(m):
 *         len_va_id[i] = va_id_cpp[i].size()             # <<<<<<<<<<<<<<
 *         for j in range(len_va_id[i]):
 *             va_id_a[i, j] = va_id_cpp[i][j]
 */
    __pyx_t_25 = __pyx_v_i;
    *((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )) = (__pyx_v_va_id_cpp[__pyx_v_i]).size();

    /* "cyext_acv.pyx":2193
 *     for i in range(m):
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):             # <<<<<<<<<<<<<<
 *             va_id_a[i, j] = va_id_cpp[i][j]
 * 
 */
    __pyx_t_25 = __pyx_v_i;
    __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )));
    __pyx_t_27 = __pyx_t_26;
    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
      __pyx_v_j = __pyx_t_28;

      /* "cyext_acv.pyx":2194
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):
 *             va_id_a[i, j] = va_id_cpp[i][j]             # <<<<<<<<<<<<<<
 * 
 *     set_size = m - 1
 */
      __pyx_t_25 = __pyx_v_i;
      __pyx_t_29 = __pyx_v_j;
      *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_29 * __pyx_v_va_id_a.strides[1]) )) = ((__pyx_v_va_id_cpp[__pyx_v_i])[__pyx_v_j]);
    }
  }

  /* "cyext_acv.pyx":2196
 *             va_id_a[i, j] = va_id_cpp[i][j]
 * 
 *     set_size = m - 1             # <<<<<<<<<<<<<<
 *     pow_set_size = 2**set_size
 *     for va in range(m):
 */
  __pyx_v_set_size = (__pyx_v_m - 1);

  /* "cyext_acv.pyx":2197
 * 
 *     set_size = m - 1
 *     pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 *     for va in range(m):
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())
 */
  __pyx_v_pow_set_size = __Pyx_pow_long(2, ((long)__pyx_v_set_size));

  /* "cyext_acv.pyx":2198
 *     set_size = m - 1
 *     pow_set_size = 2**set_size
 *     for va in range(m):             # <<<<<<<<<<<<<<
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())
 *         Sm.erase(Sm.begin() + va)
 */
  __pyx_t_16 = __pyx_v_m;
  __pyx_t_19 = __pyx_t_16;
  for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
    __pyx_v_va = __pyx_t_20;

    /* "cyext_acv.pyx":2199
 *     pow_set_size = 2**set_size
 *     for va in range(m):
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())             # <<<<<<<<<<<<<<
 *         Sm.erase(Sm.begin() + va)
 * 
 */
    try {
      __pyx_v_Sm.assign(__pyx_v_va_id_cpp.begin(), __pyx_v_va_id_cpp.end());
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2199, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":2200
 *     for va in range(m):
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())
 *         Sm.erase(Sm.begin() + va)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(set_size):
 */
    (void)(__pyx_v_Sm.erase((__pyx_v_Sm.begin() + __pyx_v_va)));

    /* "cyext_acv.pyx":2202
 *         Sm.erase(Sm.begin() + va)
 * 
 *         for i in range(set_size):             # <<<<<<<<<<<<<<
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):
 */
    __pyx_t_28 = __pyx_v_set_size;
    __pyx_t_30 = __pyx_t_28;
    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
      __pyx_v_i = __pyx_t_31;

      /* "cyext_acv.pyx":2203
 * 
 *         for i in range(set_size):
 *             sm_size[i] = Sm[i].size()             # <<<<<<<<<<<<<<
 *             for j in range(sm_size[i]):
 *                 sm_buf[i, j] = Sm[i][j]
 */
      __pyx_t_29 = __pyx_v_i;
      *((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_29 * __pyx_v_sm_size.strides[0]) )) = (__pyx_v_Sm[__pyx_v_i]).size();

      /* "cyext_acv.pyx":2204
 *         for i in range(set_size):
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):             # <<<<<<<<<<<<<<
 *                 sm_buf[i, j] = Sm[i][j]
 * 
 */
      __pyx_t_29 = __pyx_v_i;
      __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_29 * __pyx_v_sm_size.strides[0]) )));
      __pyx_t_27 = __pyx_t_26;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_27; __pyx_t_32+=1) {
        __pyx_v_j = __pyx_t_32;

        /* "cyext_acv.pyx":2205
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):
 *                 sm_buf[i, j] = Sm[i][j]             # <<<<<<<<<<<<<<
 * 
 *         for counter in range(0, pow_set_size):
 */
        __pyx_t_29 = __pyx_v_i;
        __pyx_t_25 = __pyx_v_j;
        *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sm_buf.data + __pyx_t_29 * __pyx_v_sm_buf.strides[0]) ) + __pyx_t_25 * __pyx_v_sm_buf.strides[1]) )) = ((__pyx_v_Sm[__pyx_v_i])[__pyx_v_j]);
      }
    }

    /* "cyext_acv.pyx":2207
 *                 sm_buf[i, j] = Sm[i][j]
 * 
 *         for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *             va_size = 0
 *             S_size = 0
 */
    __pyx_t_28 = __pyx_v_pow_set_size;
    __pyx_t_30 = __pyx_t_28;
    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
      __pyx_v_counter = __pyx_t_31;

      /* "cyext_acv.pyx":2208
 * 
 *         for counter in range(0, pow_set_size):
 *             va_size = 0             # <<<<<<<<<<<<<<
 *             S_size = 0
 *             for ci in range(set_size):
 */
      __pyx_v_va_size = 0;

      /* "cyext_acv.pyx":2209
 *         for counter in range(0, pow_set_size):
 *             va_size = 0
 *             S_size = 0             # <<<<<<<<<<<<<<
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):
 */
      __pyx_v_S_size = 0;

      /* "cyext_acv.pyx":2210
 *             va_size = 0
 *             S_size = 0
 *             for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):
 */
      __pyx_t_32 = __pyx_v_set_size;
      __pyx_t_33 = __pyx_t_32;
      for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
        __pyx_v_ci = __pyx_t_34;

        /* "cyext_acv.pyx":2211
 *             S_size = 0
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 */
        __pyx_t_13 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":2212
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):             # <<<<<<<<<<<<<<
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1
 */
          __pyx_t_25 = __pyx_v_ci;
          __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_25 * __pyx_v_sm_size.strides[0]) )));
          __pyx_t_27 = __pyx_t_26;
          for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_27; __pyx_t_35+=1) {
            __pyx_v_cj = __pyx_t_35;

            /* "cyext_acv.pyx":2213
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]             # <<<<<<<<<<<<<<
 *                         S_size = S_size + 1
 *                     va_size = va_size + 1
 */
            __pyx_t_25 = __pyx_v_ci;
            __pyx_t_29 = __pyx_v_cj;
            __pyx_t_36 = __pyx_v_S_size;
            *((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_36)) )) = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sm_buf.data + __pyx_t_25 * __pyx_v_sm_buf.strides[0]) ) + __pyx_t_29 * __pyx_v_sm_buf.strides[1]) )));

            /* "cyext_acv.pyx":2214
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                     va_size = va_size + 1
 * 
 */
            __pyx_v_S_size = (__pyx_v_S_size + 1);
          }

          /* "cyext_acv.pyx":2215
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1
 *                     va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *             for i in range(len_va_id[va]):
 */
          __pyx_v_va_size = (__pyx_v_va_size + 1);

          /* "cyext_acv.pyx":2211
 *             S_size = 0
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 */
        }
      }

      /* "cyext_acv.pyx":2217
 *                     va_size = va_size + 1
 * 
 *             for i in range(len_va_id[va]):             # <<<<<<<<<<<<<<
 *                 S[S_size + i] = va_id_a[va, i]
 * 
 */
      __pyx_t_29 = __pyx_v_va;
      __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_29 * __pyx_v_len_va_id.strides[0]) )));
      __pyx_t_27 = __pyx_t_26;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_27; __pyx_t_32+=1) {
        __pyx_v_i = __pyx_t_32;

        /* "cyext_acv.pyx":2218
 * 
 *             for i in range(len_va_id[va]):
 *                 S[S_size + i] = va_id_a[va, i]             # <<<<<<<<<<<<<<
 * 
 *             weight = 1./binomialC(m - 1, va_size)
 */
        __pyx_t_29 = __pyx_v_va;
        __pyx_t_25 = __pyx_v_i;
        __pyx_t_36 = (__pyx_v_S_size + __pyx_v_i);
        *((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_36)) )) = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_29 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_25 * __pyx_v_va_id_a.strides[1]) )));
      }

      /* "cyext_acv.pyx":2220
 *                 S[S_size + i] = va_id_a[va, i]
 * 
 *             weight = 1./binomialC(m - 1, va_size)             # <<<<<<<<<<<<<<
 *             v_plus = compute_sdp_swing(X, fX, y_pred, S, S_size+len_va_id[va], data, values,
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 */
      __pyx_v_weight = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));

      /* "cyext_acv.pyx":2221
 * 
 *             weight = 1./binomialC(m - 1, va_size)
 *             v_plus = compute_sdp_swing(X, fX, y_pred, S, S_size+len_va_id[va], data, values,             # <<<<<<<<<<<<<<
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 *                       thresholds, num_threads)
 */
      __pyx_t_25 = __pyx_v_va;

      /* "cyext_acv.pyx":2223
 *             v_plus = compute_sdp_swing(X, fX, y_pred, S, S_size+len_va_id[va], data, values,
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 *                       thresholds, num_threads)             # <<<<<<<<<<<<<<
 * 
 *             v_minus = compute_sdp_swing(X, fX, y_pred, S, S_size, data, values,
 */
      __pyx_t_3 = __pyx_f_9cyext_acv_compute_sdp_swing(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, (__pyx_v_S_size + (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )))), __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_thresholds, __pyx_v_num_threads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "cyext_acv.pyx":2221
 * 
 *             weight = 1./binomialC(m - 1, va_size)
 *             v_plus = compute_sdp_swing(X, fX, y_pred, S, S_size+len_va_id[va], data, values,             # <<<<<<<<<<<<<<
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 *                       thresholds, num_threads)
 */
      __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2221, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_v_plus, 1);
      __pyx_v_v_plus = __pyx_t_7;
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;

      /* "cyext_acv.pyx":2225
 *                       thresholds, num_threads)
 * 
 *             v_minus = compute_sdp_swing(X, fX, y_pred, S, S_size, data, values,             # <<<<<<<<<<<<<<
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 *                       thresholds, num_threads)
 */
      __pyx_t_3 = __pyx_f_9cyext_acv_compute_sdp_swing(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_S_size, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_thresholds, __pyx_v_num_threads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2225, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_v_minus, 1);
      __pyx_v_v_minus = __pyx_t_7;
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;

      /* "cyext_acv.pyx":2229
 *                       thresholds, num_threads)
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 dif_pos = 1 if (v_plus[i] - v_minus[i]) > 0 else 0
 *                 dif_neg = 1 if (v_plus[i] - v_minus[i]) < 0 else 0
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {
            __pyx_t_32 = __pyx_v_N;
            if ((1 == 0)) abort();
            {
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   (x)
                    #define unlikely(x) (x)
                #endif
                __pyx_t_34 = (__pyx_t_32 - 0 + 1 - 1/abs(1)) / 1;
                if (__pyx_t_34 > 0)
                {
                    #ifdef _OPENMP
                    #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_29, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40)
                    #endif /* _OPENMP */
                    {
                        #ifdef _OPENMP
                        #pragma omp for lastprivate(__pyx_v_dif_neg) lastprivate(__pyx_v_dif_null) lastprivate(__pyx_v_dif_pos) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j)
                        #endif /* _OPENMP */
                        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_34; __pyx_t_33++){
                            {
                                __pyx_v_i = (unsigned int)(0 + 1 * __pyx_t_33);
                                /* Initialize private variables to invalid values */
                                __pyx_v_dif_neg = ((unsigned int)0xbad0bad0);
                                __pyx_v_dif_null = ((unsigned int)0xbad0bad0);
                                __pyx_v_dif_pos = ((unsigned int)0xbad0bad0);
                                __pyx_v_j = ((unsigned int)0xbad0bad0);

                                /* "cyext_acv.pyx":2230
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 dif_pos = 1 if (v_plus[i] - v_minus[i]) > 0 else 0             # <<<<<<<<<<<<<<
 *                 dif_neg = 1 if (v_plus[i] - v_minus[i]) < 0 else 0
 *                 dif_null = 1 if (v_plus[i] - v_minus[i]) == 0 else 0
 */
                                __pyx_t_25 = __pyx_v_i;
                                __pyx_t_29 = __pyx_v_i;
                                if (((((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_25 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_29 * __pyx_v_v_minus.strides[0]) )))) > 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_pos = __pyx_t_35;

                                /* "cyext_acv.pyx":2231
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 dif_pos = 1 if (v_plus[i] - v_minus[i]) > 0 else 0
 *                 dif_neg = 1 if (v_plus[i] - v_minus[i]) < 0 else 0             # <<<<<<<<<<<<<<
 *                 dif_null = 1 if (v_plus[i] - v_minus[i]) == 0 else 0
 * 
 */
                                __pyx_t_29 = __pyx_v_i;
                                __pyx_t_25 = __pyx_v_i;
                                if (((((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_29 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_25 * __pyx_v_v_minus.strides[0]) )))) < 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_neg = __pyx_t_35;

                                /* "cyext_acv.pyx":2232
 *                 dif_pos = 1 if (v_plus[i] - v_minus[i]) > 0 else 0
 *                 dif_neg = 1 if (v_plus[i] - v_minus[i]) < 0 else 0
 *                 dif_null = 1 if (v_plus[i] - v_minus[i]) == 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(len_va_id[va]):
 */
                                __pyx_t_25 = __pyx_v_i;
                                __pyx_t_29 = __pyx_v_i;
                                if (((((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_25 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_29 * __pyx_v_v_minus.strides[0]) )))) == 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_null = __pyx_t_35;

                                /* "cyext_acv.pyx":2234
 *                 dif_null = 1 if (v_plus[i] - v_minus[i]) == 0 else 0
 * 
 *                 for j in range(len_va_id[va]):             # <<<<<<<<<<<<<<
 *                     phi[i, va_id_a[va, j]] += weight * (v_plus[i] - v_minus[i])
 * 
 */
                                __pyx_t_29 = __pyx_v_va;
                                __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_29 * __pyx_v_len_va_id.strides[0]) )));
                                __pyx_t_27 = __pyx_t_26;
                                for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_27; __pyx_t_35+=1) {
                                  __pyx_v_j = __pyx_t_35;

                                  /* "cyext_acv.pyx":2235
 * 
 *                 for j in range(len_va_id[va]):
 *                     phi[i, va_id_a[va, j]] += weight * (v_plus[i] - v_minus[i])             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_36 = __pyx_v_va;
                                  __pyx_t_37 = __pyx_v_j;
                                  __pyx_t_38 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_36 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_37 * __pyx_v_va_id_a.strides[1]) )));
                                  *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_38 * __pyx_v_phi.strides[0]) ) + __pyx_t_39 * __pyx_v_phi.strides[1]) )) += (__pyx_v_weight * ((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_29 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_25 * __pyx_v_v_minus.strides[0]) )))));

                                  /* "cyext_acv.pyx":2238
 * 
 * 
 *                     swings[i, va_id_a[va, j], 0] += (dif_pos * (v_plus[i] - v_minus[i]) * weight) / m             # <<<<<<<<<<<<<<
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_plus[i] - v_minus[i]) * weight) / m
 * 
 */
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_37 = __pyx_v_va;
                                  __pyx_t_36 = __pyx_v_j;
                                  __pyx_t_38 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_37 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_36 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_40 = 0;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings.data + __pyx_t_38 * __pyx_v_swings.strides[0]) ) + __pyx_t_39 * __pyx_v_swings.strides[1]) ) + __pyx_t_40 * __pyx_v_swings.strides[2]) )) += (((__pyx_v_dif_pos * ((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_25 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_29 * __pyx_v_v_minus.strides[0]) ))))) * __pyx_v_weight) / __pyx_v_m);

                                  /* "cyext_acv.pyx":2239
 * 
 *                     swings[i, va_id_a[va, j], 0] += (dif_pos * (v_plus[i] - v_minus[i]) * weight) / m
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_plus[i] - v_minus[i]) * weight) / m             # <<<<<<<<<<<<<<
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 */
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_36 = __pyx_v_va;
                                  __pyx_t_37 = __pyx_v_j;
                                  __pyx_t_38 = __pyx_v_i;
                                  __pyx_t_40 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_36 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_37 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_39 = 1;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings.data + __pyx_t_38 * __pyx_v_swings.strides[0]) ) + __pyx_t_40 * __pyx_v_swings.strides[1]) ) + __pyx_t_39 * __pyx_v_swings.strides[2]) )) += (((__pyx_v_dif_neg * ((*((double *) ( /* dim=0 */ (__pyx_v_v_plus.data + __pyx_t_29 * __pyx_v_v_plus.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_v_minus.data + __pyx_t_25 * __pyx_v_v_minus.strides[0]) ))))) * __pyx_v_weight) / __pyx_v_m);

                                  /* "cyext_acv.pyx":2241
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_plus[i] - v_minus[i]) * weight) / m
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos             # <<<<<<<<<<<<<<
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 */
                                  __pyx_t_25 = __pyx_v_va;
                                  __pyx_t_29 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_29 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_40 = 0;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_37 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_40 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_pos;

                                  /* "cyext_acv.pyx":2242
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg             # <<<<<<<<<<<<<<
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 * 
 */
                                  __pyx_t_29 = __pyx_v_va;
                                  __pyx_t_25 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_40 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_29 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_25 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_39 = 1;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_37 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_40 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_neg;

                                  /* "cyext_acv.pyx":2243
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null             # <<<<<<<<<<<<<<
 * 
 *     return np.array(phi)/m, np.array(swings), np.array(swings_prop)
 */
                                  __pyx_t_25 = __pyx_v_va;
                                  __pyx_t_29 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_29 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_40 = 2;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_37 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_40 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_null;
                                }
                            }
                        }
                    }
                }
            }
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   __builtin_expect(!!(x), 1)
                #define unlikely(x) __builtin_expect(!!(x), 0)
            #endif
          }

          /* "cyext_acv.pyx":2229
 *                       thresholds, num_threads)
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 dif_pos = 1 if (v_plus[i] - v_minus[i]) > 0 else 0
 *                 dif_neg = 1 if (v_plus[i] - v_minus[i]) < 0 else 0
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L39;
            }
            __pyx_L39:;
          }
      }
    }
  }

  /* "cyext_acv.pyx":2245
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 * 
 *     return np.array(phi)/m, np.array(swings), np.array(swings_prop)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_phi, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_swings, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_swings_prop, 3, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_41 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_41 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_41)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_41);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_41) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_41, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_41); __pyx_t_41 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2121
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef swing_sv_clf(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const long[:] fX,
 *     const long[:] y_pred,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_v_plus, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_v_minus, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_swings_prop, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_swings, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sm_buf, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sm_size, 1);
  __Pyx_XDECREF(__pyx_v_C_buff);
  __Pyx_XDECREF(__pyx_v_coal_va);
  __Pyx_XDECREF(__pyx_v_va_id);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_c);
  __PYX_XDEC_MEMVIEW(&__pyx_v_va_id_a, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_va_id, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_27swing_sv_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_27swing_sv_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_thresholds;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("swing_sv_clf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_thresholds,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 1); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 2); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 3); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 4); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 5); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 6); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 7); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 8); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 9); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 10); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, 11); __PYX_ERR(0, 2121, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "swing_sv_clf") < 0)) __PYX_ERR(0, 2121, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 2121, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[1], 0); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 2122, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[2], 0); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 2123, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 2124, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[4], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 2125, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[5], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 2126, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[6], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 2127, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[7], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 2128, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(values[8], 0); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 2129, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[9]);
    __pyx_v_thresholds = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_thresholds == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2129, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2129, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("swing_sv_clf", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2121, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 2129, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_26swing_sv_clf(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_thresholds, __pyx_v_num_threads);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_26swing_sv_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_thresholds, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("swing_sv_clf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 2121, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_swing_sv_clf(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_thresholds, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2252
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef compute_sdp_swing_diff(const double[:, :] X, const long[:] fX,             # <<<<<<<<<<<<<<
 *             const long[:] y_pred, long[::1] S, long[:, :] va_id, long va, long[:] len_va_id, unsigned long S_size, const double[:, :] data,
 *             const double[:, :, :] values,const  double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_f_9cyext_acv_compute_sdp_swing_diff(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_S, __Pyx_memviewslice __pyx_v_va_id, long __pyx_v_va, __Pyx_memviewslice __pyx_v_len_va_id, unsigned long __pyx_v_S_size, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, double const __pyx_v_thresholds, CYTHON_UNUSED int __pyx_v_num_threads) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_m = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_csm;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_lm;
  CYTHON_UNUSED unsigned int __pyx_v_lm_m;
  unsigned int __pyx_v_lm_s;
  unsigned int __pyx_v_lm_s_m;
  unsigned int __pyx_v_p_s_m;
  unsigned int __pyx_v_p_su_m;
  unsigned int __pyx_v_p_sd_m;
  unsigned int __pyx_v_nv_bool;
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  double __pyx_v_ss;
  double __pyx_v_ss_u;
  double __pyx_v_ss_a;
  double __pyx_v_ss_d;
  __Pyx_memviewslice __pyx_v_p_n_m = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n_m = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n_m = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n_m;
  double __pyx_v_n_u_m;
  double __pyx_v_n_d_m;
  double __pyx_v_ss_m;
  double __pyx_v_ss_u_m;
  double __pyx_v_ss_a_m;
  double __pyx_v_ss_d_m;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  int __pyx_t_11;
  unsigned int __pyx_t_12;
  size_t __pyx_t_13;
  long __pyx_t_14;
  long __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  unsigned int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned long __pyx_t_20;
  unsigned long __pyx_t_21;
  unsigned int __pyx_t_22;
  int __pyx_t_23;
  size_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  size_t __pyx_t_26;
  size_t __pyx_t_27;
  size_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  int __pyx_t_31;
  long __pyx_t_32;
  long __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  unsigned int __pyx_t_38;
  unsigned int __pyx_t_39;
  unsigned int __pyx_t_40;
  double __pyx_t_41;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_swing_diff", 0);

  /* "cyext_acv.pyx":2258
 *             const double thresholds,int num_threads):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2259
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef double[:] out
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2260
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef double[:] out
 * 
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":2263
 *     cdef double[:] out
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":2269
 * 
 *     cdef double[:, :, :, :] mean_forest
 *     mean_forest = np.zeros((n_trees, max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_m
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2272
 * 
 *     cdef double[:, :, :, :] mean_forest_m
 *     mean_forest_m = np.zeros((n_trees, max_leaves, N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_m = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2275
 * 
 *     cdef double[:] sdp
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2275, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2282
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2283
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_u_n = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2284
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 *     cdef double ss, ss_u, ss_a, ss_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_d_n = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2289
 * 
 *     cdef double[:, :, :] p_n_m, p_u_n_m, p_d_n_m
 *     p_n_m = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n_m = np.zeros((n_trees, max_leaves, N))
 *     p_d_n_m = np.zeros((n_trees, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n_m = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2290
 *     cdef double[:, :, :] p_n_m, p_u_n_m, p_d_n_m
 *     p_n_m = np.zeros((n_trees, max_leaves, N))
 *     p_u_n_m = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n_m = np.zeros((n_trees, max_leaves, N))
 *     cdef double n_m, n_u_m, n_d_m
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_u_n_m = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2291
 *     p_n_m = np.zeros((n_trees, max_leaves, N))
 *     p_u_n_m = np.zeros((n_trees, max_leaves, N))
 *     p_d_n_m = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n_m, n_u_m, n_d_m
 *     cdef double ss_m, ss_u_m, ss_a_m, ss_d_m
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_d_n_m = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2295
 *     cdef double ss_m, ss_u_m, ss_a_m, ss_d_m
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __pyx_t_10 = __pyx_v_n_trees;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_b = __pyx_t_12;

    /* "cyext_acv.pyx":2296
 * 
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):             # <<<<<<<<<<<<<<
 *             for i in prange(N, nogil=True, schedule='dynamic'):
 *                 csm = 0
 */
    __pyx_t_13 = __pyx_v_b;
    __pyx_t_14 = (*((long const  *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_13 * __pyx_v_leaves_nb.strides[0]) )));
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_leaf_numb = __pyx_t_16;

      /* "cyext_acv.pyx":2297
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 csm = 0
 *                 o_all = 0
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {
            __pyx_t_17 = __pyx_v_N;
            if ((1 == 0)) abort();
            {
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   (x)
                    #define unlikely(x) (x)
                #endif
                __pyx_t_19 = (__pyx_t_17 - 0 + 1 - 1/abs(1)) / 1;
                if (__pyx_t_19 > 0)
                {
                    #ifdef _OPENMP
                    #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_s_m) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_sd_m) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_su_m) reduction(+:__pyx_v_p_u) private(__pyx_t_13, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41)
                    #endif /* _OPENMP */
                    {
                        #ifdef _OPENMP
                        #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_csm) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_m) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_lm_s_m) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) schedule(dynamic)
                        #endif /* _OPENMP */
                        for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_19; __pyx_t_18++){
                            {
                                __pyx_v_i = (unsigned int)(0 + 1 * __pyx_t_18);
                                /* Initialize private variables to invalid values */
                                __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_csm = ((unsigned int)0xbad0bad0);
                                __pyx_v_j = ((unsigned int)0xbad0bad0);
                                __pyx_v_lm = ((unsigned int)0xbad0bad0);
                                __pyx_v_lm_m = ((unsigned int)0xbad0bad0);
                                __pyx_v_lm_s = ((unsigned int)0xbad0bad0);
                                __pyx_v_lm_s_m = ((unsigned int)0xbad0bad0);
                                __pyx_v_nv = ((unsigned int)0xbad0bad0);
                                __pyx_v_nv_bool = ((unsigned int)0xbad0bad0);
                                __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                __pyx_v_s = ((unsigned int)0xbad0bad0);

                                /* "cyext_acv.pyx":2298
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, schedule='dynamic'):
 *                 csm = 0             # <<<<<<<<<<<<<<
 *                 o_all = 0
 *                 for s in range(S_size):
 */
                                __pyx_v_csm = 0;

                                /* "cyext_acv.pyx":2299
 *             for i in prange(N, nogil=True, schedule='dynamic'):
 *                 csm = 0
 *                 o_all = 0             # <<<<<<<<<<<<<<
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                __pyx_v_o_all = 0;

                                /* "cyext_acv.pyx":2300
 *                 csm = 0
 *                 o_all = 0
 *                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 */
                                __pyx_t_20 = __pyx_v_S_size;
                                __pyx_t_21 = __pyx_t_20;
                                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                                  __pyx_v_s = __pyx_t_22;

                                  /* "cyext_acv.pyx":2301
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  __pyx_t_13 = __pyx_v_s;
                                  __pyx_t_24 = __pyx_v_i;
                                  __pyx_t_25 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_13)) )));
                                  __pyx_t_26 = __pyx_v_s;
                                  __pyx_t_27 = __pyx_v_b;
                                  __pyx_t_28 = __pyx_v_leaf_numb;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_26)) )));
                                  __pyx_t_30 = 1;
                                  __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_24 * __pyx_v_X.strides[0]) ) + __pyx_t_25 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_30 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (!__pyx_t_31) {
                                  } else {
                                    __pyx_t_23 = __pyx_t_31;
                                    goto __pyx_L19_bool_binop_done;
                                  }
                                  __pyx_t_26 = __pyx_v_s;
                                  __pyx_t_28 = __pyx_v_i;
                                  __pyx_t_30 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_26)) )));
                                  __pyx_t_27 = __pyx_v_s;
                                  __pyx_t_13 = __pyx_v_b;
                                  __pyx_t_24 = __pyx_v_leaf_numb;
                                  __pyx_t_29 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_27)) )));
                                  __pyx_t_25 = 0;
                                  __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_28 * __pyx_v_X.strides[0]) ) + __pyx_t_30 * __pyx_v_X.strides[1]) ))) < (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_13 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_23 = __pyx_t_31;
                                  __pyx_L19_bool_binop_done:;
                                  if (__pyx_t_23) {

                                    /* "cyext_acv.pyx":2302
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                 if o_all > 0:
 *                     continue
 */
                                    __pyx_v_o_all = (__pyx_v_o_all + 1);

                                    /* "cyext_acv.pyx":2301
 *                 o_all = 0
 *                 for s in range(S_size):
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":2303
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
                                __pyx_t_23 = ((__pyx_v_o_all > 0) != 0);
                                if (__pyx_t_23) {

                                  /* "cyext_acv.pyx":2304
 *                         o_all = o_all + 1
 *                 if o_all > 0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     nv_bool = 0
 */
                                  goto __pyx_L12_continue;

                                  /* "cyext_acv.pyx":2303
 *                     if ((X[i, S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[i, S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                         o_all = o_all + 1
 *                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 else:
 */
                                }

                                /* "cyext_acv.pyx":2306
 *                     continue
 *                 else:
 *                     nv_bool = 0             # <<<<<<<<<<<<<<
 *                     for nv in range(len_va_id[va]):
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 */
                                /*else*/ {
                                  __pyx_v_nv_bool = 0;

                                  /* "cyext_acv.pyx":2307
 *                 else:
 *                     nv_bool = 0
 *                     for nv in range(len_va_id[va]):             # <<<<<<<<<<<<<<
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                             nv_bool = nv_bool + 1
 */
                                  __pyx_t_25 = __pyx_v_va;
                                  __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )));
                                  __pyx_t_33 = __pyx_t_32;
                                  for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_33; __pyx_t_22+=1) {
                                    __pyx_v_nv = __pyx_t_22;

                                    /* "cyext_acv.pyx":2308
 *                     nv_bool = 0
 *                     for nv in range(len_va_id[va]):
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):             # <<<<<<<<<<<<<<
 *                             nv_bool = nv_bool + 1
 *                             continue
 */
                                    __pyx_t_25 = __pyx_v_va;
                                    __pyx_t_27 = __pyx_v_nv;
                                    __pyx_t_24 = __pyx_v_i;
                                    __pyx_t_29 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_25 * __pyx_v_va_id.strides[0]) ) + __pyx_t_27 * __pyx_v_va_id.strides[1]) )));
                                    __pyx_t_30 = __pyx_v_va;
                                    __pyx_t_13 = __pyx_v_nv;
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_28 = __pyx_v_leaf_numb;
                                    __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_30 * __pyx_v_va_id.strides[0]) ) + __pyx_t_13 * __pyx_v_va_id.strides[1]) )));
                                    __pyx_t_35 = 1;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_24 * __pyx_v_X.strides[0]) ) + __pyx_t_29 * __pyx_v_X.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (!__pyx_t_31) {
                                    } else {
                                      __pyx_t_23 = __pyx_t_31;
                                      goto __pyx_L25_bool_binop_done;
                                    }
                                    __pyx_t_30 = __pyx_v_va;
                                    __pyx_t_13 = __pyx_v_nv;
                                    __pyx_t_28 = __pyx_v_i;
                                    __pyx_t_35 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_30 * __pyx_v_va_id.strides[0]) ) + __pyx_t_13 * __pyx_v_va_id.strides[1]) )));
                                    __pyx_t_34 = __pyx_v_va;
                                    __pyx_t_26 = __pyx_v_nv;
                                    __pyx_t_27 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_25 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_34 * __pyx_v_va_id.strides[0]) ) + __pyx_t_26 * __pyx_v_va_id.strides[1]) )));
                                    __pyx_t_29 = 0;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_28 * __pyx_v_X.strides[0]) ) + __pyx_t_35 * __pyx_v_X.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_23 = __pyx_t_31;
                                    __pyx_L25_bool_binop_done:;
                                    if (__pyx_t_23) {

                                      /* "cyext_acv.pyx":2309
 *                     for nv in range(len_va_id[va]):
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                             nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                      __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                      /* "cyext_acv.pyx":2310
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                             nv_bool = nv_bool + 1
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                     if nv_bool == 0:
 */
                                      goto __pyx_L22_continue;

                                      /* "cyext_acv.pyx":2308
 *                     nv_bool = 0
 *                     for nv in range(len_va_id[va]):
 *                         if (X[i, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (X[i, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):             # <<<<<<<<<<<<<<
 *                             nv_bool = nv_bool + 1
 *                             continue
 */
                                    }
                                    __pyx_L22_continue:;
                                  }

                                  /* "cyext_acv.pyx":2312
 *                             continue
 * 
 *                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                         csm = 1
 *                     else:
 */
                                  __pyx_t_23 = ((__pyx_v_nv_bool == 0) != 0);
                                  if (__pyx_t_23) {

                                    /* "cyext_acv.pyx":2313
 * 
 *                     if nv_bool == 0:
 *                         csm = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         continue
 */
                                    __pyx_v_csm = 1;

                                    /* "cyext_acv.pyx":2312
 *                             continue
 * 
 *                     if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                         csm = 1
 *                     else:
 */
                                    goto __pyx_L27;
                                  }

                                  /* "cyext_acv.pyx":2315
 *                         csm = 1
 *                     else:
 *                         continue             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  /*else*/ {
                                    goto __pyx_L12_continue;
                                  }
                                  __pyx_L27:;
                                }

                                /* "cyext_acv.pyx":2318
 * 
 * 
 *                 p = 0             # <<<<<<<<<<<<<<
 *                 p_u = 0
 *                 p_d = 0
 */
                                __pyx_v_p = 0;

                                /* "cyext_acv.pyx":2319
 * 
 *                 p = 0
 *                 p_u = 0             # <<<<<<<<<<<<<<
 *                 p_d = 0
 * 
 */
                                __pyx_v_p_u = 0;

                                /* "cyext_acv.pyx":2320
 *                 p = 0
 *                 p_u = 0
 *                 p_d = 0             # <<<<<<<<<<<<<<
 * 
 *                 p_s = 0
 */
                                __pyx_v_p_d = 0;

                                /* "cyext_acv.pyx":2322
 *                 p_d = 0
 * 
 *                 p_s = 0             # <<<<<<<<<<<<<<
 *                 p_su = 0
 *                 p_sd = 0
 */
                                __pyx_v_p_s = 0;

                                /* "cyext_acv.pyx":2323
 * 
 *                 p_s = 0
 *                 p_su = 0             # <<<<<<<<<<<<<<
 *                 p_sd = 0
 * 
 */
                                __pyx_v_p_su = 0;

                                /* "cyext_acv.pyx":2324
 *                 p_s = 0
 *                 p_su = 0
 *                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                 p_s_m = 0
 */
                                __pyx_v_p_sd = 0;

                                /* "cyext_acv.pyx":2326
 *                 p_sd = 0
 * 
 *                 p_s_m = 0             # <<<<<<<<<<<<<<
 *                 p_su_m = 0
 *                 p_sd_m = 0
 */
                                __pyx_v_p_s_m = 0;

                                /* "cyext_acv.pyx":2327
 * 
 *                 p_s_m = 0
 *                 p_su_m = 0             # <<<<<<<<<<<<<<
 *                 p_sd_m = 0
 * 
 */
                                __pyx_v_p_su_m = 0;

                                /* "cyext_acv.pyx":2328
 *                 p_s_m = 0
 *                 p_su_m = 0
 *                 p_sd_m = 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(data.shape[0]):
 */
                                __pyx_v_p_sd_m = 0;

                                /* "cyext_acv.pyx":2330
 *                 p_sd_m = 0
 * 
 *                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                     a_it = 0
 *                     b_it = 0
 */
                                __pyx_t_36 = (__pyx_v_data.shape[0]);
                                __pyx_t_37 = __pyx_t_36;
                                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_37; __pyx_t_22+=1) {
                                  __pyx_v_j = __pyx_t_22;

                                  /* "cyext_acv.pyx":2331
 * 
 *                 for j in range(data.shape[0]):
 *                     a_it = 0             # <<<<<<<<<<<<<<
 *                     b_it = 0
 * 
 */
                                  __pyx_v_a_it = 0;

                                  /* "cyext_acv.pyx":2332
 *                 for j in range(data.shape[0]):
 *                     a_it = 0
 *                     b_it = 0             # <<<<<<<<<<<<<<
 * 
 *                     lm = 0
 */
                                  __pyx_v_b_it = 0;

                                  /* "cyext_acv.pyx":2334
 *                     b_it = 0
 * 
 *                     lm = 0             # <<<<<<<<<<<<<<
 *                     lm_s = 0
 * 
 */
                                  __pyx_v_lm = 0;

                                  /* "cyext_acv.pyx":2335
 * 
 *                     lm = 0
 *                     lm_s = 0             # <<<<<<<<<<<<<<
 * 
 *                     lm_m = 0
 */
                                  __pyx_v_lm_s = 0;

                                  /* "cyext_acv.pyx":2337
 *                     lm_s = 0
 * 
 *                     lm_m = 0             # <<<<<<<<<<<<<<
 *                     lm_s_m = 0
 * 
 */
                                  __pyx_v_lm_m = 0;

                                  /* "cyext_acv.pyx":2338
 * 
 *                     lm_m = 0
 *                     lm_s_m = 0             # <<<<<<<<<<<<<<
 * 
 *                     for s in range(m):
 */
                                  __pyx_v_lm_s_m = 0;

                                  /* "cyext_acv.pyx":2340
 *                     lm_s_m = 0
 * 
 *                     for s in range(m):             # <<<<<<<<<<<<<<
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 */
                                  __pyx_t_38 = __pyx_v_m;
                                  __pyx_t_39 = __pyx_t_38;
                                  for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
                                    __pyx_v_s = __pyx_t_40;

                                    /* "cyext_acv.pyx":2341
 * 
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    __pyx_t_26 = __pyx_v_j;
                                    __pyx_t_24 = __pyx_v_s;
                                    __pyx_t_27 = __pyx_v_b;
                                    __pyx_t_13 = __pyx_v_leaf_numb;
                                    __pyx_t_28 = __pyx_v_s;
                                    __pyx_t_34 = 1;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_26 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_13 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_31) {
                                    } else {
                                      __pyx_t_23 = __pyx_t_31;
                                      goto __pyx_L33_bool_binop_done;
                                    }
                                    __pyx_t_28 = __pyx_v_j;
                                    __pyx_t_13 = __pyx_v_s;
                                    __pyx_t_27 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_34 = 0;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_28 * __pyx_v_data.strides[0]) ) + __pyx_t_13 * __pyx_v_data.strides[1]) ))) >= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_23 = __pyx_t_31;
                                    __pyx_L33_bool_binop_done:;
                                    if (__pyx_t_23) {

                                      /* "cyext_acv.pyx":2342
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                      __pyx_v_a_it = (__pyx_v_a_it + 1);

                                      /* "cyext_acv.pyx":2341
 * 
 *                     for s in range(m):
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":2343
 *                         if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                             a_it = a_it + 1
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1
 */
                                  __pyx_t_20 = __pyx_v_S_size;
                                  __pyx_t_21 = __pyx_t_20;
                                  for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_21; __pyx_t_38+=1) {
                                    __pyx_v_s = __pyx_t_38;

                                    /* "cyext_acv.pyx":2344
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    __pyx_t_26 = __pyx_v_s;
                                    __pyx_t_24 = __pyx_v_j;
                                    __pyx_t_34 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_26)) )));
                                    __pyx_t_27 = __pyx_v_s;
                                    __pyx_t_13 = __pyx_v_b;
                                    __pyx_t_28 = __pyx_v_leaf_numb;
                                    __pyx_t_29 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_27)) )));
                                    __pyx_t_25 = 1;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_24 * __pyx_v_data.strides[0]) ) + __pyx_t_34 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_13 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_25 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    if (__pyx_t_31) {
                                    } else {
                                      __pyx_t_23 = __pyx_t_31;
                                      goto __pyx_L38_bool_binop_done;
                                    }
                                    __pyx_t_27 = __pyx_v_s;
                                    __pyx_t_28 = __pyx_v_j;
                                    __pyx_t_25 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_27)) )));
                                    __pyx_t_13 = __pyx_v_s;
                                    __pyx_t_26 = __pyx_v_b;
                                    __pyx_t_24 = __pyx_v_leaf_numb;
                                    __pyx_t_29 = (*((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_13)) )));
                                    __pyx_t_34 = 0;
                                    __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_28 * __pyx_v_data.strides[0]) ) + __pyx_t_25 * __pyx_v_data.strides[1]) ))) >= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_26 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                    __pyx_t_23 = __pyx_t_31;
                                    __pyx_L38_bool_binop_done:;
                                    if (__pyx_t_23) {

                                      /* "cyext_acv.pyx":2345
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                     if a_it == m:
 */
                                      __pyx_v_b_it = (__pyx_v_b_it + 1);

                                      /* "cyext_acv.pyx":2344
 *                             a_it = a_it + 1
 *                     for s in range(S_size):
 *                         if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                             b_it = b_it + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":2347
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  __pyx_t_23 = ((__pyx_v_a_it == __pyx_v_m) != 0);
                                  if (__pyx_t_23) {

                                    /* "cyext_acv.pyx":2348
 * 
 *                     if a_it == m:
 *                         lm = 1             # <<<<<<<<<<<<<<
 * 
 *                     if b_it == S_size:
 */
                                    __pyx_v_lm = 1;

                                    /* "cyext_acv.pyx":2347
 *                             b_it = b_it + 1
 * 
 *                     if a_it == m:             # <<<<<<<<<<<<<<
 *                         lm = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":2350
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  __pyx_t_23 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                  if (__pyx_t_23) {

                                    /* "cyext_acv.pyx":2351
 * 
 *                     if b_it == S_size:
 *                         lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                         nv_bool = 0
 */
                                    __pyx_v_lm_s = 1;

                                    /* "cyext_acv.pyx":2353
 *                         lm_s = 1
 * 
 *                         nv_bool = 0             # <<<<<<<<<<<<<<
 *                         for nv in range(len_va_id[va]):
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 */
                                    __pyx_v_nv_bool = 0;

                                    /* "cyext_acv.pyx":2354
 * 
 *                         nv_bool = 0
 *                         for nv in range(len_va_id[va]):             # <<<<<<<<<<<<<<
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                                 nv_bool = nv_bool + 1
 */
                                    __pyx_t_34 = __pyx_v_va;
                                    __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_34 * __pyx_v_len_va_id.strides[0]) )));
                                    __pyx_t_33 = __pyx_t_32;
                                    for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_33; __pyx_t_38+=1) {
                                      __pyx_v_nv = __pyx_t_38;

                                      /* "cyext_acv.pyx":2355
 *                         nv_bool = 0
 *                         for nv in range(len_va_id[va]):
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      __pyx_t_34 = __pyx_v_va;
                                      __pyx_t_13 = __pyx_v_nv;
                                      __pyx_t_24 = __pyx_v_j;
                                      __pyx_t_29 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_34 * __pyx_v_va_id.strides[0]) ) + __pyx_t_13 * __pyx_v_va_id.strides[1]) )));
                                      __pyx_t_25 = __pyx_v_va;
                                      __pyx_t_26 = __pyx_v_nv;
                                      __pyx_t_27 = __pyx_v_b;
                                      __pyx_t_28 = __pyx_v_leaf_numb;
                                      __pyx_t_30 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_25 * __pyx_v_va_id.strides[0]) ) + __pyx_t_26 * __pyx_v_va_id.strides[1]) )));
                                      __pyx_t_35 = 1;
                                      __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_24 * __pyx_v_data.strides[0]) ) + __pyx_t_29 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_27 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_28 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_30 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_31) {
                                      } else {
                                        __pyx_t_23 = __pyx_t_31;
                                        goto __pyx_L45_bool_binop_done;
                                      }
                                      __pyx_t_25 = __pyx_v_va;
                                      __pyx_t_26 = __pyx_v_nv;
                                      __pyx_t_28 = __pyx_v_j;
                                      __pyx_t_35 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_25 * __pyx_v_va_id.strides[0]) ) + __pyx_t_26 * __pyx_v_va_id.strides[1]) )));
                                      __pyx_t_30 = __pyx_v_va;
                                      __pyx_t_27 = __pyx_v_nv;
                                      __pyx_t_13 = __pyx_v_b;
                                      __pyx_t_24 = __pyx_v_leaf_numb;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id.data + __pyx_t_30 * __pyx_v_va_id.strides[0]) ) + __pyx_t_27 * __pyx_v_va_id.strides[1]) )));
                                      __pyx_t_29 = 0;
                                      __pyx_t_31 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_28 * __pyx_v_data.strides[0]) ) + __pyx_t_35 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_13 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_29 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_23 = __pyx_t_31;
                                      __pyx_L45_bool_binop_done:;
                                      if (__pyx_t_23) {

                                        /* "cyext_acv.pyx":2356
 *                         for nv in range(len_va_id[va]):
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                                 nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                        __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                        /* "cyext_acv.pyx":2357
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):
 *                                 nv_bool = nv_bool + 1
 *                                 continue             # <<<<<<<<<<<<<<
 * 
 *                         if nv_bool == 0:
 */
                                        goto __pyx_L42_continue;

                                        /* "cyext_acv.pyx":2355
 *                         nv_bool = 0
 *                         for nv in range(len_va_id[va]):
 *                             if (data[j, va_id[va, nv]] >  partition_leaves_trees[b, leaf_numb, va_id[va, nv], 1]) or (data[j, va_id[va, nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va, nv], 0]):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      }
                                      __pyx_L42_continue:;
                                    }

                                    /* "cyext_acv.pyx":2359
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             lm_s_m = 1
 * 
 */
                                    __pyx_t_23 = ((__pyx_v_nv_bool == 0) != 0);
                                    if (__pyx_t_23) {

                                      /* "cyext_acv.pyx":2360
 * 
 *                         if nv_bool == 0:
 *                             lm_s_m = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                      __pyx_v_lm_s_m = 1;

                                      /* "cyext_acv.pyx":2359
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             lm_s_m = 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":2350
 *                         lm = 1
 * 
 *                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                         lm_s = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":2363
 * 
 * 
 *                     p += lm             # <<<<<<<<<<<<<<
 *                     p_s += lm_s
 *                     p_s_m += lm_s_m
 */
                                  __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                  /* "cyext_acv.pyx":2364
 * 
 *                     p += lm
 *                     p_s += lm_s             # <<<<<<<<<<<<<<
 *                     p_s_m += lm_s_m
 * 
 */
                                  __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                  /* "cyext_acv.pyx":2365
 *                     p += lm
 *                     p_s += lm_s
 *                     p_s_m += lm_s_m             # <<<<<<<<<<<<<<
 * 
 *                     if fX[i] == y_pred[j]:
 */
                                  __pyx_v_p_s_m = (__pyx_v_p_s_m + __pyx_v_lm_s_m);

                                  /* "cyext_acv.pyx":2367
 *                     p_s_m += lm_s_m
 * 
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u += lm
 *                         p_su += lm_s
 */
                                  __pyx_t_27 = __pyx_v_i;
                                  __pyx_t_24 = __pyx_v_j;
                                  __pyx_t_23 = (((*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_27 * __pyx_v_fX.strides[0]) ))) == (*((long const  *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_24 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                  if (__pyx_t_23) {

                                    /* "cyext_acv.pyx":2368
 * 
 *                     if fX[i] == y_pred[j]:
 *                         p_u += lm             # <<<<<<<<<<<<<<
 *                         p_su += lm_s
 *                         p_su_m += lm_s_m
 */
                                    __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                    /* "cyext_acv.pyx":2369
 *                     if fX[i] == y_pred[j]:
 *                         p_u += lm
 *                         p_su += lm_s             # <<<<<<<<<<<<<<
 *                         p_su_m += lm_s_m
 *                     else:
 */
                                    __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                    /* "cyext_acv.pyx":2370
 *                         p_u += lm
 *                         p_su += lm_s
 *                         p_su_m += lm_s_m             # <<<<<<<<<<<<<<
 *                     else:
 *                         p_d += lm
 */
                                    __pyx_v_p_su_m = (__pyx_v_p_su_m + __pyx_v_lm_s_m);

                                    /* "cyext_acv.pyx":2367
 *                     p_s_m += lm_s_m
 * 
 *                     if fX[i] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                         p_u += lm
 *                         p_su += lm_s
 */
                                    goto __pyx_L48;
                                  }

                                  /* "cyext_acv.pyx":2372
 *                         p_su_m += lm_s_m
 *                     else:
 *                         p_d += lm             # <<<<<<<<<<<<<<
 *                         p_sd += lm_s
 *                         p_sd_m += lm_s_m
 */
                                  /*else*/ {
                                    __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                    /* "cyext_acv.pyx":2373
 *                     else:
 *                         p_d += lm
 *                         p_sd += lm_s             # <<<<<<<<<<<<<<
 *                         p_sd_m += lm_s_m
 * 
 */
                                    __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);

                                    /* "cyext_acv.pyx":2374
 *                         p_d += lm
 *                         p_sd += lm_s
 *                         p_sd_m += lm_s_m             # <<<<<<<<<<<<<<
 * 
 *                 mean_forest[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 */
                                    __pyx_v_p_sd_m = (__pyx_v_p_sd_m + __pyx_v_lm_s_m);
                                  }
                                  __pyx_L48:;
                                }

                                /* "cyext_acv.pyx":2376
 *                         p_sd_m += lm_s_m
 * 
 *                 mean_forest[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_13 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_30 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_29 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_13 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_p * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_30 * __pyx_v_values.strides[1]) ) + __pyx_t_29 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                __pyx_t_29 = 0;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_29 * __pyx_v_mean_forest.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2377
 * 
 *                 mean_forest[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_13 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_29 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_30 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_13 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_p_u * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_29 * __pyx_v_values.strides[1]) ) + __pyx_t_30 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                __pyx_t_30 = 1;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2378
 *                 mean_forest[b, leaf_numb, i, 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s) if p_s != 0 else 0
 *                 mean_forest[b, leaf_numb, i, 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su) if p_su != 0 else 0
 *                 mean_forest[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_24 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_13 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_30 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_24 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_29 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_13 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_p_d * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_30 * __pyx_v_values.strides[1]) ) + __pyx_t_29 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                __pyx_t_29 = 2;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_29 * __pyx_v_mean_forest.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2380
 *                 mean_forest[b, leaf_numb, i, 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd) if p_sd != 0 else 0
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                if (((__pyx_v_p_s != 0) != 0)) {
                                  __pyx_t_41 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_24 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_n.strides[2]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2381
 * 
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su != 0) != 0)) {
                                  __pyx_t_41 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_13 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[2]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2382
 *                 p_n[b, leaf_numb, i] += (1.*p)/p_s if p_s != 0 else 0
 *                 p_u_n[b, leaf_numb, i] += (1.*p_u)/p_su if p_su != 0 else 0
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 mean_forest_m[b, leaf_numb, i, 0] += csm*(p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s_m) if p_s_m != 0 else 0
 */
                                if (((__pyx_v_p_sd != 0) != 0)) {
                                  __pyx_t_41 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_24 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_d_n.strides[2]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2384
 *                 p_d_n[b, leaf_numb, i] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 *                 mean_forest_m[b, leaf_numb, i, 0] += csm*(p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s_m) if p_s_m != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_m[b, leaf_numb, i, 1] += csm*(p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su_m) if p_su_m != 0 else 0
 *                 mean_forest_m[b, leaf_numb, i, 2] += csm*(p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd_m) if p_sd_m != 0 else 0
 */
                                if (((__pyx_v_p_s_m != 0) != 0)) {
                                  __pyx_t_13 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_29 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_13 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_30 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_csm * (__pyx_v_p * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_29 * __pyx_v_values.strides[1]) ) + __pyx_t_30 * __pyx_v_values.strides[2]) ))))) / __pyx_v_p_s_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                __pyx_t_30 = 0;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_24 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2385
 * 
 *                 mean_forest_m[b, leaf_numb, i, 0] += csm*(p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s_m) if p_s_m != 0 else 0
 *                 mean_forest_m[b, leaf_numb, i, 1] += csm*(p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su_m) if p_su_m != 0 else 0             # <<<<<<<<<<<<<<
 *                 mean_forest_m[b, leaf_numb, i, 2] += csm*(p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd_m) if p_sd_m != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su_m != 0) != 0)) {
                                  __pyx_t_13 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_30 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_13 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_29 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_csm * (__pyx_v_p_u * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_30 * __pyx_v_values.strides[1]) ) + __pyx_t_29 * __pyx_v_values.strides[2]) ))))) / __pyx_v_p_su_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                __pyx_t_29 = 1;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_24 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_29 * __pyx_v_mean_forest_m.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2386
 *                 mean_forest_m[b, leaf_numb, i, 0] += csm*(p * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_s_m) if p_s_m != 0 else 0
 *                 mean_forest_m[b, leaf_numb, i, 1] += csm*(p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_su_m) if p_su_m != 0 else 0
 *                 mean_forest_m[b, leaf_numb, i, 2] += csm*(p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd_m) if p_sd_m != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 p_n_m[b, leaf_numb, i] += csm*(1.*p)/p_s_m if p_s_m != 0 else 0
 */
                                if (((__pyx_v_p_sd_m != 0) != 0)) {
                                  __pyx_t_13 = __pyx_v_b;
                                  __pyx_t_27 = __pyx_v_leaf_numb;
                                  __pyx_t_24 = __pyx_v_i;
                                  __pyx_t_26 = __pyx_v_b;
                                  __pyx_t_29 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_13 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_27 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                  __pyx_t_30 = (*((long const  *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_24 * __pyx_v_fX.strides[0]) )));
                                  __pyx_t_41 = ((__pyx_v_csm * (__pyx_v_p_d * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_26 * __pyx_v_values.strides[0]) ) + __pyx_t_29 * __pyx_v_values.strides[1]) ) + __pyx_t_30 * __pyx_v_values.strides[2]) ))))) / __pyx_v_p_sd_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                __pyx_t_30 = 2;
                                *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_24 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2388
 *                 mean_forest_m[b, leaf_numb, i, 2] += csm*(p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[i]]) / (p_sd_m) if p_sd_m != 0 else 0
 * 
 *                 p_n_m[b, leaf_numb, i] += csm*(1.*p)/p_s_m if p_s_m != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_u_n_m[b, leaf_numb, i] += csm*(1.*p_u)/p_su_m if p_su_m != 0 else 0
 *                 p_d_n_m[b, leaf_numb, i] += csm*(1.*p_d)/p_sd_m if p_sd_m != 0 else 0
 */
                                if (((__pyx_v_p_s_m != 0) != 0)) {
                                  __pyx_t_41 = ((__pyx_v_csm * (1. * __pyx_v_p)) / __pyx_v_p_s_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n_m.data + __pyx_t_13 * __pyx_v_p_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n_m.strides[1]) ) + __pyx_t_24 * __pyx_v_p_n_m.strides[2]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2389
 * 
 *                 p_n_m[b, leaf_numb, i] += csm*(1.*p)/p_s_m if p_s_m != 0 else 0
 *                 p_u_n_m[b, leaf_numb, i] += csm*(1.*p_u)/p_su_m if p_su_m != 0 else 0             # <<<<<<<<<<<<<<
 *                 p_d_n_m[b, leaf_numb, i] += csm*(1.*p_d)/p_sd_m if p_sd_m != 0 else 0
 * 
 */
                                if (((__pyx_v_p_su_m != 0) != 0)) {
                                  __pyx_t_41 = ((__pyx_v_csm * (1. * __pyx_v_p_u)) / __pyx_v_p_su_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_24 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_13 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n_m.data + __pyx_t_24 * __pyx_v_p_u_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n_m.strides[1]) ) + __pyx_t_13 * __pyx_v_p_u_n_m.strides[2]) )) += __pyx_t_41;

                                /* "cyext_acv.pyx":2390
 *                 p_n_m[b, leaf_numb, i] += csm*(1.*p)/p_s_m if p_s_m != 0 else 0
 *                 p_u_n_m[b, leaf_numb, i] += csm*(1.*p_u)/p_su_m if p_su_m != 0 else 0
 *                 p_d_n_m[b, leaf_numb, i] += csm*(1.*p_d)/p_sd_m if p_sd_m != 0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                if (((__pyx_v_p_sd_m != 0) != 0)) {
                                  __pyx_t_41 = ((__pyx_v_csm * (1. * __pyx_v_p_d)) / __pyx_v_p_sd_m);
                                } else {
                                  __pyx_t_41 = 0.0;
                                }
                                __pyx_t_13 = __pyx_v_b;
                                __pyx_t_27 = __pyx_v_leaf_numb;
                                __pyx_t_24 = __pyx_v_i;
                                *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n_m.data + __pyx_t_13 * __pyx_v_p_d_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n_m.strides[1]) ) + __pyx_t_24 * __pyx_v_p_d_n_m.strides[2]) )) += __pyx_t_41;
                                goto __pyx_L50;
                                __pyx_L12_continue:;
                                goto __pyx_L50;
                                __pyx_L50:;
                            }
                        }
                    }
                }
            }
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   __builtin_expect(!!(x), 1)
                #define unlikely(x) __builtin_expect(!!(x), 0)
            #endif
          }

          /* "cyext_acv.pyx":2297
 *     for b in range(n_trees):
 *         for leaf_numb in range(leaves_nb[b]):
 *             for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 csm = 0
 *                 o_all = 0
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L11;
            }
            __pyx_L11:;
          }
      }
    }
  }

  /* "cyext_acv.pyx":2393
 * 
 * 
 *     if S_size !=0 and S_size + len_va_id[va] != m:             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             ss_u = 0
 */
  __pyx_t_31 = ((__pyx_v_S_size != 0) != 0);
  if (__pyx_t_31) {
  } else {
    __pyx_t_23 = __pyx_t_31;
    goto __pyx_L52_bool_binop_done;
  }
  __pyx_t_30 = __pyx_v_va;
  __pyx_t_31 = (((__pyx_v_S_size + (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_30 * __pyx_v_len_va_id.strides[0]) )))) != __pyx_v_m) != 0);
  __pyx_t_23 = __pyx_t_31;
  __pyx_L52_bool_binop_done:;
  if (__pyx_t_23) {

    /* "cyext_acv.pyx":2394
 * 
 *     if S_size !=0 and S_size + len_va_id[va] != m:
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             ss_u = 0
 *             ss_d = 0
 */
    __pyx_t_12 = __pyx_v_N;
    __pyx_t_16 = __pyx_t_12;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_16; __pyx_t_19+=1) {
      __pyx_v_i = __pyx_t_19;

      /* "cyext_acv.pyx":2395
 *     if S_size !=0 and S_size + len_va_id[va] != m:
 *         for i in range(N):
 *             ss_u = 0             # <<<<<<<<<<<<<<
 *             ss_d = 0
 *             ss_a = 0
 */
      __pyx_v_ss_u = 0.0;

      /* "cyext_acv.pyx":2396
 *         for i in range(N):
 *             ss_u = 0
 *             ss_d = 0             # <<<<<<<<<<<<<<
 *             ss_a = 0
 *             ss_u_m = 0
 */
      __pyx_v_ss_d = 0.0;

      /* "cyext_acv.pyx":2397
 *             ss_u = 0
 *             ss_d = 0
 *             ss_a = 0             # <<<<<<<<<<<<<<
 *             ss_u_m = 0
 *             ss_d_m = 0
 */
      __pyx_v_ss_a = 0.0;

      /* "cyext_acv.pyx":2398
 *             ss_d = 0
 *             ss_a = 0
 *             ss_u_m = 0             # <<<<<<<<<<<<<<
 *             ss_d_m = 0
 *             ss_a_m = 0
 */
      __pyx_v_ss_u_m = 0.0;

      /* "cyext_acv.pyx":2399
 *             ss_a = 0
 *             ss_u_m = 0
 *             ss_d_m = 0             # <<<<<<<<<<<<<<
 *             ss_a_m = 0
 *             for b in range(n_trees):
 */
      __pyx_v_ss_d_m = 0.0;

      /* "cyext_acv.pyx":2400
 *             ss_u_m = 0
 *             ss_d_m = 0
 *             ss_a_m = 0             # <<<<<<<<<<<<<<
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 */
      __pyx_v_ss_a_m = 0.0;

      /* "cyext_acv.pyx":2401
 *             ss_d_m = 0
 *             ss_a_m = 0
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0
 */
      __pyx_t_10 = __pyx_v_n_trees;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_11; __pyx_t_18+=1) {
        __pyx_v_b = __pyx_t_18;

        /* "cyext_acv.pyx":2402
 *             ss_a_m = 0
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 n = 0
 *                 n_u = 0
 */
        __pyx_t_24 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_24 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":2403
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0             # <<<<<<<<<<<<<<
 *                 n_u = 0
 *                 n_d = 0
 */
        __pyx_v_n = 0.0;

        /* "cyext_acv.pyx":2404
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0
 *                 n_u = 0             # <<<<<<<<<<<<<<
 *                 n_d = 0
 *                 n_m = 0
 */
        __pyx_v_n_u = 0.0;

        /* "cyext_acv.pyx":2405
 *                 n = 0
 *                 n_u = 0
 *                 n_d = 0             # <<<<<<<<<<<<<<
 *                 n_m = 0
 *                 n_u_m = 0
 */
        __pyx_v_n_d = 0.0;

        /* "cyext_acv.pyx":2406
 *                 n_u = 0
 *                 n_d = 0
 *                 n_m = 0             # <<<<<<<<<<<<<<
 *                 n_u_m = 0
 *                 n_d_m = 0
 */
        __pyx_v_n_m = 0.0;

        /* "cyext_acv.pyx":2407
 *                 n_d = 0
 *                 n_m = 0
 *                 n_u_m = 0             # <<<<<<<<<<<<<<
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_v_n_u_m = 0.0;

        /* "cyext_acv.pyx":2408
 *                 n_m = 0
 *                 n_u_m = 0
 *                 n_d_m = 0             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]
 */
        __pyx_v_n_d_m = 0.0;

        /* "cyext_acv.pyx":2409
 *                 n_u_m = 0
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]
 */
        __pyx_t_17 = __pyx_v_nb_leaf;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2410
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_u += p_u_n[b, leaf_numb, i]
 *                     n_d += p_d_n[b, leaf_numb, i]
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_24 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_n.strides[2]) ))));

          /* "cyext_acv.pyx":2411
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_d += p_d_n[b, leaf_numb, i]
 * 
 */
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_24 = __pyx_v_i;
          __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_13 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[2]) ))));

          /* "cyext_acv.pyx":2412
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]
 *                     n_d += p_d_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 * 
 *                     n_m += p_n_m[b, leaf_numb, i]
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_24 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_d_n.strides[2]) ))));

          /* "cyext_acv.pyx":2414
 *                     n_d += p_d_n[b, leaf_numb, i]
 * 
 *                     n_m += p_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 */
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_24 = __pyx_v_i;
          __pyx_v_n_m = (__pyx_v_n_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n_m.data + __pyx_t_13 * __pyx_v_p_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n_m.strides[1]) ) + __pyx_t_24 * __pyx_v_p_n_m.strides[2]) ))));

          /* "cyext_acv.pyx":2415
 * 
 *                     n_m += p_n_m[b, leaf_numb, i]
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 * 
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n_u_m = (__pyx_v_n_u_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n_m.data + __pyx_t_24 * __pyx_v_p_u_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n_m.strides[1]) ) + __pyx_t_13 * __pyx_v_p_u_n_m.strides[2]) ))));

          /* "cyext_acv.pyx":2416
 *                     n_m += p_n_m[b, leaf_numb, i]
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(nb_leaf):
 */
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_24 = __pyx_v_i;
          __pyx_v_n_d_m = (__pyx_v_n_d_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n_m.data + __pyx_t_13 * __pyx_v_p_d_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n_m.strides[1]) ) + __pyx_t_24 * __pyx_v_p_d_n_m.strides[2]) ))));
        }

        /* "cyext_acv.pyx":2418
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 * 
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 */
        __pyx_t_17 = __pyx_v_nb_leaf;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2419
 * 
 *                 for leaf_numb in range(nb_leaf):
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 */
          if (((__pyx_v_n != 0.0) != 0)) {
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_30 = 0;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_24 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_41);

          /* "cyext_acv.pyx":2420
 *                 for leaf_numb in range(nb_leaf):
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 * 
 */
          if (((__pyx_v_n_u != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 1;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n_u);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_41);

          /* "cyext_acv.pyx":2421
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 */
          if (((__pyx_v_n_d != 0.0) != 0)) {
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_30 = 2;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_24 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n_d);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_41);

          /* "cyext_acv.pyx":2423
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 * 
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 */
          if (((__pyx_v_n != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 0;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_13 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_a_m = (__pyx_v_ss_a_m + __pyx_t_41);

          /* "cyext_acv.pyx":2424
 * 
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 * 
 */
          if (((__pyx_v_n_u != 0.0) != 0)) {
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_30 = 1;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_24 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_u_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_u_m = (__pyx_v_ss_u_m + __pyx_t_41);

          /* "cyext_acv.pyx":2425
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 */
          if (((__pyx_v_n_d != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 2;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_13 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_d_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_d_m = (__pyx_v_ss_d_m + __pyx_t_41);
        }
      }

      /* "cyext_acv.pyx":2427
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 * 
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0             # <<<<<<<<<<<<<<
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 */
      if ((((__pyx_v_ss_u_m - __pyx_v_ss_d_m) != 0.0) != 0)) {
        __pyx_t_41 = ((__pyx_v_ss_a_m - __pyx_v_ss_d_m) / (__pyx_v_ss_u_m - __pyx_v_ss_d_m));
      } else {
        __pyx_t_41 = 0.0;
      }
      __pyx_v_ss_m = __pyx_t_41;

      /* "cyext_acv.pyx":2428
 * 
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
        __pyx_t_41 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
      } else {
        __pyx_t_41 = 0.0;
      }
      __pyx_v_ss = __pyx_t_41;

      /* "cyext_acv.pyx":2431
 * 
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L63_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L63_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2432
 * 
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2431
 * 
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
        goto __pyx_L62;
      }

      /* "cyext_acv.pyx":2433
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L65_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L65_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2434
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 1.0;

        /* "cyext_acv.pyx":2433
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
        goto __pyx_L62;
      }

      /* "cyext_acv.pyx":2435
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L67_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L67_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2436
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2435
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
        goto __pyx_L62;
      }

      /* "cyext_acv.pyx":2437
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 *     elif S_size == 0:
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L69_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L69_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2438
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1             # <<<<<<<<<<<<<<
 *     elif S_size == 0:
 *         ss = 0
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = -1.0;

        /* "cyext_acv.pyx":2437
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 *     elif S_size == 0:
 */
      }
      __pyx_L62:;
    }

    /* "cyext_acv.pyx":2393
 * 
 * 
 *     if S_size !=0 and S_size + len_va_id[va] != m:             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             ss_u = 0
 */
    goto __pyx_L51;
  }

  /* "cyext_acv.pyx":2439
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         ss = 0
 *         for i in range(N):
 */
  __pyx_t_23 = ((__pyx_v_S_size == 0) != 0);
  if (__pyx_t_23) {

    /* "cyext_acv.pyx":2440
 *                 sdp[i] = -1
 *     elif S_size == 0:
 *         ss = 0             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             ss_u_m = 0
 */
    __pyx_v_ss = 0.0;

    /* "cyext_acv.pyx":2441
 *     elif S_size == 0:
 *         ss = 0
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             ss_u_m = 0
 *             ss_d_m = 0
 */
    __pyx_t_12 = __pyx_v_N;
    __pyx_t_16 = __pyx_t_12;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_16; __pyx_t_19+=1) {
      __pyx_v_i = __pyx_t_19;

      /* "cyext_acv.pyx":2442
 *         ss = 0
 *         for i in range(N):
 *             ss_u_m = 0             # <<<<<<<<<<<<<<
 *             ss_d_m = 0
 *             ss_a_m = 0
 */
      __pyx_v_ss_u_m = 0.0;

      /* "cyext_acv.pyx":2443
 *         for i in range(N):
 *             ss_u_m = 0
 *             ss_d_m = 0             # <<<<<<<<<<<<<<
 *             ss_a_m = 0
 *             for b in range(n_trees):
 */
      __pyx_v_ss_d_m = 0.0;

      /* "cyext_acv.pyx":2444
 *             ss_u_m = 0
 *             ss_d_m = 0
 *             ss_a_m = 0             # <<<<<<<<<<<<<<
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 */
      __pyx_v_ss_a_m = 0.0;

      /* "cyext_acv.pyx":2445
 *             ss_d_m = 0
 *             ss_a_m = 0
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 n_m = 0
 */
      __pyx_t_10 = __pyx_v_n_trees;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_11; __pyx_t_18+=1) {
        __pyx_v_b = __pyx_t_18;

        /* "cyext_acv.pyx":2446
 *             ss_a_m = 0
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 n_m = 0
 *                 n_u_m = 0
 */
        __pyx_t_24 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_24 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":2447
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 n_m = 0             # <<<<<<<<<<<<<<
 *                 n_u_m = 0
 *                 n_d_m = 0
 */
        __pyx_v_n_m = 0.0;

        /* "cyext_acv.pyx":2448
 *                 nb_leaf = leaves_nb[b]
 *                 n_m = 0
 *                 n_u_m = 0             # <<<<<<<<<<<<<<
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_v_n_u_m = 0.0;

        /* "cyext_acv.pyx":2449
 *                 n_m = 0
 *                 n_u_m = 0
 *                 n_d_m = 0             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     n_m += p_n_m[b, leaf_numb, i]
 */
        __pyx_v_n_d_m = 0.0;

        /* "cyext_acv.pyx":2450
 *                 n_u_m = 0
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     n_m += p_n_m[b, leaf_numb, i]
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]
 */
        __pyx_t_17 = __pyx_v_nb_leaf;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2451
 *                 n_d_m = 0
 *                 for leaf_numb in range(nb_leaf):
 *                     n_m += p_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n_m = (__pyx_v_n_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n_m.data + __pyx_t_24 * __pyx_v_p_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n_m.strides[1]) ) + __pyx_t_13 * __pyx_v_p_n_m.strides[2]) ))));

          /* "cyext_acv.pyx":2452
 *                 for leaf_numb in range(nb_leaf):
 *                     n_m += p_n_m[b, leaf_numb, i]
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 * 
 */
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_24 = __pyx_v_i;
          __pyx_v_n_u_m = (__pyx_v_n_u_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n_m.data + __pyx_t_13 * __pyx_v_p_u_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n_m.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n_m.strides[2]) ))));

          /* "cyext_acv.pyx":2453
 *                     n_m += p_n_m[b, leaf_numb, i]
 *                     n_u_m += p_u_n_m[b, leaf_numb, i]
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(max_leaves):
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n_d_m = (__pyx_v_n_d_m + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n_m.data + __pyx_t_24 * __pyx_v_p_d_n_m.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n_m.strides[1]) ) + __pyx_t_13 * __pyx_v_p_d_n_m.strides[2]) ))));
        }

        /* "cyext_acv.pyx":2455
 *                     n_d_m += p_d_n_m[b, leaf_numb, i]
 * 
 *                 for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 */
        __pyx_t_17 = __pyx_v_max_leaves;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2456
 * 
 *                 for leaf_numb in range(max_leaves):
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 */
          if (((__pyx_v_n != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 0;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_13 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_a_m = (__pyx_v_ss_a_m + __pyx_t_41);

          /* "cyext_acv.pyx":2457
 *                 for leaf_numb in range(max_leaves):
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 */
          if (((__pyx_v_n_u != 0.0) != 0)) {
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_30 = 1;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_24 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_u_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_u_m = (__pyx_v_ss_u_m + __pyx_t_41);

          /* "cyext_acv.pyx":2458
 *                     ss_a_m += mean_forest_m[b, leaf_numb, i, 0]/n_m if n !=0 else 0
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0             # <<<<<<<<<<<<<<
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 * 
 */
          if (((__pyx_v_n_d != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 2;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_m.data + __pyx_t_13 * __pyx_v_mean_forest_m.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest_m.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest_m.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest_m.strides[3]) ))) / __pyx_v_n_d_m);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_d_m = (__pyx_v_ss_d_m + __pyx_t_41);
        }
      }

      /* "cyext_acv.pyx":2459
 *                     ss_u_m += mean_forest_m[b, leaf_numb, i, 1]/n_u_m if n_u !=0 else 0
 *                     ss_d_m += mean_forest_m[b, leaf_numb, i, 2]/n_d_m if n_d !=0 else 0
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *             if (ss < thresholds and ss_m < thresholds):
 */
      if ((((__pyx_v_ss_u_m - __pyx_v_ss_d_m) != 0.0) != 0)) {
        __pyx_t_41 = ((__pyx_v_ss_a_m - __pyx_v_ss_d_m) / (__pyx_v_ss_u_m - __pyx_v_ss_d_m));
      } else {
        __pyx_t_41 = 0.0;
      }
      __pyx_v_ss_m = __pyx_t_41;

      /* "cyext_acv.pyx":2461
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L80_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L80_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2462
 * 
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2461
 *             ss_m = (ss_a_m - ss_d_m)/(ss_u_m - ss_d_m) if ss_u_m - ss_d_m !=0 else 0
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
        goto __pyx_L79;
      }

      /* "cyext_acv.pyx":2463
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L82_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L82_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2464
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 1.0;

        /* "cyext_acv.pyx":2463
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
        goto __pyx_L79;
      }

      /* "cyext_acv.pyx":2465
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L84_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L84_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2466
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2465
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
        goto __pyx_L79;
      }

      /* "cyext_acv.pyx":2467
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 * 
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L86_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L86_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2468
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     elif S_size + len_va_id[va] == m:
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = -1.0;

        /* "cyext_acv.pyx":2467
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 * 
 */
      }
      __pyx_L79:;
    }

    /* "cyext_acv.pyx":2439
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1
 *     elif S_size == 0:             # <<<<<<<<<<<<<<
 *         ss = 0
 *         for i in range(N):
 */
    goto __pyx_L51;
  }

  /* "cyext_acv.pyx":2470
 *                 sdp[i] = -1
 * 
 *     elif S_size + len_va_id[va] == m:             # <<<<<<<<<<<<<<
 *         ss_m = 1
 *         for i in range(N):
 */
  __pyx_t_30 = __pyx_v_va;
  __pyx_t_23 = (((__pyx_v_S_size + (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_30 * __pyx_v_len_va_id.strides[0]) )))) == __pyx_v_m) != 0);
  if (likely(__pyx_t_23)) {

    /* "cyext_acv.pyx":2471
 * 
 *     elif S_size + len_va_id[va] == m:
 *         ss_m = 1             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             ss_u = 0
 */
    __pyx_v_ss_m = 1.0;

    /* "cyext_acv.pyx":2472
 *     elif S_size + len_va_id[va] == m:
 *         ss_m = 1
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             ss_u = 0
 *             ss_d = 0
 */
    __pyx_t_12 = __pyx_v_N;
    __pyx_t_16 = __pyx_t_12;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_16; __pyx_t_19+=1) {
      __pyx_v_i = __pyx_t_19;

      /* "cyext_acv.pyx":2473
 *         ss_m = 1
 *         for i in range(N):
 *             ss_u = 0             # <<<<<<<<<<<<<<
 *             ss_d = 0
 *             ss_a = 0
 */
      __pyx_v_ss_u = 0.0;

      /* "cyext_acv.pyx":2474
 *         for i in range(N):
 *             ss_u = 0
 *             ss_d = 0             # <<<<<<<<<<<<<<
 *             ss_a = 0
 *             for b in range(n_trees):
 */
      __pyx_v_ss_d = 0.0;

      /* "cyext_acv.pyx":2475
 *             ss_u = 0
 *             ss_d = 0
 *             ss_a = 0             # <<<<<<<<<<<<<<
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 */
      __pyx_v_ss_a = 0.0;

      /* "cyext_acv.pyx":2476
 *             ss_d = 0
 *             ss_a = 0
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0
 */
      __pyx_t_10 = __pyx_v_n_trees;
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_11; __pyx_t_18+=1) {
        __pyx_v_b = __pyx_t_18;

        /* "cyext_acv.pyx":2477
 *             ss_a = 0
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 n = 0
 *                 n_u = 0
 */
        __pyx_t_24 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_24 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":2478
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0             # <<<<<<<<<<<<<<
 *                 n_u = 0
 *                 n_d = 0
 */
        __pyx_v_n = 0.0;

        /* "cyext_acv.pyx":2479
 *                 nb_leaf = leaves_nb[b]
 *                 n = 0
 *                 n_u = 0             # <<<<<<<<<<<<<<
 *                 n_d = 0
 *                 for leaf_numb in range(nb_leaf):
 */
        __pyx_v_n_u = 0.0;

        /* "cyext_acv.pyx":2480
 *                 n = 0
 *                 n_u = 0
 *                 n_d = 0             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]
 */
        __pyx_v_n_d = 0.0;

        /* "cyext_acv.pyx":2481
 *                 n_u = 0
 *                 n_d = 0
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]
 */
        __pyx_t_17 = __pyx_v_nb_leaf;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2482
 *                 n_d = 0
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_u += p_u_n[b, leaf_numb, i]
 *                     n_d += p_d_n[b, leaf_numb, i]
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_24 * __pyx_v_p_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_n.strides[2]) ))));

          /* "cyext_acv.pyx":2483
 *                 for leaf_numb in range(nb_leaf):
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 *                     n_d += p_d_n[b, leaf_numb, i]
 * 
 */
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_24 = __pyx_v_i;
          __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_13 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_24 * __pyx_v_p_u_n.strides[2]) ))));

          /* "cyext_acv.pyx":2484
 *                     n += p_n[b, leaf_numb, i]
 *                     n_u += p_u_n[b, leaf_numb, i]
 *                     n_d += p_d_n[b, leaf_numb, i]             # <<<<<<<<<<<<<<
 * 
 *                 for leaf_numb in range(max_leaves):
 */
          __pyx_t_24 = __pyx_v_b;
          __pyx_t_27 = __pyx_v_leaf_numb;
          __pyx_t_13 = __pyx_v_i;
          __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_24 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_27 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_13 * __pyx_v_p_d_n.strides[2]) ))));
        }

        /* "cyext_acv.pyx":2486
 *                     n_d += p_d_n[b, leaf_numb, i]
 * 
 *                 for leaf_numb in range(max_leaves):             # <<<<<<<<<<<<<<
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 */
        __pyx_t_17 = __pyx_v_max_leaves;
        __pyx_t_22 = __pyx_t_17;
        for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_22; __pyx_t_38+=1) {
          __pyx_v_leaf_numb = __pyx_t_38;

          /* "cyext_acv.pyx":2487
 * 
 *                 for leaf_numb in range(max_leaves):
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 */
          if (((__pyx_v_n != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 0;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_41);

          /* "cyext_acv.pyx":2488
 *                 for leaf_numb in range(max_leaves):
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
          if (((__pyx_v_n_u != 0.0) != 0)) {
            __pyx_t_24 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_13 = __pyx_v_i;
            __pyx_t_30 = 1;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_24 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_13 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n_u);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_41);

          /* "cyext_acv.pyx":2489
 *                     ss_a += mean_forest[b, leaf_numb, i, 0]/n if n !=0 else 0
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0             # <<<<<<<<<<<<<<
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 */
          if (((__pyx_v_n_d != 0.0) != 0)) {
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_27 = __pyx_v_leaf_numb;
            __pyx_t_24 = __pyx_v_i;
            __pyx_t_30 = 2;
            __pyx_t_41 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_13 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_27 * __pyx_v_mean_forest.strides[1]) ) + __pyx_t_24 * __pyx_v_mean_forest.strides[2]) ) + __pyx_t_30 * __pyx_v_mean_forest.strides[3]) ))) / __pyx_v_n_d);
          } else {
            __pyx_t_41 = 0.0;
          }
          __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_41);
        }
      }

      /* "cyext_acv.pyx":2490
 *                     ss_u += mean_forest[b, leaf_numb, i, 1]/n_u if n_u !=0 else 0
 *                     ss_d += mean_forest[b, leaf_numb, i, 2]/n_d if n_d !=0 else 0
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *             if (ss < thresholds and ss_m < thresholds):
 */
      if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
        __pyx_t_41 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
      } else {
        __pyx_t_41 = 0.0;
      }
      __pyx_v_ss = __pyx_t_41;

      /* "cyext_acv.pyx":2492
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L97_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L97_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2493
 * 
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2492
 *             ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 * 
 *             if (ss < thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 */
        goto __pyx_L96;
      }

      /* "cyext_acv.pyx":2494
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss < __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L99_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L99_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2495
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 1.0;

        /* "cyext_acv.pyx":2494
 *             if (ss < thresholds and ss_m < thresholds):
 *                 sdp[i] = 0
 *             elif (ss < thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 */
        goto __pyx_L96;
      }

      /* "cyext_acv.pyx":2496
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L101_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m >= __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L101_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2497
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0             # <<<<<<<<<<<<<<
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = 0.0;

        /* "cyext_acv.pyx":2496
 *             elif (ss < thresholds and ss_m >= thresholds):
 *                 sdp[i] = 1
 *             elif (ss >= thresholds and ss_m >= thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 */
        goto __pyx_L96;
      }

      /* "cyext_acv.pyx":2498
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 *     else:
 */
      __pyx_t_31 = ((__pyx_v_ss >= __pyx_v_thresholds) != 0);
      if (__pyx_t_31) {
      } else {
        __pyx_t_23 = __pyx_t_31;
        goto __pyx_L103_bool_binop_done;
      }
      __pyx_t_31 = ((__pyx_v_ss_m < __pyx_v_thresholds) != 0);
      __pyx_t_23 = __pyx_t_31;
      __pyx_L103_bool_binop_done:;
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":2499
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):
 *                 sdp[i] = -1             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError('Check condition in the computation of SDP')
 */
        __pyx_t_24 = __pyx_v_i;
        *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_24 * __pyx_v_sdp.strides[0]) )) = -1.0;

        /* "cyext_acv.pyx":2498
 *             elif (ss >= thresholds and ss_m >= thresholds):
 *                 sdp[i] = 0
 *             elif (ss >= thresholds and ss_m < thresholds):             # <<<<<<<<<<<<<<
 *                 sdp[i] = -1
 *     else:
 */
      }
      __pyx_L96:;
    }

    /* "cyext_acv.pyx":2470
 *                 sdp[i] = -1
 * 
 *     elif S_size + len_va_id[va] == m:             # <<<<<<<<<<<<<<
 *         ss_m = 1
 *         for i in range(N):
 */
    goto __pyx_L51;
  }

  /* "cyext_acv.pyx":2501
 *                 sdp[i] = -1
 *     else:
 *         raise ValueError('Check condition in the computation of SDP')             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2501, __pyx_L1_error)
  }
  __pyx_L51:;

  /* "cyext_acv.pyx":2503
 *         raise ValueError('Check condition in the computation of SDP')
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2252
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef compute_sdp_swing_diff(const double[:, :] X, const long[:] fX,             # <<<<<<<<<<<<<<
 *             const long[:] y_pred, long[::1] S, long[:, :] va_id, long va, long[:] len_va_id, unsigned long S_size, const double[:, :] data,
 *             const double[:, :, :] values,const  double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_swing_diff", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_m, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n_m, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n_m, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n_m, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2510
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef swing_sv_clf_direct(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const long[:] fX,
 *     const long[:] y_pred,
 */

static PyObject *__pyx_pw_9cyext_acv_29swing_sv_clf_direct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_swing_sv_clf_direct(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double const __pyx_v_thresholds, int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  CYTHON_UNUSED unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_v_value = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_S_size;
  unsigned int __pyx_v_dif_pos;
  unsigned int __pyx_v_dif_neg;
  unsigned int __pyx_v_dif_null;
  unsigned int __pyx_v_va_size;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_set_size;
  std::vector<std::vector<int> >  __pyx_v_Sm;
  std::vector<std::vector<int> >  __pyx_v_va_id_cpp;
  double __pyx_v_weight;
  __Pyx_memviewslice __pyx_v_swings_prop = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_swings = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sm_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sm_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C_buff = NULL;
  PyObject *__pyx_v_coal_va = NULL;
  PyObject *__pyx_v_va_id = NULL;
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_c = NULL;
  __Pyx_memviewslice __pyx_v_va_id_a = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_va_id = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  unsigned int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  int __pyx_t_21;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  std::vector<std::vector<int> >  __pyx_t_24;
  size_t __pyx_t_25;
  long __pyx_t_26;
  long __pyx_t_27;
  unsigned int __pyx_t_28;
  size_t __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  unsigned int __pyx_t_32;
  unsigned int __pyx_t_33;
  unsigned int __pyx_t_34;
  unsigned int __pyx_t_35;
  size_t __pyx_t_36;
  size_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  Py_ssize_t __pyx_t_39;
  PyObject *__pyx_t_40 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("swing_sv_clf_direct", 0);

  /* "cyext_acv.pyx":2521
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":2522
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2523
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :] phi
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2526
 * 
 *     cdef double[:, :] phi
 *     phi = np.zeros((N, m))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] v_value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 2526, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":2529
 * 
 *     cdef double[:] v_value
 *     v_value = np.zeros((N))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v_value = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":2532
 * 
 *     cdef long[::1] S
 *     S = np.zeros((m), dtype=np.int)             # <<<<<<<<<<<<<<
 *     cdef unsigned int S_size, dif_pos, dif_neg, dif_null, va_size, pow_set_size
 *     cdef unsigned int counter, i, va, j, ci, cj, set_size, nv
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dc_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_S = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2540
 * 
 *     cdef long[:, :, :] swings_prop
 *     swings_prop = np.zeros((N, m, 3), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] swings,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_swings_prop = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2543
 * 
 *     cdef double[:, :, :] swings,
 *     swings = np.zeros((N, m, 2))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :] sm_buf
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_2);
  __pyx_t_5 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 2543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_swings = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":2546
 * 
 *     cdef long[:, :] sm_buf
 *     sm_buf = np.zeros((data.shape[1], data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] sm_size
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_sm_buf = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":2549
 * 
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 *     if C[0] != []:
 *         C_buff = C.copy()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sm_size = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":2550
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2550, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_13) {

    /* "cyext_acv.pyx":2551
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:
 *         C_buff = C.copy()             # <<<<<<<<<<<<<<
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []
 */
    __pyx_t_5 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_copy, __pyx_v_C); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_C_buff = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "cyext_acv.pyx":2552
 *     if C[0] != []:
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = []
 *         remove_va = []
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 2552, __pyx_L1_error)
    }
    __pyx_t_14 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2552, __pyx_L1_error)
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
      __pyx_v_ci = __pyx_t_16;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 2552, __pyx_L1_error)
      }
      __pyx_t_4 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_17 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_cj = __pyx_t_19;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 2552, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 2552, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __pyx_v_coal_va = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "cyext_acv.pyx":2553
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []             # <<<<<<<<<<<<<<
 *         remove_va = []
 * 
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_va_id = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "cyext_acv.pyx":2554
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = []
 *         remove_va = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(m):
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_remove_va = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "cyext_acv.pyx":2556
 *         remove_va = []
 * 
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if i not in coal_va:
 *                 remove_va.append([i])
 */
    __pyx_t_16 = __pyx_v_m;
    __pyx_t_19 = __pyx_t_16;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;

      /* "cyext_acv.pyx":2557
 * 
 *         for i in range(m):
 *             if i not in coal_va:             # <<<<<<<<<<<<<<
 *                 remove_va.append([i])
 *                 va_id.append([i])
 */
      __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_v_coal_va, Py_NE)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2557, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_21 = (__pyx_t_13 != 0);
      if (__pyx_t_21) {

        /* "cyext_acv.pyx":2558
 *         for i in range(m):
 *             if i not in coal_va:
 *                 remove_va.append([i])             # <<<<<<<<<<<<<<
 *                 va_id.append([i])
 *             else:
 */
        __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_5);
        PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_remove_va, __pyx_t_4); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2558, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "cyext_acv.pyx":2559
 *             if i not in coal_va:
 *                 remove_va.append([i])
 *                 va_id.append([i])             # <<<<<<<<<<<<<<
 *             else:
 *                 for c in C_buff:
 */
        __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4);
        PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_va_id, __pyx_t_5); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2559, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "cyext_acv.pyx":2557
 * 
 *         for i in range(m):
 *             if i not in coal_va:             # <<<<<<<<<<<<<<
 *                 remove_va.append([i])
 *                 va_id.append([i])
 */
        goto __pyx_L10;
      }

      /* "cyext_acv.pyx":2561
 *                 va_id.append([i])
 *             else:
 *                 for c in C_buff:             # <<<<<<<<<<<<<<
 *                     if i in c:
 *                         va_id.append(c)
 */
      /*else*/ {
        if (likely(PyList_CheckExact(__pyx_v_C_buff)) || PyTuple_CheckExact(__pyx_v_C_buff)) {
          __pyx_t_5 = __pyx_v_C_buff; __Pyx_INCREF(__pyx_t_5); __pyx_t_14 = 0;
          __pyx_t_23 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_C_buff); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2561, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_23 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 2561, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_23)) {
            if (likely(PyList_CheckExact(__pyx_t_5))) {
              if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2561, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2561, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 2561, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2561, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_23(__pyx_t_5);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 2561, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "cyext_acv.pyx":2562
 *             else:
 *                 for c in C_buff:
 *                     if i in c:             # <<<<<<<<<<<<<<
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 */
          __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2562, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_21 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_v_c, Py_EQ)); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 2562, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_13 = (__pyx_t_21 != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":2563
 *                 for c in C_buff:
 *                     if i in c:
 *                         va_id.append(c)             # <<<<<<<<<<<<<<
 *                         C_buff.remove(c)
 *                         continue
 */
            __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_va_id, __pyx_v_c); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 2563, __pyx_L1_error)

            /* "cyext_acv.pyx":2564
 *                     if i in c:
 *                         va_id.append(c)
 *                         C_buff.remove(c)             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_C_buff, __pyx_n_s_remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2564, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_c);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2564, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "cyext_acv.pyx":2565
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L11_continue;

            /* "cyext_acv.pyx":2562
 *             else:
 *                 for c in C_buff:
 *                     if i in c:             # <<<<<<<<<<<<<<
 *                         va_id.append(c)
 *                         C_buff.remove(c)
 */
          }

          /* "cyext_acv.pyx":2561
 *                 va_id.append([i])
 *             else:
 *                 for c in C_buff:             # <<<<<<<<<<<<<<
 *                     if i in c:
 *                         va_id.append(c)
 */
          __pyx_L11_continue:;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_L10:;
    }

    /* "cyext_acv.pyx":2550
 *     cdef long[:] sm_size
 *     sm_size = np.zeros((data.shape[1]), dtype=np.int)
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         C_buff = C.copy()
 *         coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 */
    goto __pyx_L3;
  }

  /* "cyext_acv.pyx":2568
 * 
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_16 = __pyx_v_m;
    __pyx_t_19 = __pyx_t_16;
    for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
      __pyx_v_i = __pyx_t_20;
      __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2568, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
  }
  __pyx_L3:;

  /* "cyext_acv.pyx":2570
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     va_id_cpp = va_id
 * 
 */
  __pyx_t_14 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2570, __pyx_L1_error)
  __pyx_v_m = __pyx_t_14;

  /* "cyext_acv.pyx":2571
 * 
 *     m = len(va_id)
 *     va_id_cpp = va_id             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :] va_id_a
 */
  __pyx_t_24 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(__pyx_v_va_id); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2571, __pyx_L1_error)
  __pyx_v_va_id_cpp = __pyx_t_24;

  /* "cyext_acv.pyx":2576
 *     cdef long[:] len_va_id
 * 
 *     va_id_a = np.empty((data.shape[1], data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 2576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_va_id_a = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":2577
 * 
 *     va_id_a = np.empty((data.shape[1], data.shape[1]), dtype=np.int)
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(m):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 2577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_len_va_id = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":2579
 *     len_va_id = np.empty((data.shape[1]), dtype=np.int)
 * 
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):
 */
  __pyx_t_16 = __pyx_v_m;
  __pyx_t_19 = __pyx_t_16;
  for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
    __pyx_v_i = __pyx_t_20;

    /* "cyext_acv.pyx":2580
 * 
 *     for i in range(m):
 *         len_va_id[i] = va_id_cpp[i].size()             # <<<<<<<<<<<<<<
 *         for j in range(len_va_id[i]):
 *             va_id_a[i, j] = va_id_cpp[i][j]
 */
    __pyx_t_25 = __pyx_v_i;
    *((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )) = (__pyx_v_va_id_cpp[__pyx_v_i]).size();

    /* "cyext_acv.pyx":2581
 *     for i in range(m):
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):             # <<<<<<<<<<<<<<
 *             va_id_a[i, j] = va_id_cpp[i][j]
 * 
 */
    __pyx_t_25 = __pyx_v_i;
    __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_25 * __pyx_v_len_va_id.strides[0]) )));
    __pyx_t_27 = __pyx_t_26;
    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
      __pyx_v_j = __pyx_t_28;

      /* "cyext_acv.pyx":2582
 *         len_va_id[i] = va_id_cpp[i].size()
 *         for j in range(len_va_id[i]):
 *             va_id_a[i, j] = va_id_cpp[i][j]             # <<<<<<<<<<<<<<
 * 
 *     set_size = m - 1
 */
      __pyx_t_25 = __pyx_v_i;
      __pyx_t_29 = __pyx_v_j;
      *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_29 * __pyx_v_va_id_a.strides[1]) )) = ((__pyx_v_va_id_cpp[__pyx_v_i])[__pyx_v_j]);
    }
  }

  /* "cyext_acv.pyx":2584
 *             va_id_a[i, j] = va_id_cpp[i][j]
 * 
 *     set_size = m - 1             # <<<<<<<<<<<<<<
 *     pow_set_size = 2**set_size
 * 
 */
  __pyx_v_set_size = (__pyx_v_m - 1);

  /* "cyext_acv.pyx":2585
 * 
 *     set_size = m - 1
 *     pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *     for va in range(m):
 */
  __pyx_v_pow_set_size = __Pyx_pow_long(2, ((long)__pyx_v_set_size));

  /* "cyext_acv.pyx":2587
 *     pow_set_size = 2**set_size
 * 
 *     for va in range(m):             # <<<<<<<<<<<<<<
 * #         Sm = va_id.copy()
 * #         Sm.remove(va_id[va])
 */
  __pyx_t_16 = __pyx_v_m;
  __pyx_t_19 = __pyx_t_16;
  for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
    __pyx_v_va = __pyx_t_20;

    /* "cyext_acv.pyx":2590
 * #         Sm = va_id.copy()
 * #         Sm.remove(va_id[va])
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())             # <<<<<<<<<<<<<<
 *         Sm.erase(Sm.begin() + va)
 * 
 */
    try {
      __pyx_v_Sm.assign(__pyx_v_va_id_cpp.begin(), __pyx_v_va_id_cpp.end());
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 2590, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":2591
 * #         Sm.remove(va_id[va])
 *         Sm.assign(va_id_cpp.begin(), va_id_cpp.end())
 *         Sm.erase(Sm.begin() + va)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(set_size):
 */
    (void)(__pyx_v_Sm.erase((__pyx_v_Sm.begin() + __pyx_v_va)));

    /* "cyext_acv.pyx":2593
 *         Sm.erase(Sm.begin() + va)
 * 
 *         for i in range(set_size):             # <<<<<<<<<<<<<<
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):
 */
    __pyx_t_28 = __pyx_v_set_size;
    __pyx_t_30 = __pyx_t_28;
    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
      __pyx_v_i = __pyx_t_31;

      /* "cyext_acv.pyx":2594
 * 
 *         for i in range(set_size):
 *             sm_size[i] = Sm[i].size()             # <<<<<<<<<<<<<<
 *             for j in range(sm_size[i]):
 *                 sm_buf[i, j] = Sm[i][j]
 */
      __pyx_t_29 = __pyx_v_i;
      *((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_29 * __pyx_v_sm_size.strides[0]) )) = (__pyx_v_Sm[__pyx_v_i]).size();

      /* "cyext_acv.pyx":2595
 *         for i in range(set_size):
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):             # <<<<<<<<<<<<<<
 *                 sm_buf[i, j] = Sm[i][j]
 * 
 */
      __pyx_t_29 = __pyx_v_i;
      __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_29 * __pyx_v_sm_size.strides[0]) )));
      __pyx_t_27 = __pyx_t_26;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_27; __pyx_t_32+=1) {
        __pyx_v_j = __pyx_t_32;

        /* "cyext_acv.pyx":2596
 *             sm_size[i] = Sm[i].size()
 *             for j in range(sm_size[i]):
 *                 sm_buf[i, j] = Sm[i][j]             # <<<<<<<<<<<<<<
 * 
 *         for counter in range(0, pow_set_size):
 */
        __pyx_t_29 = __pyx_v_i;
        __pyx_t_25 = __pyx_v_j;
        *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sm_buf.data + __pyx_t_29 * __pyx_v_sm_buf.strides[0]) ) + __pyx_t_25 * __pyx_v_sm_buf.strides[1]) )) = ((__pyx_v_Sm[__pyx_v_i])[__pyx_v_j]);
      }
    }

    /* "cyext_acv.pyx":2598
 *                 sm_buf[i, j] = Sm[i][j]
 * 
 *         for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *             va_size = 0
 *             S_size = 0
 */
    __pyx_t_28 = __pyx_v_pow_set_size;
    __pyx_t_30 = __pyx_t_28;
    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
      __pyx_v_counter = __pyx_t_31;

      /* "cyext_acv.pyx":2599
 * 
 *         for counter in range(0, pow_set_size):
 *             va_size = 0             # <<<<<<<<<<<<<<
 *             S_size = 0
 *             for ci in range(set_size):
 */
      __pyx_v_va_size = 0;

      /* "cyext_acv.pyx":2600
 *         for counter in range(0, pow_set_size):
 *             va_size = 0
 *             S_size = 0             # <<<<<<<<<<<<<<
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):
 */
      __pyx_v_S_size = 0;

      /* "cyext_acv.pyx":2601
 *             va_size = 0
 *             S_size = 0
 *             for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):
 */
      __pyx_t_32 = __pyx_v_set_size;
      __pyx_t_33 = __pyx_t_32;
      for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
        __pyx_v_ci = __pyx_t_34;

        /* "cyext_acv.pyx":2602
 *             S_size = 0
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 */
        __pyx_t_13 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":2603
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):             # <<<<<<<<<<<<<<
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1
 */
          __pyx_t_25 = __pyx_v_ci;
          __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_sm_size.data + __pyx_t_25 * __pyx_v_sm_size.strides[0]) )));
          __pyx_t_27 = __pyx_t_26;
          for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_27; __pyx_t_35+=1) {
            __pyx_v_cj = __pyx_t_35;

            /* "cyext_acv.pyx":2604
 *                 if((counter & (1 << ci)) > 0):
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]             # <<<<<<<<<<<<<<
 *                         S_size = S_size + 1
 *                     va_size = va_size + 1
 */
            __pyx_t_25 = __pyx_v_ci;
            __pyx_t_29 = __pyx_v_cj;
            __pyx_t_36 = __pyx_v_S_size;
            *((long *) ( /* dim=0 */ ((char *) (((long *) __pyx_v_S.data) + __pyx_t_36)) )) = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sm_buf.data + __pyx_t_25 * __pyx_v_sm_buf.strides[0]) ) + __pyx_t_29 * __pyx_v_sm_buf.strides[1]) )));

            /* "cyext_acv.pyx":2605
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                     va_size = va_size + 1
 * 
 */
            __pyx_v_S_size = (__pyx_v_S_size + 1);
          }

          /* "cyext_acv.pyx":2606
 *                         S[S_size] = sm_buf[ci, cj]
 *                         S_size = S_size + 1
 *                     va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *             weight = 1./binomialC(m - 1, va_size)
 */
          __pyx_v_va_size = (__pyx_v_va_size + 1);

          /* "cyext_acv.pyx":2602
 *             S_size = 0
 *             for ci in range(set_size):
 *                 if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                     for cj in range(sm_size[ci]):
 *                         S[S_size] = sm_buf[ci, cj]
 */
        }
      }

      /* "cyext_acv.pyx":2608
 *                     va_size = va_size + 1
 * 
 *             weight = 1./binomialC(m - 1, va_size)             # <<<<<<<<<<<<<<
 *             v_value = compute_sdp_swing_diff(X, fX, y_pred, S, va_id_a, va, len_va_id, S_size, data, values,
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 */
      __pyx_v_weight = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));

      /* "cyext_acv.pyx":2609
 * 
 *             weight = 1./binomialC(m - 1, va_size)
 *             v_value = compute_sdp_swing_diff(X, fX, y_pred, S, va_id_a, va, len_va_id, S_size, data, values,             # <<<<<<<<<<<<<<
 *                       partition_leaves_trees, leaf_idx_trees, leaves_nb, scaling,
 *                       thresholds, num_threads)
 */
      __pyx_t_5 = __pyx_f_9cyext_acv_compute_sdp_swing_diff(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_S, __pyx_v_va_id_a, __pyx_v_va, __pyx_v_len_va_id, __pyx_v_S_size, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_thresholds, __pyx_v_num_threads); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2609, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 2609, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_v_value, 1);
      __pyx_v_v_value = __pyx_t_7;
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;

      /* "cyext_acv.pyx":2614
 * 
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 dif_pos = 1 if v_value[i] > 0 else 0
 *                 dif_neg = 1 if v_value[i] < 0 else 0
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {
            __pyx_t_32 = __pyx_v_N;
            if ((1 == 0)) abort();
            {
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   (x)
                    #define unlikely(x) (x)
                #endif
                __pyx_t_34 = (__pyx_t_32 - 0 + 1 - 1/abs(1)) / 1;
                if (__pyx_t_34 > 0)
                {
                    #ifdef _OPENMP
                    #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_29, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39)
                    #endif /* _OPENMP */
                    {
                        #ifdef _OPENMP
                        #pragma omp for lastprivate(__pyx_v_dif_neg) lastprivate(__pyx_v_dif_null) lastprivate(__pyx_v_dif_pos) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j)
                        #endif /* _OPENMP */
                        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_34; __pyx_t_33++){
                            {
                                __pyx_v_i = (unsigned int)(0 + 1 * __pyx_t_33);
                                /* Initialize private variables to invalid values */
                                __pyx_v_dif_neg = ((unsigned int)0xbad0bad0);
                                __pyx_v_dif_null = ((unsigned int)0xbad0bad0);
                                __pyx_v_dif_pos = ((unsigned int)0xbad0bad0);
                                __pyx_v_j = ((unsigned int)0xbad0bad0);

                                /* "cyext_acv.pyx":2615
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 dif_pos = 1 if v_value[i] > 0 else 0             # <<<<<<<<<<<<<<
 *                 dif_neg = 1 if v_value[i] < 0 else 0
 *                 dif_null = 1 if v_value[i] == 0 else 0
 */
                                __pyx_t_29 = __pyx_v_i;
                                if ((((*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) ))) > 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_pos = __pyx_t_35;

                                /* "cyext_acv.pyx":2616
 *             for i in prange(N, nogil=True, num_threads=num_threads):
 *                 dif_pos = 1 if v_value[i] > 0 else 0
 *                 dif_neg = 1 if v_value[i] < 0 else 0             # <<<<<<<<<<<<<<
 *                 dif_null = 1 if v_value[i] == 0 else 0
 * 
 */
                                __pyx_t_29 = __pyx_v_i;
                                if ((((*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) ))) < 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_neg = __pyx_t_35;

                                /* "cyext_acv.pyx":2617
 *                 dif_pos = 1 if v_value[i] > 0 else 0
 *                 dif_neg = 1 if v_value[i] < 0 else 0
 *                 dif_null = 1 if v_value[i] == 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 for j in range(len_va_id[va]):
 */
                                __pyx_t_29 = __pyx_v_i;
                                if ((((*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) ))) == 0.0) != 0)) {
                                  __pyx_t_35 = 1;
                                } else {
                                  __pyx_t_35 = 0;
                                }
                                __pyx_v_dif_null = __pyx_t_35;

                                /* "cyext_acv.pyx":2619
 *                 dif_null = 1 if v_value[i] == 0 else 0
 * 
 *                 for j in range(len_va_id[va]):             # <<<<<<<<<<<<<<
 *                     phi[i, va_id_a[va, j]] += weight * v_value[i]
 * 
 */
                                __pyx_t_29 = __pyx_v_va;
                                __pyx_t_26 = (*((long *) ( /* dim=0 */ (__pyx_v_len_va_id.data + __pyx_t_29 * __pyx_v_len_va_id.strides[0]) )));
                                __pyx_t_27 = __pyx_t_26;
                                for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_27; __pyx_t_35+=1) {
                                  __pyx_v_j = __pyx_t_35;

                                  /* "cyext_acv.pyx":2620
 * 
 *                 for j in range(len_va_id[va]):
 *                     phi[i, va_id_a[va, j]] += weight * v_value[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_va;
                                  __pyx_t_36 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_38 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_36 * __pyx_v_va_id_a.strides[1]) )));
                                  *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_37 * __pyx_v_phi.strides[0]) ) + __pyx_t_38 * __pyx_v_phi.strides[1]) )) += (__pyx_v_weight * (*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) ))));

                                  /* "cyext_acv.pyx":2623
 * 
 * 
 *                     swings[i, va_id_a[va, j], 0] += (dif_pos * (v_value[i]) * weight) / m             # <<<<<<<<<<<<<<
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_value[i]) * weight) / m
 * 
 */
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_36 = __pyx_v_va;
                                  __pyx_t_25 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_38 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_36 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_25 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_39 = 0;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings.data + __pyx_t_37 * __pyx_v_swings.strides[0]) ) + __pyx_t_38 * __pyx_v_swings.strides[1]) ) + __pyx_t_39 * __pyx_v_swings.strides[2]) )) += (((__pyx_v_dif_pos * (*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) )))) * __pyx_v_weight) / __pyx_v_m);

                                  /* "cyext_acv.pyx":2624
 * 
 *                     swings[i, va_id_a[va, j], 0] += (dif_pos * (v_value[i]) * weight) / m
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_value[i]) * weight) / m             # <<<<<<<<<<<<<<
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 */
                                  __pyx_t_29 = __pyx_v_i;
                                  __pyx_t_25 = __pyx_v_va;
                                  __pyx_t_36 = __pyx_v_j;
                                  __pyx_t_37 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_25 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_36 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_38 = 1;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings.data + __pyx_t_37 * __pyx_v_swings.strides[0]) ) + __pyx_t_39 * __pyx_v_swings.strides[1]) ) + __pyx_t_38 * __pyx_v_swings.strides[2]) )) += (((__pyx_v_dif_neg * (*((double *) ( /* dim=0 */ (__pyx_v_v_value.data + __pyx_t_29 * __pyx_v_v_value.strides[0]) )))) * __pyx_v_weight) / __pyx_v_m);

                                  /* "cyext_acv.pyx":2626
 *                     swings[i, va_id_a[va, j], 1] += (dif_neg * (v_value[i]) * weight) / m
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos             # <<<<<<<<<<<<<<
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 */
                                  __pyx_t_29 = __pyx_v_va;
                                  __pyx_t_36 = __pyx_v_j;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_38 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_29 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_36 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_39 = 0;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_25 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_38 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_pos;

                                  /* "cyext_acv.pyx":2627
 * 
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg             # <<<<<<<<<<<<<<
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 * 
 */
                                  __pyx_t_36 = __pyx_v_va;
                                  __pyx_t_29 = __pyx_v_j;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_39 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_36 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_29 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_38 = 1;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_25 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_38 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_neg;

                                  /* "cyext_acv.pyx":2628
 *                     swings_prop[i, va_id_a[va, j], 0] += dif_pos
 *                     swings_prop[i, va_id_a[va, j], 1] += dif_neg
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null             # <<<<<<<<<<<<<<
 * 
 *     return np.array(phi)/m, np.array(swings), np.array(swings_prop)
 */
                                  __pyx_t_29 = __pyx_v_va;
                                  __pyx_t_36 = __pyx_v_j;
                                  __pyx_t_25 = __pyx_v_i;
                                  __pyx_t_38 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_va_id_a.data + __pyx_t_29 * __pyx_v_va_id_a.strides[0]) ) + __pyx_t_36 * __pyx_v_va_id_a.strides[1]) )));
                                  __pyx_t_39 = 2;
                                  *((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_swings_prop.data + __pyx_t_25 * __pyx_v_swings_prop.strides[0]) ) + __pyx_t_38 * __pyx_v_swings_prop.strides[1]) ) + __pyx_t_39 * __pyx_v_swings_prop.strides[2]) )) += __pyx_v_dif_null;
                                }
                            }
                        }
                    }
                }
            }
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   __builtin_expect(!!(x), 1)
                #define unlikely(x) __builtin_expect(!!(x), 0)
            #endif
          }

          /* "cyext_acv.pyx":2614
 * 
 * 
 *             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                 dif_pos = 1 if v_value[i] > 0 else 0
 *                 dif_neg = 1 if v_value[i] < 0 else 0
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L37;
            }
            __pyx_L37:;
          }
      }
    }
  }

  /* "cyext_acv.pyx":2630
 *                     swings_prop[i, va_id_a[va, j], 2] += dif_null
 * 
 *     return np.array(phi)/m, np.array(swings), np.array(swings_prop)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_phi, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_swings, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_swings_prop, 3, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_40 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_40 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_40)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_40);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_40) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_40, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_40); __pyx_t_40 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2510
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef swing_sv_clf_direct(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const long[:] fX,
 *     const long[:] y_pred,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf_direct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_v_value, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_swings_prop, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_swings, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sm_buf, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sm_size, 1);
  __Pyx_XDECREF(__pyx_v_C_buff);
  __Pyx_XDECREF(__pyx_v_coal_va);
  __Pyx_XDECREF(__pyx_v_va_id);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_c);
  __PYX_XDEC_MEMVIEW(&__pyx_v_va_id_a, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_va_id, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_29swing_sv_clf_direct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_29swing_sv_clf_direct(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_thresholds;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("swing_sv_clf_direct (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_thresholds,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 1); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 2); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 3); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 4); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 5); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 6); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 7); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 8); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 9); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 10); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, 11); __PYX_ERR(0, 2510, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "swing_sv_clf_direct") < 0)) __PYX_ERR(0, 2510, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 2510, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[1], 0); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 2511, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[2], 0); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 2512, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 2513, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[4], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 2514, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[5], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 2515, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[6], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 2516, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[7], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 2517, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(values[8], 0); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 2518, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[9]);
    __pyx_v_thresholds = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_thresholds == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2518, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2518, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("swing_sv_clf_direct", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2510, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf_direct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 2518, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_28swing_sv_clf_direct(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_thresholds, __pyx_v_num_threads);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_28swing_sv_clf_direct(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_thresholds, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("swing_sv_clf_direct", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 2510, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_swing_sv_clf_direct(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_thresholds, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.swing_sv_clf_direct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2637
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg_cpp_pa_coal(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_31global_sdp_reg_cpp_pa_coal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_reg_cpp_pa_coal(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, CYTHON_UNUSED int __pyx_v_num_threads, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  int __pyx_v_n_trees;
  __Pyx_memviewslice __pyx_v_leaves_tree = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_tree_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_part_l = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_value;
  double __pyx_v_value_l;
  __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_nb_leaf_l;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  double __pyx_v_ss;
  int __pyx_v_b;
  int __pyx_v_l;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_max_size;
  int __pyx_v_size;
  int __pyx_v_leaf_numb_l;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_9;
  unsigned int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  long __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_26;
  __Pyx_memviewslice __pyx_t_27 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_28;
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<long> ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  int __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  int __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  int __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  __Pyx_memviewslice __pyx_t_41 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  int __pyx_t_45;
  int __pyx_t_46;
  int __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  unsigned int __pyx_t_49;
  unsigned int __pyx_t_50;
  unsigned int __pyx_t_51;
  Py_ssize_t __pyx_t_52;
  double __pyx_t_53;
  int __pyx_t_54;
  unsigned int __pyx_t_55;
  unsigned int __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  Py_ssize_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  Py_ssize_t __pyx_t_64;
  Py_ssize_t __pyx_t_65;
  Py_ssize_t __pyx_t_66;
  Py_ssize_t __pyx_t_67;
  long __pyx_t_68;
  long __pyx_t_69;
  PyObject *__pyx_t_70 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg_cpp_pa_coal", 0);

  /* "cyext_acv.pyx":2643
 *             int num_threads, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2644
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2646
 *     cdef unsigned int m = X.shape[1]
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :, :] leaves_tree_l
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":2655
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int].iterator t
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":2661
 *     cdef double[:] sdp
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":2662
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm, lm_u, lm_d, lm_s
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":2670
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_len_s_star = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2675
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2675, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":2676
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_9 = __pyx_v_N;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "cyext_acv.pyx":2677
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":2678
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2678, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":2681
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_R_buf = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2683
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2683, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_12) {

    /* "cyext_acv.pyx":2684
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 2684, __pyx_L1_error)
    }
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2684, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_ci = __pyx_t_15;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 2684, __pyx_L1_error)
      }
      __pyx_t_2 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_16 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2684, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_17 = __pyx_t_16;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_cj = __pyx_t_18;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 2684, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2684, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":2685
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in range(m)]
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __pyx_v_m;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 2685, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_19 = (__pyx_t_12 != 0);
      if (__pyx_t_19) {
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_2);
        PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 2685, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __pyx_t_4 = PyNumber_Add(__pyx_t_7, __pyx_v_C); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "cyext_acv.pyx":2683
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":2687
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __pyx_v_m;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_7);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2687, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":2689
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_13 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2689, __pyx_L1_error)
  __pyx_v_m = __pyx_t_13;

  /* "cyext_acv.pyx":2690
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_power = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "cyext_acv.pyx":2691
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":2692
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_20 = (__pyx_v_m + 1);
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_21; __pyx_t_11+=1) {
    __pyx_v_size = __pyx_t_11;

    /* "cyext_acv.pyx":2693
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "cyext_acv.pyx":2694
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_combinations); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_22 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_22 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2694, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2694, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_22); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_22, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_22, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_7 = __pyx_t_4; __Pyx_INCREF(__pyx_t_7); __pyx_t_13 = 0;
      __pyx_t_23 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_23 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 2694, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_23)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_4); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 2694, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_13); __Pyx_INCREF(__pyx_t_4); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 2694, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_23(__pyx_t_7);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2694, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "cyext_acv.pyx":2695
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_24 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_24)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_24);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_4 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_24, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_4); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 2695, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "cyext_acv.pyx":2696
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**25:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":2694
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "cyext_acv.pyx":2697
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**25:
 *             break
 */
    __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 2697, __pyx_L1_error)

    /* "cyext_acv.pyx":2698
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**25:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_19 = ((__pyx_v_max_size >= 0x2000000) != 0);
    if (__pyx_t_19) {

      /* "cyext_acv.pyx":2699
 *         power.append(power_b)
 *         if max_size >= 2**25:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L16_break;

      /* "cyext_acv.pyx":2698
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**25:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L16_break:;

  /* "cyext_acv.pyx":2701
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_26 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2701, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_26;

  /* "cyext_acv.pyx":2703
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ones); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_24) < 0) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_27 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_27.memview)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_27;
  __pyx_t_27.memview = NULL;
  __pyx_t_27.data = NULL;

  /* "cyext_acv.pyx":2707
 * 
 *     # cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 *     mean_forest = np.zeros((X.shape[0], 3))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_S = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2708
 *     # cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 *     mean_forest = np.zeros((X.shape[0], 3))             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in range(m + 1):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_24);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":2710
 *     mean_forest = np.zeros((X.shape[0], 3))
 * 
 *     for s_0 in range(m + 1):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __pyx_t_20 = (__pyx_v_m + 1);
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_21; __pyx_t_11+=1) {
    __pyx_v_s_0 = __pyx_t_11;

    /* "cyext_acv.pyx":2711
 * 
 *     for s_0 in range(m + 1):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_28 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_29; __pyx_t_22+=1) {
      __pyx_v_s_1 = __pyx_t_22;

      /* "cyext_acv.pyx":2712
 *     for s_0 in range(m + 1):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_30 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
        __pyx_v_i = __pyx_t_32;

        /* "cyext_acv.pyx":2713
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":2715
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":2716
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":2717
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":2718
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 *                 for j in range(3):
 */
      __pyx_t_9 = __pyx_v_N;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_10; __pyx_t_32+=1) {
        __pyx_v_i = __pyx_t_32;

        /* "cyext_acv.pyx":2719
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 *                 for j in range(3):
 *                     mean_forest[R_buf[i], j] = 0
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);

        /* "cyext_acv.pyx":2720
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 *                 for j in range(3):             # <<<<<<<<<<<<<<
 *                     mean_forest[R_buf[i], j] = 0
 * 
 */
        for (__pyx_t_34 = 0; __pyx_t_34 < 3; __pyx_t_34+=1) {
          __pyx_v_j = __pyx_t_34;

          /* "cyext_acv.pyx":2721
 *                 R_buf[i] = R[i]
 *                 for j in range(3):
 *                     mean_forest[R_buf[i], j] = 0             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
          __pyx_t_33 = __pyx_v_i;
          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_36 = __pyx_v_j;
          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_35 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_36 * __pyx_v_mean_forest.strides[1]) )) = 0.0;
        }
      }

      /* "cyext_acv.pyx":2723
 *                     mean_forest[R_buf[i], j] = 0
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 for l in range(n_trees):
 *                     if b == l:
 */
      __pyx_t_32 = __pyx_v_n_trees;
      __pyx_t_34 = __pyx_t_32;
      for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_34; __pyx_t_37+=1) {
        __pyx_v_b = __pyx_t_37;

        /* "cyext_acv.pyx":2724
 * 
 *             for b in range(n_trees):
 *                 for l in range(n_trees):             # <<<<<<<<<<<<<<
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]
 */
        __pyx_t_38 = __pyx_v_n_trees;
        __pyx_t_39 = __pyx_t_38;
        for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
          __pyx_v_l = __pyx_t_40;

          /* "cyext_acv.pyx":2725
 *             for b in range(n_trees):
 *                 for l in range(n_trees):
 *                     if b == l:             # <<<<<<<<<<<<<<
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 */
          __pyx_t_19 = ((__pyx_v_b == __pyx_v_l) != 0);
          if (__pyx_t_19) {

            /* "cyext_acv.pyx":2726
 *                 for l in range(n_trees):
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                         nb_leaf = leaves_nb[b]
 * 
 */
            __pyx_t_41.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_41.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_41, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_41.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_41.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_41.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_41.suboffsets[0] = -1;

__pyx_t_41.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_41.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_41.suboffsets[1] = -1;

__pyx_t_41.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_41.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_41.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
            __pyx_v_leaves_tree = __pyx_t_41;
            __pyx_t_41.memview = NULL;
            __pyx_t_41.data = NULL;

            /* "cyext_acv.pyx":2727
 *                     if b == l:
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 * 
 *                         for leaf_numb in range(nb_leaf):
 */
            __pyx_t_33 = __pyx_v_b;
            __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_33 * __pyx_v_leaves_nb.strides[0]) )));

            /* "cyext_acv.pyx":2729
 *                         nb_leaf = leaves_nb[b]
 * 
 *                         for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                             leaf_part = leaves_tree[leaf_numb]
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
            __pyx_t_9 = __pyx_v_nb_leaf;
            __pyx_t_10 = __pyx_t_9;
            for (__pyx_t_42 = 0; __pyx_t_42 < __pyx_t_10; __pyx_t_42+=1) {
              __pyx_v_leaf_numb = __pyx_t_42;

              /* "cyext_acv.pyx":2730
 * 
 *                         for leaf_numb in range(nb_leaf):
 *                             leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 */
              __pyx_t_5.data = __pyx_v_leaves_tree.data;
              __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
              __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
              __pyx_v_leaf_part = __pyx_t_5;
              __pyx_t_5.memview = NULL;
              __pyx_t_5.data = NULL;

              /* "cyext_acv.pyx":2731
 *                         for leaf_numb in range(nb_leaf):
 *                             leaf_part = leaves_tree[leaf_numb]
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)
 */
              __pyx_t_33 = __pyx_v_b;
              __pyx_t_36 = __pyx_v_leaf_numb;
              __pyx_t_35 = __pyx_v_b;
              __pyx_t_43 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_idx_trees.strides[1]) )));
              __pyx_t_44 = 0;
              __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_35 * __pyx_v_values.strides[0]) ) + __pyx_t_43 * __pyx_v_values.strides[1]) ) + __pyx_t_44 * __pyx_v_values.strides[2]) )));

              /* "cyext_acv.pyx":2733
 *                             value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
              __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_24);
              __Pyx_GIVEREF(__pyx_t_7);
              PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2733, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
              __pyx_v_lm = __pyx_t_8;
              __pyx_t_8.memview = NULL;
              __pyx_t_8.data = NULL;

              /* "cyext_acv.pyx":2734
 * 
 *                             lm = np.zeros(data.shape[0], dtype=np.int)
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                             for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_24);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_24, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2734, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
              __pyx_v_lm_s = __pyx_t_8;
              __pyx_t_8.memview = NULL;
              __pyx_t_8.data = NULL;

              /* "cyext_acv.pyx":2736
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                             for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                 a_it = 0
 *                                 b_it = 0
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {
                    if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 2736, __pyx_L40_error) }
                    __pyx_t_13 = (__pyx_v_data.shape[0]);
                    if ((1 == 0)) abort();
                    {
                        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                            #undef likely
                            #undef unlikely
                            #define likely(x)   (x)
                            #define unlikely(x) (x)
                        #endif
                        __pyx_t_15 = (__pyx_t_13 - 0 + 1 - 1/abs(1)) / 1;
                        if (__pyx_t_15 > 0)
                        {
                            #ifdef _OPENMP
                            #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_19, __pyx_t_33, __pyx_t_35, __pyx_t_36, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48)
                            #endif /* _OPENMP */
                            {
                                #ifdef _OPENMP
                                #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                                #endif /* _OPENMP */
                                for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_15; __pyx_t_14++){
                                    {
                                        __pyx_v_i = (int)(0 + 1 * __pyx_t_14);
                                        /* Initialize private variables to invalid values */
                                        __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                        __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                        __pyx_v_s = ((int)0xbad0bad0);

                                        /* "cyext_acv.pyx":2737
 * 
 *                             for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                                 a_it = 0             # <<<<<<<<<<<<<<
 *                                 b_it = 0
 *                                 for s in range(data.shape[1]):
 */
                                        __pyx_v_a_it = 0;

                                        /* "cyext_acv.pyx":2738
 *                             for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                                 a_it = 0
 *                                 b_it = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(data.shape[1]):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 */
                                        __pyx_v_b_it = 0;

                                        /* "cyext_acv.pyx":2739
 *                                 a_it = 0
 *                                 b_it = 0
 *                                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1
 */
                                        __pyx_t_16 = (__pyx_v_data.shape[1]);
                                        __pyx_t_17 = __pyx_t_16;
                                        for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_17; __pyx_t_45+=1) {
                                          __pyx_v_s = __pyx_t_45;

                                          /* "cyext_acv.pyx":2740
 *                                 b_it = 0
 *                                 for s in range(data.shape[1]):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 */
                                          __pyx_t_36 = __pyx_v_i;
                                          __pyx_t_33 = __pyx_v_s;
                                          __pyx_t_44 = __pyx_v_s;
                                          __pyx_t_43 = 1;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_33 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_44 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (__pyx_t_12) {
                                          } else {
                                            __pyx_t_19 = __pyx_t_12;
                                            goto __pyx_L49_bool_binop_done;
                                          }
                                          __pyx_t_43 = __pyx_v_i;
                                          __pyx_t_44 = __pyx_v_s;
                                          __pyx_t_33 = __pyx_v_s;
                                          __pyx_t_36 = 0;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_43 * __pyx_v_data.strides[0]) ) + __pyx_t_44 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_33 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_36 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_19 = __pyx_t_12;
                                          __pyx_L49_bool_binop_done:;
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2741
 *                                 for s in range(data.shape[1]):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 */
                                            __pyx_v_a_it = (__pyx_v_a_it + 1);

                                            /* "cyext_acv.pyx":2740
 *                                 b_it = 0
 *                                 for s in range(data.shape[1]):
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):             # <<<<<<<<<<<<<<
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":2742
 *                                     if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0])):
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                         b_it = b_it + 1
 */
                                        __pyx_t_45 = __pyx_v_S_size;
                                        __pyx_t_46 = __pyx_t_45;
                                        for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_46; __pyx_t_47+=1) {
                                          __pyx_v_s = __pyx_t_47;

                                          /* "cyext_acv.pyx":2743
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          __pyx_t_36 = __pyx_v_s;
                                          __pyx_t_33 = __pyx_v_i;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_36 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_43 = __pyx_v_s;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_48 = 1;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_33 * __pyx_v_data.strides[0]) ) + __pyx_t_44 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_35 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_48 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (__pyx_t_12) {
                                          } else {
                                            __pyx_t_19 = __pyx_t_12;
                                            goto __pyx_L54_bool_binop_done;
                                          }
                                          __pyx_t_43 = __pyx_v_s;
                                          __pyx_t_48 = __pyx_v_i;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_36 = __pyx_v_s;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_36 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_33 = 0;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_48 * __pyx_v_data.strides[0]) ) + __pyx_t_35 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_44 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_19 = __pyx_t_12;
                                          __pyx_L54_bool_binop_done:;
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2744
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):
 *                                         b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                 if a_it == data.shape[1]:
 */
                                            __pyx_v_b_it = (__pyx_v_b_it + 1);

                                            /* "cyext_acv.pyx":2743
 *                                         a_it = a_it + 1
 *                                 for s in range(S_size):
 *                                     if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         b_it = b_it + 1
 * 
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":2746
 *                                         b_it = b_it + 1
 * 
 *                                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                     lm[i] = 1
 * 
 */
                                        __pyx_t_19 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                        if (__pyx_t_19) {

                                          /* "cyext_acv.pyx":2747
 * 
 *                                 if a_it == data.shape[1]:
 *                                     lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 if b_it == S_size:
 */
                                          __pyx_t_36 = __pyx_v_i;
                                          *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_36 * __pyx_v_lm.strides[0]) )) = 1;

                                          /* "cyext_acv.pyx":2746
 *                                         b_it = b_it + 1
 * 
 *                                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                     lm[i] = 1
 * 
 */
                                        }

                                        /* "cyext_acv.pyx":2749
 *                                     lm[i] = 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     lm_s[i] = 1
 * 
 */
                                        __pyx_t_19 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                        if (__pyx_t_19) {

                                          /* "cyext_acv.pyx":2750
 * 
 *                                 if b_it == S_size:
 *                                     lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                             for i in prange(N, nogil=True, num_threads=num_threads):
 */
                                          __pyx_t_36 = __pyx_v_i;
                                          *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_36 * __pyx_v_lm_s.strides[0]) )) = 1;

                                          /* "cyext_acv.pyx":2749
 *                                     lm[i] = 1
 * 
 *                                 if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                     lm_s[i] = 1
 * 
 */
                                        }
                                    }
                                }
                            }
                        }
                    }
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   __builtin_expect(!!(x), 1)
                        #define unlikely(x) __builtin_expect(!!(x), 0)
                    #endif
                  }

                  /* "cyext_acv.pyx":2736
 *                             lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                             for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                 a_it = 0
 *                                 b_it = 0
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L41;
                    }
                    __pyx_L40_error: {
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L1_error;
                    }
                    __pyx_L41:;
                  }
              }

              /* "cyext_acv.pyx":2752
 *                                     lm_s[i] = 1
 * 
 *                             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                 o_all = 0
 *                                 for s in range(S_size):
 */
              {
                  #ifdef WITH_THREAD
                  PyThreadState *_save;
                  Py_UNBLOCK_THREADS
                  __Pyx_FastGIL_Remember();
                  #endif
                  /*try:*/ {
                    __pyx_t_49 = __pyx_v_N;
                    if ((1 == 0)) abort();
                    {
                        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                            #undef likely
                            #undef unlikely
                            #define likely(x)   (x)
                            #define unlikely(x) (x)
                        #endif
                        __pyx_t_51 = (__pyx_t_49 - 0 + 1 - 1/abs(1)) / 1;
                        if (__pyx_t_51 > 0)
                        {
                            #ifdef _OPENMP
                            #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) num_threads(__pyx_v_num_threads) private(__pyx_t_12, __pyx_t_14, __pyx_t_15, __pyx_t_19, __pyx_t_33, __pyx_t_35, __pyx_t_36, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_52, __pyx_t_53)
                            #endif /* _OPENMP */
                            {
                                #ifdef _OPENMP
                                #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                                #endif /* _OPENMP */
                                for (__pyx_t_50 = 0; __pyx_t_50 < __pyx_t_51; __pyx_t_50++){
                                    {
                                        __pyx_v_i = (int)(0 + 1 * __pyx_t_50);
                                        /* Initialize private variables to invalid values */
                                        __pyx_v_j = ((int)0xbad0bad0);
                                        __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                        __pyx_v_s = ((int)0xbad0bad0);

                                        /* "cyext_acv.pyx":2753
 * 
 *                             for i in prange(N, nogil=True, num_threads=num_threads):
 *                                 o_all = 0             # <<<<<<<<<<<<<<
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 */
                                        __pyx_v_o_all = 0;

                                        /* "cyext_acv.pyx":2754
 *                             for i in prange(N, nogil=True, num_threads=num_threads):
 *                                 o_all = 0
 *                                 for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 */
                                        __pyx_t_45 = __pyx_v_S_size;
                                        __pyx_t_46 = __pyx_t_45;
                                        for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_46; __pyx_t_47+=1) {
                                          __pyx_v_s = __pyx_t_47;

                                          /* "cyext_acv.pyx":2755
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 */
                                          __pyx_t_36 = __pyx_v_i;
                                          __pyx_t_33 = __pyx_v_s;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_35 = __pyx_v_s;
                                          __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_52 = 1;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_44 * __pyx_v_X.strides[0]) ) + __pyx_t_43 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_48 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_52 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          if (!__pyx_t_12) {
                                          } else {
                                            __pyx_t_19 = __pyx_t_12;
                                            goto __pyx_L72_bool_binop_done;
                                          }
                                          __pyx_t_35 = __pyx_v_i;
                                          __pyx_t_52 = __pyx_v_s;
                                          __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_52 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_36 = __pyx_v_s;
                                          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_36 * __pyx_v_S.strides[0]) )));
                                          __pyx_t_44 = 0;
                                          __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_48 * __pyx_v_X.strides[0]) ) + __pyx_t_33 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_43 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_44 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                          __pyx_t_19 = __pyx_t_12;
                                          __pyx_L72_bool_binop_done:;
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2756
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                                 if o_all > 0:
 *                                     continue
 */
                                            __pyx_v_o_all = (__pyx_v_o_all + 1);

                                            /* "cyext_acv.pyx":2755
 *                                 o_all = 0
 *                                 for s in range(S_size):
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 */
                                          }
                                        }

                                        /* "cyext_acv.pyx":2757
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                        __pyx_t_19 = ((__pyx_v_o_all > 0) != 0);
                                        if (__pyx_t_19) {

                                          /* "cyext_acv.pyx":2758
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                                 p = 0
 */
                                          goto __pyx_L65_continue;

                                          /* "cyext_acv.pyx":2757
 *                                     if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0])):
 *                                         o_all = o_all + 1
 *                                 if o_all > 0:             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                        }

                                        /* "cyext_acv.pyx":2760
 *                                     continue
 * 
 *                                 p = 0             # <<<<<<<<<<<<<<
 *                                 p_u = 0
 *                                 p_d = 0
 */
                                        __pyx_v_p = 0;

                                        /* "cyext_acv.pyx":2761
 * 
 *                                 p = 0
 *                                 p_u = 0             # <<<<<<<<<<<<<<
 *                                 p_d = 0
 *                                 p_s = 0
 */
                                        __pyx_v_p_u = 0;

                                        /* "cyext_acv.pyx":2762
 *                                 p = 0
 *                                 p_u = 0
 *                                 p_d = 0             # <<<<<<<<<<<<<<
 *                                 p_s = 0
 *                                 p_su = 0
 */
                                        __pyx_v_p_d = 0;

                                        /* "cyext_acv.pyx":2763
 *                                 p_u = 0
 *                                 p_d = 0
 *                                 p_s = 0             # <<<<<<<<<<<<<<
 *                                 p_su = 0
 *                                 p_sd = 0
 */
                                        __pyx_v_p_s = 0;

                                        /* "cyext_acv.pyx":2764
 *                                 p_d = 0
 *                                 p_s = 0
 *                                 p_su = 0             # <<<<<<<<<<<<<<
 *                                 p_sd = 0
 * 
 */
                                        __pyx_v_p_su = 0;

                                        /* "cyext_acv.pyx":2765
 *                                 p_s = 0
 *                                 p_su = 0
 *                                 p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                                 for j in range(data.shape[0]):
 */
                                        __pyx_v_p_sd = 0;

                                        /* "cyext_acv.pyx":2767
 *                                 p_sd = 0
 * 
 *                                 for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 */
                                        __pyx_t_15 = (__pyx_v_data.shape[0]);
                                        __pyx_t_14 = __pyx_t_15;
                                        for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_14; __pyx_t_45+=1) {
                                          __pyx_v_j = __pyx_t_45;

                                          /* "cyext_acv.pyx":2768
 * 
 *                                 for j in range(data.shape[0]):
 *                                     p += lm[j]             # <<<<<<<<<<<<<<
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 */
                                          __pyx_t_36 = __pyx_v_j;
                                          __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_36 * __pyx_v_lm.strides[0]) ))));

                                          /* "cyext_acv.pyx":2769
 *                                 for j in range(data.shape[0]):
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]
 */
                                          __pyx_t_36 = __pyx_v_j;
                                          __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_36 * __pyx_v_lm_s.strides[0]) ))));

                                          /* "cyext_acv.pyx":2770
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]
 */
                                          __pyx_t_36 = __pyx_v_i;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_43 = __pyx_v_j;
                                          __pyx_t_52 = __pyx_v_i;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_33 = __pyx_v_j;
                                          __pyx_t_19 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_44 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_43 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_35 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_33 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2771
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_su += lm_s[j]
 *                                     else:
 */
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_33 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":2772
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         p_d += lm[j]
 */
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) ))));

                                            /* "cyext_acv.pyx":2770
 *                                     p += lm[j]
 *                                     p_s += lm_s[j]
 *                                     if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                         p_u += lm[j]
 *                                         p_su += lm_s[j]
 */
                                            goto __pyx_L77;
                                          }

                                          /* "cyext_acv.pyx":2774
 *                                         p_su += lm_s[j]
 *                                     else:
 *                                         p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_sd += lm_s[j]
 * 
 */
                                          /*else*/ {
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_33 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":2775
 *                                     else:
 *                                         p_d += lm[j]
 *                                         p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                                 mean_forest[R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 */
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) ))));
                                          }
                                          __pyx_L77:;
                                        }

                                        /* "cyext_acv.pyx":2777
 *                                         p_sd += lm_s[j]
 * 
 *                                 mean_forest[R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                 mean_forest[R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0
 *                                 mean_forest[R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 */
                                        if (((__pyx_v_p_s != 0) != 0)) {
                                          __pyx_t_33 = __pyx_v_i;
                                          __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_53 = ((((__pyx_v_p * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_s) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_52 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p) * __pyx_v_value) / __pyx_v_p_s));
                                        } else {
                                          __pyx_t_53 = 0.0;
                                        }
                                        __pyx_t_33 = __pyx_v_i;
                                        __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_35 = 0;
                                        *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_52 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_35 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;

                                        /* "cyext_acv.pyx":2778
 * 
 *                                 mean_forest[R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 *                                 mean_forest[R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                 mean_forest[R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 *                     else:
 */
                                        if (((__pyx_v_p_su != 0) != 0)) {
                                          __pyx_t_33 = __pyx_v_i;
                                          __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_53 = ((((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_su) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_35 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_u) * __pyx_v_value) / __pyx_v_p_su));
                                        } else {
                                          __pyx_t_53 = 0.0;
                                        }
                                        __pyx_t_33 = __pyx_v_i;
                                        __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_52 = 1;
                                        *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_35 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_52 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;

                                        /* "cyext_acv.pyx":2779
 *                                 mean_forest[R_buf[i], 0] += (p * value*value) / (p_s) - (2 * fX[R_buf[i]] * p * value)/(p_s) if p_s != 0 else 0
 *                                 mean_forest[R_buf[i], 1] += (p_u * value*value) / (p_su) - (2 * fX[R_buf[i]] * p_u * value)/(p_su) if p_su != 0 else 0
 *                                 mean_forest[R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]
 */
                                        if (((__pyx_v_p_sd != 0) != 0)) {
                                          __pyx_t_33 = __pyx_v_i;
                                          __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_53 = ((((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value) / __pyx_v_p_sd) - ((((2.0 * (*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_52 * __pyx_v_fX.strides[0]) )))) * __pyx_v_p_d) * __pyx_v_value) / __pyx_v_p_sd));
                                        } else {
                                          __pyx_t_53 = 0.0;
                                        }
                                        __pyx_t_33 = __pyx_v_i;
                                        __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                        __pyx_t_35 = 2;
                                        *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_52 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_35 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;
                                        goto __pyx_L79;
                                        __pyx_L65_continue:;
                                        goto __pyx_L79;
                                        __pyx_L79:;
                                    }
                                }
                            }
                        }
                    }
                    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                        #undef likely
                        #undef unlikely
                        #define likely(x)   __builtin_expect(!!(x), 1)
                        #define unlikely(x) __builtin_expect(!!(x), 0)
                    #endif
                  }

                  /* "cyext_acv.pyx":2752
 *                                     lm_s[i] = 1
 * 
 *                             for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                 o_all = 0
 *                                 for s in range(S_size):
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      #ifdef WITH_THREAD
                      __Pyx_FastGIL_Forget();
                      Py_BLOCK_THREADS
                      #endif
                      goto __pyx_L64;
                    }
                    __pyx_L64:;
                  }
              }
            }

            /* "cyext_acv.pyx":2725
 *             for b in range(n_trees):
 *                 for l in range(n_trees):
 *                     if b == l:             # <<<<<<<<<<<<<<
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 */
            goto __pyx_L34;
          }

          /* "cyext_acv.pyx":2781
 *                                 mean_forest[R_buf[i], 2] += (p_d * value*value) / (p_sd) - (2 * fX[R_buf[i]] * p_d * value)/(p_sd) if p_sd != 0 else 0
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]             # <<<<<<<<<<<<<<
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]
 */
          /*else*/ {
            __pyx_t_41.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_41.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_41, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_b;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_41.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_41.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_41.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_41.suboffsets[0] = -1;

__pyx_t_41.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_41.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_41.suboffsets[1] = -1;

__pyx_t_41.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_41.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_41.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
            __pyx_v_leaves_tree = __pyx_t_41;
            __pyx_t_41.memview = NULL;
            __pyx_t_41.data = NULL;

            /* "cyext_acv.pyx":2782
 *                     else:
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                         leaves_tree_l = partition_leaves_trees[l]
 *                         nb_leaf_l = leaves_nb[l]
 */
            __pyx_t_33 = __pyx_v_b;
            __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_33 * __pyx_v_leaves_nb.strides[0]) )));

            /* "cyext_acv.pyx":2783
 *                         leaves_tree = partition_leaves_trees[b]
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]             # <<<<<<<<<<<<<<
 *                         nb_leaf_l = leaves_nb[l]
 * 
 */
            __pyx_t_41.data = __pyx_v_partition_leaves_trees.data;
            __pyx_t_41.memview = __pyx_v_partition_leaves_trees.memview;
            __PYX_INC_MEMVIEW(&__pyx_t_41, 0);
            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_leaves_trees.strides[0];
        __pyx_t_41.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_41.shape[0] = __pyx_v_partition_leaves_trees.shape[1];
__pyx_t_41.strides[0] = __pyx_v_partition_leaves_trees.strides[1];
    __pyx_t_41.suboffsets[0] = -1;

__pyx_t_41.shape[1] = __pyx_v_partition_leaves_trees.shape[2];
__pyx_t_41.strides[1] = __pyx_v_partition_leaves_trees.strides[2];
    __pyx_t_41.suboffsets[1] = -1;

__pyx_t_41.shape[2] = __pyx_v_partition_leaves_trees.shape[3];
__pyx_t_41.strides[2] = __pyx_v_partition_leaves_trees.strides[3];
    __pyx_t_41.suboffsets[2] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
            __pyx_v_leaves_tree_l = __pyx_t_41;
            __pyx_t_41.memview = NULL;
            __pyx_t_41.data = NULL;

            /* "cyext_acv.pyx":2784
 *                         nb_leaf = leaves_nb[b]
 *                         leaves_tree_l = partition_leaves_trees[l]
 *                         nb_leaf_l = leaves_nb[l]             # <<<<<<<<<<<<<<
 * 
 *                         for leaf_numb in range(nb_leaf):
 */
            __pyx_t_33 = __pyx_v_l;
            __pyx_v_nb_leaf_l = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_33 * __pyx_v_leaves_nb.strides[0]) )));

            /* "cyext_acv.pyx":2786
 *                         nb_leaf_l = leaves_nb[l]
 * 
 *                         for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                             for leaf_numb_l in range(nb_leaf_l):
 * 
 */
            __pyx_t_9 = __pyx_v_nb_leaf;
            __pyx_t_10 = __pyx_t_9;
            for (__pyx_t_42 = 0; __pyx_t_42 < __pyx_t_10; __pyx_t_42+=1) {
              __pyx_v_leaf_numb = __pyx_t_42;

              /* "cyext_acv.pyx":2787
 * 
 *                         for leaf_numb in range(nb_leaf):
 *                             for leaf_numb_l in range(nb_leaf_l):             # <<<<<<<<<<<<<<
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]
 */
              __pyx_t_51 = __pyx_v_nb_leaf_l;
              __pyx_t_50 = __pyx_t_51;
              for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_50; __pyx_t_45+=1) {
                __pyx_v_leaf_numb_l = __pyx_t_45;

                /* "cyext_acv.pyx":2789
 *                             for leaf_numb_l in range(nb_leaf_l):
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]             # <<<<<<<<<<<<<<
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 */
                __pyx_t_5.data = __pyx_v_leaves_tree.data;
                __pyx_t_5.memview = __pyx_v_leaves_tree.memview;
                __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
                __pyx_v_leaf_part = __pyx_t_5;
                __pyx_t_5.memview = NULL;
                __pyx_t_5.data = NULL;

                /* "cyext_acv.pyx":2790
 * 
 *                                 leaf_part = leaves_tree[leaf_numb]
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]             # <<<<<<<<<<<<<<
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 */
                __pyx_t_5.data = __pyx_v_leaves_tree_l.data;
                __pyx_t_5.memview = __pyx_v_leaves_tree_l.memview;
                __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_leaf_numb_l;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_leaves_tree_l.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_leaves_tree_l.shape[1];
__pyx_t_5.strides[0] = __pyx_v_leaves_tree_l.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_5.shape[1] = __pyx_v_leaves_tree_l.shape[2];
__pyx_t_5.strides[1] = __pyx_v_leaves_tree_l.strides[2];
    __pyx_t_5.suboffsets[1] = -1;

__PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
                __pyx_v_leaf_part_l = __pyx_t_5;
                __pyx_t_5.memview = NULL;
                __pyx_t_5.data = NULL;

                /* "cyext_acv.pyx":2791
 *                                 leaf_part = leaves_tree[leaf_numb]
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]             # <<<<<<<<<<<<<<
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 */
                __pyx_t_33 = __pyx_v_b;
                __pyx_t_35 = __pyx_v_leaf_numb;
                __pyx_t_52 = __pyx_v_b;
                __pyx_t_43 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_idx_trees.strides[1]) )));
                __pyx_t_36 = 0;
                __pyx_v_value = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_52 * __pyx_v_values.strides[0]) ) + __pyx_t_43 * __pyx_v_values.strides[1]) ) + __pyx_t_36 * __pyx_v_values.strides[2]) )));

                /* "cyext_acv.pyx":2792
 *                                 leaf_part_l = leaves_tree_l[leaf_numb_l]
 *                                 value = values[b, leaf_idx_trees[b, leaf_numb], 0]
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]             # <<<<<<<<<<<<<<
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)
 */
                __pyx_t_35 = __pyx_v_l;
                __pyx_t_33 = __pyx_v_leaf_numb_l;
                __pyx_t_36 = __pyx_v_l;
                __pyx_t_43 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_35 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[1]) )));
                __pyx_t_52 = 0;
                __pyx_v_value_l = (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_36 * __pyx_v_values.strides[0]) ) + __pyx_t_43 * __pyx_v_values.strides[1]) ) + __pyx_t_52 * __pyx_v_values.strides[2]) )));

                /* "cyext_acv.pyx":2794
 *                                 value_l = values[l, leaf_idx_trees[l, leaf_numb_l], 0]
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 */
                __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_24);
                __Pyx_GIVEREF(__pyx_t_1);
                PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_1);
                __pyx_t_1 = 0;
                __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_24, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2794, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
                __pyx_v_lm = __pyx_t_8;
                __pyx_t_8.memview = NULL;
                __pyx_t_8.data = NULL;

                /* "cyext_acv.pyx":2795
 * 
 *                                 lm = np.zeros(data.shape[0], dtype=np.int)
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *                                 for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 */
                __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_24);
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_2);
                __pyx_t_2 = 0;
                __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_24, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2795, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
                __pyx_v_lm_s = __pyx_t_8;
                __pyx_t_8.memview = NULL;
                __pyx_t_8.data = NULL;

                /* "cyext_acv.pyx":2797
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                                 for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                     a_it = 0
 *                                     b_it = 0
 */
                {
                    #ifdef WITH_THREAD
                    PyThreadState *_save;
                    Py_UNBLOCK_THREADS
                    __Pyx_FastGIL_Remember();
                    #endif
                    /*try:*/ {
                      if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 2797, __pyx_L87_error) }
                      __pyx_t_15 = (__pyx_v_data.shape[0]);
                      if ((1 == 0)) abort();
                      {
                          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                              #undef likely
                              #undef unlikely
                              #define likely(x)   (x)
                              #define unlikely(x) (x)
                          #endif
                          __pyx_t_13 = (__pyx_t_15 - 0 + 1 - 1/abs(1)) / 1;
                          if (__pyx_t_13 > 0)
                          {
                              #ifdef _OPENMP
                              #pragma omp parallel num_threads(__pyx_v_num_threads) private(__pyx_t_12, __pyx_t_16, __pyx_t_17, __pyx_t_19, __pyx_t_33, __pyx_t_35, __pyx_t_36, __pyx_t_43, __pyx_t_44, __pyx_t_46, __pyx_t_47, __pyx_t_52, __pyx_t_54)
                              #endif /* _OPENMP */
                              {
                                  #ifdef _OPENMP
                                  #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_s)
                                  #endif /* _OPENMP */
                                  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14++){
                                      {
                                          __pyx_v_i = (int)(0 + 1 * __pyx_t_14);
                                          /* Initialize private variables to invalid values */
                                          __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                          __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                          __pyx_v_s = ((int)0xbad0bad0);

                                          /* "cyext_acv.pyx":2798
 * 
 *                                 for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                                     a_it = 0             # <<<<<<<<<<<<<<
 *                                     b_it = 0
 *                                     for s in range(data.shape[1]):
 */
                                          __pyx_v_a_it = 0;

                                          /* "cyext_acv.pyx":2799
 *                                 for i in prange(data.shape[0], nogil=True, num_threads=num_threads):
 *                                     a_it = 0
 *                                     b_it = 0             # <<<<<<<<<<<<<<
 *                                     for s in range(data.shape[1]):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 */
                                          __pyx_v_b_it = 0;

                                          /* "cyext_acv.pyx":2800
 *                                     a_it = 0
 *                                     b_it = 0
 *                                     for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1
 */
                                          __pyx_t_16 = (__pyx_v_data.shape[1]);
                                          __pyx_t_17 = __pyx_t_16;
                                          for (__pyx_t_46 = 0; __pyx_t_46 < __pyx_t_17; __pyx_t_46+=1) {
                                            __pyx_v_s = __pyx_t_46;

                                            /* "cyext_acv.pyx":2801
 *                                     b_it = 0
 *                                     for s in range(data.shape[1]):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 */
                                            __pyx_t_33 = __pyx_v_i;
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_52 = __pyx_v_s;
                                            __pyx_t_43 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_33 * __pyx_v_data.strides[0]) ) + __pyx_t_35 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_52 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L96_bool_binop_done;
                                            }
                                            __pyx_t_43 = __pyx_v_i;
                                            __pyx_t_52 = __pyx_v_s;
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_33 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_43 * __pyx_v_data.strides[0]) ) + __pyx_t_52 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_35 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L96_bool_binop_done;
                                            }
                                            __pyx_t_33 = __pyx_v_i;
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_52 = __pyx_v_s;
                                            __pyx_t_43 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_33 * __pyx_v_data.strides[0]) ) + __pyx_t_35 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_52 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L96_bool_binop_done;
                                            }
                                            __pyx_t_43 = __pyx_v_i;
                                            __pyx_t_52 = __pyx_v_s;
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_33 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_43 * __pyx_v_data.strides[0]) ) + __pyx_t_52 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_35 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_19 = __pyx_t_12;
                                            __pyx_L96_bool_binop_done:;
                                            if (__pyx_t_19) {

                                              /* "cyext_acv.pyx":2802
 *                                     for s in range(data.shape[1]):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 */
                                              __pyx_v_a_it = (__pyx_v_a_it + 1);

                                              /* "cyext_acv.pyx":2801
 *                                     b_it = 0
 *                                     for s in range(data.shape[1]):
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):             # <<<<<<<<<<<<<<
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":2803
 *                                         if ((data[i, s] <= leaf_part[s, 1]) and (data[i, s] >= leaf_part[s, 0]) and (data[i, s] <= leaf_part_l[s, 1]) and (data[i, s] >= leaf_part_l[s, 0])):
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                             b_it = b_it + 1
 */
                                          __pyx_t_46 = __pyx_v_S_size;
                                          __pyx_t_47 = __pyx_t_46;
                                          for (__pyx_t_54 = 0; __pyx_t_54 < __pyx_t_47; __pyx_t_54+=1) {
                                            __pyx_v_s = __pyx_t_54;

                                            /* "cyext_acv.pyx":2804
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             b_it = b_it + 1
 * 
 */
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_43 = __pyx_v_s;
                                            __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_44 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_52 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_36 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_44 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L103_bool_binop_done;
                                            }
                                            __pyx_t_43 = __pyx_v_s;
                                            __pyx_t_44 = __pyx_v_i;
                                            __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_44 * __pyx_v_data.strides[0]) ) + __pyx_t_36 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_52 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L103_bool_binop_done;
                                            }
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_35 = __pyx_v_i;
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_43 = __pyx_v_s;
                                            __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_44 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_52 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_36 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_44 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L103_bool_binop_done;
                                            }
                                            __pyx_t_43 = __pyx_v_s;
                                            __pyx_t_44 = __pyx_v_i;
                                            __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_35 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_44 * __pyx_v_data.strides[0]) ) + __pyx_t_36 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_52 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_35 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_19 = __pyx_t_12;
                                            __pyx_L103_bool_binop_done:;
                                            if (__pyx_t_19) {

                                              /* "cyext_acv.pyx":2805
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):
 *                                             b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                                     if a_it == data.shape[1]:
 */
                                              __pyx_v_b_it = (__pyx_v_b_it + 1);

                                              /* "cyext_acv.pyx":2804
 *                                             a_it = a_it + 1
 *                                     for s in range(S_size):
 *                                         if ((data[i, S[s]] <= leaf_part[S[s], 1]) and (data[i, S[s]] >= leaf_part[S[s], 0]) and (data[i, S[s]] <= leaf_part_l[S[s], 1]) and (data[i, S[s]] >= leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             b_it = b_it + 1
 * 
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":2807
 *                                             b_it = b_it + 1
 * 
 *                                     if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                         lm[i] = 1
 * 
 */
                                          __pyx_t_19 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2808
 * 
 *                                     if a_it == data.shape[1]:
 *                                         lm[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                     if b_it == S_size:
 */
                                            __pyx_t_33 = __pyx_v_i;
                                            *((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_33 * __pyx_v_lm.strides[0]) )) = 1;

                                            /* "cyext_acv.pyx":2807
 *                                             b_it = b_it + 1
 * 
 *                                     if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                         lm[i] = 1
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":2810
 *                                         lm[i] = 1
 * 
 *                                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                         lm_s[i] = 1
 * 
 */
                                          __pyx_t_19 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2811
 * 
 *                                     if b_it == S_size:
 *                                         lm_s[i] = 1             # <<<<<<<<<<<<<<
 * 
 *                                 for i in prange(N, nogil=True, num_threads=num_threads):
 */
                                            __pyx_t_33 = __pyx_v_i;
                                            *((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) )) = 1;

                                            /* "cyext_acv.pyx":2810
 *                                         lm[i] = 1
 * 
 *                                     if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                         lm_s[i] = 1
 * 
 */
                                          }
                                      }
                                  }
                              }
                          }
                      }
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   __builtin_expect(!!(x), 1)
                          #define unlikely(x) __builtin_expect(!!(x), 0)
                      #endif
                    }

                    /* "cyext_acv.pyx":2797
 *                                 lm_s = np.zeros(data.shape[0], dtype=np.int)
 * 
 *                                 for i in prange(data.shape[0], nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                     a_it = 0
 *                                     b_it = 0
 */
                    /*finally:*/ {
                      /*normal exit:*/{
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L88;
                      }
                      __pyx_L87_error: {
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L1_error;
                      }
                      __pyx_L88:;
                    }
                }

                /* "cyext_acv.pyx":2813
 *                                         lm_s[i] = 1
 * 
 *                                 for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                     o_all = 0
 *                                     for s in range(S_size):
 */
                {
                    #ifdef WITH_THREAD
                    PyThreadState *_save;
                    Py_UNBLOCK_THREADS
                    __Pyx_FastGIL_Remember();
                    #endif
                    /*try:*/ {
                      __pyx_t_49 = __pyx_v_N;
                      if ((1 == 0)) abort();
                      {
                          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                              #undef likely
                              #undef unlikely
                              #define likely(x)   (x)
                              #define unlikely(x) (x)
                          #endif
                          __pyx_t_56 = (__pyx_t_49 - 0 + 1 - 1/abs(1)) / 1;
                          if (__pyx_t_56 > 0)
                          {
                              #ifdef _OPENMP
                              #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) num_threads(__pyx_v_num_threads) private(__pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_19, __pyx_t_33, __pyx_t_35, __pyx_t_36, __pyx_t_43, __pyx_t_44, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_52, __pyx_t_53, __pyx_t_54)
                              #endif /* _OPENMP */
                              {
                                  #ifdef _OPENMP
                                  #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s)
                                  #endif /* _OPENMP */
                                  for (__pyx_t_55 = 0; __pyx_t_55 < __pyx_t_56; __pyx_t_55++){
                                      {
                                          __pyx_v_i = (int)(0 + 1 * __pyx_t_55);
                                          /* Initialize private variables to invalid values */
                                          __pyx_v_j = ((int)0xbad0bad0);
                                          __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                          __pyx_v_s = ((int)0xbad0bad0);

                                          /* "cyext_acv.pyx":2814
 * 
 *                                 for i in prange(N, nogil=True, num_threads=num_threads):
 *                                     o_all = 0             # <<<<<<<<<<<<<<
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 */
                                          __pyx_v_o_all = 0;

                                          /* "cyext_acv.pyx":2815
 *                                 for i in prange(N, nogil=True, num_threads=num_threads):
 *                                     o_all = 0
 *                                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 */
                                          __pyx_t_46 = __pyx_v_S_size;
                                          __pyx_t_47 = __pyx_t_46;
                                          for (__pyx_t_54 = 0; __pyx_t_54 < __pyx_t_47; __pyx_t_54+=1) {
                                            __pyx_v_s = __pyx_t_54;

                                            /* "cyext_acv.pyx":2816
 *                                     o_all = 0
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 */
                                            __pyx_t_33 = __pyx_v_i;
                                            __pyx_t_35 = __pyx_v_s;
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_36 = __pyx_v_s;
                                            __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_36 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_48 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_52 * __pyx_v_X.strides[0]) ) + __pyx_t_43 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_44 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_48 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (!__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L123_bool_binop_done;
                                            }
                                            __pyx_t_36 = __pyx_v_i;
                                            __pyx_t_48 = __pyx_v_s;
                                            __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_52 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_44 * __pyx_v_X.strides[0]) ) + __pyx_t_35 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part.data + __pyx_t_43 * __pyx_v_leaf_part.strides[0]) ) + __pyx_t_52 * __pyx_v_leaf_part.strides[1]) )))) != 0);
                                            if (!__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L123_bool_binop_done;
                                            }
                                            __pyx_t_33 = __pyx_v_i;
                                            __pyx_t_52 = __pyx_v_s;
                                            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_52 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_36 = __pyx_v_s;
                                            __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_36 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_44 = 1;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_43 * __pyx_v_X.strides[0]) ) + __pyx_t_48 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_35 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_44 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            if (!__pyx_t_12) {
                                            } else {
                                              __pyx_t_19 = __pyx_t_12;
                                              goto __pyx_L123_bool_binop_done;
                                            }
                                            __pyx_t_36 = __pyx_v_i;
                                            __pyx_t_44 = __pyx_v_s;
                                            __pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_52 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_44 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_33 = __pyx_v_s;
                                            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                            __pyx_t_43 = 0;
                                            __pyx_t_12 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_35 * __pyx_v_X.strides[0]) ) + __pyx_t_52 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_part_l.data + __pyx_t_48 * __pyx_v_leaf_part_l.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_part_l.strides[1]) )))) != 0);
                                            __pyx_t_19 = __pyx_t_12;
                                            __pyx_L123_bool_binop_done:;
                                            if (__pyx_t_19) {

                                              /* "cyext_acv.pyx":2817
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                                     if o_all > 0:
 *                                         continue
 */
                                              __pyx_v_o_all = (__pyx_v_o_all + 1);

                                              /* "cyext_acv.pyx":2816
 *                                     o_all = 0
 *                                     for s in range(S_size):
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):             # <<<<<<<<<<<<<<
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 */
                                            }
                                          }

                                          /* "cyext_acv.pyx":2818
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:             # <<<<<<<<<<<<<<
 *                                         continue
 * 
 */
                                          __pyx_t_19 = ((__pyx_v_o_all > 0) != 0);
                                          if (__pyx_t_19) {

                                            /* "cyext_acv.pyx":2819
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:
 *                                         continue             # <<<<<<<<<<<<<<
 * 
 *                                     p = 0
 */
                                            goto __pyx_L116_continue;

                                            /* "cyext_acv.pyx":2818
 *                                         if ((X[R_buf[i], S[s]] > leaf_part[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part[S[s], 0]) or (X[R_buf[i], S[s]] > leaf_part_l[S[s], 1]) or (X[R_buf[i], S[s]] < leaf_part_l[S[s], 0])):
 *                                             o_all = o_all + 1
 *                                     if o_all > 0:             # <<<<<<<<<<<<<<
 *                                         continue
 * 
 */
                                          }

                                          /* "cyext_acv.pyx":2821
 *                                         continue
 * 
 *                                     p = 0             # <<<<<<<<<<<<<<
 *                                     p_u = 0
 *                                     p_d = 0
 */
                                          __pyx_v_p = 0;

                                          /* "cyext_acv.pyx":2822
 * 
 *                                     p = 0
 *                                     p_u = 0             # <<<<<<<<<<<<<<
 *                                     p_d = 0
 *                                     p_s = 0
 */
                                          __pyx_v_p_u = 0;

                                          /* "cyext_acv.pyx":2823
 *                                     p = 0
 *                                     p_u = 0
 *                                     p_d = 0             # <<<<<<<<<<<<<<
 *                                     p_s = 0
 *                                     p_su = 0
 */
                                          __pyx_v_p_d = 0;

                                          /* "cyext_acv.pyx":2824
 *                                     p_u = 0
 *                                     p_d = 0
 *                                     p_s = 0             # <<<<<<<<<<<<<<
 *                                     p_su = 0
 *                                     p_sd = 0
 */
                                          __pyx_v_p_s = 0;

                                          /* "cyext_acv.pyx":2825
 *                                     p_d = 0
 *                                     p_s = 0
 *                                     p_su = 0             # <<<<<<<<<<<<<<
 *                                     p_sd = 0
 * 
 */
                                          __pyx_v_p_su = 0;

                                          /* "cyext_acv.pyx":2826
 *                                     p_s = 0
 *                                     p_su = 0
 *                                     p_sd = 0             # <<<<<<<<<<<<<<
 * 
 *                                     for j in range(data.shape[0]):
 */
                                          __pyx_v_p_sd = 0;

                                          /* "cyext_acv.pyx":2828
 *                                     p_sd = 0
 * 
 *                                     for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 */
                                          __pyx_t_13 = (__pyx_v_data.shape[0]);
                                          __pyx_t_14 = __pyx_t_13;
                                          for (__pyx_t_46 = 0; __pyx_t_46 < __pyx_t_14; __pyx_t_46+=1) {
                                            __pyx_v_j = __pyx_t_46;

                                            /* "cyext_acv.pyx":2829
 * 
 *                                     for j in range(data.shape[0]):
 *                                         p += lm[j]             # <<<<<<<<<<<<<<
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 */
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p = (__pyx_v_p + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_33 * __pyx_v_lm.strides[0]) ))));

                                            /* "cyext_acv.pyx":2830
 *                                     for j in range(data.shape[0]):
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]             # <<<<<<<<<<<<<<
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]
 */
                                            __pyx_t_33 = __pyx_v_j;
                                            __pyx_v_p_s = (__pyx_v_p_s + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) ))));

                                            /* "cyext_acv.pyx":2831
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]
 */
                                            __pyx_t_33 = __pyx_v_i;
                                            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_48 = __pyx_v_j;
                                            __pyx_t_44 = __pyx_v_i;
                                            __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_44 * __pyx_v_R_buf.strides[0]) )));
                                            __pyx_t_52 = __pyx_v_j;
                                            __pyx_t_19 = (((((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_43 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_48 * __pyx_v_y_pred.strides[0]) )))) * ((*((double *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_36 * __pyx_v_fX.strides[0]) ))) - (*((double *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_52 * __pyx_v_y_pred.strides[0]) ))))) > __pyx_v_tX) != 0);
                                            if (__pyx_t_19) {

                                              /* "cyext_acv.pyx":2832
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]             # <<<<<<<<<<<<<<
 *                                             p_su += lm_s[j]
 *                                         else:
 */
                                              __pyx_t_52 = __pyx_v_j;
                                              __pyx_v_p_u = (__pyx_v_p_u + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_52 * __pyx_v_lm.strides[0]) ))));

                                              /* "cyext_acv.pyx":2833
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]             # <<<<<<<<<<<<<<
 *                                         else:
 *                                             p_d += lm[j]
 */
                                              __pyx_t_52 = __pyx_v_j;
                                              __pyx_v_p_su = (__pyx_v_p_su + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_52 * __pyx_v_lm_s.strides[0]) ))));

                                              /* "cyext_acv.pyx":2831
 *                                         p += lm[j]
 *                                         p_s += lm_s[j]
 *                                         if (fX[R_buf[i]] - y_pred[j])*(fX[R_buf[i]] - y_pred[j]) > tX:             # <<<<<<<<<<<<<<
 *                                             p_u += lm[j]
 *                                             p_su += lm_s[j]
 */
                                              goto __pyx_L130;
                                            }

                                            /* "cyext_acv.pyx":2835
 *                                             p_su += lm_s[j]
 *                                         else:
 *                                             p_d += lm[j]             # <<<<<<<<<<<<<<
 *                                             p_sd += lm_s[j]
 * 
 */
                                            /*else*/ {
                                              __pyx_t_52 = __pyx_v_j;
                                              __pyx_v_p_d = (__pyx_v_p_d + (*((long *) ( /* dim=0 */ (__pyx_v_lm.data + __pyx_t_52 * __pyx_v_lm.strides[0]) ))));

                                              /* "cyext_acv.pyx":2836
 *                                         else:
 *                                             p_d += lm[j]
 *                                             p_sd += lm_s[j]             # <<<<<<<<<<<<<<
 * 
 *                                     mean_forest[R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 */
                                              __pyx_t_52 = __pyx_v_j;
                                              __pyx_v_p_sd = (__pyx_v_p_sd + (*((long *) ( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_52 * __pyx_v_lm_s.strides[0]) ))));
                                            }
                                            __pyx_L130:;
                                          }

                                          /* "cyext_acv.pyx":2838
 *                                             p_sd += lm_s[j]
 * 
 *                                     mean_forest[R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                                     mean_forest[R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                                     mean_forest[R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 */
                                          if (((__pyx_v_p_s != 0) != 0)) {
                                            __pyx_t_53 = (((__pyx_v_p * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_s);
                                          } else {
                                            __pyx_t_53 = 0.0;
                                          }
                                          __pyx_t_52 = __pyx_v_i;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_36 = 0;
                                          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_44 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_36 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;

                                          /* "cyext_acv.pyx":2839
 * 
 *                                     mean_forest[R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                                     mean_forest[R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                                     mean_forest[R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                          if (((__pyx_v_p_su != 0) != 0)) {
                                            __pyx_t_53 = (((__pyx_v_p_u * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_su);
                                          } else {
                                            __pyx_t_53 = 0.0;
                                          }
                                          __pyx_t_52 = __pyx_v_i;
                                          __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_44 = 1;
                                          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_36 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_44 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;

                                          /* "cyext_acv.pyx":2840
 *                                     mean_forest[R_buf[i], 0] += (p * value*value_l) / (p_s)  if p_s != 0 else 0
 *                                     mean_forest[R_buf[i], 1] += (p_u * value*value_l) / (p_su)  if p_su != 0 else 0
 *                                     mean_forest[R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                          if (((__pyx_v_p_sd != 0) != 0)) {
                                            __pyx_t_53 = (((__pyx_v_p_d * __pyx_v_value) * __pyx_v_value_l) / __pyx_v_p_sd);
                                          } else {
                                            __pyx_t_53 = 0.0;
                                          }
                                          __pyx_t_52 = __pyx_v_i;
                                          __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
                                          __pyx_t_36 = 2;
                                          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_44 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_36 * __pyx_v_mean_forest.strides[1]) )) += __pyx_t_53;
                                          goto __pyx_L132;
                                          __pyx_L116_continue:;
                                          goto __pyx_L132;
                                          __pyx_L132:;
                                      }
                                  }
                              }
                          }
                      }
                      #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                          #undef likely
                          #undef unlikely
                          #define likely(x)   __builtin_expect(!!(x), 1)
                          #define unlikely(x) __builtin_expect(!!(x), 0)
                      #endif
                    }

                    /* "cyext_acv.pyx":2813
 *                                         lm_s[i] = 1
 * 
 *                                 for i in prange(N, nogil=True, num_threads=num_threads):             # <<<<<<<<<<<<<<
 *                                     o_all = 0
 *                                     for s in range(S_size):
 */
                    /*finally:*/ {
                      /*normal exit:*/{
                        #ifdef WITH_THREAD
                        __Pyx_FastGIL_Forget();
                        Py_BLOCK_THREADS
                        #endif
                        goto __pyx_L115;
                      }
                      __pyx_L115:;
                    }
                }
              }
            }
          }
          __pyx_L34:;
        }
      }

      /* "cyext_acv.pyx":2842
 *                                     mean_forest[R_buf[i], 2] += (p_d * value*value_l) / (p_sd) if p_sd != 0 else 0
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 ss = (mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 0])/(mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2]) if mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2] !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 */
      __pyx_t_9 = __pyx_v_N;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_10; __pyx_t_32+=1) {
        __pyx_v_i = __pyx_t_32;

        /* "cyext_acv.pyx":2843
 * 
 *             for i in range(N):
 *                 ss = (mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 0])/(mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2]) if mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2] !=0 else 0             # <<<<<<<<<<<<<<
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 */
        __pyx_t_52 = __pyx_v_i;
        __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_44 = 1;
        __pyx_t_48 = __pyx_v_i;
        __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_43 = 2;
        if (((((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_36 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_44 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_33 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_43 * __pyx_v_mean_forest.strides[1]) )))) != 0.0) != 0)) {
          __pyx_t_35 = __pyx_v_i;
          __pyx_t_57 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_35 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_58 = 1;
          __pyx_t_59 = __pyx_v_i;
          __pyx_t_60 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_59 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_61 = 0;
          __pyx_t_62 = __pyx_v_i;
          __pyx_t_63 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_62 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_64 = 1;
          __pyx_t_65 = __pyx_v_i;
          __pyx_t_66 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_65 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_67 = 2;
          __pyx_t_53 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_57 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_58 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_60 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_61 * __pyx_v_mean_forest.strides[1]) )))) / ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_63 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_64 * __pyx_v_mean_forest.strides[1]) ))) - (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest.data + __pyx_t_66 * __pyx_v_mean_forest.strides[0]) ) + __pyx_t_67 * __pyx_v_mean_forest.strides[1]) )))));
        } else {
          __pyx_t_53 = 0.0;
        }
        __pyx_v_ss = __pyx_t_53;

        /* "cyext_acv.pyx":2844
 *             for i in range(N):
 *                 ss = (mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 0])/(mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2]) if mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2] !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_12 = ((__pyx_v_ss >= __pyx_v_pi_level) != 0);
        if (__pyx_t_12) {
        } else {
          __pyx_t_19 = __pyx_t_12;
          goto __pyx_L136_bool_binop_done;
        }
        __pyx_t_48 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_12 = ((__pyx_v_ss >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )))) != 0);
        __pyx_t_19 = __pyx_t_12;
        __pyx_L136_bool_binop_done:;
        if (__pyx_t_19) {

          /* "cyext_acv.pyx":2845
 *                 ss = (mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 0])/(mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2]) if mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2] !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_48 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

          /* "cyext_acv.pyx":2846
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_48 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":2847
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_34 = __pyx_v_S_size;
          __pyx_t_37 = __pyx_t_34;
          for (__pyx_t_38 = 0; __pyx_t_38 < __pyx_t_37; __pyx_t_38+=1) {
            __pyx_v_s = __pyx_t_38;

            /* "cyext_acv.pyx":2848
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_48 = __pyx_v_s;
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_52 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_33 * __pyx_v_s_star.strides[0]) ) + __pyx_t_52 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":2844
 *             for i in range(N):
 *                 ss = (mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 0])/(mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2]) if mean_forest[R_buf[i], 1] - mean_forest[R_buf[i], 2] !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":2850
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_9 = __pyx_v_N;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_10; __pyx_t_22+=1) {
      __pyx_v_i = __pyx_t_22;

      /* "cyext_acv.pyx":2851
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 * 
 */
      __pyx_t_48 = __pyx_v_i;
      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_19 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_19) {

        /* "cyext_acv.pyx":2852
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 * 
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 2852, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":2854
 *                 r.push_back(R[i])
 * 
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):
 */
        __pyx_t_13 = (__pyx_v_X.shape[1]);
        __pyx_t_48 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_14 = __pyx_t_13;
        for (__pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_32 < __pyx_t_14; __pyx_t_32+=1) {
          __pyx_v_s = __pyx_t_32;

          /* "cyext_acv.pyx":2855
 * 
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                     s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
          __pyx_t_52 = __pyx_v_i;
          __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_52 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_44 = __pyx_v_s;
          *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_33 * __pyx_v_s_star.strides[0]) ) + __pyx_t_44 * __pyx_v_s_star.strides[1]) )) = -1L;
        }

        /* "cyext_acv.pyx":2856
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_48 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_68 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_69 = __pyx_t_68;
        for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_69; __pyx_t_32+=1) {
          __pyx_v_s = __pyx_t_32;

          /* "cyext_acv.pyx":2857
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_48 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_48 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_52 = __pyx_v_s;
          __pyx_t_44 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_43 * __pyx_v_s_star.strides[0]) ) + __pyx_t_52 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_44 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":2851
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 * 
 */
      }
    }

    /* "cyext_acv.pyx":2859
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_30 = __pyx_v_r.size();
    __pyx_t_31 = __pyx_t_30;
    for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_31; __pyx_t_22+=1) {
      __pyx_v_i = __pyx_t_22;

      /* "cyext_acv.pyx":2860
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":2861
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":2863
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_12 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_12) {
    } else {
      goto __pyx_L151_next_and;
    }
    __pyx_t_12 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_19 = __pyx_t_12;
      goto __pyx_L150_bool_binop_done;
    }
    __pyx_L151_next_and:;
    __pyx_t_12 = (__pyx_v_stop != 0);
    __pyx_t_19 = __pyx_t_12;
    __pyx_L150_bool_binop_done:;
    if (__pyx_t_19) {

      /* "cyext_acv.pyx":2864
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L21_break;

      /* "cyext_acv.pyx":2863
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L21_break:;

  /* "cyext_acv.pyx":2866
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_24))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_24);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_24);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_24, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_24, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_24, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_7, __pyx_t_24); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_np); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_24);
  __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_long); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_24); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_np); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_24);
  __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_70 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_long); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_dtype, __pyx_t_70) < 0) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_70); __pyx_t_70 = 0;
  __pyx_t_70 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_24); if (unlikely(!__pyx_t_70)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_70);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_24 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_70);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_70);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_24);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_70 = 0;
  __pyx_t_24 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2637
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_reg_cpp_pa_coal(double[:, :] X, double[:] fX, double tX,             # <<<<<<<<<<<<<<
 *             double[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_24);
  __PYX_XDEC_MEMVIEW(&__pyx_t_27, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_41, 1);
  __Pyx_XDECREF(__pyx_t_70);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cpp_pa_coal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_tree_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_part_l, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_31global_sdp_reg_cpp_pa_coal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_31global_sdp_reg_cpp_pa_coal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_tX;
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_num_threads;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_reg_cpp_pa_coal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_tX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_num_threads,&__pyx_n_s_stop,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 1); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 2); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 3); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 4); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 5); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 6); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 7); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 8); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 9); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 10); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 11); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 12); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, 13); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_reg_cpp_pa_coal") < 0)) __PYX_ERR(0, 2637, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 2637, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 2637, __pyx_L3_error)
    __pyx_v_tX = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_tX == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2637, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 2638, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 2638, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 2639, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 2639, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 2640, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 2640, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[9], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 2640, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[10]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2640, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[13]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_reg_cpp_pa_coal", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2637, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cpp_pa_coal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 2640, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_30global_sdp_reg_cpp_pa_coal(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_num_threads, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_30global_sdp_reg_cpp_pa_coal(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, double __pyx_v_tX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_num_threads, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_reg_cpp_pa_coal", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 2637, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_reg_cpp_pa_coal(__pyx_v_X, __pyx_v_fX, __pyx_v_tX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_num_threads, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_reg_cpp_pa_coal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":2872
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_33global_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  double __pyx_v_ss;
  double __pyx_v_ss_a;
  double __pyx_v_ss_u;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_lm;
  int __pyx_v_lm_s;
  int __pyx_v_max_size;
  int __pyx_v_size;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  CYTHON_UNUSED long __pyx_v_power_set_size;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  long __pyx_t_22;
  long __pyx_t_23;
  int __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_t_26;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_27;
  __Pyx_memviewslice __pyx_t_28 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_29 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_30;
  std::vector<std::vector<long> > ::size_type __pyx_t_31;
  std::vector<long> ::size_type __pyx_t_32;
  std::vector<long> ::size_type __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  unsigned int __pyx_t_38;
  unsigned int __pyx_t_39;
  int __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  int __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  unsigned int __pyx_t_46;
  int __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  Py_ssize_t __pyx_t_51;
  int __pyx_t_52;
  double __pyx_t_53;
  PyObject *__pyx_t_54 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf", 0);

  /* "cyext_acv.pyx":2878
 *             int minimal, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":2879
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":2880
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":2881
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":2888
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 2888, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":2891
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, leaf_idx_trees.shape[1], N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_leaf_idx_trees.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":2895
 *     cdef double[:] sdp
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2896
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":2904
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 2904, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_len_s_star = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":2909
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2909, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":2910
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":2911
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":2912
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2912, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":2915
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 2915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_R_buf = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":2917
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2917, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_14) {

    /* "cyext_acv.pyx":2918
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 2918, __pyx_L1_error)
    }
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2918, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_ci = __pyx_t_17;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 2918, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_18 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2918, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_cj = __pyx_t_20;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 2918, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 2918, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":2919
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in range(m)]
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 2919, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_21 = (__pyx_t_14 != 0);
      if (__pyx_t_21) {
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6);
        PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
        __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 2919, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_v_C); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":2917
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":2921
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2921, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 2921, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":2923
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2923, __pyx_L1_error)
  __pyx_v_m = __pyx_t_15;

  /* "cyext_acv.pyx":2924
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(minimal, m + 1):
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_power = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "cyext_acv.pyx":2925
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(minimal, m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":2926
 *     power = []
 *     max_size = 0
 *     for size in range(minimal, m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_22 = (__pyx_v_m + 1);
  __pyx_t_23 = __pyx_t_22;
  for (__pyx_t_13 = __pyx_v_minimal; __pyx_t_13 < __pyx_t_23; __pyx_t_13+=1) {
    __pyx_v_size = __pyx_t_13;

    /* "cyext_acv.pyx":2927
 *     max_size = 0
 *     for size in range(minimal, m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":2928
 *     for size in range(minimal, m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_itertools); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_combinations); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_24 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_24 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2928, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2928, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_24, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_24, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
      __pyx_t_25 = NULL;
    } else {
      __pyx_t_15 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_25 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 2928, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_25)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 2928, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2928, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 2928, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2928, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_25(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2928, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "cyext_acv.pyx":2929
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_7); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 2929, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "cyext_acv.pyx":2930
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":2928
 *     for size in range(minimal, m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyext_acv.pyx":2931
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 2931, __pyx_L1_error)

    /* "cyext_acv.pyx":2932
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_21 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":2933
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L16_break;

      /* "cyext_acv.pyx":2932
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L16_break:;

  /* "cyext_acv.pyx":2935
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2935, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_27;

  /* "cyext_acv.pyx":2937
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 2937, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_28;
  __pyx_t_28.memview = NULL;
  __pyx_t_28.data = NULL;

  /* "cyext_acv.pyx":2940
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":2941
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_u_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":2942
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 2942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_d_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":2945
 *     cdef double n, n_u, n_d
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":2946
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(len(power))):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 2946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_S = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":2948
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(len(power))):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_power); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2948, __pyx_L1_error)
  __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_1 = __pyx_t_7; __Pyx_INCREF(__pyx_t_1); __pyx_t_15 = 0;
    __pyx_t_25 = NULL;
  } else {
    __pyx_t_15 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 2948, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_25)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 2948, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 2948, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2948, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_25(__pyx_t_1);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2948, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2948, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_s_0 = __pyx_t_13;

    /* "cyext_acv.pyx":2949
 * 
 *     for s_0 in tqdm(range(len(power))):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_30 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_31 = __pyx_t_30;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_31; __pyx_t_13+=1) {
      __pyx_v_s_1 = __pyx_t_13;

      /* "cyext_acv.pyx":2950
 *     for s_0 in tqdm(range(len(power))):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_32 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_33 = __pyx_t_32;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_33; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":2951
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_34 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":2953
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":2954
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":2955
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":2956
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 * 
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":2957
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
        __pyx_t_34 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":2959
 *                 R_buf[i] = R[i]
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 */
      __pyx_t_24 = __pyx_v_n_trees;
      __pyx_t_35 = __pyx_t_24;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
        __pyx_v_b = __pyx_t_36;

        /* "cyext_acv.pyx":2960
 * 
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 */
        __pyx_t_34 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_34 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":2961
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 */
        __pyx_t_11 = __pyx_v_nb_leaf;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_12; __pyx_t_37+=1) {
          __pyx_v_leaf_numb = __pyx_t_37;

          /* "cyext_acv.pyx":2962
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 */
          __pyx_t_38 = __pyx_v_N;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_i = __pyx_t_40;

            /* "cyext_acv.pyx":2963
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_41 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_41 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_43 * __pyx_v_p_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":2964
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_43 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_43 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_41 * __pyx_v_p_u_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":2965
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_41 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_41 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_43 * __pyx_v_p_d_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":2966
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):             # <<<<<<<<<<<<<<
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 */
            for (__pyx_t_44 = 0; __pyx_t_44 < 3; __pyx_t_44+=1) {
              __pyx_v_j = __pyx_t_44;

              /* "cyext_acv.pyx":2967
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
              __pyx_t_34 = __pyx_v_i;
              __pyx_t_43 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_45 = __pyx_v_j;
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) )) = 0.0;
            }
          }
        }
      }

      /* "cyext_acv.pyx":2969
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
      __pyx_t_24 = __pyx_v_n_trees;
      __pyx_t_35 = __pyx_t_24;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
        __pyx_v_b = __pyx_t_36;

        /* "cyext_acv.pyx":2970
 * 
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):
 */
        __pyx_t_34 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_34 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":2971
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         o_all = 0
 */
        {
            #ifdef WITH_THREAD
            PyThreadState *_save;
            Py_UNBLOCK_THREADS
            __Pyx_FastGIL_Remember();
            #endif
            /*try:*/ {
              __pyx_t_11 = __pyx_v_nb_leaf;
              if ((1 == 0)) abort();
              {
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   (x)
                      #define unlikely(x) (x)
                  #endif
                  __pyx_t_38 = (__pyx_t_11 - 0 + 1 - 1/abs(1)) / 1;
                  if (__pyx_t_38 > 0)
                  {
                      #ifdef _OPENMP
                      #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_14, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_21, __pyx_t_34, __pyx_t_37, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53)
                      #endif /* _OPENMP */
                      {
                          #ifdef _OPENMP
                          #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) schedule(dynamic)
                          #endif /* _OPENMP */
                          for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_38; __pyx_t_12++){
                              {
                                  __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_12);
                                  /* Initialize private variables to invalid values */
                                  __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                  __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                  __pyx_v_i = ((int)0xbad0bad0);
                                  __pyx_v_j = ((int)0xbad0bad0);
                                  __pyx_v_lm = ((int)0xbad0bad0);
                                  __pyx_v_lm_s = ((int)0xbad0bad0);
                                  __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                  __pyx_v_s = ((int)0xbad0bad0);

                                  /* "cyext_acv.pyx":2972
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
                                  __pyx_t_39 = __pyx_v_N;
                                  __pyx_t_46 = __pyx_t_39;
                                  for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_46; __pyx_t_37+=1) {
                                    __pyx_v_i = __pyx_t_37;

                                    /* "cyext_acv.pyx":2973
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":2974
 *                     for i in range(N):
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_40 = __pyx_v_S_size;
                                    __pyx_t_44 = __pyx_t_40;
                                    for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_44; __pyx_t_47+=1) {
                                      __pyx_v_s = __pyx_t_47;

                                      /* "cyext_acv.pyx":2975
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      __pyx_t_34 = __pyx_v_i;
                                      __pyx_t_45 = __pyx_v_s;
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_45 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_43 = __pyx_v_s;
                                      __pyx_t_48 = __pyx_v_b;
                                      __pyx_t_49 = __pyx_v_leaf_numb;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_51 = 1;
                                      __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_41 * __pyx_v_X.strides[0]) ) + __pyx_t_42 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_14) {
                                      } else {
                                        __pyx_t_21 = __pyx_t_14;
                                        goto __pyx_L52_bool_binop_done;
                                      }
                                      __pyx_t_43 = __pyx_v_i;
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_48 = __pyx_v_s;
                                      __pyx_t_45 = __pyx_v_b;
                                      __pyx_t_34 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_41 = 0;
                                      __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_50 * __pyx_v_X.strides[0]) ) + __pyx_t_49 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_21 = __pyx_t_14;
                                      __pyx_L52_bool_binop_done:;
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":2976
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                         if o_all > 0:
 *                             continue
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":2975
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":2977
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    __pyx_t_21 = ((__pyx_v_o_all > 0) != 0);
                                    if (__pyx_t_21) {

                                      /* "cyext_acv.pyx":2978
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         p = 0
 */
                                      goto __pyx_L47_continue;

                                      /* "cyext_acv.pyx":2977
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":2980
 *                             continue
 * 
 *                         p = 0             # <<<<<<<<<<<<<<
 *                         p_u = 0
 *                         p_d = 0
 */
                                    __pyx_v_p = 0;

                                    /* "cyext_acv.pyx":2981
 * 
 *                         p = 0
 *                         p_u = 0             # <<<<<<<<<<<<<<
 *                         p_d = 0
 *                         p_s = 0
 */
                                    __pyx_v_p_u = 0;

                                    /* "cyext_acv.pyx":2982
 *                         p = 0
 *                         p_u = 0
 *                         p_d = 0             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_su = 0
 */
                                    __pyx_v_p_d = 0;

                                    /* "cyext_acv.pyx":2983
 *                         p_u = 0
 *                         p_d = 0
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_su = 0
 *                         p_sd = 0
 */
                                    __pyx_v_p_s = 0;

                                    /* "cyext_acv.pyx":2984
 *                         p_d = 0
 *                         p_s = 0
 *                         p_su = 0             # <<<<<<<<<<<<<<
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 */
                                    __pyx_v_p_su = 0;

                                    /* "cyext_acv.pyx":2985
 *                         p_s = 0
 *                         p_su = 0
 *                         p_sd = 0             # <<<<<<<<<<<<<<
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 */
                                    __pyx_v_p_sd = 0;

                                    /* "cyext_acv.pyx":2986
 *                         p_su = 0
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
                                    __pyx_t_16 = (__pyx_v_data.shape[0]);
                                    __pyx_t_17 = __pyx_t_16;
                                    for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_17; __pyx_t_40+=1) {
                                      __pyx_v_j = __pyx_t_40;

                                      /* "cyext_acv.pyx":2987
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             b_it = 0
 *                             lm = 0
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":2988
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             lm = 0
 *                             lm_s = 0
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":2989
 *                             a_it = 0
 *                             b_it = 0
 *                             lm = 0             # <<<<<<<<<<<<<<
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 */
                                      __pyx_v_lm = 0;

                                      /* "cyext_acv.pyx":2990
 *                             b_it = 0
 *                             lm = 0
 *                             lm_s = 0             # <<<<<<<<<<<<<<
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                      __pyx_v_lm_s = 0;

                                      /* "cyext_acv.pyx":2991
 *                             lm = 0
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_18 = (__pyx_v_data.shape[1]);
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_19; __pyx_t_44+=1) {
                                        __pyx_v_s = __pyx_t_44;

                                        /* "cyext_acv.pyx":2992
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        __pyx_t_48 = __pyx_v_j;
                                        __pyx_t_41 = __pyx_v_s;
                                        __pyx_t_42 = __pyx_v_b;
                                        __pyx_t_34 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = __pyx_v_s;
                                        __pyx_t_51 = 1;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_48 * __pyx_v_data.strides[0]) ) + __pyx_t_41 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_14) {
                                        } else {
                                          __pyx_t_21 = __pyx_t_14;
                                          goto __pyx_L60_bool_binop_done;
                                        }
                                        __pyx_t_51 = __pyx_v_j;
                                        __pyx_t_45 = __pyx_v_s;
                                        __pyx_t_34 = __pyx_v_b;
                                        __pyx_t_42 = __pyx_v_leaf_numb;
                                        __pyx_t_41 = __pyx_v_s;
                                        __pyx_t_48 = 0;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_51 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_21 = __pyx_t_14;
                                        __pyx_L60_bool_binop_done:;
                                        if (__pyx_t_21) {

                                          /* "cyext_acv.pyx":2993
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":2992
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":2994
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1
 */
                                      __pyx_t_44 = __pyx_v_S_size;
                                      __pyx_t_47 = __pyx_t_44;
                                      for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_47; __pyx_t_52+=1) {
                                        __pyx_v_s = __pyx_t_52;

                                        /* "cyext_acv.pyx":2995
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        __pyx_t_48 = __pyx_v_s;
                                        __pyx_t_41 = __pyx_v_j;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_34 = __pyx_v_s;
                                        __pyx_t_45 = __pyx_v_b;
                                        __pyx_t_51 = __pyx_v_leaf_numb;
                                        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_49 = 1;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_41 * __pyx_v_data.strides[0]) ) + __pyx_t_42 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_43 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_14) {
                                        } else {
                                          __pyx_t_21 = __pyx_t_14;
                                          goto __pyx_L65_bool_binop_done;
                                        }
                                        __pyx_t_34 = __pyx_v_s;
                                        __pyx_t_49 = __pyx_v_j;
                                        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_51 = __pyx_v_s;
                                        __pyx_t_45 = __pyx_v_b;
                                        __pyx_t_48 = __pyx_v_leaf_numb;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_41 = 0;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_49 * __pyx_v_data.strides[0]) ) + __pyx_t_43 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_21 = __pyx_t_14;
                                        __pyx_L65_bool_binop_done:;
                                        if (__pyx_t_21) {

                                          /* "cyext_acv.pyx":2996
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if a_it == data.shape[1]:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":2995
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":2998
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      __pyx_t_21 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":2999
 * 
 *                             if a_it == data.shape[1]:
 *                                 lm = 1             # <<<<<<<<<<<<<<
 * 
 *                             if b_it == S_size:
 */
                                        __pyx_v_lm = 1;

                                        /* "cyext_acv.pyx":2998
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3001
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      __pyx_t_21 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3002
 * 
 *                             if b_it == S_size:
 *                                 lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                             p += lm
 */
                                        __pyx_v_lm_s = 1;

                                        /* "cyext_acv.pyx":3001
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3004
 *                                 lm_s = 1
 * 
 *                             p += lm             # <<<<<<<<<<<<<<
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 */
                                      __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                      /* "cyext_acv.pyx":3005
 * 
 *                             p += lm
 *                             p_s += lm_s             # <<<<<<<<<<<<<<
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 */
                                      __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                      /* "cyext_acv.pyx":3006
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                      __pyx_t_51 = __pyx_v_i;
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_51 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = __pyx_v_j;
                                      __pyx_t_21 = (((*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_41 * __pyx_v_fX.strides[0]) ))) == (*((long *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_42 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3007
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm             # <<<<<<<<<<<<<<
 *                                 p_su += lm_s
 *                             else:
 */
                                        __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3008
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 *                                 p_su += lm_s             # <<<<<<<<<<<<<<
 *                             else:
 *                                 p_d += lm
 */
                                        __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                        /* "cyext_acv.pyx":3006
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                        goto __pyx_L69;
                                      }

                                      /* "cyext_acv.pyx":3010
 *                                 p_su += lm_s
 *                             else:
 *                                 p_d += lm             # <<<<<<<<<<<<<<
 *                                 p_sd += lm_s
 * 
 */
                                      /*else*/ {
                                        __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3011
 *                             else:
 *                                 p_d += lm
 *                                 p_sd += lm_s             # <<<<<<<<<<<<<<
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 */
                                        __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);
                                      }
                                      __pyx_L69:;
                                    }

                                    /* "cyext_acv.pyx":3013
 *                                 p_sd += lm_s
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_51 = __pyx_v_leaf_numb;
                                      __pyx_t_41 = __pyx_v_i;
                                      __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_45 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_48 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_45 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_43 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_51 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_43 = 0;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_51 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3014
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_41 = __pyx_v_b;
                                      __pyx_t_43 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = __pyx_v_i;
                                      __pyx_t_51 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_48 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_41 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_51 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p_u * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_48 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_45 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_42 = __pyx_v_i;
                                    __pyx_t_51 = __pyx_v_b;
                                    __pyx_t_43 = __pyx_v_leaf_numb;
                                    __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_45 = 1;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_51 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3015
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_45 = __pyx_v_leaf_numb;
                                      __pyx_t_41 = __pyx_v_i;
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_51 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_45 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_43 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p_d * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_51 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_48 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_43 = __pyx_v_b;
                                    __pyx_t_45 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_48 = 2;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3017
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_48 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_n.strides[2]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3018
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_45 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_45 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_u_n.strides[2]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3019
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_48 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_d_n.strides[2]) )) += __pyx_t_53;
                                    __pyx_L47_continue:;
                                  }
                              }
                          }
                      }
                  }
              }
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   __builtin_expect(!!(x), 1)
                  #define unlikely(x) __builtin_expect(!!(x), 0)
              #endif
            }

            /* "cyext_acv.pyx":2971
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         o_all = 0
 */
            /*finally:*/ {
              /*normal exit:*/{
                #ifdef WITH_THREAD
                __Pyx_FastGIL_Forget();
                Py_BLOCK_THREADS
                #endif
                goto __pyx_L42;
              }
              __pyx_L42:;
            }
        }
      }

      /* "cyext_acv.pyx":3021
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 ss_u = 0
 *                 ss_d = 0
 */
      __pyx_t_38 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_38;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3022
 * 
 *             for i in range(N):
 *                 ss_u = 0             # <<<<<<<<<<<<<<
 *                 ss_d = 0
 *                 ss_a = 0
 */
        __pyx_v_ss_u = 0.0;

        /* "cyext_acv.pyx":3023
 *             for i in range(N):
 *                 ss_u = 0
 *                 ss_d = 0             # <<<<<<<<<<<<<<
 *                 ss_a = 0
 *                 for b in range(n_trees):
 */
        __pyx_v_ss_d = 0.0;

        /* "cyext_acv.pyx":3024
 *                 ss_u = 0
 *                 ss_d = 0
 *                 ss_a = 0             # <<<<<<<<<<<<<<
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 */
        __pyx_v_ss_a = 0.0;

        /* "cyext_acv.pyx":3025
 *                 ss_d = 0
 *                 ss_a = 0
 *                 for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0
 */
        __pyx_t_35 = __pyx_v_n_trees;
        __pyx_t_36 = __pyx_t_35;
        for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
          __pyx_v_b = __pyx_t_37;

          /* "cyext_acv.pyx":3026
 *                 ss_a = 0
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                     n = 0
 *                     n_u = 0
 */
          __pyx_t_41 = __pyx_v_b;
          __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_41 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":3027
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0             # <<<<<<<<<<<<<<
 *                     n_u = 0
 *                     n_d = 0
 */
          __pyx_v_n = 0.0;

          /* "cyext_acv.pyx":3028
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0
 *                     n_u = 0             # <<<<<<<<<<<<<<
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):
 */
          __pyx_v_n_u = 0.0;

          /* "cyext_acv.pyx":3029
 *                     n = 0
 *                     n_u = 0
 *                     n_d = 0             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 */
          __pyx_v_n_d = 0.0;

          /* "cyext_acv.pyx":3030
 *                     n_u = 0
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 */
          __pyx_t_11 = __pyx_v_nb_leaf;
          __pyx_t_39 = __pyx_t_11;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_leaf_numb = __pyx_t_40;

            /* "cyext_acv.pyx":3031
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_45 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_45 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_n.strides[2]) ))));

            /* "cyext_acv.pyx":3032
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 * 
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_48 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_48 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_u_n.strides[2]) ))));

            /* "cyext_acv.pyx":3033
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 * 
 *                     for leaf_numb in range(nb_leaf):
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_45 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_45 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_d_n.strides[2]) ))));
          }

          /* "cyext_acv.pyx":3035
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 * 
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 */
          __pyx_t_11 = __pyx_v_nb_leaf;
          __pyx_t_39 = __pyx_t_11;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_leaf_numb = __pyx_t_40;

            /* "cyext_acv.pyx":3036
 * 
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 */
            if (((__pyx_v_n != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_48 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_43 = 0;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_53);

            /* "cyext_acv.pyx":3037
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 * 
 */
            if (((__pyx_v_n_u != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_43 = __pyx_v_b;
              __pyx_t_45 = __pyx_v_leaf_numb;
              __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_48 = 1;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_u);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_53);

            /* "cyext_acv.pyx":3038
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
            if (((__pyx_v_n_d != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_48 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_43 = 2;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_d);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_53);
          }
        }

        /* "cyext_acv.pyx":3040
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 */
        if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
          __pyx_t_53 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
        } else {
          __pyx_t_53 = 0.0;
        }
        __pyx_v_ss = __pyx_t_53;

        /* "cyext_acv.pyx":3041
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_41 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_21 = ((__pyx_v_ss >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":3042
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

          /* "cyext_acv.pyx":3043
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3044
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_35 = __pyx_v_S_size;
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_s = __pyx_t_37;

            /* "cyext_acv.pyx":3045
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_41 = __pyx_v_s;
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_42 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_45 * __pyx_v_s_star.strides[0]) ) + __pyx_t_42 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_41 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3041
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":3047
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_21 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":3048
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":3049
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3050
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_35 = __pyx_v_S_size;
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_s = __pyx_t_37;

            /* "cyext_acv.pyx":3051
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_41 = __pyx_v_s;
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_45 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_45 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_41 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3052
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_16 = (__pyx_v_X.shape[1]);
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_17 = __pyx_t_16;
          for (__pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_35 < __pyx_t_17; __pyx_t_35+=1) {
            __pyx_v_s = __pyx_t_35;

            /* "cyext_acv.pyx":3053
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_45 = __pyx_v_i;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_45 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_48 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_48 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":3047
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":3055
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_38 = __pyx_v_N;
    __pyx_t_12 = __pyx_t_38;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3056
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_41 = __pyx_v_i;
      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_21 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_21) {

        /* "cyext_acv.pyx":3057
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3057, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3058
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_41 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_23 = __pyx_t_22;
        for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
          __pyx_v_s = __pyx_t_24;

          /* "cyext_acv.pyx":3059
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_45 = __pyx_v_s;
          __pyx_t_48 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_43 * __pyx_v_s_star.strides[0]) ) + __pyx_t_45 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_48 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":3056
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":3061
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_32 = __pyx_v_r.size();
    __pyx_t_33 = __pyx_t_32;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_33; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3062
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":3063
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":3065
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_14 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_14) {
    } else {
      goto __pyx_L97_next_and;
    }
    __pyx_t_14 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_14) {
    } else {
      __pyx_t_21 = __pyx_t_14;
      goto __pyx_L96_bool_binop_done;
    }
    __pyx_L97_next_and:;
    __pyx_t_14 = (__pyx_v_stop != 0);
    __pyx_t_21 = __pyx_t_14;
    __pyx_L96_bool_binop_done:;
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":3066
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L21_break;

      /* "cyext_acv.pyx":3065
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":2948
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(len(power))):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L21_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":3068
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_54 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_54) < 0) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __pyx_t_54 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_54);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_54);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_3);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_54 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":2872
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_28, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_29, 1);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_33global_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_33global_sdp_clf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_clf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 1); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 2); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 3); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 4); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 5); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 6); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 7); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 8); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 9); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 10); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 11); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, 12); __PYX_ERR(0, 2872, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_clf") < 0)) __PYX_ERR(0, 2872, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 2872, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 2872, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 2873, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 2873, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 2874, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 2874, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 2875, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 2875, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 2875, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[9]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2875, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2876, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[12]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2876, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_clf", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2872, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 2875, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_32global_sdp_clf(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_32global_sdp_clf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 2872, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_clf(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3075
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf_ptrees(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_35global_sdp_clf_ptrees(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf_ptrees(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  CYTHON_UNUSED unsigned int __pyx_v_N_b;
  int __pyx_v_n_trees;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  double __pyx_v_ss;
  double __pyx_v_ss_a;
  double __pyx_v_ss_u;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_lm;
  int __pyx_v_lm_s;
  int __pyx_v_max_size;
  int __pyx_v_size;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  long __pyx_t_22;
  long __pyx_t_23;
  int __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_t_26;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_27;
  __Pyx_memviewslice __pyx_t_28 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<std::vector<long> > ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  std::vector<long> ::size_type __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  int __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  unsigned int __pyx_t_37;
  unsigned int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  Py_ssize_t __pyx_t_44;
  int __pyx_t_45;
  int __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  int __pyx_t_51;
  double __pyx_t_52;
  PyObject *__pyx_t_53 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf_ptrees", 0);

  /* "cyext_acv.pyx":3081
 *             int minimal, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int N_b = X.shape[0]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3082
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N_b = X.shape[0]
 * 
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":3083
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int N_b = X.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_N_b = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3085
 *     cdef unsigned int N_b = X.shape[0]
 * 
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":3092
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":3095
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, leaf_idx_trees.shape[1], N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_leaf_idx_trees.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 3095, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":3099
 *     cdef double[:] sdp
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3100
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3108
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_len_s_star = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3113
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3113, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3114
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":3115
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":3116
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3116, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3119
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_R_buf = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3121
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 3121, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_14) {

    /* "cyext_acv.pyx":3122
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 3122, __pyx_L1_error)
    }
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3122, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_ci = __pyx_t_17;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 3122, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_18 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3122, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_cj = __pyx_t_20;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 3122, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 3122, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":3123
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in range(m)]
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_14 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3123, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_21 = (__pyx_t_14 != 0);
      if (__pyx_t_21) {
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3123, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3123, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6);
        PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
        __pyx_t_6 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 3123, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_v_C); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":3121
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":3125
 *         va_id = [[i] for i in range(m) if i not in remove_va] + C
 *     else:
 *         va_id = [[i] for i in range(m)]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 3125, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_va_id = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":3127
 *         va_id = [[i] for i in range(m)]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3127, __pyx_L1_error)
  __pyx_v_m = __pyx_t_15;

  /* "cyext_acv.pyx":3128
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_power = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "cyext_acv.pyx":3129
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":3130
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_22 = (__pyx_v_m + 1);
  __pyx_t_23 = __pyx_t_22;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_23; __pyx_t_13+=1) {
    __pyx_v_size = __pyx_t_13;

    /* "cyext_acv.pyx":3131
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "cyext_acv.pyx":3132
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_itertools); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_combinations); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_24 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_24 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3132, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_6};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3132, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_24, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_24, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
      __pyx_t_25 = NULL;
    } else {
      __pyx_t_15 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_25 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 3132, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_25)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3132, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3132, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3132, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3132, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_25(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 3132, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "cyext_acv.pyx":3133
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_7); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 3133, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "cyext_acv.pyx":3134
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":3132
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyext_acv.pyx":3135
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 3135, __pyx_L1_error)

    /* "cyext_acv.pyx":3136
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_21 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":3137
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L16_break;

      /* "cyext_acv.pyx":3136
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L16_break:;

  /* "cyext_acv.pyx":3139
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3139, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_27;

  /* "cyext_acv.pyx":3141
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 3141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_28;
  __pyx_t_28.memview = NULL;
  __pyx_t_28.data = NULL;

  /* "cyext_acv.pyx":3144
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":3145
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_S = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3147
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __pyx_t_6 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_15 = 0;
    __pyx_t_25 = NULL;
  } else {
    __pyx_t_15 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 3147, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_25)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3147, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3147, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_25(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3147, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3147, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_s_0 = __pyx_t_13;

    /* "cyext_acv.pyx":3148
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_29 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_30 = __pyx_t_29;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_30; __pyx_t_13+=1) {
      __pyx_v_s_1 = __pyx_t_13;

      /* "cyext_acv.pyx":3149
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_31 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_32 = __pyx_t_31;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_32; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3150
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":3152
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":3153
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":3154
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":3155
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 *                 # for j in range(3):
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3156
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 *                 # for j in range(3):
 *                 #   mean_forest[R_buf[i], j] = 0
 */
        __pyx_t_33 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":3160
 *                 #   mean_forest[R_buf[i], j] = 0
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 */
      __pyx_t_24 = __pyx_v_n_trees;
      __pyx_t_34 = __pyx_t_24;
      for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_34; __pyx_t_35+=1) {
        __pyx_v_b = __pyx_t_35;

        /* "cyext_acv.pyx":3161
 * 
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 */
        __pyx_t_33 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_33 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":3162
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         for j in range(3):
 */
        __pyx_t_11 = __pyx_v_nb_leaf;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_12; __pyx_t_36+=1) {
          __pyx_v_leaf_numb = __pyx_t_36;

          /* "cyext_acv.pyx":3163
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 */
          __pyx_t_37 = __pyx_v_N;
          __pyx_t_38 = __pyx_t_37;
          for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
            __pyx_v_i = __pyx_t_39;

            /* "cyext_acv.pyx":3164
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 *                         for j in range(3):             # <<<<<<<<<<<<<<
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 */
            for (__pyx_t_40 = 0; __pyx_t_40 < 3; __pyx_t_40+=1) {
              __pyx_v_j = __pyx_t_40;

              /* "cyext_acv.pyx":3165
 *                     for i in range(N):
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0             # <<<<<<<<<<<<<<
 * 
 *             for b in prange(n_trees, nogil=True, schedule='dynamic'):
 */
              __pyx_t_33 = __pyx_v_i;
              __pyx_t_41 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_44 = __pyx_v_j;
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_41 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[3]) )) = 0.0;
            }
          }
        }
      }

      /* "cyext_acv.pyx":3167
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 *             for b in prange(n_trees, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {
            __pyx_t_24 = __pyx_v_n_trees;
            if ((1 == 0)) abort();
            {
                #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                    #undef likely
                    #undef unlikely
                    #define likely(x)   (x)
                    #define unlikely(x) (x)
                #endif
                __pyx_t_35 = (__pyx_t_24 - 0 + 1 - 1/abs(1)) / 1;
                if (__pyx_t_35 > 0)
                {
                    #ifdef _OPENMP
                    #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_11, __pyx_t_12, __pyx_t_14, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_21, __pyx_t_33, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52)
                    #endif /* _OPENMP */
                    {
                        #ifdef _OPENMP
                        #pragma omp for lastprivate(__pyx_v_a_it) firstprivate(__pyx_v_b) lastprivate(__pyx_v_b) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_nb_leaf) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) schedule(dynamic)
                        #endif /* _OPENMP */
                        for (__pyx_t_34 = 0; __pyx_t_34 < __pyx_t_35; __pyx_t_34++){
                            {
                                __pyx_v_b = (int)(0 + 1 * __pyx_t_34);
                                /* Initialize private variables to invalid values */
                                __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                __pyx_v_i = ((int)0xbad0bad0);
                                __pyx_v_j = ((int)0xbad0bad0);
                                __pyx_v_leaf_numb = ((int)0xbad0bad0);
                                __pyx_v_lm = ((int)0xbad0bad0);
                                __pyx_v_lm_s = ((int)0xbad0bad0);
                                __pyx_v_nb_leaf = ((unsigned int)0xbad0bad0);
                                __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                __pyx_v_s = ((int)0xbad0bad0);

                                /* "cyext_acv.pyx":3168
 * 
 *             for b in prange(n_trees, nogil=True, schedule='dynamic'):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 */
                                __pyx_t_33 = __pyx_v_b;
                                __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_33 * __pyx_v_leaves_nb.strides[0]) )));

                                /* "cyext_acv.pyx":3169
 *             for b in prange(n_trees, nogil=True, schedule='dynamic'):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         o_all = 0
 */
                                __pyx_t_11 = __pyx_v_nb_leaf;
                                __pyx_t_12 = __pyx_t_11;
                                for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_12; __pyx_t_36+=1) {
                                  __pyx_v_leaf_numb = __pyx_t_36;

                                  /* "cyext_acv.pyx":3170
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
                                  __pyx_t_37 = __pyx_v_N;
                                  __pyx_t_38 = __pyx_t_37;
                                  for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
                                    __pyx_v_i = __pyx_t_39;

                                    /* "cyext_acv.pyx":3171
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":3172
 *                     for i in range(N):
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_40 = __pyx_v_S_size;
                                    __pyx_t_45 = __pyx_t_40;
                                    for (__pyx_t_46 = 0; __pyx_t_46 < __pyx_t_45; __pyx_t_46+=1) {
                                      __pyx_v_s = __pyx_t_46;

                                      /* "cyext_acv.pyx":3173
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      __pyx_t_33 = __pyx_v_i;
                                      __pyx_t_44 = __pyx_v_s;
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_44 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_41 = __pyx_v_s;
                                      __pyx_t_47 = __pyx_v_b;
                                      __pyx_t_48 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_41 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_50 = 1;
                                      __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_43 * __pyx_v_X.strides[0]) ) + __pyx_t_42 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_14) {
                                      } else {
                                        __pyx_t_21 = __pyx_t_14;
                                        goto __pyx_L52_bool_binop_done;
                                      }
                                      __pyx_t_41 = __pyx_v_i;
                                      __pyx_t_50 = __pyx_v_s;
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_47 = __pyx_v_s;
                                      __pyx_t_44 = __pyx_v_b;
                                      __pyx_t_33 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_47 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_43 = 0;
                                      __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_49 * __pyx_v_X.strides[0]) ) + __pyx_t_48 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_43 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_21 = __pyx_t_14;
                                      __pyx_L52_bool_binop_done:;
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3174
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                         if o_all > 0:
 *                             continue
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":3173
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":3175
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    __pyx_t_21 = ((__pyx_v_o_all > 0) != 0);
                                    if (__pyx_t_21) {

                                      /* "cyext_acv.pyx":3176
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         p = 0
 */
                                      goto __pyx_L47_continue;

                                      /* "cyext_acv.pyx":3175
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":3178
 *                             continue
 * 
 *                         p = 0             # <<<<<<<<<<<<<<
 *                         p_u = 0
 *                         p_d = 0
 */
                                    __pyx_v_p = 0;

                                    /* "cyext_acv.pyx":3179
 * 
 *                         p = 0
 *                         p_u = 0             # <<<<<<<<<<<<<<
 *                         p_d = 0
 *                         p_s = 0
 */
                                    __pyx_v_p_u = 0;

                                    /* "cyext_acv.pyx":3180
 *                         p = 0
 *                         p_u = 0
 *                         p_d = 0             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_su = 0
 */
                                    __pyx_v_p_d = 0;

                                    /* "cyext_acv.pyx":3181
 *                         p_u = 0
 *                         p_d = 0
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_su = 0
 *                         p_sd = 0
 */
                                    __pyx_v_p_s = 0;

                                    /* "cyext_acv.pyx":3182
 *                         p_d = 0
 *                         p_s = 0
 *                         p_su = 0             # <<<<<<<<<<<<<<
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 */
                                    __pyx_v_p_su = 0;

                                    /* "cyext_acv.pyx":3183
 *                         p_s = 0
 *                         p_su = 0
 *                         p_sd = 0             # <<<<<<<<<<<<<<
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 */
                                    __pyx_v_p_sd = 0;

                                    /* "cyext_acv.pyx":3184
 *                         p_su = 0
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
                                    __pyx_t_16 = (__pyx_v_data.shape[0]);
                                    __pyx_t_17 = __pyx_t_16;
                                    for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_17; __pyx_t_40+=1) {
                                      __pyx_v_j = __pyx_t_40;

                                      /* "cyext_acv.pyx":3185
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             b_it = 0
 *                             lm = 0
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":3186
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             lm = 0
 *                             lm_s = 0
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":3187
 *                             a_it = 0
 *                             b_it = 0
 *                             lm = 0             # <<<<<<<<<<<<<<
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 */
                                      __pyx_v_lm = 0;

                                      /* "cyext_acv.pyx":3188
 *                             b_it = 0
 *                             lm = 0
 *                             lm_s = 0             # <<<<<<<<<<<<<<
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                      __pyx_v_lm_s = 0;

                                      /* "cyext_acv.pyx":3189
 *                             lm = 0
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_18 = (__pyx_v_data.shape[1]);
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_19; __pyx_t_45+=1) {
                                        __pyx_v_s = __pyx_t_45;

                                        /* "cyext_acv.pyx":3190
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        __pyx_t_47 = __pyx_v_j;
                                        __pyx_t_43 = __pyx_v_s;
                                        __pyx_t_42 = __pyx_v_b;
                                        __pyx_t_33 = __pyx_v_leaf_numb;
                                        __pyx_t_44 = __pyx_v_s;
                                        __pyx_t_50 = 1;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_47 * __pyx_v_data.strides[0]) ) + __pyx_t_43 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_14) {
                                        } else {
                                          __pyx_t_21 = __pyx_t_14;
                                          goto __pyx_L60_bool_binop_done;
                                        }
                                        __pyx_t_50 = __pyx_v_j;
                                        __pyx_t_44 = __pyx_v_s;
                                        __pyx_t_33 = __pyx_v_b;
                                        __pyx_t_42 = __pyx_v_leaf_numb;
                                        __pyx_t_43 = __pyx_v_s;
                                        __pyx_t_47 = 0;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_50 * __pyx_v_data.strides[0]) ) + __pyx_t_44 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_43 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_21 = __pyx_t_14;
                                        __pyx_L60_bool_binop_done:;
                                        if (__pyx_t_21) {

                                          /* "cyext_acv.pyx":3191
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":3190
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":3192
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1
 */
                                      __pyx_t_45 = __pyx_v_S_size;
                                      __pyx_t_46 = __pyx_t_45;
                                      for (__pyx_t_51 = 0; __pyx_t_51 < __pyx_t_46; __pyx_t_51+=1) {
                                        __pyx_v_s = __pyx_t_51;

                                        /* "cyext_acv.pyx":3193
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        __pyx_t_47 = __pyx_v_s;
                                        __pyx_t_43 = __pyx_v_j;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_47 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_33 = __pyx_v_s;
                                        __pyx_t_44 = __pyx_v_b;
                                        __pyx_t_50 = __pyx_v_leaf_numb;
                                        __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_48 = 1;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_43 * __pyx_v_data.strides[0]) ) + __pyx_t_42 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_14) {
                                        } else {
                                          __pyx_t_21 = __pyx_t_14;
                                          goto __pyx_L65_bool_binop_done;
                                        }
                                        __pyx_t_33 = __pyx_v_s;
                                        __pyx_t_48 = __pyx_v_j;
                                        __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_50 = __pyx_v_s;
                                        __pyx_t_44 = __pyx_v_b;
                                        __pyx_t_47 = __pyx_v_leaf_numb;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_50 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_43 = 0;
                                        __pyx_t_14 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_48 * __pyx_v_data.strides[0]) ) + __pyx_t_41 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_44 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_43 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_21 = __pyx_t_14;
                                        __pyx_L65_bool_binop_done:;
                                        if (__pyx_t_21) {

                                          /* "cyext_acv.pyx":3194
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if a_it == data.shape[1]:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":3193
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":3196
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      __pyx_t_21 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3197
 * 
 *                             if a_it == data.shape[1]:
 *                                 lm = 1             # <<<<<<<<<<<<<<
 * 
 *                             if b_it == S_size:
 */
                                        __pyx_v_lm = 1;

                                        /* "cyext_acv.pyx":3196
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3199
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      __pyx_t_21 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3200
 * 
 *                             if b_it == S_size:
 *                                 lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                             p += lm
 */
                                        __pyx_v_lm_s = 1;

                                        /* "cyext_acv.pyx":3199
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3202
 *                                 lm_s = 1
 * 
 *                             p += lm             # <<<<<<<<<<<<<<
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 */
                                      __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                      /* "cyext_acv.pyx":3203
 * 
 *                             p += lm
 *                             p_s += lm_s             # <<<<<<<<<<<<<<
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 */
                                      __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                      /* "cyext_acv.pyx":3204
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                      __pyx_t_50 = __pyx_v_i;
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_50 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = __pyx_v_j;
                                      __pyx_t_21 = (((*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_43 * __pyx_v_fX.strides[0]) ))) == (*((long *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_42 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                      if (__pyx_t_21) {

                                        /* "cyext_acv.pyx":3205
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm             # <<<<<<<<<<<<<<
 *                                 p_su += lm_s
 *                             else:
 */
                                        __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3206
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 *                                 p_su += lm_s             # <<<<<<<<<<<<<<
 *                             else:
 *                                 p_d += lm
 */
                                        __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                        /* "cyext_acv.pyx":3204
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                        goto __pyx_L69;
                                      }

                                      /* "cyext_acv.pyx":3208
 *                                 p_su += lm_s
 *                             else:
 *                                 p_d += lm             # <<<<<<<<<<<<<<
 *                                 p_sd += lm_s
 * 
 */
                                      /*else*/ {
                                        __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3209
 *                             else:
 *                                 p_d += lm
 *                                 p_sd += lm_s             # <<<<<<<<<<<<<<
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_s) if p_s != 0 else 0
 */
                                        __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);
                                      }
                                      __pyx_L69:;
                                    }

                                    /* "cyext_acv.pyx":3211
 *                                 p_sd += lm_s
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_s) if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_su) if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_sd) if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_50 = __pyx_v_leaf_numb;
                                      __pyx_t_43 = __pyx_v_i;
                                      __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_44 = __pyx_v_b;
                                      __pyx_t_33 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_47 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_52 = ((__pyx_v_p * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_44 * __pyx_v_values.strides[0]) ) + __pyx_t_33 * __pyx_v_values.strides[1]) ) + __pyx_t_41 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_52 = 0.0;
                                    }
                                    __pyx_t_43 = __pyx_v_i;
                                    __pyx_t_47 = __pyx_v_b;
                                    __pyx_t_50 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_41 = 0;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_47 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_50 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_52;

                                    /* "cyext_acv.pyx":3212
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_s) if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_su) if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_sd) if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_43 = __pyx_v_b;
                                      __pyx_t_41 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = __pyx_v_i;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_47 = __pyx_v_b;
                                      __pyx_t_33 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_43 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_41 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_50 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_52 = ((__pyx_v_p_u * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_47 * __pyx_v_values.strides[0]) ) + __pyx_t_33 * __pyx_v_values.strides[1]) ) + __pyx_t_44 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_52 = 0.0;
                                    }
                                    __pyx_t_42 = __pyx_v_i;
                                    __pyx_t_50 = __pyx_v_b;
                                    __pyx_t_41 = __pyx_v_leaf_numb;
                                    __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_44 = 1;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_50 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_52;

                                    /* "cyext_acv.pyx":3213
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_s) if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_su) if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_sd) if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_44 = __pyx_v_leaf_numb;
                                      __pyx_t_43 = __pyx_v_i;
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_50 = __pyx_v_b;
                                      __pyx_t_33 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_44 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_41 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_52 = ((__pyx_v_p_d * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_50 * __pyx_v_values.strides[0]) ) + __pyx_t_33 * __pyx_v_values.strides[1]) ) + __pyx_t_47 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_52 = 0.0;
                                    }
                                    __pyx_t_43 = __pyx_v_i;
                                    __pyx_t_41 = __pyx_v_b;
                                    __pyx_t_44 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_47 = 2;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_41 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_52;
                                    __pyx_L47_continue:;
                                  }
                                }
                            }
                        }
                    }
                }
            }
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   __builtin_expect(!!(x), 1)
                #define unlikely(x) __builtin_expect(!!(x), 0)
            #endif
          }

          /* "cyext_acv.pyx":3167
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 *             for b in prange(n_trees, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L40;
            }
            __pyx_L40:;
          }
      }

      /* "cyext_acv.pyx":3215
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) / (p_sd) if p_sd != 0 else 0
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 ss_u = 0
 *                 ss_d = 0
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_12; __pyx_t_35+=1) {
        __pyx_v_i = __pyx_t_35;

        /* "cyext_acv.pyx":3216
 * 
 *             for i in range(N):
 *                 ss_u = 0             # <<<<<<<<<<<<<<
 *                 ss_d = 0
 *                 ss_a = 0
 */
        __pyx_v_ss_u = 0.0;

        /* "cyext_acv.pyx":3217
 *             for i in range(N):
 *                 ss_u = 0
 *                 ss_d = 0             # <<<<<<<<<<<<<<
 *                 ss_a = 0
 *                 for b in range(n_trees):
 */
        __pyx_v_ss_d = 0.0;

        /* "cyext_acv.pyx":3218
 *                 ss_u = 0
 *                 ss_d = 0
 *                 ss_a = 0             # <<<<<<<<<<<<<<
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 */
        __pyx_v_ss_a = 0.0;

        /* "cyext_acv.pyx":3219
 *                 ss_d = 0
 *                 ss_a = 0
 *                 for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):
 */
        __pyx_t_34 = __pyx_v_n_trees;
        __pyx_t_24 = __pyx_t_34;
        for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_24; __pyx_t_36+=1) {
          __pyx_v_b = __pyx_t_36;

          /* "cyext_acv.pyx":3220
 *                 ss_a = 0
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]
 */
          __pyx_t_43 = __pyx_v_b;
          __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_43 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":3221
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]
 */
          __pyx_t_37 = __pyx_v_nb_leaf;
          __pyx_t_38 = __pyx_t_37;
          for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
            __pyx_v_leaf_numb = __pyx_t_39;

            /* "cyext_acv.pyx":3222
 *                     nb_leaf = leaves_nb[b]
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]             # <<<<<<<<<<<<<<
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]
 */
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_47 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_41 = 0;
            __pyx_v_ss_a = (__pyx_v_ss_a + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_47 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[3]) ))));

            /* "cyext_acv.pyx":3223
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]             # <<<<<<<<<<<<<<
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_41 = __pyx_v_b;
            __pyx_t_44 = __pyx_v_leaf_numb;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_47 = 1;
            __pyx_v_ss_u = (__pyx_v_ss_u + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_41 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_47 * __pyx_v_mean_forest_b.strides[3]) ))));

            /* "cyext_acv.pyx":3224
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]             # <<<<<<<<<<<<<<
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 */
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_47 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_41 = 2;
            __pyx_v_ss_d = (__pyx_v_ss_d + (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_47 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_44 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[3]) ))));
          }
        }

        /* "cyext_acv.pyx":3225
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 */
        if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
          __pyx_t_52 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
        } else {
          __pyx_t_52 = 0.0;
        }
        __pyx_v_ss = __pyx_t_52;

        /* "cyext_acv.pyx":3226
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_14 = ((__pyx_v_ss >= __pyx_v_pi_level) != 0);
        if (__pyx_t_14) {
        } else {
          __pyx_t_21 = __pyx_t_14;
          goto __pyx_L79_bool_binop_done;
        }
        __pyx_t_43 = __pyx_v_i;
        __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_14 = ((__pyx_v_ss >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_41 * __pyx_v_sdp.strides[0]) )))) != 0);
        __pyx_t_21 = __pyx_t_14;
        __pyx_L79_bool_binop_done:;
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":3227
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_43 = __pyx_v_i;
          __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_41 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

          /* "cyext_acv.pyx":3228
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_43 = __pyx_v_i;
          __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_41 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3229
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_34 = __pyx_v_S_size;
          __pyx_t_24 = __pyx_t_34;
          for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_24; __pyx_t_36+=1) {
            __pyx_v_s = __pyx_t_36;

            /* "cyext_acv.pyx":3230
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_43 = __pyx_v_s;
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_44 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_42 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_44 * __pyx_v_s_star.strides[0]) ) + __pyx_t_42 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3226
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= pi_level and ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":3232
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_21 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_21) {

          /* "cyext_acv.pyx":3233
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_43 = __pyx_v_i;
          __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_41 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":3234
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_43 = __pyx_v_i;
          __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_41 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3235
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_34 = __pyx_v_S_size;
          __pyx_t_24 = __pyx_t_34;
          for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_24; __pyx_t_36+=1) {
            __pyx_v_s = __pyx_t_36;

            /* "cyext_acv.pyx":3236
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_43 = __pyx_v_s;
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_44 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_44 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3232
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":3238
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_11 = __pyx_v_N;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3239
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
      __pyx_t_43 = __pyx_v_i;
      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_21 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_41 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_21) {

        /* "cyext_acv.pyx":3240
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                     s_star[R_buf[i], s] = -1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3240, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3241
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):
 */
        __pyx_t_16 = (__pyx_v_X.shape[1]);
        __pyx_t_43 = __pyx_v_i;
        __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_17 = __pyx_t_16;
        for (__pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_41 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_35 < __pyx_t_17; __pyx_t_35+=1) {
          __pyx_v_s = __pyx_t_35;

          /* "cyext_acv.pyx":3242
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                     s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
          __pyx_t_44 = __pyx_v_i;
          __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_44 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_47 = __pyx_v_s;
          *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_47 * __pyx_v_s_star.strides[1]) )) = -1L;
        }

        /* "cyext_acv.pyx":3243
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_43 = __pyx_v_i;
        __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_41 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_23 = __pyx_t_22;
        for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_23; __pyx_t_35+=1) {
          __pyx_v_s = __pyx_t_35;

          /* "cyext_acv.pyx":3244
 *                     s_star[R_buf[i], s] = -1
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_43 = __pyx_v_i;
          __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_44 = __pyx_v_s;
          __pyx_t_47 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_41 * __pyx_v_s_star.strides[0]) ) + __pyx_t_44 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_47 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":3239
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
      }
    }

    /* "cyext_acv.pyx":3246
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_31 = __pyx_v_r.size();
    __pyx_t_32 = __pyx_t_31;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_32; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3247
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":3248
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if R.size() == 0 and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":3250
 *             R.pop_back()
 * 
 *         if R.size() == 0 and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_14 = ((__pyx_v_R.size() == 0) != 0);
    if (__pyx_t_14) {
    } else {
      __pyx_t_21 = __pyx_t_14;
      goto __pyx_L96_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_stop != 0);
    __pyx_t_21 = __pyx_t_14;
    __pyx_L96_bool_binop_done:;
    if (__pyx_t_21) {

      /* "cyext_acv.pyx":3251
 * 
 *         if R.size() == 0 and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L21_break;

      /* "cyext_acv.pyx":3250
 *             R.pop_back()
 * 
 *         if R.size() == 0 and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":3147
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L21_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":3253
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_long); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_53 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_long); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_53) < 0) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_53); __pyx_t_53 = 0;
  __pyx_t_53 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_53)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_53);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_53);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_53 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":3075
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf_ptrees(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_28, 1);
  __Pyx_XDECREF(__pyx_t_53);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_ptrees", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_35global_sdp_clf_ptrees(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_35global_sdp_clf_ptrees(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_clf_ptrees (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 1); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 2); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 3); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 4); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 5); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 6); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 7); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 8); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 9); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 10); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 11); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, 12); __PYX_ERR(0, 3075, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_clf_ptrees") < 0)) __PYX_ERR(0, 3075, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 3075, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 3075, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 3076, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 3076, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 3077, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 3077, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 3078, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 3078, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 3078, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[9]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 3078, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3079, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[12]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3079, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_clf_ptrees", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3075, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_ptrees", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 3078, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_34global_sdp_clf_ptrees(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_34global_sdp_clf_ptrees(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf_ptrees", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 3075, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_clf_ptrees(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_ptrees", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3260
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_acv_leaves_adap(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_37shap_values_acv_leaves_adap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_shap_values_acv_leaves_adap(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int &__pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, __Pyx_memviewslice __pyx_v_size, std::vector<std::vector<long> >  const __pyx_v_C, int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_x = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_memviewslice __pyx_v_s_star_i = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_n_star_i = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_18;
  unsigned int __pyx_t_19;
  unsigned int __pyx_t_20;
  unsigned int __pyx_t_21;
  int __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_acv_leaves_adap", 0);

  /* "cyext_acv.pyx":3270
 *     const vector[vector[long]] C, int num_threads):
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":3271
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3272
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] phi
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":3275
 * 
 *     cdef double[:, :, :] phi
 *     phi = np.zeros((N, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :] phi_x
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 3275, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":3281
 *     cdef const long[:] s_star_i, n_star_i
 * 
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         if size[i] != m:
 *             s_star_i = S_star[i, :size[i]]
 */
  __pyx_t_8 = __pyx_v_N;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "cyext_acv.pyx":3282
 * 
 *     for i in range(N):
 *         if size[i] != m:             # <<<<<<<<<<<<<<
 *             s_star_i = S_star[i, :size[i]]
 *             n_star_i = N_star[i, size[i]:]
 */
    __pyx_t_11 = __pyx_v_i;
    __pyx_t_12 = (((*((long const  *) ( /* dim=0 */ (__pyx_v_size.data + __pyx_t_11 * __pyx_v_size.strides[0]) ))) != __pyx_v_m) != 0);
    if (__pyx_t_12) {

      /* "cyext_acv.pyx":3283
 *     for i in range(N):
 *         if size[i] != m:
 *             s_star_i = S_star[i, :size[i]]             # <<<<<<<<<<<<<<
 *             n_star_i = N_star[i, size[i]:]
 *             phi_x = single_shap_values_acv_leaves(X[i], data, values, partition_leaves_trees, leaf_idx_trees,
 */
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_13.data = __pyx_v_S_star.data;
      __pyx_t_13.memview = __pyx_v_S_star.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_S_star.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_13,
    __pyx_v_S_star.shape[1], __pyx_v_S_star.strides[1], __pyx_v_S_star.suboffsets[1],
    1,
    0,
    &__pyx_t_14,
    0,
    (*((long const  *) ( /* dim=0 */ (__pyx_v_size.data + __pyx_t_11 * __pyx_v_size.strides[0]) ))),
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 3283, __pyx_L1_error)
}

__PYX_XDEC_MEMVIEW(&__pyx_v_s_star_i, 1);
      __pyx_v_s_star_i = __pyx_t_13;
      __pyx_t_13.memview = NULL;
      __pyx_t_13.data = NULL;

      /* "cyext_acv.pyx":3284
 *         if size[i] != m:
 *             s_star_i = S_star[i, :size[i]]
 *             n_star_i = N_star[i, size[i]:]             # <<<<<<<<<<<<<<
 *             phi_x = single_shap_values_acv_leaves(X[i], data, values, partition_leaves_trees, leaf_idx_trees,
 *                             leaves_nb, scaling, node_idx_trees, s_star_i, n_star_i, C, num_threads)
 */
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_15.data = __pyx_v_N_star.data;
      __pyx_t_15.memview = __pyx_v_N_star.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_15, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_N_star.strides[0];
        __pyx_t_15.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_15,
    __pyx_v_N_star.shape[1], __pyx_v_N_star.strides[1], __pyx_v_N_star.suboffsets[1],
    1,
    0,
    &__pyx_t_14,
    (*((long const  *) ( /* dim=0 */ (__pyx_v_size.data + __pyx_t_11 * __pyx_v_size.strides[0]) ))),
    0,
    0,
    1,
    0,
    0,
    1) < 0))
{
    __PYX_ERR(0, 3284, __pyx_L1_error)
}

__PYX_XDEC_MEMVIEW(&__pyx_v_n_star_i, 1);
      __pyx_v_n_star_i = __pyx_t_15;
      __pyx_t_15.memview = NULL;
      __pyx_t_15.data = NULL;

      /* "cyext_acv.pyx":3285
 *             s_star_i = S_star[i, :size[i]]
 *             n_star_i = N_star[i, size[i]:]
 *             phi_x = single_shap_values_acv_leaves(X[i], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                             leaves_nb, scaling, node_idx_trees, s_star_i, n_star_i, C, num_threads)
 *         else:
 */
      __pyx_t_16.data = __pyx_v_X.data;
      __pyx_t_16.memview = __pyx_v_X.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_X.shape[1];
__pyx_t_16.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_1 = __pyx_f_9cyext_acv_single_shap_values_acv_leaves(__pyx_t_16, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_s_star_i, __pyx_v_n_star_i, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3285, __pyx_L1_error)

      /* "cyext_acv.pyx":3286
 *             n_star_i = N_star[i, size[i]:]
 *             phi_x = single_shap_values_acv_leaves(X[i], data, values, partition_leaves_trees, leaf_idx_trees,
 *                             leaves_nb, scaling, node_idx_trees, s_star_i, n_star_i, C, num_threads)             # <<<<<<<<<<<<<<
 *         else:
 *             # Warning here....
 */
      __Pyx_GOTREF(__pyx_t_1);
      __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "cyext_acv.pyx":3285
 *             s_star_i = S_star[i, :size[i]]
 *             n_star_i = N_star[i, size[i]:]
 *             phi_x = single_shap_values_acv_leaves(X[i], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                             leaves_nb, scaling, node_idx_trees, s_star_i, n_star_i, C, num_threads)
 *         else:
 */
      __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 3285, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_phi_x, 1);
      __pyx_v_phi_x = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "cyext_acv.pyx":3282
 * 
 *     for i in range(N):
 *         if size[i] != m:             # <<<<<<<<<<<<<<
 *             s_star_i = S_star[i, :size[i]]
 *             n_star_i = N_star[i, size[i]:]
 */
      goto __pyx_L5;
    }

    /* "cyext_acv.pyx":3289
 *         else:
 *             # Warning here....
 *             phi_x = single_shap_values(X[i], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                             leaves_nb, scaling, node_idx_trees, C, num_threads)
 * 
 */
    /*else*/ {
      __pyx_t_16.data = __pyx_v_X.data;
      __pyx_t_16.memview = __pyx_v_X.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_X.shape[1];
__pyx_t_16.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_1 = __pyx_f_9cyext_acv_single_shap_values(__pyx_t_16, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3289, __pyx_L1_error)

      /* "cyext_acv.pyx":3290
 *             # Warning here....
 *             phi_x = single_shap_values(X[i], data, values, partition_leaves_trees, leaf_idx_trees,
 *                             leaves_nb, scaling, node_idx_trees, C, num_threads)             # <<<<<<<<<<<<<<
 * 
 *         for j in range(m):
 */
      __Pyx_GOTREF(__pyx_t_1);
      __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "cyext_acv.pyx":3289
 *         else:
 *             # Warning here....
 *             phi_x = single_shap_values(X[i], data, values, partition_leaves_trees, leaf_idx_trees,             # <<<<<<<<<<<<<<
 *                             leaves_nb, scaling, node_idx_trees, C, num_threads)
 * 
 */
      __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 3289, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_phi_x, 1);
      __pyx_v_phi_x = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;
    }
    __pyx_L5:;

    /* "cyext_acv.pyx":3292
 *                             leaves_nb, scaling, node_idx_trees, C, num_threads)
 * 
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             for k in range(d):
 *                 phi[i, j, k] =  phi_x[j, k]
 */
    __pyx_t_18 = __pyx_v_m;
    __pyx_t_19 = __pyx_t_18;
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_19; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "cyext_acv.pyx":3293
 * 
 *         for j in range(m):
 *             for k in range(d):             # <<<<<<<<<<<<<<
 *                 phi[i, j, k] =  phi_x[j, k]
 *     return np.array(phi)
 */
      __pyx_t_20 = __pyx_v_d;
      __pyx_t_21 = __pyx_t_20;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
        __pyx_v_k = __pyx_t_22;

        /* "cyext_acv.pyx":3294
 *         for j in range(m):
 *             for k in range(d):
 *                 phi[i, j, k] =  phi_x[j, k]             # <<<<<<<<<<<<<<
 *     return np.array(phi)
 * 
 */
        __pyx_t_11 = __pyx_v_j;
        __pyx_t_23 = __pyx_v_k;
        __pyx_t_24 = __pyx_v_i;
        __pyx_t_25 = __pyx_v_j;
        __pyx_t_26 = __pyx_v_k;
        *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_24 * __pyx_v_phi.strides[0]) ) + __pyx_t_25 * __pyx_v_phi.strides[1]) ) + __pyx_t_26 * __pyx_v_phi.strides[2]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_x.data + __pyx_t_11 * __pyx_v_phi_x.strides[0]) ) + __pyx_t_23 * __pyx_v_phi_x.strides[1]) )));
      }
    }
  }

  /* "cyext_acv.pyx":3295
 *             for k in range(d):
 *                 phi[i, j, k] =  phi_x[j, k]
 *     return np.array(phi)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_phi, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":3260
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef shap_values_acv_leaves_adap(const double[:, :] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves_adap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_x, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star_i, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_n_star_i, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_37shap_values_acv_leaves_adap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_37shap_values_acv_leaves_adap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  __Pyx_memviewslice __pyx_v_S_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_N_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_size = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shap_values_acv_leaves_adap (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_S_star,&__pyx_n_s_N_star,&__pyx_n_s_size,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 1); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 2); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 3); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 4); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 5); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 6); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 7); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 8); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_N_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 9); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 10); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 11); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, 12); __PYX_ERR(0, 3260, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "shap_values_acv_leaves_adap") < 0)) __PYX_ERR(0, 3260, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 3260, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 3261, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 3262, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 3263, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 3264, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 3265, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_scaling == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3266, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3267, __pyx_L3_error)
    __pyx_v_S_star = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[8], 0); if (unlikely(!__pyx_v_S_star.memview)) __PYX_ERR(0, 3267, __pyx_L3_error)
    __pyx_v_N_star = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[9], 0); if (unlikely(!__pyx_v_N_star.memview)) __PYX_ERR(0, 3267, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[10], 0); if (unlikely(!__pyx_v_size.memview)) __PYX_ERR(0, 3267, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[11]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3268, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3268, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shap_values_acv_leaves_adap", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3260, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves_adap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_36shap_values_acv_leaves_adap(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_size, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_36shap_values_acv_leaves_adap(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, __Pyx_memviewslice __pyx_v_size, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shap_values_acv_leaves_adap", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S_star.memview)) { __Pyx_RaiseUnboundLocalError("S_star"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_N_star.memview)) { __Pyx_RaiseUnboundLocalError("N_star"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  if (unlikely(!__pyx_v_size.memview)) { __Pyx_RaiseUnboundLocalError("size"); __PYX_ERR(0, 3260, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_shap_values_acv_leaves_adap(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_size, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.shap_values_acv_leaves_adap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_N_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_size, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3301
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef single_shap_values_acv_leaves(const double[:] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_39single_shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_single_shap_values_acv_leaves(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  const __pyx_v_C, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_ns;
  unsigned int __pyx_v_add;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_i2;
  unsigned int __pyx_v_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_csi;
  unsigned int __pyx_v_cs;
  unsigned int __pyx_v_len_n_star;
  double __pyx_v_lm;
  __Pyx_memviewslice __pyx_v_N_star_a = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_S_size;
  int __pyx_v_va_size;
  int __pyx_v_b_it;
  int __pyx_v_nv_bool;
  long __pyx_v_set_size;
  std::vector<std::vector<long> >  __pyx_v_C_buff;
  std::vector<std::vector<long> >  __pyx_v_va_id;
  std::vector<std::vector<long> >  __pyx_v_buff;
  std::vector<long>  __pyx_v_coal_va;
  std::vector<long>  __pyx_v_remove_va;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id_v2;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_C_b;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id;
  std::vector<std::vector<std::vector<std::vector<long> > > >  __pyx_v_node_id_va;
  double __pyx_v_p_s;
  double __pyx_v_p_si;
  double __pyx_v_coef;
  double __pyx_v_coef_0;
  double __pyx_v_p_off;
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_si = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_12;
  unsigned int __pyx_t_13;
  unsigned int __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_16;
  std::vector<std::vector<long> > ::size_type __pyx_t_17;
  std::vector<std::vector<long> > ::size_type __pyx_t_18;
  std::vector<long> ::size_type __pyx_t_19;
  std::vector<long> ::size_type __pyx_t_20;
  std::vector<long>  __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  size_t __pyx_t_24;
  std::vector<std::vector<long> >  __pyx_t_25;
  unsigned int __pyx_t_26;
  unsigned int __pyx_t_27;
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  unsigned int __pyx_t_32;
  size_t __pyx_t_33;
  size_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  unsigned int __pyx_t_37;
  size_t __pyx_t_38;
  size_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  int __pyx_t_41;
  std::vector<long> ::size_type __pyx_t_42;
  std::vector<long> ::size_type __pyx_t_43;
  unsigned int __pyx_t_44;
  unsigned int __pyx_t_45;
  Py_ssize_t __pyx_t_46;
  Py_ssize_t __pyx_t_47;
  size_t __pyx_t_48;
  size_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  size_t __pyx_t_51;
  size_t __pyx_t_52;
  size_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  long __pyx_t_56;
  long __pyx_t_57;
  unsigned int __pyx_t_58;
  unsigned int __pyx_t_59;
  unsigned int __pyx_t_60;
  size_t __pyx_t_61;
  size_t __pyx_t_62;
  size_t __pyx_t_63;
  size_t __pyx_t_64;
  size_t __pyx_t_65;
  size_t __pyx_t_66;
  size_t __pyx_t_67;
  size_t __pyx_t_68;
  size_t __pyx_t_69;
  size_t __pyx_t_70;
  size_t __pyx_t_71;
  size_t __pyx_t_72;
  double __pyx_t_73;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_shap_values_acv_leaves", 0);

  /* "cyext_acv.pyx":3312
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":3313
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int m = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3314
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":3315
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :] phi
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":3318
 * 
 *     cdef double[:, :] phi
 *     phi = np.zeros((m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[ :, :,  :, :] phi_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":3321
 * 
 *     cdef double[ :, :,  :, :] phi_b
 *     phi_b = np.zeros((max_leaves, 2**scaling, m, d))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_b = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3325
 * 
 *     cdef double[:, :, :] phi_n
 *     phi_n = np.zeros((max_leaves, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :, :, ::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_n = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3328
 * 
 *     cdef long[:, :, :, ::1] S
 *     S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_S = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":3340
 *     cdef vector[vector[long]] C_buff, va_id, buff, Sm
 *     cdef vector[long] coal_va, remove_va, node_id_b
 *     buff.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     for i in range(max_leaves):
 *         buff[i].resize(1)
 */
  try {
    __pyx_v_buff.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3340, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3341
 *     cdef vector[long] coal_va, remove_va, node_id_b
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):             # <<<<<<<<<<<<<<
 *         buff[i].resize(1)
 * 
 */
  __pyx_t_12 = __pyx_v_max_leaves;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":3342
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):
 *         buff[i].resize(1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    try {
      (__pyx_v_buff[__pyx_v_i]).resize(1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 3342, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":3346
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, node_id
 *     node_id_v2.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     C_b.resize(max_leaves)
 *     # Sm.resize(max_leaves)
 */
  try {
    __pyx_v_node_id_v2.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3346, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3347
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, node_id
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     # Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)
 */
  try {
    __pyx_v_C_b.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3347, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3349
 *     C_b.resize(max_leaves)
 *     # Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 */
  try {
    __pyx_v_node_id.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3349, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3352
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)             # <<<<<<<<<<<<<<
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)
 */
  try {
    __pyx_v_node_id_va.resize(__pyx_v_m);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3352, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3353
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         node_id_va[i].resize(max_leaves)
 * 
 */
  __pyx_t_12 = __pyx_v_m;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "cyext_acv.pyx":3354
 *     node_id_va.resize(m)
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     N_star_a = np.array(N_star)
 */
    try {
      (__pyx_v_node_id_va[__pyx_v_i]).resize(__pyx_v_max_leaves);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 3354, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":3356
 *         node_id_va[i].resize(max_leaves)
 * 
 *     N_star_a = np.array(N_star)             # <<<<<<<<<<<<<<
 *     len_n_star = N_star_a.shape[0]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_N_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_15 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_15.memview)) __PYX_ERR(0, 3356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_N_star_a = __pyx_t_15;
  __pyx_t_15.memview = NULL;
  __pyx_t_15.data = NULL;

  /* "cyext_acv.pyx":3357
 * 
 *     N_star_a = np.array(N_star)
 *     len_n_star = N_star_a.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if C[0].size() != 0:
 */
  __pyx_v_len_n_star = (__pyx_v_N_star_a.shape[0]);

  /* "cyext_acv.pyx":3359
 *     len_n_star = N_star_a.shape[0]
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 */
  __pyx_t_16 = (((__pyx_v_C[0]).size() != 0) != 0);
  if (__pyx_t_16) {

    /* "cyext_acv.pyx":3360
 * 
 *     if C[0].size() != 0:
 *         C_buff = C             # <<<<<<<<<<<<<<
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 * 
 */
    __pyx_v_C_buff = __pyx_v_C;

    /* "cyext_acv.pyx":3361
 *     if C[0].size() != 0:
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(S_star.shape[0]):
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_17 = __pyx_v_C.size();
    __pyx_t_18 = __pyx_t_17;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_18; __pyx_t_12+=1) {
      __pyx_v_ci = __pyx_t_12;
      __pyx_t_19 = (__pyx_v_C[__pyx_v_ci]).size();
      __pyx_t_20 = __pyx_t_19;
      for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_20; __pyx_t_13+=1) {
        __pyx_v_cj = __pyx_t_13;
        __pyx_t_7 = __Pyx_PyInt_From_long(((__pyx_v_C[__pyx_v_ci])[__pyx_v_cj])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 3361, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __pyx_t_21 = __pyx_convert_vector_from_py_long(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_coal_va = __pyx_t_21;

    /* "cyext_acv.pyx":3363
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 * 
 *         for i in range(S_star.shape[0]):             # <<<<<<<<<<<<<<
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])
 */
    __pyx_t_22 = (__pyx_v_S_star.shape[0]);
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_23; __pyx_t_12+=1) {
      __pyx_v_i = __pyx_t_12;

      /* "cyext_acv.pyx":3364
 * 
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 */
      __pyx_t_24 = __pyx_v_i;
      __pyx_t_16 = ((!((std::find<std::vector<long> ::iterator,long>(__pyx_v_coal_va.begin(), __pyx_v_coal_va.end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_24 * __pyx_v_S_star.strides[0]) )))) != __pyx_v_coal_va.end()) != 0)) != 0);
      if (__pyx_t_16) {

        /* "cyext_acv.pyx":3365
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])             # <<<<<<<<<<<<<<
 *                 buff[0][0] = S_star[i]
 *                 va_id.push_back(buff[0])
 */
        __pyx_t_24 = __pyx_v_i;
        try {
          __pyx_v_remove_va.push_back((*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_24 * __pyx_v_S_star.strides[0]) ))));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3365, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3366
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]             # <<<<<<<<<<<<<<
 *                 va_id.push_back(buff[0])
 *             else:
 */
        __pyx_t_24 = __pyx_v_i;
        ((__pyx_v_buff[0])[0]) = (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_24 * __pyx_v_S_star.strides[0]) )));

        /* "cyext_acv.pyx":3367
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 *                 va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 for ci in range(C_buff.size()):
 */
        try {
          __pyx_v_va_id.push_back((__pyx_v_buff[0]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3367, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3364
 * 
 *         for i in range(S_star.shape[0]):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), S_star[i]) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(S_star[i])
 *                 buff[0][0] = S_star[i]
 */
        goto __pyx_L14;
      }

      /* "cyext_acv.pyx":3369
 *                 va_id.push_back(buff[0])
 *             else:
 *                 for ci in range(C_buff.size()):             # <<<<<<<<<<<<<<
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 */
      /*else*/ {
        __pyx_t_17 = __pyx_v_C_buff.size();
        __pyx_t_18 = __pyx_t_17;
        for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_18; __pyx_t_13+=1) {
          __pyx_v_ci = __pyx_t_13;

          /* "cyext_acv.pyx":3370
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          __pyx_t_24 = __pyx_v_i;
          __pyx_t_16 = ((std::find<std::vector<long> ::iterator,long>((__pyx_v_C_buff[__pyx_v_ci]).begin(), (__pyx_v_C_buff[__pyx_v_ci]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_24 * __pyx_v_S_star.strides[0]) )))) != (__pyx_v_C_buff[__pyx_v_ci]).end()) != 0);
          if (__pyx_t_16) {

            /* "cyext_acv.pyx":3371
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])             # <<<<<<<<<<<<<<
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 */
            try {
              __pyx_v_va_id.push_back((__pyx_v_C_buff[__pyx_v_ci]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 3371, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":3372
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])             # <<<<<<<<<<<<<<
 *                         C_buff.pop_back()
 *                         break
 */
            (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(__pyx_v_C_buff.begin(), __pyx_v_C_buff.end(), (__pyx_v_C_buff[__pyx_v_ci])));

            /* "cyext_acv.pyx":3373
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_v_C_buff.pop_back();

            /* "cyext_acv.pyx":3374
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 *                         break             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L16_break;

            /* "cyext_acv.pyx":3370
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), S_star[i]) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          }
        }
        __pyx_L16_break:;
      }
      __pyx_L14:;
    }

    /* "cyext_acv.pyx":3359
 *     len_n_star = N_star_a.shape[0]
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         coal_va = [C[ci][cj] for ci in range(C.size()) for cj in range(C[ci].size())]
 */
    goto __pyx_L7;
  }

  /* "cyext_acv.pyx":3377
 * 
 *     else:
 *         va_id = [[S_star[i]] for i in range(S_star.shape[0])]             # <<<<<<<<<<<<<<
 * 
 *     m = va_id.size()
 */
  /*else*/ {
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_22 = (__pyx_v_S_star.shape[0]);
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_23; __pyx_t_12+=1) {
      __pyx_v_i = __pyx_t_12;
      __pyx_t_24 = __pyx_v_i;
      __pyx_t_7 = __Pyx_PyInt_From_long((*((long const  *) ( /* dim=0 */ (__pyx_v_S_star.data + __pyx_t_24 * __pyx_v_S_star.strides[0]) )))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_7);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 3377, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_25 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(__pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3377, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_va_id = __pyx_t_25;
  }
  __pyx_L7:;

  /* "cyext_acv.pyx":3379
 *         va_id = [[S_star[i]] for i in range(S_star.shape[0])]
 * 
 *     m = va_id.size()             # <<<<<<<<<<<<<<
 * 
 *     cdef double p_s, p_si, coef, coef_0, p_off
 */
  __pyx_v_m = __pyx_v_va_id.size();

  /* "cyext_acv.pyx":3384
 *     cdef double[:, :, :] lm_s, lm_si
 *     cdef double[:, :] lm_star
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_8 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_s = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3385
 *     cdef double[:, :] lm_star
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_8);
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_si = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3386
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_star = np.zeros((va_id.size(), max_leaves))             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_lm_star = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":3388
 *     lm_star = np.zeros((va_id.size(), max_leaves))
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for j in range(data.shape[1]):
 *             for i2 in range(d):
 */
  __pyx_t_12 = __pyx_v_n_trees;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_b = __pyx_t_14;

    /* "cyext_acv.pyx":3389
 * 
 *     for b in range(n_trees):
 *         for j in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 */
    __pyx_t_22 = (__pyx_v_data.shape[1]);
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_23; __pyx_t_26+=1) {
      __pyx_v_j = __pyx_t_26;

      /* "cyext_acv.pyx":3390
 *     for b in range(n_trees):
 *         for j in range(data.shape[1]):
 *             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi_n[leaf_numb, j, i2] = 0
 */
      __pyx_t_27 = __pyx_v_d;
      __pyx_t_28 = __pyx_t_27;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_28; __pyx_t_29+=1) {
        __pyx_v_i2 = __pyx_t_29;

        /* "cyext_acv.pyx":3391
 *         for j in range(data.shape[1]):
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                     phi_n[leaf_numb, j, i2] = 0
 *                     for counter in range(phi_b.shape[1]):
 */
        __pyx_t_30 = (__pyx_v_phi_b.shape[0]);
        __pyx_t_31 = __pyx_t_30;
        for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
          __pyx_v_leaf_numb = __pyx_t_32;

          /* "cyext_acv.pyx":3392
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi_n[leaf_numb, j, i2] = 0             # <<<<<<<<<<<<<<
 *                     for counter in range(phi_b.shape[1]):
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 */
          __pyx_t_24 = __pyx_v_leaf_numb;
          __pyx_t_33 = __pyx_v_j;
          __pyx_t_34 = __pyx_v_i2;
          *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_24 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_33 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_34 * __pyx_v_phi_n.strides[2]) )) = 0.0;

          /* "cyext_acv.pyx":3393
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi_n[leaf_numb, j, i2] = 0
 *                     for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 */
          __pyx_t_35 = (__pyx_v_phi_b.shape[1]);
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_counter = __pyx_t_37;

            /* "cyext_acv.pyx":3394
 *                     phi_n[leaf_numb, j, i2] = 0
 *                     for counter in range(phi_b.shape[1]):
 *                         phi_b[leaf_numb, counter, j, i2] = 0             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
            __pyx_t_34 = __pyx_v_leaf_numb;
            __pyx_t_33 = __pyx_v_counter;
            __pyx_t_24 = __pyx_v_j;
            __pyx_t_38 = __pyx_v_i2;
            *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_34 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_33 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_24 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_38 * __pyx_v_phi_b.strides[3]) )) = 0.0;
          }
        }
      }
    }

    /* "cyext_acv.pyx":3395
 *                     for counter in range(phi_b.shape[1]):
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 */
    __pyx_t_38 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ ((char *) (((long const  *) __pyx_v_leaves_nb.data) + __pyx_t_38)) )));

    /* "cyext_acv.pyx":3396
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_26 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              int __pyx_parallel_temp0 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp1 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp2 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp3 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp4 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp5 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp6 = ((double)__PYX_NAN());
              double __pyx_parallel_temp7 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp8 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp9 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp10 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp11 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp12 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp13 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp14 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp15 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp16 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp17 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp18 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp19 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp20 = ((double)__PYX_NAN());
              double __pyx_parallel_temp21 = ((double)__PYX_NAN());
              double __pyx_parallel_temp22 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp23 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp24 = ((unsigned int)0xbad0bad0);
              long __pyx_parallel_temp25 = ((long)0xbad0bad0);
              unsigned int __pyx_parallel_temp26 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp27 = ((int)0xbad0bad0);
              const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
              PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
              int __pyx_parallel_why;
              __pyx_parallel_why = 0;
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_28 = (__pyx_t_26 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_28 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57, __pyx_t_58, __pyx_t_59, __pyx_t_60, __pyx_t_61, __pyx_t_62, __pyx_t_63, __pyx_t_64, __pyx_t_65, __pyx_t_66, __pyx_t_67, __pyx_t_68, __pyx_t_69, __pyx_t_70, __pyx_t_71, __pyx_t_72, __pyx_t_73) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      Py_BEGIN_ALLOW_THREADS
                      #endif /* _OPENMP */
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_S_size) lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_add) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_ci) lastprivate(__pyx_v_cj) lastprivate(__pyx_v_coef) lastprivate(__pyx_v_coef_0) lastprivate(__pyx_v_counter) lastprivate(__pyx_v_cs) lastprivate(__pyx_v_csi) lastprivate(__pyx_v_i) lastprivate(__pyx_v_i2) lastprivate(__pyx_v_l) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_ns) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_off) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_si) lastprivate(__pyx_v_pow_set_size) lastprivate(__pyx_v_s) lastprivate(__pyx_v_set_size) lastprivate(__pyx_v_va) lastprivate(__pyx_v_va_size) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_28; __pyx_t_27++){
                          if (__pyx_parallel_why < 2)
                          {
                              __pyx_v_leaf_numb = (unsigned int)(0 + 1 * __pyx_t_27);
                              /* Initialize private variables to invalid values */
                              __pyx_v_S_size = ((int)0xbad0bad0);
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_add = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((int)0xbad0bad0);
                              __pyx_v_ci = ((unsigned int)0xbad0bad0);
                              __pyx_v_cj = ((unsigned int)0xbad0bad0);
                              __pyx_v_coef = ((double)__PYX_NAN());
                              __pyx_v_coef_0 = ((double)__PYX_NAN());
                              __pyx_v_counter = ((unsigned int)0xbad0bad0);
                              __pyx_v_cs = ((unsigned int)0xbad0bad0);
                              __pyx_v_csi = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((unsigned int)0xbad0bad0);
                              __pyx_v_i2 = ((unsigned int)0xbad0bad0);
                              __pyx_v_l = ((unsigned int)0xbad0bad0);
                              __pyx_v_lm = ((double)__PYX_NAN());
                              __pyx_v_ns = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv_bool = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_off = ((double)__PYX_NAN());
                              __pyx_v_p_s = ((double)__PYX_NAN());
                              __pyx_v_p_si = ((double)__PYX_NAN());
                              __pyx_v_pow_set_size = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_set_size = ((long)0xbad0bad0);
                              __pyx_v_va = ((unsigned int)0xbad0bad0);
                              __pyx_v_va_size = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":3397
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 */
                              (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).clear();

                              /* "cyext_acv.pyx":3398
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):             # <<<<<<<<<<<<<<
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 */
                              __pyx_t_22 = (__pyx_v_N_star.shape[0]);
                              __pyx_t_23 = __pyx_t_22;
                              for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_23; __pyx_t_29+=1) {
                                __pyx_v_i = __pyx_t_29;

                                /* "cyext_acv.pyx":3399
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):             # <<<<<<<<<<<<<<
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 */
                                __pyx_t_38 = __pyx_v_i;
                                __pyx_t_16 = ((std::find<std::vector<long> ::iterator,long>(((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_N_star.data + __pyx_t_38 * __pyx_v_N_star.strides[0]) )))) != ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end()) != 0);
                                if (__pyx_t_16) {

                                  /* "cyext_acv.pyx":3400
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])             # <<<<<<<<<<<<<<
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0
 */
                                  __pyx_t_38 = __pyx_v_i;
                                  (void)(std::remove<std::vector<long> ::iterator,long>(((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).end(), (*((long const  *) ( /* dim=0 */ (__pyx_v_N_star.data + __pyx_t_38 * __pyx_v_N_star.strides[0]) )))));

                                  /* "cyext_acv.pyx":3401
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *             lm = 0
 *             for i in range(data.shape[0]):
 */
                                  ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).pop_back();

                                  /* "cyext_acv.pyx":3399
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 *                 if (std_find[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i]) != node_idx_trees[b][leaf_numb].end()):             # <<<<<<<<<<<<<<
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 */
                                }
                              }

                              /* "cyext_acv.pyx":3402
 *                     std_remove[vector[long].iterator, long](node_idx_trees[b][leaf_numb].begin(), node_idx_trees[b][leaf_numb].end(), N_star[i])
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 */
                              __pyx_v_lm = 0.0;

                              /* "cyext_acv.pyx":3403
 *                     node_idx_trees[b][leaf_numb].pop_back()
 *             lm = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 */
                              __pyx_t_22 = (__pyx_v_data.shape[0]);
                              __pyx_t_23 = __pyx_t_22;
                              for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_23; __pyx_t_29+=1) {
                                __pyx_v_i = __pyx_t_29;

                                /* "cyext_acv.pyx":3404
 *             lm = 0
 *             for i in range(data.shape[0]):
 *                 a_it = 0             # <<<<<<<<<<<<<<
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 */
                                __pyx_v_a_it = 0;

                                /* "cyext_acv.pyx":3405
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 *                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 */
                                __pyx_t_30 = (__pyx_v_data.shape[1]);
                                __pyx_t_31 = __pyx_t_30;
                                for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
                                  __pyx_v_s = __pyx_t_32;

                                  /* "cyext_acv.pyx":3406
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  __pyx_t_38 = __pyx_v_i;
                                  __pyx_t_24 = __pyx_v_s;
                                  __pyx_t_33 = __pyx_v_b;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_39 = __pyx_v_s;
                                  __pyx_t_40 = 1;
                                  __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_38 * __pyx_v_data.strides[0]) ) + __pyx_t_24 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_41) {
                                  } else {
                                    __pyx_t_16 = __pyx_t_41;
                                    goto __pyx_L47_bool_binop_done;
                                  }
                                  __pyx_t_39 = __pyx_v_i;
                                  __pyx_t_34 = __pyx_v_s;
                                  __pyx_t_33 = __pyx_v_b;
                                  __pyx_t_24 = __pyx_v_leaf_numb;
                                  __pyx_t_38 = __pyx_v_s;
                                  __pyx_t_40 = 0;
                                  __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_39 * __pyx_v_data.strides[0]) ) + __pyx_t_34 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_24 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_16 = __pyx_t_41;
                                  __pyx_L47_bool_binop_done:;
                                  if (__pyx_t_16) {

                                    /* "cyext_acv.pyx":3407
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1
 */
                                    __pyx_v_a_it = (__pyx_v_a_it + 1);

                                    /* "cyext_acv.pyx":3406
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":3408
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                __pyx_t_16 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                if (__pyx_t_16) {

                                  /* "cyext_acv.pyx":3409
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1.0);

                                  /* "cyext_acv.pyx":3408
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":3412
 * 
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                              __pyx_t_16 = (((__pyx_v_C[0]).size() != 0) != 0);
                              if (__pyx_t_16) {

                                /* "cyext_acv.pyx":3413
 * 
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C             # <<<<<<<<<<<<<<
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 */
                                (__pyx_v_C_b[__pyx_v_leaf_numb]) = __pyx_v_C;

                                /* "cyext_acv.pyx":3414
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 */
                                __pyx_t_19 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_20; __pyx_t_29+=1) {
                                  __pyx_v_nv = __pyx_t_29;

                                  /* "cyext_acv.pyx":3415
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0             # <<<<<<<<<<<<<<
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 */
                                  __pyx_v_add = 0;

                                  /* "cyext_acv.pyx":3416
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 *                     for ns in range(remove_va.size()):             # <<<<<<<<<<<<<<
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 */
                                  __pyx_t_42 = __pyx_v_remove_va.size();
                                  __pyx_t_43 = __pyx_t_42;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_43; __pyx_t_32+=1) {
                                    __pyx_v_ns = __pyx_t_32;

                                    /* "cyext_acv.pyx":3417
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    __pyx_t_16 = (((((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]) == (__pyx_v_remove_va[__pyx_v_ns])) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3418
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1             # <<<<<<<<<<<<<<
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 */
                                      __pyx_v_add = 1;

                                      /* "cyext_acv.pyx":3419
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]             # <<<<<<<<<<<<<<
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 */
                                      ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]);

                                      /* "cyext_acv.pyx":3420
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 *                             break
 *                     if add == 0:
 */
                                      try {
                                        (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                      } catch(...) {
                                        #ifdef WITH_THREAD
                                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                        #endif
                                        __Pyx_CppExn2PyErr();
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        __PYX_ERR(0, 3420, __pyx_L37_error)
                                      }

                                      /* "cyext_acv.pyx":3421
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break             # <<<<<<<<<<<<<<
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 */
                                      goto __pyx_L54_break;

                                      /* "cyext_acv.pyx":3417
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    }
                                  }
                                  __pyx_L54_break:;

                                  /* "cyext_acv.pyx":3422
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  __pyx_t_16 = ((__pyx_v_add == 0) != 0);
                                  if (__pyx_t_16) {

                                    /* "cyext_acv.pyx":3423
 *                             break
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 */
                                    __pyx_t_17 = (__pyx_v_C_b[__pyx_v_leaf_numb]).size();
                                    __pyx_t_18 = __pyx_t_17;
                                    for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_18; __pyx_t_32+=1) {
                                      __pyx_v_ci = __pyx_t_32;

                                      /* "cyext_acv.pyx":3424
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 */
                                      __pyx_t_42 = ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_43 = __pyx_t_42;
                                      for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_43; __pyx_t_37+=1) {
                                        __pyx_v_cj = __pyx_t_37;

                                        /* "cyext_acv.pyx":3425
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        __pyx_t_16 = (((((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]) == (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv])) != 0);
                                        if (__pyx_t_16) {

                                          /* "cyext_acv.pyx":3426
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1             # <<<<<<<<<<<<<<
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 */
                                          __pyx_v_add = 1;

                                          /* "cyext_acv.pyx":3427
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                     break
 *                             if add == 1:
 */
                                          try {
                                            (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back(((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]));
                                          } catch(...) {
                                            #ifdef WITH_THREAD
                                            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                            #endif
                                            __Pyx_CppExn2PyErr();
                                            #ifdef WITH_THREAD
                                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                            #endif
                                            __PYX_ERR(0, 3427, __pyx_L37_error)
                                          }

                                          /* "cyext_acv.pyx":3428
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break             # <<<<<<<<<<<<<<
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 */
                                          goto __pyx_L60_break;

                                          /* "cyext_acv.pyx":3425
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        }
                                      }
                                      __pyx_L60_break:;

                                      /* "cyext_acv.pyx":3429
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      __pyx_t_16 = ((__pyx_v_add == 1) != 0);
                                      if (__pyx_t_16) {

                                        /* "cyext_acv.pyx":3430
 *                                     break
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 */
                                        (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >((__pyx_v_C_b[__pyx_v_leaf_numb]).begin(), (__pyx_v_C_b[__pyx_v_leaf_numb]).end(), ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])));

                                        /* "cyext_acv.pyx":3431
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 */
                                        (__pyx_v_C_b[__pyx_v_leaf_numb]).pop_back();

                                        /* "cyext_acv.pyx":3432
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 *                                 break             # <<<<<<<<<<<<<<
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 */
                                        goto __pyx_L58_break;

                                        /* "cyext_acv.pyx":3429
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      }
                                    }
                                    __pyx_L58_break:;

                                    /* "cyext_acv.pyx":3422
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":3433
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]             # <<<<<<<<<<<<<<
 *             else:
 *                 node_id[leaf_numb].clear()
 */
                                (__pyx_v_node_id[__pyx_v_leaf_numb]) = (__pyx_v_node_id_v2[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":3412
 * 
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                                goto __pyx_L50;
                              }

                              /* "cyext_acv.pyx":3435
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 *                 node_id[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 */
                              /*else*/ {
                                (__pyx_v_node_id[__pyx_v_leaf_numb]).clear();

                                /* "cyext_acv.pyx":3436
 *             else:
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 */
                                __pyx_t_19 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_20 = __pyx_t_19;
                                for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_20; __pyx_t_29+=1) {
                                  __pyx_v_i = __pyx_t_29;

                                  /* "cyext_acv.pyx":3437
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]             # <<<<<<<<<<<<<<
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 */
                                  ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_i]);

                                  /* "cyext_acv.pyx":3438
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 * 
 *             for va in range(va_id.size()):
 */
                                  try {
                                    (__pyx_v_node_id[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                  } catch(...) {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    __Pyx_CppExn2PyErr();
                                    #ifdef WITH_THREAD
                                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                    #endif
                                    __PYX_ERR(0, 3438, __pyx_L37_error)
                                  }
                                }
                              }
                              __pyx_L50:;

                              /* "cyext_acv.pyx":3440
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 *             for va in range(va_id.size()):             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 */
                              __pyx_t_17 = __pyx_v_va_id.size();
                              __pyx_t_18 = __pyx_t_17;
                              for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_18; __pyx_t_29+=1) {
                                __pyx_v_va = __pyx_t_29;

                                /* "cyext_acv.pyx":3441
 * 
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]             # <<<<<<<<<<<<<<
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 * 
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]) = (__pyx_v_node_id[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":3442
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 * 
 *                     lm_star[va, leaf_numb] = 0
 */
                                __pyx_t_16 = ((!((std::find<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])) != ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end()) != 0)) != 0);
                                if (__pyx_t_16) {

                                  /* "cyext_acv.pyx":3444
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 * 
 *                     lm_star[va, leaf_numb] = 0             # <<<<<<<<<<<<<<
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 */
                                  __pyx_t_38 = __pyx_v_va;
                                  __pyx_t_24 = __pyx_v_leaf_numb;
                                  *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_38 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_24 * __pyx_v_lm_star.strides[1]) )) = 0.0;

                                  /* "cyext_acv.pyx":3445
 * 
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(len_n_star):
 */
                                  __pyx_t_22 = (__pyx_v_data.shape[0]);
                                  __pyx_t_23 = __pyx_t_22;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_23; __pyx_t_32+=1) {
                                    __pyx_v_i = __pyx_t_32;

                                    /* "cyext_acv.pyx":3446
 *                     lm_star[va, leaf_numb] = 0
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":3447
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_37 = __pyx_v_len_n_star;
                                    __pyx_t_44 = __pyx_t_37;
                                    for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_44; __pyx_t_45+=1) {
                                      __pyx_v_s = __pyx_t_45;

                                      /* "cyext_acv.pyx":3448
 *                         b_it = 0
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_38 = __pyx_v_i;
                                      __pyx_t_40 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_24 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_33 = __pyx_v_s;
                                      __pyx_t_34 = __pyx_v_b;
                                      __pyx_t_39 = __pyx_v_leaf_numb;
                                      __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_33 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_47 = 1;
                                      __pyx_t_48 = __pyx_v_s;
                                      __pyx_t_49 = __pyx_v_i;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_48 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_52 = __pyx_v_b;
                                      __pyx_t_53 = __pyx_v_leaf_numb;
                                      __pyx_t_54 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_51 * __pyx_v_N_star_a.strides[0]) )));
                                      __pyx_t_55 = 0;
                                      __pyx_t_16 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_38 * __pyx_v_data.strides[0]) ) + __pyx_t_40 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_49 * __pyx_v_data.strides[0]) ) + __pyx_t_50 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_16) {

                                        /* "cyext_acv.pyx":3449
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == len_n_star:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":3448
 *                         b_it = 0
 *                         for s in range(len_n_star):
 *                             if ((data[i, N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (data[i, N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":3451
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1
 * 
 */
                                    __pyx_t_16 = ((__pyx_v_b_it == __pyx_v_len_n_star) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3452
 * 
 *                         if b_it == len_n_star:
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                      __pyx_t_51 = __pyx_v_va;
                                      __pyx_t_53 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = __pyx_v_va;
                                      __pyx_t_48 = __pyx_v_leaf_numb;
                                      *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_52 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_48 * __pyx_v_lm_star.strides[1]) )) = ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_51 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_53 * __pyx_v_lm_star.strides[1]) ))) + 1.0);

                                      /* "cyext_acv.pyx":3451
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_star[va, leaf_numb] = lm_star[va, leaf_numb] +  1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3455
 * 
 * 
 *                     p_s = 0             # <<<<<<<<<<<<<<
 *                     p_si = 0
 * 
 */
                                  __pyx_v_p_s = 0.0;

                                  /* "cyext_acv.pyx":3456
 * 
 *                     p_s = 0
 *                     p_si = 0             # <<<<<<<<<<<<<<
 * 
 *                     o_all = 0
 */
                                  __pyx_v_p_si = 0.0;

                                  /* "cyext_acv.pyx":3458
 *                     p_si = 0
 * 
 *                     o_all = 0             # <<<<<<<<<<<<<<
 *                     csi = 0
 *                     for s in range(len_n_star):
 */
                                  __pyx_v_o_all = 0;

                                  /* "cyext_acv.pyx":3459
 * 
 *                     o_all = 0
 *                     csi = 0             # <<<<<<<<<<<<<<
 *                     for s in range(len_n_star):
 *                         if ((X[N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 */
                                  __pyx_v_csi = 0;

                                  /* "cyext_acv.pyx":3460
 *                     o_all = 0
 *                     csi = 0
 *                     for s in range(len_n_star):             # <<<<<<<<<<<<<<
 *                         if ((X[N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                             o_all = o_all + 1
 */
                                  __pyx_t_32 = __pyx_v_len_n_star;
                                  __pyx_t_37 = __pyx_t_32;
                                  for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_37; __pyx_t_44+=1) {
                                    __pyx_v_s = __pyx_t_44;

                                    /* "cyext_acv.pyx":3461
 *                     csi = 0
 *                     for s in range(len_n_star):
 *                         if ((X[N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    __pyx_t_53 = __pyx_v_s;
                                    __pyx_t_55 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_53 * __pyx_v_N_star_a.strides[0]) )));
                                    __pyx_t_51 = __pyx_v_s;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_52 = __pyx_v_leaf_numb;
                                    __pyx_t_54 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_51 * __pyx_v_N_star_a.strides[0]) )));
                                    __pyx_t_50 = 1;
                                    __pyx_t_49 = __pyx_v_s;
                                    __pyx_t_47 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_49 * __pyx_v_N_star_a.strides[0]) )));
                                    __pyx_t_33 = __pyx_v_s;
                                    __pyx_t_39 = __pyx_v_b;
                                    __pyx_t_34 = __pyx_v_leaf_numb;
                                    __pyx_t_46 = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_33 * __pyx_v_N_star_a.strides[0]) )));
                                    __pyx_t_40 = 0;
                                    __pyx_t_16 = ((((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_55 * __pyx_v_X.strides[0]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_47 * __pyx_v_X.strides[0]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3462
 *                     for s in range(len_n_star):
 *                         if ((X[N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):
 *                             o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                     if o_all == len_n_star:
 */
                                      __pyx_v_o_all = (__pyx_v_o_all + 1);

                                      /* "cyext_acv.pyx":3461
 *                     csi = 0
 *                     for s in range(len_n_star):
 *                         if ((X[N_star_a[s]] <= partition_leaves_trees[b, leaf_numb, N_star_a[s], 1]) * (X[N_star_a[s]] > partition_leaves_trees[b, leaf_numb, N_star_a[s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3464
 *                             o_all = o_all + 1
 * 
 *                     if o_all == len_n_star:             # <<<<<<<<<<<<<<
 *                         csi = 1
 * 
 */
                                  __pyx_t_16 = ((__pyx_v_o_all == __pyx_v_len_n_star) != 0);
                                  if (__pyx_t_16) {

                                    /* "cyext_acv.pyx":3465
 * 
 *                     if o_all == len_n_star:
 *                         csi = 1             # <<<<<<<<<<<<<<
 * 
 *                     p_s = lm/data.shape[0]
 */
                                    __pyx_v_csi = 1;

                                    /* "cyext_acv.pyx":3464
 *                             o_all = o_all + 1
 * 
 *                     if o_all == len_n_star:             # <<<<<<<<<<<<<<
 *                         csi = 1
 * 
 */
                                  }

                                  /* "cyext_acv.pyx":3467
 *                         csi = 1
 * 
 *                     p_s = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                     p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                     for nv in range(va_id[va].size()):
 */
                                  __pyx_v_p_s = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                  /* "cyext_acv.pyx":3468
 * 
 *                     p_s = lm/data.shape[0]
 *                     p_si = (csi * lm)/lm_star[va, leaf_numb]             # <<<<<<<<<<<<<<
 *                     for nv in range(va_id[va].size()):
 *                         for i2 in range(d):
 */
                                  __pyx_t_33 = __pyx_v_va;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_v_p_si = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_star.data + __pyx_t_33 * __pyx_v_lm_star.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_star.strides[1]) ))));

                                  /* "cyext_acv.pyx":3469
 *                     p_s = lm/data.shape[0]
 *                     p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                     for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                         for i2 in range(d):
 *                             phi_n[leaf_numb, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                  __pyx_t_19 = (__pyx_v_va_id[__pyx_v_va]).size();
                                  __pyx_t_20 = __pyx_t_19;
                                  for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_20; __pyx_t_32+=1) {
                                    __pyx_v_nv = __pyx_t_32;

                                    /* "cyext_acv.pyx":3470
 *                     p_si = (csi * lm)/lm_star[va, leaf_numb]
 *                     for nv in range(va_id[va].size()):
 *                         for i2 in range(d):             # <<<<<<<<<<<<<<
 *                             phi_n[leaf_numb, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 *                     continue
 */
                                    __pyx_t_37 = __pyx_v_d;
                                    __pyx_t_44 = __pyx_t_37;
                                    for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_44; __pyx_t_45+=1) {
                                      __pyx_v_i2 = __pyx_t_45;

                                      /* "cyext_acv.pyx":3471
 *                     for nv in range(va_id[va].size()):
 *                         for i2 in range(d):
 *                             phi_n[leaf_numb, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                      __pyx_t_34 = __pyx_v_b;
                                      __pyx_t_33 = __pyx_v_leaf_numb;
                                      __pyx_t_39 = __pyx_v_b;
                                      __pyx_t_40 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_34 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_49 = __pyx_v_i2;
                                      __pyx_t_51 = __pyx_v_leaf_numb;
                                      __pyx_t_46 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_52 = __pyx_v_i2;
                                      *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_51 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_46 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_52 * __pyx_v_phi_n.strides[2]) )) += ((__pyx_v_p_si - __pyx_v_p_s) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_39 * __pyx_v_values.strides[0]) ) + __pyx_t_40 * __pyx_v_values.strides[1]) ) + __pyx_t_49 * __pyx_v_values.strides[2]) ))));
                                    }
                                  }

                                  /* "cyext_acv.pyx":3472
 *                         for i2 in range(d):
 *                             phi_n[leaf_numb, va_id[va][nv], i2] += (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 */
                                  goto __pyx_L65_continue;

                                  /* "cyext_acv.pyx":3442
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 * 
 *                     lm_star[va, leaf_numb] = 0
 */
                                }

                                /* "cyext_acv.pyx":3474
 *                     continue
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 */
                                (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])));

                                /* "cyext_acv.pyx":3475
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 *                 node_id_va[va][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).pop_back();

                                /* "cyext_acv.pyx":3477
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()             # <<<<<<<<<<<<<<
 *                 pow_set_size = 2**set_size
 * 
 */
                                __pyx_v_set_size = ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).size();

                                /* "cyext_acv.pyx":3478
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 *                 pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *                 for counter in range(0, pow_set_size):
 */
                                __pyx_v_pow_set_size = __Pyx_pow_long(2, __pyx_v_set_size);

                                /* "cyext_acv.pyx":3480
 *                 pow_set_size = 2**set_size
 * 
 *                 for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *                     va_size = 0
 *                     S_size = 0
 */
                                __pyx_t_32 = __pyx_v_pow_set_size;
                                __pyx_t_37 = __pyx_t_32;
                                for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_37; __pyx_t_44+=1) {
                                  __pyx_v_counter = __pyx_t_44;

                                  /* "cyext_acv.pyx":3481
 * 
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0             # <<<<<<<<<<<<<<
 *                     S_size = 0
 *                     for ci in range(set_size):
 */
                                  __pyx_v_va_size = 0;

                                  /* "cyext_acv.pyx":3482
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0
 *                     S_size = 0             # <<<<<<<<<<<<<<
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 */
                                  __pyx_v_S_size = 0;

                                  /* "cyext_acv.pyx":3483
 *                     va_size = 0
 *                     S_size = 0
 *                     for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 */
                                  __pyx_t_56 = __pyx_v_set_size;
                                  __pyx_t_57 = __pyx_t_56;
                                  for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_57; __pyx_t_45+=1) {
                                    __pyx_v_ci = __pyx_t_45;

                                    /* "cyext_acv.pyx":3484
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    __pyx_t_16 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3485
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 */
                                      __pyx_t_19 = (((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_20 = __pyx_t_19;
                                      for (__pyx_t_58 = 0; __pyx_t_58 < __pyx_t_20; __pyx_t_58+=1) {
                                        __pyx_v_cj = __pyx_t_58;

                                        /* "cyext_acv.pyx":3486
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]             # <<<<<<<<<<<<<<
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1
 */
                                        __pyx_t_33 = __pyx_v_va;
                                        __pyx_t_34 = __pyx_v_counter;
                                        __pyx_t_49 = __pyx_v_leaf_numb;
                                        __pyx_t_40 = __pyx_v_S_size;
                                        *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) ) + __pyx_t_34 * __pyx_v_S.strides[1]) ) + __pyx_t_49 * __pyx_v_S.strides[2]) )) + __pyx_t_40)) )) = ((((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]);

                                        /* "cyext_acv.pyx":3487
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                             va_size = va_size + 1
 * 
 */
                                        __pyx_v_S_size = (__pyx_v_S_size + 1);
                                      }

                                      /* "cyext_acv.pyx":3488
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 */
                                      __pyx_v_va_size = (__pyx_v_va_size + 1);

                                      /* "cyext_acv.pyx":3484
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3490
 *                             va_size = va_size + 1
 * 
 *                     lm_s[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     lm_si[va, leaf_numb, counter] = 0
 * 
 */
                                  __pyx_t_49 = __pyx_v_va;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_33 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_49 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_33 * __pyx_v_lm_s.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":3491
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 * 
 *                     for nv in range(len_n_star):
 */
                                  __pyx_t_33 = __pyx_v_va;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_49 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_33 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_49 * __pyx_v_lm_si.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":3493
 *                     lm_si[va, leaf_numb, counter] = 0
 * 
 *                     for nv in range(len_n_star):             # <<<<<<<<<<<<<<
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]
 * 
 */
                                  __pyx_t_45 = __pyx_v_len_n_star;
                                  __pyx_t_58 = __pyx_t_45;
                                  for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_58; __pyx_t_59+=1) {
                                    __pyx_v_nv = __pyx_t_59;

                                    /* "cyext_acv.pyx":3494
 * 
 *                     for nv in range(len_n_star):
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]             # <<<<<<<<<<<<<<
 * 
 *                     for i in range(data.shape[0]):
 */
                                    __pyx_t_49 = __pyx_v_nv;
                                    __pyx_t_34 = __pyx_v_va;
                                    __pyx_t_33 = __pyx_v_counter;
                                    __pyx_t_39 = __pyx_v_leaf_numb;
                                    __pyx_t_52 = (__pyx_v_S_size + __pyx_v_nv);
                                    *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) ) + __pyx_t_33 * __pyx_v_S.strides[1]) ) + __pyx_t_39 * __pyx_v_S.strides[2]) )) + __pyx_t_52)) )) = (*((long *) ( /* dim=0 */ (__pyx_v_N_star_a.data + __pyx_t_49 * __pyx_v_N_star_a.strides[0]) )));
                                  }

                                  /* "cyext_acv.pyx":3496
 *                         S[va, counter, leaf_numb, S_size+nv] = N_star_a[nv]
 * 
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 */
                                  __pyx_t_22 = (__pyx_v_data.shape[0]);
                                  __pyx_t_23 = __pyx_t_22;
                                  for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_23; __pyx_t_45+=1) {
                                    __pyx_v_i = __pyx_t_45;

                                    /* "cyext_acv.pyx":3497
 * 
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":3498
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_58 = (__pyx_v_S_size + __pyx_v_len_n_star);
                                    __pyx_t_59 = __pyx_t_58;
                                    for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_59; __pyx_t_60+=1) {
                                      __pyx_v_s = __pyx_t_60;

                                      /* "cyext_acv.pyx":3499
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_49 = __pyx_v_va;
                                      __pyx_t_52 = __pyx_v_counter;
                                      __pyx_t_39 = __pyx_v_leaf_numb;
                                      __pyx_t_33 = __pyx_v_s;
                                      __pyx_t_34 = __pyx_v_i;
                                      __pyx_t_40 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_49 * __pyx_v_S.strides[0]) ) + __pyx_t_52 * __pyx_v_S.strides[1]) ) + __pyx_t_39 * __pyx_v_S.strides[2]) )) + __pyx_t_33)) )));
                                      __pyx_t_51 = __pyx_v_va;
                                      __pyx_t_48 = __pyx_v_counter;
                                      __pyx_t_53 = __pyx_v_leaf_numb;
                                      __pyx_t_24 = __pyx_v_s;
                                      __pyx_t_38 = __pyx_v_b;
                                      __pyx_t_61 = __pyx_v_leaf_numb;
                                      __pyx_t_46 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) ) + __pyx_t_48 * __pyx_v_S.strides[1]) ) + __pyx_t_53 * __pyx_v_S.strides[2]) )) + __pyx_t_24)) )));
                                      __pyx_t_47 = 1;
                                      __pyx_t_62 = __pyx_v_va;
                                      __pyx_t_63 = __pyx_v_counter;
                                      __pyx_t_64 = __pyx_v_leaf_numb;
                                      __pyx_t_65 = __pyx_v_s;
                                      __pyx_t_66 = __pyx_v_i;
                                      __pyx_t_50 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_62 * __pyx_v_S.strides[0]) ) + __pyx_t_63 * __pyx_v_S.strides[1]) ) + __pyx_t_64 * __pyx_v_S.strides[2]) )) + __pyx_t_65)) )));
                                      __pyx_t_67 = __pyx_v_va;
                                      __pyx_t_68 = __pyx_v_counter;
                                      __pyx_t_69 = __pyx_v_leaf_numb;
                                      __pyx_t_70 = __pyx_v_s;
                                      __pyx_t_71 = __pyx_v_b;
                                      __pyx_t_72 = __pyx_v_leaf_numb;
                                      __pyx_t_54 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_67 * __pyx_v_S.strides[0]) ) + __pyx_t_68 * __pyx_v_S.strides[1]) ) + __pyx_t_69 * __pyx_v_S.strides[2]) )) + __pyx_t_70)) )));
                                      __pyx_t_55 = 0;
                                      __pyx_t_16 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_34 * __pyx_v_data.strides[0]) ) + __pyx_t_40 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_61 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_66 * __pyx_v_data.strides[0]) ) + __pyx_t_50 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_71 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_72 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_16) {

                                        /* "cyext_acv.pyx":3500
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == S_size + len_n_star:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":3499
 *                         b_it = 0
 *                         for s in range(S_size + len_n_star):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":3502
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 */
                                    __pyx_t_16 = ((__pyx_v_b_it == (__pyx_v_S_size + __pyx_v_len_n_star)) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3503
 * 
 *                         if b_it == S_size + len_n_star:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 */
                                      __pyx_t_70 = __pyx_v_va;
                                      __pyx_t_69 = __pyx_v_leaf_numb;
                                      __pyx_t_68 = __pyx_v_counter;
                                      __pyx_t_67 = __pyx_v_va;
                                      __pyx_t_72 = __pyx_v_leaf_numb;
                                      __pyx_t_71 = __pyx_v_counter;
                                      *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_67 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_72 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_71 * __pyx_v_lm_s.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_70 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_69 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_68 * __pyx_v_lm_s.strides[2]) ))) + 1.0);

                                      /* "cyext_acv.pyx":3504
 *                         if b_it == S_size + len_n_star:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":3505
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_19 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_20 = __pyx_t_19;
                                      for (__pyx_t_58 = 0; __pyx_t_58 < __pyx_t_20; __pyx_t_58+=1) {
                                        __pyx_v_nv = __pyx_t_58;

                                        /* "cyext_acv.pyx":3506
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_68 = __pyx_v_i;
                                        __pyx_t_55 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_69 = __pyx_v_b;
                                        __pyx_t_70 = __pyx_v_leaf_numb;
                                        __pyx_t_54 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_50 = 1;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_68 * __pyx_v_data.strides[0]) ) + __pyx_t_55 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_70 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_41) {
                                        } else {
                                          __pyx_t_16 = __pyx_t_41;
                                          goto __pyx_L100_bool_binop_done;
                                        }
                                        __pyx_t_70 = __pyx_v_i;
                                        __pyx_t_50 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_69 = __pyx_v_b;
                                        __pyx_t_68 = __pyx_v_leaf_numb;
                                        __pyx_t_54 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_55 = 0;
                                        __pyx_t_41 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_70 * __pyx_v_data.strides[0]) ) + __pyx_t_50 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_69 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_68 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_16 = __pyx_t_41;
                                        __pyx_L100_bool_binop_done:;
                                        if (__pyx_t_16) {

                                          /* "cyext_acv.pyx":3507
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":3508
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L97_continue;

                                          /* "cyext_acv.pyx":3506
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L97_continue:;
                                      }

                                      /* "cyext_acv.pyx":3510
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      __pyx_t_16 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_16) {

                                        /* "cyext_acv.pyx":3511
 * 
 *                             if nv_bool == 0:
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                        __pyx_t_68 = __pyx_v_va;
                                        __pyx_t_69 = __pyx_v_leaf_numb;
                                        __pyx_t_70 = __pyx_v_counter;
                                        __pyx_t_71 = __pyx_v_va;
                                        __pyx_t_72 = __pyx_v_leaf_numb;
                                        __pyx_t_67 = __pyx_v_counter;
                                        *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_71 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_72 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_67 * __pyx_v_lm_si.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_68 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_69 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_70 * __pyx_v_lm_si.strides[2]) ))) + 1.0);

                                        /* "cyext_acv.pyx":3510
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3502
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 *                             nv_bool = 0
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3515
 * 
 * 
 *                     p_s = 0             # <<<<<<<<<<<<<<
 *                     p_si = 0
 * 
 */
                                  __pyx_v_p_s = 0.0;

                                  /* "cyext_acv.pyx":3516
 * 
 *                     p_s = 0
 *                     p_si = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  __pyx_v_p_si = 0.0;

                                  /* "cyext_acv.pyx":3519
 * 
 * 
 *                     csi = 0             # <<<<<<<<<<<<<<
 *                     cs = 0
 *                     o_all = 0
 */
                                  __pyx_v_csi = 0;

                                  /* "cyext_acv.pyx":3520
 * 
 *                     csi = 0
 *                     cs = 0             # <<<<<<<<<<<<<<
 *                     o_all = 0
 *                     for s in range(S_size + len_n_star):
 */
                                  __pyx_v_cs = 0;

                                  /* "cyext_acv.pyx":3521
 *                     csi = 0
 *                     cs = 0
 *                     o_all = 0             # <<<<<<<<<<<<<<
 *                     for s in range(S_size + len_n_star):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                  __pyx_v_o_all = 0;

                                  /* "cyext_acv.pyx":3522
 *                     cs = 0
 *                     o_all = 0
 *                     for s in range(S_size + len_n_star):             # <<<<<<<<<<<<<<
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                             o_all = o_all + 1
 */
                                  __pyx_t_45 = (__pyx_v_S_size + __pyx_v_len_n_star);
                                  __pyx_t_58 = __pyx_t_45;
                                  for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_58; __pyx_t_59+=1) {
                                    __pyx_v_s = __pyx_t_59;

                                    /* "cyext_acv.pyx":3523
 *                     o_all = 0
 *                     for s in range(S_size + len_n_star):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    __pyx_t_70 = __pyx_v_va;
                                    __pyx_t_69 = __pyx_v_counter;
                                    __pyx_t_68 = __pyx_v_leaf_numb;
                                    __pyx_t_67 = __pyx_v_s;
                                    __pyx_t_55 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_70 * __pyx_v_S.strides[0]) ) + __pyx_t_69 * __pyx_v_S.strides[1]) ) + __pyx_t_68 * __pyx_v_S.strides[2]) )) + __pyx_t_67)) )));
                                    __pyx_t_72 = __pyx_v_va;
                                    __pyx_t_71 = __pyx_v_counter;
                                    __pyx_t_65 = __pyx_v_leaf_numb;
                                    __pyx_t_64 = __pyx_v_s;
                                    __pyx_t_63 = __pyx_v_b;
                                    __pyx_t_62 = __pyx_v_leaf_numb;
                                    __pyx_t_54 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_72 * __pyx_v_S.strides[0]) ) + __pyx_t_71 * __pyx_v_S.strides[1]) ) + __pyx_t_65 * __pyx_v_S.strides[2]) )) + __pyx_t_64)) )));
                                    __pyx_t_50 = 1;
                                    __pyx_t_66 = __pyx_v_va;
                                    __pyx_t_24 = __pyx_v_counter;
                                    __pyx_t_53 = __pyx_v_leaf_numb;
                                    __pyx_t_48 = __pyx_v_s;
                                    __pyx_t_47 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_66 * __pyx_v_S.strides[0]) ) + __pyx_t_24 * __pyx_v_S.strides[1]) ) + __pyx_t_53 * __pyx_v_S.strides[2]) )) + __pyx_t_48)) )));
                                    __pyx_t_51 = __pyx_v_va;
                                    __pyx_t_61 = __pyx_v_counter;
                                    __pyx_t_38 = __pyx_v_leaf_numb;
                                    __pyx_t_33 = __pyx_v_s;
                                    __pyx_t_39 = __pyx_v_b;
                                    __pyx_t_52 = __pyx_v_leaf_numb;
                                    __pyx_t_46 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) ) + __pyx_t_61 * __pyx_v_S.strides[1]) ) + __pyx_t_38 * __pyx_v_S.strides[2]) )) + __pyx_t_33)) )));
                                    __pyx_t_40 = 0;
                                    __pyx_t_16 = ((((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_55 * __pyx_v_X.strides[0]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_62 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_47 * __pyx_v_X.strides[0]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_39 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3524
 *                     for s in range(S_size + len_n_star):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                             o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                     if o_all == S_size + len_n_star:
 */
                                      __pyx_v_o_all = (__pyx_v_o_all + 1);

                                      /* "cyext_acv.pyx":3523
 *                     o_all = 0
 *                     for s in range(S_size + len_n_star):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3526
 *                             o_all = o_all + 1
 * 
 *                     if o_all == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                         cs = 1
 *                         nv_bool = 0
 */
                                  __pyx_t_16 = ((__pyx_v_o_all == (__pyx_v_S_size + __pyx_v_len_n_star)) != 0);
                                  if (__pyx_t_16) {

                                    /* "cyext_acv.pyx":3527
 * 
 *                     if o_all == S_size + len_n_star:
 *                         cs = 1             # <<<<<<<<<<<<<<
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 */
                                    __pyx_v_cs = 1;

                                    /* "cyext_acv.pyx":3528
 *                     if o_all == S_size + len_n_star:
 *                         cs = 1
 *                         nv_bool = 0             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                    __pyx_v_nv_bool = 0;

                                    /* "cyext_acv.pyx":3529
 *                         cs = 1
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1
 */
                                    __pyx_t_19 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_20 = __pyx_t_19;
                                    for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_20; __pyx_t_45+=1) {
                                      __pyx_v_nv = __pyx_t_45;

                                      /* "cyext_acv.pyx":3530
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      __pyx_t_40 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_33 = __pyx_v_b;
                                      __pyx_t_38 = __pyx_v_leaf_numb;
                                      __pyx_t_46 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_47 = 1;
                                      __pyx_t_41 = (((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_40 * __pyx_v_X.strides[0]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_47 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_41) {
                                      } else {
                                        __pyx_t_16 = __pyx_t_41;
                                        goto __pyx_L110_bool_binop_done;
                                      }
                                      __pyx_t_47 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_38 = __pyx_v_b;
                                      __pyx_t_33 = __pyx_v_leaf_numb;
                                      __pyx_t_46 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_40 = 0;
                                      __pyx_t_41 = (((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_47 * __pyx_v_X.strides[0]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_38 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_46 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_40 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_16 = __pyx_t_41;
                                      __pyx_L110_bool_binop_done:;
                                      if (__pyx_t_16) {

                                        /* "cyext_acv.pyx":3531
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                        __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                        /* "cyext_acv.pyx":3532
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1
 *                                 continue             # <<<<<<<<<<<<<<
 * 
 *                         if nv_bool == 0:
 */
                                        goto __pyx_L107_continue;

                                        /* "cyext_acv.pyx":3530
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      }
                                      __pyx_L107_continue:;
                                    }

                                    /* "cyext_acv.pyx":3534
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    __pyx_t_16 = ((__pyx_v_nv_bool == 0) != 0);
                                    if (__pyx_t_16) {

                                      /* "cyext_acv.pyx":3535
 * 
 *                         if nv_bool == 0:
 *                             csi = 1             # <<<<<<<<<<<<<<
 * 
 *                     coef = 0
 */
                                      __pyx_v_csi = 1;

                                      /* "cyext_acv.pyx":3534
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":3526
 *                             o_all = o_all + 1
 * 
 *                     if o_all == S_size + len_n_star:             # <<<<<<<<<<<<<<
 *                         cs = 1
 *                         nv_bool = 0
 */
                                  }

                                  /* "cyext_acv.pyx":3537
 *                             csi = 1
 * 
 *                     coef = 0             # <<<<<<<<<<<<<<
 *                     for l in range(1, m - set_size):
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 */
                                  __pyx_v_coef = 0.0;

                                  /* "cyext_acv.pyx":3538
 * 
 *                     coef = 0
 *                     for l in range(1, m - set_size):             # <<<<<<<<<<<<<<
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 */
                                  __pyx_t_56 = (__pyx_v_m - __pyx_v_set_size);
                                  __pyx_t_57 = __pyx_t_56;
                                  for (__pyx_t_45 = 1; __pyx_t_45 < __pyx_t_57; __pyx_t_45+=1) {
                                    __pyx_v_l = __pyx_t_45;

                                    /* "cyext_acv.pyx":3539
 *                     coef = 0
 *                     for l in range(1, m - set_size):
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 */
                                    if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0) != 0.0) != 0)) {
                                      __pyx_t_73 = (__pyx_v_coef + ((1. * __pyx_f_9cyext_acv_binomialC(((__pyx_v_m - __pyx_v_set_size) - 1), __pyx_v_l, 0)) / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0)));
                                    } else {
                                      __pyx_t_73 = 0.0;
                                    }
                                    __pyx_v_coef = __pyx_t_73;
                                  }

                                  /* "cyext_acv.pyx":3541
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0) != 0.0) != 0)) {
                                    __pyx_t_73 = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));
                                  } else {
                                    __pyx_t_73 = 0.0;
                                  }
                                  __pyx_v_coef_0 = __pyx_t_73;

                                  /* "cyext_acv.pyx":3544
 * 
 * 
 *                     p_s = (cs * lm)/lm_s[va, leaf_numb, counter]             # <<<<<<<<<<<<<<
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 */
                                  __pyx_t_33 = __pyx_v_va;
                                  __pyx_t_38 = __pyx_v_leaf_numb;
                                  __pyx_t_61 = __pyx_v_counter;
                                  __pyx_v_p_s = ((__pyx_v_cs * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_61 * __pyx_v_lm_s.strides[2]) ))));

                                  /* "cyext_acv.pyx":3545
 * 
 *                     p_s = (cs * lm)/lm_s[va, leaf_numb, counter]
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter]             # <<<<<<<<<<<<<<
 * 
 *                     if S_size != 0:
 */
                                  __pyx_t_61 = __pyx_v_va;
                                  __pyx_t_38 = __pyx_v_leaf_numb;
                                  __pyx_t_33 = __pyx_v_counter;
                                  __pyx_v_p_si = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_61 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_38 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_33 * __pyx_v_lm_si.strides[2]) ))));

                                  /* "cyext_acv.pyx":3547
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 *                     if S_size != 0:             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 */
                                  __pyx_t_16 = ((__pyx_v_S_size != 0) != 0);
                                  if (__pyx_t_16) {

                                    /* "cyext_acv.pyx":3548
 * 
 *                     if S_size != 0:
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                    __pyx_t_19 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_20 = __pyx_t_19;
                                    for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_20; __pyx_t_45+=1) {
                                      __pyx_v_nv = __pyx_t_45;

                                      /* "cyext_acv.pyx":3549
 *                     if S_size != 0:
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                      __pyx_t_58 = __pyx_v_d;
                                      __pyx_t_59 = __pyx_t_58;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_59; __pyx_t_60+=1) {
                                        __pyx_v_i2 = __pyx_t_60;

                                        /* "cyext_acv.pyx":3550
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
                                        __pyx_t_33 = __pyx_v_b;
                                        __pyx_t_38 = __pyx_v_leaf_numb;
                                        __pyx_t_61 = __pyx_v_b;
                                        __pyx_t_40 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_38 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_51 = __pyx_v_i2;
                                        __pyx_t_52 = __pyx_v_leaf_numb;
                                        __pyx_t_39 = __pyx_v_counter;
                                        __pyx_t_46 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_48 = __pyx_v_i2;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_52 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_46 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_48 * __pyx_v_phi_b.strides[3]) )) += (((__pyx_v_coef_0 + __pyx_v_coef) * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_61 * __pyx_v_values.strides[0]) ) + __pyx_t_40 * __pyx_v_values.strides[1]) ) + __pyx_t_51 * __pyx_v_values.strides[2]) ))));
                                      }
                                    }

                                    /* "cyext_acv.pyx":3547
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter]
 * 
 *                     if S_size != 0:             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 */
                                    goto __pyx_L115;
                                  }

                                  /* "cyext_acv.pyx":3553
 * 
 *                     else:
 *                         p_off = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 */
                                  /*else*/ {
                                    __pyx_v_p_off = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                    /* "cyext_acv.pyx":3554
 *                     else:
 *                         p_off = lm/data.shape[0]
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                    __pyx_t_19 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_20 = __pyx_t_19;
                                    for (__pyx_t_45 = 0; __pyx_t_45 < __pyx_t_20; __pyx_t_45+=1) {
                                      __pyx_v_nv = __pyx_t_45;

                                      /* "cyext_acv.pyx":3555
 *                         p_off = lm/data.shape[0]
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                      __pyx_t_58 = __pyx_v_d;
                                      __pyx_t_59 = __pyx_t_58;
                                      for (__pyx_t_60 = 0; __pyx_t_60 < __pyx_t_59; __pyx_t_60+=1) {
                                        __pyx_v_i2 = __pyx_t_60;

                                        /* "cyext_acv.pyx":3556
 *                         for nv in range(va_id[va].size()):
 *                             for i2 in range(d):
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *         for j in range(data.shape[1]):
 */
                                        __pyx_t_38 = __pyx_v_b;
                                        __pyx_t_33 = __pyx_v_leaf_numb;
                                        __pyx_t_51 = __pyx_v_b;
                                        __pyx_t_40 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_38 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_33 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_61 = __pyx_v_i2;
                                        __pyx_t_48 = __pyx_v_b;
                                        __pyx_t_39 = __pyx_v_leaf_numb;
                                        __pyx_t_52 = __pyx_v_b;
                                        __pyx_t_46 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_48 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_39 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                        __pyx_t_53 = __pyx_v_i2;
                                        __pyx_t_24 = __pyx_v_leaf_numb;
                                        __pyx_t_66 = __pyx_v_counter;
                                        __pyx_t_47 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_64 = __pyx_v_i2;
                                        *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_24 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_66 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_47 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_64 * __pyx_v_phi_b.strides[3]) )) += (((__pyx_v_p_si - __pyx_v_p_off) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_51 * __pyx_v_values.strides[0]) ) + __pyx_t_40 * __pyx_v_values.strides[1]) ) + __pyx_t_61 * __pyx_v_values.strides[2]) )))) + ((__pyx_v_coef * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_52 * __pyx_v_values.strides[0]) ) + __pyx_t_46 * __pyx_v_values.strides[1]) ) + __pyx_t_53 * __pyx_v_values.strides[2]) )))));
                                      }
                                    }
                                  }
                                  __pyx_L115:;
                                }
                                __pyx_L65_continue:;
                              }
                              goto __pyx_L125;
                              __pyx_L37_error:;
                              {
                                  #ifdef WITH_THREAD
                                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                  #endif
                                  #ifdef _OPENMP
                                  #pragma omp flush(__pyx_parallel_exc_type)
                                  #endif /* _OPENMP */
                                  if (!__pyx_parallel_exc_type) {
                                    __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                    __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                    __Pyx_GOTREF(__pyx_parallel_exc_type);
                                  }
                                  #ifdef WITH_THREAD
                                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                  #endif
                              }
                              __pyx_parallel_why = 4;
                              goto __pyx_L124;
                              __pyx_L124:;
                              #ifdef _OPENMP
                              #pragma omp critical(__pyx_parallel_lastprivates3)
                              #endif /* _OPENMP */
                              {
                                  __pyx_parallel_temp0 = __pyx_v_S_size;
                                  __pyx_parallel_temp1 = __pyx_v_a_it;
                                  __pyx_parallel_temp2 = __pyx_v_add;
                                  __pyx_parallel_temp3 = __pyx_v_b_it;
                                  __pyx_parallel_temp4 = __pyx_v_ci;
                                  __pyx_parallel_temp5 = __pyx_v_cj;
                                  __pyx_parallel_temp6 = __pyx_v_coef;
                                  __pyx_parallel_temp7 = __pyx_v_coef_0;
                                  __pyx_parallel_temp8 = __pyx_v_counter;
                                  __pyx_parallel_temp9 = __pyx_v_cs;
                                  __pyx_parallel_temp10 = __pyx_v_csi;
                                  __pyx_parallel_temp11 = __pyx_v_i;
                                  __pyx_parallel_temp12 = __pyx_v_i2;
                                  __pyx_parallel_temp13 = __pyx_v_l;
                                  __pyx_parallel_temp14 = __pyx_v_leaf_numb;
                                  __pyx_parallel_temp15 = __pyx_v_lm;
                                  __pyx_parallel_temp16 = __pyx_v_ns;
                                  __pyx_parallel_temp17 = __pyx_v_nv;
                                  __pyx_parallel_temp18 = __pyx_v_nv_bool;
                                  __pyx_parallel_temp19 = __pyx_v_o_all;
                                  __pyx_parallel_temp20 = __pyx_v_p_off;
                                  __pyx_parallel_temp21 = __pyx_v_p_s;
                                  __pyx_parallel_temp22 = __pyx_v_p_si;
                                  __pyx_parallel_temp23 = __pyx_v_pow_set_size;
                                  __pyx_parallel_temp24 = __pyx_v_s;
                                  __pyx_parallel_temp25 = __pyx_v_set_size;
                                  __pyx_parallel_temp26 = __pyx_v_va;
                                  __pyx_parallel_temp27 = __pyx_v_va_size;
                              }
                              __pyx_L125:;
                              #ifdef _OPENMP
                              #pragma omp flush(__pyx_parallel_why)
                              #endif /* _OPENMP */
                          }
                      }
                      #ifdef _OPENMP
                      Py_END_ALLOW_THREADS
                      #else
{
#ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      #endif /* _OPENMP */
                      /* Clean up any temporaries */
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      #ifndef _OPENMP
}
#endif /* _OPENMP */
                  }
              }
              if (__pyx_parallel_exc_type) {
                /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
                __pyx_parallel_why = 4;
              }
              if (__pyx_parallel_why) {
                __pyx_v_S_size = __pyx_parallel_temp0;
                __pyx_v_a_it = __pyx_parallel_temp1;
                __pyx_v_add = __pyx_parallel_temp2;
                __pyx_v_b_it = __pyx_parallel_temp3;
                __pyx_v_ci = __pyx_parallel_temp4;
                __pyx_v_cj = __pyx_parallel_temp5;
                __pyx_v_coef = __pyx_parallel_temp6;
                __pyx_v_coef_0 = __pyx_parallel_temp7;
                __pyx_v_counter = __pyx_parallel_temp8;
                __pyx_v_cs = __pyx_parallel_temp9;
                __pyx_v_csi = __pyx_parallel_temp10;
                __pyx_v_i = __pyx_parallel_temp11;
                __pyx_v_i2 = __pyx_parallel_temp12;
                __pyx_v_l = __pyx_parallel_temp13;
                __pyx_v_leaf_numb = __pyx_parallel_temp14;
                __pyx_v_lm = __pyx_parallel_temp15;
                __pyx_v_ns = __pyx_parallel_temp16;
                __pyx_v_nv = __pyx_parallel_temp17;
                __pyx_v_nv_bool = __pyx_parallel_temp18;
                __pyx_v_o_all = __pyx_parallel_temp19;
                __pyx_v_p_off = __pyx_parallel_temp20;
                __pyx_v_p_s = __pyx_parallel_temp21;
                __pyx_v_p_si = __pyx_parallel_temp22;
                __pyx_v_pow_set_size = __pyx_parallel_temp23;
                __pyx_v_s = __pyx_parallel_temp24;
                __pyx_v_set_size = __pyx_parallel_temp25;
                __pyx_v_va = __pyx_parallel_temp26;
                __pyx_v_va_size = __pyx_parallel_temp27;
                switch (__pyx_parallel_why) {
                      case 4:
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      __Pyx_GIVEREF(__pyx_parallel_exc_type);
                      __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                      __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                  }
                  goto __pyx_L33_error;
                }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":3396
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             for i in range(N_star.shape[0]):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L34;
          }
          __pyx_L33_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L34:;
        }
    }

    /* "cyext_acv.pyx":3558
 *                                 phi_b[leaf_numb, counter, va_id[va][nv], i2] += (p_si-p_off)*values[b, leaf_idx_trees[b, leaf_numb], i2] + coef * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 *         for j in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 */
    __pyx_t_22 = (__pyx_v_data.shape[1]);
    __pyx_t_23 = __pyx_t_22;
    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_23; __pyx_t_28+=1) {
      __pyx_v_j = __pyx_t_28;

      /* "cyext_acv.pyx":3559
 * 
 *         for j in range(data.shape[1]):
 *             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi[j, i2] += phi_n[leaf_numb, j, i2]
 */
      __pyx_t_27 = __pyx_v_d;
      __pyx_t_26 = __pyx_t_27;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_26; __pyx_t_29+=1) {
        __pyx_v_i2 = __pyx_t_29;

        /* "cyext_acv.pyx":3560
 *         for j in range(data.shape[1]):
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                     phi[j, i2] += phi_n[leaf_numb, j, i2]
 *                     for counter in range(phi_b.shape[1]):
 */
        __pyx_t_30 = (__pyx_v_phi_b.shape[0]);
        __pyx_t_31 = __pyx_t_30;
        for (__pyx_t_32 = 0; __pyx_t_32 < __pyx_t_31; __pyx_t_32+=1) {
          __pyx_v_leaf_numb = __pyx_t_32;

          /* "cyext_acv.pyx":3561
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi[j, i2] += phi_n[leaf_numb, j, i2]             # <<<<<<<<<<<<<<
 *                     for counter in range(phi_b.shape[1]):
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 */
          __pyx_t_39 = __pyx_v_leaf_numb;
          __pyx_t_48 = __pyx_v_j;
          __pyx_t_53 = __pyx_v_i2;
          __pyx_t_52 = __pyx_v_j;
          __pyx_t_33 = __pyx_v_i2;
          *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_52 * __pyx_v_phi.strides[0]) ) + __pyx_t_33 * __pyx_v_phi.strides[1]) )) += (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_n.data + __pyx_t_39 * __pyx_v_phi_n.strides[0]) ) + __pyx_t_48 * __pyx_v_phi_n.strides[1]) ) + __pyx_t_53 * __pyx_v_phi_n.strides[2]) )));

          /* "cyext_acv.pyx":3562
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     phi[j, i2] += phi_n[leaf_numb, j, i2]
 *                     for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 * 
 */
          __pyx_t_35 = (__pyx_v_phi_b.shape[1]);
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_counter = __pyx_t_37;

            /* "cyext_acv.pyx":3563
 *                     phi[j, i2] += phi_n[leaf_numb, j, i2]
 *                     for counter in range(phi_b.shape[1]):
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(phi)/m
 */
            __pyx_t_53 = __pyx_v_leaf_numb;
            __pyx_t_48 = __pyx_v_counter;
            __pyx_t_39 = __pyx_v_j;
            __pyx_t_33 = __pyx_v_i2;
            __pyx_t_52 = __pyx_v_j;
            __pyx_t_38 = __pyx_v_i2;
            *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_52 * __pyx_v_phi.strides[0]) ) + __pyx_t_38 * __pyx_v_phi.strides[1]) )) += (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_53 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_48 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_39 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_33 * __pyx_v_phi_b.strides[3]) )));
          }
        }
      }
    }
  }

  /* "cyext_acv.pyx":3565
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 * 
 *     return np.asarray(phi)/m             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_phi, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":3301
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef single_shap_values_acv_leaves(const double[:] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __Pyx_AddTraceback("cyext_acv.single_shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_N_star_a, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_si, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_39single_shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_39single_shap_values_acv_leaves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  __Pyx_memviewslice __pyx_v_S_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_N_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("single_shap_values_acv_leaves (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_S_star,&__pyx_n_s_N_star,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 1); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 2); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 3); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 4); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 5); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 6); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 7); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 8); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_N_star)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 9); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 10); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, 11); __PYX_ERR(0, 3301, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "single_shap_values_acv_leaves") < 0)) __PYX_ERR(0, 3301, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 3301, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 3302, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 3303, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 3304, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 3305, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 3306, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_scaling == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3307, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3308, __pyx_L3_error)
    __pyx_v_S_star = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[8], 0); if (unlikely(!__pyx_v_S_star.memview)) __PYX_ERR(0, 3308, __pyx_L3_error)
    __pyx_v_N_star = __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(values[9], 0); if (unlikely(!__pyx_v_N_star.memview)) __PYX_ERR(0, 3308, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[10]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3309, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3309, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("single_shap_values_acv_leaves", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3301, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.single_shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_38single_shap_values_acv_leaves(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_38single_shap_values_acv_leaves(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, __Pyx_memviewslice __pyx_v_S_star, __Pyx_memviewslice __pyx_v_N_star, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_shap_values_acv_leaves", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_S_star.memview)) { __Pyx_RaiseUnboundLocalError("S_star"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  if (unlikely(!__pyx_v_N_star.memview)) { __Pyx_RaiseUnboundLocalError("N_star"); __PYX_ERR(0, 3301, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_single_shap_values_acv_leaves(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_S_star, __pyx_v_N_star, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.single_shap_values_acv_leaves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_N_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3573
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef single_shap_values(const double[:] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

static PyObject *__pyx_pw_9cyext_acv_41single_shap_values(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_single_shap_values(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  const __pyx_v_C, CYTHON_UNUSED int __pyx_v_num_threads, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_max_leaves;
  __Pyx_memviewslice __pyx_v_phi = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_phi_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  CYTHON_UNUSED unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_va;
  unsigned int __pyx_v_counter;
  unsigned int __pyx_v_ci;
  unsigned int __pyx_v_cj;
  unsigned int __pyx_v_pow_set_size;
  unsigned int __pyx_v_nv;
  unsigned int __pyx_v_ns;
  unsigned int __pyx_v_add;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_leaf_numb;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_s;
  unsigned int __pyx_v_j;
  unsigned int __pyx_v_i2;
  unsigned int __pyx_v_l;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_csi;
  unsigned int __pyx_v_cs;
  double __pyx_v_lm;
  std::vector<std::vector<long> >  __pyx_v_C_buff;
  std::vector<std::vector<long> >  __pyx_v_va_id;
  std::vector<std::vector<long> >  __pyx_v_buff;
  std::vector<long>  __pyx_v_coal_va;
  std::vector<long>  __pyx_v_remove_va;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id_v2;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_C_b;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_Sm;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_id;
  std::vector<std::vector<std::vector<std::vector<long> > > >  __pyx_v_node_id_va;
  int __pyx_v_S_size;
  int __pyx_v_va_size;
  int __pyx_v_b_it;
  int __pyx_v_nv_bool;
  long __pyx_v_set_size;
  double __pyx_v_p_s;
  double __pyx_v_p_si;
  double __pyx_v_coef;
  double __pyx_v_coef_0;
  __Pyx_memviewslice __pyx_v_lm_s = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lm_si = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  unsigned int __pyx_t_13;
  int __pyx_t_14;
  std::vector<std::vector<long> > ::size_type __pyx_t_15;
  std::vector<std::vector<long> > ::size_type __pyx_t_16;
  std::vector<long> ::size_type __pyx_t_17;
  std::vector<long> ::size_type __pyx_t_18;
  unsigned int __pyx_t_19;
  __Pyx_memviewslice __pyx_t_20 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  unsigned int __pyx_t_25;
  unsigned int __pyx_t_26;
  unsigned int __pyx_t_27;
  unsigned int __pyx_t_28;
  unsigned int __pyx_t_29;
  unsigned int __pyx_t_30;
  unsigned int __pyx_t_31;
  size_t __pyx_t_32;
  size_t __pyx_t_33;
  size_t __pyx_t_34;
  size_t __pyx_t_35;
  size_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  int __pyx_t_38;
  std::vector<long> ::size_type __pyx_t_39;
  std::vector<long> ::size_type __pyx_t_40;
  long __pyx_t_41;
  long __pyx_t_42;
  unsigned int __pyx_t_43;
  int __pyx_t_44;
  int __pyx_t_45;
  size_t __pyx_t_46;
  size_t __pyx_t_47;
  size_t __pyx_t_48;
  size_t __pyx_t_49;
  size_t __pyx_t_50;
  size_t __pyx_t_51;
  Py_ssize_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  size_t __pyx_t_54;
  size_t __pyx_t_55;
  size_t __pyx_t_56;
  size_t __pyx_t_57;
  size_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  size_t __pyx_t_60;
  size_t __pyx_t_61;
  size_t __pyx_t_62;
  size_t __pyx_t_63;
  size_t __pyx_t_64;
  size_t __pyx_t_65;
  Py_ssize_t __pyx_t_66;
  Py_ssize_t __pyx_t_67;
  double __pyx_t_68;
  unsigned int __pyx_t_69;
  unsigned int __pyx_t_70;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_shap_values", 0);

  /* "cyext_acv.pyx":3584
 * 
 * 
 *     cdef unsigned int d = values.shape[2]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]
 */
  __pyx_v_d = (__pyx_v_values.shape[2]);

  /* "cyext_acv.pyx":3585
 * 
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int m = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_m = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3586
 *     cdef unsigned int d = values.shape[2]
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 * 
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":3587
 *     cdef unsigned int m = X.shape[0]
 *     cdef unsigned int n_trees = values.shape[0]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :] phi
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":3590
 * 
 *     cdef double[:, :] phi
 *     phi = np.zeros((m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[ :, :, :, :] phi_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 3590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":3593
 * 
 *     cdef double[ :, :, :, :] phi_b
 *     phi_b = np.zeros((max_leaves, 2**scaling, m, d))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:, :, :, ::1] S
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_d); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_phi_b = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3596
 * 
 *     cdef long[:, :, :, ::1] S
 *     S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3596, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_S = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3604
 *     cdef vector[vector[long]] C_buff, va_id, buff
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     for i in range(max_leaves):
 *         buff[i].resize(1)
 */
  try {
    __pyx_v_buff.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3604, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3605
 *     cdef vector[long] coal_va, remove_va, node_id_b, buff_l
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):             # <<<<<<<<<<<<<<
 *         buff[i].resize(1)
 * 
 */
  __pyx_t_11 = __pyx_v_max_leaves;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":3606
 *     buff.resize(max_leaves)
 *     for i in range(max_leaves):
 *         buff[i].resize(1)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 */
    try {
      (__pyx_v_buff[__pyx_v_i]).resize(1);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 3606, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":3609
 * 
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 */
  try {
    __pyx_v_node_id_v2.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3609, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3610
 *     cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)
 */
  try {
    __pyx_v_C_b.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3610, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3611
 *     node_id_v2.resize(max_leaves)
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)             # <<<<<<<<<<<<<<
 *     node_id.resize(max_leaves)
 * 
 */
  try {
    __pyx_v_Sm.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3611, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3612
 *     C_b.resize(max_leaves)
 *     Sm.resize(max_leaves)
 *     node_id.resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 */
  try {
    __pyx_v_node_id.resize(__pyx_v_max_leaves);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3612, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3615
 * 
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)             # <<<<<<<<<<<<<<
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)
 */
  try {
    __pyx_v_node_id_va.resize(__pyx_v_m);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3615, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3616
 *     cdef vector[vector[vector[vector[long]]]] node_id_va
 *     node_id_va.resize(m)
 *     for i in range(m):             # <<<<<<<<<<<<<<
 *         node_id_va[i].resize(max_leaves)
 * 
 */
  __pyx_t_11 = __pyx_v_m;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":3617
 *     node_id_va.resize(m)
 *     for i in range(m):
 *         node_id_va[i].resize(max_leaves)             # <<<<<<<<<<<<<<
 * 
 *     cdef int S_size, va_size,  b_it, nv_bool
 */
    try {
      (__pyx_v_node_id_va[__pyx_v_i]).resize(__pyx_v_max_leaves);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 3617, __pyx_L1_error)
    }
  }

  /* "cyext_acv.pyx":3623
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
  __pyx_t_14 = (((__pyx_v_C[0]).size() != 0) != 0);
  if (__pyx_t_14) {

    /* "cyext_acv.pyx":3624
 * 
 *     if C[0].size() != 0:
 *         C_buff = C             # <<<<<<<<<<<<<<
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 */
    __pyx_v_C_buff = __pyx_v_C;

    /* "cyext_acv.pyx":3625
 *     if C[0].size() != 0:
 *         C_buff = C
 *         for ci in range(C.size()):             # <<<<<<<<<<<<<<
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])
 */
    __pyx_t_15 = __pyx_v_C.size();
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_16; __pyx_t_11+=1) {
      __pyx_v_ci = __pyx_t_11;

      /* "cyext_acv.pyx":3626
 *         C_buff = C
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):             # <<<<<<<<<<<<<<
 *                 coal_va.push_back(C[ci][cj])
 * 
 */
      __pyx_t_17 = (__pyx_v_C[__pyx_v_ci]).size();
      __pyx_t_18 = __pyx_t_17;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_18; __pyx_t_12+=1) {
        __pyx_v_cj = __pyx_t_12;

        /* "cyext_acv.pyx":3627
 *         for ci in range(C.size()):
 *             for cj in range(C[ci].size()):
 *                 coal_va.push_back(C[ci][cj])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(m):
 */
        try {
          __pyx_v_coal_va.push_back(((__pyx_v_C[__pyx_v_ci])[__pyx_v_cj]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3627, __pyx_L1_error)
        }
      }
    }

    /* "cyext_acv.pyx":3629
 *                 coal_va.push_back(C[ci][cj])
 * 
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 */
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3630
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
      __pyx_t_14 = ((!((std::find<std::vector<long> ::iterator,long>(__pyx_v_coal_va.begin(), __pyx_v_coal_va.end(), __pyx_v_i) != __pyx_v_coal_va.end()) != 0)) != 0);
      if (__pyx_t_14) {

        /* "cyext_acv.pyx":3631
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)             # <<<<<<<<<<<<<<
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])
 */
        try {
          __pyx_v_remove_va.push_back(__pyx_v_i);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3631, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3632
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i             # <<<<<<<<<<<<<<
 *                 va_id.push_back(buff[0])
 *             else:
 */
        ((__pyx_v_buff[0])[0]) = __pyx_v_i;

        /* "cyext_acv.pyx":3633
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 *                 va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 for ci in range(C_buff.size()):
 */
        try {
          __pyx_v_va_id.push_back((__pyx_v_buff[0]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3633, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3630
 * 
 *         for i in range(m):
 *             if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():             # <<<<<<<<<<<<<<
 *                 remove_va.push_back(i)
 *                 buff[0][0] = i
 */
        goto __pyx_L14;
      }

      /* "cyext_acv.pyx":3635
 *                 va_id.push_back(buff[0])
 *             else:
 *                 for ci in range(C_buff.size()):             # <<<<<<<<<<<<<<
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 */
      /*else*/ {
        __pyx_t_15 = __pyx_v_C_buff.size();
        __pyx_t_16 = __pyx_t_15;
        for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_16; __pyx_t_19+=1) {
          __pyx_v_ci = __pyx_t_19;

          /* "cyext_acv.pyx":3636
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          __pyx_t_14 = ((std::find<std::vector<long> ::iterator,long>((__pyx_v_C_buff[__pyx_v_ci]).begin(), (__pyx_v_C_buff[__pyx_v_ci]).end(), __pyx_v_i) != (__pyx_v_C_buff[__pyx_v_ci]).end()) != 0);
          if (__pyx_t_14) {

            /* "cyext_acv.pyx":3637
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])             # <<<<<<<<<<<<<<
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 */
            try {
              __pyx_v_va_id.push_back((__pyx_v_C_buff[__pyx_v_ci]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 3637, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":3638
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])             # <<<<<<<<<<<<<<
 *                         C_buff.pop_back()
 *                         break
 */
            (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(__pyx_v_C_buff.begin(), __pyx_v_C_buff.end(), (__pyx_v_C_buff[__pyx_v_ci])));

            /* "cyext_acv.pyx":3639
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
            __pyx_v_C_buff.pop_back();

            /* "cyext_acv.pyx":3640
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 *                         C_buff.pop_back()
 *                         break             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
            goto __pyx_L16_break;

            /* "cyext_acv.pyx":3636
 *             else:
 *                 for ci in range(C_buff.size()):
 *                     if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):             # <<<<<<<<<<<<<<
 *                         va_id.push_back(C_buff[ci])
 *                         std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
 */
          }
        }
        __pyx_L16_break:;
      }
      __pyx_L14:;
    }

    /* "cyext_acv.pyx":3623
 *     cdef long set_size
 * 
 *     if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *         C_buff = C
 *         for ci in range(C.size()):
 */
    goto __pyx_L7;
  }

  /* "cyext_acv.pyx":3643
 * 
 *     else:
 *         for i in range(m):             # <<<<<<<<<<<<<<
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])
 */
  /*else*/ {
    __pyx_t_11 = __pyx_v_m;
    __pyx_t_12 = __pyx_t_11;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3644
 *     else:
 *         for i in range(m):
 *             buff[0][0] = i             # <<<<<<<<<<<<<<
 *             va_id.push_back(buff[0])
 * 
 */
      ((__pyx_v_buff[0])[0]) = __pyx_v_i;

      /* "cyext_acv.pyx":3645
 *         for i in range(m):
 *             buff[0][0] = i
 *             va_id.push_back(buff[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef double p_s, p_si, coef, coef_0
 */
      try {
        __pyx_v_va_id.push_back((__pyx_v_buff[0]));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 3645, __pyx_L1_error)
      }
    }
  }
  __pyx_L7:;

  /* "cyext_acv.pyx":3649
 *     cdef double p_s, p_si, coef, coef_0
 *     cdef double[:, :, :] lm_s, lm_si
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_7);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_8 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_20 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_20.memview)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_lm_s = __pyx_t_20;
  __pyx_t_20.memview = NULL;
  __pyx_t_20.data = NULL;

  /* "cyext_acv.pyx":3650
 *     cdef double[:, :, :] lm_s, lm_si
 *     lm_s = np.zeros((va_id.size(), max_leaves, 2**scaling))
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_va_id.size()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyInt_From_long(__Pyx_pow_long(2, ((long)__pyx_v_scaling))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_20 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_20.memview)) __PYX_ERR(0, 3650, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_lm_si = __pyx_t_20;
  __pyx_t_20.memview = NULL;
  __pyx_t_20.data = NULL;

  /* "cyext_acv.pyx":3652
 *     lm_si = np.zeros((va_id.size(), max_leaves, 2**scaling))
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 */
  __pyx_t_11 = __pyx_v_n_trees;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_b = __pyx_t_13;

    /* "cyext_acv.pyx":3653
 * 
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *             for counter in range(phi_b.shape[1]):
 *                 for j in range(m):
 */
    __pyx_t_21 = (__pyx_v_phi_b.shape[0]);
    __pyx_t_22 = __pyx_t_21;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_22; __pyx_t_19+=1) {
      __pyx_v_leaf_numb = __pyx_t_19;

      /* "cyext_acv.pyx":3654
 *     for b in range(n_trees):
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                 for j in range(m):
 *                     for i2 in range(d):
 */
      __pyx_t_23 = (__pyx_v_phi_b.shape[1]);
      __pyx_t_24 = __pyx_t_23;
      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_24; __pyx_t_25+=1) {
        __pyx_v_counter = __pyx_t_25;

        /* "cyext_acv.pyx":3655
 *         for leaf_numb in range(phi_b.shape[0]):
 *             for counter in range(phi_b.shape[1]):
 *                 for j in range(m):             # <<<<<<<<<<<<<<
 *                     for i2 in range(d):
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 */
        __pyx_t_26 = __pyx_v_m;
        __pyx_t_27 = __pyx_t_26;
        for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
          __pyx_v_j = __pyx_t_28;

          /* "cyext_acv.pyx":3656
 *             for counter in range(phi_b.shape[1]):
 *                 for j in range(m):
 *                     for i2 in range(d):             # <<<<<<<<<<<<<<
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 */
          __pyx_t_29 = __pyx_v_d;
          __pyx_t_30 = __pyx_t_29;
          for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_30; __pyx_t_31+=1) {
            __pyx_v_i2 = __pyx_t_31;

            /* "cyext_acv.pyx":3657
 *                 for j in range(m):
 *                     for i2 in range(d):
 *                         phi_b[leaf_numb, counter, j, i2] = 0             # <<<<<<<<<<<<<<
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
            __pyx_t_32 = __pyx_v_leaf_numb;
            __pyx_t_33 = __pyx_v_counter;
            __pyx_t_34 = __pyx_v_j;
            __pyx_t_35 = __pyx_v_i2;
            *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_32 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_33 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_34 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_35 * __pyx_v_phi_b.strides[3]) )) = 0.0;
          }
        }
      }
    }

    /* "cyext_acv.pyx":3658
 *                     for i2 in range(d):
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 */
    __pyx_t_35 = __pyx_v_b;
    __pyx_v_nb_leaf = (*((long const  *) ( /* dim=0 */ ((char *) (((long const  *) __pyx_v_leaves_nb.data) + __pyx_t_35)) )));

    /* "cyext_acv.pyx":3659
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          __pyx_t_19 = __pyx_v_nb_leaf;
          if ((1 == 0)) abort();
          {
              int __pyx_parallel_temp0 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp1 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp2 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp3 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp4 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp5 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp6 = ((double)__PYX_NAN());
              double __pyx_parallel_temp7 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp8 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp9 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp10 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp11 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp12 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp13 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp14 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp15 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp16 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp17 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp18 = ((int)0xbad0bad0);
              unsigned int __pyx_parallel_temp19 = ((unsigned int)0xbad0bad0);
              double __pyx_parallel_temp20 = ((double)__PYX_NAN());
              double __pyx_parallel_temp21 = ((double)__PYX_NAN());
              unsigned int __pyx_parallel_temp22 = ((unsigned int)0xbad0bad0);
              unsigned int __pyx_parallel_temp23 = ((unsigned int)0xbad0bad0);
              long __pyx_parallel_temp24 = ((long)0xbad0bad0);
              unsigned int __pyx_parallel_temp25 = ((unsigned int)0xbad0bad0);
              int __pyx_parallel_temp26 = ((int)0xbad0bad0);
              const char *__pyx_parallel_filename = NULL; int __pyx_parallel_lineno = 0, __pyx_parallel_clineno = 0;
              PyObject *__pyx_parallel_exc_type = NULL, *__pyx_parallel_exc_value = NULL, *__pyx_parallel_exc_tb = NULL;
              int __pyx_parallel_why;
              __pyx_parallel_why = 0;
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_26 = (__pyx_t_19 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_26 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_t_38, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53, __pyx_t_54, __pyx_t_55, __pyx_t_56, __pyx_t_57, __pyx_t_58, __pyx_t_59, __pyx_t_60, __pyx_t_61, __pyx_t_62, __pyx_t_63, __pyx_t_64, __pyx_t_65, __pyx_t_66, __pyx_t_67, __pyx_t_68, __pyx_t_69, __pyx_t_70) private(__pyx_filename, __pyx_lineno, __pyx_clineno) shared(__pyx_parallel_why, __pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      Py_BEGIN_ALLOW_THREADS
                      #endif /* _OPENMP */
                      #ifdef _OPENMP
                      #pragma omp for lastprivate(__pyx_v_S_size) lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_add) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_ci) lastprivate(__pyx_v_cj) lastprivate(__pyx_v_coef) lastprivate(__pyx_v_coef_0) lastprivate(__pyx_v_counter) lastprivate(__pyx_v_cs) lastprivate(__pyx_v_csi) lastprivate(__pyx_v_i) lastprivate(__pyx_v_i2) lastprivate(__pyx_v_l) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_ns) lastprivate(__pyx_v_nv) lastprivate(__pyx_v_nv_bool) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_p_s) lastprivate(__pyx_v_p_si) lastprivate(__pyx_v_pow_set_size) lastprivate(__pyx_v_s) lastprivate(__pyx_v_set_size) lastprivate(__pyx_v_va) lastprivate(__pyx_v_va_size) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_25 = 0; __pyx_t_25 < __pyx_t_26; __pyx_t_25++){
                          if (__pyx_parallel_why < 2)
                          {
                              __pyx_v_leaf_numb = (unsigned int)(0 + 1 * __pyx_t_25);
                              /* Initialize private variables to invalid values */
                              __pyx_v_S_size = ((int)0xbad0bad0);
                              __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                              __pyx_v_add = ((unsigned int)0xbad0bad0);
                              __pyx_v_b_it = ((int)0xbad0bad0);
                              __pyx_v_ci = ((unsigned int)0xbad0bad0);
                              __pyx_v_cj = ((unsigned int)0xbad0bad0);
                              __pyx_v_coef = ((double)__PYX_NAN());
                              __pyx_v_coef_0 = ((double)__PYX_NAN());
                              __pyx_v_counter = ((unsigned int)0xbad0bad0);
                              __pyx_v_cs = ((unsigned int)0xbad0bad0);
                              __pyx_v_csi = ((unsigned int)0xbad0bad0);
                              __pyx_v_i = ((unsigned int)0xbad0bad0);
                              __pyx_v_i2 = ((unsigned int)0xbad0bad0);
                              __pyx_v_l = ((unsigned int)0xbad0bad0);
                              __pyx_v_lm = ((double)__PYX_NAN());
                              __pyx_v_ns = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv = ((unsigned int)0xbad0bad0);
                              __pyx_v_nv_bool = ((int)0xbad0bad0);
                              __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                              __pyx_v_p_s = ((double)__PYX_NAN());
                              __pyx_v_p_si = ((double)__PYX_NAN());
                              __pyx_v_pow_set_size = ((unsigned int)0xbad0bad0);
                              __pyx_v_s = ((unsigned int)0xbad0bad0);
                              __pyx_v_set_size = ((long)0xbad0bad0);
                              __pyx_v_va = ((unsigned int)0xbad0bad0);
                              __pyx_v_va_size = ((int)0xbad0bad0);

                              /* "cyext_acv.pyx":3660
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *             lm = 0
 *             for i in range(data.shape[0]):
 */
                              (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).clear();

                              /* "cyext_acv.pyx":3661
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0             # <<<<<<<<<<<<<<
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 */
                              __pyx_v_lm = 0.0;

                              /* "cyext_acv.pyx":3662
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 *             for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 */
                              __pyx_t_21 = (__pyx_v_data.shape[0]);
                              __pyx_t_22 = __pyx_t_21;
                              for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_22; __pyx_t_27+=1) {
                                __pyx_v_i = __pyx_t_27;

                                /* "cyext_acv.pyx":3663
 *             lm = 0
 *             for i in range(data.shape[0]):
 *                 a_it = 0             # <<<<<<<<<<<<<<
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 */
                                __pyx_v_a_it = 0;

                                /* "cyext_acv.pyx":3664
 *             for i in range(data.shape[0]):
 *                 a_it = 0
 *                 for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 */
                                __pyx_t_23 = (__pyx_v_data.shape[1]);
                                __pyx_t_24 = __pyx_t_23;
                                for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_24; __pyx_t_28+=1) {
                                  __pyx_v_s = __pyx_t_28;

                                  /* "cyext_acv.pyx":3665
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  __pyx_t_35 = __pyx_v_i;
                                  __pyx_t_34 = __pyx_v_s;
                                  __pyx_t_33 = __pyx_v_b;
                                  __pyx_t_32 = __pyx_v_leaf_numb;
                                  __pyx_t_36 = __pyx_v_s;
                                  __pyx_t_37 = 1;
                                  __pyx_t_38 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_35 * __pyx_v_data.strides[0]) ) + __pyx_t_34 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_32 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_36 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  if (__pyx_t_38) {
                                  } else {
                                    __pyx_t_14 = __pyx_t_38;
                                    goto __pyx_L44_bool_binop_done;
                                  }
                                  __pyx_t_36 = __pyx_v_i;
                                  __pyx_t_32 = __pyx_v_s;
                                  __pyx_t_33 = __pyx_v_b;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_35 = __pyx_v_s;
                                  __pyx_t_37 = 0;
                                  __pyx_t_38 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_32 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_33 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                  __pyx_t_14 = __pyx_t_38;
                                  __pyx_L44_bool_binop_done:;
                                  if (__pyx_t_14) {

                                    /* "cyext_acv.pyx":3666
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1
 */
                                    __pyx_v_a_it = (__pyx_v_a_it + 1);

                                    /* "cyext_acv.pyx":3665
 *                 a_it = 0
 *                 for s in range(data.shape[1]):
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):             # <<<<<<<<<<<<<<
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 */
                                  }
                                }

                                /* "cyext_acv.pyx":3667
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                __pyx_t_14 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                if (__pyx_t_14) {

                                  /* "cyext_acv.pyx":3668
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:
 *                     lm = lm + 1             # <<<<<<<<<<<<<<
 * 
 *             if C[0].size() != 0:
 */
                                  __pyx_v_lm = (__pyx_v_lm + 1.0);

                                  /* "cyext_acv.pyx":3667
 *                     if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
 *                         a_it = a_it + 1
 *                 if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                     lm = lm + 1
 * 
 */
                                }
                              }

                              /* "cyext_acv.pyx":3670
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                              __pyx_t_14 = (((__pyx_v_C[0]).size() != 0) != 0);
                              if (__pyx_t_14) {

                                /* "cyext_acv.pyx":3671
 * 
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C             # <<<<<<<<<<<<<<
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 */
                                (__pyx_v_C_b[__pyx_v_leaf_numb]) = __pyx_v_C;

                                /* "cyext_acv.pyx":3672
 *             if C[0].size() != 0:
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 */
                                __pyx_t_17 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_18 = __pyx_t_17;
                                for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_18; __pyx_t_27+=1) {
                                  __pyx_v_nv = __pyx_t_27;

                                  /* "cyext_acv.pyx":3673
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0             # <<<<<<<<<<<<<<
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 */
                                  __pyx_v_add = 0;

                                  /* "cyext_acv.pyx":3674
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 *                     add = 0
 *                     for ns in range(remove_va.size()):             # <<<<<<<<<<<<<<
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 */
                                  __pyx_t_39 = __pyx_v_remove_va.size();
                                  __pyx_t_40 = __pyx_t_39;
                                  for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_40; __pyx_t_28+=1) {
                                    __pyx_v_ns = __pyx_t_28;

                                    /* "cyext_acv.pyx":3675
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    __pyx_t_14 = (((((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]) == (__pyx_v_remove_va[__pyx_v_ns])) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3676
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1             # <<<<<<<<<<<<<<
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 */
                                      __pyx_v_add = 1;

                                      /* "cyext_acv.pyx":3677
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]             # <<<<<<<<<<<<<<
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 */
                                      ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv]);

                                      /* "cyext_acv.pyx":3678
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 *                             break
 *                     if add == 0:
 */
                                      try {
                                        (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                      } catch(...) {
                                        #ifdef WITH_THREAD
                                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                        #endif
                                        __Pyx_CppExn2PyErr();
                                        #ifdef WITH_THREAD
                                        __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                        #endif
                                        __PYX_ERR(0, 3678, __pyx_L37_error)
                                      }

                                      /* "cyext_acv.pyx":3679
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break             # <<<<<<<<<<<<<<
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 */
                                      goto __pyx_L51_break;

                                      /* "cyext_acv.pyx":3675
 *                     add = 0
 *                     for ns in range(remove_va.size()):
 *                         if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:             # <<<<<<<<<<<<<<
 *                             add = 1
 *                             buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
 */
                                    }
                                  }
                                  __pyx_L51_break:;

                                  /* "cyext_acv.pyx":3680
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  __pyx_t_14 = ((__pyx_v_add == 0) != 0);
                                  if (__pyx_t_14) {

                                    /* "cyext_acv.pyx":3681
 *                             break
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 */
                                    __pyx_t_15 = (__pyx_v_C_b[__pyx_v_leaf_numb]).size();
                                    __pyx_t_16 = __pyx_t_15;
                                    for (__pyx_t_28 = 0; __pyx_t_28 < __pyx_t_16; __pyx_t_28+=1) {
                                      __pyx_v_ci = __pyx_t_28;

                                      /* "cyext_acv.pyx":3682
 *                     if add == 0:
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 */
                                      __pyx_t_39 = ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_40 = __pyx_t_39;
                                      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_40; __pyx_t_29+=1) {
                                        __pyx_v_cj = __pyx_t_29;

                                        /* "cyext_acv.pyx":3683
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        __pyx_t_14 = (((((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]) == (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_nv])) != 0);
                                        if (__pyx_t_14) {

                                          /* "cyext_acv.pyx":3684
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1             # <<<<<<<<<<<<<<
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 */
                                          __pyx_v_add = 1;

                                          /* "cyext_acv.pyx":3685
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                     break
 *                             if add == 1:
 */
                                          try {
                                            (__pyx_v_node_id_v2[__pyx_v_leaf_numb]).push_back(((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci]));
                                          } catch(...) {
                                            #ifdef WITH_THREAD
                                            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                            #endif
                                            __Pyx_CppExn2PyErr();
                                            #ifdef WITH_THREAD
                                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                            #endif
                                            __PYX_ERR(0, 3685, __pyx_L37_error)
                                          }

                                          /* "cyext_acv.pyx":3686
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break             # <<<<<<<<<<<<<<
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 */
                                          goto __pyx_L57_break;

                                          /* "cyext_acv.pyx":3683
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 *                                 if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:             # <<<<<<<<<<<<<<
 *                                     add = 1
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 */
                                        }
                                      }
                                      __pyx_L57_break:;

                                      /* "cyext_acv.pyx":3687
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      __pyx_t_14 = ((__pyx_v_add == 1) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3688
 *                                     break
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])             # <<<<<<<<<<<<<<
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 */
                                        (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >((__pyx_v_C_b[__pyx_v_leaf_numb]).begin(), (__pyx_v_C_b[__pyx_v_leaf_numb]).end(), ((__pyx_v_C_b[__pyx_v_leaf_numb])[__pyx_v_ci])));

                                        /* "cyext_acv.pyx":3689
 *                             if add == 1:
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 */
                                        (__pyx_v_C_b[__pyx_v_leaf_numb]).pop_back();

                                        /* "cyext_acv.pyx":3690
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 *                                 break             # <<<<<<<<<<<<<<
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 */
                                        goto __pyx_L55_break;

                                        /* "cyext_acv.pyx":3687
 *                                     node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
 *                                     break
 *                             if add == 1:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
 *                                 C_b[leaf_numb].pop_back()
 */
                                      }
                                    }
                                    __pyx_L55_break:;

                                    /* "cyext_acv.pyx":3680
 *                             node_id_v2[leaf_numb].push_back(buff[leaf_numb])
 *                             break
 *                     if add == 0:             # <<<<<<<<<<<<<<
 *                         for ci in range(C_b[leaf_numb].size()):
 *                             for cj in range(C_b[leaf_numb][ci].size()):
 */
                                  }
                                }

                                /* "cyext_acv.pyx":3691
 *                                 C_b[leaf_numb].pop_back()
 *                                 break
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]             # <<<<<<<<<<<<<<
 *             else:
 *                 node_id[leaf_numb].clear()
 */
                                (__pyx_v_node_id[__pyx_v_leaf_numb]) = (__pyx_v_node_id_v2[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":3670
 *                     lm = lm + 1
 * 
 *             if C[0].size() != 0:             # <<<<<<<<<<<<<<
 *                 C_b[leaf_numb] = C
 *                 for nv in range(node_idx_trees[b][leaf_numb].size()):
 */
                                goto __pyx_L47;
                              }

                              /* "cyext_acv.pyx":3693
 *                 node_id[leaf_numb] = node_id_v2[leaf_numb]
 *             else:
 *                 node_id[leaf_numb].clear()             # <<<<<<<<<<<<<<
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 */
                              /*else*/ {
                                (__pyx_v_node_id[__pyx_v_leaf_numb]).clear();

                                /* "cyext_acv.pyx":3694
 *             else:
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):             # <<<<<<<<<<<<<<
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 */
                                __pyx_t_17 = ((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb]).size();
                                __pyx_t_18 = __pyx_t_17;
                                for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_18; __pyx_t_27+=1) {
                                  __pyx_v_i = __pyx_t_27;

                                  /* "cyext_acv.pyx":3695
 *                 node_id[leaf_numb].clear()
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]             # <<<<<<<<<<<<<<
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])
 * 
 */
                                  ((__pyx_v_buff[__pyx_v_leaf_numb])[0]) = (((__pyx_v_node_idx_trees[__pyx_v_b])[__pyx_v_leaf_numb])[__pyx_v_i]);

                                  /* "cyext_acv.pyx":3696
 *                 for i in range(node_idx_trees[b][leaf_numb].size()):
 *                     buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
 *                     node_id[leaf_numb].push_back(buff[leaf_numb])             # <<<<<<<<<<<<<<
 * 
 * 
 */
                                  try {
                                    (__pyx_v_node_id[__pyx_v_leaf_numb]).push_back((__pyx_v_buff[__pyx_v_leaf_numb]));
                                  } catch(...) {
                                    #ifdef WITH_THREAD
                                    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                    #endif
                                    __Pyx_CppExn2PyErr();
                                    #ifdef WITH_THREAD
                                    __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                    #endif
                                    __PYX_ERR(0, 3696, __pyx_L37_error)
                                  }
                                }
                              }
                              __pyx_L47:;

                              /* "cyext_acv.pyx":3699
 * 
 * 
 *             for va in range(va_id.size()):             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 */
                              __pyx_t_15 = __pyx_v_va_id.size();
                              __pyx_t_16 = __pyx_t_15;
                              for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_16; __pyx_t_27+=1) {
                                __pyx_v_va = __pyx_t_27;

                                /* "cyext_acv.pyx":3700
 * 
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]             # <<<<<<<<<<<<<<
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]) = (__pyx_v_node_id[__pyx_v_leaf_numb]);

                                /* "cyext_acv.pyx":3701
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                __pyx_t_14 = ((!((std::find<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])) != ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end()) != 0)) != 0);
                                if (__pyx_t_14) {

                                  /* "cyext_acv.pyx":3702
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 */
                                  goto __pyx_L62_continue;

                                  /* "cyext_acv.pyx":3701
 *             for va in range(va_id.size()):
 *                 node_id_va[va][leaf_numb] = node_id[leaf_numb]
 *                 if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
                                }

                                /* "cyext_acv.pyx":3704
 *                     continue
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])             # <<<<<<<<<<<<<<
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 */
                                (void)(std::remove<std::vector<std::vector<long> > ::iterator,std::vector<long> >(((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).begin(), ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).end(), (__pyx_v_va_id[__pyx_v_va])));

                                /* "cyext_acv.pyx":3705
 * 
 *                 std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
 *                 node_id_va[va][leaf_numb].pop_back()             # <<<<<<<<<<<<<<
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 */
                                ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).pop_back();

                                /* "cyext_acv.pyx":3707
 *                 node_id_va[va][leaf_numb].pop_back()
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()             # <<<<<<<<<<<<<<
 *                 pow_set_size = 2**set_size
 * 
 */
                                __pyx_v_set_size = ((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb]).size();

                                /* "cyext_acv.pyx":3708
 * 
 *                 set_size = node_id_va[va][leaf_numb].size()
 *                 pow_set_size = 2**set_size             # <<<<<<<<<<<<<<
 * 
 *                 for counter in range(0, pow_set_size):
 */
                                __pyx_v_pow_set_size = __Pyx_pow_long(2, __pyx_v_set_size);

                                /* "cyext_acv.pyx":3710
 *                 pow_set_size = 2**set_size
 * 
 *                 for counter in range(0, pow_set_size):             # <<<<<<<<<<<<<<
 *                     va_size = 0
 *                     S_size = 0
 */
                                __pyx_t_28 = __pyx_v_pow_set_size;
                                __pyx_t_29 = __pyx_t_28;
                                for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_29; __pyx_t_30+=1) {
                                  __pyx_v_counter = __pyx_t_30;

                                  /* "cyext_acv.pyx":3711
 * 
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0             # <<<<<<<<<<<<<<
 *                     S_size = 0
 *                     for ci in range(set_size):
 */
                                  __pyx_v_va_size = 0;

                                  /* "cyext_acv.pyx":3712
 *                 for counter in range(0, pow_set_size):
 *                     va_size = 0
 *                     S_size = 0             # <<<<<<<<<<<<<<
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 */
                                  __pyx_v_S_size = 0;

                                  /* "cyext_acv.pyx":3713
 *                     va_size = 0
 *                     S_size = 0
 *                     for ci in range(set_size):             # <<<<<<<<<<<<<<
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 */
                                  __pyx_t_41 = __pyx_v_set_size;
                                  __pyx_t_42 = __pyx_t_41;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_42; __pyx_t_31+=1) {
                                    __pyx_v_ci = __pyx_t_31;

                                    /* "cyext_acv.pyx":3714
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    __pyx_t_14 = (((__pyx_v_counter & (1 << __pyx_v_ci)) > 0) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3715
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):             # <<<<<<<<<<<<<<
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 */
                                      __pyx_t_17 = (((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci]).size();
                                      __pyx_t_18 = __pyx_t_17;
                                      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_18; __pyx_t_43+=1) {
                                        __pyx_v_cj = __pyx_t_43;

                                        /* "cyext_acv.pyx":3716
 *                         if((counter & (1 << ci)) > 0):
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]             # <<<<<<<<<<<<<<
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1
 */
                                        __pyx_t_35 = __pyx_v_va;
                                        __pyx_t_34 = __pyx_v_counter;
                                        __pyx_t_33 = __pyx_v_leaf_numb;
                                        __pyx_t_37 = __pyx_v_S_size;
                                        *((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_35 * __pyx_v_S.strides[0]) ) + __pyx_t_34 * __pyx_v_S.strides[1]) ) + __pyx_t_33 * __pyx_v_S.strides[2]) )) + __pyx_t_37)) )) = ((((__pyx_v_node_id_va[__pyx_v_va])[__pyx_v_leaf_numb])[__pyx_v_ci])[__pyx_v_cj]);

                                        /* "cyext_acv.pyx":3717
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1             # <<<<<<<<<<<<<<
 *                             va_size = va_size + 1
 * 
 */
                                        __pyx_v_S_size = (__pyx_v_S_size + 1);
                                      }

                                      /* "cyext_acv.pyx":3718
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 *                                 S_size = S_size + 1
 *                             va_size = va_size + 1             # <<<<<<<<<<<<<<
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 */
                                      __pyx_v_va_size = (__pyx_v_va_size + 1);

                                      /* "cyext_acv.pyx":3714
 *                     S_size = 0
 *                     for ci in range(set_size):
 *                         if((counter & (1 << ci)) > 0):             # <<<<<<<<<<<<<<
 *                             for cj in range(node_id_va[va][leaf_numb][ci].size()):
 *                                 S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3720
 *                             va_size = va_size + 1
 * 
 *                     lm_s[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):
 */
                                  __pyx_t_33 = __pyx_v_va;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_35 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_33 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_35 * __pyx_v_lm_s.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":3721
 * 
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0             # <<<<<<<<<<<<<<
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 */
                                  __pyx_t_35 = __pyx_v_va;
                                  __pyx_t_34 = __pyx_v_leaf_numb;
                                  __pyx_t_33 = __pyx_v_counter;
                                  *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_35 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_33 * __pyx_v_lm_si.strides[2]) )) = 0.0;

                                  /* "cyext_acv.pyx":3722
 *                     lm_s[va, leaf_numb, counter] = 0
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                         b_it = 0
 *                         for s in range(S_size):
 */
                                  __pyx_t_21 = (__pyx_v_data.shape[0]);
                                  __pyx_t_22 = __pyx_t_21;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_22; __pyx_t_31+=1) {
                                    __pyx_v_i = __pyx_t_31;

                                    /* "cyext_acv.pyx":3723
 *                     lm_si[va, leaf_numb, counter] = 0
 *                     for i in range(data.shape[0]):
 *                         b_it = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                    __pyx_v_b_it = 0;

                                    /* "cyext_acv.pyx":3724
 *                     for i in range(data.shape[0]):
 *                         b_it = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1
 */
                                    __pyx_t_44 = __pyx_v_S_size;
                                    __pyx_t_45 = __pyx_t_44;
                                    for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_45; __pyx_t_43+=1) {
                                      __pyx_v_s = __pyx_t_43;

                                      /* "cyext_acv.pyx":3725
 *                         b_it = 0
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      __pyx_t_33 = __pyx_v_va;
                                      __pyx_t_34 = __pyx_v_counter;
                                      __pyx_t_35 = __pyx_v_leaf_numb;
                                      __pyx_t_32 = __pyx_v_s;
                                      __pyx_t_36 = __pyx_v_i;
                                      __pyx_t_37 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_33 * __pyx_v_S.strides[0]) ) + __pyx_t_34 * __pyx_v_S.strides[1]) ) + __pyx_t_35 * __pyx_v_S.strides[2]) )) + __pyx_t_32)) )));
                                      __pyx_t_46 = __pyx_v_va;
                                      __pyx_t_47 = __pyx_v_counter;
                                      __pyx_t_48 = __pyx_v_leaf_numb;
                                      __pyx_t_49 = __pyx_v_s;
                                      __pyx_t_50 = __pyx_v_b;
                                      __pyx_t_51 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) ) + __pyx_t_47 * __pyx_v_S.strides[1]) ) + __pyx_t_48 * __pyx_v_S.strides[2]) )) + __pyx_t_49)) )));
                                      __pyx_t_53 = 1;
                                      __pyx_t_54 = __pyx_v_va;
                                      __pyx_t_55 = __pyx_v_counter;
                                      __pyx_t_56 = __pyx_v_leaf_numb;
                                      __pyx_t_57 = __pyx_v_s;
                                      __pyx_t_58 = __pyx_v_i;
                                      __pyx_t_59 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_54 * __pyx_v_S.strides[0]) ) + __pyx_t_55 * __pyx_v_S.strides[1]) ) + __pyx_t_56 * __pyx_v_S.strides[2]) )) + __pyx_t_57)) )));
                                      __pyx_t_60 = __pyx_v_va;
                                      __pyx_t_61 = __pyx_v_counter;
                                      __pyx_t_62 = __pyx_v_leaf_numb;
                                      __pyx_t_63 = __pyx_v_s;
                                      __pyx_t_64 = __pyx_v_b;
                                      __pyx_t_65 = __pyx_v_leaf_numb;
                                      __pyx_t_66 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_60 * __pyx_v_S.strides[0]) ) + __pyx_t_61 * __pyx_v_S.strides[1]) ) + __pyx_t_62 * __pyx_v_S.strides[2]) )) + __pyx_t_63)) )));
                                      __pyx_t_67 = 0;
                                      __pyx_t_14 = ((((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_36 * __pyx_v_data.strides[0]) ) + __pyx_t_37 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_58 * __pyx_v_data.strides[0]) ) + __pyx_t_59 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_64 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_65 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_67 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3726
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                                 b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                         if b_it == S_size:
 */
                                        __pyx_v_b_it = (__pyx_v_b_it + 1);

                                        /* "cyext_acv.pyx":3725
 *                         b_it = 0
 *                         for s in range(S_size):
 *                             if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                                 b_it = b_it + 1
 * 
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":3728
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 */
                                    __pyx_t_14 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3729
 * 
 *                         if b_it == S_size:
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                             nv_bool = 0
 */
                                      __pyx_t_63 = __pyx_v_va;
                                      __pyx_t_62 = __pyx_v_leaf_numb;
                                      __pyx_t_61 = __pyx_v_counter;
                                      __pyx_t_60 = __pyx_v_va;
                                      __pyx_t_65 = __pyx_v_leaf_numb;
                                      __pyx_t_64 = __pyx_v_counter;
                                      *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_60 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_65 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_64 * __pyx_v_lm_s.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_63 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_62 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_61 * __pyx_v_lm_s.strides[2]) ))) + 1.0);

                                      /* "cyext_acv.pyx":3731
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 *                             nv_bool = 0             # <<<<<<<<<<<<<<
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                      __pyx_v_nv_bool = 0;

                                      /* "cyext_acv.pyx":3732
 * 
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 */
                                      __pyx_t_17 = (__pyx_v_va_id[__pyx_v_va]).size();
                                      __pyx_t_18 = __pyx_t_17;
                                      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_18; __pyx_t_43+=1) {
                                        __pyx_v_nv = __pyx_t_43;

                                        /* "cyext_acv.pyx":3733
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        __pyx_t_61 = __pyx_v_i;
                                        __pyx_t_67 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_62 = __pyx_v_b;
                                        __pyx_t_63 = __pyx_v_leaf_numb;
                                        __pyx_t_66 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_59 = 1;
                                        __pyx_t_38 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_61 * __pyx_v_data.strides[0]) ) + __pyx_t_67 * __pyx_v_data.strides[1]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_62 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_63 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_59 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (!__pyx_t_38) {
                                        } else {
                                          __pyx_t_14 = __pyx_t_38;
                                          goto __pyx_L81_bool_binop_done;
                                        }
                                        __pyx_t_63 = __pyx_v_i;
                                        __pyx_t_59 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_62 = __pyx_v_b;
                                        __pyx_t_61 = __pyx_v_leaf_numb;
                                        __pyx_t_66 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                        __pyx_t_67 = 0;
                                        __pyx_t_38 = (((*((double const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_63 * __pyx_v_data.strides[0]) ) + __pyx_t_59 * __pyx_v_data.strides[1]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_62 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_61 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_67 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_14 = __pyx_t_38;
                                        __pyx_L81_bool_binop_done:;
                                        if (__pyx_t_14) {

                                          /* "cyext_acv.pyx":3734
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                     continue
 * 
 */
                                          __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                          /* "cyext_acv.pyx":3735
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                     nv_bool = nv_bool + 1
 *                                     continue             # <<<<<<<<<<<<<<
 * 
 *                             if nv_bool == 0:
 */
                                          goto __pyx_L78_continue;

                                          /* "cyext_acv.pyx":3733
 *                             nv_bool = 0
 *                             for nv in range(va_id[va].size()):
 *                                 if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                     nv_bool = nv_bool + 1
 *                                     continue
 */
                                        }
                                        __pyx_L78_continue:;
                                      }

                                      /* "cyext_acv.pyx":3737
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      __pyx_t_14 = ((__pyx_v_nv_bool == 0) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3738
 * 
 *                             if nv_bool == 0:
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1             # <<<<<<<<<<<<<<
 * 
 *                     csi = 0
 */
                                        __pyx_t_61 = __pyx_v_va;
                                        __pyx_t_62 = __pyx_v_leaf_numb;
                                        __pyx_t_63 = __pyx_v_counter;
                                        __pyx_t_64 = __pyx_v_va;
                                        __pyx_t_65 = __pyx_v_leaf_numb;
                                        __pyx_t_60 = __pyx_v_counter;
                                        *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_64 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_65 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_60 * __pyx_v_lm_si.strides[2]) )) = ((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_61 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_62 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_63 * __pyx_v_lm_si.strides[2]) ))) + 1.0);

                                        /* "cyext_acv.pyx":3737
 *                                     continue
 * 
 *                             if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3728
 *                                 b_it = b_it + 1
 * 
 *                         if b_it == S_size:             # <<<<<<<<<<<<<<
 *                             lm_s[va, leaf_numb, counter] = lm_s[va, leaf_numb, counter] + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3740
 *                                 lm_si[va, leaf_numb, counter] = lm_si[va, leaf_numb, counter] + 1
 * 
 *                     csi = 0             # <<<<<<<<<<<<<<
 *                     cs = 0
 * 
 */
                                  __pyx_v_csi = 0;

                                  /* "cyext_acv.pyx":3741
 * 
 *                     csi = 0
 *                     cs = 0             # <<<<<<<<<<<<<<
 * 
 *                     o_all = 0
 */
                                  __pyx_v_cs = 0;

                                  /* "cyext_acv.pyx":3743
 *                     cs = 0
 * 
 *                     o_all = 0             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 */
                                  __pyx_v_o_all = 0;

                                  /* "cyext_acv.pyx":3744
 * 
 *                     o_all = 0
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                             o_all = o_all + 1
 */
                                  __pyx_t_44 = __pyx_v_S_size;
                                  __pyx_t_45 = __pyx_t_44;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_45; __pyx_t_31+=1) {
                                    __pyx_v_s = __pyx_t_31;

                                    /* "cyext_acv.pyx":3745
 *                     o_all = 0
 *                     for s in range(S_size):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    __pyx_t_63 = __pyx_v_va;
                                    __pyx_t_62 = __pyx_v_counter;
                                    __pyx_t_61 = __pyx_v_leaf_numb;
                                    __pyx_t_60 = __pyx_v_s;
                                    __pyx_t_67 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_63 * __pyx_v_S.strides[0]) ) + __pyx_t_62 * __pyx_v_S.strides[1]) ) + __pyx_t_61 * __pyx_v_S.strides[2]) )) + __pyx_t_60)) )));
                                    __pyx_t_65 = __pyx_v_va;
                                    __pyx_t_64 = __pyx_v_counter;
                                    __pyx_t_57 = __pyx_v_leaf_numb;
                                    __pyx_t_56 = __pyx_v_s;
                                    __pyx_t_55 = __pyx_v_b;
                                    __pyx_t_54 = __pyx_v_leaf_numb;
                                    __pyx_t_66 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_65 * __pyx_v_S.strides[0]) ) + __pyx_t_64 * __pyx_v_S.strides[1]) ) + __pyx_t_57 * __pyx_v_S.strides[2]) )) + __pyx_t_56)) )));
                                    __pyx_t_59 = 1;
                                    __pyx_t_58 = __pyx_v_va;
                                    __pyx_t_49 = __pyx_v_counter;
                                    __pyx_t_48 = __pyx_v_leaf_numb;
                                    __pyx_t_47 = __pyx_v_s;
                                    __pyx_t_53 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_58 * __pyx_v_S.strides[0]) ) + __pyx_t_49 * __pyx_v_S.strides[1]) ) + __pyx_t_48 * __pyx_v_S.strides[2]) )) + __pyx_t_47)) )));
                                    __pyx_t_46 = __pyx_v_va;
                                    __pyx_t_51 = __pyx_v_counter;
                                    __pyx_t_50 = __pyx_v_leaf_numb;
                                    __pyx_t_32 = __pyx_v_s;
                                    __pyx_t_35 = __pyx_v_b;
                                    __pyx_t_34 = __pyx_v_leaf_numb;
                                    __pyx_t_52 = (*((long *) ( /* dim=3 */ ((char *) (((long *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_S.data + __pyx_t_46 * __pyx_v_S.strides[0]) ) + __pyx_t_51 * __pyx_v_S.strides[1]) ) + __pyx_t_50 * __pyx_v_S.strides[2]) )) + __pyx_t_32)) )));
                                    __pyx_t_37 = 0;
                                    __pyx_t_14 = ((((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_67 * __pyx_v_X.strides[0]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_55 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_54 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_66 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_59 * __pyx_v_partition_leaves_trees.strides[3]) )))) * ((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_53 * __pyx_v_X.strides[0]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_35 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[3]) ))))) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3746
 *                     for s in range(S_size):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
 *                             o_all = o_all + 1             # <<<<<<<<<<<<<<
 * 
 *                     if o_all == S_size:
 */
                                      __pyx_v_o_all = (__pyx_v_o_all + 1);

                                      /* "cyext_acv.pyx":3745
 *                     o_all = 0
 *                     for s in range(S_size):
 *                         if ((X[S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):             # <<<<<<<<<<<<<<
 *                             o_all = o_all + 1
 * 
 */
                                    }
                                  }

                                  /* "cyext_acv.pyx":3748
 *                             o_all = o_all + 1
 * 
 *                     if o_all == S_size:             # <<<<<<<<<<<<<<
 *                         cs = 1
 *                         nv_bool = 0
 */
                                  __pyx_t_14 = ((__pyx_v_o_all == __pyx_v_S_size) != 0);
                                  if (__pyx_t_14) {

                                    /* "cyext_acv.pyx":3749
 * 
 *                     if o_all == S_size:
 *                         cs = 1             # <<<<<<<<<<<<<<
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 */
                                    __pyx_v_cs = 1;

                                    /* "cyext_acv.pyx":3750
 *                     if o_all == S_size:
 *                         cs = 1
 *                         nv_bool = 0             # <<<<<<<<<<<<<<
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 */
                                    __pyx_v_nv_bool = 0;

                                    /* "cyext_acv.pyx":3751
 *                         cs = 1
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1
 */
                                    __pyx_t_17 = (__pyx_v_va_id[__pyx_v_va]).size();
                                    __pyx_t_18 = __pyx_t_17;
                                    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_18; __pyx_t_31+=1) {
                                      __pyx_v_nv = __pyx_t_31;

                                      /* "cyext_acv.pyx":3752
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      __pyx_t_37 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_32 = __pyx_v_b;
                                      __pyx_t_50 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_53 = 1;
                                      __pyx_t_38 = (((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_37 * __pyx_v_X.strides[0]) ))) > (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_32 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_53 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_38) {
                                      } else {
                                        __pyx_t_14 = __pyx_t_38;
                                        goto __pyx_L91_bool_binop_done;
                                      }
                                      __pyx_t_53 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_50 = __pyx_v_b;
                                      __pyx_t_32 = __pyx_v_leaf_numb;
                                      __pyx_t_52 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_37 = 0;
                                      __pyx_t_38 = (((*((double const  *) ( /* dim=0 */ (__pyx_v_X.data + __pyx_t_53 * __pyx_v_X.strides[0]) ))) <= (*((double const  *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_32 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_52 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_37 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_14 = __pyx_t_38;
                                      __pyx_L91_bool_binop_done:;
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3753
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1             # <<<<<<<<<<<<<<
 *                                 continue
 * 
 */
                                        __pyx_v_nv_bool = (__pyx_v_nv_bool + 1);

                                        /* "cyext_acv.pyx":3754
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
 *                                 nv_bool = nv_bool + 1
 *                                 continue             # <<<<<<<<<<<<<<
 * 
 *                         if nv_bool == 0:
 */
                                        goto __pyx_L88_continue;

                                        /* "cyext_acv.pyx":3752
 *                         nv_bool = 0
 *                         for nv in range(va_id[va].size()):
 *                             if ((X[va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):             # <<<<<<<<<<<<<<
 *                                 nv_bool = nv_bool + 1
 *                                 continue
 */
                                      }
                                      __pyx_L88_continue:;
                                    }

                                    /* "cyext_acv.pyx":3756
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    __pyx_t_14 = ((__pyx_v_nv_bool == 0) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3757
 * 
 *                         if nv_bool == 0:
 *                             csi = 1             # <<<<<<<<<<<<<<
 * 
 *                     coef = 0
 */
                                      __pyx_v_csi = 1;

                                      /* "cyext_acv.pyx":3756
 *                                 continue
 * 
 *                         if nv_bool == 0:             # <<<<<<<<<<<<<<
 *                             csi = 1
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":3748
 *                             o_all = o_all + 1
 * 
 *                     if o_all == S_size:             # <<<<<<<<<<<<<<
 *                         cs = 1
 *                         nv_bool = 0
 */
                                  }

                                  /* "cyext_acv.pyx":3759
 *                             csi = 1
 * 
 *                     coef = 0             # <<<<<<<<<<<<<<
 *                     for l in range(1, m - set_size):
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 */
                                  __pyx_v_coef = 0.0;

                                  /* "cyext_acv.pyx":3760
 * 
 *                     coef = 0
 *                     for l in range(1, m - set_size):             # <<<<<<<<<<<<<<
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 */
                                  __pyx_t_41 = (__pyx_v_m - __pyx_v_set_size);
                                  __pyx_t_42 = __pyx_t_41;
                                  for (__pyx_t_31 = 1; __pyx_t_31 < __pyx_t_42; __pyx_t_31+=1) {
                                    __pyx_v_l = __pyx_t_31;

                                    /* "cyext_acv.pyx":3761
 *                     coef = 0
 *                     for l in range(1, m - set_size):
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 */
                                    if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0) != 0.0) != 0)) {
                                      __pyx_t_68 = (__pyx_v_coef + ((1. * __pyx_f_9cyext_acv_binomialC(((__pyx_v_m - __pyx_v_set_size) - 1), __pyx_v_l, 0)) / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), (__pyx_v_l + __pyx_v_va_size), 0)));
                                    } else {
                                      __pyx_t_68 = 0.0;
                                    }
                                    __pyx_v_coef = __pyx_t_68;
                                  }

                                  /* "cyext_acv.pyx":3763
 *                         coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0
 * 
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                     if S_size == 0:
 */
                                  if (((__pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0) != 0.0) != 0)) {
                                    __pyx_t_68 = (1. / __pyx_f_9cyext_acv_binomialC((__pyx_v_m - 1), __pyx_v_va_size, 0));
                                  } else {
                                    __pyx_t_68 = 0.0;
                                  }
                                  __pyx_v_coef_0 = __pyx_t_68;

                                  /* "cyext_acv.pyx":3765
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 * 
 *                     if S_size == 0:             # <<<<<<<<<<<<<<
 *                         p_s = lm/data.shape[0]
 *                     else:
 */
                                  __pyx_t_14 = ((__pyx_v_S_size == 0) != 0);
                                  if (__pyx_t_14) {

                                    /* "cyext_acv.pyx":3766
 * 
 *                     if S_size == 0:
 *                         p_s = lm/data.shape[0]             # <<<<<<<<<<<<<<
 *                     else:
 *                         p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0
 */
                                    __pyx_v_p_s = (__pyx_v_lm / (__pyx_v_data.shape[0]));

                                    /* "cyext_acv.pyx":3765
 *                     coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0
 * 
 *                     if S_size == 0:             # <<<<<<<<<<<<<<
 *                         p_s = lm/data.shape[0]
 *                     else:
 */
                                    goto __pyx_L96;
                                  }

                                  /* "cyext_acv.pyx":3768
 *                         p_s = lm/data.shape[0]
 *                     else:
 *                         p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0
 * 
 */
                                  /*else*/ {
                                    __pyx_t_32 = __pyx_v_va;
                                    __pyx_t_50 = __pyx_v_leaf_numb;
                                    __pyx_t_51 = __pyx_v_counter;
                                    if ((((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_32 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_50 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_51 * __pyx_v_lm_s.strides[2]) ))) != 0.0) != 0)) {
                                      __pyx_t_46 = __pyx_v_va;
                                      __pyx_t_34 = __pyx_v_leaf_numb;
                                      __pyx_t_35 = __pyx_v_counter;
                                      __pyx_t_68 = ((__pyx_v_cs * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_s.data + __pyx_t_46 * __pyx_v_lm_s.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_s.strides[1]) ) + __pyx_t_35 * __pyx_v_lm_s.strides[2]) ))));
                                    } else {
                                      __pyx_t_68 = 0.0;
                                    }
                                    __pyx_v_p_s = __pyx_t_68;
                                  }
                                  __pyx_L96:;

                                  /* "cyext_acv.pyx":3769
 *                     else:
 *                         p_s = (cs * lm)/lm_s[va, leaf_numb, counter] if lm_s[va, leaf_numb, counter] !=0 else 0
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                     for nv in range(va_id[va].size()):
 */
                                  __pyx_t_51 = __pyx_v_va;
                                  __pyx_t_50 = __pyx_v_leaf_numb;
                                  __pyx_t_32 = __pyx_v_counter;
                                  if ((((*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_51 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_50 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_32 * __pyx_v_lm_si.strides[2]) ))) != 0.0) != 0)) {
                                    __pyx_t_35 = __pyx_v_va;
                                    __pyx_t_34 = __pyx_v_leaf_numb;
                                    __pyx_t_46 = __pyx_v_counter;
                                    __pyx_t_68 = ((__pyx_v_csi * __pyx_v_lm) / (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_lm_si.data + __pyx_t_35 * __pyx_v_lm_si.strides[0]) ) + __pyx_t_34 * __pyx_v_lm_si.strides[1]) ) + __pyx_t_46 * __pyx_v_lm_si.strides[2]) ))));
                                  } else {
                                    __pyx_t_68 = 0.0;
                                  }
                                  __pyx_v_p_si = __pyx_t_68;

                                  /* "cyext_acv.pyx":3771
 *                     p_si = (csi * lm)/lm_si[va, leaf_numb, counter] if lm_si[va, leaf_numb, counter] !=0 else 0
 * 
 *                     for nv in range(va_id[va].size()):             # <<<<<<<<<<<<<<
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 */
                                  __pyx_t_17 = (__pyx_v_va_id[__pyx_v_va]).size();
                                  __pyx_t_18 = __pyx_t_17;
                                  for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_18; __pyx_t_31+=1) {
                                    __pyx_v_nv = __pyx_t_31;

                                    /* "cyext_acv.pyx":3772
 * 
 *                     for nv in range(va_id[va].size()):
 *                         for i2 in range(d):             # <<<<<<<<<<<<<<
 *                             phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 */
                                    __pyx_t_43 = __pyx_v_d;
                                    __pyx_t_69 = __pyx_t_43;
                                    for (__pyx_t_70 = 0; __pyx_t_70 < __pyx_t_69; __pyx_t_70+=1) {
                                      __pyx_v_i2 = __pyx_t_70;

                                      /* "cyext_acv.pyx":3773
 *                     for nv in range(va_id[va].size()):
 *                         for i2 in range(d):
 *                             phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]             # <<<<<<<<<<<<<<
 * 
 *         for j in range(m):
 */
                                      __pyx_t_32 = __pyx_v_b;
                                      __pyx_t_50 = __pyx_v_leaf_numb;
                                      __pyx_t_51 = __pyx_v_b;
                                      __pyx_t_37 = (*((long const  *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_32 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_50 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_46 = __pyx_v_i2;
                                      __pyx_t_34 = __pyx_v_leaf_numb;
                                      __pyx_t_35 = __pyx_v_counter;
                                      __pyx_t_52 = ((__pyx_v_va_id[__pyx_v_va])[__pyx_v_nv]);
                                      __pyx_t_47 = __pyx_v_i2;
                                      *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_34 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_35 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_52 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_47 * __pyx_v_phi_b.strides[3]) )) += (((__pyx_v_coef_0 + __pyx_v_coef) * (__pyx_v_p_si - __pyx_v_p_s)) * (*((double const  *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_51 * __pyx_v_values.strides[0]) ) + __pyx_t_37 * __pyx_v_values.strides[1]) ) + __pyx_t_46 * __pyx_v_values.strides[2]) ))));
                                    }
                                  }
                                }
                                __pyx_L62_continue:;
                              }
                              goto __pyx_L102;
                              __pyx_L37_error:;
                              {
                                  #ifdef WITH_THREAD
                                  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                                  #endif
                                  #ifdef _OPENMP
                                  #pragma omp flush(__pyx_parallel_exc_type)
                                  #endif /* _OPENMP */
                                  if (!__pyx_parallel_exc_type) {
                                    __Pyx_ErrFetchWithState(&__pyx_parallel_exc_type, &__pyx_parallel_exc_value, &__pyx_parallel_exc_tb);
                                    __pyx_parallel_filename = __pyx_filename; __pyx_parallel_lineno = __pyx_lineno; __pyx_parallel_clineno = __pyx_clineno;
                                    __Pyx_GOTREF(__pyx_parallel_exc_type);
                                  }
                                  #ifdef WITH_THREAD
                                  __Pyx_PyGILState_Release(__pyx_gilstate_save);
                                  #endif
                              }
                              __pyx_parallel_why = 4;
                              goto __pyx_L101;
                              __pyx_L101:;
                              #ifdef _OPENMP
                              #pragma omp critical(__pyx_parallel_lastprivates4)
                              #endif /* _OPENMP */
                              {
                                  __pyx_parallel_temp0 = __pyx_v_S_size;
                                  __pyx_parallel_temp1 = __pyx_v_a_it;
                                  __pyx_parallel_temp2 = __pyx_v_add;
                                  __pyx_parallel_temp3 = __pyx_v_b_it;
                                  __pyx_parallel_temp4 = __pyx_v_ci;
                                  __pyx_parallel_temp5 = __pyx_v_cj;
                                  __pyx_parallel_temp6 = __pyx_v_coef;
                                  __pyx_parallel_temp7 = __pyx_v_coef_0;
                                  __pyx_parallel_temp8 = __pyx_v_counter;
                                  __pyx_parallel_temp9 = __pyx_v_cs;
                                  __pyx_parallel_temp10 = __pyx_v_csi;
                                  __pyx_parallel_temp11 = __pyx_v_i;
                                  __pyx_parallel_temp12 = __pyx_v_i2;
                                  __pyx_parallel_temp13 = __pyx_v_l;
                                  __pyx_parallel_temp14 = __pyx_v_leaf_numb;
                                  __pyx_parallel_temp15 = __pyx_v_lm;
                                  __pyx_parallel_temp16 = __pyx_v_ns;
                                  __pyx_parallel_temp17 = __pyx_v_nv;
                                  __pyx_parallel_temp18 = __pyx_v_nv_bool;
                                  __pyx_parallel_temp19 = __pyx_v_o_all;
                                  __pyx_parallel_temp20 = __pyx_v_p_s;
                                  __pyx_parallel_temp21 = __pyx_v_p_si;
                                  __pyx_parallel_temp22 = __pyx_v_pow_set_size;
                                  __pyx_parallel_temp23 = __pyx_v_s;
                                  __pyx_parallel_temp24 = __pyx_v_set_size;
                                  __pyx_parallel_temp25 = __pyx_v_va;
                                  __pyx_parallel_temp26 = __pyx_v_va_size;
                              }
                              __pyx_L102:;
                              #ifdef _OPENMP
                              #pragma omp flush(__pyx_parallel_why)
                              #endif /* _OPENMP */
                          }
                      }
                      #ifdef _OPENMP
                      Py_END_ALLOW_THREADS
                      #else
{
#ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      #endif /* _OPENMP */
                      /* Clean up any temporaries */
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                      #ifndef _OPENMP
}
#endif /* _OPENMP */
                  }
              }
              if (__pyx_parallel_exc_type) {
                /* This may have been overridden by a continue, break or return in another thread. Prefer the error. */
                __pyx_parallel_why = 4;
              }
              if (__pyx_parallel_why) {
                __pyx_v_S_size = __pyx_parallel_temp0;
                __pyx_v_a_it = __pyx_parallel_temp1;
                __pyx_v_add = __pyx_parallel_temp2;
                __pyx_v_b_it = __pyx_parallel_temp3;
                __pyx_v_ci = __pyx_parallel_temp4;
                __pyx_v_cj = __pyx_parallel_temp5;
                __pyx_v_coef = __pyx_parallel_temp6;
                __pyx_v_coef_0 = __pyx_parallel_temp7;
                __pyx_v_counter = __pyx_parallel_temp8;
                __pyx_v_cs = __pyx_parallel_temp9;
                __pyx_v_csi = __pyx_parallel_temp10;
                __pyx_v_i = __pyx_parallel_temp11;
                __pyx_v_i2 = __pyx_parallel_temp12;
                __pyx_v_l = __pyx_parallel_temp13;
                __pyx_v_leaf_numb = __pyx_parallel_temp14;
                __pyx_v_lm = __pyx_parallel_temp15;
                __pyx_v_ns = __pyx_parallel_temp16;
                __pyx_v_nv = __pyx_parallel_temp17;
                __pyx_v_nv_bool = __pyx_parallel_temp18;
                __pyx_v_o_all = __pyx_parallel_temp19;
                __pyx_v_p_s = __pyx_parallel_temp20;
                __pyx_v_p_si = __pyx_parallel_temp21;
                __pyx_v_pow_set_size = __pyx_parallel_temp22;
                __pyx_v_s = __pyx_parallel_temp23;
                __pyx_v_set_size = __pyx_parallel_temp24;
                __pyx_v_va = __pyx_parallel_temp25;
                __pyx_v_va_size = __pyx_parallel_temp26;
                switch (__pyx_parallel_why) {
                      case 4:
                  {
                      #ifdef WITH_THREAD
                      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                      #endif
                      __Pyx_GIVEREF(__pyx_parallel_exc_type);
                      __Pyx_ErrRestoreWithState(__pyx_parallel_exc_type, __pyx_parallel_exc_value, __pyx_parallel_exc_tb);
                      __pyx_filename = __pyx_parallel_filename; __pyx_lineno = __pyx_parallel_lineno; __pyx_clineno = __pyx_parallel_clineno;
                      #ifdef WITH_THREAD
                      __Pyx_PyGILState_Release(__pyx_gilstate_save);
                      #endif
                  }
                  goto __pyx_L33_error;
                }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":3659
 *                         phi_b[leaf_numb, counter, j, i2] = 0
 *         nb_leaf = leaves_nb[b]
 *         for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             node_id_v2[leaf_numb].clear()
 *             lm = 0
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L34;
          }
          __pyx_L33_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L34:;
        }
    }

    /* "cyext_acv.pyx":3775
 *                             phi_b[leaf_numb, counter, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]
 * 
 *         for j in range(m):             # <<<<<<<<<<<<<<
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 */
    __pyx_t_26 = __pyx_v_m;
    __pyx_t_25 = __pyx_t_26;
    for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_25; __pyx_t_19+=1) {
      __pyx_v_j = __pyx_t_19;

      /* "cyext_acv.pyx":3776
 * 
 *         for j in range(m):
 *             for i2 in range(d):             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     for counter in range(phi_b.shape[1]):
 */
      __pyx_t_27 = __pyx_v_d;
      __pyx_t_28 = __pyx_t_27;
      for (__pyx_t_29 = 0; __pyx_t_29 < __pyx_t_28; __pyx_t_29+=1) {
        __pyx_v_i2 = __pyx_t_29;

        /* "cyext_acv.pyx":3777
 *         for j in range(m):
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):             # <<<<<<<<<<<<<<
 *                     for counter in range(phi_b.shape[1]):
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 */
        __pyx_t_21 = (__pyx_v_phi_b.shape[0]);
        __pyx_t_22 = __pyx_t_21;
        for (__pyx_t_30 = 0; __pyx_t_30 < __pyx_t_22; __pyx_t_30+=1) {
          __pyx_v_leaf_numb = __pyx_t_30;

          /* "cyext_acv.pyx":3778
 *             for i2 in range(d):
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     for counter in range(phi_b.shape[1]):             # <<<<<<<<<<<<<<
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 * 
 */
          __pyx_t_23 = (__pyx_v_phi_b.shape[1]);
          __pyx_t_24 = __pyx_t_23;
          for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_24; __pyx_t_31+=1) {
            __pyx_v_counter = __pyx_t_31;

            /* "cyext_acv.pyx":3779
 *                 for leaf_numb in range(phi_b.shape[0]):
 *                     for counter in range(phi_b.shape[1]):
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(phi)/m
 */
            __pyx_t_50 = __pyx_v_leaf_numb;
            __pyx_t_32 = __pyx_v_counter;
            __pyx_t_46 = __pyx_v_j;
            __pyx_t_51 = __pyx_v_i2;
            __pyx_t_47 = __pyx_v_j;
            __pyx_t_35 = __pyx_v_i2;
            *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi.data + __pyx_t_47 * __pyx_v_phi.strides[0]) ) + __pyx_t_35 * __pyx_v_phi.strides[1]) )) += (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_phi_b.data + __pyx_t_50 * __pyx_v_phi_b.strides[0]) ) + __pyx_t_32 * __pyx_v_phi_b.strides[1]) ) + __pyx_t_46 * __pyx_v_phi_b.strides[2]) ) + __pyx_t_51 * __pyx_v_phi_b.strides[3]) )));
          }
        }
      }
    }
  }

  /* "cyext_acv.pyx":3781
 *                         phi[j, i2] += phi_b[leaf_numb, counter, j, i2]
 * 
 *     return np.asarray(phi)/m             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_phi, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":3573
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef single_shap_values(const double[:] X,             # <<<<<<<<<<<<<<
 *     const double[:, :] data,
 *     const double[:, :, :] values,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_20, 1);
  __Pyx_AddTraceback("cyext_acv.single_shap_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_phi_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_s, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lm_si, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_41single_shap_values(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_41single_shap_values(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_scaling;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees;
  std::vector<std::vector<long> >  __pyx_v_C;
  int __pyx_v_num_threads;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("single_shap_values (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_node_idx_trees,&__pyx_n_s_C,&__pyx_n_s_num_threads,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 1); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 2); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 3); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 4); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 5); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 6); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 7); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 8); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_threads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, 9); __PYX_ERR(0, 3573, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "single_shap_values") < 0)) __PYX_ERR(0, 3573, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(values[0], 0); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 3573, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(values[1], 0); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 3574, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(values[2], 0); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 3575, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(values[3], 0); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 3576, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(values[4], 0); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 3577, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(values[5], 0); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 3578, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_scaling == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3579, __pyx_L3_error)
    __pyx_v_node_idx_trees = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(values[7]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3580, __pyx_L3_error)
    __pyx_v_C = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(values[8]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3581, __pyx_L3_error)
    __pyx_v_num_threads = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_num_threads == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3581, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("single_shap_values", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3573, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.single_shap_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_40single_shap_values(__pyx_self, __pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_40single_shap_values(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, int __pyx_v_scaling, std::vector<std::vector<std::vector<long> > >  __pyx_v_node_idx_trees, std::vector<std::vector<long> >  __pyx_v_C, int __pyx_v_num_threads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_shap_values", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 3573, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_single_shap_values(__pyx_v_X, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_node_idx_trees, __pyx_v_C, __pyx_v_num_threads, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.single_shap_values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3787
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf_approx(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

static PyObject *__pyx_pw_9cyext_acv_43global_sdp_clf_approx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_clf_approx(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, PyObject *__pyx_v_search_space, int __pyx_v_stop, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  unsigned int __pyx_v_max_leaves;
  int __pyx_v_n_trees;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_mean_forest = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_mean_forest_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_v_a_it;
  unsigned int __pyx_v_b_it;
  unsigned int __pyx_v_o_all;
  unsigned int __pyx_v_p;
  unsigned int __pyx_v_p_s;
  unsigned int __pyx_v_nb_leaf;
  unsigned int __pyx_v_p_u;
  unsigned int __pyx_v_p_d;
  unsigned int __pyx_v_p_su;
  unsigned int __pyx_v_p_sd;
  double __pyx_v_ss;
  double __pyx_v_ss_a;
  double __pyx_v_ss_u;
  double __pyx_v_ss_d;
  int __pyx_v_b;
  int __pyx_v_leaf_numb;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_lm;
  int __pyx_v_lm_s;
  int __pyx_v_max_size;
  int __pyx_v_size;
  __Pyx_memviewslice __pyx_v_S = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_len_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_remove_va = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_u_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_p_d_n = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_n;
  double __pyx_v_n_u;
  double __pyx_v_n_d;
  CYTHON_UNUSED long __pyx_v_power_set_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  long __pyx_t_22;
  long __pyx_t_23;
  int __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_t_26;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_27;
  __Pyx_memviewslice __pyx_t_28 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_29 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_30;
  std::vector<std::vector<long> > ::size_type __pyx_t_31;
  std::vector<long> ::size_type __pyx_t_32;
  std::vector<long> ::size_type __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  unsigned int __pyx_t_38;
  unsigned int __pyx_t_39;
  int __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  Py_ssize_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  int __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  unsigned int __pyx_t_46;
  int __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  Py_ssize_t __pyx_t_51;
  int __pyx_t_52;
  double __pyx_t_53;
  PyObject *__pyx_t_54 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf_approx", 0);

  /* "cyext_acv.pyx":3793
 *             int minimal, list search_space, bint stop):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":3794
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":3795
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]             # <<<<<<<<<<<<<<
 *     cdef int n_trees = values.shape[0]
 *     cdef double[:, :, :] leaves_tree
 */
  __pyx_v_max_leaves = (__pyx_v_partition_leaves_trees.shape[1]);

  /* "cyext_acv.pyx":3796
 *     cdef unsigned int m = X.shape[1]
 *     cdef unsigned int max_leaves = partition_leaves_trees.shape[1]
 *     cdef int n_trees = values.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double[:, :, :] leaves_tree
 *     cdef double[:, :] leaf_part
 */
  __pyx_v_n_trees = (__pyx_v_values.shape[0]);

  /* "cyext_acv.pyx":3803
 * 
 *     cdef double[:, :] mean_forest
 *     mean_forest = np.zeros((N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
  __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":3806
 * 
 *     cdef double[:, :, :, :] mean_forest_b
 *     mean_forest_b = np.zeros((n_trees, leaf_idx_trees.shape[1], N, 3))             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_leaf_idx_trees.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 3806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mean_forest_b = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":3810
 *     cdef double[:] sdp
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3811
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] lm_u, lm_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 3811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":3819
 * 
 *     cdef long[:] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_len_s_star = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3824
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3824, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3825
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":3826
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":3827
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 3827, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":3830
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_R_buf = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3832
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 3832, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_14) {

    /* "cyext_acv.pyx":3833
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 3833, __pyx_L1_error)
    }
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3833, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_ci = __pyx_t_17;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 3833, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_18 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_cj = __pyx_t_20;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 3833, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3833, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 3833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":3834
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]             # <<<<<<<<<<<<<<
 *         for ci in range(len(C)):
 *             i = 0
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 3834, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_6); __pyx_t_15 = 0;
    for (;;) {
      if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_7); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3834, __pyx_L1_error)
      #else
      __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_i = __pyx_t_13;
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_14 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_21 = (__pyx_t_14 != 0);
      if (__pyx_t_21) {
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3834, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3834, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_7);
        PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
        __pyx_t_7 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 3834, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":3835
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):             # <<<<<<<<<<<<<<
 *             i = 0
 *             for cj in range(len(C[ci])):
 */
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 3835, __pyx_L1_error)
    }
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3835, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_ci = __pyx_t_17;

      /* "cyext_acv.pyx":3836
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 *             i = 0             # <<<<<<<<<<<<<<
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 */
      __pyx_v_i = 0;

      /* "cyext_acv.pyx":3837
 *         for ci in range(len(C)):
 *             i = 0
 *             for cj in range(len(C[ci])):             # <<<<<<<<<<<<<<
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 */
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 3837, __pyx_L1_error)
      }
      __pyx_t_2 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_18 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_19 = __pyx_t_18;
      for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
        __pyx_v_cj = __pyx_t_20;

        /* "cyext_acv.pyx":3838
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 3838, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3838, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_21 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_search_space, Py_EQ)); if (unlikely(__pyx_t_21 < 0)) __PYX_ERR(0, 3838, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_14 = (__pyx_t_21 != 0);
        if (__pyx_t_14) {

          /* "cyext_acv.pyx":3839
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     break
 *             if i != 0:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "cyext_acv.pyx":3840
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 *                     break             # <<<<<<<<<<<<<<
 *             if i != 0:
 *                 va_id += [C[ci]]
 */
          goto __pyx_L16_break;

          /* "cyext_acv.pyx":3838
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        }
      }
      __pyx_L16_break:;

      /* "cyext_acv.pyx":3841
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      __pyx_t_14 = ((__pyx_v_i != 0) != 0);
      if (__pyx_t_14) {

        /* "cyext_acv.pyx":3842
 *                     break
 *             if i != 0:
 *                 va_id += [C[ci]]             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in search_space]
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 3842, __pyx_L1_error)
        }
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        PyList_SET_ITEM(__pyx_t_2, 0, PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_va_id, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3842, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v_va_id, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "cyext_acv.pyx":3841
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      }
    }

    /* "cyext_acv.pyx":3832
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":3844
 *                 va_id += [C[ci]]
 *     else:
 *         va_id = [[i] for i in search_space]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 3844, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
    for (;;) {
      if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_1); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3844, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3844, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_i = __pyx_t_13;
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
      __pyx_t_1 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 3844, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":3846
 *         va_id = [[i] for i in search_space]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3846, __pyx_L1_error)
  __pyx_v_m = __pyx_t_15;

  /* "cyext_acv.pyx":3847
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(minimal, m + 1):
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_power = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "cyext_acv.pyx":3848
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(minimal, m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":3849
 *     power = []
 *     max_size = 0
 *     for size in range(minimal, m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_22 = (__pyx_v_m + 1);
  __pyx_t_23 = __pyx_t_22;
  for (__pyx_t_13 = __pyx_v_minimal; __pyx_t_13 < __pyx_t_23; __pyx_t_13+=1) {
    __pyx_v_size = __pyx_t_13;

    /* "cyext_acv.pyx":3850
 *     max_size = 0
 *     for size in range(minimal, m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "cyext_acv.pyx":3851
 *     for size in range(minimal, m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_combinations); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_24 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_24 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_24, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_24, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_7 = __pyx_t_6; __Pyx_INCREF(__pyx_t_7); __pyx_t_15 = 0;
      __pyx_t_25 = NULL;
    } else {
      __pyx_t_15 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3851, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_25 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 3851, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_25)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3851, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3851, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3851, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_25(__pyx_t_7);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 3851, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "cyext_acv.pyx":3852
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_6); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 3852, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "cyext_acv.pyx":3853
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":3851
 *     for size in range(minimal, m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "cyext_acv.pyx":3854
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_26 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_26 == ((int)-1))) __PYX_ERR(0, 3854, __pyx_L1_error)

    /* "cyext_acv.pyx":3855
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
    __pyx_t_14 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_14) {

      /* "cyext_acv.pyx":3856
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 */
      goto __pyx_L22_break;

      /* "cyext_acv.pyx":3855
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
    }
  }
  __pyx_L22_break:;

  /* "cyext_acv.pyx":3857
 *         if max_size >= 2**15:
 *             break
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3857, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_27;

  /* "cyext_acv.pyx":3859
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_28 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_28.memview)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s_star = __pyx_t_28;
  __pyx_t_28.memview = NULL;
  __pyx_t_28.data = NULL;

  /* "cyext_acv.pyx":3862
 * 
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 3862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":3863
 *     cdef double[:, :, :] p_n, p_u_n, p_d_n
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     p_d_n = np.zeros((n_trees, max_leaves, N))
 *     cdef double n, n_u, n_d
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 3863, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_u_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":3864
 *     p_n = np.zeros((n_trees, max_leaves, N))
 *     p_u_n = np.zeros((n_trees, max_leaves, N))
 *     p_d_n = np.zeros((n_trees, max_leaves, N))             # <<<<<<<<<<<<<<
 *     cdef double n, n_u, n_d
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_leaves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_7);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_29 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_29.memview)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_d_n = __pyx_t_29;
  __pyx_t_29.memview = NULL;
  __pyx_t_29.data = NULL;

  /* "cyext_acv.pyx":3867
 *     cdef double n, n_u, n_d
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":3868
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(len(power))):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 3868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_S = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":3870
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(len(power))):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyList_GET_SIZE(__pyx_v_power); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3870, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_1 = __pyx_t_6; __Pyx_INCREF(__pyx_t_1); __pyx_t_15 = 0;
    __pyx_t_25 = NULL;
  } else {
    __pyx_t_15 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_25 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 3870, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_25)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3870, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 3870, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_25(__pyx_t_1);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3870, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3870, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_s_0 = __pyx_t_13;

    /* "cyext_acv.pyx":3871
 * 
 *     for s_0 in tqdm(range(len(power))):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_30 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_31 = __pyx_t_30;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_31; __pyx_t_13+=1) {
      __pyx_v_s_1 = __pyx_t_13;

      /* "cyext_acv.pyx":3872
 *     for s_0 in tqdm(range(len(power))):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_32 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_33 = __pyx_t_32;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_33; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3873
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        __pyx_t_34 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":3875
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":3876
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             for i in range(N):
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":3877
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":3878
 *             r.clear()
 *             N = R.size()
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 * 
 */
      __pyx_t_11 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_11;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3879
 *             N = R.size()
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
        __pyx_t_34 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":3881
 *                 R_buf[i] = R[i]
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 */
      __pyx_t_24 = __pyx_v_n_trees;
      __pyx_t_35 = __pyx_t_24;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
        __pyx_v_b = __pyx_t_36;

        /* "cyext_acv.pyx":3882
 * 
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 */
        __pyx_t_34 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_34 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":3883
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 */
        __pyx_t_11 = __pyx_v_nb_leaf;
        __pyx_t_12 = __pyx_t_11;
        for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_12; __pyx_t_37+=1) {
          __pyx_v_leaf_numb = __pyx_t_37;

          /* "cyext_acv.pyx":3884
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 */
          __pyx_t_38 = __pyx_v_N;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_i = __pyx_t_40;

            /* "cyext_acv.pyx":3885
 *                 for leaf_numb in range(nb_leaf):
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_41 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_41 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_43 * __pyx_v_p_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":3886
 *                     for i in range(N):
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_43 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_43 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_41 * __pyx_v_p_u_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":3887
 *                         p_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0             # <<<<<<<<<<<<<<
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 */
            __pyx_t_34 = __pyx_v_i;
            __pyx_t_41 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
            *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_41 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_43 * __pyx_v_p_d_n.strides[2]) )) = 0.0;

            /* "cyext_acv.pyx":3888
 *                         p_u_n[b, leaf_numb, R_buf[i]] = 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):             # <<<<<<<<<<<<<<
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 */
            for (__pyx_t_44 = 0; __pyx_t_44 < 3; __pyx_t_44+=1) {
              __pyx_v_j = __pyx_t_44;

              /* "cyext_acv.pyx":3889
 *                         p_d_n[b, leaf_numb, R_buf[i]] = 0
 *                         for j in range(3):
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0             # <<<<<<<<<<<<<<
 * 
 *             for b in range(n_trees):
 */
              __pyx_t_34 = __pyx_v_i;
              __pyx_t_43 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_45 = __pyx_v_j;
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) )) = 0.0;
            }
          }
        }
      }

      /* "cyext_acv.pyx":3891
 *                             mean_forest_b[b, leaf_numb, R_buf[i], j] = 0
 * 
 *             for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 */
      __pyx_t_24 = __pyx_v_n_trees;
      __pyx_t_35 = __pyx_t_24;
      for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_35; __pyx_t_36+=1) {
        __pyx_v_b = __pyx_t_36;

        /* "cyext_acv.pyx":3892
 * 
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):
 */
        __pyx_t_34 = __pyx_v_b;
        __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_34 * __pyx_v_leaves_nb.strides[0]) )));

        /* "cyext_acv.pyx":3893
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         o_all = 0
 */
        {
            #ifdef WITH_THREAD
            PyThreadState *_save;
            Py_UNBLOCK_THREADS
            __Pyx_FastGIL_Remember();
            #endif
            /*try:*/ {
              __pyx_t_11 = __pyx_v_nb_leaf;
              if ((1 == 0)) abort();
              {
                  #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                      #undef likely
                      #undef unlikely
                      #define likely(x)   (x)
                      #define unlikely(x) (x)
                  #endif
                  __pyx_t_38 = (__pyx_t_11 - 0 + 1 - 1/abs(1)) / 1;
                  if (__pyx_t_38 > 0)
                  {
                      #ifdef _OPENMP
                      #pragma omp parallel reduction(+:__pyx_v_p) reduction(+:__pyx_v_p_d) reduction(+:__pyx_v_p_s) reduction(+:__pyx_v_p_sd) reduction(+:__pyx_v_p_su) reduction(+:__pyx_v_p_u) private(__pyx_t_14, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_21, __pyx_t_34, __pyx_t_37, __pyx_t_39, __pyx_t_40, __pyx_t_41, __pyx_t_42, __pyx_t_43, __pyx_t_44, __pyx_t_45, __pyx_t_46, __pyx_t_47, __pyx_t_48, __pyx_t_49, __pyx_t_50, __pyx_t_51, __pyx_t_52, __pyx_t_53)
                      #endif /* _OPENMP */
                      {
                          #ifdef _OPENMP
                          #pragma omp for lastprivate(__pyx_v_a_it) lastprivate(__pyx_v_b_it) lastprivate(__pyx_v_i) lastprivate(__pyx_v_j) firstprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_leaf_numb) lastprivate(__pyx_v_lm) lastprivate(__pyx_v_lm_s) lastprivate(__pyx_v_o_all) lastprivate(__pyx_v_s) schedule(dynamic)
                          #endif /* _OPENMP */
                          for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_38; __pyx_t_12++){
                              {
                                  __pyx_v_leaf_numb = (int)(0 + 1 * __pyx_t_12);
                                  /* Initialize private variables to invalid values */
                                  __pyx_v_a_it = ((unsigned int)0xbad0bad0);
                                  __pyx_v_b_it = ((unsigned int)0xbad0bad0);
                                  __pyx_v_i = ((int)0xbad0bad0);
                                  __pyx_v_j = ((int)0xbad0bad0);
                                  __pyx_v_lm = ((int)0xbad0bad0);
                                  __pyx_v_lm_s = ((int)0xbad0bad0);
                                  __pyx_v_o_all = ((unsigned int)0xbad0bad0);
                                  __pyx_v_s = ((int)0xbad0bad0);

                                  /* "cyext_acv.pyx":3894
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         o_all = 0
 *                         for s in range(S_size):
 */
                                  __pyx_t_39 = __pyx_v_N;
                                  __pyx_t_46 = __pyx_t_39;
                                  for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_46; __pyx_t_37+=1) {
                                    __pyx_v_i = __pyx_t_37;

                                    /* "cyext_acv.pyx":3895
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
 *                     for i in range(N):
 *                         o_all = 0             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                    __pyx_v_o_all = 0;

                                    /* "cyext_acv.pyx":3896
 *                     for i in range(N):
 *                         o_all = 0
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 */
                                    __pyx_t_40 = __pyx_v_S_size;
                                    __pyx_t_44 = __pyx_t_40;
                                    for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_44; __pyx_t_47+=1) {
                                      __pyx_v_s = __pyx_t_47;

                                      /* "cyext_acv.pyx":3897
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      __pyx_t_34 = __pyx_v_i;
                                      __pyx_t_45 = __pyx_v_s;
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_34 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_45 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_43 = __pyx_v_s;
                                      __pyx_t_48 = __pyx_v_b;
                                      __pyx_t_49 = __pyx_v_leaf_numb;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_43 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_51 = 1;
                                      __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_41 * __pyx_v_X.strides[0]) ) + __pyx_t_42 * __pyx_v_X.strides[1]) ))) > (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_50 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      if (!__pyx_t_21) {
                                      } else {
                                        __pyx_t_14 = __pyx_t_21;
                                        goto __pyx_L58_bool_binop_done;
                                      }
                                      __pyx_t_43 = __pyx_v_i;
                                      __pyx_t_51 = __pyx_v_s;
                                      __pyx_t_50 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_49 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_48 = __pyx_v_s;
                                      __pyx_t_45 = __pyx_v_b;
                                      __pyx_t_34 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                      __pyx_t_41 = 0;
                                      __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_X.data + __pyx_t_50 * __pyx_v_X.strides[0]) ) + __pyx_t_49 * __pyx_v_X.strides[1]) ))) < (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                      __pyx_t_14 = __pyx_t_21;
                                      __pyx_L58_bool_binop_done:;
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3898
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1             # <<<<<<<<<<<<<<
 *                         if o_all > 0:
 *                             continue
 */
                                        __pyx_v_o_all = (__pyx_v_o_all + 1);

                                        /* "cyext_acv.pyx":3897
 *                         o_all = 0
 *                         for s in range(S_size):
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 */
                                      }
                                    }

                                    /* "cyext_acv.pyx":3899
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    __pyx_t_14 = ((__pyx_v_o_all > 0) != 0);
                                    if (__pyx_t_14) {

                                      /* "cyext_acv.pyx":3900
 *                                 o_all = o_all + 1
 *                         if o_all > 0:
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         p = 0
 */
                                      goto __pyx_L53_continue;

                                      /* "cyext_acv.pyx":3899
 *                             if ((X[R_buf[i], S[s]] > partition_leaves_trees[b, leaf_numb, S[s], 1]) or (X[R_buf[i], S[s]] < partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                 o_all = o_all + 1
 *                         if o_all > 0:             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
                                    }

                                    /* "cyext_acv.pyx":3902
 *                             continue
 * 
 *                         p = 0             # <<<<<<<<<<<<<<
 *                         p_u = 0
 *                         p_d = 0
 */
                                    __pyx_v_p = 0;

                                    /* "cyext_acv.pyx":3903
 * 
 *                         p = 0
 *                         p_u = 0             # <<<<<<<<<<<<<<
 *                         p_d = 0
 *                         p_s = 0
 */
                                    __pyx_v_p_u = 0;

                                    /* "cyext_acv.pyx":3904
 *                         p = 0
 *                         p_u = 0
 *                         p_d = 0             # <<<<<<<<<<<<<<
 *                         p_s = 0
 *                         p_su = 0
 */
                                    __pyx_v_p_d = 0;

                                    /* "cyext_acv.pyx":3905
 *                         p_u = 0
 *                         p_d = 0
 *                         p_s = 0             # <<<<<<<<<<<<<<
 *                         p_su = 0
 *                         p_sd = 0
 */
                                    __pyx_v_p_s = 0;

                                    /* "cyext_acv.pyx":3906
 *                         p_d = 0
 *                         p_s = 0
 *                         p_su = 0             # <<<<<<<<<<<<<<
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 */
                                    __pyx_v_p_su = 0;

                                    /* "cyext_acv.pyx":3907
 *                         p_s = 0
 *                         p_su = 0
 *                         p_sd = 0             # <<<<<<<<<<<<<<
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 */
                                    __pyx_v_p_sd = 0;

                                    /* "cyext_acv.pyx":3908
 *                         p_su = 0
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                             a_it = 0
 *                             b_it = 0
 */
                                    __pyx_t_16 = (__pyx_v_data.shape[0]);
                                    __pyx_t_17 = __pyx_t_16;
                                    for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_17; __pyx_t_40+=1) {
                                      __pyx_v_j = __pyx_t_40;

                                      /* "cyext_acv.pyx":3909
 *                         p_sd = 0
 *                         for j in range(data.shape[0]):
 *                             a_it = 0             # <<<<<<<<<<<<<<
 *                             b_it = 0
 *                             lm = 0
 */
                                      __pyx_v_a_it = 0;

                                      /* "cyext_acv.pyx":3910
 *                         for j in range(data.shape[0]):
 *                             a_it = 0
 *                             b_it = 0             # <<<<<<<<<<<<<<
 *                             lm = 0
 *                             lm_s = 0
 */
                                      __pyx_v_b_it = 0;

                                      /* "cyext_acv.pyx":3911
 *                             a_it = 0
 *                             b_it = 0
 *                             lm = 0             # <<<<<<<<<<<<<<
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 */
                                      __pyx_v_lm = 0;

                                      /* "cyext_acv.pyx":3912
 *                             b_it = 0
 *                             lm = 0
 *                             lm_s = 0             # <<<<<<<<<<<<<<
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 */
                                      __pyx_v_lm_s = 0;

                                      /* "cyext_acv.pyx":3913
 *                             lm = 0
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 */
                                      __pyx_t_18 = (__pyx_v_data.shape[1]);
                                      __pyx_t_19 = __pyx_t_18;
                                      for (__pyx_t_44 = 0; __pyx_t_44 < __pyx_t_19; __pyx_t_44+=1) {
                                        __pyx_v_s = __pyx_t_44;

                                        /* "cyext_acv.pyx":3914
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        __pyx_t_48 = __pyx_v_j;
                                        __pyx_t_41 = __pyx_v_s;
                                        __pyx_t_42 = __pyx_v_b;
                                        __pyx_t_34 = __pyx_v_leaf_numb;
                                        __pyx_t_45 = __pyx_v_s;
                                        __pyx_t_51 = 1;
                                        __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_48 * __pyx_v_data.strides[0]) ) + __pyx_t_41 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_21) {
                                        } else {
                                          __pyx_t_14 = __pyx_t_21;
                                          goto __pyx_L66_bool_binop_done;
                                        }
                                        __pyx_t_51 = __pyx_v_j;
                                        __pyx_t_45 = __pyx_v_s;
                                        __pyx_t_34 = __pyx_v_b;
                                        __pyx_t_42 = __pyx_v_leaf_numb;
                                        __pyx_t_41 = __pyx_v_s;
                                        __pyx_t_48 = 0;
                                        __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_51 * __pyx_v_data.strides[0]) ) + __pyx_t_45 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_34 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_14 = __pyx_t_21;
                                        __pyx_L66_bool_binop_done:;
                                        if (__pyx_t_14) {

                                          /* "cyext_acv.pyx":3915
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1             # <<<<<<<<<<<<<<
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 */
                                          __pyx_v_a_it = (__pyx_v_a_it + 1);

                                          /* "cyext_acv.pyx":3914
 *                             lm_s = 0
 *                             for s in range(data.shape[1]):
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):             # <<<<<<<<<<<<<<
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":3916
 *                                 if ((data[j, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[j, s] >= partition_leaves_trees[b, leaf_numb, s, 0])):
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):             # <<<<<<<<<<<<<<
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1
 */
                                      __pyx_t_44 = __pyx_v_S_size;
                                      __pyx_t_47 = __pyx_t_44;
                                      for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_47; __pyx_t_52+=1) {
                                        __pyx_v_s = __pyx_t_52;

                                        /* "cyext_acv.pyx":3917
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        __pyx_t_48 = __pyx_v_s;
                                        __pyx_t_41 = __pyx_v_j;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_48 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_34 = __pyx_v_s;
                                        __pyx_t_45 = __pyx_v_b;
                                        __pyx_t_51 = __pyx_v_leaf_numb;
                                        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_49 = 1;
                                        __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_41 * __pyx_v_data.strides[0]) ) + __pyx_t_42 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_43 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_49 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        if (__pyx_t_21) {
                                        } else {
                                          __pyx_t_14 = __pyx_t_21;
                                          goto __pyx_L71_bool_binop_done;
                                        }
                                        __pyx_t_34 = __pyx_v_s;
                                        __pyx_t_49 = __pyx_v_j;
                                        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_34 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_51 = __pyx_v_s;
                                        __pyx_t_45 = __pyx_v_b;
                                        __pyx_t_48 = __pyx_v_leaf_numb;
                                        __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_51 * __pyx_v_S.strides[0]) )));
                                        __pyx_t_41 = 0;
                                        __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_49 * __pyx_v_data.strides[0]) ) + __pyx_t_43 * __pyx_v_data.strides[1]) ))) >= (*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition_leaves_trees.data + __pyx_t_45 * __pyx_v_partition_leaves_trees.strides[0]) ) + __pyx_t_48 * __pyx_v_partition_leaves_trees.strides[1]) ) + __pyx_t_42 * __pyx_v_partition_leaves_trees.strides[2]) ) + __pyx_t_41 * __pyx_v_partition_leaves_trees.strides[3]) )))) != 0);
                                        __pyx_t_14 = __pyx_t_21;
                                        __pyx_L71_bool_binop_done:;
                                        if (__pyx_t_14) {

                                          /* "cyext_acv.pyx":3918
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):
 *                                     b_it = b_it + 1             # <<<<<<<<<<<<<<
 * 
 *                             if a_it == data.shape[1]:
 */
                                          __pyx_v_b_it = (__pyx_v_b_it + 1);

                                          /* "cyext_acv.pyx":3917
 *                                     a_it = a_it + 1
 *                             for s in range(S_size):
 *                                 if ((data[j, S[s]] <= partition_leaves_trees[b, leaf_numb, S[s], 1]) and (data[j, S[s]] >= partition_leaves_trees[b, leaf_numb, S[s], 0])):             # <<<<<<<<<<<<<<
 *                                     b_it = b_it + 1
 * 
 */
                                        }
                                      }

                                      /* "cyext_acv.pyx":3920
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      __pyx_t_14 = ((__pyx_v_a_it == (__pyx_v_data.shape[1])) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3921
 * 
 *                             if a_it == data.shape[1]:
 *                                 lm = 1             # <<<<<<<<<<<<<<
 * 
 *                             if b_it == S_size:
 */
                                        __pyx_v_lm = 1;

                                        /* "cyext_acv.pyx":3920
 *                                     b_it = b_it + 1
 * 
 *                             if a_it == data.shape[1]:             # <<<<<<<<<<<<<<
 *                                 lm = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3923
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      __pyx_t_14 = ((__pyx_v_b_it == __pyx_v_S_size) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3924
 * 
 *                             if b_it == S_size:
 *                                 lm_s = 1             # <<<<<<<<<<<<<<
 * 
 *                             p += lm
 */
                                        __pyx_v_lm_s = 1;

                                        /* "cyext_acv.pyx":3923
 *                                 lm = 1
 * 
 *                             if b_it == S_size:             # <<<<<<<<<<<<<<
 *                                 lm_s = 1
 * 
 */
                                      }

                                      /* "cyext_acv.pyx":3926
 *                                 lm_s = 1
 * 
 *                             p += lm             # <<<<<<<<<<<<<<
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 */
                                      __pyx_v_p = (__pyx_v_p + __pyx_v_lm);

                                      /* "cyext_acv.pyx":3927
 * 
 *                             p += lm
 *                             p_s += lm_s             # <<<<<<<<<<<<<<
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 */
                                      __pyx_v_p_s = (__pyx_v_p_s + __pyx_v_lm_s);

                                      /* "cyext_acv.pyx":3928
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                      __pyx_t_51 = __pyx_v_i;
                                      __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_51 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_42 = __pyx_v_j;
                                      __pyx_t_14 = (((*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_41 * __pyx_v_fX.strides[0]) ))) == (*((long *) ( /* dim=0 */ (__pyx_v_y_pred.data + __pyx_t_42 * __pyx_v_y_pred.strides[0]) )))) != 0);
                                      if (__pyx_t_14) {

                                        /* "cyext_acv.pyx":3929
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm             # <<<<<<<<<<<<<<
 *                                 p_su += lm_s
 *                             else:
 */
                                        __pyx_v_p_u = (__pyx_v_p_u + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3930
 *                             if fX[R_buf[i]] == y_pred[j]:
 *                                 p_u += lm
 *                                 p_su += lm_s             # <<<<<<<<<<<<<<
 *                             else:
 *                                 p_d += lm
 */
                                        __pyx_v_p_su = (__pyx_v_p_su + __pyx_v_lm_s);

                                        /* "cyext_acv.pyx":3928
 *                             p += lm
 *                             p_s += lm_s
 *                             if fX[R_buf[i]] == y_pred[j]:             # <<<<<<<<<<<<<<
 *                                 p_u += lm
 *                                 p_su += lm_s
 */
                                        goto __pyx_L75;
                                      }

                                      /* "cyext_acv.pyx":3932
 *                                 p_su += lm_s
 *                             else:
 *                                 p_d += lm             # <<<<<<<<<<<<<<
 *                                 p_sd += lm_s
 * 
 */
                                      /*else*/ {
                                        __pyx_v_p_d = (__pyx_v_p_d + __pyx_v_lm);

                                        /* "cyext_acv.pyx":3933
 *                             else:
 *                                 p_d += lm
 *                                 p_sd += lm_s             # <<<<<<<<<<<<<<
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 */
                                        __pyx_v_p_sd = (__pyx_v_p_sd + __pyx_v_lm_s);
                                      }
                                      __pyx_L75:;
                                    }

                                    /* "cyext_acv.pyx":3935
 *                                 p_sd += lm_s
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_51 = __pyx_v_leaf_numb;
                                      __pyx_t_41 = __pyx_v_i;
                                      __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_45 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_51 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_48 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_45 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_43 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_51 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_43 = 0;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_51 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3936
 * 
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_41 = __pyx_v_b;
                                      __pyx_t_43 = __pyx_v_leaf_numb;
                                      __pyx_t_42 = __pyx_v_i;
                                      __pyx_t_51 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_48 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_41 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_43 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_51 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p_u * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_48 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_45 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_42 = __pyx_v_i;
                                    __pyx_t_51 = __pyx_v_b;
                                    __pyx_t_43 = __pyx_v_leaf_numb;
                                    __pyx_t_41 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_42 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_45 = 1;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_51 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_41 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3937
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 0] += (p * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]]) /p_s if p_s != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 1] += (p_u * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_su if p_su != 0 else 0
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_42 = __pyx_v_b;
                                      __pyx_t_45 = __pyx_v_leaf_numb;
                                      __pyx_t_41 = __pyx_v_i;
                                      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                      __pyx_t_51 = __pyx_v_b;
                                      __pyx_t_34 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_leaf_idx_trees.data + __pyx_t_42 * __pyx_v_leaf_idx_trees.strides[0]) ) + __pyx_t_45 * __pyx_v_leaf_idx_trees.strides[1]) )));
                                      __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_fX.data + __pyx_t_43 * __pyx_v_fX.strides[0]) )));
                                      __pyx_t_53 = ((__pyx_v_p_d * (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_values.data + __pyx_t_51 * __pyx_v_values.strides[0]) ) + __pyx_t_34 * __pyx_v_values.strides[1]) ) + __pyx_t_48 * __pyx_v_values.strides[2]) )))) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_43 = __pyx_v_b;
                                    __pyx_t_45 = __pyx_v_leaf_numb;
                                    __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    __pyx_t_48 = 2;
                                    *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[3]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3939
 *                         mean_forest_b[b, leaf_numb, R_buf[i], 2] += (p_d * values[b, leaf_idx_trees[b, leaf_numb], fX[R_buf[i]]])/p_sd if p_sd != 0 else 0
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 */
                                    if (((__pyx_v_p_s != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p) / __pyx_v_p_s);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_48 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_n.strides[2]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3940
 * 
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0             # <<<<<<<<<<<<<<
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 */
                                    if (((__pyx_v_p_su != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p_u) / __pyx_v_p_su);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_45 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_45 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_u_n.strides[2]) )) += __pyx_t_53;

                                    /* "cyext_acv.pyx":3941
 *                         p_n[b, leaf_numb, R_buf[i]] += (1.*p)/p_s if p_s != 0 else 0
 *                         p_u_n[b, leaf_numb, R_buf[i]] += (1.*p_u)/p_su if p_su != 0 else 0
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
                                    if (((__pyx_v_p_sd != 0) != 0)) {
                                      __pyx_t_53 = ((1. * __pyx_v_p_d) / __pyx_v_p_sd);
                                    } else {
                                      __pyx_t_53 = 0.0;
                                    }
                                    __pyx_t_41 = __pyx_v_i;
                                    __pyx_t_48 = __pyx_v_b;
                                    __pyx_t_42 = __pyx_v_leaf_numb;
                                    __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
                                    *((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_48 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_d_n.strides[2]) )) += __pyx_t_53;
                                    __pyx_L53_continue:;
                                  }
                              }
                          }
                      }
                  }
              }
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   __builtin_expect(!!(x), 1)
                  #define unlikely(x) __builtin_expect(!!(x), 0)
              #endif
            }

            /* "cyext_acv.pyx":3893
 *             for b in range(n_trees):
 *                 nb_leaf = leaves_nb[b]
 *                 for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                     for i in range(N):
 *                         o_all = 0
 */
            /*finally:*/ {
              /*normal exit:*/{
                #ifdef WITH_THREAD
                __Pyx_FastGIL_Forget();
                Py_BLOCK_THREADS
                #endif
                goto __pyx_L48;
              }
              __pyx_L48:;
            }
        }
      }

      /* "cyext_acv.pyx":3943
 *                         p_d_n[b, leaf_numb, R_buf[i]] += (1.*p_d)/p_sd if p_sd != 0 else 0
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 ss_u = 0
 *                 ss_d = 0
 */
      __pyx_t_38 = __pyx_v_N;
      __pyx_t_12 = __pyx_t_38;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_12; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "cyext_acv.pyx":3944
 * 
 *             for i in range(N):
 *                 ss_u = 0             # <<<<<<<<<<<<<<
 *                 ss_d = 0
 *                 ss_a = 0
 */
        __pyx_v_ss_u = 0.0;

        /* "cyext_acv.pyx":3945
 *             for i in range(N):
 *                 ss_u = 0
 *                 ss_d = 0             # <<<<<<<<<<<<<<
 *                 ss_a = 0
 *                 for b in range(n_trees):
 */
        __pyx_v_ss_d = 0.0;

        /* "cyext_acv.pyx":3946
 *                 ss_u = 0
 *                 ss_d = 0
 *                 ss_a = 0             # <<<<<<<<<<<<<<
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 */
        __pyx_v_ss_a = 0.0;

        /* "cyext_acv.pyx":3947
 *                 ss_d = 0
 *                 ss_a = 0
 *                 for b in range(n_trees):             # <<<<<<<<<<<<<<
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0
 */
        __pyx_t_35 = __pyx_v_n_trees;
        __pyx_t_36 = __pyx_t_35;
        for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
          __pyx_v_b = __pyx_t_37;

          /* "cyext_acv.pyx":3948
 *                 ss_a = 0
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]             # <<<<<<<<<<<<<<
 *                     n = 0
 *                     n_u = 0
 */
          __pyx_t_41 = __pyx_v_b;
          __pyx_v_nb_leaf = (*((long *) ( /* dim=0 */ (__pyx_v_leaves_nb.data + __pyx_t_41 * __pyx_v_leaves_nb.strides[0]) )));

          /* "cyext_acv.pyx":3949
 *                 for b in range(n_trees):
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0             # <<<<<<<<<<<<<<
 *                     n_u = 0
 *                     n_d = 0
 */
          __pyx_v_n = 0.0;

          /* "cyext_acv.pyx":3950
 *                     nb_leaf = leaves_nb[b]
 *                     n = 0
 *                     n_u = 0             # <<<<<<<<<<<<<<
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):
 */
          __pyx_v_n_u = 0.0;

          /* "cyext_acv.pyx":3951
 *                     n = 0
 *                     n_u = 0
 *                     n_d = 0             # <<<<<<<<<<<<<<
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 */
          __pyx_v_n_d = 0.0;

          /* "cyext_acv.pyx":3952
 *                     n_u = 0
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 */
          __pyx_t_11 = __pyx_v_nb_leaf;
          __pyx_t_39 = __pyx_t_11;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_leaf_numb = __pyx_t_40;

            /* "cyext_acv.pyx":3953
 *                     n_d = 0
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_45 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n = (__pyx_v_n + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_n.data + __pyx_t_45 * __pyx_v_p_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_n.strides[2]) ))));

            /* "cyext_acv.pyx":3954
 *                     for leaf_numb in range(nb_leaf):
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 * 
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_48 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n_u = (__pyx_v_n_u + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_u_n.data + __pyx_t_48 * __pyx_v_p_u_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_u_n.strides[1]) ) + __pyx_t_45 * __pyx_v_p_u_n.strides[2]) ))));

            /* "cyext_acv.pyx":3955
 *                         n += p_n[b, leaf_numb, R_buf[i]]
 *                         n_u += p_u_n[b, leaf_numb, R_buf[i]]
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]             # <<<<<<<<<<<<<<
 * 
 *                     for leaf_numb in range(nb_leaf):
 */
            __pyx_t_41 = __pyx_v_i;
            __pyx_t_45 = __pyx_v_b;
            __pyx_t_42 = __pyx_v_leaf_numb;
            __pyx_t_48 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
            __pyx_v_n_d = (__pyx_v_n_d + (*((double *) ( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_p_d_n.data + __pyx_t_45 * __pyx_v_p_d_n.strides[0]) ) + __pyx_t_42 * __pyx_v_p_d_n.strides[1]) ) + __pyx_t_48 * __pyx_v_p_d_n.strides[2]) ))));
          }

          /* "cyext_acv.pyx":3957
 *                         n_d += p_d_n[b, leaf_numb, R_buf[i]]
 * 
 *                     for leaf_numb in range(nb_leaf):             # <<<<<<<<<<<<<<
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 */
          __pyx_t_11 = __pyx_v_nb_leaf;
          __pyx_t_39 = __pyx_t_11;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_leaf_numb = __pyx_t_40;

            /* "cyext_acv.pyx":3958
 * 
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 */
            if (((__pyx_v_n != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_48 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_43 = 0;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_a = (__pyx_v_ss_a + __pyx_t_53);

            /* "cyext_acv.pyx":3959
 *                     for leaf_numb in range(nb_leaf):
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0             # <<<<<<<<<<<<<<
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 * 
 */
            if (((__pyx_v_n_u != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_43 = __pyx_v_b;
              __pyx_t_45 = __pyx_v_leaf_numb;
              __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_48 = 1;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_43 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_48 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_u);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_u = (__pyx_v_ss_u + __pyx_t_53);

            /* "cyext_acv.pyx":3960
 *                         ss_a += mean_forest_b[b, leaf_numb, R_buf[i], 0]/n if n !=0 else 0
 *                         ss_u += mean_forest_b[b, leaf_numb, R_buf[i], 1]/n_u if n_u !=0 else 0
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0             # <<<<<<<<<<<<<<
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 */
            if (((__pyx_v_n_d != 0.0) != 0)) {
              __pyx_t_41 = __pyx_v_i;
              __pyx_t_48 = __pyx_v_b;
              __pyx_t_42 = __pyx_v_leaf_numb;
              __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
              __pyx_t_43 = 2;
              __pyx_t_53 = ((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_mean_forest_b.data + __pyx_t_48 * __pyx_v_mean_forest_b.strides[0]) ) + __pyx_t_42 * __pyx_v_mean_forest_b.strides[1]) ) + __pyx_t_45 * __pyx_v_mean_forest_b.strides[2]) ) + __pyx_t_43 * __pyx_v_mean_forest_b.strides[3]) ))) / __pyx_v_n_d);
            } else {
              __pyx_t_53 = 0.0;
            }
            __pyx_v_ss_d = (__pyx_v_ss_d + __pyx_t_53);
          }
        }

        /* "cyext_acv.pyx":3962
 *                         ss_d += mean_forest_b[b, leaf_numb, R_buf[i], 2]/n_d if n_d !=0 else 0
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0             # <<<<<<<<<<<<<<
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 */
        if ((((__pyx_v_ss_u - __pyx_v_ss_d) != 0.0) != 0)) {
          __pyx_t_53 = ((__pyx_v_ss_a - __pyx_v_ss_d) / (__pyx_v_ss_u - __pyx_v_ss_d));
        } else {
          __pyx_t_53 = 0.0;
        }
        __pyx_v_ss = __pyx_t_53;

        /* "cyext_acv.pyx":3963
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_41 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_14 = ((__pyx_v_ss >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_14) {

          /* "cyext_acv.pyx":3964
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )) = __pyx_v_ss;

          /* "cyext_acv.pyx":3965
 *                 if ss >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3966
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_35 = __pyx_v_S_size;
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_s = __pyx_t_37;

            /* "cyext_acv.pyx":3967
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_41 = __pyx_v_s;
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_45 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_42 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_45 * __pyx_v_s_star.strides[0]) ) + __pyx_t_42 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_41 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3963
 * 
 *                 ss = (ss_a - ss_d)/(ss_u - ss_d) if ss_u - ss_d !=0 else 0
 *                 if ss >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = ss
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":3969
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_14 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_14) {

          /* "cyext_acv.pyx":3970
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":3971
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          *((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )) = __pyx_v_S_size;

          /* "cyext_acv.pyx":3972
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_35 = __pyx_v_S_size;
          __pyx_t_36 = __pyx_t_35;
          for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
            __pyx_v_s = __pyx_t_37;

            /* "cyext_acv.pyx":3973
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_41 = __pyx_v_s;
            __pyx_t_43 = __pyx_v_i;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_43 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_45 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_45 * __pyx_v_s_star.strides[1]) )) = (*((long *) ( /* dim=0 */ (__pyx_v_S.data + __pyx_t_41 * __pyx_v_S.strides[0]) )));
          }

          /* "cyext_acv.pyx":3974
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_16 = (__pyx_v_X.shape[1]);
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_17 = __pyx_t_16;
          for (__pyx_t_35 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) ))); __pyx_t_35 < __pyx_t_17; __pyx_t_35+=1) {
            __pyx_v_s = __pyx_t_35;

            /* "cyext_acv.pyx":3975
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_45 = __pyx_v_i;
            __pyx_t_42 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_45 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_48 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_42 * __pyx_v_s_star.strides[0]) ) + __pyx_t_48 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":3969
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":3977
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_38 = __pyx_v_N;
    __pyx_t_12 = __pyx_t_38;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3978
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_41 = __pyx_v_i;
      __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_14 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_43 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_14) {

        /* "cyext_acv.pyx":3979
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 3979, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":3980
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_41 = __pyx_v_i;
        __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_22 = (*((long *) ( /* dim=0 */ (__pyx_v_len_s_star.data + __pyx_t_43 * __pyx_v_len_s_star.strides[0]) )));
        __pyx_t_23 = __pyx_t_22;
        for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_23; __pyx_t_24+=1) {
          __pyx_v_s = __pyx_t_24;

          /* "cyext_acv.pyx":3981
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_41 = __pyx_v_i;
          __pyx_t_43 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_41 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_45 = __pyx_v_s;
          __pyx_t_48 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_43 * __pyx_v_s_star.strides[0]) ) + __pyx_t_45 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_48 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":3978
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":3983
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_32 = __pyx_v_r.size();
    __pyx_t_33 = __pyx_t_32;
    for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_33; __pyx_t_13+=1) {
      __pyx_v_i = __pyx_t_13;

      /* "cyext_acv.pyx":3984
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":3985
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":3987
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_21 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_21) {
    } else {
      goto __pyx_L103_next_and;
    }
    __pyx_t_21 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_21) {
    } else {
      __pyx_t_14 = __pyx_t_21;
      goto __pyx_L102_bool_binop_done;
    }
    __pyx_L103_next_and:;
    __pyx_t_21 = (__pyx_v_stop != 0);
    __pyx_t_14 = __pyx_t_21;
    __pyx_L102_bool_binop_done:;
    if (__pyx_t_14) {

      /* "cyext_acv.pyx":3988
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L27_break;

      /* "cyext_acv.pyx":3987
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":3870
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(len(power))):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L27_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":3990
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_long); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_len_s_star, 1, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_54 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_long); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_54) < 0) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_54); __pyx_t_54 = 0;
  __pyx_t_54 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_54)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_54);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_54);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_54 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":3787
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_clf_approx(double[:, :] X, long[:] fX,             # <<<<<<<<<<<<<<
 *             long[:] y_pred, double[:, :] data,
 *             double[:, :, :] values, double[:, :, :, :] partition_leaves_trees,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_28, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_29, 1);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_approx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_mean_forest_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_S, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_len_s_star, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_u_n, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_p_d_n, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_43global_sdp_clf_approx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_43global_sdp_clf_approx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_fX = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_pred = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_partition_leaves_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaf_idx_trees = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_leaves_nb = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_scaling = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  PyObject *__pyx_v_search_space = 0;
  int __pyx_v_stop;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_clf_approx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_fX,&__pyx_n_s_y_pred,&__pyx_n_s_data,&__pyx_n_s_values,&__pyx_n_s_partition_leaves_trees,&__pyx_n_s_leaf_idx_trees,&__pyx_n_s_leaves_nb,&__pyx_n_s_scaling,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_search_space,&__pyx_n_s_stop,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fX)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 1); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_pred)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 2); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 3); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 4); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_partition_leaves_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 5); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_idx_trees)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 6); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaves_nb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 7); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_scaling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 8); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 9); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 10); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 11); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_space)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 12); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, 13); __PYX_ERR(0, 3787, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_clf_approx") < 0)) __PYX_ERR(0, 3787, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 3787, __pyx_L3_error)
    __pyx_v_fX = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_fX.memview)) __PYX_ERR(0, 3787, __pyx_L3_error)
    __pyx_v_y_pred = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_pred.memview)) __PYX_ERR(0, 3788, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 3788, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_values.memview)) __PYX_ERR(0, 3789, __pyx_L3_error)
    __pyx_v_partition_leaves_trees = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_partition_leaves_trees.memview)) __PYX_ERR(0, 3789, __pyx_L3_error)
    __pyx_v_leaf_idx_trees = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaf_idx_trees.memview)) __PYX_ERR(0, 3790, __pyx_L3_error)
    __pyx_v_leaves_nb = __Pyx_PyObject_to_MemoryviewSlice_ds_long(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_leaves_nb.memview)) __PYX_ERR(0, 3790, __pyx_L3_error)
    __pyx_v_scaling = __Pyx_PyObject_to_MemoryviewSlice_ds_double(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_scaling.memview)) __PYX_ERR(0, 3790, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[9]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 3790, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3791, __pyx_L3_error)
    __pyx_v_search_space = ((PyObject*)values[12]);
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[13]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3791, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_clf_approx", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3787, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_approx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 3790, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_search_space), (&PyList_Type), 1, "search_space", 1))) __PYX_ERR(0, 3791, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_42global_sdp_clf_approx(__pyx_self, __pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_search_space, __pyx_v_stop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_42global_sdp_clf_approx(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_fX, __Pyx_memviewslice __pyx_v_y_pred, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_values, __Pyx_memviewslice __pyx_v_partition_leaves_trees, __Pyx_memviewslice __pyx_v_leaf_idx_trees, __Pyx_memviewslice __pyx_v_leaves_nb, __Pyx_memviewslice __pyx_v_scaling, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, PyObject *__pyx_v_search_space, int __pyx_v_stop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_clf_approx", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_fX.memview)) { __Pyx_RaiseUnboundLocalError("fX"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_pred.memview)) { __Pyx_RaiseUnboundLocalError("y_pred"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_values.memview)) { __Pyx_RaiseUnboundLocalError("values"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_partition_leaves_trees.memview)) { __Pyx_RaiseUnboundLocalError("partition_leaves_trees"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaf_idx_trees.memview)) { __Pyx_RaiseUnboundLocalError("leaf_idx_trees"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_leaves_nb.memview)) { __Pyx_RaiseUnboundLocalError("leaves_nb"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_scaling.memview)) { __Pyx_RaiseUnboundLocalError("scaling"); __PYX_ERR(0, 3787, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_clf_approx(__pyx_v_X, __pyx_v_fX, __pyx_v_y_pred, __pyx_v_data, __pyx_v_values, __pyx_v_partition_leaves_trees, __pyx_v_leaf_idx_trees, __pyx_v_leaves_nb, __pyx_v_scaling, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_search_space, __pyx_v_stop, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_clf_approx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_fX, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_pred, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_leaves_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaf_idx_trees, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_leaves_nb, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_scaling, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":3994
 * 
 * 
 * cpdef double binomialC(unsigned long N, unsigned long k) nogil:             # <<<<<<<<<<<<<<
 *     cdef double r
 *     r = _comb_int_long(N, k)
 */

static PyObject *__pyx_pw_9cyext_acv_45binomialC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_9cyext_acv_binomialC(unsigned long __pyx_v_N, unsigned long __pyx_v_k, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_r;
  double __pyx_r;
  int __pyx_t_1;

  /* "cyext_acv.pyx":3996
 * cpdef double binomialC(unsigned long N, unsigned long k) nogil:
 *     cdef double r
 *     r = _comb_int_long(N, k)             # <<<<<<<<<<<<<<
 *     if r != 0:
 *         return r
 */
  __pyx_v_r = __pyx_f_9cyext_acv__comb_int_long(__pyx_v_N, __pyx_v_k, 0);

  /* "cyext_acv.pyx":3997
 *     cdef double r
 *     r = _comb_int_long(N, k)
 *     if r != 0:             # <<<<<<<<<<<<<<
 *         return r
 * 
 */
  __pyx_t_1 = ((__pyx_v_r != 0.0) != 0);
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":3998
 *     r = _comb_int_long(N, k)
 *     if r != 0:
 *         return r             # <<<<<<<<<<<<<<
 * 
 * cpdef double _comb_int_long(unsigned long N, unsigned long k) nogil:
 */
    __pyx_r = __pyx_v_r;
    goto __pyx_L0;

    /* "cyext_acv.pyx":3997
 *     cdef double r
 *     r = _comb_int_long(N, k)
 *     if r != 0:             # <<<<<<<<<<<<<<
 *         return r
 * 
 */
  }

  /* "cyext_acv.pyx":3994
 * 
 * 
 * cpdef double binomialC(unsigned long N, unsigned long k) nogil:             # <<<<<<<<<<<<<<
 *     cdef double r
 *     r = _comb_int_long(N, k)
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_45binomialC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_45binomialC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_N;
  unsigned long __pyx_v_k;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binomialC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_N,&__pyx_n_s_k,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_N)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("binomialC", 1, 2, 2, 1); __PYX_ERR(0, 3994, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "binomialC") < 0)) __PYX_ERR(0, 3994, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_N = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_N == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 3994, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_k == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 3994, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binomialC", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3994, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.binomialC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_44binomialC(__pyx_self, __pyx_v_N, __pyx_v_k);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_44binomialC(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_N, unsigned long __pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binomialC", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_9cyext_acv_binomialC(__pyx_v_N, __pyx_v_k, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.binomialC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4000
 *         return r
 * 
 * cpdef double _comb_int_long(unsigned long N, unsigned long k) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Compute binom(N, k) for integers.
 */

static PyObject *__pyx_pw_9cyext_acv_47_comb_int_long(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_9cyext_acv__comb_int_long(unsigned long __pyx_v_N, unsigned long __pyx_v_k, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_val;
  unsigned PY_LONG_LONG __pyx_v_j;
  unsigned PY_LONG_LONG __pyx_v_M;
  unsigned PY_LONG_LONG __pyx_v_nterms;
  double __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  unsigned long __pyx_t_3;
  unsigned long __pyx_t_4;
  unsigned long __pyx_t_5;
  unsigned PY_LONG_LONG __pyx_t_6;
  unsigned PY_LONG_LONG __pyx_t_7;
  unsigned PY_LONG_LONG __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4008
 *     cdef unsigned long long j, M, nterms
 * 
 *     if k > N or N == ULONG_MAX:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_k > __pyx_v_N) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_N == ULONG_MAX) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "cyext_acv.pyx":4009
 * 
 *     if k > N or N == ULONG_MAX:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     M = N + 1
 */
    __pyx_r = 0.0;
    goto __pyx_L0;

    /* "cyext_acv.pyx":4008
 *     cdef unsigned long long j, M, nterms
 * 
 *     if k > N or N == ULONG_MAX:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "cyext_acv.pyx":4011
 *         return 0
 * 
 *     M = N + 1             # <<<<<<<<<<<<<<
 *     nterms = min(k, N - k)
 * 
 */
  __pyx_v_M = (__pyx_v_N + 1);

  /* "cyext_acv.pyx":4012
 * 
 *     M = N + 1
 *     nterms = min(k, N - k)             # <<<<<<<<<<<<<<
 * 
 *     val = 1
 */
  __pyx_t_3 = (__pyx_v_N - __pyx_v_k);
  __pyx_t_4 = __pyx_v_k;
  if (((__pyx_t_3 < __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_nterms = __pyx_t_5;

  /* "cyext_acv.pyx":4014
 *     nterms = min(k, N - k)
 * 
 *     val = 1             # <<<<<<<<<<<<<<
 * 
 *     for j in range(1, nterms + 1):
 */
  __pyx_v_val = 1.0;

  /* "cyext_acv.pyx":4016
 *     val = 1
 * 
 *     for j in range(1, nterms + 1):             # <<<<<<<<<<<<<<
 *         # Overflow check
 *         # if val > ULONG_MAX // (M - j):
 */
  __pyx_t_6 = (__pyx_v_nterms + 1);
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 1; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_j = __pyx_t_8;

    /* "cyext_acv.pyx":4021
 *         #   return 0
 * 
 *         val *= M - j             # <<<<<<<<<<<<<<
 *         val //= j
 * 
 */
    __pyx_v_val = (__pyx_v_val * (__pyx_v_M - __pyx_v_j));

    /* "cyext_acv.pyx":4022
 * 
 *         val *= M - j
 *         val //= j             # <<<<<<<<<<<<<<
 * 
 *     return val
 */
    if (unlikely(__pyx_v_j == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4022, __pyx_L1_error)
    }
    __pyx_v_val = floor(__pyx_v_val / __pyx_v_j);
  }

  /* "cyext_acv.pyx":4024
 *         val //= j
 * 
 *     return val             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4000
 *         return r
 * 
 * cpdef double _comb_int_long(unsigned long N, unsigned long k) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Compute binom(N, k) for integers.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv._comb_int_long", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_47_comb_int_long(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9cyext_acv_46_comb_int_long[] = "\n    Compute binom(N, k) for integers.\n    Returns 0 if error/overflow encountered.\n    ";
static PyObject *__pyx_pw_9cyext_acv_47_comb_int_long(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_N;
  unsigned long __pyx_v_k;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_comb_int_long (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_N,&__pyx_n_s_k,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_N)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_comb_int_long", 1, 2, 2, 1); __PYX_ERR(0, 4000, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_comb_int_long") < 0)) __PYX_ERR(0, 4000, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_N = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_N == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 4000, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_k == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 4000, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_comb_int_long", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4000, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv._comb_int_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_46_comb_int_long(__pyx_self, __pyx_v_N, __pyx_v_k);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_46_comb_int_long(CYTHON_UNUSED PyObject *__pyx_self, unsigned long __pyx_v_N, unsigned long __pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_comb_int_long", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_9cyext_acv__comb_int_long(__pyx_v_N, __pyx_v_k, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv._comb_int_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4030
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rf_v0(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rf_v0(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  std::vector<int>  __pyx_v_in_data;
  std::vector<int>  __pyx_v_in_data_b;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  std::vector<int> ::size_type __pyx_t_9;
  std::vector<int> ::size_type __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  size_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  std::vector<int> ::size_type __pyx_t_17;
  std::vector<int> ::size_type __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4034
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4035
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     sdp = 0
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4037
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int b, level, it_node, i
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4042
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4043
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4044
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4044, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4045
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         in_data.clear()
 *         in_data_b.clear()
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4046
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4047
 *         nodes_child.clear()
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.push_back(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4048
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4049
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.push_back(i)             # <<<<<<<<<<<<<<
 *             in_data_b.push_back(i)
 * 
 */
      try {
        __pyx_v_in_data.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4049, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4050
 *         for i in range(N):
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4050, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4052
 *             in_data_b.push_back(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 * 
 */
    __pyx_t_7 = __pyx_v_max_depth;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_8; __pyx_t_4+=1) {
      __pyx_v_level = __pyx_t_4;

      /* "cyext_acv.pyx":4053
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_9 = __pyx_v_nodes_level.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_10; __pyx_t_5+=1) {
        __pyx_v_it_node = __pyx_t_5;

        /* "cyext_acv.pyx":4055
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_11 = __pyx_v_b;
        __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_13 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4056
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )));
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_14 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":4057
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_16 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4057, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4059
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_17 = __pyx_v_in_data.size();
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_18; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4060
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_12 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_19 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4061
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4062
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4060
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4063
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4056
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4066
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_19 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4066, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4067
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_17 = __pyx_v_in_data.size();
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_18; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4068
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_19 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_16 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4069
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4070
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4068
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4071
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4055
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4074
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_12 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4074, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4075
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4075, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4077
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_13 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4078
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4077
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4080
 *                     break
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4082
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
    __pyx_t_13 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_13) {

      /* "cyext_acv.pyx":4083
 * 
 *         if classifier == 1:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *         else:
 */
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":4084
 *         if classifier == 1:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(in_data.size()):
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))))));
      }

      /* "cyext_acv.pyx":4082
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
      goto __pyx_L20;
    }

    /* "cyext_acv.pyx":4086
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *         else:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":4087
 *         else:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_t_19 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_19)) ))))) <= __pyx_v_t)));
      }
    }
    __pyx_L20:;
  }

  /* "cyext_acv.pyx":4089
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4030
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rf_v0(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":4095
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[int] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_49compute_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf_v0(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf_v0", 0);

  /* "cyext_acv.pyx":4099
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4100
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4102
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf_v0(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_6 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_8 = (__pyx_t_6 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_8 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_10, __pyx_t_11) firstprivate(__pyx_t_9)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_8; __pyx_t_7++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_7);

                            /* "cyext_acv.pyx":4103
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rf_v0(X[i], y_X[i], data, y_data, S,             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 */
                            __pyx_t_9.data = __pyx_v_X.data;
                            __pyx_t_9.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_X.shape[1];
__pyx_t_9.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_10 = __pyx_v_i;

                            /* "cyext_acv.pyx":4105
 *             sdp[i] = single_compute_sdp_rf_v0(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
                            __pyx_t_11 = __pyx_v_i;
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_11)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rf_v0(__pyx_t_9, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_10)) ))), __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                            __pyx_t_9.memview = NULL;
                            __pyx_t_9.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":4102
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf_v0(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":4106
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4095
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[int] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_49compute_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_49compute_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<int>  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rf_v0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 1); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 2); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 3); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 4); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 5); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 6); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 7); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 8); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 9); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 10); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 11); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, 12); __PYX_ERR(0, 4095, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rf_v0") < 0)) __PYX_ERR(0, 4095, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4095, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4095, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4095, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4095, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_int(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4095, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4096, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4096, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4096, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4096, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4097, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4097, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4097, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4097, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_v0", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4095, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_48compute_sdp_rf_v0(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_48compute_sdp_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf_v0", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4095, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rf_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4113
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

static PyObject *__pyx_pw_9cyext_acv_51global_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_rf_v0(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_ba = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_in_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_max_size;
  int __pyx_v_size;
  std::vector<int>  __pyx_v_S;
  std::vector<int>  __pyx_v_len_s_star;
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_X_arr = NULL;
  PyObject *__pyx_v_y_X_arr = NULL;
  PyObject *__pyx_v_remove_va = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  std::vector<int>  __pyx_t_7;
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  long __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_26;
  __Pyx_memviewslice __pyx_t_27 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_28;
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<long> ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  __Pyx_memviewslice __pyx_t_34 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_35 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  PyObject *__pyx_t_41 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_rf_v0", 0);

  /* "cyext_acv.pyx":4118
 *             int minimal, bint stop, list search_space):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4119
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp, sdp_b, sdp_ba
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":4124
 *     cdef double[:] sdp_global
 *     cdef double[:] in_data
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4125
 *     cdef double[:] in_data
 *     sdp = np.zeros((N))
 *     sdp_b = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 *     in_data = np.zeros(N)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_b = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4126
 *     sdp = np.zeros((N))
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 *     in_data = np.zeros(N)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4127
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))
 *     in_data = np.zeros(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, a_it, b_it, o_all, p, p_s, nb_leaf, p_u, p_d, p_su, p_sd, down, up
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_in_data = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4134
 * 
 *     cdef vector[int] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_len_s_star = __pyx_t_7;

  /* "cyext_acv.pyx":4139
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 4139, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":4140
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_8 = __pyx_v_N;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "cyext_acv.pyx":4141
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":4142
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 4142, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":4145
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     X_arr = np.array(X, dtype=np.double)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_R_buf = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":4147
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     X_arr = np.array(X, dtype=np.double)             # <<<<<<<<<<<<<<
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_X, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_double); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_X_arr = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":4148
 * 
 *     X_arr = np.array(X, dtype=np.double)
 *     y_X_arr = np.array(y_X, dtype=np.double)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_y_X, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_double); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_y_X_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4150
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 4150, __pyx_L1_error)
  }
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_12) {

    /* "cyext_acv.pyx":4151
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 4151, __pyx_L1_error)
    }
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4151, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_ci = __pyx_t_15;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 4151, __pyx_L1_error)
      }
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_16 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4151, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_17 = __pyx_t_16;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_cj = __pyx_t_18;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4151, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 4151, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":4152
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]             # <<<<<<<<<<<<<<
 *         for ci in range(len(C)):
 *             i = 0
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 4152, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
    for (;;) {
      if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4152, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_i = __pyx_t_10;
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = (__pyx_t_12 != 0);
      if (__pyx_t_19) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4152, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4152, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 4152, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":4153
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):             # <<<<<<<<<<<<<<
 *             i = 0
 *             for cj in range(len(C[ci])):
 */
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 4153, __pyx_L1_error)
    }
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4153, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_ci = __pyx_t_15;

      /* "cyext_acv.pyx":4154
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 *             i = 0             # <<<<<<<<<<<<<<
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 */
      __pyx_v_i = 0;

      /* "cyext_acv.pyx":4155
 *         for ci in range(len(C)):
 *             i = 0
 *             for cj in range(len(C[ci])):             # <<<<<<<<<<<<<<
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 */
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 4155, __pyx_L1_error)
      }
      __pyx_t_2 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_16 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4155, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_17 = __pyx_t_16;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_cj = __pyx_t_18;

        /* "cyext_acv.pyx":4156
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4156, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_19 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_search_space, Py_EQ)); if (unlikely(__pyx_t_19 < 0)) __PYX_ERR(0, 4156, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = (__pyx_t_19 != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":4157
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     break
 *             if i != 0:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "cyext_acv.pyx":4158
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 *                     break             # <<<<<<<<<<<<<<
 *             if i != 0:
 *                 va_id += [C[ci]]
 */
          goto __pyx_L16_break;

          /* "cyext_acv.pyx":4156
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        }
      }
      __pyx_L16_break:;

      /* "cyext_acv.pyx":4159
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      __pyx_t_12 = ((__pyx_v_i != 0) != 0);
      if (__pyx_t_12) {

        /* "cyext_acv.pyx":4160
 *                     break
 *             if i != 0:
 *                 va_id += [C[ci]]             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in search_space]
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4160, __pyx_L1_error)
        }
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        PyList_SET_ITEM(__pyx_t_2, 0, PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_va_id, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v_va_id, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "cyext_acv.pyx":4159
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      }
    }

    /* "cyext_acv.pyx":4150
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":4162
 *                 va_id += [C[ci]]
 *     else:
 *         va_id = [[i] for i in search_space]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 4162, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
    for (;;) {
      if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_4); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4162, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4162, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_i = __pyx_t_10;
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 4162, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":4164
 *         va_id = [[i] for i in search_space]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_13 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4164, __pyx_L1_error)
  __pyx_v_m = __pyx_t_13;

  /* "cyext_acv.pyx":4165
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_power = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4166
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":4167
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_20 = (__pyx_v_m + 1);
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_21; __pyx_t_10+=1) {
    __pyx_v_size = __pyx_t_10;

    /* "cyext_acv.pyx":4168
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":4169
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_combinations); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_22 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_22 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_22); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_22, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_22, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
      __pyx_t_23 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_23 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4169, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_23)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4169, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4169, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4169, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_23(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4169, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "cyext_acv.pyx":4170
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_4);
      __pyx_t_6 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_24, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_24 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_24)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_24);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_24, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_1); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 4170, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyext_acv.pyx":4171
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":4169
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyext_acv.pyx":4172
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 4172, __pyx_L1_error)

    /* "cyext_acv.pyx":4173
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_12 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_12) {

      /* "cyext_acv.pyx":4174
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L22_break;

      /* "cyext_acv.pyx":4173
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L22_break:;

  /* "cyext_acv.pyx":4176
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_26 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4176, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_26;

  /* "cyext_acv.pyx":4178
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_24) < 0) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_27 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_27.memview)) __PYX_ERR(0, 4178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_s_star = __pyx_t_27;
  __pyx_t_27.memview = NULL;
  __pyx_t_27.data = NULL;

  /* "cyext_acv.pyx":4181
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":4182
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_S = __pyx_t_7;

  /* "cyext_acv.pyx":4184
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_24);
  __pyx_t_2 = 0;
  __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_1, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_24);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_13 = 0;
    __pyx_t_23 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_23 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4184, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_23)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4184, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 4184, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_23(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4184, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4184, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_s_0 = __pyx_t_10;

    /* "cyext_acv.pyx":4185
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_28 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_29; __pyx_t_10+=1) {
      __pyx_v_s_1 = __pyx_t_10;

      /* "cyext_acv.pyx":4186
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_30 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_31; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":4187
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        (__pyx_v_S[__pyx_v_i]) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":4189
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":4190
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             in_data = np.zeros(X.shape[0])
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":4191
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             in_data = np.zeros(X.shape[0])
 * 
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":4192
 *             r.clear()
 *             N = R.size()
 *             in_data = np.zeros(X.shape[0])             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_np); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_24 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_24);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4192, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_in_data, 1);
      __pyx_v_in_data = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":4194
 *             in_data = np.zeros(X.shape[0])
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 *                 in_data[R_buf[i]] = 1
 */
      __pyx_t_8 = __pyx_v_N;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_9; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":4195
 * 
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 *                 in_data[R_buf[i]] = 1
 * 
 */
        __pyx_t_32 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);

        /* "cyext_acv.pyx":4196
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 *                 in_data[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 * 
 *                 # sdp_b[R_buf[i]] = single_compute_sdp_rf_v0(X[R_buf[i]],  y_X[R_buf[i]], data, y_data, S[:S_size], features, thresholds,  children_left, children_right,
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
        *((double *) ( /* dim=0 */ (__pyx_v_in_data.data + __pyx_t_33 * __pyx_v_in_data.strides[0]) )) = 1.0;
      }

      /* "cyext_acv.pyx":4201
 *                 #                 max_depth, min_node_size, classifier, t)
 * 
 *             sdp_ba = compute_sdp_rf_v0(X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_in_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, ((PyObject*)&PyBool_Type)) < 0) __PYX_ERR(0, 4201, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_X_arr, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_34 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_34.memview)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyext_acv.pyx":4202
 * 
 *             sdp_ba = compute_sdp_rf_v0(X_arr[np.array(in_data, dtype=bool)],
 *                                     y_X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_in_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, ((PyObject*)&PyBool_Type)) < 0) __PYX_ERR(0, 4202, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_y_X_arr, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_35 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_35.memview)) __PYX_ERR(0, 4202, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyext_acv.pyx":4203
 *             sdp_ba = compute_sdp_rf_v0(X_arr[np.array(in_data, dtype=bool)],
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_S); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, __pyx_v_S_size, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4203, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyext_acv.pyx":4201
 *                 #                 max_depth, min_node_size, classifier, t)
 * 
 *             sdp_ba = compute_sdp_rf_v0(X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 */
      __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rf_v0(__pyx_t_34, __pyx_t_35, __pyx_v_data, __pyx_v_y_data, __pyx_t_7, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __PYX_XDEC_MEMVIEW(&__pyx_t_34, 1);
      __pyx_t_34.memview = NULL;
      __pyx_t_34.data = NULL;
      __PYX_XDEC_MEMVIEW(&__pyx_t_35, 1);
      __pyx_t_35.memview = NULL;
      __pyx_t_35.data = NULL;
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4201, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_ba, 1);
      __pyx_v_sdp_ba = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":4206
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 */
      __pyx_t_8 = __pyx_v_N;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_9; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":4207
 * 
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]             # <<<<<<<<<<<<<<
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = __pyx_v_i;
        __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
        *((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_36 * __pyx_v_sdp_b.strides[0]) )) = (*((double *) ( /* dim=0 */ (__pyx_v_sdp_ba.data + __pyx_t_32 * __pyx_v_sdp_ba.strides[0]) )));

        /* "cyext_acv.pyx":4208
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_36 = __pyx_v_i;
        __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_12 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_33 * __pyx_v_sdp_b.strides[0]) ))) >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_37 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":4209
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_32 = __pyx_v_i;
          __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_33 * __pyx_v_sdp.strides[0]) )) = (*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_37 * __pyx_v_sdp_b.strides[0]) )));

          /* "cyext_acv.pyx":4210
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_36 = __pyx_v_i;
          (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]) = __pyx_v_S_size;

          /* "cyext_acv.pyx":4211
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_38 = __pyx_v_S_size;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_s = __pyx_t_40;

            /* "cyext_acv.pyx":4212
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_32 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_37 * __pyx_v_s_star.strides[0]) ) + __pyx_t_32 * __pyx_v_s_star.strides[1]) )) = (__pyx_v_S[__pyx_v_s]);
          }

          /* "cyext_acv.pyx":4208
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":4214
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_12 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":4215
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":4216
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_36 = __pyx_v_i;
          (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]) = __pyx_v_S_size;

          /* "cyext_acv.pyx":4217
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_38 = __pyx_v_S_size;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_s = __pyx_t_40;

            /* "cyext_acv.pyx":4218
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_37 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_32 * __pyx_v_s_star.strides[0]) ) + __pyx_t_37 * __pyx_v_s_star.strides[1]) )) = (__pyx_v_S[__pyx_v_s]);
          }

          /* "cyext_acv.pyx":4219
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_14 = (__pyx_v_X.shape[1]);
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_15 = __pyx_t_14;
          for (__pyx_t_38 = (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]); __pyx_t_38 < __pyx_t_15; __pyx_t_38+=1) {
            __pyx_v_s = __pyx_t_38;

            /* "cyext_acv.pyx":4220
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_37 = __pyx_v_i;
            __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_37 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_33 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_32 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":4214
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":4222
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_8 = __pyx_v_N;
    __pyx_t_9 = __pyx_t_8;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "cyext_acv.pyx":4223
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_36 = __pyx_v_i;
      __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_12 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_37 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_12) {

        /* "cyext_acv.pyx":4224
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 4224, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":4225
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_36 = __pyx_v_i;
        __pyx_t_22 = (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]);
        __pyx_t_38 = __pyx_t_22;
        for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
          __pyx_v_s = __pyx_t_39;

          /* "cyext_acv.pyx":4226
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_33 = __pyx_v_s;
          __pyx_t_32 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_37 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_32 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":4223
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":4228
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_30 = __pyx_v_r.size();
    __pyx_t_31 = __pyx_t_30;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_31; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "cyext_acv.pyx":4229
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":4230
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":4232
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_19 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_19) {
    } else {
      goto __pyx_L53_next_and;
    }
    __pyx_t_19 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_19) {
    } else {
      __pyx_t_12 = __pyx_t_19;
      goto __pyx_L52_bool_binop_done;
    }
    __pyx_L53_next_and:;
    __pyx_t_19 = (__pyx_v_stop != 0);
    __pyx_t_12 = __pyx_t_19;
    __pyx_L52_bool_binop_done:;
    if (__pyx_t_12) {

      /* "cyext_acv.pyx":4233
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L27_break;

      /* "cyext_acv.pyx":4232
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":4184
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L27_break:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cyext_acv.pyx":4235
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_24)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_24, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_len_s_star); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_41) < 0) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_24))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_24);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_24);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_24, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_24, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_24, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = PyTuple_New(4); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_24, 2, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_24, 3, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_t_41 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_24;
  __pyx_t_24 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4113
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_24);
  __PYX_XDEC_MEMVIEW(&__pyx_t_27, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_34, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_35, 1);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_ba, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_in_data, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_X_arr);
  __Pyx_XDECREF(__pyx_v_y_X_arr);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_51global_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_51global_sdp_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  PyObject *__pyx_v_search_space = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_rf_v0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,&__pyx_n_s_search_space,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 1); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 2); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 3); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 4); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 5); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 6); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 7); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 8); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 9); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 10); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 11); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 12); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 13); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 14); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 15); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_space)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, 16); __PYX_ERR(0, 4113, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_rf_v0") < 0)) __PYX_ERR(0, 4113, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4113, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4113, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4113, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4113, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4114, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4114, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4114, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4114, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4115, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4115, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4115, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4115, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[12]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4115, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4116, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[15]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4116, __pyx_L3_error)
    __pyx_v_search_space = ((PyObject*)values[16]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_rf_v0", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4113, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 4115, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_search_space), (&PyList_Type), 1, "search_space", 1))) __PYX_ERR(0, 4116, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_50global_sdp_rf_v0(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_50global_sdp_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_rf_v0", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4113, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_rf_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4242
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_exp_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_exp_rf(__Pyx_memviewslice &__pyx_v_x, CYTHON_UNUSED double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, CYTHON_UNUSED double &__pyx_v_t) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  int __pyx_v_b;
  CYTHON_UNUSED int __pyx_v_level;
  int __pyx_v_i;
  int __pyx_v_it_node;
  std::set<int>  __pyx_v_nodes_level;
  std::set<int>  __pyx_v_nodes_child;
  std::set<int>  __pyx_v_in_data;
  std::set<int>  __pyx_v_in_data_b;
  std::set<int> ::iterator __pyx_v_it;
  std::set<int> ::iterator __pyx_v_it_point;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4246
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4247
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4250
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int b, level, i, it_node
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4256
 *     cdef set[int].iterator it, it_point
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         nodes_level.clear()
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4257
 * 
 *     for b in range(n_trees):
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.insert(0)
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4258
 *     for b in range(n_trees):
 *         nodes_child.clear()
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.insert(0)
 * 
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4259
 *         nodes_child.clear()
 *         nodes_level.clear()
 *         nodes_level.insert(0)             # <<<<<<<<<<<<<<
 * 
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.insert(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4259, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4261
 *         nodes_level.insert(0)
 * 
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4262
 * 
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.insert(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4263
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.insert(i)
 *             in_data_b.insert(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4264
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.insert(i)             # <<<<<<<<<<<<<<
 *             in_data_b.insert(i)
 * 
 */
      try {
        __pyx_v_in_data.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4264, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4265
 *         for i in range(N):
 *             in_data.insert(i)
 *             in_data_b.insert(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4265, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4267
 *             in_data_b.insert(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 */
    __pyx_t_6 = __pyx_v_max_depth;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_level = __pyx_t_8;

      /* "cyext_acv.pyx":4268
 * 
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()             # <<<<<<<<<<<<<<
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 */
      __pyx_v_it_point = __pyx_v_nodes_level.begin();

      /* "cyext_acv.pyx":4269
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):             # <<<<<<<<<<<<<<
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it_point != __pyx_v_nodes_level.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":4270
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 */
        __pyx_v_it_node = (*__pyx_v_it_point);

        /* "cyext_acv.pyx":4271
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
        __pyx_t_10 = __pyx_v_b;
        __pyx_t_11 = __pyx_v_it_node;
        __pyx_t_9 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":4272
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_14 = __pyx_v_it_node;
          __pyx_t_9 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_12 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_9) {

            /* "cyext_acv.pyx":4273
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_14 * __pyx_v_children_left.strides[0]) ) + __pyx_t_13 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4273, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4275
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":4276
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":4277
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_13 = __pyx_v_b;
              __pyx_t_14 = __pyx_v_it_node;
              __pyx_t_10 = (*__pyx_v_it);
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_13 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_10 * __pyx_v_data.strides[0]) ) + __pyx_t_11 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_12 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_15 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":4278
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":4277
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":4279
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":4280
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4272
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4283
 * 
 *                     else:
 *                         nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
          /*else*/ {
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_12 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4283, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4285
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":4286
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":4287
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_14 = (*__pyx_v_it);
              __pyx_t_13 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_12 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_10 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_14 * __pyx_v_data.strides[0]) ) + __pyx_t_13 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_10 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":4288
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":4287
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":4289
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 *                 else:
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":4290
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4271
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4292
 *                         in_data = in_data_b
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 */
        /*else*/ {
          __pyx_t_10 = __pyx_v_b;
          __pyx_t_11 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_10 * __pyx_v_children_left.strides[0]) ) + __pyx_t_11 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4292, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4293
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 *                      nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_10 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4293, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4295
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        __pyx_t_9 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":4296
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 *                 inc(it_point)
 * 
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4295
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        }

        /* "cyext_acv.pyx":4297
 *                 if in_data.size() < min_node_size:
 *                     break
 *                 inc(it_point)             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
        (void)((++__pyx_v_it_point));
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4299
 *                 inc(it_point)
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4301
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
    __pyx_t_9 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_9) {

      /* "cyext_acv.pyx":4302
 * 
 *         if classifier == 1:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 */
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":4303
 *         if classifier == 1:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":4304
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]             # <<<<<<<<<<<<<<
 *                 inc(it)
 *         else:
 */
        __pyx_t_10 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_10)) )))));

        /* "cyext_acv.pyx":4305
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 *                 inc(it)             # <<<<<<<<<<<<<<
 *         else:
 *             it = in_data.begin()
 */
        (void)((++__pyx_v_it));
      }

      /* "cyext_acv.pyx":4301
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
      goto __pyx_L20;
    }

    /* "cyext_acv.pyx":4307
 *                 inc(it)
 *         else:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 */
    /*else*/ {
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":4308
 *         else:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":4309
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]             # <<<<<<<<<<<<<<
 *                 inc(it)
 * 
 */
        __pyx_t_10 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_10)) )))));

        /* "cyext_acv.pyx":4310
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * y_data[deref(it)]
 *                 inc(it)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        (void)((++__pyx_v_it));
      }
    }
    __pyx_L20:;
  }

  /* "cyext_acv.pyx":4312
 *                 inc(it)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4242
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_exp_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_exp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":4318
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_53compute_exp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_exp_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_rf", 0);

  /* "cyext_acv.pyx":4322
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4323
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4325
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_exp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_6 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_8 = (__pyx_t_6 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_8 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_10, __pyx_t_11) firstprivate(__pyx_t_9)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_8; __pyx_t_7++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_7);

                            /* "cyext_acv.pyx":4326
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_exp_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 */
                            __pyx_t_9.data = __pyx_v_X.data;
                            __pyx_t_9.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_X.shape[1];
__pyx_t_9.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_10 = __pyx_v_i;

                            /* "cyext_acv.pyx":4328
 *             sdp[i] = single_compute_exp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
                            __pyx_t_11 = __pyx_v_i;
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_11)) )) = __pyx_f_9cyext_acv_single_compute_exp_rf(__pyx_t_9, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_10)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                            __pyx_t_9.memview = NULL;
                            __pyx_t_9.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":4325
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_exp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":4329
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4318
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_exp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_53compute_exp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_53compute_exp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_exp_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 1); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 2); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 3); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 4); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 5); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 6); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 7); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 8); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 9); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 10); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 11); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, 12); __PYX_ERR(0, 4318, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_exp_rf") < 0)) __PYX_ERR(0, 4318, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4318, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4318, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4318, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4318, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4318, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4319, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4319, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4319, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4319, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4320, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4320, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4320, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4320, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_exp_rf", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_exp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_52compute_exp_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_52compute_exp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_exp_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4318, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_exp_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_exp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4335
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_quantile_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 */

static double __pyx_f_9cyext_acv_single_compute_quantile_rf(__Pyx_memviewslice &__pyx_v_x, CYTHON_UNUSED double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, CYTHON_UNUSED int &__pyx_v_classifier, CYTHON_UNUSED double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_weights, __Pyx_memviewslice &__pyx_v_samples, __Pyx_memviewslice &__pyx_v_samples_child, double &__pyx_v_quantile) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_s;
  CYTHON_UNUSED double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  PyObject *__pyx_v_weights_py = NULL;
  PyObject *__pyx_v_sorter = NULL;
  PyObject *__pyx_v_y_quantile = NULL;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  std::vector<int> ::size_type __pyx_t_10;
  std::vector<int> ::size_type __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  size_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  unsigned int __pyx_t_17;
  unsigned int __pyx_t_18;
  size_t __pyx_t_19;
  size_t __pyx_t_20;
  size_t __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  double __pyx_t_28;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_compute_quantile_rf", 0);

  /* "cyext_acv.pyx":4339
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4340
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     sdp = 0
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4342
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4348
 *     cdef vector[int] nodes_level, nodes_child
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4349
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4350
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 * 
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 4350, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4351
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4353
 *         nodes_child.clear()
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             samples[i] = 1
 *             samples_child[i] = 1
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4354
 * 
 *         for i in range(N):
 *             samples[i] = 1             # <<<<<<<<<<<<<<
 *             samples_child[i] = 1
 * 
 */
      __pyx_t_7 = __pyx_v_i;
      *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_7)) )) = 1.0;

      /* "cyext_acv.pyx":4355
 *         for i in range(N):
 *             samples[i] = 1
 *             samples_child[i] = 1             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      __pyx_t_7 = __pyx_v_i;
      *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_7)) )) = 1.0;
    }

    /* "cyext_acv.pyx":4357
 *             samples_child[i] = 1
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 *                 s = 0
 */
    __pyx_t_8 = __pyx_v_max_depth;
    __pyx_t_9 = __pyx_t_8;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_9; __pyx_t_4+=1) {
      __pyx_v_level = __pyx_t_4;

      /* "cyext_acv.pyx":4358
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 *                 s = 0
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_10 = __pyx_v_nodes_level.size();
      __pyx_t_11 = __pyx_t_10;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_11; __pyx_t_5+=1) {
        __pyx_v_it_node = __pyx_t_5;

        /* "cyext_acv.pyx":4359
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):
 *                 s = 0             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 */
        __pyx_v_s = 0.0;

        /* "cyext_acv.pyx":4360
 *             for it_node in range(nodes_level.size()):
 *                 s = 0
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_7 = __pyx_v_b;
        __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_13 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_7 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4361
 *                 s = 0
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_7 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_7 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )));
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_14 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":4362
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(N):
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_16 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4362, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4364
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         for i in range(N):             # <<<<<<<<<<<<<<
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]])
 *                             s += samples_child[i]
 */
            __pyx_t_6 = __pyx_v_N;
            __pyx_t_17 = __pyx_t_6;
            for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
              __pyx_v_i = __pyx_t_18;

              /* "cyext_acv.pyx":4365
 * 
 *                         for i in range(N):
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                             s += samples_child[i]
 *                     else:
 */
              __pyx_t_15 = __pyx_v_i;
              __pyx_t_7 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_7 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_20 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_21 = __pyx_v_i;
              *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_21)) )) = ((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_15)) ))) * ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_12 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_20 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))));

              /* "cyext_acv.pyx":4366
 *                         for i in range(N):
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]])
 *                             s += samples_child[i]             # <<<<<<<<<<<<<<
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 */
              __pyx_t_20 = __pyx_v_i;
              __pyx_v_s = (__pyx_v_s + (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_20)) ))));
            }

            /* "cyext_acv.pyx":4361
 *                 s = 0
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4368
 *                             s += samples_child[i]
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(N):
 */
          /*else*/ {
            __pyx_t_20 = __pyx_v_b;
            __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_20 * __pyx_v_children_right.strides[0]) ) + __pyx_t_14 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4368, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4370
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         for i in range(N):             # <<<<<<<<<<<<<<
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]])
 *                             s += samples_child[i]
 */
            __pyx_t_6 = __pyx_v_N;
            __pyx_t_17 = __pyx_t_6;
            for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
              __pyx_v_i = __pyx_t_18;

              /* "cyext_acv.pyx":4371
 * 
 *                         for i in range(N):
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                             s += samples_child[i]
 *                 else:
 */
              __pyx_t_20 = __pyx_v_i;
              __pyx_t_7 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_19 = __pyx_v_i;
              __pyx_t_16 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_7 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_21 = __pyx_v_i;
              *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_21)) )) = ((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_20)) ))) * ((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_19 * __pyx_v_data.strides[0]) ) + __pyx_t_16 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))));

              /* "cyext_acv.pyx":4372
 *                         for i in range(N):
 *                             samples_child[i] = samples[i] * (data[i, features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]])
 *                             s += samples_child[i]             # <<<<<<<<<<<<<<
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
              __pyx_t_15 = __pyx_v_i;
              __pyx_v_s = (__pyx_v_s + (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_15)) ))));
            }
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4360
 *             for it_node in range(nodes_level.size()):
 *                 s = 0
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4374
 *                             s += samples_child[i]
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_12 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 4374, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4375
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                      for i in range(N):
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 4375, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4377
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                      for i in range(N):             # <<<<<<<<<<<<<<
 *                         samples_child[i] = samples[i]
 *                         s += samples_child[i]
 */
          __pyx_t_6 = __pyx_v_N;
          __pyx_t_17 = __pyx_t_6;
          for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
            __pyx_v_i = __pyx_t_18;

            /* "cyext_acv.pyx":4378
 * 
 *                      for i in range(N):
 *                         samples_child[i] = samples[i]             # <<<<<<<<<<<<<<
 *                         s += samples_child[i]
 * 
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_7 = __pyx_v_i;
            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_7)) )) = (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_15)) )));

            /* "cyext_acv.pyx":4379
 *                      for i in range(N):
 *                         samples_child[i] = samples[i]
 *                         s += samples_child[i]             # <<<<<<<<<<<<<<
 * 
 *                 if s < min_node_size:
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_v_s = (__pyx_v_s + (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_15)) ))));
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4381
 *                         s += samples_child[i]
 * 
 *                 if s < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 else:
 */
        __pyx_t_13 = ((__pyx_v_s < __pyx_v_min_node_size) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4382
 * 
 *                 if s < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 *                 else:
 *                     for i in range(N):
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4381
 *                         s += samples_child[i]
 * 
 *                 if s < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 else:
 */
        }

        /* "cyext_acv.pyx":4384
 *                     break
 *                 else:
 *                     for i in range(N):             # <<<<<<<<<<<<<<
 *                         samples[i] = samples_child[i]
 * 
 */
        /*else*/ {
          __pyx_t_6 = __pyx_v_N;
          __pyx_t_17 = __pyx_t_6;
          for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
            __pyx_v_i = __pyx_t_18;

            /* "cyext_acv.pyx":4385
 *                 else:
 *                     for i in range(N):
 *                         samples[i] = samples_child[i]             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_7 = __pyx_v_i;
            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_7)) )) = (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples_child.data) + __pyx_t_15)) )));
          }
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4387
 *                         samples[i] = samples_child[i]
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         s = 0
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4389
 *             nodes_level = nodes_child
 * 
 *         s = 0             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             s += samples[i]
 */
    __pyx_v_s = 0.0;

    /* "cyext_acv.pyx":4390
 * 
 *         s = 0
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             s += samples[i]
 * 
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4391
 *         s = 0
 *         for i in range(N):
 *             s += samples[i]             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
      __pyx_t_15 = __pyx_v_i;
      __pyx_v_s = (__pyx_v_s + (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_15)) ))));
    }

    /* "cyext_acv.pyx":4393
 *             s += samples[i]
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             weights[i] += samples[i] * (1/s)
 * 
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4394
 * 
 *         for i in range(N):
 *             weights[i] += samples[i] * (1/s)             # <<<<<<<<<<<<<<
 * 
 *     weights_py = np.array(weights)/n_trees
 */
      __pyx_t_15 = __pyx_v_i;
      __pyx_t_7 = __pyx_v_i;
      *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_weights.data) + __pyx_t_7)) )) += ((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_samples.data) + __pyx_t_15)) ))) * (1.0 / __pyx_v_s));
    }
  }

  /* "cyext_acv.pyx":4396
 *             weights[i] += samples[i] * (1/s)
 * 
 *     weights_py = np.array(weights)/n_trees             # <<<<<<<<<<<<<<
 *     sorter = np.argsort(y_data)
 *     y_quantile = weighted_percentile(y_data, quantile, weights_py, sorter)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_np); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_n_s_array); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_memoryview_fromslice(__pyx_v_weights, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_25 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_24))) {
    __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_24);
    if (likely(__pyx_t_25)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_24);
      __Pyx_INCREF(__pyx_t_25);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_24, function);
    }
  }
  __pyx_t_22 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_24, __pyx_t_25, __pyx_t_23) : __Pyx_PyObject_CallOneArg(__pyx_t_24, __pyx_t_23);
  __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyInt_From_unsigned_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_23 = __Pyx_PyNumber_Divide(__pyx_t_22, __pyx_t_24); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_v_weights_py = __pyx_t_23;
  __pyx_t_23 = 0;

  /* "cyext_acv.pyx":4397
 * 
 *     weights_py = np.array(weights)/n_trees
 *     sorter = np.argsort(y_data)             # <<<<<<<<<<<<<<
 *     y_quantile = weighted_percentile(y_data, quantile, weights_py, sorter)
 *     return y_quantile
 */
  __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_np); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_argsort); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 4397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __pyx_memoryview_fromslice(__pyx_v_y_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_22))) {
    __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_22);
    if (likely(__pyx_t_25)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
      __Pyx_INCREF(__pyx_t_25);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_22, function);
    }
  }
  __pyx_t_23 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_22, __pyx_t_25, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_22, __pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_v_sorter = __pyx_t_23;
  __pyx_t_23 = 0;

  /* "cyext_acv.pyx":4398
 *     weights_py = np.array(weights)/n_trees
 *     sorter = np.argsort(y_data)
 *     y_quantile = weighted_percentile(y_data, quantile, weights_py, sorter)             # <<<<<<<<<<<<<<
 *     return y_quantile
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_weighted_percentile); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 4398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_24 = __pyx_memoryview_fromslice(__pyx_v_y_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 4398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = PyFloat_FromDouble(__pyx_v_quantile); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 4398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_26 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_22))) {
    __pyx_t_26 = PyMethod_GET_SELF(__pyx_t_22);
    if (likely(__pyx_t_26)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
      __Pyx_INCREF(__pyx_t_26);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_22, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_22)) {
    PyObject *__pyx_temp[5] = {__pyx_t_26, __pyx_t_24, __pyx_t_25, __pyx_v_weights_py, __pyx_v_sorter};
    __pyx_t_23 = __Pyx_PyFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4398, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_22)) {
    PyObject *__pyx_temp[5] = {__pyx_t_26, __pyx_t_24, __pyx_t_25, __pyx_v_weights_py, __pyx_v_sorter};
    __pyx_t_23 = __Pyx_PyCFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4398, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  } else
  #endif
  {
    __pyx_t_27 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 4398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    if (__pyx_t_26) {
      __Pyx_GIVEREF(__pyx_t_26); PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_26); __pyx_t_26 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_27, 0+__pyx_t_8, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_27, 1+__pyx_t_8, __pyx_t_25);
    __Pyx_INCREF(__pyx_v_weights_py);
    __Pyx_GIVEREF(__pyx_v_weights_py);
    PyTuple_SET_ITEM(__pyx_t_27, 2+__pyx_t_8, __pyx_v_weights_py);
    __Pyx_INCREF(__pyx_v_sorter);
    __Pyx_GIVEREF(__pyx_v_sorter);
    PyTuple_SET_ITEM(__pyx_t_27, 3+__pyx_t_8, __pyx_v_sorter);
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_27, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 4398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  }
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_v_y_quantile = __pyx_t_23;
  __pyx_t_23 = 0;

  /* "cyext_acv.pyx":4399
 *     sorter = np.argsort(y_data)
 *     y_quantile = weighted_percentile(y_data, quantile, weights_py, sorter)
 *     return y_quantile             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_t_28 = __pyx_PyFloat_AsDouble(__pyx_v_y_quantile); if (unlikely((__pyx_t_28 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4399, __pyx_L1_error)
  __pyx_r = __pyx_t_28;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4335
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_quantile_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_WriteUnraisable("cyext_acv.single_compute_quantile_rf", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_weights_py);
  __Pyx_XDECREF(__pyx_v_sorter);
  __Pyx_XDECREF(__pyx_v_y_quantile);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4405
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_quantile_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 */

static PyObject *__pyx_pw_9cyext_acv_55compute_quantile_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_quantile_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, double &__pyx_v_quantile, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_samples = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_samples_child = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_quantile_rf", 0);

  /* "cyext_acv.pyx":4409
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, ::1] weights, samples, samples_child
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4410
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4412
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         samples = np.ones((N, data.shape[0]))
 *         samples_child = np.ones((N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_weights = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4413
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))
 *         samples = np.ones((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         samples_child = np.ones((N, data.shape[0]))
 *         cdef int i
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_samples = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4414
 *         weights = np.zeros((N, data.shape[0]))
 *         samples = np.ones((N, data.shape[0]))
 *         samples_child = np.ones((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(N):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_samples_child = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4416
 *         samples_child = np.ones((N, data.shape[0]))
 *         cdef int i
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_quantile_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_8 = __pyx_v_N;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "cyext_acv.pyx":4417
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_quantile_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 */
    __pyx_t_11.data = __pyx_v_X.data;
    __pyx_t_11.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_11, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_11.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_11.shape[0] = __pyx_v_X.shape[1];
__pyx_t_11.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_11.suboffsets[0] = -1;

__pyx_t_12 = __pyx_v_i;

    /* "cyext_acv.pyx":4419
 *             sdp[i] = single_compute_quantile_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
    __pyx_t_5.data = __pyx_v_weights.data;
    __pyx_t_5.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_5.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_13.data = __pyx_v_samples.data;
    __pyx_t_13.memview = __pyx_v_samples.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_samples.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_13.shape[0] = __pyx_v_samples.shape[1];
__pyx_t_13.strides[0] = __pyx_v_samples.strides[1];
    __pyx_t_13.suboffsets[0] = -1;

__pyx_t_14.data = __pyx_v_samples_child.data;
    __pyx_t_14.memview = __pyx_v_samples_child.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_samples_child.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_samples_child.shape[1];
__pyx_t_14.strides[0] = __pyx_v_samples_child.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":4417
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_quantile_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_15)) )) = __pyx_f_9cyext_acv_single_compute_quantile_rf(__pyx_t_11, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_12)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_5, __pyx_t_13, __pyx_t_14, __pyx_v_quantile);
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
    __pyx_t_5.memview = NULL;
    __pyx_t_5.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
    __pyx_t_13.memview = NULL;
    __pyx_t_13.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
    __pyx_t_14.memview = NULL;
    __pyx_t_14.data = NULL;
  }

  /* "cyext_acv.pyx":4420
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4405
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_quantile_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __Pyx_AddTraceback("cyext_acv.compute_quantile_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_samples, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_samples_child, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_55compute_quantile_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_55compute_quantile_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_quantile;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_quantile_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_quantile,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 1); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 2); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 3); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 4); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 5); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 6); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 7); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 8); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 9); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 10); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 11); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 12); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quantile)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, 13); __PYX_ERR(0, 4405, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_quantile_rf") < 0)) __PYX_ERR(0, 4405, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4405, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4405, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4405, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4405, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4405, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4406, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4406, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4406, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4406, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4407, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4407, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4407, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4407, __pyx_L3_error)
    __pyx_v_quantile = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_quantile == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4407, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_quantile_rf", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4405, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_quantile_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_54compute_quantile_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_quantile);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_54compute_quantile_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_quantile) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_quantile_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4405, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_quantile_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_quantile, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_quantile_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4427
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_quantile_diff_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 */

static double __pyx_f_9cyext_acv_single_compute_quantile_diff_rf(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, CYTHON_UNUSED int &__pyx_v_classifier, CYTHON_UNUSED double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_weights, CYTHON_UNUSED __Pyx_memviewslice &__pyx_v_samples, CYTHON_UNUSED __Pyx_memviewslice &__pyx_v_samples_child, double &__pyx_v_quantile) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  CYTHON_UNUSED double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  std::vector<int>  __pyx_v_in_data;
  std::vector<int>  __pyx_v_in_data_b;
  __Pyx_memviewslice __pyx_v_local_diff = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_weights_py = NULL;
  PyObject *__pyx_v_sorter = NULL;
  PyObject *__pyx_v_y_quantile = NULL;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  unsigned int __pyx_t_10;
  unsigned int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  std::vector<int> ::size_type __pyx_t_14;
  std::vector<int> ::size_type __pyx_t_15;
  size_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  size_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  std::vector<int> ::size_type __pyx_t_22;
  std::vector<int> ::size_type __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  double __pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("single_compute_quantile_diff_rf", 0);

  /* "cyext_acv.pyx":4431
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4432
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     sdp = 0
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4434
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int b, level, it_node, i
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4439
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     cdef double[::1] local_diff = np.zeros(N)             # <<<<<<<<<<<<<<
 * 
 *     for b in range(n_trees):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_local_diff = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4441
 *     cdef double[::1] local_diff = np.zeros(N)
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_6 = __pyx_v_n_trees;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_b = __pyx_t_8;

    /* "cyext_acv.pyx":4442
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4443
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 4443, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4444
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         in_data.clear()
 *         in_data_b.clear()
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4445
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4446
 *         nodes_child.clear()
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.push_back(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4447
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)
 */
    __pyx_t_9 = __pyx_v_N;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "cyext_acv.pyx":4448
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.push_back(i)             # <<<<<<<<<<<<<<
 *             in_data_b.push_back(i)
 * 
 */
      try {
        __pyx_v_in_data.push_back(__pyx_v_i);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 4448, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4449
 *         for i in range(N):
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.push_back(__pyx_v_i);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 4449, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4451
 *             in_data_b.push_back(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 * 
 */
    __pyx_t_12 = __pyx_v_max_depth;
    __pyx_t_13 = __pyx_t_12;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_13; __pyx_t_9+=1) {
      __pyx_v_level = __pyx_t_9;

      /* "cyext_acv.pyx":4452
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_14 = __pyx_v_nodes_level.size();
      __pyx_t_15 = __pyx_t_14;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_15; __pyx_t_10+=1) {
        __pyx_v_it_node = __pyx_t_10;

        /* "cyext_acv.pyx":4454
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_16 = __pyx_v_b;
        __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_18 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_16 * __pyx_v_features.strides[0]) ) + __pyx_t_17 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_18) {

          /* "cyext_acv.pyx":4455
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_16 = __pyx_v_b;
          __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_19 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_16 * __pyx_v_features.strides[0]) ) + __pyx_t_17 * __pyx_v_features.strides[1]) )));
          __pyx_t_20 = __pyx_v_b;
          __pyx_t_21 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_18 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_19 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_20 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_21 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_18) {

            /* "cyext_acv.pyx":4456
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
            __pyx_t_20 = __pyx_v_b;
            __pyx_t_21 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_20 * __pyx_v_children_left.strides[0]) ) + __pyx_t_21 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4456, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4458
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_22 = __pyx_v_in_data.size();
            __pyx_t_23 = __pyx_t_22;
            for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_23; __pyx_t_11+=1) {
              __pyx_v_i = __pyx_t_11;

              /* "cyext_acv.pyx":4459
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_20 = __pyx_v_b;
              __pyx_t_21 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_17 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_19 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_20 * __pyx_v_features.strides[0]) ) + __pyx_t_21 * __pyx_v_features.strides[1]) )));
              __pyx_t_16 = __pyx_v_b;
              __pyx_t_24 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_18 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_17 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_16 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_24 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_18) {

                /* "cyext_acv.pyx":4460
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4461
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4459
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4462
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4455
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4465
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_16 = __pyx_v_b;
            __pyx_t_24 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_16 * __pyx_v_children_right.strides[0]) ) + __pyx_t_24 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4465, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4466
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_22 = __pyx_v_in_data.size();
            __pyx_t_23 = __pyx_t_22;
            for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_23; __pyx_t_11+=1) {
              __pyx_v_i = __pyx_t_11;

              /* "cyext_acv.pyx":4467
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_16 = __pyx_v_b;
              __pyx_t_24 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_21 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_19 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_16 * __pyx_v_features.strides[0]) ) + __pyx_t_24 * __pyx_v_features.strides[1]) )));
              __pyx_t_20 = __pyx_v_b;
              __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_18 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_21 * __pyx_v_data.strides[0]) ) + __pyx_t_19 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_20 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_17 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_18) {

                /* "cyext_acv.pyx":4468
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4469
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4467
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4470
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4454
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4473
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_20 = __pyx_v_b;
          __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_20 * __pyx_v_children_left.strides[0]) ) + __pyx_t_17 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 4473, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4474
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_20 = __pyx_v_b;
          __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_20 * __pyx_v_children_right.strides[0]) ) + __pyx_t_17 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 4474, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4476
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_18 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_18) {

          /* "cyext_acv.pyx":4477
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4476
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4479
 *                     break
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         for i in range(in_data.size()):
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4481
 *             nodes_level = nodes_child
 * 
 *         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *             weights[in_data[i]] +=  1./in_data.size()
 *             # local_diff[i] = (y_data[i] - y_x)*(y_data[i] - y_x)
 */
    __pyx_t_14 = __pyx_v_in_data.size();
    __pyx_t_15 = __pyx_t_14;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_15; __pyx_t_9+=1) {
      __pyx_v_i = __pyx_t_9;

      /* "cyext_acv.pyx":4482
 * 
 *         for i in range(in_data.size()):
 *             weights[in_data[i]] +=  1./in_data.size()             # <<<<<<<<<<<<<<
 *             # local_diff[i] = (y_data[i] - y_x)*(y_data[i] - y_x)
 *             local_diff[in_data[i]] = y_x - y_data[in_data[i]]
 */
      __pyx_t_17 = (__pyx_v_in_data[__pyx_v_i]);
      *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_weights.data) + __pyx_t_17)) )) += (1. / __pyx_v_in_data.size());

      /* "cyext_acv.pyx":4484
 *             weights[in_data[i]] +=  1./in_data.size()
 *             # local_diff[i] = (y_data[i] - y_x)*(y_data[i] - y_x)
 *             local_diff[in_data[i]] = y_x - y_data[in_data[i]]             # <<<<<<<<<<<<<<
 *     weights_py = np.array(weights)/n_trees
 *     sorter = np.argsort(local_diff)
 */
      __pyx_t_17 = (__pyx_v_in_data[__pyx_v_i]);
      __pyx_t_24 = (__pyx_v_in_data[__pyx_v_i]);
      *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_local_diff.data) + __pyx_t_24)) )) = (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_17)) ))));
    }
  }

  /* "cyext_acv.pyx":4485
 *             # local_diff[i] = (y_data[i] - y_x)*(y_data[i] - y_x)
 *             local_diff[in_data[i]] = y_x - y_data[in_data[i]]
 *     weights_py = np.array(weights)/n_trees             # <<<<<<<<<<<<<<
 *     sorter = np.argsort(local_diff)
 *     y_quantile = weighted_percentile(local_diff, quantile, weights_py, sorter)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_weights, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_n_trees); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_weights_py = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cyext_acv.pyx":4486
 *             local_diff[in_data[i]] = y_x - y_data[in_data[i]]
 *     weights_py = np.array(weights)/n_trees
 *     sorter = np.argsort(local_diff)             # <<<<<<<<<<<<<<
 *     y_quantile = weighted_percentile(local_diff, quantile, weights_py, sorter)
 *     return y_quantile
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_argsort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_local_diff, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sorter = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cyext_acv.pyx":4487
 *     weights_py = np.array(weights)/n_trees
 *     sorter = np.argsort(local_diff)
 *     y_quantile = weighted_percentile(local_diff, quantile, weights_py, sorter)             # <<<<<<<<<<<<<<
 *     return y_quantile
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_weighted_percentile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_local_diff, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_quantile); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_25 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_25)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_25);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_25, __pyx_t_2, __pyx_t_4, __pyx_v_weights_py, __pyx_v_sorter};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4487, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_25, __pyx_t_2, __pyx_t_4, __pyx_v_weights_py, __pyx_v_sorter};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4487, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_26 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 4487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__pyx_t_25) {
      __Pyx_GIVEREF(__pyx_t_25); PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_25); __pyx_t_25 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 0+__pyx_t_12, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_26, 1+__pyx_t_12, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_weights_py);
    __Pyx_GIVEREF(__pyx_v_weights_py);
    PyTuple_SET_ITEM(__pyx_t_26, 2+__pyx_t_12, __pyx_v_weights_py);
    __Pyx_INCREF(__pyx_v_sorter);
    __Pyx_GIVEREF(__pyx_v_sorter);
    PyTuple_SET_ITEM(__pyx_t_26, 3+__pyx_t_12, __pyx_v_sorter);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_y_quantile = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cyext_acv.pyx":4488
 *     sorter = np.argsort(local_diff)
 *     y_quantile = weighted_percentile(local_diff, quantile, weights_py, sorter)
 *     return y_quantile             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_t_27 = __pyx_PyFloat_AsDouble(__pyx_v_y_quantile); if (unlikely((__pyx_t_27 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4488, __pyx_L1_error)
  __pyx_r = __pyx_t_27;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4427
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_quantile_diff_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[::1] & weights, double[::1] & samples, double[::1] & samples_child, double & quantile):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_WriteUnraisable("cyext_acv.single_compute_quantile_diff_rf", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_local_diff, 1);
  __Pyx_XDECREF(__pyx_v_weights_py);
  __Pyx_XDECREF(__pyx_v_sorter);
  __Pyx_XDECREF(__pyx_v_y_quantile);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4494
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_quantile_diff_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 */

static PyObject *__pyx_pw_9cyext_acv_57compute_quantile_diff_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_quantile_diff_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, double &__pyx_v_quantile, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_samples = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_samples_child = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_quantile_diff_rf", 0);

  /* "cyext_acv.pyx":4498
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, ::1] weights, samples, samples_child
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4499
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4501
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         samples = np.ones((N, data.shape[0]))
 *         samples_child = np.ones((N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_weights = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4502
 *         cdef double[:, ::1] weights, samples, samples_child
 *         weights = np.zeros((N, data.shape[0]))
 *         samples = np.ones((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         samples_child = np.ones((N, data.shape[0]))
 *         cdef int i
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_samples = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4503
 *         weights = np.zeros((N, data.shape[0]))
 *         samples = np.ones((N, data.shape[0]))
 *         samples_child = np.ones((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(N):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ones); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_6 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_samples_child = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4505
 *         samples_child = np.ones((N, data.shape[0]))
 *         cdef int i
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_quantile_diff_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_8 = __pyx_v_N;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "cyext_acv.pyx":4506
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_quantile_diff_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 */
    __pyx_t_11.data = __pyx_v_X.data;
    __pyx_t_11.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_11, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_11.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_11.shape[0] = __pyx_v_X.shape[1];
__pyx_t_11.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_11.suboffsets[0] = -1;

__pyx_t_12 = __pyx_v_i;

    /* "cyext_acv.pyx":4508
 *             sdp[i] = single_compute_quantile_diff_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
    __pyx_t_5.data = __pyx_v_weights.data;
    __pyx_t_5.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_5, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_5.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_5.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_5.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_5.suboffsets[0] = -1;

__pyx_t_13.data = __pyx_v_samples.data;
    __pyx_t_13.memview = __pyx_v_samples.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_samples.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_13.shape[0] = __pyx_v_samples.shape[1];
__pyx_t_13.strides[0] = __pyx_v_samples.strides[1];
    __pyx_t_13.suboffsets[0] = -1;

__pyx_t_14.data = __pyx_v_samples_child.data;
    __pyx_t_14.memview = __pyx_v_samples_child.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_samples_child.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_samples_child.shape[1];
__pyx_t_14.strides[0] = __pyx_v_samples_child.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":4506
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_quantile_diff_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_15)) )) = __pyx_f_9cyext_acv_single_compute_quantile_diff_rf(__pyx_t_11, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_12)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_5, __pyx_t_13, __pyx_t_14, __pyx_v_quantile);
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
    __pyx_t_5.memview = NULL;
    __pyx_t_5.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
    __pyx_t_13.memview = NULL;
    __pyx_t_13.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
    __pyx_t_14.memview = NULL;
    __pyx_t_14.data = NULL;
  }

  /* "cyext_acv.pyx":4509
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, weights[i, :], samples[i, :], samples_child[i, :], quantile)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4494
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_quantile_diff_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & quantile):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __Pyx_AddTraceback("cyext_acv.compute_quantile_diff_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_samples, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_samples_child, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_57compute_quantile_diff_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_57compute_quantile_diff_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_quantile;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_quantile_diff_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_quantile,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 1); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 2); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 3); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 4); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 5); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 6); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 7); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 8); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 9); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 10); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 11); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 12); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quantile)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, 13); __PYX_ERR(0, 4494, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_quantile_diff_rf") < 0)) __PYX_ERR(0, 4494, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4494, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4494, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4494, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4494, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4494, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4495, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4495, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4495, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4495, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4496, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4496, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4496, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4496, __pyx_L3_error)
    __pyx_v_quantile = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_quantile == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4496, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_quantile_diff_rf", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4494, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_quantile_diff_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_56compute_quantile_diff_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_quantile);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_56compute_quantile_diff_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_quantile) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_quantile_diff_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4494, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_quantile_diff_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_quantile, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_quantile_diff_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4515
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_cdf_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_cdf_rf(__Pyx_memviewslice &__pyx_v_x, CYTHON_UNUSED double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, CYTHON_UNUSED int &__pyx_v_classifier, double &__pyx_v_t) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  std::vector<int>  __pyx_v_in_data;
  std::vector<int>  __pyx_v_in_data_b;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  std::vector<int> ::size_type __pyx_t_9;
  std::vector<int> ::size_type __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  size_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  std::vector<int> ::size_type __pyx_t_17;
  std::vector<int> ::size_type __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4519
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4520
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     sdp = 0
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4522
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int b, level, it_node, i
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4527
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4528
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4529
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4529, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4530
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         in_data.clear()
 *         in_data_b.clear()
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4531
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4532
 *         nodes_child.clear()
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.push_back(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4533
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4534
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.push_back(i)             # <<<<<<<<<<<<<<
 *             in_data_b.push_back(i)
 * 
 */
      try {
        __pyx_v_in_data.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4534, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4535
 *         for i in range(N):
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4535, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4537
 *             in_data_b.push_back(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 * 
 */
    __pyx_t_7 = __pyx_v_max_depth;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_8; __pyx_t_4+=1) {
      __pyx_v_level = __pyx_t_4;

      /* "cyext_acv.pyx":4538
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_9 = __pyx_v_nodes_level.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_10; __pyx_t_5+=1) {
        __pyx_v_it_node = __pyx_t_5;

        /* "cyext_acv.pyx":4540
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_11 = __pyx_v_b;
        __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_13 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4541
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )));
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_14 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":4542
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_16 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4542, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4544
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_17 = __pyx_v_in_data.size();
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_18; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4545
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_12 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_19 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4546
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4547
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4545
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4548
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4541
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4551
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_19 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4551, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4552
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_17 = __pyx_v_in_data.size();
            __pyx_t_18 = __pyx_t_17;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_18; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4553
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_19 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_19 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_16 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4554
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4555
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4553
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4556
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4540
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4559
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_12 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4559, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4560
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4560, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4562
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_13 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4563
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4562
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4565
 *                     break
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         for i in range(in_data.size()):
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4567
 *             nodes_level = nodes_child
 * 
 *         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *             sdp  += (1./(n_trees*in_data.size())) * (y_data[in_data[i]] <= t)
 * 
 */
    __pyx_t_9 = __pyx_v_in_data.size();
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "cyext_acv.pyx":4568
 * 
 *         for i in range(in_data.size()):
 *             sdp  += (1./(n_trees*in_data.size())) * (y_data[in_data[i]] <= t)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
      __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
      __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * ((*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))) <= __pyx_v_t)));
    }
  }

  /* "cyext_acv.pyx":4570
 *             sdp  += (1./(n_trees*in_data.size())) * (y_data[in_data[i]] <= t)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4515
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_cdf_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_cdf_rf", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":4576
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_cdf_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_59compute_cdf_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_cdf_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_cdf_rf", 0);

  /* "cyext_acv.pyx":4580
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4581
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4583
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_cdf_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_6 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_8 = (__pyx_t_6 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_8 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_10, __pyx_t_11) firstprivate(__pyx_t_9)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i)
                    #endif /* _OPENMP */
                    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_8; __pyx_t_7++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_7);

                            /* "cyext_acv.pyx":4584
 *         cdef int i
 *         for i in prange(N, nogil=True):
 *             sdp[i] = single_compute_cdf_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 */
                            __pyx_t_9.data = __pyx_v_X.data;
                            __pyx_t_9.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_X.shape[1];
__pyx_t_9.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_10 = __pyx_v_i;

                            /* "cyext_acv.pyx":4586
 *             sdp[i] = single_compute_cdf_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
                            __pyx_t_11 = __pyx_v_i;
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_11)) )) = __pyx_f_9cyext_acv_single_compute_cdf_rf(__pyx_t_9, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_10)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                            __pyx_t_9.memview = NULL;
                            __pyx_t_9.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":4583
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_cdf_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":4587
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4576
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_cdf_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_cdf_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_59compute_cdf_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_59compute_cdf_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_cdf_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 1); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 2); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 3); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 4); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 5); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 6); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 7); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 8); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 9); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 10); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 11); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, 12); __PYX_ERR(0, 4576, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_cdf_rf") < 0)) __PYX_ERR(0, 4576, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4576, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4576, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4576, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4576, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4576, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4577, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4577, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4577, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4577, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4578, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4578, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4578, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4578, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_cdf_rf", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4576, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_cdf_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_58compute_cdf_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_58compute_cdf_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_cdf_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4576, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_cdf_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_cdf_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4594
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef sufficient_expl_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

static PyObject *__pyx_pw_9cyext_acv_61sufficient_expl_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_sufficient_expl_rf_v0(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, CYTHON_UNUSED int __pyx_v_stop, PyObject *__pyx_v_search_space, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_K;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_max_size;
  int __pyx_v_size;
  int __pyx_v_subset;
  int __pyx_v_a;
  int __pyx_v_k;
  std::vector<int>  __pyx_v_S;
  CYTHON_UNUSED std::vector<int>  __pyx_v_len_s_star;
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  PyObject *__pyx_v_remove_va = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_sufficient;
  std::vector<std::vector<double> >  __pyx_v_sufficient_sdp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  std::vector<int>  __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  long __pyx_t_17;
  long __pyx_t_18;
  int __pyx_t_19;
  PyObject *(*__pyx_t_20)(PyObject *);
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_23;
  __Pyx_memviewslice __pyx_t_24 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_25;
  unsigned int __pyx_t_26;
  std::vector<std::vector<double> >  __pyx_t_27;
  std::vector<std::vector<long> > ::size_type __pyx_t_28;
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<long> ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  std::vector<std::vector<long> > ::size_type __pyx_t_33;
  std::vector<std::vector<long> > ::size_type __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sufficient_expl_rf_v0", 0);

  /* "cyext_acv.pyx":4599
 *             int minimal, bint stop, list search_space):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4600
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp, sdp_b, sdp_ba
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":4604
 *     cdef double[:] sdp, sdp_b, sdp_ba
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4605
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, a_it, b_it, o_all, p, p_s, nb_leaf, p_u, p_d, p_su, p_sd, down, up
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4605, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":4610
 *     cdef double ss, ss_a, ss_u, ss_d, K
 *     cdef int b, leaf_numb, i, s, s_0, s_1, S_size, j, max_size, size, subset, a, k
 *     K = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] S, len_s_star
 */
  __pyx_v_K = 0.0;

  /* "cyext_acv.pyx":4613
 * 
 *     cdef vector[int] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_len_s_star = __pyx_t_7;

  /* "cyext_acv.pyx":4617
 *     cdef list power, va_id
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 4617, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 4617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {

    /* "cyext_acv.pyx":4618
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 4618, __pyx_L1_error)
    }
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4618, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_ci = __pyx_t_11;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 4618, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_12 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4618, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
        __pyx_v_cj = __pyx_t_14;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4618, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4618, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 4618, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":4619
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]             # <<<<<<<<<<<<<<
 *         for ci in range(len(C)):
 *             i = 0
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 4619, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_6); __pyx_t_9 = 0;
    for (;;) {
      if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4619, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4619, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_i = __pyx_t_15;
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4619, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 4619, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_16 = (__pyx_t_8 != 0);
      if (__pyx_t_16) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 4619, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":4620
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):             # <<<<<<<<<<<<<<
 *             i = 0
 *             for cj in range(len(C[ci])):
 */
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 4620, __pyx_L1_error)
    }
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4620, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_ci = __pyx_t_11;

      /* "cyext_acv.pyx":4621
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 *             i = 0             # <<<<<<<<<<<<<<
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 */
      __pyx_v_i = 0;

      /* "cyext_acv.pyx":4622
 *         for ci in range(len(C)):
 *             i = 0
 *             for cj in range(len(C[ci])):             # <<<<<<<<<<<<<<
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 */
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 4622, __pyx_L1_error)
      }
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_12 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4622, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
        __pyx_v_cj = __pyx_t_14;

        /* "cyext_acv.pyx":4623
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4623, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_16 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_search_space, Py_EQ)); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 4623, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_8 = (__pyx_t_16 != 0);
        if (__pyx_t_8) {

          /* "cyext_acv.pyx":4624
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     break
 *             if i != 0:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "cyext_acv.pyx":4625
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 *                     break             # <<<<<<<<<<<<<<
 *             if i != 0:
 *                 va_id += [C[ci]]
 */
          goto __pyx_L14_break;

          /* "cyext_acv.pyx":4623
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        }
      }
      __pyx_L14_break:;

      /* "cyext_acv.pyx":4626
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      __pyx_t_8 = ((__pyx_v_i != 0) != 0);
      if (__pyx_t_8) {

        /* "cyext_acv.pyx":4627
 *                     break
 *             if i != 0:
 *                 va_id += [C[ci]]             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in search_space]
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 4627, __pyx_L1_error)
        }
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        PyList_SET_ITEM(__pyx_t_1, 0, PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_va_id, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_va_id, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "cyext_acv.pyx":4626
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      }
    }

    /* "cyext_acv.pyx":4617
 *     cdef list power, va_id
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
    goto __pyx_L3;
  }

  /* "cyext_acv.pyx":4629
 *                 va_id += [C[ci]]
 *     else:
 *         va_id = [[i] for i in search_space]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 4629, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    for (;;) {
      if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4629, __pyx_L1_error)
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_i = __pyx_t_15;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 4629, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "cyext_acv.pyx":4631
 *         va_id = [[i] for i in search_space]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4631, __pyx_L1_error)
  __pyx_v_m = __pyx_t_9;

  /* "cyext_acv.pyx":4632
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_power = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "cyext_acv.pyx":4633
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":4634
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_17 = (__pyx_v_m + 1);
  __pyx_t_18 = __pyx_t_17;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_18; __pyx_t_15+=1) {
    __pyx_v_size = __pyx_t_15;

    /* "cyext_acv.pyx":4635
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "cyext_acv.pyx":4636
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_combinations); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    __pyx_t_19 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_19 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_va_id, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4636, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_va_id, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4636, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_19, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_19, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_3 = __pyx_t_6; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
      __pyx_t_20 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_20 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 4636, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_20)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4636, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4636, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4636, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4636, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_20(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4636, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "cyext_acv.pyx":4637
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2);
      __pyx_t_4 = 0;
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_21 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_21)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_21);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_21, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_6); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 4637, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "cyext_acv.pyx":4638
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":4636
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyext_acv.pyx":4639
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 4639, __pyx_L1_error)

    /* "cyext_acv.pyx":4640
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_8 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_8) {

      /* "cyext_acv.pyx":4641
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L20_break;

      /* "cyext_acv.pyx":4640
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L20_break:;

  /* "cyext_acv.pyx":4643
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_23 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4643, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_23;

  /* "cyext_acv.pyx":4645
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_21) < 0) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_24 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_24.memview)) __PYX_ERR(0, 4645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_s_star = __pyx_t_24;
  __pyx_t_24.memview = NULL;
  __pyx_t_24.data = NULL;

  /* "cyext_acv.pyx":4648
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":4649
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_S = __pyx_t_7;

  /* "cyext_acv.pyx":4651
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]             # <<<<<<<<<<<<<<
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_25 = __pyx_v_N;
  __pyx_t_26 = __pyx_t_25;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_26; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_int_neg_1);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 4651, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_23 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sufficient = __pyx_t_23;

  /* "cyext_acv.pyx":4652
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_25 = __pyx_v_N;
  __pyx_t_26 = __pyx_t_25;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_26; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_int_neg_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 4652, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sufficient_sdp = __pyx_t_27;

  /* "cyext_acv.pyx":4654
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_21);
  __pyx_t_2 = 0;
  __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_21) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_21);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_20 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_20 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 4654, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_20)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4654, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4654, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_20(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4654, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4654, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_s_0 = __pyx_t_15;

    /* "cyext_acv.pyx":4655
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_28 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_29; __pyx_t_15+=1) {
      __pyx_v_s_1 = __pyx_t_15;

      /* "cyext_acv.pyx":4656
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_30 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_31; __pyx_t_19+=1) {
        __pyx_v_i = __pyx_t_19;

        /* "cyext_acv.pyx":4657
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        (__pyx_v_S[__pyx_v_i]) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":4659
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             sdp = compute_sdp_rf_v0(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":4660
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_v0(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_S); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_21 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, __pyx_v_S_size, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_21); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "cyext_acv.pyx":4661
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_v0(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
      __pyx_t_21 = __pyx_f_9cyext_acv_compute_sdp_rf_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_t_7, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);

      /* "cyext_acv.pyx":4660
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_v0(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_21, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 4660, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
      __pyx_v_sdp = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":4663
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 if sdp[i] >= pi_level:
 *                     subset = 0
 */
      __pyx_t_25 = __pyx_v_N;
      __pyx_t_26 = __pyx_t_25;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_26; __pyx_t_19+=1) {
        __pyx_v_i = __pyx_t_19;

        /* "cyext_acv.pyx":4664
 * 
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:             # <<<<<<<<<<<<<<
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_8 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
        if (__pyx_t_8) {

          /* "cyext_acv.pyx":4665
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:
 *                     subset = 0             # <<<<<<<<<<<<<<
 *                     for j in range(sufficient[i].size()):
 *                         a = 0
 */
          __pyx_v_subset = 0;

          /* "cyext_acv.pyx":4666
 *                 if sdp[i] >= pi_level:
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):             # <<<<<<<<<<<<<<
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 */
          __pyx_t_33 = (__pyx_v_sufficient[__pyx_v_i]).size();
          __pyx_t_34 = __pyx_t_33;
          for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_34; __pyx_t_35+=1) {
            __pyx_v_j = __pyx_t_35;

            /* "cyext_acv.pyx":4667
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 *                         a = 0             # <<<<<<<<<<<<<<
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 */
            __pyx_v_a = 0;

            /* "cyext_acv.pyx":4668
 *                     for j in range(sufficient[i].size()):
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):             # <<<<<<<<<<<<<<
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 */
            __pyx_t_30 = ((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j]).size();
            __pyx_t_31 = __pyx_t_30;
            for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_31; __pyx_t_36+=1) {
              __pyx_v_k = __pyx_t_36;

              /* "cyext_acv.pyx":4669
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():             # <<<<<<<<<<<<<<
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 */
              __pyx_t_8 = ((std::find<std::vector<long> ::iterator,long>(((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).begin(), ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).end(), (((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j])[__pyx_v_k])) != ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).end()) != 0);
              if (__pyx_t_8) {

                /* "cyext_acv.pyx":4670
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1             # <<<<<<<<<<<<<<
 *                         if a == sufficient[i][j].size():
 *                             subset = 1
 */
                __pyx_v_a = (__pyx_v_a + 1);

                /* "cyext_acv.pyx":4669
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():             # <<<<<<<<<<<<<<
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 */
              }
            }

            /* "cyext_acv.pyx":4671
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 *                         if a == sufficient[i][j].size():             # <<<<<<<<<<<<<<
 *                             subset = 1
 *                             break
 */
            __pyx_t_8 = ((__pyx_v_a == ((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j]).size()) != 0);
            if (__pyx_t_8) {

              /* "cyext_acv.pyx":4672
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 *                             subset = 1             # <<<<<<<<<<<<<<
 *                             break
 * 
 */
              __pyx_v_subset = 1;

              /* "cyext_acv.pyx":4673
 *                         if a == sufficient[i][j].size():
 *                             subset = 1
 *                             break             # <<<<<<<<<<<<<<
 * 
 *                     if subset == 0:
 */
              goto __pyx_L38_break;

              /* "cyext_acv.pyx":4671
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 *                         if a == sufficient[i][j].size():             # <<<<<<<<<<<<<<
 *                             subset = 1
 *                             break
 */
            }
          }
          __pyx_L38_break:;

          /* "cyext_acv.pyx":4675
 *                             break
 * 
 *                     if subset == 0:             # <<<<<<<<<<<<<<
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 */
          __pyx_t_8 = ((__pyx_v_subset == 0) != 0);
          if (__pyx_t_8) {

            /* "cyext_acv.pyx":4676
 * 
 *                     if subset == 0:
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])             # <<<<<<<<<<<<<<
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):
 */
            try {
              (__pyx_v_sufficient[__pyx_v_i]).push_back(((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4676, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4677
 *                     if subset == 0:
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1
 */
            __pyx_t_32 = __pyx_v_i;
            try {
              (__pyx_v_sufficient_sdp[__pyx_v_i]).push_back((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 4677, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4678
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             sdp_global[S[s]] += 1
 *                         K += 1
 */
            __pyx_t_35 = __pyx_v_S_size;
            __pyx_t_36 = __pyx_t_35;
            for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
              __pyx_v_s = __pyx_t_37;

              /* "cyext_acv.pyx":4679
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1             # <<<<<<<<<<<<<<
 *                         K += 1
 * 
 */
              __pyx_t_32 = (__pyx_v_S[__pyx_v_s]);
              *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_32 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
            }

            /* "cyext_acv.pyx":4680
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1
 *                         K += 1             # <<<<<<<<<<<<<<
 * 
 *     return sufficient, sufficient_sdp, np.asarray(sdp_global)/K
 */
            __pyx_v_K = (__pyx_v_K + 1.0);

            /* "cyext_acv.pyx":4675
 *                             break
 * 
 *                     if subset == 0:             # <<<<<<<<<<<<<<
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 */
          }

          /* "cyext_acv.pyx":4664
 * 
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:             # <<<<<<<<<<<<<<
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 */
        }
      }
    }

    /* "cyext_acv.pyx":4654
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4682
 *                         K += 1
 * 
 *     return sufficient, sufficient_sdp, np.asarray(sdp_global)/K             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_sufficient); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(__pyx_v_sufficient_sdp); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_K); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_21 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4594
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef sufficient_expl_rf_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_21);
  __PYX_XDEC_MEMVIEW(&__pyx_t_24, 1);
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_61sufficient_expl_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_61sufficient_expl_rf_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  PyObject *__pyx_v_search_space = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sufficient_expl_rf_v0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,&__pyx_n_s_search_space,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 1); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 2); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 3); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 4); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 5); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 6); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 7); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 8); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 9); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 10); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 11); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 12); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 13); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 14); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 15); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_space)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, 16); __PYX_ERR(0, 4594, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sufficient_expl_rf_v0") < 0)) __PYX_ERR(0, 4594, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4594, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4594, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4594, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4594, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4595, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4595, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4595, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4595, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4596, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4596, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4596, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4596, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[12]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4596, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4597, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[15]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4597, __pyx_L3_error)
    __pyx_v_search_space = ((PyObject*)values[16]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf_v0", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4594, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 4596, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_search_space), (&PyList_Type), 1, "search_space", 1))) __PYX_ERR(0, 4597, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_60sufficient_expl_rf_v0(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_60sufficient_expl_rf_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sufficient_expl_rf_v0", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4594, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_sufficient_expl_rf_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4689
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_v0(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_partition) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  std::vector<int>  __pyx_v_in_data;
  std::vector<int>  __pyx_v_in_data_b;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  std::vector<int> ::size_type __pyx_t_9;
  std::vector<int> ::size_type __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  size_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  std::vector<int> ::size_type __pyx_t_18;
  std::vector<int> ::size_type __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4693
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4694
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4697
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int b, level, it_node, i
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4702
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4703
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4704
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4704, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4705
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         in_data.clear()
 *         in_data_b.clear()
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4706
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4707
 *         nodes_child.clear()
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.push_back(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4708
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4709
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.push_back(i)             # <<<<<<<<<<<<<<
 *             in_data_b.push_back(i)
 * 
 */
      try {
        __pyx_v_in_data.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4709, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4710
 *         for i in range(N):
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4710, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4712
 *             in_data_b.push_back(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 * 
 */
    __pyx_t_7 = __pyx_v_max_depth;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_8; __pyx_t_4+=1) {
      __pyx_v_level = __pyx_t_4;

      /* "cyext_acv.pyx":4713
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_9 = __pyx_v_nodes_level.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_10; __pyx_t_5+=1) {
        __pyx_v_it_node = __pyx_t_5;

        /* "cyext_acv.pyx":4715
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_11 = __pyx_v_b;
        __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_13 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4716
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )));
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_14 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":4717
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_16 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4717, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4719
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
            __pyx_t_14 = 1;
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_14 * __pyx_v_partition.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_17 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_13) {

              /* "cyext_acv.pyx":4720
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_12 = 1;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_14 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_17 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":4719
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }

            /* "cyext_acv.pyx":4722
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_18 = __pyx_v_in_data.size();
            __pyx_t_19 = __pyx_t_18;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_19; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4723
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_17 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_16 * __pyx_v_data.strides[0]) ) + __pyx_t_12 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4724
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4725
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4723
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4726
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4716
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":4729
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_14 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4729, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4731
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_17 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
            __pyx_t_12 = 0;
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_17 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_13) {

              /* "cyext_acv.pyx":4732
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
              __pyx_t_17 = 0;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_17 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":4731
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }

            /* "cyext_acv.pyx":4734
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_18 = __pyx_v_in_data.size();
            __pyx_t_19 = __pyx_t_18;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_19; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":4735
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_14 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_17 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_14 * __pyx_v_data.strides[0]) ) + __pyx_t_17 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":4736
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":4737
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":4735
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":4738
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4715
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":4741
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_12 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4741, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4742
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4742, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":4744
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_13 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4745
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":4744
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":4747
 *                     break
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":4749
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
    __pyx_t_13 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_13) {

      /* "cyext_acv.pyx":4750
 * 
 *         if classifier == 1:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *         else:
 */
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":4751
 *         if classifier == 1:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(in_data.size()):
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))))));
      }

      /* "cyext_acv.pyx":4749
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
      goto __pyx_L22;
    }

    /* "cyext_acv.pyx":4753
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *         else:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":4754
 *         else:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_16)) ))))) <= __pyx_v_t)));
      }
    }
    __pyx_L22:;
  }

  /* "cyext_acv.pyx":4756
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4689
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_v0(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":4762
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_63compute_sdp_rule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule_v0(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule_v0", 0);

  /* "cyext_acv.pyx":4766
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4767
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4767, __pyx_L1_error)

  /* "cyext_acv.pyx":4770
 * 
 *         cdef double [:, :, :] partition_byobs
 *         partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 4770, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":4772
 *         partition_byobs = buffer
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4773
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 4773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":4775
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_8 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_10 = (__pyx_t_8 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_10 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_12, __pyx_t_14) firstprivate(__pyx_t_11, __pyx_t_13)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_10; __pyx_t_9++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_9);

                            /* "cyext_acv.pyx":4776
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
                            __pyx_t_11.data = __pyx_v_X.data;
                            __pyx_t_11.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_11, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_11.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_11.shape[0] = __pyx_v_X.shape[1];
__pyx_t_11.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_11.suboffsets[0] = -1;

__pyx_t_12 = __pyx_v_i;

                            /* "cyext_acv.pyx":4778
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs)
 */
                            __pyx_t_13.data = __pyx_v_partition_byobs.data;
                            __pyx_t_13.memview = __pyx_v_partition_byobs.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_13.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_13.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_13.suboffsets[0] = -1;

__pyx_t_13.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_13.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_13.suboffsets[1] = -1;

__pyx_t_14 = __pyx_v_i;

                            /* "cyext_acv.pyx":4776
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_14)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__pyx_t_11, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_12)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_13);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_11, 0);
                            __pyx_t_11.memview = NULL;
                            __pyx_t_11.data = NULL;
                            __PYX_XDEC_MEMVIEW(&__pyx_t_13, 0);
                            __pyx_t_13.memview = NULL;
                            __pyx_t_13.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":4775
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":4780
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *         return np.array(sdp), np.array(partition_byobs)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4762
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_63compute_sdp_rule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_63compute_sdp_rule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rule_v0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 1); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 2); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 3); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 4); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 5); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 6); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 7); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 8); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 9); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 10); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 11); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, 12); __PYX_ERR(0, 4762, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rule_v0") < 0)) __PYX_ERR(0, 4762, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4762, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4762, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4762, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4762, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4762, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4763, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4763, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4763, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4763, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4764, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4764, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4764, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4764, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_v0", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4762, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_62compute_sdp_rule_v0(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_62compute_sdp_rule_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule_v0", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4762, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rule_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4787
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_65compute_sdp_maxrule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_v0(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_v0", 0);

  /* "cyext_acv.pyx":4791
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4792
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4792, __pyx_L1_error)

  /* "cyext_acv.pyx":4795
 * 
 *         cdef double [:, :, :] partition_byobs
 *         partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 4795, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":4797
 *         partition_byobs = buffer
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4798
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 *         buffer_data[:, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :] partition_byobs_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__4, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4798, __pyx_L1_error)

  /* "cyext_acv.pyx":4801
 * 
 *         cdef double [:, :, :, :] partition_byobs_data
 *         partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":4803
 *         partition_byobs_data = buffer_data
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 *         cdef double[:] sdp = np.zeros(N)
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4804
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef double[:] sdp = np.zeros(N)
 *         cdef int i, j, k
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 4804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_data = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":4805
 *         cdef int N = X.shape[0]
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 *         cdef double[:] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 4805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":4809
 * 
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],
 *                             features, thresholds, children_left, children_right,
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":4810
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    __pyx_t_10.data = __pyx_v_X.data;
    __pyx_t_10.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_X.shape[1];
__pyx_t_10.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_14 = __pyx_v_i;

    /* "cyext_acv.pyx":4812
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_9.data = __pyx_v_partition_byobs.data;
    __pyx_t_9.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_9.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_9.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_9.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_9.suboffsets[1] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":4810
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_15 * __pyx_v_sdp.strides[0]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__pyx_t_10, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_14)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_9);
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
    __pyx_t_9.memview = NULL;
    __pyx_t_9.data = NULL;

    /* "cyext_acv.pyx":4814
 *                             max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                             features, thresholds, children_left, children_right,
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 4814, __pyx_L8_error) }
          __pyx_t_16 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_18 = (__pyx_t_16 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_18 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_14, __pyx_t_15, __pyx_t_19) firstprivate(__pyx_t_10, __pyx_t_9)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_18; __pyx_t_17++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_17);

                              /* "cyext_acv.pyx":4815
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                              __pyx_t_10.data = __pyx_v_data.data;
                              __pyx_t_10.memview = __pyx_v_data.memview;
                              __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
                              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_data.shape[1];
__pyx_t_10.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_14 = __pyx_v_i;

                              /* "cyext_acv.pyx":4817
 *                 sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data)
 */
                              __pyx_t_9.data = __pyx_v_partition_byobs_data.data;
                              __pyx_t_9.memview = __pyx_v_partition_byobs_data.memview;
                              __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_9.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_9.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_9.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_9.suboffsets[1] = -1;

__pyx_t_15 = __pyx_v_i;

                              /* "cyext_acv.pyx":4815
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                              __pyx_t_19 = __pyx_v_j;
                              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_15 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_19 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__pyx_t_10, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_14)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_9);
                              __PYX_XDEC_MEMVIEW(&__pyx_t_10, 0);
                              __pyx_t_10.memview = NULL;
                              __pyx_t_10.data = NULL;
                              __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                              __pyx_t_9.memview = NULL;
                              __pyx_t_9.data = NULL;
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":4814
 *                             max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                             features, thresholds, children_left, children_right,
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":4819
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs_data, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_20 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_7 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_20, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4787
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_v0(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_65compute_sdp_maxrule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_65compute_sdp_maxrule_v0(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_v0 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 1); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 2); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 3); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 4); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 5); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 6); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 7); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 8); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 9); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 10); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 11); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 12); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, 13); __PYX_ERR(0, 4787, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule_v0") < 0)) __PYX_ERR(0, 4787, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4787, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4787, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4787, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4787, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4787, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4788, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4788, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4788, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4788, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4789, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4789, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4789, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4789, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4789, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_v0", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4787, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_64compute_sdp_maxrule_v0(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_64compute_sdp_maxrule_v0(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_v0", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4787, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule_v0(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_v0", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4826
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_biased(double[:] & x, double & y_x, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :, :, :] & partition) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__Pyx_memviewslice &__pyx_v_x, CYTHON_UNUSED double &__pyx_v_y_x, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, CYTHON_UNUSED int &__pyx_v_min_node_size, CYTHON_UNUSED int &__pyx_v_classifier, CYTHON_UNUSED double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_partition) {
  unsigned int __pyx_v_n_trees;
  double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  unsigned int __pyx_t_6;
  std::vector<int> ::size_type __pyx_t_7;
  std::vector<int> ::size_type __pyx_t_8;
  unsigned int __pyx_t_9;
  size_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  size_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  size_t __pyx_t_16;
  size_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4830
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :, :, :] & partition) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4833
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int b, level, it_node, i
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4838
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4839
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4840
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 * 
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4840, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4841
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4843
 *         nodes_child.clear()
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
    __pyx_t_4 = __pyx_v_max_depth;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_level = __pyx_t_6;

      /* "cyext_acv.pyx":4844
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 */
      __pyx_t_7 = __pyx_v_nodes_level.size();
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_it_node = __pyx_t_9;

        /* "cyext_acv.pyx":4845
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 * 
 */
        __pyx_t_10 = __pyx_v_b;
        __pyx_t_11 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_12 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":4846
 *             for it_node in range(nodes_level.size()):
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 * 
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          __pyx_t_10 = __pyx_v_b;
          __pyx_t_11 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )));
          __pyx_t_14 = __pyx_v_b;
          __pyx_t_15 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_12 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_13 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_14 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_15 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_12) {

            /* "cyext_acv.pyx":4848
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 * 
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level, 1] >= thresholds[b, nodes_level[it_node]]:
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_15 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_14 * __pyx_v_children_left.strides[0]) ) + __pyx_t_15 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4848, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4850
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level, 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], b, level,  1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_15 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
            __pyx_t_10 = __pyx_v_b;
            __pyx_t_16 = __pyx_v_level;
            __pyx_t_13 = 1;
            __pyx_t_17 = __pyx_v_b;
            __pyx_t_18 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_12 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_11 * __pyx_v_partition.strides[0]) ) + __pyx_t_10 * __pyx_v_partition.strides[1]) ) + __pyx_t_16 * __pyx_v_partition.strides[2]) ) + __pyx_t_13 * __pyx_v_partition.strides[3]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_17 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_18 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_12) {

              /* "cyext_acv.pyx":4851
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level, 1] >= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], b, level,  1] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
              __pyx_t_17 = __pyx_v_b;
              __pyx_t_18 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_14 = __pyx_v_b;
              __pyx_t_15 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
              __pyx_t_16 = __pyx_v_b;
              __pyx_t_10 = __pyx_v_level;
              __pyx_t_11 = 1;
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_13 * __pyx_v_partition.strides[0]) ) + __pyx_t_16 * __pyx_v_partition.strides[1]) ) + __pyx_t_10 * __pyx_v_partition.strides[2]) ) + __pyx_t_11 * __pyx_v_partition.strides[3]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_17 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_18 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":4850
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level, 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], b, level,  1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }

            /* "cyext_acv.pyx":4846
 *             for it_node in range(nodes_level.size()):
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 * 
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
            goto __pyx_L10;
          }

          /* "cyext_acv.pyx":4854
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level,  0] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_17 = __pyx_v_b;
            __pyx_t_18 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_17 * __pyx_v_children_right.strides[0]) ) + __pyx_t_18 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4854, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":4856
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level,  0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], b, level, 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_17 = __pyx_v_b;
            __pyx_t_18 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_15 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_17 * __pyx_v_features.strides[0]) ) + __pyx_t_18 * __pyx_v_features.strides[1]) )));
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_10 = __pyx_v_level;
            __pyx_t_11 = 0;
            __pyx_t_16 = __pyx_v_b;
            __pyx_t_13 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_12 = (((*((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_15 * __pyx_v_partition.strides[0]) ) + __pyx_t_14 * __pyx_v_partition.strides[1]) ) + __pyx_t_10 * __pyx_v_partition.strides[2]) ) + __pyx_t_11 * __pyx_v_partition.strides[3]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_16 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_13 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_12) {

              /* "cyext_acv.pyx":4857
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level,  0] <= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], b, level, 0] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
              __pyx_t_16 = __pyx_v_b;
              __pyx_t_13 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_17 = __pyx_v_b;
              __pyx_t_18 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_17 * __pyx_v_features.strides[0]) ) + __pyx_t_18 * __pyx_v_features.strides[1]) )));
              __pyx_t_10 = __pyx_v_b;
              __pyx_t_14 = __pyx_v_level;
              __pyx_t_15 = 0;
              *((double *) ( /* dim=3 */ (( /* dim=2 */ (( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_11 * __pyx_v_partition.strides[0]) ) + __pyx_t_10 * __pyx_v_partition.strides[1]) ) + __pyx_t_14 * __pyx_v_partition.strides[2]) ) + __pyx_t_15 * __pyx_v_partition.strides[3]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_16 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_13 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":4856
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], b, level,  0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], b, level, 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }
          }
          __pyx_L10:;

          /* "cyext_acv.pyx":4845
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 * 
 */
          goto __pyx_L9;
        }

        /* "cyext_acv.pyx":4860
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_16 = __pyx_v_b;
          __pyx_t_13 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_16 * __pyx_v_children_left.strides[0]) ) + __pyx_t_13 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4860, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":4861
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          __pyx_t_16 = __pyx_v_b;
          __pyx_t_13 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_16 * __pyx_v_children_right.strides[0]) ) + __pyx_t_13 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 4861, __pyx_L1_error)
          }
        }
        __pyx_L9:;
      }

      /* "cyext_acv.pyx":4863
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }
  }

  /* "cyext_acv.pyx":4865
 *             nodes_level = nodes_child
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * cpdef compile_rules(rule_data, data, S, min_node_size):
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4826
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_biased(double[:] & x, double & y_x, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :, :, :] & partition) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":4867
 *     return sdp
 * 
 * cpdef compile_rules(rule_data, data, S, min_node_size):             # <<<<<<<<<<<<<<
 *     ntree = rule_data.shape[1]
 *     rule_trees_up = []
 */

static PyObject *__pyx_pw_9cyext_acv_67compile_rules(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compile_rules(PyObject *__pyx_v_rule_data, PyObject *__pyx_v_data, PyObject *__pyx_v_S, PyObject *__pyx_v_min_node_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ntree = NULL;
  PyObject *__pyx_v_rule_trees_up = NULL;
  PyObject *__pyx_v_rule_trees_dn = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_tn = NULL;
  PyObject *__pyx_v_depth = NULL;
  PyObject *__pyx_v_part_up = NULL;
  PyObject *__pyx_v_part_dn = NULL;
  PyObject *__pyx_v_where = NULL;
  PyObject *__pyx_v_rule_trees = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compile_rules", 0);

  /* "cyext_acv.pyx":4868
 * 
 * cpdef compile_rules(rule_data, data, S, min_node_size):
 *     ntree = rule_data.shape[1]             # <<<<<<<<<<<<<<
 *     rule_trees_up = []
 *     rule_trees_dn = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_rule_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ntree = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":4869
 * cpdef compile_rules(rule_data, data, S, min_node_size):
 *     ntree = rule_data.shape[1]
 *     rule_trees_up = []             # <<<<<<<<<<<<<<
 *     rule_trees_dn = []
 *     for i in range(ntree):
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_rule_trees_up = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":4870
 *     ntree = rule_data.shape[1]
 *     rule_trees_up = []
 *     rule_trees_dn = []             # <<<<<<<<<<<<<<
 *     for i in range(ntree):
 *         tn = 0
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_rule_trees_dn = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":4871
 *     rule_trees_up = []
 *     rule_trees_dn = []
 *     for i in range(ntree):             # <<<<<<<<<<<<<<
 *         tn = 0
 *         depth = rule_data.shape[2]
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_ntree); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4871, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4871, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4871, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4871, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":4872
 *     rule_trees_dn = []
 *     for i in range(ntree):
 *         tn = 0             # <<<<<<<<<<<<<<
 *         depth = rule_data.shape[2]
 *         while(tn <= min_node_size):
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_XDECREF_SET(__pyx_v_tn, __pyx_int_0);

    /* "cyext_acv.pyx":4873
 *     for i in range(ntree):
 *         tn = 0
 *         depth = rule_data.shape[2]             # <<<<<<<<<<<<<<
 *         while(tn <= min_node_size):
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rule_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_depth, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "cyext_acv.pyx":4874
 *         tn = 0
 *         depth = rule_data.shape[2]
 *         while(tn <= min_node_size):             # <<<<<<<<<<<<<<
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 */
    while (1) {
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_tn, __pyx_v_min_node_size, Py_LE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4874, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 4874, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!__pyx_t_6) break;

      /* "cyext_acv.pyx":4875
 *         depth = rule_data.shape[2]
 *         while(tn <= min_node_size):
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)             # <<<<<<<<<<<<<<
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PySlice_New(Py_None, __pyx_v_depth, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_slice__2);
      __Pyx_GIVEREF(__pyx_slice__2);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_slice__2);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_i);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_5);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_1);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_rule_data, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4875, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_part_up, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "cyext_acv.pyx":4876
 *         while(tn <= min_node_size):
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)             # <<<<<<<<<<<<<<
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_max); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PySlice_New(Py_None, __pyx_v_depth, Py_None); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_slice__2);
      __Pyx_GIVEREF(__pyx_slice__2);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_slice__2);
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_i);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_8);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_0);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_rule_data, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4876, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_part_dn, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "cyext_acv.pyx":4877
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])             # <<<<<<<<<<<<<<
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_prod); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "cyext_acv.pyx":4878
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)             # <<<<<<<<<<<<<<
 *             tn = np.sum(where)
 *             depth -= 1
 */
      if (likely(PyList_CheckExact(__pyx_v_S)) || PyTuple_CheckExact(__pyx_v_S)) {
        __pyx_t_5 = __pyx_v_S; __Pyx_INCREF(__pyx_t_5); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_S); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4878, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4878, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_11); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4878, __pyx_L1_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4878, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_11); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 4878, __pyx_L1_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4878, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          }
        } else {
          __pyx_t_11 = __pyx_t_10(__pyx_t_5);
          if (unlikely(!__pyx_t_11)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 4878, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "cyext_acv.pyx":4877
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])             # <<<<<<<<<<<<<<
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)
 */
        __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_slice__2);
        __Pyx_GIVEREF(__pyx_slice__2);
        PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_slice__2);
        __Pyx_INCREF(__pyx_v_s);
        __Pyx_GIVEREF(__pyx_v_s);
        PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_s);
        __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_v_data, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_part_up, __pyx_v_s); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_12, __pyx_t_11, Py_LE); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_slice__2);
        __Pyx_GIVEREF(__pyx_slice__2);
        PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_slice__2);
        __Pyx_INCREF(__pyx_v_s);
        __Pyx_GIVEREF(__pyx_v_s);
        PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_s);
        __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_v_data, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_part_dn, __pyx_v_s); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_14 = PyObject_RichCompare(__pyx_t_12, __pyx_t_11, Py_GE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyNumber_Multiply(__pyx_t_13, __pyx_t_14); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_11))) __PYX_ERR(0, 4877, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "cyext_acv.pyx":4878
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)             # <<<<<<<<<<<<<<
 *             tn = np.sum(where)
 *             depth -= 1
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "cyext_acv.pyx":4877
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])             # <<<<<<<<<<<<<<
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)
 */
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "cyext_acv.pyx":4878
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)             # <<<<<<<<<<<<<<
 *             tn = np.sum(where)
 *             depth -= 1
 */
      __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_axis, __pyx_int_0) < 0) __PYX_ERR(0, 4878, __pyx_L1_error)

      /* "cyext_acv.pyx":4877
 *             part_up = np.min(rule_data[:, i, :depth, 1], axis=-1)
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])             # <<<<<<<<<<<<<<
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)
 */
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "cyext_acv.pyx":4878
 *             part_dn = np.max(rule_data[:, i, :depth, 0], axis=-1)
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)             # <<<<<<<<<<<<<<
 *             tn = np.sum(where)
 *             depth -= 1
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_2 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_11, ((PyObject*)&PyBool_Type)) : __Pyx_PyObject_CallOneArg(__pyx_t_8, ((PyObject*)&PyBool_Type));
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_where, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "cyext_acv.pyx":4879
 *             where = np.prod([(data[:, s] <= part_up[s]) * (data[:, s] >= part_dn[s])
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)             # <<<<<<<<<<<<<<
 *             depth -= 1
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_sum); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_8, __pyx_v_where) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_where);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF_SET(__pyx_v_tn, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "cyext_acv.pyx":4880
 *                                              for s in S], axis=0).astype(bool)
 *             tn = np.sum(where)
 *             depth -= 1             # <<<<<<<<<<<<<<
 * 
 *         rule_trees_up.append(np.expand_dims(part_up, -1))
 */
      __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_depth, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_depth, __pyx_t_2);
      __pyx_t_2 = 0;
    }

    /* "cyext_acv.pyx":4882
 *             depth -= 1
 * 
 *         rule_trees_up.append(np.expand_dims(part_up, -1))             # <<<<<<<<<<<<<<
 *         rule_trees_dn.append(np.expand_dims(part_dn, -1))
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_np); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_expand_dims); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_v_part_up)) { __Pyx_RaiseUnboundLocalError("part_up"); __PYX_ERR(0, 4882, __pyx_L1_error) }
    __pyx_t_11 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_part_up, __pyx_int_neg_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4882, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_part_up, __pyx_int_neg_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4882, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(__pyx_v_part_up);
      __Pyx_GIVEREF(__pyx_v_part_up);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_15, __pyx_v_part_up);
      __Pyx_INCREF(__pyx_int_neg_1);
      __Pyx_GIVEREF(__pyx_int_neg_1);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_15, __pyx_int_neg_1);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_rule_trees_up, __pyx_t_2); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 4882, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyext_acv.pyx":4883
 * 
 *         rule_trees_up.append(np.expand_dims(part_up, -1))
 *         rule_trees_dn.append(np.expand_dims(part_dn, -1))             # <<<<<<<<<<<<<<
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_expand_dims); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_v_part_dn)) { __Pyx_RaiseUnboundLocalError("part_dn"); __PYX_ERR(0, 4883, __pyx_L1_error) }
    __pyx_t_8 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_part_dn, __pyx_int_neg_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4883, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_part_dn, __pyx_int_neg_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4883, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_11 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_part_dn);
      __Pyx_GIVEREF(__pyx_v_part_dn);
      PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_15, __pyx_v_part_dn);
      __Pyx_INCREF(__pyx_int_neg_1);
      __Pyx_GIVEREF(__pyx_int_neg_1);
      PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_15, __pyx_int_neg_1);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_16 = __Pyx_PyObject_Append(__pyx_v_rule_trees_dn, __pyx_t_2); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 4883, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "cyext_acv.pyx":4871
 *     rule_trees_up = []
 *     rule_trees_dn = []
 *     for i in range(ntree):             # <<<<<<<<<<<<<<
 *         tn = 0
 *         depth = rule_data.shape[2]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4884
 *         rule_trees_up.append(np.expand_dims(part_up, -1))
 *         rule_trees_dn.append(np.expand_dims(part_dn, -1))
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)             # <<<<<<<<<<<<<<
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_rule_trees_up);
  __Pyx_GIVEREF(__pyx_v_rule_trees_up);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_rule_trees_up);
  __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4884, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4884, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_11, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_rule_trees_up, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4885
 *         rule_trees_dn.append(np.expand_dims(part_dn, -1))
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)             # <<<<<<<<<<<<<<
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],
 *                                 axis=-1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_max); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_rule_trees_dn);
  __Pyx_GIVEREF(__pyx_v_rule_trees_dn);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_rule_trees_dn);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4885, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4885, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_rule_trees_dn, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4886
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],             # <<<<<<<<<<<<<<
 *                                 axis=-1)
 *     return rule_trees
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_expand_dims); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_rule_trees_dn);
  __Pyx_GIVEREF(__pyx_v_rule_trees_dn);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_rule_trees_dn);
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4886, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_expand_dims); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_rule_trees_up);
  __Pyx_GIVEREF(__pyx_v_rule_trees_up);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_rule_trees_up);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4886, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_11);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_7);
  __pyx_t_11 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":4887
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],
 *                                 axis=-1)             # <<<<<<<<<<<<<<
 *     return rule_trees
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_axis, __pyx_int_neg_1) < 0) __PYX_ERR(0, 4887, __pyx_L1_error)

  /* "cyext_acv.pyx":4886
 *     rule_trees_up = np.min(np.concatenate(rule_trees_up, axis=-1), axis=-1)
 *     rule_trees_dn = np.max(np.concatenate(rule_trees_dn, axis=-1), axis=-1)
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],             # <<<<<<<<<<<<<<
 *                                 axis=-1)
 *     return rule_trees
 */
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rule_trees = __pyx_t_11;
  __pyx_t_11 = 0;

  /* "cyext_acv.pyx":4888
 *     rule_trees = np.concatenate([np.expand_dims(rule_trees_dn, axis=-1), np.expand_dims(rule_trees_up, axis=-1)],
 *                                 axis=-1)
 *     return rule_trees             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_rule_trees);
  __pyx_r = __pyx_v_rule_trees;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4867
 *     return sdp
 * 
 * cpdef compile_rules(rule_data, data, S, min_node_size):             # <<<<<<<<<<<<<<
 *     ntree = rule_data.shape[1]
 *     rule_trees_up = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("cyext_acv.compile_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ntree);
  __Pyx_XDECREF(__pyx_v_rule_trees_up);
  __Pyx_XDECREF(__pyx_v_rule_trees_dn);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_tn);
  __Pyx_XDECREF(__pyx_v_depth);
  __Pyx_XDECREF(__pyx_v_part_up);
  __Pyx_XDECREF(__pyx_v_part_dn);
  __Pyx_XDECREF(__pyx_v_where);
  __Pyx_XDECREF(__pyx_v_rule_trees);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_67compile_rules(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_67compile_rules(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rule_data = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_S = 0;
  PyObject *__pyx_v_min_node_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compile_rules (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rule_data,&__pyx_n_s_data,&__pyx_n_s_S,&__pyx_n_s_min_node_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rule_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compile_rules", 1, 4, 4, 1); __PYX_ERR(0, 4867, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compile_rules", 1, 4, 4, 2); __PYX_ERR(0, 4867, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compile_rules", 1, 4, 4, 3); __PYX_ERR(0, 4867, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compile_rules") < 0)) __PYX_ERR(0, 4867, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_rule_data = values[0];
    __pyx_v_data = values[1];
    __pyx_v_S = values[2];
    __pyx_v_min_node_size = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compile_rules", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4867, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compile_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_66compile_rules(__pyx_self, __pyx_v_rule_data, __pyx_v_data, __pyx_v_S, __pyx_v_min_node_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_66compile_rules(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rule_data, PyObject *__pyx_v_data, PyObject *__pyx_v_S, PyObject *__pyx_v_min_node_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compile_rules", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_9cyext_acv_compile_rules(__pyx_v_rule_data, __pyx_v_data, __pyx_v_S, __pyx_v_min_node_size, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compile_rules", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4895
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule_biased(double[:, :] X, double[::1] y_X, double[:, :] data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_69compute_sdp_rule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule_biased(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_rules = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_14;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule_biased", 0);

  /* "cyext_acv.pyx":4899
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_children_left.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_max_depth); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4900
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer[:, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :] partition_byobs
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__5, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4900, __pyx_L1_error)

  /* "cyext_acv.pyx":4903
 * 
 *         cdef double [:, :, :, :, :] partition_byobs
 *         partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 4903, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":4905
 *         partition_byobs = buffer
 * 
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "cyext_acv.pyx":4907
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4908
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(N):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 4908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":4910
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_10 = __pyx_v_N;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "cyext_acv.pyx":4911
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    __pyx_t_13.data = __pyx_v_X.data;
    __pyx_t_13.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_13.shape[0] = __pyx_v_X.shape[1];
__pyx_t_13.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_13.suboffsets[0] = -1;

__pyx_t_14 = __pyx_v_i;

    /* "cyext_acv.pyx":4913
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
    __pyx_t_15.data = __pyx_v_partition_byobs.data;
    __pyx_t_15.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_15, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_15.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_15.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_15.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_15.suboffsets[0] = -1;

__pyx_t_15.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_15.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_15.suboffsets[1] = -1;

__pyx_t_15.shape[2] = __pyx_v_partition_byobs.shape[3];
__pyx_t_15.strides[2] = __pyx_v_partition_byobs.strides[3];
    __pyx_t_15.suboffsets[2] = -1;

__pyx_t_15.shape[3] = __pyx_v_partition_byobs.shape[4];
__pyx_t_15.strides[3] = __pyx_v_partition_byobs.strides[4];
    __pyx_t_15.suboffsets[3] = -1;

__pyx_t_16 = __pyx_v_i;

    /* "cyext_acv.pyx":4911
 *         cdef int i
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_16)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__pyx_t_13, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_14)) ))), (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_15);
    __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
    __pyx_t_13.memview = NULL;
    __pyx_t_13.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
    __pyx_t_15.memview = NULL;
    __pyx_t_15.data = NULL;
  }

  /* "cyext_acv.pyx":4915
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)
 * 
 */
  __pyx_t_10 = __pyx_v_N;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "cyext_acv.pyx":4916
 * 
 *         for i in range(N):
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_15.data = __pyx_v_partition_byobs.data;
    __pyx_t_15.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_15, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_15.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_15.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_15.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_15.suboffsets[0] = -1;

__pyx_t_15.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_15.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_15.suboffsets[1] = -1;

__pyx_t_15.shape[2] = __pyx_v_partition_byobs.shape[3];
__pyx_t_15.strides[2] = __pyx_v_partition_byobs.strides[3];
    __pyx_t_15.suboffsets[2] = -1;

__pyx_t_15.shape[3] = __pyx_v_partition_byobs.shape[4];
__pyx_t_15.strides[3] = __pyx_v_partition_byobs.strides[4];
    __pyx_t_15.suboffsets[3] = -1;

__pyx_t_5 = __pyx_memoryview_fromslice(__pyx_t_15, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
    __pyx_t_15.memview = NULL;
    __pyx_t_15.data = NULL;
    __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __pyx_convert_vector_to_py_int((__pyx_v_S[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_min_node_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __pyx_f_9cyext_acv_compile_rules(__pyx_t_5, __pyx_t_7, __pyx_t_1, __pyx_t_6, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__Pyx_SetItemInt(__pyx_v_rules, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 0, 0) < 0)) __PYX_ERR(0, 4916, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "cyext_acv.pyx":4919
 * 
 * 
 *         return np.array(sdp), rules             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_rules);
  __Pyx_GIVEREF(__pyx_v_rules);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_rules);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4895
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule_biased(double[:, :] X, double[::1] y_X, double[:, :] data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_rules);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_69compute_sdp_rule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_69compute_sdp_rule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rule_biased (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 1); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 2); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 3); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 4); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 5); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 6); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 7); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 8); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 9); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 10); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, 11); __PYX_ERR(0, 4895, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rule_biased") < 0)) __PYX_ERR(0, 4895, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4895, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4895, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4895, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[3]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4895, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4896, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4896, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4896, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4896, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4897, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4897, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4897, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4897, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rule_biased", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_68compute_sdp_rule_biased(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_68compute_sdp_rule_biased(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule_biased", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4895, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rule_biased(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4926
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_biased(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_71compute_sdp_maxrule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_biased(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_rules = NULL;
  PyObject *__pyx_v_rules_data = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  int __pyx_t_24;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased", 0);

  /* "cyext_acv.pyx":4930
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_children_left.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_max_depth); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4931
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer[:, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :] partition_byobs
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__5, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4931, __pyx_L1_error)

  /* "cyext_acv.pyx":4934
 * 
 *         cdef double [:, :, :, :, :] partition_byobs
 *         partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 4934, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":4936
 *         partition_byobs = buffer
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_children_left.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_max_depth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = PyTuple_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_9) < 0) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":4937
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer_data[:, :, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :, :] partition_byobs_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__6, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 4937, __pyx_L1_error)

  /* "cyext_acv.pyx":4940
 * 
 *         cdef double [:, :, :, :, :, :] partition_byobs_data
 *         partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))
 */
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 4940, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":4942
 *         partition_byobs_data = buffer_data
 * 
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_int_2);
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_4) < 0) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "cyext_acv.pyx":4943
 * 
 *         rules = np.zeros(shape=(X.shape[0], X.shape[1], 2))
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_7);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_int_2);
  __pyx_t_9 = 0;
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_rules_data = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "cyext_acv.pyx":4945
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":4946
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 4946, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_sdp = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":4947
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         cdef int i, j
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 4947, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_sdp_data = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":4950
 * 
 *         cdef int i, j
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_13 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "cyext_acv.pyx":4951
 *         cdef int i, j
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    __pyx_t_16.data = __pyx_v_X.data;
    __pyx_t_16.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_X.shape[1];
__pyx_t_16.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

    /* "cyext_acv.pyx":4953
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_18.data = __pyx_v_partition_byobs.data;
    __pyx_t_18.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_18.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_18.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_byobs.shape[3];
__pyx_t_18.strides[2] = __pyx_v_partition_byobs.strides[3];
    __pyx_t_18.suboffsets[2] = -1;

__pyx_t_18.shape[3] = __pyx_v_partition_byobs.shape[4];
__pyx_t_18.strides[3] = __pyx_v_partition_byobs.strides[4];
    __pyx_t_18.suboffsets[3] = -1;

__pyx_t_19 = __pyx_v_i;

    /* "cyext_acv.pyx":4951
 *         cdef int i, j
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_biased(X[i], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_19)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_18);
    __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
    __pyx_t_16.memview = NULL;
    __pyx_t_16.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
    __pyx_t_18.memview = NULL;
    __pyx_t_18.data = NULL;

    /* "cyext_acv.pyx":4955
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                             features, thresholds, children_left, children_right,
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 4955, __pyx_L8_error) }
          __pyx_t_20 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_22 = (__pyx_t_20 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_22 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_17, __pyx_t_19, __pyx_t_23) firstprivate(__pyx_t_16, __pyx_t_18)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_22; __pyx_t_21++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_21);

                              /* "cyext_acv.pyx":4956
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                              __pyx_t_16.data = __pyx_v_data.data;
                              __pyx_t_16.memview = __pyx_v_data.memview;
                              __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
                              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_data.shape[1];
__pyx_t_16.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

                              /* "cyext_acv.pyx":4958
 *                 sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
                              __pyx_t_18.data = __pyx_v_partition_byobs_data.data;
                              __pyx_t_18.memview = __pyx_v_partition_byobs_data.memview;
                              __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
                              {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_18.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_18.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_byobs_data.shape[4];
__pyx_t_18.strides[2] = __pyx_v_partition_byobs_data.strides[4];
    __pyx_t_18.suboffsets[2] = -1;

__pyx_t_18.shape[3] = __pyx_v_partition_byobs_data.shape[5];
__pyx_t_18.strides[3] = __pyx_v_partition_byobs_data.strides[5];
    __pyx_t_18.suboffsets[3] = -1;

__pyx_t_19 = __pyx_v_i;

                              /* "cyext_acv.pyx":4956
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                             features, thresholds, children_left, children_right,
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                              __pyx_t_23 = __pyx_v_j;
                              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_19 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_23 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_18);
                              __PYX_XDEC_MEMVIEW(&__pyx_t_16, 0);
                              __pyx_t_16.memview = NULL;
                              __pyx_t_16.data = NULL;
                              __PYX_XDEC_MEMVIEW(&__pyx_t_18, 0);
                              __pyx_t_18.memview = NULL;
                              __pyx_t_18.data = NULL;
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":4955
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                             features, thresholds, children_left, children_right,
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":4960
 *                             max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)
 *             for j in range(data.shape[0]):
 */
  __pyx_t_13 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "cyext_acv.pyx":4961
 * 
 *         for i in range(N):
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)             # <<<<<<<<<<<<<<
 *             for j in range(data.shape[0]):
 *                 rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 */
    __pyx_t_18.data = __pyx_v_partition_byobs.data;
    __pyx_t_18.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_18.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_18.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_byobs.shape[3];
__pyx_t_18.strides[2] = __pyx_v_partition_byobs.strides[3];
    __pyx_t_18.suboffsets[2] = -1;

__pyx_t_18.shape[3] = __pyx_v_partition_byobs.shape[4];
__pyx_t_18.strides[3] = __pyx_v_partition_byobs.strides[4];
    __pyx_t_18.suboffsets[3] = -1;

__pyx_t_5 = __pyx_memoryview_fromslice(__pyx_t_18, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
    __pyx_t_18.memview = NULL;
    __pyx_t_18.data = NULL;
    __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_convert_vector_to_py_int((__pyx_v_S[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_min_node_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __pyx_f_9cyext_acv_compile_rules(__pyx_t_5, __pyx_t_4, __pyx_t_3, __pyx_t_7, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__Pyx_SetItemInt(__pyx_v_rules, __pyx_v_i, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 0, 0) < 0)) __PYX_ERR(0, 4961, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "cyext_acv.pyx":4962
 *         for i in range(N):
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)
 *             for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 * 
 */
    __pyx_t_22 = (__pyx_v_data.shape[0]);
    __pyx_t_21 = __pyx_t_22;
    for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_21; __pyx_t_24+=1) {
      __pyx_v_j = __pyx_t_24;

      /* "cyext_acv.pyx":4963
 *             rules[i] = compile_rules(partition_byobs[i], data, S[i], min_node_size)
 *             for j in range(data.shape[0]):
 *                 rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), rules, np.array(sdp_data), rules_data
 */
      __pyx_t_18.data = __pyx_v_partition_byobs_data.data;
      __pyx_t_18.memview = __pyx_v_partition_byobs_data.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_18.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_18.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_18.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_18.suboffsets[1] = -1;

__pyx_t_18.shape[2] = __pyx_v_partition_byobs_data.shape[4];
__pyx_t_18.strides[2] = __pyx_v_partition_byobs_data.strides[4];
    __pyx_t_18.suboffsets[2] = -1;

__pyx_t_18.shape[3] = __pyx_v_partition_byobs_data.shape[5];
__pyx_t_18.strides[3] = __pyx_v_partition_byobs_data.strides[5];
    __pyx_t_18.suboffsets[3] = -1;

__pyx_t_1 = __pyx_memoryview_fromslice(__pyx_t_18, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
      __pyx_t_18.memview = NULL;
      __pyx_t_18.data = NULL;
      __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __pyx_convert_vector_to_py_int((__pyx_v_S[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_min_node_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __pyx_f_9cyext_acv_compile_rules(__pyx_t_1, __pyx_t_7, __pyx_t_3, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_3 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_rules_data, __pyx_t_7, __pyx_t_5) < 0)) __PYX_ERR(0, 4963, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }

  /* "cyext_acv.pyx":4965
 *                 rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 * 
 *         return np.array(sdp), rules, np.array(sdp_data), rules_data             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_rules);
  __Pyx_GIVEREF(__pyx_v_rules);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_rules);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_rules_data);
  __Pyx_GIVEREF(__pyx_v_rules_data);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_rules_data);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4926
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_biased(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __Pyx_XDECREF(__pyx_v_rules);
  __Pyx_XDECREF(__pyx_v_rules_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_71compute_sdp_maxrule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_71compute_sdp_maxrule_biased(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 1); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 2); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 3); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 4); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 5); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 6); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 7); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 8); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 9); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 10); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 11); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 12); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, 13); __PYX_ERR(0, 4926, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule_biased") < 0)) __PYX_ERR(0, 4926, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 4926, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 4926, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 4926, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 4926, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 4926, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 4927, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 4927, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 4927, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 4927, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4928, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4928, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 4928, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4928, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 4928, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4926, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_70compute_sdp_maxrule_biased(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_70compute_sdp_maxrule_biased(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 4926, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule_biased(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":4971
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_v0_weights(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition, double[:] & w) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rule_v0_weights(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_partition, __Pyx_memviewslice &__pyx_v_w) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  unsigned int __pyx_v_b;
  CYTHON_UNUSED unsigned int __pyx_v_level;
  unsigned int __pyx_v_it_node;
  unsigned int __pyx_v_i;
  std::vector<int>  __pyx_v_nodes_level;
  std::vector<int>  __pyx_v_nodes_child;
  std::vector<int>  __pyx_v_in_data;
  std::vector<int>  __pyx_v_in_data_b;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  std::vector<int> ::size_type __pyx_t_9;
  std::vector<int> ::size_type __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  size_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  std::vector<int> ::size_type __pyx_t_18;
  std::vector<int> ::size_type __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":4975
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition, double[:] & w) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":4976
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":4979
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int b, level, it_node, i
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":4983
 *     cdef vector[int] nodes_level, nodes_child, in_data, in_data_b
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":4984
 * 
 *     for b in range(n_trees):
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":4985
 *     for b in range(n_trees):
 *         nodes_level.clear()
 *         nodes_level.push_back(0)             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.push_back(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 4985, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":4986
 *         nodes_level.clear()
 *         nodes_level.push_back(0)
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         in_data.clear()
 *         in_data_b.clear()
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":4987
 *         nodes_level.push_back(0)
 *         nodes_child.clear()
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":4988
 *         nodes_child.clear()
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.push_back(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":4989
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":4990
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.push_back(i)             # <<<<<<<<<<<<<<
 *             in_data_b.push_back(i)
 * 
 */
      try {
        __pyx_v_in_data.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4990, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":4991
 *         for i in range(N):
 *             in_data.push_back(i)
 *             in_data_b.push_back(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.push_back(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 4991, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":4993
 *             in_data_b.push_back(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             for it_node in range(nodes_level.size()):
 * 
 */
    __pyx_t_7 = __pyx_v_max_depth;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_8; __pyx_t_4+=1) {
      __pyx_v_level = __pyx_t_4;

      /* "cyext_acv.pyx":4994
 * 
 *         for level in range(max_depth):
 *             for it_node in range(nodes_level.size()):             # <<<<<<<<<<<<<<
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 */
      __pyx_t_9 = __pyx_v_nodes_level.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_10; __pyx_t_5+=1) {
        __pyx_v_it_node = __pyx_t_5;

        /* "cyext_acv.pyx":4996
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
        __pyx_t_11 = __pyx_v_b;
        __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
        __pyx_t_13 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":4997
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_12 * __pyx_v_features.strides[1]) )));
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          __pyx_t_13 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_14 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_13) {

            /* "cyext_acv.pyx":4998
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_16 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 4998, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5000
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
            __pyx_t_14 = 1;
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_14 * __pyx_v_partition.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_17 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_13) {

              /* "cyext_acv.pyx":5001
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_12 = 1;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_14 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_17 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5000
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 1] >= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }

            /* "cyext_acv.pyx":5003
 *                             partition[features[b, nodes_level[it_node]], 1] = thresholds[b, nodes_level[it_node]]
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_18 = __pyx_v_in_data.size();
            __pyx_t_19 = __pyx_t_18;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_19; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":5004
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_17 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_17 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_16 * __pyx_v_data.strides[0]) ) + __pyx_t_12 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":5005
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":5006
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":5004
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] > thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":5007
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":4997
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":5010
 * 
 *                     else:
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:
 */
          /*else*/ {
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            try {
              __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_14 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5010, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5012
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_17 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
            __pyx_t_12 = 0;
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
            __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_17 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_13) {

              /* "cyext_acv.pyx":5013
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]             # <<<<<<<<<<<<<<
 * 
 *                         for i in range(in_data.size()):
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_14 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_15 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
              __pyx_t_17 = 0;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_17 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_16 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5012
 *                         nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                         if partition[features[b, nodes_level[it_node]], 0] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 */
            }

            /* "cyext_acv.pyx":5015
 *                             partition[features[b, nodes_level[it_node]], 0] = thresholds[b, nodes_level[it_node]]
 * 
 *                         for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 */
            __pyx_t_18 = __pyx_v_in_data.size();
            __pyx_t_19 = __pyx_t_18;
            for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_19; __pyx_t_6+=1) {
              __pyx_v_i = __pyx_t_6;

              /* "cyext_acv.pyx":5016
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_16 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_14 = (__pyx_v_in_data[__pyx_v_i]);
              __pyx_t_17 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_16 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
              __pyx_t_13 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_14 * __pyx_v_data.strides[0]) ) + __pyx_t_17 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_13) {

                /* "cyext_acv.pyx":5017
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])             # <<<<<<<<<<<<<<
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b
 */
                (void)(std::remove<std::vector<int> ::iterator,int>(__pyx_v_in_data_b.begin(), __pyx_v_in_data_b.end(), (__pyx_v_in_data[__pyx_v_i])));

                /* "cyext_acv.pyx":5018
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
                __pyx_v_in_data_b.pop_back();

                /* "cyext_acv.pyx":5016
 * 
 *                         for i in range(in_data.size()):
 *                             if data[in_data[i], features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:             # <<<<<<<<<<<<<<
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 */
              }
            }

            /* "cyext_acv.pyx":5019
 *                                 std_remove[vector[int].iterator, int](in_data_b.begin(), in_data_b.end(), in_data[i])
 *                                 in_data_b.pop_back()
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":4996
 *             for it_node in range(nodes_level.size()):
 * 
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, nodes_level[it_node]]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, nodes_level[it_node]]] <= thresholds[b, nodes_level[it_node]]:
 *                         nodes_child.push_back(children_left[b, nodes_level[it_node]])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":5022
 * 
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 */
        /*else*/ {
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_15 * __pyx_v_children_left.strides[0]) ) + __pyx_t_12 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5022, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":5023
 *                 else:
 *                      nodes_child.push_back(children_left[b, nodes_level[it_node]])
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = (__pyx_v_nodes_level[__pyx_v_it_node]);
          try {
            __pyx_v_nodes_child.push_back((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5023, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":5025
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        __pyx_t_13 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_13) {

          /* "cyext_acv.pyx":5026
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":5025
 *                      nodes_child.push_back(children_right[b, nodes_level[it_node]])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
        }
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":5028
 *                     break
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":5030
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
    __pyx_t_13 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_13) {

      /* "cyext_acv.pyx":5031
 * 
 *         if classifier == 1:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 */
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":5032
 *         if classifier == 1:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])             # <<<<<<<<<<<<<<
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 *         else:
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))))));

        /* "cyext_acv.pyx":5033
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(in_data.size()):
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        *((double *) ( /* dim=0 */ (__pyx_v_w.data + __pyx_t_12 * __pyx_v_w.strides[0]) )) += (1. / (__pyx_v_n_trees * __pyx_v_in_data.size()));
      }

      /* "cyext_acv.pyx":5030
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[in_data[i]])
 */
      goto __pyx_L22;
    }

    /* "cyext_acv.pyx":5035
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 *         else:
 *             for i in range(in_data.size()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 */
    /*else*/ {
      __pyx_t_9 = __pyx_v_in_data.size();
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_10; __pyx_t_4+=1) {
        __pyx_v_i = __pyx_t_4;

        /* "cyext_acv.pyx":5036
 *         else:
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)             # <<<<<<<<<<<<<<
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 *     return sdp
 */
        __pyx_t_12 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_16)) ))))) <= __pyx_v_t)));

        /* "cyext_acv.pyx":5037
 *             for i in range(in_data.size()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))             # <<<<<<<<<<<<<<
 *     return sdp
 * 
 */
        __pyx_t_16 = (__pyx_v_in_data[__pyx_v_i]);
        *((double *) ( /* dim=0 */ (__pyx_v_w.data + __pyx_t_16 * __pyx_v_w.strides[0]) )) += (1. / (__pyx_v_n_trees * __pyx_v_in_data.size()));
      }
    }
    __pyx_L22:;
  }

  /* "cyext_acv.pyx":5038
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[in_data[i]])*(y_x - y_data[in_data[i]]) <= t)
 *                 w[in_data[i]] += (1./(n_trees*in_data.size()))
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":4971
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_v0_weights(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition, double[:] & w) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rule_v0_weights", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":5044
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_opti(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_73compute_sdp_maxrule_opti(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_opti(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED PyObject *__pyx_v_rules_data = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_opti", 0);

  /* "cyext_acv.pyx":5048
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5049
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs = buffer
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5049, __pyx_L1_error)

  /* "cyext_acv.pyx":5051
 *         buffer[:, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :] partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 5051, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":5053
 *         cdef double [:, :, :] partition_byobs = buffer
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5054
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 *         buffer_data[:, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__4, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5054, __pyx_L1_error)

  /* "cyext_acv.pyx":5056
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 5056, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":5058
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_2);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":5060
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5061
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":5062
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp_data = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5064
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i, j
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_weights = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5067
 *         cdef int i, j
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":5068
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    __pyx_t_14.data = __pyx_v_X.data;
    __pyx_t_14.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_X.shape[1];
__pyx_t_14.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":5070
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_10.data = __pyx_v_partition_byobs.data;
    __pyx_t_10.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_16.data = __pyx_v_weights.data;
    __pyx_t_16.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_16.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

    /* "cyext_acv.pyx":5068
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_17)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0_weights(__pyx_t_14, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_15)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10, __pyx_t_16);
    __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
    __pyx_t_14.memview = NULL;
    __pyx_t_14.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
    __pyx_t_16.memview = NULL;
    __pyx_t_16.data = NULL;

    /* "cyext_acv.pyx":5072
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 5072, __pyx_L8_error) }
          __pyx_t_18 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_20 = (__pyx_t_18 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_20 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_15, __pyx_t_17, __pyx_t_21, __pyx_t_22) firstprivate(__pyx_t_10, __pyx_t_16)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_20; __pyx_t_19++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_19);

                              /* "cyext_acv.pyx":5073
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              __pyx_t_15 = __pyx_v_i;
                              __pyx_t_17 = __pyx_v_j;
                              __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_weights.data + __pyx_t_15 * __pyx_v_weights.strides[0]) ) + __pyx_t_17 * __pyx_v_weights.strides[1]) ))) != 0.0) != 0);
                              if (__pyx_t_21) {

                                /* "cyext_acv.pyx":5074
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_16.data = __pyx_v_data.data;
                                __pyx_t_16.memview = __pyx_v_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_data.shape[1];
__pyx_t_16.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

                                /* "cyext_acv.pyx":5076
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)
 */
                                __pyx_t_10.data = __pyx_v_partition_byobs_data.data;
                                __pyx_t_10.memview = __pyx_v_partition_byobs_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_15 = __pyx_v_i;

                                /* "cyext_acv.pyx":5074
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_22 = __pyx_v_j;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_15 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_22 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10);
                                __PYX_XDEC_MEMVIEW(&__pyx_t_16, 0);
                                __pyx_t_16.memview = NULL;
                                __pyx_t_16.data = NULL;
                                __PYX_XDEC_MEMVIEW(&__pyx_t_10, 0);
                                __pyx_t_10.memview = NULL;
                                __pyx_t_10.data = NULL;

                                /* "cyext_acv.pyx":5073
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":5072
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_v0(data[j], y_X[i], data, y_data, S[i],
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":5078
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs_data, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_23)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_23);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_23, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_weights, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_23);
    if (likely(__pyx_t_24)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_23, function);
    }
  }
  __pyx_t_5 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_23, __pyx_t_24, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_23, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = PyTuple_New(5); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_23, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_23, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_23, 4, __pyx_t_5);
  __pyx_t_7 = 0;
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_23;
  __pyx_t_23 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5044
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_opti(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_opti", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __Pyx_XDECREF(__pyx_v_rules_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_73compute_sdp_maxrule_opti(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_73compute_sdp_maxrule_opti(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_opti (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 1); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 2); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 3); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 4); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 5); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 6); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 7); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 8); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 9); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 10); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 11); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 12); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, 13); __PYX_ERR(0, 5044, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule_opti") < 0)) __PYX_ERR(0, 5044, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5044, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5044, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5044, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5044, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5044, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5045, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5045, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5045, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5045, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5046, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5046, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5046, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5046, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5046, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_opti", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5044, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_opti", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_72compute_sdp_maxrule_opti(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_72compute_sdp_maxrule_opti(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_opti", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5044, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule_opti(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_opti", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5085
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_biased_v2(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_75compute_sdp_maxrule_biased_v2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_biased_v2(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_rules_data = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  int __pyx_t_23;
  __Pyx_memviewslice __pyx_t_24 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_25;
  int __pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased_v2", 0);

  /* "cyext_acv.pyx":5089
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5090
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5090, __pyx_L1_error)

  /* "cyext_acv.pyx":5093
 * 
 *         cdef double [:, :, :] partition_byobs
 *         partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 5093, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":5096
 * 
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_children_left.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_max_depth); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_8);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_9) < 0) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5097
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer_data[:, :, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :, :] partition_byobs_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__6, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5097, __pyx_L1_error)

  /* "cyext_acv.pyx":5100
 * 
 *         cdef double [:, :, :, :, :, :] partition_byobs_data
 *         partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5100, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5102
 *         partition_byobs_data = buffer_data
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_int_2);
  __pyx_t_5 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_3) < 0) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules_data = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cyext_acv.pyx":5104
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5105
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 5105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sdp = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":5106
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
  __pyx_t_9 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 5106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sdp_data = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":5108
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i, j
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shape, __pyx_t_9) < 0) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 5108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_weights = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "cyext_acv.pyx":5111
 *         cdef int i, j
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_13 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "cyext_acv.pyx":5112
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    __pyx_t_16.data = __pyx_v_X.data;
    __pyx_t_16.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_X.shape[1];
__pyx_t_16.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

    /* "cyext_acv.pyx":5114
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_12.data = __pyx_v_partition_byobs.data;
    __pyx_t_12.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_12, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_12.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_12.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_12.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_12.suboffsets[0] = -1;

__pyx_t_12.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_12.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_12.suboffsets[1] = -1;

__pyx_t_18.data = __pyx_v_weights.data;
    __pyx_t_18.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_18.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_19 = __pyx_v_i;

    /* "cyext_acv.pyx":5112
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_v0_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_19)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_v0_weights(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_12, __pyx_t_18);
    __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
    __pyx_t_16.memview = NULL;
    __pyx_t_16.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
    __pyx_t_12.memview = NULL;
    __pyx_t_12.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
    __pyx_t_18.memview = NULL;
    __pyx_t_18.data = NULL;

    /* "cyext_acv.pyx":5116
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 5116, __pyx_L8_error) }
          __pyx_t_20 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_22 = (__pyx_t_20 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_22 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_17, __pyx_t_19, __pyx_t_23, __pyx_t_25) firstprivate(__pyx_t_18, __pyx_t_24)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_21 = 0; __pyx_t_21 < __pyx_t_22; __pyx_t_21++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_21);

                              /* "cyext_acv.pyx":5117
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              __pyx_t_17 = __pyx_v_i;
                              __pyx_t_19 = __pyx_v_j;
                              __pyx_t_23 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_weights.data + __pyx_t_17 * __pyx_v_weights.strides[0]) ) + __pyx_t_19 * __pyx_v_weights.strides[1]) ))) != 0.0) != 0);
                              if (__pyx_t_23) {

                                /* "cyext_acv.pyx":5118
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_18.data = __pyx_v_data.data;
                                __pyx_t_18.memview = __pyx_v_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_18, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_18.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_18.shape[0] = __pyx_v_data.shape[1];
__pyx_t_18.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_18.suboffsets[0] = -1;

__pyx_t_19 = __pyx_v_i;

                                /* "cyext_acv.pyx":5120
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
                                __pyx_t_24.data = __pyx_v_partition_byobs_data.data;
                                __pyx_t_24.memview = __pyx_v_partition_byobs_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_24, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_24.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_24.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_24.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_24.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_24.suboffsets[0] = -1;

__pyx_t_24.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_24.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_24.suboffsets[1] = -1;

__pyx_t_24.shape[2] = __pyx_v_partition_byobs_data.shape[4];
__pyx_t_24.strides[2] = __pyx_v_partition_byobs_data.strides[4];
    __pyx_t_24.suboffsets[2] = -1;

__pyx_t_24.shape[3] = __pyx_v_partition_byobs_data.shape[5];
__pyx_t_24.strides[3] = __pyx_v_partition_byobs_data.strides[5];
    __pyx_t_24.suboffsets[3] = -1;

__pyx_t_17 = __pyx_v_i;

                                /* "cyext_acv.pyx":5118
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_25 = __pyx_v_j;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_17 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_25 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_biased(__pyx_t_18, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_19)) ))), (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_24);
                                __PYX_XDEC_MEMVIEW(&__pyx_t_18, 0);
                                __pyx_t_18.memview = NULL;
                                __pyx_t_18.data = NULL;
                                __PYX_XDEC_MEMVIEW(&__pyx_t_24, 0);
                                __pyx_t_24.memview = NULL;
                                __pyx_t_24.data = NULL;

                                /* "cyext_acv.pyx":5117
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":5116
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule_biased(data[j], y_X[i], S[i],
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":5122
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             for j in range(data.shape[0]):
 *                 if weights[i, j] != 0:
 */
  __pyx_t_13 = __pyx_v_N;
  __pyx_t_14 = __pyx_t_13;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "cyext_acv.pyx":5123
 * 
 *         for i in range(N):
 *             for j in range(data.shape[0]):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 */
    __pyx_t_22 = (__pyx_v_data.shape[0]);
    __pyx_t_21 = __pyx_t_22;
    for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_21; __pyx_t_26+=1) {
      __pyx_v_j = __pyx_t_26;

      /* "cyext_acv.pyx":5124
 *         for i in range(N):
 *             for j in range(data.shape[0]):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 * 
 */
      __pyx_t_19 = __pyx_v_i;
      __pyx_t_25 = __pyx_v_j;
      __pyx_t_23 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_weights.data + __pyx_t_19 * __pyx_v_weights.strides[0]) ) + __pyx_t_25 * __pyx_v_weights.strides[1]) ))) != 0.0) != 0);
      if (__pyx_t_23) {

        /* "cyext_acv.pyx":5125
 *             for j in range(data.shape[0]):
 *                 if weights[i, j] != 0:
 *                     rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), rules_data, np.array(weights)
 */
        __pyx_t_24.data = __pyx_v_partition_byobs_data.data;
        __pyx_t_24.memview = __pyx_v_partition_byobs_data.memview;
        __PYX_INC_MEMVIEW(&__pyx_t_24, 0);
        {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_24.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_24.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_24.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_24.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_24.suboffsets[0] = -1;

__pyx_t_24.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_24.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_24.suboffsets[1] = -1;

__pyx_t_24.shape[2] = __pyx_v_partition_byobs_data.shape[4];
__pyx_t_24.strides[2] = __pyx_v_partition_byobs_data.strides[4];
    __pyx_t_24.suboffsets[2] = -1;

__pyx_t_24.shape[3] = __pyx_v_partition_byobs_data.shape[5];
__pyx_t_24.strides[3] = __pyx_v_partition_byobs_data.strides[5];
    __pyx_t_24.suboffsets[3] = -1;

__pyx_t_9 = __pyx_memoryview_fromslice(__pyx_t_24, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __PYX_XDEC_MEMVIEW(&__pyx_t_24, 1);
        __pyx_t_24.memview = NULL;
        __pyx_t_24.data = NULL;
        __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __pyx_convert_vector_to_py_int((__pyx_v_S[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_min_node_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __pyx_f_9cyext_acv_compile_rules(__pyx_t_9, __pyx_t_3, __pyx_t_1, __pyx_t_7, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_1 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_rules_data, __pyx_t_3, __pyx_t_8) < 0)) __PYX_ERR(0, 5125, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "cyext_acv.pyx":5124
 *         for i in range(N):
 *             for j in range(data.shape[0]):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 * 
 */
      }
    }
  }

  /* "cyext_acv.pyx":5127
 *                     rules_data[i, j] = compile_rules(partition_byobs_data[i, j], data, S[i], min_node_size)
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), rules_data, np.array(weights)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_weights, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_9 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_7);
  __Pyx_INCREF(__pyx_v_rules_data);
  __Pyx_GIVEREF(__pyx_v_rules_data);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_rules_data);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_9);
  __pyx_t_8 = 0;
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5085
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_biased_v2(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_24, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased_v2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __Pyx_XDECREF(__pyx_v_rules_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_75compute_sdp_maxrule_biased_v2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_75compute_sdp_maxrule_biased_v2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased_v2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 1); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 2); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 3); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 4); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 5); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 6); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 7); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 8); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 9); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 10); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 11); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 12); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, 13); __PYX_ERR(0, 5085, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule_biased_v2") < 0)) __PYX_ERR(0, 5085, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5085, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5085, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5085, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5085, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5085, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5086, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5086, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5086, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5086, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5087, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5087, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5087, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5087, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5087, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_biased_v2", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5085, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased_v2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_74compute_sdp_maxrule_biased_v2(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_74compute_sdp_maxrule_biased_v2(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_biased_v2", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5085, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule_biased_v2(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_biased_v2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5134
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rule(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_partition) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  int __pyx_v_b;
  CYTHON_UNUSED int __pyx_v_level;
  int __pyx_v_i;
  int __pyx_v_it_node;
  std::set<int>  __pyx_v_nodes_level;
  std::set<int>  __pyx_v_nodes_child;
  std::set<int>  __pyx_v_in_data;
  std::set<int>  __pyx_v_in_data_b;
  std::set<int> ::iterator __pyx_v_it;
  std::set<int> ::iterator __pyx_v_it_point;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":5138
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":5139
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":5142
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int b, level, i, it_node
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":5148
 *     cdef set[int].iterator it, it_point
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         nodes_level.clear()
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":5149
 * 
 *     for b in range(n_trees):
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.insert(0)
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":5150
 *     for b in range(n_trees):
 *         nodes_child.clear()
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.insert(0)
 * 
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":5151
 *         nodes_child.clear()
 *         nodes_level.clear()
 *         nodes_level.insert(0)             # <<<<<<<<<<<<<<
 * 
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.insert(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 5151, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":5153
 *         nodes_level.insert(0)
 * 
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":5154
 * 
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.insert(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":5155
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.insert(i)
 *             in_data_b.insert(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":5156
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.insert(i)             # <<<<<<<<<<<<<<
 *             in_data_b.insert(i)
 * 
 */
      try {
        __pyx_v_in_data.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5156, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":5157
 *         for i in range(N):
 *             in_data.insert(i)
 *             in_data_b.insert(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5157, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":5159
 *             in_data_b.insert(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 */
    __pyx_t_6 = __pyx_v_max_depth;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_level = __pyx_t_8;

      /* "cyext_acv.pyx":5160
 * 
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()             # <<<<<<<<<<<<<<
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 */
      __pyx_v_it_point = __pyx_v_nodes_level.begin();

      /* "cyext_acv.pyx":5161
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):             # <<<<<<<<<<<<<<
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it_point != __pyx_v_nodes_level.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5162
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 */
        __pyx_v_it_node = (*__pyx_v_it_point);

        /* "cyext_acv.pyx":5163
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
        __pyx_t_10 = __pyx_v_b;
        __pyx_t_11 = __pyx_v_it_node;
        __pyx_t_9 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5164
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_14 = __pyx_v_it_node;
          __pyx_t_9 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_12 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_9) {

            /* "cyext_acv.pyx":5165
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_14 * __pyx_v_children_left.strides[0]) ) + __pyx_t_13 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5165, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5167
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 */
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_14 = __pyx_v_it_node;
            __pyx_t_10 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_13 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
            __pyx_t_11 = 1;
            __pyx_t_12 = __pyx_v_b;
            __pyx_t_15 = __pyx_v_it_node;
            __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_10 * __pyx_v_partition.strides[0]) ) + __pyx_t_11 * __pyx_v_partition.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_12 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_15 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_9) {

              /* "cyext_acv.pyx":5168
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = __pyx_v_it_node;
              __pyx_t_14 = __pyx_v_b;
              __pyx_t_13 = __pyx_v_it_node;
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_13 * __pyx_v_features.strides[1]) )));
              __pyx_t_10 = 1;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_11 * __pyx_v_partition.strides[0]) ) + __pyx_t_10 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5167
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 */
            }

            /* "cyext_acv.pyx":5170
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5171
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5172
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_13 = (*__pyx_v_it);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_12 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
              __pyx_t_10 = __pyx_v_b;
              __pyx_t_11 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_13 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_10 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_11 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5173
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5172
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5174
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5175
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":5164
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":5178
 * 
 *                     else:
 *                         nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:
 */
          /*else*/ {
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_10 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_10 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5178, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5180
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 */
            __pyx_t_10 = __pyx_v_b;
            __pyx_t_11 = __pyx_v_it_node;
            __pyx_t_15 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )));
            __pyx_t_12 = 0;
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_15 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_14 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_13 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_9) {

              /* "cyext_acv.pyx":5181
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
              __pyx_t_13 = __pyx_v_b;
              __pyx_t_14 = __pyx_v_it_node;
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_10 = __pyx_v_it_node;
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = 0;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_15 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5180
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 */
            }

            /* "cyext_acv.pyx":5183
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5184
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5185
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_14 = __pyx_v_b;
              __pyx_t_13 = __pyx_v_it_node;
              __pyx_t_10 = (*__pyx_v_it);
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_13 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_10 * __pyx_v_data.strides[0]) ) + __pyx_t_11 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5186
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5185
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5187
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 *                 else:
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5188
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":5163
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":5190
 *                         in_data = in_data_b
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 */
        /*else*/ {
          __pyx_t_12 = __pyx_v_b;
          __pyx_t_15 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_12 * __pyx_v_children_left.strides[0]) ) + __pyx_t_15 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5190, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":5191
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 *                      nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5191, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":5193
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        __pyx_t_9 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5194
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 *                 inc(it_point)
 * 
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":5193
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        }

        /* "cyext_acv.pyx":5195
 *                 if in_data.size() < min_node_size:
 *                     break
 *                 inc(it_point)             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
        (void)((++__pyx_v_it_point));
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":5197
 *                 inc(it_point)
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":5199
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
    __pyx_t_9 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_9) {

      /* "cyext_acv.pyx":5200
 * 
 *         if classifier == 1:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 */
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5201
 *         if classifier == 1:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5202
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])             # <<<<<<<<<<<<<<
 *                 inc(it)
 *         else:
 */
        __pyx_t_12 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))))));

        /* "cyext_acv.pyx":5203
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 inc(it)             # <<<<<<<<<<<<<<
 *         else:
 *             it = in_data.begin()
 */
        (void)((++__pyx_v_it));
      }

      /* "cyext_acv.pyx":5199
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
      goto __pyx_L22;
    }

    /* "cyext_acv.pyx":5205
 *                 inc(it)
 *         else:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 */
    /*else*/ {
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5206
 *         else:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5207
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)             # <<<<<<<<<<<<<<
 *                 inc(it)
 * 
 */
        __pyx_t_12 = (*__pyx_v_it);
        __pyx_t_15 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_15)) ))))) <= __pyx_v_t)));

        /* "cyext_acv.pyx":5208
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 inc(it)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        (void)((++__pyx_v_it));
      }
    }
    __pyx_L22:;
  }

  /* "cyext_acv.pyx":5210
 *                 inc(it)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5134
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rule", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":5216
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule(double[:, :] & X, double[::1] & y_X,  double[:, :] & data, double[::1] & y_data, vector[vector[int]] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_77compute_sdp_rule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rule(__Pyx_memviewslice &__pyx_v_X, __Pyx_memviewslice &__pyx_v_y_X, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<std::vector<int> >  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_12;
  __Pyx_memviewslice __pyx_t_13 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule", 0);

  /* "cyext_acv.pyx":5220
 *         int & max_depth, int & min_node_size, int & classifier, double & t):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 *         cdef double [:, :, :] partition_byobs = buffer
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5221
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 *         cdef double [:, :, :] partition_byobs = buffer
 * 
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5221, __pyx_L1_error)

  /* "cyext_acv.pyx":5222
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10
 *         cdef double [:, :, :] partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 5222, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":5224
 *         cdef double [:, :, :] partition_byobs = buffer
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5225
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 5225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cyext_acv.pyx":5228
 *         cdef int i
 * 
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_8 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_10 = (__pyx_t_8 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_10 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_12, __pyx_t_14) firstprivate(__pyx_t_11, __pyx_t_13)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_10; __pyx_t_9++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_9);

                            /* "cyext_acv.pyx":5229
 * 
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rule(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
                            __pyx_t_11.data = __pyx_v_X.data;
                            __pyx_t_11.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_11, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_11.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_11.shape[0] = __pyx_v_X.shape[1];
__pyx_t_11.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_11.suboffsets[0] = -1;

__pyx_t_12 = __pyx_v_i;

                            /* "cyext_acv.pyx":5231
 *             sdp[i] = single_compute_sdp_rule(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs)
 */
                            __pyx_t_13.data = __pyx_v_partition_byobs.data;
                            __pyx_t_13.memview = __pyx_v_partition_byobs.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_13, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_13.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_13.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_13.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_13.suboffsets[0] = -1;

__pyx_t_13.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_13.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_13.suboffsets[1] = -1;

__pyx_t_14 = __pyx_v_i;

                            /* "cyext_acv.pyx":5229
 * 
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rule(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 */
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_14)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule(__pyx_t_11, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_12)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_13);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_11, 0);
                            __pyx_t_11.memview = NULL;
                            __pyx_t_11.data = NULL;
                            __PYX_XDEC_MEMVIEW(&__pyx_t_13, 0);
                            __pyx_t_13.memview = NULL;
                            __pyx_t_13.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":5228
 *         cdef int i
 * 
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":5233
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i])
 * 
 *         return np.array(sdp), np.array(partition_byobs)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5216
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rule(double[:, :] & X, double[::1] & y_X,  double[:, :] & data, double[::1] & y_data, vector[vector[int]] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_13, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_77compute_sdp_rule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_77compute_sdp_rule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 1); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 2); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 3); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 4); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 5); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 6); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 7); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 8); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 9); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 10); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 11); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, 12); __PYX_ERR(0, 5216, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rule") < 0)) __PYX_ERR(0, 5216, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5216, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5216, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5216, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5216, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5216, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5217, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5217, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5217, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5217, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5218, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5218, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5218, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5218, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rule", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5216, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_76compute_sdp_rule(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_76compute_sdp_rule(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rule", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5216, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rule(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5239
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_weights(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition,
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rule_weights(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, __Pyx_memviewslice &__pyx_v_partition, __Pyx_memviewslice &__pyx_v_w) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  int __pyx_v_b;
  CYTHON_UNUSED int __pyx_v_level;
  int __pyx_v_i;
  int __pyx_v_it_node;
  std::set<int>  __pyx_v_nodes_level;
  std::set<int>  __pyx_v_nodes_child;
  std::set<int>  __pyx_v_in_data;
  std::set<int>  __pyx_v_in_data_b;
  std::set<int> ::iterator __pyx_v_it;
  std::set<int> ::iterator __pyx_v_it_point;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":5244
 *         double[:] & w) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":5245
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":5248
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int b, level, i, it_node
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":5254
 *     cdef set[int].iterator it, it_point
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         nodes_level.clear()
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":5255
 * 
 *     for b in range(n_trees):
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.insert(0)
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":5256
 *     for b in range(n_trees):
 *         nodes_child.clear()
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.insert(0)
 * 
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":5257
 *         nodes_child.clear()
 *         nodes_level.clear()
 *         nodes_level.insert(0)             # <<<<<<<<<<<<<<
 * 
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.insert(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 5257, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":5259
 *         nodes_level.insert(0)
 * 
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":5260
 * 
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.insert(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":5261
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.insert(i)
 *             in_data_b.insert(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":5262
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.insert(i)             # <<<<<<<<<<<<<<
 *             in_data_b.insert(i)
 * 
 */
      try {
        __pyx_v_in_data.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5262, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":5263
 *         for i in range(N):
 *             in_data.insert(i)
 *             in_data_b.insert(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5263, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":5265
 *             in_data_b.insert(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 */
    __pyx_t_6 = __pyx_v_max_depth;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_level = __pyx_t_8;

      /* "cyext_acv.pyx":5266
 * 
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()             # <<<<<<<<<<<<<<
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 */
      __pyx_v_it_point = __pyx_v_nodes_level.begin();

      /* "cyext_acv.pyx":5267
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):             # <<<<<<<<<<<<<<
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it_point != __pyx_v_nodes_level.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5268
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 */
        __pyx_v_it_node = (*__pyx_v_it_point);

        /* "cyext_acv.pyx":5269
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
        __pyx_t_10 = __pyx_v_b;
        __pyx_t_11 = __pyx_v_it_node;
        __pyx_t_9 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5270
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_14 = __pyx_v_it_node;
          __pyx_t_9 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_12 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_9) {

            /* "cyext_acv.pyx":5271
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_14 * __pyx_v_children_left.strides[0]) ) + __pyx_t_13 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5271, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5273
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 */
            __pyx_t_13 = __pyx_v_b;
            __pyx_t_14 = __pyx_v_it_node;
            __pyx_t_10 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_13 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
            __pyx_t_11 = 1;
            __pyx_t_12 = __pyx_v_b;
            __pyx_t_15 = __pyx_v_it_node;
            __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_10 * __pyx_v_partition.strides[0]) ) + __pyx_t_11 * __pyx_v_partition.strides[1]) ))) >= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_12 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_15 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_9) {

              /* "cyext_acv.pyx":5274
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = __pyx_v_it_node;
              __pyx_t_14 = __pyx_v_b;
              __pyx_t_13 = __pyx_v_it_node;
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_13 * __pyx_v_features.strides[1]) )));
              __pyx_t_10 = 1;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_11 * __pyx_v_partition.strides[0]) ) + __pyx_t_10 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5273
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         if partition[features[b, it_node], 1] >= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 */
            }

            /* "cyext_acv.pyx":5276
 *                             partition[features[b, it_node], 1] = thresholds[b, it_node]
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5277
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5278
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_13 = (*__pyx_v_it);
              __pyx_t_14 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_12 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
              __pyx_t_10 = __pyx_v_b;
              __pyx_t_11 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_13 * __pyx_v_data.strides[0]) ) + __pyx_t_14 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_10 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_11 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5279
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5278
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5280
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5281
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":5270
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":5284
 * 
 *                     else:
 *                         nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:
 */
          /*else*/ {
            __pyx_t_11 = __pyx_v_b;
            __pyx_t_10 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_10 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5284, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5286
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 */
            __pyx_t_10 = __pyx_v_b;
            __pyx_t_11 = __pyx_v_it_node;
            __pyx_t_15 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )));
            __pyx_t_12 = 0;
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_15 * __pyx_v_partition.strides[0]) ) + __pyx_t_12 * __pyx_v_partition.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_14 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_13 * __pyx_v_thresholds.strides[1]) )))) != 0);
            if (__pyx_t_9) {

              /* "cyext_acv.pyx":5287
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
              __pyx_t_13 = __pyx_v_b;
              __pyx_t_14 = __pyx_v_it_node;
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_10 = __pyx_v_it_node;
              __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = 0;
              *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_partition.data + __pyx_t_12 * __pyx_v_partition.strides[0]) ) + __pyx_t_15 * __pyx_v_partition.strides[1]) )) = (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )));

              /* "cyext_acv.pyx":5286
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         if partition[features[b, it_node], 0] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 */
            }

            /* "cyext_acv.pyx":5289
 *                             partition[features[b, it_node], 0] = thresholds[b, it_node]
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5290
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5291
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_14 = __pyx_v_b;
              __pyx_t_13 = __pyx_v_it_node;
              __pyx_t_10 = (*__pyx_v_it);
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_14 * __pyx_v_features.strides[0]) ) + __pyx_t_13 * __pyx_v_features.strides[1]) )));
              __pyx_t_15 = __pyx_v_b;
              __pyx_t_12 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_10 * __pyx_v_data.strides[0]) ) + __pyx_t_11 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_15 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_12 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5292
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5291
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5293
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 *                 else:
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5294
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":5269
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":5296
 *                         in_data = in_data_b
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 */
        /*else*/ {
          __pyx_t_12 = __pyx_v_b;
          __pyx_t_15 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_12 * __pyx_v_children_left.strides[0]) ) + __pyx_t_15 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5296, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":5297
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 *                      nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_15 = __pyx_v_b;
          __pyx_t_12 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5297, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":5299
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        __pyx_t_9 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5300
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 *                 inc(it_point)
 * 
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":5299
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        }

        /* "cyext_acv.pyx":5301
 *                 if in_data.size() < min_node_size:
 *                     break
 *                 inc(it_point)             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
        (void)((++__pyx_v_it_point));
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":5303
 *                 inc(it_point)
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":5305
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
    __pyx_t_9 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_9) {

      /* "cyext_acv.pyx":5306
 * 
 *         if classifier == 1:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 */
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5307
 *         if classifier == 1:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5308
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])             # <<<<<<<<<<<<<<
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 *                 inc(it)
 */
        __pyx_t_12 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) ))))));

        /* "cyext_acv.pyx":5309
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))             # <<<<<<<<<<<<<<
 *                 inc(it)
 *         else:
 */
        __pyx_t_12 = (*__pyx_v_it);
        *((double *) ( /* dim=0 */ (__pyx_v_w.data + __pyx_t_12 * __pyx_v_w.strides[0]) )) += (1. / (__pyx_v_n_trees * __pyx_v_in_data.size()));

        /* "cyext_acv.pyx":5310
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 *                 inc(it)             # <<<<<<<<<<<<<<
 *         else:
 *             it = in_data.begin()
 */
        (void)((++__pyx_v_it));
      }

      /* "cyext_acv.pyx":5305
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
      goto __pyx_L22;
    }

    /* "cyext_acv.pyx":5312
 *                 inc(it)
 *         else:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 */
    /*else*/ {
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5313
 *         else:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5314
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)             # <<<<<<<<<<<<<<
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 *                 inc(it)
 */
        __pyx_t_12 = (*__pyx_v_it);
        __pyx_t_15 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_12)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_15)) ))))) <= __pyx_v_t)));

        /* "cyext_acv.pyx":5315
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))             # <<<<<<<<<<<<<<
 *                 inc(it)
 * 
 */
        __pyx_t_15 = (*__pyx_v_it);
        *((double *) ( /* dim=0 */ (__pyx_v_w.data + __pyx_t_15 * __pyx_v_w.strides[0]) )) += (1. / (__pyx_v_n_trees * __pyx_v_in_data.size()));

        /* "cyext_acv.pyx":5316
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 w[deref(it)] += (1./(n_trees*in_data.size()))
 *                 inc(it)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        (void)((++__pyx_v_it));
      }
    }
    __pyx_L22:;
  }

  /* "cyext_acv.pyx":5318
 *                 inc(it)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5239
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rule_weights(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t, double[:, :] & partition,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rule_weights", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":5325
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_79compute_sdp_maxrule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED PyObject *__pyx_v_rules_data = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule", 0);

  /* "cyext_acv.pyx":5329
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5330
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs = buffer
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5330, __pyx_L1_error)

  /* "cyext_acv.pyx":5332
 *         buffer[:, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :] partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 5332, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":5334
 *         cdef double [:, :, :] partition_byobs = buffer
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5335
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 *         buffer_data[:, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__4, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5335, __pyx_L1_error)

  /* "cyext_acv.pyx":5337
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 5337, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":5339
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_2);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":5341
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5342
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 5342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":5343
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp_data = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5345
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i, j
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_weights = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5348
 *         cdef int i, j
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __pyx_t_11 = __pyx_v_N;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":5349
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    __pyx_t_14.data = __pyx_v_X.data;
    __pyx_t_14.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_X.shape[1];
__pyx_t_14.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":5351
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_10.data = __pyx_v_partition_byobs.data;
    __pyx_t_10.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_16.data = __pyx_v_weights.data;
    __pyx_t_16.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_16.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

    /* "cyext_acv.pyx":5349
 * 
 *         for i in range(N):
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_17)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_weights(__pyx_t_14, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_15)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10, __pyx_t_16);
    __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
    __pyx_t_14.memview = NULL;
    __pyx_t_14.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
    __pyx_t_16.memview = NULL;
    __pyx_t_16.data = NULL;

    /* "cyext_acv.pyx":5353
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 5353, __pyx_L8_error) }
          __pyx_t_18 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_20 = (__pyx_t_18 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_20 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_15, __pyx_t_17, __pyx_t_21, __pyx_t_22) firstprivate(__pyx_t_10, __pyx_t_16)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_20; __pyx_t_19++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_19);

                              /* "cyext_acv.pyx":5354
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              __pyx_t_15 = __pyx_v_i;
                              __pyx_t_17 = __pyx_v_j;
                              __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_weights.data + __pyx_t_15 * __pyx_v_weights.strides[0]) ) + __pyx_t_17 * __pyx_v_weights.strides[1]) ))) != 0.0) != 0);
                              if (__pyx_t_21) {

                                /* "cyext_acv.pyx":5355
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_16.data = __pyx_v_data.data;
                                __pyx_t_16.memview = __pyx_v_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_data.shape[1];
__pyx_t_16.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

                                /* "cyext_acv.pyx":5357
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)
 */
                                __pyx_t_10.data = __pyx_v_partition_byobs_data.data;
                                __pyx_t_10.memview = __pyx_v_partition_byobs_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_15 = __pyx_v_i;

                                /* "cyext_acv.pyx":5355
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_22 = __pyx_v_j;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_15 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_22 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10);
                                __PYX_XDEC_MEMVIEW(&__pyx_t_16, 0);
                                __pyx_t_16.memview = NULL;
                                __pyx_t_16.data = NULL;
                                __PYX_XDEC_MEMVIEW(&__pyx_t_10, 0);
                                __pyx_t_10.memview = NULL;
                                __pyx_t_10.data = NULL;

                                /* "cyext_acv.pyx":5354
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":5353
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }
  }

  /* "cyext_acv.pyx":5359
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs_data, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_23)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_23);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_23, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_weights, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_23);
    if (likely(__pyx_t_24)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_23, function);
    }
  }
  __pyx_t_5 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_23, __pyx_t_24, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_23, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = PyTuple_New(5); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_23, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_23, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_23, 4, __pyx_t_5);
  __pyx_t_7 = 0;
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_23;
  __pyx_t_23 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5325
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __Pyx_XDECREF(__pyx_v_rules_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_79compute_sdp_maxrule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_79compute_sdp_maxrule(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 1); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 2); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 3); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 4); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 5); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 6); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 7); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 8); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 9); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 10); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 11); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 12); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, 13); __PYX_ERR(0, 5325, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule") < 0)) __PYX_ERR(0, 5325, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5325, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5325, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5325, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5325, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5325, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5326, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5326, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5326, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5326, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5327, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5327, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5327, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5327, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5327, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_78compute_sdp_maxrule(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_78compute_sdp_maxrule(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5325, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5365
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_verbose(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

static PyObject *__pyx_pw_9cyext_acv_81compute_sdp_maxrule_verbose(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_maxrule_verbose(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED double &__pyx_v_pi, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_buffer = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_buffer_data = NULL;
  __Pyx_memviewslice __pyx_v_partition_byobs_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED PyObject *__pyx_v_rules_data = NULL;
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_weights = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  __Pyx_memviewslice __pyx_t_14 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_verbose", 0);

  /* "cyext_acv.pyx":5369
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer[:, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5370
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs = buffer
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer, __pyx_tuple__3, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5370, __pyx_L1_error)

  /* "cyext_acv.pyx":5372
 *         buffer[:, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :] partition_byobs = buffer             # <<<<<<<<<<<<<<
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(__pyx_v_buffer, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 5372, __pyx_L1_error)
  __pyx_v_partition_byobs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cyext_acv.pyx":5374
 *         cdef double [:, :, :] partition_byobs = buffer
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_int_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_1e_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_buffer_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5375
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 *         buffer_data[:, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 */
  if (unlikely(PyObject_SetItem(__pyx_v_buffer_data, __pyx_tuple__4, __pyx_float_neg_1e_10) < 0)) __PYX_ERR(0, 5375, __pyx_L1_error)

  /* "cyext_acv.pyx":5377
 *         buffer_data[:, :, :, 0] = -1e+10
 * 
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data             # <<<<<<<<<<<<<<
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(__pyx_v_buffer_data, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 5377, __pyx_L1_error)
  __pyx_v_partition_byobs_data = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "cyext_acv.pyx":5379
 *         cdef double [:, :, :, :] partition_byobs_data = buffer_data
 * 
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))             # <<<<<<<<<<<<<<
 * 
 *         cdef int N = X.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_2);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rules_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":5381
 *         rules_data = np.zeros(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5382
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(0, 5382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;

  /* "cyext_acv.pyx":5383
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))             # <<<<<<<<<<<<<<
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sdp_data = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5385
 *         cdef double[:, :] sdp_data = np.zeros((N, data.shape[0]))
 * 
 *         cdef double[:, :] weights = np.zeros(shape=(N, data.shape[0]))             # <<<<<<<<<<<<<<
 *         cdef int i, j
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_t_7) < 0) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 5385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_weights = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "cyext_acv.pyx":5388
 *         cdef int i, j
 * 
 *         for i in tqdm(range(N)):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5388, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 5388, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 5388, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_12(__pyx_t_2);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 5388, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5388, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_i = __pyx_t_13;

    /* "cyext_acv.pyx":5389
 * 
 *         for i in tqdm(range(N)):
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    __pyx_t_14.data = __pyx_v_X.data;
    __pyx_t_14.memview = __pyx_v_X.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_14, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_14.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_14.shape[0] = __pyx_v_X.shape[1];
__pyx_t_14.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_14.suboffsets[0] = -1;

__pyx_t_15 = __pyx_v_i;

    /* "cyext_acv.pyx":5391
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])             # <<<<<<<<<<<<<<
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 */
    __pyx_t_10.data = __pyx_v_partition_byobs.data;
    __pyx_t_10.memview = __pyx_v_partition_byobs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs.shape[1];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs.strides[1];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs.shape[2];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs.strides[2];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_16.data = __pyx_v_weights.data;
    __pyx_t_16.memview = __pyx_v_weights.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
    {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_weights.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_weights.shape[1];
__pyx_t_16.strides[0] = __pyx_v_weights.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

    /* "cyext_acv.pyx":5389
 * 
 *         for i in tqdm(range(N)):
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 */
    *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_17)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule_weights(__pyx_t_14, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_15)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10, __pyx_t_16);
    __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
    __pyx_t_14.memview = NULL;
    __pyx_t_14.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
    __pyx_t_16.memview = NULL;
    __pyx_t_16.data = NULL;

    /* "cyext_acv.pyx":5393
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundMemoryviewSliceNogil("data"); __PYX_ERR(0, 5393, __pyx_L8_error) }
          __pyx_t_18 = (__pyx_v_data.shape[0]);
          if ((1 == 0)) abort();
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              __pyx_t_20 = (__pyx_t_18 - 0 + 1 - 1/abs(1)) / 1;
              if (__pyx_t_20 > 0)
              {
                  #ifdef _OPENMP
                  #pragma omp parallel private(__pyx_t_15, __pyx_t_17, __pyx_t_21, __pyx_t_22) firstprivate(__pyx_t_10, __pyx_t_16)
                  #endif /* _OPENMP */
                  {
                      #ifdef _OPENMP
                      #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(dynamic)
                      #endif /* _OPENMP */
                      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_20; __pyx_t_19++){
                          {
                              __pyx_v_j = (int)(0 + 1 * __pyx_t_19);

                              /* "cyext_acv.pyx":5394
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              __pyx_t_15 = __pyx_v_i;
                              __pyx_t_17 = __pyx_v_j;
                              __pyx_t_21 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_weights.data + __pyx_t_15 * __pyx_v_weights.strides[0]) ) + __pyx_t_17 * __pyx_v_weights.strides[1]) ))) != 0.0) != 0);
                              if (__pyx_t_21) {

                                /* "cyext_acv.pyx":5395
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_16.data = __pyx_v_data.data;
                                __pyx_t_16.memview = __pyx_v_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_16, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_data.strides[0];
        __pyx_t_16.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_16.shape[0] = __pyx_v_data.shape[1];
__pyx_t_16.strides[0] = __pyx_v_data.strides[1];
    __pyx_t_16.suboffsets[0] = -1;

__pyx_t_17 = __pyx_v_i;

                                /* "cyext_acv.pyx":5397
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])             # <<<<<<<<<<<<<<
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)
 */
                                __pyx_t_10.data = __pyx_v_partition_byobs_data.data;
                                __pyx_t_10.memview = __pyx_v_partition_byobs_data.memview;
                                __PYX_INC_MEMVIEW(&__pyx_t_10, 0);
                                {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[0];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

{
    Py_ssize_t __pyx_tmp_idx = __pyx_v_j;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_partition_byobs_data.strides[1];
        __pyx_t_10.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_10.shape[0] = __pyx_v_partition_byobs_data.shape[2];
__pyx_t_10.strides[0] = __pyx_v_partition_byobs_data.strides[2];
    __pyx_t_10.suboffsets[0] = -1;

__pyx_t_10.shape[1] = __pyx_v_partition_byobs_data.shape[3];
__pyx_t_10.strides[1] = __pyx_v_partition_byobs_data.strides[3];
    __pyx_t_10.suboffsets[1] = -1;

__pyx_t_15 = __pyx_v_i;

                                /* "cyext_acv.pyx":5395
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                                 features, thresholds, children_left, children_right,
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 */
                                __pyx_t_22 = __pyx_v_j;
                                *((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_sdp_data.data + __pyx_t_15 * __pyx_v_sdp_data.strides[0]) ) + __pyx_t_22 * __pyx_v_sdp_data.strides[1]) )) = __pyx_f_9cyext_acv_single_compute_sdp_rule(__pyx_t_16, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_17)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_t_10);
                                __PYX_XDEC_MEMVIEW(&__pyx_t_16, 0);
                                __pyx_t_16.memview = NULL;
                                __pyx_t_16.data = NULL;
                                __PYX_XDEC_MEMVIEW(&__pyx_t_10, 0);
                                __pyx_t_10.memview = NULL;
                                __pyx_t_10.data = NULL;

                                /* "cyext_acv.pyx":5394
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):
 *                 if weights[i, j] != 0:             # <<<<<<<<<<<<<<
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 *                                 features, thresholds, children_left, children_right,
 */
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "cyext_acv.pyx":5393
 *                         max_depth, min_node_size, classifier, t, partition_byobs[i], weights[i])
 * 
 *             for j in prange(data.shape[0], nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *                 if weights[i, j] != 0:
 *                     sdp_data[i, j] = single_compute_sdp_rule(data[j], y_X[i], data, y_data, S[i],
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L1_error;
          }
          __pyx_L9:;
        }
    }

    /* "cyext_acv.pyx":5388
 *         cdef int i, j
 * 
 *         for i in tqdm(range(N)):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rule_weights(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyext_acv.pyx":5399
 *                                 max_depth, min_node_size, classifier, t, partition_byobs_data[i, j])
 * 
 *         return np.array(sdp), np.array(partition_byobs), np.array(sdp_data), np.array(partition_byobs_data), np.array(weights)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs, 3, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_sdp_data, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_partition_byobs_data, 4, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_23 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_23)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_23);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_23, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_weights, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_24 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_23);
    if (likely(__pyx_t_24)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_23, function);
    }
  }
  __pyx_t_5 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_23, __pyx_t_24, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_23, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = PyTuple_New(5); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_23, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_23, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_23, 4, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_23;
  __pyx_t_23 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5365
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_maxrule_verbose(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, double & pi):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_14, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_verbose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs, 1);
  __Pyx_XDECREF(__pyx_v_buffer_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_partition_byobs_data, 1);
  __Pyx_XDECREF(__pyx_v_rules_data);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_weights, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_81compute_sdp_maxrule_verbose(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_81compute_sdp_maxrule_verbose(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  double __pyx_v_pi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_verbose (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_pi,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 1); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 2); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 3); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 4); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 5); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 6); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 7); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 8); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 9); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 10); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 11); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 12); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, 13); __PYX_ERR(0, 5365, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_maxrule_verbose") < 0)) __PYX_ERR(0, 5365, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5365, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5365, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5365, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5365, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5365, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5366, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5366, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5366, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5366, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5367, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5367, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5367, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5367, __pyx_L3_error)
    __pyx_v_pi = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5367, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_maxrule_verbose", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5365, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_verbose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_80compute_sdp_maxrule_verbose(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_80compute_sdp_maxrule_verbose(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, double __pyx_v_pi) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_maxrule_verbose", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5365, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_maxrule_verbose(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_pi, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_maxrule_verbose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5406
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

static double __pyx_f_9cyext_acv_single_compute_sdp_rf(__Pyx_memviewslice &__pyx_v_x, double &__pyx_v_y_x, __Pyx_memviewslice &__pyx_v_data, __Pyx_memviewslice &__pyx_v_y_data, std::vector<int>  &__pyx_v_S, __Pyx_memviewslice &__pyx_v_features, __Pyx_memviewslice &__pyx_v_thresholds, __Pyx_memviewslice &__pyx_v_children_left, __Pyx_memviewslice &__pyx_v_children_right, int &__pyx_v_max_depth, int &__pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t) {
  unsigned int __pyx_v_n_trees;
  unsigned int __pyx_v_N;
  double __pyx_v_sdp;
  int __pyx_v_b;
  CYTHON_UNUSED int __pyx_v_level;
  int __pyx_v_i;
  int __pyx_v_it_node;
  std::set<int>  __pyx_v_nodes_level;
  std::set<int>  __pyx_v_nodes_child;
  std::set<int>  __pyx_v_in_data;
  std::set<int>  __pyx_v_in_data_b;
  std::set<int> ::iterator __pyx_v_it;
  std::set<int> ::iterator __pyx_v_it_point;
  double __pyx_r;
  unsigned int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cyext_acv.pyx":5410
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 * 
 *     cdef unsigned int n_trees = features.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int N = data.shape[0]
 *     cdef double s, sdp
 */
  __pyx_v_n_trees = (__pyx_v_features.shape[0]);

  /* "cyext_acv.pyx":5411
 * 
 *     cdef unsigned int n_trees = features.shape[0]
 *     cdef unsigned int N = data.shape[0]             # <<<<<<<<<<<<<<
 *     cdef double s, sdp
 *     cdef int o
 */
  __pyx_v_N = (__pyx_v_data.shape[0]);

  /* "cyext_acv.pyx":5414
 *     cdef double s, sdp
 *     cdef int o
 *     sdp = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int b, level, i, it_node
 */
  __pyx_v_sdp = 0.0;

  /* "cyext_acv.pyx":5420
 *     cdef set[int].iterator it, it_point
 * 
 *     for b in range(n_trees):             # <<<<<<<<<<<<<<
 *         nodes_child.clear()
 *         nodes_level.clear()
 */
  __pyx_t_1 = __pyx_v_n_trees;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_b = __pyx_t_3;

    /* "cyext_acv.pyx":5421
 * 
 *     for b in range(n_trees):
 *         nodes_child.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.clear()
 *         nodes_level.insert(0)
 */
    __pyx_v_nodes_child.clear();

    /* "cyext_acv.pyx":5422
 *     for b in range(n_trees):
 *         nodes_child.clear()
 *         nodes_level.clear()             # <<<<<<<<<<<<<<
 *         nodes_level.insert(0)
 * 
 */
    __pyx_v_nodes_level.clear();

    /* "cyext_acv.pyx":5423
 *         nodes_child.clear()
 *         nodes_level.clear()
 *         nodes_level.insert(0)             # <<<<<<<<<<<<<<
 * 
 *         in_data.clear()
 */
    try {
      __pyx_v_nodes_level.insert(0);
    } catch(...) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      __Pyx_CppExn2PyErr();
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 5423, __pyx_L1_error)
    }

    /* "cyext_acv.pyx":5425
 *         nodes_level.insert(0)
 * 
 *         in_data.clear()             # <<<<<<<<<<<<<<
 *         in_data_b.clear()
 *         for i in range(N):
 */
    __pyx_v_in_data.clear();

    /* "cyext_acv.pyx":5426
 * 
 *         in_data.clear()
 *         in_data_b.clear()             # <<<<<<<<<<<<<<
 *         for i in range(N):
 *             in_data.insert(i)
 */
    __pyx_v_in_data_b.clear();

    /* "cyext_acv.pyx":5427
 *         in_data.clear()
 *         in_data_b.clear()
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             in_data.insert(i)
 *             in_data_b.insert(i)
 */
    __pyx_t_4 = __pyx_v_N;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "cyext_acv.pyx":5428
 *         in_data_b.clear()
 *         for i in range(N):
 *             in_data.insert(i)             # <<<<<<<<<<<<<<
 *             in_data_b.insert(i)
 * 
 */
      try {
        __pyx_v_in_data.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5428, __pyx_L1_error)
      }

      /* "cyext_acv.pyx":5429
 *         for i in range(N):
 *             in_data.insert(i)
 *             in_data_b.insert(i)             # <<<<<<<<<<<<<<
 * 
 *         for level in range(max_depth):
 */
      try {
        __pyx_v_in_data_b.insert(__pyx_v_i);
      } catch(...) {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        __Pyx_CppExn2PyErr();
        #ifdef WITH_THREAD
        __Pyx_PyGILState_Release(__pyx_gilstate_save);
        #endif
        __PYX_ERR(0, 5429, __pyx_L1_error)
      }
    }

    /* "cyext_acv.pyx":5431
 *             in_data_b.insert(i)
 * 
 *         for level in range(max_depth):             # <<<<<<<<<<<<<<
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 */
    __pyx_t_6 = __pyx_v_max_depth;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_level = __pyx_t_8;

      /* "cyext_acv.pyx":5432
 * 
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()             # <<<<<<<<<<<<<<
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 */
      __pyx_v_it_point = __pyx_v_nodes_level.begin();

      /* "cyext_acv.pyx":5433
 *         for level in range(max_depth):
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):             # <<<<<<<<<<<<<<
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it_point != __pyx_v_nodes_level.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5434
 *             it_point = nodes_level.begin()
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)             # <<<<<<<<<<<<<<
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 */
        __pyx_v_it_node = (*__pyx_v_it_point);

        /* "cyext_acv.pyx":5435
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
        __pyx_t_10 = __pyx_v_b;
        __pyx_t_11 = __pyx_v_it_node;
        __pyx_t_9 = ((std::find<std::vector<int> ::iterator,int>(__pyx_v_S.begin(), __pyx_v_S.end(), (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_10 * __pyx_v_features.strides[0]) ) + __pyx_t_11 * __pyx_v_features.strides[1]) )))) != __pyx_v_S.end()) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5436
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          __pyx_t_12 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_11 * __pyx_v_features.strides[0]) ) + __pyx_t_10 * __pyx_v_features.strides[1]) )));
          __pyx_t_13 = __pyx_v_b;
          __pyx_t_14 = __pyx_v_it_node;
          __pyx_t_9 = (((*((double *) ( /* dim=0 */ (__pyx_v_x.data + __pyx_t_12 * __pyx_v_x.strides[0]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_13 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_14 * __pyx_v_thresholds.strides[1]) )))) != 0);
          if (__pyx_t_9) {

            /* "cyext_acv.pyx":5437
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
            __pyx_t_14 = __pyx_v_b;
            __pyx_t_13 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_14 * __pyx_v_children_left.strides[0]) ) + __pyx_t_13 * __pyx_v_children_left.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5437, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5439
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5440
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5441
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_13 = __pyx_v_b;
              __pyx_t_14 = __pyx_v_it_node;
              __pyx_t_10 = (*__pyx_v_it);
              __pyx_t_11 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_13 * __pyx_v_features.strides[0]) ) + __pyx_t_14 * __pyx_v_features.strides[1]) )));
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_10 * __pyx_v_data.strides[0]) ) + __pyx_t_11 * __pyx_v_data.strides[1]) ))) > (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_12 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_15 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5442
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5441
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5443
 *                             if data[deref(it), features[b, it_node]] > thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 * 
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5444
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
            __pyx_v_in_data = __pyx_v_in_data_b;

            /* "cyext_acv.pyx":5436
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                         nodes_child.insert(children_left[b, it_node])
 * 
 */
            goto __pyx_L12;
          }

          /* "cyext_acv.pyx":5447
 * 
 *                     else:
 *                         nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                         it = in_data.begin()
 */
          /*else*/ {
            __pyx_t_15 = __pyx_v_b;
            __pyx_t_12 = __pyx_v_it_node;
            try {
              __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_15 * __pyx_v_children_right.strides[0]) ) + __pyx_t_12 * __pyx_v_children_right.strides[1]) ))));
            } catch(...) {
              #ifdef WITH_THREAD
              PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
              #endif
              __Pyx_CppExn2PyErr();
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              __PYX_ERR(0, 5447, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5449
 *                         nodes_child.insert(children_right[b, it_node])
 * 
 *                         it = in_data.begin()             # <<<<<<<<<<<<<<
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 */
            __pyx_v_it = __pyx_v_in_data.begin();

            /* "cyext_acv.pyx":5450
 * 
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 */
            while (1) {
              __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
              if (!__pyx_t_9) break;

              /* "cyext_acv.pyx":5451
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              __pyx_t_12 = __pyx_v_b;
              __pyx_t_15 = __pyx_v_it_node;
              __pyx_t_14 = (*__pyx_v_it);
              __pyx_t_13 = (*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_features.data + __pyx_t_12 * __pyx_v_features.strides[0]) ) + __pyx_t_15 * __pyx_v_features.strides[1]) )));
              __pyx_t_11 = __pyx_v_b;
              __pyx_t_10 = __pyx_v_it_node;
              __pyx_t_9 = (((*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_data.data + __pyx_t_14 * __pyx_v_data.strides[0]) ) + __pyx_t_13 * __pyx_v_data.strides[1]) ))) <= (*((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_thresholds.data + __pyx_t_11 * __pyx_v_thresholds.strides[0]) ) + __pyx_t_10 * __pyx_v_thresholds.strides[1]) )))) != 0);
              if (__pyx_t_9) {

                /* "cyext_acv.pyx":5452
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))             # <<<<<<<<<<<<<<
 *                             inc(it)
 *                         in_data = in_data_b
 */
                (void)(__pyx_v_in_data_b.erase((*__pyx_v_it)));

                /* "cyext_acv.pyx":5451
 *                         it = in_data.begin()
 *                         while(it != in_data.end()):
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:             # <<<<<<<<<<<<<<
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 */
              }

              /* "cyext_acv.pyx":5453
 *                             if data[deref(it), features[b, it_node]] <= thresholds[b, it_node]:
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)             # <<<<<<<<<<<<<<
 *                         in_data = in_data_b
 *                 else:
 */
              (void)((++__pyx_v_it));
            }

            /* "cyext_acv.pyx":5454
 *                                 in_data_b.erase(deref(it))
 *                             inc(it)
 *                         in_data = in_data_b             # <<<<<<<<<<<<<<
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 */
            __pyx_v_in_data = __pyx_v_in_data_b;
          }
          __pyx_L12:;

          /* "cyext_acv.pyx":5435
 *             while(it_point != nodes_level.end()):
 *                 it_node = deref(it_point)
 *                 if std_find[vector[int].iterator, int](S.begin(), S.end(), features[b, it_node]) != S.end():             # <<<<<<<<<<<<<<
 *                     if x[features[b, it_node]] <= thresholds[b, it_node]:
 *                         nodes_child.insert(children_left[b, it_node])
 */
          goto __pyx_L11;
        }

        /* "cyext_acv.pyx":5456
 *                         in_data = in_data_b
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])             # <<<<<<<<<<<<<<
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 */
        /*else*/ {
          __pyx_t_10 = __pyx_v_b;
          __pyx_t_11 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_left.data + __pyx_t_10 * __pyx_v_children_left.strides[0]) ) + __pyx_t_11 * __pyx_v_children_left.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5456, __pyx_L1_error)
          }

          /* "cyext_acv.pyx":5457
 *                 else:
 *                      nodes_child.insert(children_left[b, it_node])
 *                      nodes_child.insert(children_right[b, it_node])             # <<<<<<<<<<<<<<
 * 
 *                 if in_data.size() < min_node_size:
 */
          __pyx_t_11 = __pyx_v_b;
          __pyx_t_10 = __pyx_v_it_node;
          try {
            __pyx_v_nodes_child.insert((*((int *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_children_right.data + __pyx_t_11 * __pyx_v_children_right.strides[0]) ) + __pyx_t_10 * __pyx_v_children_right.strides[1]) ))));
          } catch(...) {
            #ifdef WITH_THREAD
            PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
            #endif
            __Pyx_CppExn2PyErr();
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            __PYX_ERR(0, 5457, __pyx_L1_error)
          }
        }
        __pyx_L11:;

        /* "cyext_acv.pyx":5459
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        __pyx_t_9 = ((__pyx_v_in_data.size() < __pyx_v_min_node_size) != 0);
        if (__pyx_t_9) {

          /* "cyext_acv.pyx":5460
 * 
 *                 if in_data.size() < min_node_size:
 *                     break             # <<<<<<<<<<<<<<
 *                 inc(it_point)
 * 
 */
          goto __pyx_L10_break;

          /* "cyext_acv.pyx":5459
 *                      nodes_child.insert(children_right[b, it_node])
 * 
 *                 if in_data.size() < min_node_size:             # <<<<<<<<<<<<<<
 *                     break
 *                 inc(it_point)
 */
        }

        /* "cyext_acv.pyx":5461
 *                 if in_data.size() < min_node_size:
 *                     break
 *                 inc(it_point)             # <<<<<<<<<<<<<<
 * 
 *             nodes_level = nodes_child
 */
        (void)((++__pyx_v_it_point));
      }
      __pyx_L10_break:;

      /* "cyext_acv.pyx":5463
 *                 inc(it_point)
 * 
 *             nodes_level = nodes_child             # <<<<<<<<<<<<<<
 * 
 *         if classifier == 1:
 */
      __pyx_v_nodes_level = __pyx_v_nodes_child;
    }

    /* "cyext_acv.pyx":5465
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
    __pyx_t_9 = ((__pyx_v_classifier == 1) != 0);
    if (__pyx_t_9) {

      /* "cyext_acv.pyx":5466
 * 
 *         if classifier == 1:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 */
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5467
 *         if classifier == 1:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5468
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])             # <<<<<<<<<<<<<<
 *                 inc(it)
 *         else:
 */
        __pyx_t_10 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (__pyx_v_y_x == (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_10)) ))))));

        /* "cyext_acv.pyx":5469
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * (y_x == y_data[deref(it)])
 *                 inc(it)             # <<<<<<<<<<<<<<
 *         else:
 *             it = in_data.begin()
 */
        (void)((++__pyx_v_it));
      }

      /* "cyext_acv.pyx":5465
 *             nodes_level = nodes_child
 * 
 *         if classifier == 1:             # <<<<<<<<<<<<<<
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 */
      goto __pyx_L20;
    }

    /* "cyext_acv.pyx":5471
 *                 inc(it)
 *         else:
 *             it = in_data.begin()             # <<<<<<<<<<<<<<
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 */
    /*else*/ {
      __pyx_v_it = __pyx_v_in_data.begin();

      /* "cyext_acv.pyx":5472
 *         else:
 *             it = in_data.begin()
 *             while(it != in_data.end()):             # <<<<<<<<<<<<<<
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 inc(it)
 */
      while (1) {
        __pyx_t_9 = ((__pyx_v_it != __pyx_v_in_data.end()) != 0);
        if (!__pyx_t_9) break;

        /* "cyext_acv.pyx":5473
 *             it = in_data.begin()
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)             # <<<<<<<<<<<<<<
 *                 inc(it)
 * 
 */
        __pyx_t_10 = (*__pyx_v_it);
        __pyx_t_11 = (*__pyx_v_it);
        __pyx_v_sdp = (__pyx_v_sdp + ((1. / (__pyx_v_n_trees * __pyx_v_in_data.size())) * (((__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_10)) )))) * (__pyx_v_y_x - (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_data.data) + __pyx_t_11)) ))))) <= __pyx_v_t)));

        /* "cyext_acv.pyx":5474
 *             while(it != in_data.end()):
 *                 sdp  += (1./(n_trees*in_data.size())) * ((y_x - y_data[deref(it)])*(y_x - y_data[deref(it)]) <= t)
 *                 inc(it)             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
        (void)((++__pyx_v_it));
      }
    }
    __pyx_L20:;
  }

  /* "cyext_acv.pyx":5476
 *                 inc(it)
 * 
 *     return sdp             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_sdp;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5406
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cdef double single_compute_sdp_rf(double[:] & x, double & y_x,  double[:, :] & data, double[::1] & y_data, vector[int] & S,             # <<<<<<<<<<<<<<
 *         int[:, :] & features, double[:, :] & thresholds,  int[:, :] & children_left, int[:, :] & children_right,
 *         int & max_depth, int & min_node_size, int & classifier, double & t) nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cyext_acv.single_compute_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "cyext_acv.pyx":5482
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_83compute_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf", 0);

  /* "cyext_acv.pyx":5486
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5487
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5489
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_6 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_8 = (__pyx_t_6 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_8 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_10, __pyx_t_11) firstprivate(__pyx_t_9)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_8; __pyx_t_7++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_7);

                            /* "cyext_acv.pyx":5490
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S[i],             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 */
                            __pyx_t_9.data = __pyx_v_X.data;
                            __pyx_t_9.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_X.shape[1];
__pyx_t_9.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_10 = __pyx_v_i;

                            /* "cyext_acv.pyx":5492
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
                            __pyx_t_11 = __pyx_v_i;
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_11)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rf(__pyx_t_9, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_10)) ))), __pyx_v_data, __pyx_v_y_data, (__pyx_v_S[__pyx_v_i]), __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                            __pyx_t_9.memview = NULL;
                            __pyx_t_9.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":5489
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S[i],
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":5493
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5482
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[vector[int]] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_83compute_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_83compute_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<int> >  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 1); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 2); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 3); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 4); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 5); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 6); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 7); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 8); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 9); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 10); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 11); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, 12); __PYX_ERR(0, 5482, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rf") < 0)) __PYX_ERR(0, 5482, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5482, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5482, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5482, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5482, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5482, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5483, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5483, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5483, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5483, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5484, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5484, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5484, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5484, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rf", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5482, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_82compute_sdp_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_82compute_sdp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<std::vector<int> >  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5482, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5499
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf_same_set(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[int] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

static PyObject *__pyx_pw_9cyext_acv_85compute_sdp_rf_same_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_compute_sdp_rf_same_set(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_N;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf_same_set", 0);

  /* "cyext_acv.pyx":5503
 *         int max_depth, int min_node_size, int & classifier, double & t):
 * 
 *         cdef int N = X.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5504
 * 
 *         cdef int N = X.shape[0]
 *         cdef double[::1] sdp = np.zeros(N)             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5506
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {
        __pyx_t_6 = __pyx_v_N;
        if ((1 == 0)) abort();
        {
            #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                #undef likely
                #undef unlikely
                #define likely(x)   (x)
                #define unlikely(x) (x)
            #endif
            __pyx_t_8 = (__pyx_t_6 - 0 + 1 - 1/abs(1)) / 1;
            if (__pyx_t_8 > 0)
            {
                #ifdef _OPENMP
                #pragma omp parallel private(__pyx_t_10, __pyx_t_11) firstprivate(__pyx_t_9)
                #endif /* _OPENMP */
                {
                    #ifdef _OPENMP
                    #pragma omp for firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) schedule(dynamic)
                    #endif /* _OPENMP */
                    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_8; __pyx_t_7++){
                        {
                            __pyx_v_i = (int)(0 + 1 * __pyx_t_7);

                            /* "cyext_acv.pyx":5507
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S,             # <<<<<<<<<<<<<<
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 */
                            __pyx_t_9.data = __pyx_v_X.data;
                            __pyx_t_9.memview = __pyx_v_X.memview;
                            __PYX_INC_MEMVIEW(&__pyx_t_9, 0);
                            {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_X.strides[0];
        __pyx_t_9.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_9.shape[0] = __pyx_v_X.shape[1];
__pyx_t_9.strides[0] = __pyx_v_X.strides[1];
    __pyx_t_9.suboffsets[0] = -1;

__pyx_t_10 = __pyx_v_i;

                            /* "cyext_acv.pyx":5509
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 *         return np.array(sdp)
 * 
 */
                            __pyx_t_11 = __pyx_v_i;
                            *((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_sdp.data) + __pyx_t_11)) )) = __pyx_f_9cyext_acv_single_compute_sdp_rf(__pyx_t_9, (*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v_y_X.data) + __pyx_t_10)) ))), __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);
                            __PYX_XDEC_MEMVIEW(&__pyx_t_9, 0);
                            __pyx_t_9.memview = NULL;
                            __pyx_t_9.data = NULL;
                        }
                    }
                }
            }
        }
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   __builtin_expect(!!(x), 1)
            #define unlikely(x) __builtin_expect(!!(x), 0)
        #endif
      }

      /* "cyext_acv.pyx":5506
 *         cdef double[::1] sdp = np.zeros(N)
 *         cdef int i
 *         for i in prange(N, nogil=True, schedule='dynamic'):             # <<<<<<<<<<<<<<
 *             sdp[i] = single_compute_sdp_rf(X[i], y_X[i], data, y_data, S,
 *                         features, thresholds, children_left, children_right,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cyext_acv.pyx":5510
 *                         features, thresholds, children_left, children_right,
 *                         max_depth, min_node_size, classifier, t)
 *         return np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5499
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef compute_sdp_rf_same_set(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data, vector[int] S,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_same_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_85compute_sdp_rf_same_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_85compute_sdp_rf_same_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<int>  __pyx_v_S;
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_sdp_rf_same_set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_S,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 1); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 2); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 3); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 4); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 5); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 6); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 7); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 8); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 9); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 10); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 11); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, 12); __PYX_ERR(0, 5499, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_sdp_rf_same_set") < 0)) __PYX_ERR(0, 5499, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5499, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5499, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5499, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5499, __pyx_L3_error)
    __pyx_v_S = __pyx_convert_vector_from_py_int(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5499, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5500, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5500, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5500, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[8], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5500, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5501, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5501, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5501, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[12]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5501, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_sdp_rf_same_set", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5499, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_same_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9cyext_acv_84compute_sdp_rf_same_set(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_84compute_sdp_rf_same_set(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, std::vector<int>  __pyx_v_S, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compute_sdp_rf_same_set", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5499, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rf_same_set(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_S, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.compute_sdp_rf_same_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5517
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

static PyObject *__pyx_pw_9cyext_acv_87global_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_global_sdp_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_b = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_ba = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_in_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_max_size;
  int __pyx_v_size;
  std::vector<int>  __pyx_v_S;
  std::vector<int>  __pyx_v_len_s_star;
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  std::vector<long>  __pyx_v_R;
  std::vector<long>  __pyx_v_r;
  __Pyx_memviewslice __pyx_v_R_buf = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_X_arr = NULL;
  PyObject *__pyx_v_y_X_arr = NULL;
  PyObject *__pyx_v_remove_va = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  std::vector<int>  __pyx_t_7;
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  long __pyx_t_20;
  long __pyx_t_21;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_26;
  __Pyx_memviewslice __pyx_t_27 = { 0, 0, { 0 }, { 0 }, { 0 } };
  std::vector<std::vector<long> > ::size_type __pyx_t_28;
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<long> ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  __Pyx_memviewslice __pyx_t_34 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_35 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  int __pyx_t_38;
  int __pyx_t_39;
  int __pyx_t_40;
  PyObject *__pyx_t_41 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_rf", 0);

  /* "cyext_acv.pyx":5522
 *             int minimal, bint stop, list search_space):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5523
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp, sdp_b, sdp_ba
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":5528
 *     cdef double[:] sdp_global
 *     cdef double[:] in_data
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5529
 *     cdef double[:] in_data
 *     sdp = np.zeros((N))
 *     sdp_b = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 *     in_data = np.zeros(N)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_b = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5530
 *     sdp = np.zeros((N))
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 *     in_data = np.zeros(N)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5531
 *     sdp_b = np.zeros((N))
 *     sdp_global = np.zeros((m))
 *     in_data = np.zeros(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, a_it, b_it, o_all, p, p_s, nb_leaf, p_u, p_d, p_su, p_sd, down, up
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_in_data = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5538
 * 
 *     cdef vector[int] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_len_s_star = __pyx_t_7;

  /* "cyext_acv.pyx":5543
 * 
 *     cdef vector[long] R, r
 *     R.resize(N)             # <<<<<<<<<<<<<<
 *     for i in range(N):
 *         R[i] = i
 */
  try {
    __pyx_v_R.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 5543, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":5544
 *     cdef vector[long] R, r
 *     R.resize(N)
 *     for i in range(N):             # <<<<<<<<<<<<<<
 *         R[i] = i
 *     r.resize(N)
 */
  __pyx_t_8 = __pyx_v_N;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "cyext_acv.pyx":5545
 *     R.resize(N)
 *     for i in range(N):
 *         R[i] = i             # <<<<<<<<<<<<<<
 *     r.resize(N)
 * 
 */
    (__pyx_v_R[__pyx_v_i]) = __pyx_v_i;
  }

  /* "cyext_acv.pyx":5546
 *     for i in range(N):
 *         R[i] = i
 *     r.resize(N)             # <<<<<<<<<<<<<<
 * 
 *     cdef long[:] R_buf
 */
  try {
    __pyx_v_r.resize(__pyx_v_N);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 5546, __pyx_L1_error)
  }

  /* "cyext_acv.pyx":5549
 * 
 *     cdef long[:] R_buf
 *     R_buf = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     X_arr = np.array(X, dtype=np.double)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 5549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_R_buf = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "cyext_acv.pyx":5551
 *     R_buf = np.zeros((N), dtype=np.int)
 * 
 *     X_arr = np.array(X, dtype=np.double)             # <<<<<<<<<<<<<<
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_X, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_double); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_X_arr = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cyext_acv.pyx":5552
 * 
 *     X_arr = np.array(X, dtype=np.double)
 *     y_X_arr = np.array(y_X, dtype=np.double)             # <<<<<<<<<<<<<<
 * 
 *     if C[0] != []:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_y_X, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_double); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_y_X_arr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5554
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 5554, __pyx_L1_error)
  }
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_12) {

    /* "cyext_acv.pyx":5555
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 5555, __pyx_L1_error)
    }
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5555, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_ci = __pyx_t_15;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 5555, __pyx_L1_error)
      }
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_16 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5555, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_17 = __pyx_t_16;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_cj = __pyx_t_18;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5555, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 5555, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":5556
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]             # <<<<<<<<<<<<<<
 *         for ci in range(len(C)):
 *             i = 0
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 5556, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
    for (;;) {
      if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5556, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5556, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_i = __pyx_t_10;
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 5556, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_19 = (__pyx_t_12 != 0);
      if (__pyx_t_19) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5556, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 5556, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cyext_acv.pyx":5557
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):             # <<<<<<<<<<<<<<
 *             i = 0
 *             for cj in range(len(C[ci])):
 */
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 5557, __pyx_L1_error)
    }
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5557, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_ci = __pyx_t_15;

      /* "cyext_acv.pyx":5558
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 *             i = 0             # <<<<<<<<<<<<<<
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 */
      __pyx_v_i = 0;

      /* "cyext_acv.pyx":5559
 *         for ci in range(len(C)):
 *             i = 0
 *             for cj in range(len(C[ci])):             # <<<<<<<<<<<<<<
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 */
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 5559, __pyx_L1_error)
      }
      __pyx_t_2 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_16 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5559, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_17 = __pyx_t_16;
      for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
        __pyx_v_cj = __pyx_t_18;

        /* "cyext_acv.pyx":5560
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5560, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_19 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_search_space, Py_EQ)); if (unlikely(__pyx_t_19 < 0)) __PYX_ERR(0, 5560, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = (__pyx_t_19 != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":5561
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     break
 *             if i != 0:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "cyext_acv.pyx":5562
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 *                     break             # <<<<<<<<<<<<<<
 *             if i != 0:
 *                 va_id += [C[ci]]
 */
          goto __pyx_L16_break;

          /* "cyext_acv.pyx":5560
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        }
      }
      __pyx_L16_break:;

      /* "cyext_acv.pyx":5563
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      __pyx_t_12 = ((__pyx_v_i != 0) != 0);
      if (__pyx_t_12) {

        /* "cyext_acv.pyx":5564
 *                     break
 *             if i != 0:
 *                 va_id += [C[ci]]             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in search_space]
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5564, __pyx_L1_error)
        }
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        PyList_SET_ITEM(__pyx_t_2, 0, PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_va_id, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v_va_id, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "cyext_acv.pyx":5563
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      }
    }

    /* "cyext_acv.pyx":5554
 *     y_X_arr = np.array(y_X, dtype=np.double)
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
    goto __pyx_L5;
  }

  /* "cyext_acv.pyx":5566
 *                 va_id += [C[ci]]
 *     else:
 *         va_id = [[i] for i in search_space]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 5566, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
    for (;;) {
      if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_4); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5566, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5566, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_i = __pyx_t_10;
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 5566, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "cyext_acv.pyx":5568
 *         va_id = [[i] for i in search_space]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_13 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5568, __pyx_L1_error)
  __pyx_v_m = __pyx_t_13;

  /* "cyext_acv.pyx":5569
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_power = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5570
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":5571
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_20 = (__pyx_v_m + 1);
  __pyx_t_21 = __pyx_t_20;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_21; __pyx_t_10+=1) {
    __pyx_v_size = __pyx_t_10;

    /* "cyext_acv.pyx":5572
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":5573
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_combinations); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_22 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_22 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_va_id, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_22); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_22, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_22, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_13 = 0;
      __pyx_t_23 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_23 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5573, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_23)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5573, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5573, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_23(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 5573, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "cyext_acv.pyx":5574
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_4);
      __pyx_t_6 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_24, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_24 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_24)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_24);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_24, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_1); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 5574, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyext_acv.pyx":5575
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":5573
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyext_acv.pyx":5576
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_25 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_25 == ((int)-1))) __PYX_ERR(0, 5576, __pyx_L1_error)

    /* "cyext_acv.pyx":5577
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_12 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_12) {

      /* "cyext_acv.pyx":5578
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L22_break;

      /* "cyext_acv.pyx":5577
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L22_break:;

  /* "cyext_acv.pyx":5580
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_26 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5580, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_26;

  /* "cyext_acv.pyx":5582
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_24) < 0) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_24); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_27 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_27.memview)) __PYX_ERR(0, 5582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_s_star = __pyx_t_27;
  __pyx_t_27.memview = NULL;
  __pyx_t_27.data = NULL;

  /* "cyext_acv.pyx":5585
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":5586
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_S = __pyx_t_7;

  /* "cyext_acv.pyx":5588
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_24 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_24);
  __pyx_t_2 = 0;
  __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_1, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_24);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_13 = 0;
    __pyx_t_23 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_23 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 5588, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_23)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5588, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5588, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 5588, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5588, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_23(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 5588, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_s_0 = __pyx_t_10;

    /* "cyext_acv.pyx":5589
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_28 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_29; __pyx_t_10+=1) {
      __pyx_v_s_1 = __pyx_t_10;

      /* "cyext_acv.pyx":5590
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_30 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_31; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":5591
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        (__pyx_v_S[__pyx_v_i]) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":5593
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             r.clear()
 *             N = R.size()
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":5594
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()             # <<<<<<<<<<<<<<
 *             N = R.size()
 *             in_data = np.zeros(X.shape[0])
 */
      __pyx_v_r.clear();

      /* "cyext_acv.pyx":5595
 *             S_size = power_cpp[s_0][s_1].size()
 *             r.clear()
 *             N = R.size()             # <<<<<<<<<<<<<<
 *             in_data = np.zeros(X.shape[0])
 * 
 */
      __pyx_v_N = __pyx_v_R.size();

      /* "cyext_acv.pyx":5596
 *             r.clear()
 *             N = R.size()
 *             in_data = np.zeros(X.shape[0])             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_np); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __pyx_t_24 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_24) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_24);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5596, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_in_data, 1);
      __pyx_v_in_data = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":5598
 *             in_data = np.zeros(X.shape[0])
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 R_buf[i] = R[i]
 *                 in_data[R_buf[i]] = 1
 */
      __pyx_t_8 = __pyx_v_N;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_9; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":5599
 * 
 *             for i in range(N):
 *                 R_buf[i] = R[i]             # <<<<<<<<<<<<<<
 *                 in_data[R_buf[i]] = 1
 * 
 */
        __pyx_t_32 = __pyx_v_i;
        *((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )) = (__pyx_v_R[__pyx_v_i]);

        /* "cyext_acv.pyx":5600
 *             for i in range(N):
 *                 R_buf[i] = R[i]
 *                 in_data[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 * 
 *                 # sdp_b[R_buf[i]] = single_compute_sdp_rf(X[R_buf[i]],  y_X[R_buf[i]], data, y_data, S[:S_size], features, thresholds,  children_left, children_right,
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
        *((double *) ( /* dim=0 */ (__pyx_v_in_data.data + __pyx_t_33 * __pyx_v_in_data.strides[0]) )) = 1.0;
      }

      /* "cyext_acv.pyx":5605
 *                 #                 max_depth, min_node_size, classifier, t)
 * 
 *             sdp_ba = compute_sdp_rf_same_set(X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_in_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, ((PyObject*)&PyBool_Type)) < 0) __PYX_ERR(0, 5605, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_X_arr, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_34 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_34.memview)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyext_acv.pyx":5606
 * 
 *             sdp_ba = compute_sdp_rf_same_set(X_arr[np.array(in_data, dtype=bool)],
 *                                     y_X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_in_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_24);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, ((PyObject*)&PyBool_Type)) < 0) __PYX_ERR(0, 5606, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_y_X_arr, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_35 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_35.memview)) __PYX_ERR(0, 5606, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "cyext_acv.pyx":5607
 *             sdp_ba = compute_sdp_rf_same_set(X_arr[np.array(in_data, dtype=bool)],
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_S); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, __pyx_v_S_size, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5607, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "cyext_acv.pyx":5605
 *                 #                 max_depth, min_node_size, classifier, t)
 * 
 *             sdp_ba = compute_sdp_rf_same_set(X_arr[np.array(in_data, dtype=bool)],             # <<<<<<<<<<<<<<
 *                                     y_X_arr[np.array(in_data, dtype=bool)],
 *                                     data, y_data, S[:S_size], features, thresholds,  children_left,
 */
      __pyx_t_1 = __pyx_f_9cyext_acv_compute_sdp_rf_same_set(__pyx_t_34, __pyx_t_35, __pyx_v_data, __pyx_v_y_data, __pyx_t_7, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __PYX_XDEC_MEMVIEW(&__pyx_t_34, 1);
      __pyx_t_34.memview = NULL;
      __pyx_t_34.data = NULL;
      __PYX_XDEC_MEMVIEW(&__pyx_t_35, 1);
      __pyx_t_35.memview = NULL;
      __pyx_t_35.data = NULL;
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5605, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_ba, 1);
      __pyx_v_sdp_ba = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":5610
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 */
      __pyx_t_8 = __pyx_v_N;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_9; __pyx_t_22+=1) {
        __pyx_v_i = __pyx_t_22;

        /* "cyext_acv.pyx":5611
 * 
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]             # <<<<<<<<<<<<<<
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = __pyx_v_i;
        __pyx_t_36 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_33 * __pyx_v_R_buf.strides[0]) )));
        *((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_36 * __pyx_v_sdp_b.strides[0]) )) = (*((double *) ( /* dim=0 */ (__pyx_v_sdp_ba.data + __pyx_t_32 * __pyx_v_sdp_ba.strides[0]) )));

        /* "cyext_acv.pyx":5612
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_36 = __pyx_v_i;
        __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
        __pyx_t_12 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_33 * __pyx_v_sdp_b.strides[0]) ))) >= (*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_37 * __pyx_v_sdp.strides[0]) )))) != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":5613
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_32 = __pyx_v_i;
          __pyx_t_33 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_32 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_33 * __pyx_v_sdp.strides[0]) )) = (*((double *) ( /* dim=0 */ (__pyx_v_sdp_b.data + __pyx_t_37 * __pyx_v_sdp_b.strides[0]) )));

          /* "cyext_acv.pyx":5614
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_36 = __pyx_v_i;
          (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]) = __pyx_v_S_size;

          /* "cyext_acv.pyx":5615
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 * 
 */
          __pyx_t_38 = __pyx_v_S_size;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_s = __pyx_t_40;

            /* "cyext_acv.pyx":5616
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 * 
 *                 if S_size == X.shape[1]:
 */
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_32 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_37 * __pyx_v_s_star.strides[0]) ) + __pyx_t_32 * __pyx_v_s_star.strides[1]) )) = (__pyx_v_S[__pyx_v_s]);
          }

          /* "cyext_acv.pyx":5612
 *             for i in range(N):
 *                 sdp_b[R_buf[i]] = sdp_ba[i]
 *                 if sdp_b[R_buf[i]] >= sdp[R_buf[i]]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = sdp_b[R_buf[i]]
 *                     len_s_star[R_buf[i]] = S_size
 */
        }

        /* "cyext_acv.pyx":5618
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        __pyx_t_12 = ((__pyx_v_S_size == (__pyx_v_X.shape[1])) != 0);
        if (__pyx_t_12) {

          /* "cyext_acv.pyx":5619
 * 
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1             # <<<<<<<<<<<<<<
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) )) = 1.0;

          /* "cyext_acv.pyx":5620
 *                 if S_size == X.shape[1]:
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size             # <<<<<<<<<<<<<<
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 */
          __pyx_t_36 = __pyx_v_i;
          (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]) = __pyx_v_S_size;

          /* "cyext_acv.pyx":5621
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 */
          __pyx_t_38 = __pyx_v_S_size;
          __pyx_t_39 = __pyx_t_38;
          for (__pyx_t_40 = 0; __pyx_t_40 < __pyx_t_39; __pyx_t_40+=1) {
            __pyx_v_s = __pyx_t_40;

            /* "cyext_acv.pyx":5622
 *                     len_s_star[R_buf[i]] = S_size
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]             # <<<<<<<<<<<<<<
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1
 */
            __pyx_t_36 = __pyx_v_i;
            __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_37 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_32 * __pyx_v_s_star.strides[0]) ) + __pyx_t_37 * __pyx_v_s_star.strides[1]) )) = (__pyx_v_S[__pyx_v_s]);
          }

          /* "cyext_acv.pyx":5623
 *                     for s in range(S_size):
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)             # <<<<<<<<<<<<<<
 *                         s_star[R_buf[i], s] = -1
 * 
 */
          __pyx_t_14 = (__pyx_v_X.shape[1]);
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_15 = __pyx_t_14;
          for (__pyx_t_38 = (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]); __pyx_t_38 < __pyx_t_15; __pyx_t_38+=1) {
            __pyx_v_s = __pyx_t_38;

            /* "cyext_acv.pyx":5624
 *                         s_star[R_buf[i], s] = S[s]
 *                     for s in range(len_s_star[R_buf[i]], X.shape[1]): # to filter (important for coalition)
 *                         s_star[R_buf[i], s] = -1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(N):
 */
            __pyx_t_37 = __pyx_v_i;
            __pyx_t_32 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_37 * __pyx_v_R_buf.strides[0]) )));
            __pyx_t_33 = __pyx_v_s;
            *((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_32 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )) = -1L;
          }

          /* "cyext_acv.pyx":5618
 *                         s_star[R_buf[i], s] = S[s]
 * 
 *                 if S_size == X.shape[1]:             # <<<<<<<<<<<<<<
 *                     sdp[R_buf[i]] = 1
 *                     len_s_star[R_buf[i]] = S_size
 */
        }
      }
    }

    /* "cyext_acv.pyx":5626
 *                         s_star[R_buf[i], s] = -1
 * 
 *         for i in range(N):             # <<<<<<<<<<<<<<
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 */
    __pyx_t_8 = __pyx_v_N;
    __pyx_t_9 = __pyx_t_8;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "cyext_acv.pyx":5627
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      __pyx_t_36 = __pyx_v_i;
      __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
      __pyx_t_12 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_37 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
      if (__pyx_t_12) {

        /* "cyext_acv.pyx":5628
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])             # <<<<<<<<<<<<<<
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 */
        try {
          __pyx_v_r.push_back((__pyx_v_R[__pyx_v_i]));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 5628, __pyx_L1_error)
        }

        /* "cyext_acv.pyx":5629
 *             if sdp[R_buf[i]] >= pi_level:
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):             # <<<<<<<<<<<<<<
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 */
        __pyx_t_36 = __pyx_v_i;
        __pyx_t_22 = (__pyx_v_len_s_star[(*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )))]);
        __pyx_t_38 = __pyx_t_22;
        for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
          __pyx_v_s = __pyx_t_39;

          /* "cyext_acv.pyx":5630
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 *                     sdp_global[s_star[R_buf[i], s]] += 1             # <<<<<<<<<<<<<<
 * 
 *         for i in range(r.size()):
 */
          __pyx_t_36 = __pyx_v_i;
          __pyx_t_37 = (*((long *) ( /* dim=0 */ (__pyx_v_R_buf.data + __pyx_t_36 * __pyx_v_R_buf.strides[0]) )));
          __pyx_t_33 = __pyx_v_s;
          __pyx_t_32 = (*((long *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_s_star.data + __pyx_t_37 * __pyx_v_s_star.strides[0]) ) + __pyx_t_33 * __pyx_v_s_star.strides[1]) )));
          *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_32 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
        }

        /* "cyext_acv.pyx":5627
 * 
 *         for i in range(N):
 *             if sdp[R_buf[i]] >= pi_level:             # <<<<<<<<<<<<<<
 *                 r.push_back(R[i])
 *                 for s in range(len_s_star[R_buf[i]]):
 */
      }
    }

    /* "cyext_acv.pyx":5632
 *                     sdp_global[s_star[R_buf[i], s]] += 1
 * 
 *         for i in range(r.size()):             # <<<<<<<<<<<<<<
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()
 */
    __pyx_t_30 = __pyx_v_r.size();
    __pyx_t_31 = __pyx_t_30;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_31; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "cyext_acv.pyx":5633
 * 
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])             # <<<<<<<<<<<<<<
 *             R.pop_back()
 * 
 */
      (void)(std::remove<std::vector<long> ::iterator,long>(__pyx_v_R.begin(), __pyx_v_R.end(), (__pyx_v_r[__pyx_v_i])));

      /* "cyext_acv.pyx":5634
 *         for i in range(r.size()):
 *             std_remove[vector[long].iterator, long](R.begin(), R.end(), r[i])
 *             R.pop_back()             # <<<<<<<<<<<<<<
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 */
      __pyx_v_R.pop_back();
    }

    /* "cyext_acv.pyx":5636
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_19 = ((__pyx_v_R.size() == 0) != 0);
    if (!__pyx_t_19) {
    } else {
      goto __pyx_L53_next_and;
    }
    __pyx_t_19 = ((__pyx_v_S_size >= ((__pyx_v_X.shape[1]) / 2)) != 0);
    if (__pyx_t_19) {
    } else {
      __pyx_t_12 = __pyx_t_19;
      goto __pyx_L52_bool_binop_done;
    }
    __pyx_L53_next_and:;
    __pyx_t_19 = (__pyx_v_stop != 0);
    __pyx_t_12 = __pyx_t_19;
    __pyx_L52_bool_binop_done:;
    if (__pyx_t_12) {

      /* "cyext_acv.pyx":5637
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)
 */
      goto __pyx_L27_break;

      /* "cyext_acv.pyx":5636
 *             R.pop_back()
 * 
 *         if (R.size() == 0 or S_size >= X.shape[1]/2) and stop:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }

    /* "cyext_acv.pyx":5588
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __pyx_L27_break:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cyext_acv.pyx":5639
 *             break
 * 
 *     return np.asarray(sdp_global)/X.shape[0], np.array(s_star, dtype=np.long), np.array(len_s_star, dtype=np.long), np.array(sdp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_24)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_24);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_24, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_X.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_s_star, 2, (PyObject *(*)(char *)) __pyx_memview_get_long, (int (*)(char *, PyObject *)) __pyx_memview_set_long, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_len_s_star); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_41 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_long); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_41) < 0) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_sdp, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_24))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_24);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_24);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_24, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_24, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_24, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = PyTuple_New(4); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 5639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_24, 2, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_24, 3, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_t_41 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_24;
  __pyx_t_24 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5517
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef global_sdp_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_XDECREF(__pyx_t_24);
  __PYX_XDEC_MEMVIEW(&__pyx_t_27, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_34, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_35, 1);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_b, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_ba, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_in_data, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __PYX_XDEC_MEMVIEW(&__pyx_v_R_buf, 1);
  __Pyx_XDECREF(__pyx_v_X_arr);
  __Pyx_XDECREF(__pyx_v_y_X_arr);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_87global_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_87global_sdp_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  PyObject *__pyx_v_search_space = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("global_sdp_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,&__pyx_n_s_search_space,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 1); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 2); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 3); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 4); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 5); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 6); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 7); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 8); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 9); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 10); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 11); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 12); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 13); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 14); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 15); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_space)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, 16); __PYX_ERR(0, 5517, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "global_sdp_rf") < 0)) __PYX_ERR(0, 5517, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5517, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5517, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5517, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5517, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5518, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5518, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5518, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5518, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5519, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5519, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5519, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5519, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[12]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5519, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5520, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[15]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5520, __pyx_L3_error)
    __pyx_v_search_space = ((PyObject*)values[16]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("global_sdp_rf", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5517, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 5519, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_search_space), (&PyList_Type), 1, "search_space", 1))) __PYX_ERR(0, 5520, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_86global_sdp_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_86global_sdp_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("global_sdp_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5517, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_global_sdp_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.global_sdp_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cyext_acv.pyx":5646
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef sufficient_expl_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

static PyObject *__pyx_pw_9cyext_acv_89sufficient_expl_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_9cyext_acv_sufficient_expl_rf(__Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int &__pyx_v_classifier, double &__pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, CYTHON_UNUSED int __pyx_v_stop, PyObject *__pyx_v_search_space, CYTHON_UNUSED int __pyx_skip_dispatch) {
  unsigned int __pyx_v_N;
  unsigned int __pyx_v_m;
  __Pyx_memviewslice __pyx_v_sdp = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_sdp_global = { 0, 0, { 0 }, { 0 }, { 0 } };
  double __pyx_v_K;
  int __pyx_v_i;
  int __pyx_v_s;
  int __pyx_v_s_0;
  int __pyx_v_s_1;
  int __pyx_v_S_size;
  int __pyx_v_j;
  int __pyx_v_max_size;
  int __pyx_v_size;
  int __pyx_v_subset;
  int __pyx_v_a;
  int __pyx_v_k;
  std::vector<int>  __pyx_v_S;
  CYTHON_UNUSED std::vector<int>  __pyx_v_len_s_star;
  PyObject *__pyx_v_power = 0;
  PyObject *__pyx_v_va_id = 0;
  PyObject *__pyx_v_remove_va = NULL;
  Py_ssize_t __pyx_v_ci;
  Py_ssize_t __pyx_v_cj;
  PyObject *__pyx_v_power_b = NULL;
  PyObject *__pyx_v_co = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_power_cpp;
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_s_star = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED long __pyx_v_power_set_size;
  std::vector<std::vector<std::vector<long> > >  __pyx_v_sufficient;
  std::vector<std::vector<double> >  __pyx_v_sufficient_sdp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  std::vector<int>  __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  long __pyx_t_17;
  long __pyx_t_18;
  int __pyx_t_19;
  PyObject *(*__pyx_t_20)(PyObject *);
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  std::vector<std::vector<std::vector<long> > >  __pyx_t_23;
  __Pyx_memviewslice __pyx_t_24 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_25;
  unsigned int __pyx_t_26;
  std::vector<std::vector<double> >  __pyx_t_27;
  std::vector<std::vector<long> > ::size_type __pyx_t_28;
  std::vector<std::vector<long> > ::size_type __pyx_t_29;
  std::vector<long> ::size_type __pyx_t_30;
  std::vector<long> ::size_type __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  std::vector<std::vector<long> > ::size_type __pyx_t_33;
  std::vector<std::vector<long> > ::size_type __pyx_t_34;
  int __pyx_t_35;
  int __pyx_t_36;
  int __pyx_t_37;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sufficient_expl_rf", 0);

  /* "cyext_acv.pyx":5651
 *             int minimal, bint stop, list search_space):
 * 
 *     cdef unsigned int N = X.shape[0]             # <<<<<<<<<<<<<<
 *     cdef unsigned int m = X.shape[1]
 * 
 */
  __pyx_v_N = (__pyx_v_X.shape[0]);

  /* "cyext_acv.pyx":5652
 * 
 *     cdef unsigned int N = X.shape[0]
 *     cdef unsigned int m = X.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     cdef double[:] sdp, sdp_b, sdp_ba
 */
  __pyx_v_m = (__pyx_v_X.shape[1]);

  /* "cyext_acv.pyx":5656
 *     cdef double[:] sdp, sdp_b, sdp_ba
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))             # <<<<<<<<<<<<<<
 *     sdp_global = np.zeros((m))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5657
 *     cdef double[:] sdp_global
 *     sdp = np.zeros((N))
 *     sdp_global = np.zeros((m))             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int it, it_s, a_it, b_it, o_all, p, p_s, nb_leaf, p_u, p_d, p_su, p_sd, down, up
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_m); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_sdp_global = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cyext_acv.pyx":5662
 *     cdef double ss, ss_a, ss_u, ss_d, K
 *     cdef int b, leaf_numb, i, s, s_0, s_1, S_size, j, max_size, size, subset, a, k
 *     K = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[int] S, len_s_star
 */
  __pyx_v_K = 0.0;

  /* "cyext_acv.pyx":5665
 * 
 *     cdef vector[int] S, len_s_star
 *     len_s_star = np.zeros((N), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef list power, va_id
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_6); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_len_s_star = __pyx_t_7;

  /* "cyext_acv.pyx":5669
 *     cdef list power, va_id
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
  if (unlikely(__pyx_v_C == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 5669, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyObject_RichCompare(PyList_GET_ITEM(__pyx_v_C, 0), __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 5669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {

    /* "cyext_acv.pyx":5670
 * 
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]             # <<<<<<<<<<<<<<
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 5670, __pyx_L1_error)
    }
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5670, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_ci = __pyx_t_11;
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 5670, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_12 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5670, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
        __pyx_v_cj = __pyx_t_14;
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5670, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 5670, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __pyx_v_remove_va = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":5671
 *     if C[0] != []:
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]             # <<<<<<<<<<<<<<
 *         for ci in range(len(C)):
 *             i = 0
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 5671, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_6); __pyx_t_9 = 0;
    for (;;) {
      if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5671, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5671, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_i = __pyx_t_15;
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_remove_va, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 5671, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_16 = (__pyx_t_8 != 0);
      if (__pyx_t_16) {
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5671, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5671, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_3);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 5671, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cyext_acv.pyx":5672
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):             # <<<<<<<<<<<<<<
 *             i = 0
 *             for cj in range(len(C[ci])):
 */
    if (unlikely(__pyx_v_C == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 5672, __pyx_L1_error)
    }
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_C); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5672, __pyx_L1_error)
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_ci = __pyx_t_11;

      /* "cyext_acv.pyx":5673
 *         va_id = [[i] for i in search_space if i not in remove_va]
 *         for ci in range(len(C)):
 *             i = 0             # <<<<<<<<<<<<<<
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 */
      __pyx_v_i = 0;

      /* "cyext_acv.pyx":5674
 *         for ci in range(len(C)):
 *             i = 0
 *             for cj in range(len(C[ci])):             # <<<<<<<<<<<<<<
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 */
      if (unlikely(__pyx_v_C == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 5674, __pyx_L1_error)
      }
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_12 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5674, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __pyx_t_12;
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
        __pyx_v_cj = __pyx_t_14;

        /* "cyext_acv.pyx":5675
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5675, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci), __pyx_v_cj, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_16 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_search_space, Py_EQ)); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 5675, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_8 = (__pyx_t_16 != 0);
        if (__pyx_t_8) {

          /* "cyext_acv.pyx":5676
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:
 *                     i += 1             # <<<<<<<<<<<<<<
 *                     break
 *             if i != 0:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "cyext_acv.pyx":5677
 *                 if C[ci][cj] in search_space:
 *                     i += 1
 *                     break             # <<<<<<<<<<<<<<
 *             if i != 0:
 *                 va_id += [C[ci]]
 */
          goto __pyx_L14_break;

          /* "cyext_acv.pyx":5675
 *             i = 0
 *             for cj in range(len(C[ci])):
 *                 if C[ci][cj] in search_space:             # <<<<<<<<<<<<<<
 *                     i += 1
 *                     break
 */
        }
      }
      __pyx_L14_break:;

      /* "cyext_acv.pyx":5678
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      __pyx_t_8 = ((__pyx_v_i != 0) != 0);
      if (__pyx_t_8) {

        /* "cyext_acv.pyx":5679
 *                     break
 *             if i != 0:
 *                 va_id += [C[ci]]             # <<<<<<<<<<<<<<
 *     else:
 *         va_id = [[i] for i in search_space]
 */
        if (unlikely(__pyx_v_C == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 5679, __pyx_L1_error)
        }
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __Pyx_GIVEREF(PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        PyList_SET_ITEM(__pyx_t_1, 0, PyList_GET_ITEM(__pyx_v_C, __pyx_v_ci));
        __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_va_id, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_va_id, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "cyext_acv.pyx":5678
 *                     i += 1
 *                     break
 *             if i != 0:             # <<<<<<<<<<<<<<
 *                 va_id += [C[ci]]
 *     else:
 */
      }
    }

    /* "cyext_acv.pyx":5669
 *     cdef list power, va_id
 * 
 *     if C[0] != []:             # <<<<<<<<<<<<<<
 *         remove_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
 *         va_id = [[i] for i in search_space if i not in remove_va]
 */
    goto __pyx_L3;
  }

  /* "cyext_acv.pyx":5681
 *                 va_id += [C[ci]]
 *     else:
 *         va_id = [[i] for i in search_space]             # <<<<<<<<<<<<<<
 * 
 *     m = len(va_id)
 */
  /*else*/ {
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_v_search_space == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 5681, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_search_space; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    for (;;) {
      if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5681, __pyx_L1_error)
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5681, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5681, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_i = __pyx_t_15;
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5681, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5681, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 5681, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_va_id = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "cyext_acv.pyx":5683
 *         va_id = [[i] for i in search_space]
 * 
 *     m = len(va_id)             # <<<<<<<<<<<<<<
 *     power = []
 *     max_size = 0
 */
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_va_id); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5683, __pyx_L1_error)
  __pyx_v_m = __pyx_t_9;

  /* "cyext_acv.pyx":5684
 * 
 *     m = len(va_id)
 *     power = []             # <<<<<<<<<<<<<<
 *     max_size = 0
 *     for size in range(m + 1):
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_power = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "cyext_acv.pyx":5685
 *     m = len(va_id)
 *     power = []
 *     max_size = 0             # <<<<<<<<<<<<<<
 *     for size in range(m + 1):
 *         power_b = []
 */
  __pyx_v_max_size = 0;

  /* "cyext_acv.pyx":5686
 *     power = []
 *     max_size = 0
 *     for size in range(m + 1):             # <<<<<<<<<<<<<<
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 */
  __pyx_t_17 = (__pyx_v_m + 1);
  __pyx_t_18 = __pyx_t_17;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_18; __pyx_t_15+=1) {
    __pyx_v_size = __pyx_t_15;

    /* "cyext_acv.pyx":5687
 *     max_size = 0
 *     for size in range(m + 1):
 *         power_b = []             # <<<<<<<<<<<<<<
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_power_b, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "cyext_acv.pyx":5688
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_combinations); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    __pyx_t_19 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_19 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_va_id, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5688, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_va_id, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5688, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_v_va_id);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_19, __pyx_v_va_id);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_19, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_3 = __pyx_t_6; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
      __pyx_t_20 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_20 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 5688, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_20)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5688, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5688, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5688, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5688, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_20(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 5688, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_co, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "cyext_acv.pyx":5689
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))             # <<<<<<<<<<<<<<
 *             max_size += 1
 *         power.append(power_b)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_List(__pyx_v_co); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2);
      __pyx_t_4 = 0;
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_21 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_21)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_21);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_21, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_power_b, __pyx_t_6); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 5689, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "cyext_acv.pyx":5690
 *         for co in itertools.combinations(va_id, size):
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1             # <<<<<<<<<<<<<<
 *         power.append(power_b)
 *         if max_size >= 2**15:
 */
      __pyx_v_max_size = (__pyx_v_max_size + 1);

      /* "cyext_acv.pyx":5688
 *     for size in range(m + 1):
 *         power_b = []
 *         for co in itertools.combinations(va_id, size):             # <<<<<<<<<<<<<<
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cyext_acv.pyx":5691
 *             power_b.append(np.array(sum(list(co),[])))
 *             max_size += 1
 *         power.append(power_b)             # <<<<<<<<<<<<<<
 *         if max_size >= 2**15:
 *             break
 */
    __pyx_t_22 = __Pyx_PyList_Append(__pyx_v_power, __pyx_v_power_b); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 5691, __pyx_L1_error)

    /* "cyext_acv.pyx":5692
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_8 = ((__pyx_v_max_size >= 0x8000) != 0);
    if (__pyx_t_8) {

      /* "cyext_acv.pyx":5693
 *         power.append(power_b)
 *         if max_size >= 2**15:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power
 */
      goto __pyx_L20_break;

      /* "cyext_acv.pyx":5692
 *             max_size += 1
 *         power.append(power_b)
 *         if max_size >= 2**15:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L20_break:;

  /* "cyext_acv.pyx":5695
 *             break
 * 
 *     cdef vector[vector[vector[long]]] power_cpp = power             # <<<<<<<<<<<<<<
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)
 */
  __pyx_t_23 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_power); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5695, __pyx_L1_error)
  __pyx_v_power_cpp = __pyx_t_23;

  /* "cyext_acv.pyx":5697
 *     cdef vector[vector[vector[long]]] power_cpp = power
 *     cdef long[:, :] s_star
 *     s_star = -1*np.ones((N, X.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_N); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyInt_FromSsize_t((__pyx_v_X.shape[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_21) < 0) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_int_neg_1, __pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_24 = __Pyx_PyObject_to_MemoryviewSlice_dsds_long(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_24.memview)) __PYX_ERR(0, 5697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_s_star = __pyx_t_24;
  __pyx_t_24.memview = NULL;
  __pyx_t_24.data = NULL;

  /* "cyext_acv.pyx":5700
 * 
 * 
 *     cdef long power_set_size = 2**m             # <<<<<<<<<<<<<<
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 */
  __pyx_v_power_set_size = __Pyx_pow_long(2, ((long)__pyx_v_m));

  /* "cyext_acv.pyx":5701
 * 
 *     cdef long power_set_size = 2**m
 *     S = np.zeros((data.shape[1]), dtype=np.int)             # <<<<<<<<<<<<<<
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t((__pyx_v_data.shape[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_S = __pyx_t_7;

  /* "cyext_acv.pyx":5703
 *     S = np.zeros((data.shape[1]), dtype=np.int)
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]             # <<<<<<<<<<<<<<
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_25 = __pyx_v_N;
  __pyx_t_26 = __pyx_t_25;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_26; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_int_neg_1);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 5703, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_23 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sufficient = __pyx_t_23;

  /* "cyext_acv.pyx":5704
 * 
 *     cdef vector[vector[vector[long]]] sufficient = [[[-1]] for i in range(N)]
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]             # <<<<<<<<<<<<<<
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_25 = __pyx_v_N;
  __pyx_t_26 = __pyx_t_25;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_26; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_int_neg_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 5704, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_27 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(__pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5704, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sufficient_sdp = __pyx_t_27;

  /* "cyext_acv.pyx":5706
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_minimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = __Pyx_PyInt_From_long((__pyx_v_m + 1)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_21);
  __pyx_t_2 = 0;
  __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_21) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_21);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_20 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_20 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 5706, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_20)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5706, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 5706, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5706, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_20(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 5706, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5706, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_s_0 = __pyx_t_15;

    /* "cyext_acv.pyx":5707
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):             # <<<<<<<<<<<<<<
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]
 */
    __pyx_t_28 = (__pyx_v_power_cpp[__pyx_v_s_0]).size();
    __pyx_t_29 = __pyx_t_28;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_29; __pyx_t_15+=1) {
      __pyx_v_s_1 = __pyx_t_15;

      /* "cyext_acv.pyx":5708
 *     for s_0 in tqdm(range(minimal, m + 1)):
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):             # <<<<<<<<<<<<<<
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 */
      __pyx_t_30 = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();
      __pyx_t_31 = __pyx_t_30;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_31; __pyx_t_19+=1) {
        __pyx_v_i = __pyx_t_19;

        /* "cyext_acv.pyx":5709
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 *                 S[i] = power_cpp[s_0][s_1][i]             # <<<<<<<<<<<<<<
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 */
        (__pyx_v_S[__pyx_v_i]) = (((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1])[__pyx_v_i]);
      }

      /* "cyext_acv.pyx":5711
 *                 S[i] = power_cpp[s_0][s_1][i]
 * 
 *             S_size = power_cpp[s_0][s_1].size()             # <<<<<<<<<<<<<<
 *             sdp = compute_sdp_rf_same_set(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)
 */
      __pyx_v_S_size = ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).size();

      /* "cyext_acv.pyx":5712
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_same_set(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_3 = __pyx_convert_vector_to_py_int(__pyx_v_S); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_21 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, __pyx_v_S_size, NULL, NULL, NULL, 0, 1, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __pyx_convert_vector_from_py_int(__pyx_t_21); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "cyext_acv.pyx":5713
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_same_set(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,
 *                                     children_right, max_depth, min_node_size, classifier, t)             # <<<<<<<<<<<<<<
 * 
 *             for i in range(N):
 */
      __pyx_t_21 = __pyx_f_9cyext_acv_compute_sdp_rf_same_set(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_t_7, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);

      /* "cyext_acv.pyx":5712
 * 
 *             S_size = power_cpp[s_0][s_1].size()
 *             sdp = compute_sdp_rf_same_set(X, y_X, data, y_data, S[:S_size], features, thresholds,  children_left,             # <<<<<<<<<<<<<<
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_double(__pyx_t_21, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 5712, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
      __pyx_v_sdp = __pyx_t_5;
      __pyx_t_5.memview = NULL;
      __pyx_t_5.data = NULL;

      /* "cyext_acv.pyx":5715
 *                                     children_right, max_depth, min_node_size, classifier, t)
 * 
 *             for i in range(N):             # <<<<<<<<<<<<<<
 *                 if sdp[i] >= pi_level:
 *                     subset = 0
 */
      __pyx_t_25 = __pyx_v_N;
      __pyx_t_26 = __pyx_t_25;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_26; __pyx_t_19+=1) {
        __pyx_v_i = __pyx_t_19;

        /* "cyext_acv.pyx":5716
 * 
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:             # <<<<<<<<<<<<<<
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 */
        __pyx_t_32 = __pyx_v_i;
        __pyx_t_8 = (((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) ))) >= __pyx_v_pi_level) != 0);
        if (__pyx_t_8) {

          /* "cyext_acv.pyx":5717
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:
 *                     subset = 0             # <<<<<<<<<<<<<<
 *                     for j in range(sufficient[i].size()):
 *                         a = 0
 */
          __pyx_v_subset = 0;

          /* "cyext_acv.pyx":5718
 *                 if sdp[i] >= pi_level:
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):             # <<<<<<<<<<<<<<
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 */
          __pyx_t_33 = (__pyx_v_sufficient[__pyx_v_i]).size();
          __pyx_t_34 = __pyx_t_33;
          for (__pyx_t_35 = 0; __pyx_t_35 < __pyx_t_34; __pyx_t_35+=1) {
            __pyx_v_j = __pyx_t_35;

            /* "cyext_acv.pyx":5719
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 *                         a = 0             # <<<<<<<<<<<<<<
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 */
            __pyx_v_a = 0;

            /* "cyext_acv.pyx":5720
 *                     for j in range(sufficient[i].size()):
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):             # <<<<<<<<<<<<<<
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 */
            __pyx_t_30 = ((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j]).size();
            __pyx_t_31 = __pyx_t_30;
            for (__pyx_t_36 = 0; __pyx_t_36 < __pyx_t_31; __pyx_t_36+=1) {
              __pyx_v_k = __pyx_t_36;

              /* "cyext_acv.pyx":5721
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():             # <<<<<<<<<<<<<<
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 */
              __pyx_t_8 = ((std::find<std::vector<long> ::iterator,long>(((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).begin(), ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).end(), (((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j])[__pyx_v_k])) != ((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]).end()) != 0);
              if (__pyx_t_8) {

                /* "cyext_acv.pyx":5722
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1             # <<<<<<<<<<<<<<
 *                         if a == sufficient[i][j].size():
 *                             subset = 1
 */
                __pyx_v_a = (__pyx_v_a + 1);

                /* "cyext_acv.pyx":5721
 *                         a = 0
 *                         for k in range(sufficient[i][j].size()):
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():             # <<<<<<<<<<<<<<
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 */
              }
            }

            /* "cyext_acv.pyx":5723
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 *                         if a == sufficient[i][j].size():             # <<<<<<<<<<<<<<
 *                             subset = 1
 *                             break
 */
            __pyx_t_8 = ((__pyx_v_a == ((__pyx_v_sufficient[__pyx_v_i])[__pyx_v_j]).size()) != 0);
            if (__pyx_t_8) {

              /* "cyext_acv.pyx":5724
 *                                 a += 1
 *                         if a == sufficient[i][j].size():
 *                             subset = 1             # <<<<<<<<<<<<<<
 *                             break
 * 
 */
              __pyx_v_subset = 1;

              /* "cyext_acv.pyx":5725
 *                         if a == sufficient[i][j].size():
 *                             subset = 1
 *                             break             # <<<<<<<<<<<<<<
 * 
 *                     if subset == 0:
 */
              goto __pyx_L38_break;

              /* "cyext_acv.pyx":5723
 *                             if std_find[vector[long].iterator, long](power_cpp[s_0][s_1].begin(), power_cpp[s_0][s_1].end(), sufficient[i][j][k]) != power_cpp[s_0][s_1].end():
 *                                 a += 1
 *                         if a == sufficient[i][j].size():             # <<<<<<<<<<<<<<
 *                             subset = 1
 *                             break
 */
            }
          }
          __pyx_L38_break:;

          /* "cyext_acv.pyx":5727
 *                             break
 * 
 *                     if subset == 0:             # <<<<<<<<<<<<<<
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 */
          __pyx_t_8 = ((__pyx_v_subset == 0) != 0);
          if (__pyx_t_8) {

            /* "cyext_acv.pyx":5728
 * 
 *                     if subset == 0:
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])             # <<<<<<<<<<<<<<
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):
 */
            try {
              (__pyx_v_sufficient[__pyx_v_i]).push_back(((__pyx_v_power_cpp[__pyx_v_s_0])[__pyx_v_s_1]));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 5728, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5729
 *                     if subset == 0:
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])             # <<<<<<<<<<<<<<
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1
 */
            __pyx_t_32 = __pyx_v_i;
            try {
              (__pyx_v_sufficient_sdp[__pyx_v_i]).push_back((*((double *) ( /* dim=0 */ (__pyx_v_sdp.data + __pyx_t_32 * __pyx_v_sdp.strides[0]) ))));
            } catch(...) {
              __Pyx_CppExn2PyErr();
              __PYX_ERR(0, 5729, __pyx_L1_error)
            }

            /* "cyext_acv.pyx":5730
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):             # <<<<<<<<<<<<<<
 *                             sdp_global[S[s]] += 1
 *                         K += 1
 */
            __pyx_t_35 = __pyx_v_S_size;
            __pyx_t_36 = __pyx_t_35;
            for (__pyx_t_37 = 0; __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
              __pyx_v_s = __pyx_t_37;

              /* "cyext_acv.pyx":5731
 *                         sufficient_sdp[i].push_back(sdp[i])
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1             # <<<<<<<<<<<<<<
 *                         K += 1
 * 
 */
              __pyx_t_32 = (__pyx_v_S[__pyx_v_s]);
              *((double *) ( /* dim=0 */ (__pyx_v_sdp_global.data + __pyx_t_32 * __pyx_v_sdp_global.strides[0]) )) += 1.0;
            }

            /* "cyext_acv.pyx":5732
 *                         for s in range(S_size):
 *                             sdp_global[S[s]] += 1
 *                         K += 1             # <<<<<<<<<<<<<<
 * 
 *     return sufficient, sufficient_sdp, np.asarray(sdp_global)/K
 */
            __pyx_v_K = (__pyx_v_K + 1.0);

            /* "cyext_acv.pyx":5727
 *                             break
 * 
 *                     if subset == 0:             # <<<<<<<<<<<<<<
 *                         sufficient[i].push_back(power_cpp[s_0][s_1])
 *                         sufficient_sdp[i].push_back(sdp[i])
 */
          }

          /* "cyext_acv.pyx":5716
 * 
 *             for i in range(N):
 *                 if sdp[i] >= pi_level:             # <<<<<<<<<<<<<<
 *                     subset = 0
 *                     for j in range(sufficient[i].size()):
 */
        }
      }
    }

    /* "cyext_acv.pyx":5706
 *     cdef vector[vector[double]] sufficient_sdp = [[-1] for i in range(N)]
 * 
 *     for s_0 in tqdm(range(minimal, m + 1)):             # <<<<<<<<<<<<<<
 *         for s_1 in range(0, power_cpp[s_0].size()):
 *             for i in range(power_cpp[s_0][s_1].size()):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":5734
 *                         K += 1
 * 
 *     return sufficient, sufficient_sdp, np.asarray(sdp_global)/K             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(__pyx_v_sufficient); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(__pyx_v_sufficient_sdp); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_sdp_global, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_K); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_21 = 0;
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cyext_acv.pyx":5646
 * @cython.nonecheck(False)
 * @cython.cdivision(True)
 * cpdef sufficient_expl_rf(double[:, :] X, double[::1] y_X,  double[:, :] data, double[::1] y_data,             # <<<<<<<<<<<<<<
 *         int[:, :] features, double[:, :] thresholds,  int[:, :] children_left, int[:, :] children_right,
 *         int max_depth, int min_node_size, int & classifier, double & t, list C, double pi_level,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_21);
  __PYX_XDEC_MEMVIEW(&__pyx_t_24, 1);
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sdp_global, 1);
  __Pyx_XDECREF(__pyx_v_power);
  __Pyx_XDECREF(__pyx_v_va_id);
  __Pyx_XDECREF(__pyx_v_remove_va);
  __Pyx_XDECREF(__pyx_v_power_b);
  __Pyx_XDECREF(__pyx_v_co);
  __PYX_XDEC_MEMVIEW(&__pyx_v_s_star, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9cyext_acv_89sufficient_expl_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_9cyext_acv_89sufficient_expl_rf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_X = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_y_data = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_features = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_thresholds = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_left = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_children_right = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_max_depth;
  int __pyx_v_min_node_size;
  int __pyx_v_classifier;
  double __pyx_v_t;
  PyObject *__pyx_v_C = 0;
  double __pyx_v_pi_level;
  int __pyx_v_minimal;
  int __pyx_v_stop;
  PyObject *__pyx_v_search_space = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sufficient_expl_rf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_X,&__pyx_n_s_y_X,&__pyx_n_s_data,&__pyx_n_s_y_data,&__pyx_n_s_features,&__pyx_n_s_thresholds,&__pyx_n_s_children_left,&__pyx_n_s_children_right,&__pyx_n_s_max_depth,&__pyx_n_s_min_node_size,&__pyx_n_s_classifier,&__pyx_n_s_t,&__pyx_n_s_C,&__pyx_n_s_pi_level,&__pyx_n_s_minimal,&__pyx_n_s_stop,&__pyx_n_s_search_space,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_X)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 1); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 2); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 3); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 4); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_thresholds)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 5); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_left)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 6); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 7); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 8); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_node_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 9); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_classifier)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 10); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 11); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 12); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pi_level)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 13); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 14); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 15); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_space)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, 16); __PYX_ERR(0, 5646, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sufficient_expl_rf") < 0)) __PYX_ERR(0, 5646, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_X = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_X.memview)) __PYX_ERR(0, 5646, __pyx_L3_error)
    __pyx_v_y_X = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_X.memview)) __PYX_ERR(0, 5646, __pyx_L3_error)
    __pyx_v_data = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_data.memview)) __PYX_ERR(0, 5646, __pyx_L3_error)
    __pyx_v_y_data = __Pyx_PyObject_to_MemoryviewSlice_dc_double(values[3], PyBUF_WRITABLE); if (unlikely(!__pyx_v_y_data.memview)) __PYX_ERR(0, 5646, __pyx_L3_error)
    __pyx_v_features = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[4], PyBUF_WRITABLE); if (unlikely(!__pyx_v_features.memview)) __PYX_ERR(0, 5647, __pyx_L3_error)
    __pyx_v_thresholds = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(values[5], PyBUF_WRITABLE); if (unlikely(!__pyx_v_thresholds.memview)) __PYX_ERR(0, 5647, __pyx_L3_error)
    __pyx_v_children_left = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[6], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_left.memview)) __PYX_ERR(0, 5647, __pyx_L3_error)
    __pyx_v_children_right = __Pyx_PyObject_to_MemoryviewSlice_dsds_int(values[7], PyBUF_WRITABLE); if (unlikely(!__pyx_v_children_right.memview)) __PYX_ERR(0, 5647, __pyx_L3_error)
    __pyx_v_max_depth = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_max_depth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5648, __pyx_L3_error)
    __pyx_v_min_node_size = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_min_node_size == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5648, __pyx_L3_error)
    __pyx_v_classifier = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_classifier == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5648, __pyx_L3_error)
    __pyx_v_t = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_t == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5648, __pyx_L3_error)
    __pyx_v_C = ((PyObject*)values[12]);
    __pyx_v_pi_level = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_pi_level == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 5648, __pyx_L3_error)
    __pyx_v_minimal = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_minimal == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5649, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyObject_IsTrue(values[15]); if (unlikely((__pyx_v_stop == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 5649, __pyx_L3_error)
    __pyx_v_search_space = ((PyObject*)values[16]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sufficient_expl_rf", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5646, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), (&PyList_Type), 1, "C", 1))) __PYX_ERR(0, 5648, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_search_space), (&PyList_Type), 1, "search_space", 1))) __PYX_ERR(0, 5649, __pyx_L1_error)
  __pyx_r = __pyx_pf_9cyext_acv_88sufficient_expl_rf(__pyx_self, __pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9cyext_acv_88sufficient_expl_rf(CYTHON_UNUSED PyObject *__pyx_self, __Pyx_memviewslice __pyx_v_X, __Pyx_memviewslice __pyx_v_y_X, __Pyx_memviewslice __pyx_v_data, __Pyx_memviewslice __pyx_v_y_data, __Pyx_memviewslice __pyx_v_features, __Pyx_memviewslice __pyx_v_thresholds, __Pyx_memviewslice __pyx_v_children_left, __Pyx_memviewslice __pyx_v_children_right, int __pyx_v_max_depth, int __pyx_v_min_node_size, int __pyx_v_classifier, double __pyx_v_t, PyObject *__pyx_v_C, double __pyx_v_pi_level, int __pyx_v_minimal, int __pyx_v_stop, PyObject *__pyx_v_search_space) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sufficient_expl_rf", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_X.memview)) { __Pyx_RaiseUnboundLocalError("X"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_X.memview)) { __Pyx_RaiseUnboundLocalError("y_X"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_data.memview)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_y_data.memview)) { __Pyx_RaiseUnboundLocalError("y_data"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_features.memview)) { __Pyx_RaiseUnboundLocalError("features"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_thresholds.memview)) { __Pyx_RaiseUnboundLocalError("thresholds"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_left.memview)) { __Pyx_RaiseUnboundLocalError("children_left"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  if (unlikely(!__pyx_v_children_right.memview)) { __Pyx_RaiseUnboundLocalError("children_right"); __PYX_ERR(0, 5646, __pyx_L1_error) }
  __pyx_t_1 = __pyx_f_9cyext_acv_sufficient_expl_rf(__pyx_v_X, __pyx_v_y_X, __pyx_v_data, __pyx_v_y_data, __pyx_v_features, __pyx_v_thresholds, __pyx_v_children_left, __pyx_v_children_right, __pyx_v_max_depth, __pyx_v_min_node_size, __pyx_v_classifier, __pyx_v_t, __pyx_v_C, __pyx_v_pi_level, __pyx_v_minimal, __pyx_v_stop, __pyx_v_search_space, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cyext_acv.sufficient_expl_rf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_X, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_y_data, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_features, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_thresholds, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_left, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_children_right, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_long")
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<long>  __pyx_convert_vector_from_py_long(PyObject *__pyx_v_o) {
  std::vector<long>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<long>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_long", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_v_item); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((long)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_long")
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::vector<long> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(PyObject *__pyx_v_o) {
  std::vector<std::vector<long> >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::vector<long> >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::vector<long>  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_long(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::vector<long> )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::vector<std::vector<long> > >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(PyObject *__pyx_v_o) {
  std::vector<std::vector<std::vector<long> > >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::vector<std::vector<long> > >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::vector<std::vector<long> >  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_std_3a__3a_vector_3c_long_3e___(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::vector<std::vector<long> > )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<int>  __pyx_convert_vector_from_py_int(PyObject *__pyx_v_o) {
  std::vector<int>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<int>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_int", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_int(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_item); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((int)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_int(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_int")
 * cdef vector[X] __pyx_convert_vector_from_py_int(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::vector<int> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(PyObject *__pyx_v_o) {
  std::vector<std::vector<int> >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::vector<int> >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::vector<int>  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_int(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::vector<int> )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_vector_3c_int_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_int(const std::vector<int>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_int", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_double")
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<double>  __pyx_convert_vector_from_py_double(PyObject *__pyx_v_o) {
  std::vector<double>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<double>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_double", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_v_item); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((double)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_double")
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::vector<double> >  __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(PyObject *__pyx_v_o) {
  std::vector<std::vector<double> >  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::vector<double> >  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::vector<double>  __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_vector_from_py_double(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::vector<double> )__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_vector_3c_double_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_long")
 * cdef object __pyx_convert_vector_to_py_long(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_long(const std::vector<long>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_long", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_long")
 * cdef object __pyx_convert_vector_to_py_long(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_long((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_long")
 * cdef object __pyx_convert_vector_to_py_long(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___(const std::vector<std::vector<long> >  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_vector_to_py_long((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(const std::vector<std::vector<std::vector<long> > >  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_vector_to_py_std_3a__3a_vector_3c_long_3e___((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_std_3a__3a_vector_3c_std_3a__3a_vector_3c_long_3e____3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_double(const std::vector<double>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_double", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_double")
 * cdef object __pyx_convert_vector_to_py_double(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = PyFloat_FromDouble((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_double")
 * cdef object __pyx_convert_vector_to_py_double(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(const std::vector<std::vector<double> >  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __pyx_convert_vector_to_py_double((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___")
 * cdef object __pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_std_3a__3a_vector_3c_double_3e___", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(2, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(2, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(2, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(2, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(2, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(2, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(2, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(2, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(2, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(2, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(2, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(2, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(2, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(2, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(2, 445, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(2, 446, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 456, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(2, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(2, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(2, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(2, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(2, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__20, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 622, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(2, 628, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__2);
            __Pyx_GIVEREF(__pyx_slice__2);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__2);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(2, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__2);
        __Pyx_GIVEREF(__pyx_slice__2);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__2);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(2, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(2, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1179
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1181
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1182
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1184
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1197
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1198
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1199
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1201
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1202
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1203
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1205
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1219
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1220
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1222
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1224
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 1224, __pyx_L1_error)

    /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1227
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1228
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1229
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1230
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1231
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1233
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1237
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1239
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1242
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1244
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1246
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1254
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1253
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(2, 1253, __pyx_L1_error)

  /* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1258
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 1258, __pyx_L1_error)

  /* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1263
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 1263, __pyx_L1_error)

    /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1265
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(2, 1265, __pyx_L1_error)
  }

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1276
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1277
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1279
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1280
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1281
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1285
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1289
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1291
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1294
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1295
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1297
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1297, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1300
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1300, __pyx_L1_error)

      /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1305
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1307
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(2, 1307, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1308
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1314
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1321
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1322
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1323
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1329
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1329, __pyx_L1_error)

    /* "View.MemoryView":1330
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1330, __pyx_L1_error)

    /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1333
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1334
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1337
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1344
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1346
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1347
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1348
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1349
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1351
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1352
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1353
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1354
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1367
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1374
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1381
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1384
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1386
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1388
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1389
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1391
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1400
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1401
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1403
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1411
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1412
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1415
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1416
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1417
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1419
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1420
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1422
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *__pyx_freelist_9cyext_acv___pyx_scope_struct__powerset[8];
static int __pyx_freecount_9cyext_acv___pyx_scope_struct__powerset = 0;

static PyObject *__pyx_tp_new_9cyext_acv___pyx_scope_struct__powerset(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9cyext_acv___pyx_scope_struct__powerset > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset)))) {
    o = (PyObject*)__pyx_freelist_9cyext_acv___pyx_scope_struct__powerset[--__pyx_freecount_9cyext_acv___pyx_scope_struct__powerset];
    memset(o, 0, sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9cyext_acv___pyx_scope_struct__powerset(PyObject *o) {
  struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *p = (struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_s);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9cyext_acv___pyx_scope_struct__powerset < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset)))) {
    __pyx_freelist_9cyext_acv___pyx_scope_struct__powerset[__pyx_freecount_9cyext_acv___pyx_scope_struct__powerset++] = ((struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9cyext_acv___pyx_scope_struct__powerset(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *p = (struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)o;
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9cyext_acv___pyx_scope_struct__powerset(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *p = (struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset *)o;
  tmp = ((PyObject*)p->__pyx_v_s);
  p->__pyx_v_s = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9cyext_acv___pyx_scope_struct__powerset = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv.__pyx_scope_struct__powerset", /*tp_name*/
  sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct__powerset), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9cyext_acv___pyx_scope_struct__powerset, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9cyext_acv___pyx_scope_struct__powerset, /*tp_traverse*/
  __pyx_tp_clear_9cyext_acv___pyx_scope_struct__powerset, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9cyext_acv___pyx_scope_struct__powerset, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *__pyx_freelist_9cyext_acv___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_9cyext_acv___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_9cyext_acv___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9cyext_acv___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9cyext_acv___pyx_scope_struct_1_genexpr[--__pyx_freecount_9cyext_acv___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9cyext_acv___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9cyext_acv___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_9cyext_acv___pyx_scope_struct_1_genexpr[__pyx_freecount_9cyext_acv___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9cyext_acv___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9cyext_acv___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9cyext_acv___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9cyext_acv___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9cyext_acv___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "cyext_acv._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"compute_exp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_1compute_exp, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_exp_normalized", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_3compute_exp_normalized, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_exp_cat", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_5compute_exp_cat, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_clf_cat", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_7compute_sdp_clf_cat, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_clf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_9compute_sdp_clf, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_reg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_13compute_sdp_reg, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_reg_cat", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_15compute_sdp_reg_cat, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_reg_cat", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_17global_sdp_reg_cat, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_reg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_19global_sdp_reg, METH_VARARGS|METH_KEYWORDS, 0},
  {"shap_values_leaves_pa", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_21shap_values_leaves_pa, METH_VARARGS|METH_KEYWORDS, 0},
  {"shap_values_leaves_normalized", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_23shap_values_leaves_normalized, METH_VARARGS|METH_KEYWORDS, 0},
  {"shap_values_acv_leaves", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_25shap_values_acv_leaves, METH_VARARGS|METH_KEYWORDS, 0},
  {"swing_sv_clf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_27swing_sv_clf, METH_VARARGS|METH_KEYWORDS, 0},
  {"swing_sv_clf_direct", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_29swing_sv_clf_direct, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_reg_cpp_pa_coal", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_31global_sdp_reg_cpp_pa_coal, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_clf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_33global_sdp_clf, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_clf_ptrees", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_35global_sdp_clf_ptrees, METH_VARARGS|METH_KEYWORDS, 0},
  {"shap_values_acv_leaves_adap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_37shap_values_acv_leaves_adap, METH_VARARGS|METH_KEYWORDS, 0},
  {"single_shap_values_acv_leaves", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_39single_shap_values_acv_leaves, METH_VARARGS|METH_KEYWORDS, 0},
  {"single_shap_values", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_41single_shap_values, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_clf_approx", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_43global_sdp_clf_approx, METH_VARARGS|METH_KEYWORDS, 0},
  {"binomialC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_45binomialC, METH_VARARGS|METH_KEYWORDS, 0},
  {"_comb_int_long", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_47_comb_int_long, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9cyext_acv_46_comb_int_long},
  {"compute_sdp_rf_v0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_49compute_sdp_rf_v0, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_rf_v0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_51global_sdp_rf_v0, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_exp_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_53compute_exp_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_quantile_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_55compute_quantile_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_quantile_diff_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_57compute_quantile_diff_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_cdf_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_59compute_cdf_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"sufficient_expl_rf_v0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_61sufficient_expl_rf_v0, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_rule_v0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_63compute_sdp_rule_v0, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule_v0", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_65compute_sdp_maxrule_v0, METH_VARARGS|METH_KEYWORDS, 0},
  {"compile_rules", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_67compile_rules, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_rule_biased", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_69compute_sdp_rule_biased, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule_biased", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_71compute_sdp_maxrule_biased, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule_opti", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_73compute_sdp_maxrule_opti, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule_biased_v2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_75compute_sdp_maxrule_biased_v2, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_rule", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_77compute_sdp_rule, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_79compute_sdp_maxrule, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_maxrule_verbose", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_81compute_sdp_maxrule_verbose, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_83compute_sdp_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"compute_sdp_rf_same_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_85compute_sdp_rf_same_set, METH_VARARGS|METH_KEYWORDS, 0},
  {"global_sdp_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_87global_sdp_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {"sufficient_expl_rf", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9cyext_acv_89sufficient_expl_rf, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cyext_acv(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cyext_acv},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cyext_acv",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Check_condition_in_the_computati, __pyx_k_Check_condition_in_the_computati, sizeof(__pyx_k_Check_condition_in_the_computati), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_s_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 0, 1, 1},
  {&__pyx_n_s_N_star, __pyx_k_N_star, sizeof(__pyx_k_N_star), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_n_s_S_star, __pyx_k_S_star, sizeof(__pyx_k_S_star), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_n_s_X, __pyx_k_X, sizeof(__pyx_k_X), 0, 0, 1, 1},
  {&__pyx_kp_s_acv_explainers_cyext_acv_cyext_a, __pyx_k_acv_explainers_cyext_acv_cyext_a, sizeof(__pyx_k_acv_explainers_cyext_acv_cyext_a), 0, 0, 1, 0},
  {&__pyx_n_s_acv_explainers_utils, __pyx_k_acv_explainers_utils, sizeof(__pyx_k_acv_explainers_utils), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_axis, __pyx_k_axis, sizeof(__pyx_k_axis), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_chain, __pyx_k_chain, sizeof(__pyx_k_chain), 0, 0, 1, 1},
  {&__pyx_n_s_children_left, __pyx_k_children_left, sizeof(__pyx_k_children_left), 0, 0, 1, 1},
  {&__pyx_n_s_children_right, __pyx_k_children_right, sizeof(__pyx_k_children_right), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_classifier, __pyx_k_classifier, sizeof(__pyx_k_classifier), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_comb, __pyx_k_comb, sizeof(__pyx_k_comb), 0, 0, 1, 1},
  {&__pyx_n_s_combinations, __pyx_k_combinations, sizeof(__pyx_k_combinations), 0, 0, 1, 1},
  {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_cyext_acv, __pyx_k_cyext_acv, sizeof(__pyx_k_cyext_acv), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_expand_dims, __pyx_k_expand_dims, sizeof(__pyx_k_expand_dims), 0, 0, 1, 1},
  {&__pyx_n_s_fX, __pyx_k_fX, sizeof(__pyx_k_fX), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_from_iterable, __pyx_k_from_iterable, sizeof(__pyx_k_from_iterable), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_iterable, __pyx_k_iterable, sizeof(__pyx_k_iterable), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_leaf_idx_trees, __pyx_k_leaf_idx_trees, sizeof(__pyx_k_leaf_idx_trees), 0, 0, 1, 1},
  {&__pyx_n_s_leaves_nb, __pyx_k_leaves_nb, sizeof(__pyx_k_leaves_nb), 0, 0, 1, 1},
  {&__pyx_n_s_long, __pyx_k_long, sizeof(__pyx_k_long), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_depth, __pyx_k_max_depth, sizeof(__pyx_k_max_depth), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_min_node_size, __pyx_k_min_node_size, sizeof(__pyx_k_min_node_size), 0, 0, 1, 1},
  {&__pyx_n_s_minimal, __pyx_k_minimal, sizeof(__pyx_k_minimal), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_node_idx_trees, __pyx_k_node_idx_trees, sizeof(__pyx_k_node_idx_trees), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_num_threads, __pyx_k_num_threads, sizeof(__pyx_k_num_threads), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_partition_leaves_trees, __pyx_k_partition_leaves_trees, sizeof(__pyx_k_partition_leaves_trees), 0, 0, 1, 1},
  {&__pyx_n_s_pi, __pyx_k_pi, sizeof(__pyx_k_pi), 0, 0, 1, 1},
  {&__pyx_n_s_pi_level, __pyx_k_pi_level, sizeof(__pyx_k_pi_level), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_powerset, __pyx_k_powerset, sizeof(__pyx_k_powerset), 0, 0, 1, 1},
  {&__pyx_n_s_powerset_locals_genexpr, __pyx_k_powerset_locals_genexpr, sizeof(__pyx_k_powerset_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_prod, __pyx_k_prod, sizeof(__pyx_k_prod), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_quantile, __pyx_k_quantile, sizeof(__pyx_k_quantile), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_rule_data, __pyx_k_rule_data, sizeof(__pyx_k_rule_data), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_scaling, __pyx_k_scaling, sizeof(__pyx_k_scaling), 0, 0, 1, 1},
  {&__pyx_n_s_scipy_special, __pyx_k_scipy_special, sizeof(__pyx_k_scipy_special), 0, 0, 1, 1},
  {&__pyx_n_s_search_space, __pyx_k_search_space, sizeof(__pyx_k_search_space), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_tX, __pyx_k_tX, sizeof(__pyx_k_tX), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_thresholds, __pyx_k_thresholds, sizeof(__pyx_k_thresholds), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tqdm, __pyx_k_tqdm, sizeof(__pyx_k_tqdm), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_weighted_percentile, __pyx_k_weighted_percentile, sizeof(__pyx_k_weighted_percentile), 0, 0, 1, 1},
  {&__pyx_n_s_y_X, __pyx_k_y_X, sizeof(__pyx_k_y_X), 0, 0, 1, 1},
  {&__pyx_n_s_y_data, __pyx_k_y_data, sizeof(__pyx_k_y_data), 0, 0, 1, 1},
  {&__pyx_n_s_y_pred, __pyx_k_y_pred, sizeof(__pyx_k_y_pred), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 2501, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 148, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(2, 151, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(2, 2, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(2, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 613, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(2, 832, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cyext_acv.pyx":2501
 *                 sdp[i] = -1
 *     else:
 *         raise ValueError('Check condition in the computation of SDP')             # <<<<<<<<<<<<<<
 * 
 *     return sdp
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Check_condition_in_the_computati); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 2501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "cyext_acv.pyx":4767
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], 2))
 *         buffer[:, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :] partition_byobs
 */
  __pyx_slice__2 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__2)) __PYX_ERR(0, 4767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__2);
  __Pyx_GIVEREF(__pyx_slice__2);
  __pyx_tuple__3 = PyTuple_Pack(3, __pyx_slice__2, __pyx_slice__2, __pyx_int_0); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 4767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "cyext_acv.pyx":4798
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], 2))
 *         buffer_data[:, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :] partition_byobs_data
 */
  __pyx_tuple__4 = PyTuple_Pack(4, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_int_0); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 4798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "cyext_acv.pyx":4900
 * 
 *         buffer = 1e+10 * np.ones(shape=(X.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer[:, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :] partition_byobs
 */
  __pyx_tuple__5 = PyTuple_Pack(5, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_int_0); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 4900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "cyext_acv.pyx":4937
 * 
 *         buffer_data = 1e+10 * np.ones(shape=(X.shape[0], data.shape[0], X.shape[1], children_left.shape[0], max_depth , 2))
 *         buffer_data[:, :, :, :, :, 0] = -1e+10             # <<<<<<<<<<<<<<
 * 
 *         cdef double [:, :, :, :, :, :] partition_byobs_data
 */
  __pyx_tuple__6 = PyTuple_Pack(6, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_slice__2, __pyx_int_0); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 4937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "../../../../tmp/build-env-vgg82hwn/lib/python3.8/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(2, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(2, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(2, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(2, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(2, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(2, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__20 = PyTuple_New(1); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(2, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__20, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(2, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */
  __pyx_tuple__26 = PyTuple_Pack(4, __pyx_n_s_iterable, __pyx_n_s_s, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acv_explainers_cyext_acv_cyext_a, __pyx_n_s_powerset, 455, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 455, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__33 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  __pyx_umethod_PyList_Type_copy.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1e_10 = PyFloat_FromDouble(1e+10); if (unlikely(!__pyx_float_1e_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_1e_10 = PyFloat_FromDouble(-1e+10); if (unlikely(!__pyx_float_neg_1e_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_9cyext_acv___pyx_scope_struct__powerset) < 0) __PYX_ERR(0, 455, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9cyext_acv___pyx_scope_struct__powerset.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9cyext_acv___pyx_scope_struct__powerset.tp_dictoffset && __pyx_type_9cyext_acv___pyx_scope_struct__powerset.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9cyext_acv___pyx_scope_struct__powerset.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9cyext_acv___pyx_scope_struct__powerset = &__pyx_type_9cyext_acv___pyx_scope_struct__powerset;
  if (PyType_Ready(&__pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9cyext_acv___pyx_scope_struct_1_genexpr = &__pyx_type_9cyext_acv___pyx_scope_struct_1_genexpr;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcyext_acv(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcyext_acv(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cyext_acv(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cyext_acv(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cyext_acv(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  static PyThread_type_lock __pyx_t_3[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cyext_acv' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cyext_acv(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cyext_acv", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cyext_acv) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cyext_acv")) {
      if (unlikely(PyDict_SetItemString(modules, "cyext_acv", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cyext_acv.pyx":5
 * from libcpp.vector cimport vector
 * from libcpp.set cimport set
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * ctypedef np.float64_t double
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":9
 * ctypedef np.float64_t double
 * cimport cython
 * from scipy.special import comb             # <<<<<<<<<<<<<<
 * import itertools
 * from tqdm import tqdm
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_comb);
  __Pyx_GIVEREF(__pyx_n_s_comb);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_comb);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scipy_special, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_comb); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_comb, __pyx_t_1) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyext_acv.pyx":10
 * cimport cython
 * from scipy.special import comb
 * import itertools             # <<<<<<<<<<<<<<
 * from tqdm import tqdm
 * from acv_explainers.utils import weighted_percentile
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_itertools, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyext_acv.pyx":11
 * from scipy.special import comb
 * import itertools
 * from tqdm import tqdm             # <<<<<<<<<<<<<<
 * from acv_explainers.utils import weighted_percentile
 * from cython.parallel cimport prange, parallel, threadid
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_tqdm);
  __Pyx_GIVEREF(__pyx_n_s_tqdm);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_tqdm);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_tqdm, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_tqdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tqdm, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cyext_acv.pyx":12
 * import itertools
 * from tqdm import tqdm
 * from acv_explainers.utils import weighted_percentile             # <<<<<<<<<<<<<<
 * from cython.parallel cimport prange, parallel, threadid
 * from cython.operator cimport dereference as deref, preincrement as inc
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_weighted_percentile);
  __Pyx_GIVEREF(__pyx_n_s_weighted_percentile);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_weighted_percentile);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_acv_explainers_utils, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_weighted_percentile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_weighted_percentile, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyext_acv.pyx":455
 *     return np.array(sdp)
 * 
 * def powerset(iterable):             # <<<<<<<<<<<<<<
 *     s = list(iterable)
 *     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9cyext_acv_11powerset, NULL, __pyx_n_s_cyext_acv); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_powerset, __pyx_t_2) < 0) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cyext_acv.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * 
 * from libcpp.vector cimport vector
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_3[0] = PyThread_allocate_lock();
  __pyx_t_3[1] = PyThread_allocate_lock();
  __pyx_t_3[2] = PyThread_allocate_lock();
  __pyx_t_3[3] = PyThread_allocate_lock();
  __pyx_t_3[4] = PyThread_allocate_lock();
  __pyx_t_3[5] = PyThread_allocate_lock();
  __pyx_t_3[6] = PyThread_allocate_lock();
  __pyx_t_3[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_3, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_2) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cyext_acv", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cyext_acv");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* None */
static void __Pyx_RaiseUnboundMemoryviewSliceNogil(const char *varname) {
    #ifdef WITH_THREAD
    PyGILState_STATE gilstate = PyGILState_Ensure();
    #endif
    __Pyx_RaiseUnboundLocalError(varname);
    #ifdef WITH_THREAD
    PyGILState_Release(gilstate);
    #endif
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* None */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_long(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_long, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 3,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 4,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_long(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_long, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_double__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 3,
                                                 &__Pyx_TypeInfo_double__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsds_double__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 4,
                                                 &__Pyx_TypeInfo_double__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_long__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_long__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_long__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_long__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_long__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_long__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_double__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_int(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_int, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(double *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj) {
    double value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(double *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 5,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* None */
  static CYTHON_INLINE long __Pyx_pow_long(long b, long e) {
    long t = b;
    switch (e) {
        case 3:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 2:
            t *= b;
        CYTHON_FALLTHROUGH;
        case 1:
            return t;
        case 0:
            return 1;
    }
    #if 1
    if (unlikely(e<0)) return 0;
    #endif
    t = 1;
    while (likely(e)) {
        t *= (b * (e&1)) | ((~e)&1);
        b *= b;
        e >>= 1;
    }
    return t;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_long(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_long(*(long *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_long(const char *itemp, PyObject *obj) {
    long value = __Pyx_PyInt_As_long(obj);
    if ((value == (long)-1) && PyErr_Occurred())
        return 0;
    *(long *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_d_dc_long(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 4,
                                                 &__Pyx_TypeInfo_long, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_long__const__(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_long(*(long const  *) itemp);
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_long(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_long, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsds_long(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 3,
                                                 &__Pyx_TypeInfo_long, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsdsdsdsdsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 6,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned long) -1;
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned PY_LONG_LONG) -1;
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CoroutineBase */
  #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
  #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
